/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#ifndef __lj_hal_kernel_h_
#define __lj_hal_kernel_h_

#include "lj_hal.h"
#include "lj_hal_kernel_hardware.h"
#include "lj_hal_kernel_hardware_fe.h"
#include "lj_hal_driver.h"
#include "lj_hal_kernel_mutex.h"
#include "lj_hal_metadata.h"
#include "lj_hal_kernel_mmu.h"

#if ljdENABLE_SW_PREEMPTION
#include "lj_hal_kernel_preemption.h"
#endif

#if ljdENABLE_VG
#include "lj_hal_kernel_vg.h"
#endif

#if ljdSECURITY || ljdENABLE_TRUST_APPLICATION
#include "lj_hal_security_interface.h"
#endif

#ifdef __cplusplus
extern "C" {
#endif

#define ljdRECOVERY_FORCE_TIMEOUT 100

#ifndef ljdENABLE_VIRTUAL_POOL
#define ljdENABLE_VIRTUAL_POOL 1
#endif

#ifndef ljdDYNAMIC_ALLOC_LOCAL_MEMORY
#define ljdDYNAMIC_ALLOC_LOCAL_MEMORY 0
#endif

#ifndef ljdSUSPEND_RESUME_FROM_DISK
#define ljdSUSPEND_RESUME_FROM_DISK   0
#endif

#if ljdENABLE_CLEAR_FENCE
#include <linux/spinlock.h>
#endif
/*******************************************************************************
 ***** Stuck Dump Level ********************************************************/

/*
 * Stuck Dump Level
 *
 * Level  1 ~  5 : NORMAL model, when recovery is enabled, stuck dump is invalid
 * Level 11 ~ 15 : FORCE model, whether recovery is enabled or not, driver will
 *                 dump as level set
 *
 * NONE : Dump nothing when stuck happens.
 *
 * NEARBY_MEMORY : Dump GPU state and memory near stuck point.
 * USER_COMMAND  : Beside NEARBY_MEMORY, dump context buffer and user command
 *                 buffer.
 * STALL_COMMAND : Beside USER_COMMAND, commit will be stall to make sure command
 *                 causing stuck isn't missed.
 * ALL_COMMAND   : Beside USER_COMMAND, dump kernel command buffer.
 * ALL_CORE      : Dump all the cores with ALL_COMMAND level.
 */
typedef enum _ljeSTUCK_DUMP_LEVEL {
    ljvSTUCK_DUMP_NONE = 0,

    ljvSTUCK_DUMP_NEARBY_MEMORY = 1,
    ljvSTUCK_DUMP_USER_COMMAND,
    ljvSTUCK_DUMP_STALL_COMMAND,
    ljvSTUCK_DUMP_ALL_COMMAND,
    ljvSTUCK_DUMP_ALL_CORE,

    ljvSTUCK_FORCE_DUMP_NEARBY_MEMORY = 11,
    ljvSTUCK_FORCE_DUMP_USER_COMMAND,
    ljvSTUCK_FORCE_DUMP_STALL_COMMAND,
    ljvSTUCK_FORCE_DUMP_ALL_COMMAND,
    ljvSTUCK_FORCE_DUMP_ALL_CORE,
} ljeSTUCK_DUMP_LEVEL;

/******************************************************************************
 ***** Page table *************************************************************/

#define ljvPAGE_TABLE_DIRTY_BIT_OTHER (1 << 0)
#define ljvPAGE_TABLE_DIRTY_BIT_FE    (1 << 1)

/******************************************************************************
 ***** GPU Virtualization ****************************************************/

typedef enum _ljeVGPU_TYPE {
    ljvVGPU_NONE = 0,
    ljvVGPU_MDEV,
    ljvVGPU_SRIOV,
} ljeVGPU_TYPE;

/******************************************************************************
 ***** Process Database Management ********************************************/

typedef enum _ljeDATABASE_TYPE {
    ljvDB_VIDEO_MEMORY = 1,    /* Video memory created. */
    ljvDB_COMMAND_BUFFER,      /* Command Buffer. */
    ljvDB_NON_PAGED,           /* Non paged memory. */
    ljvDB_CONTIGUOUS,          /* Contiguous memory. */
    ljvDB_SIGNAL,              /* Signal. */
    ljvDB_VIDEO_MEMORY_LOCKED, /* Video memory locked. */
    ljvDB_CONTEXT,             /* Context */
    ljvDB_IDLE,                /* GPU idle. */
    ljvDB_MAP_MEMORY,          /* Map memory */
    ljvDB_SHBUF,               /* Shared buffer. */
#if ljdENABLE_SW_PREEMPTION
    ljvDB_PRIORITY,
#endif
#if ljdENABLE_CLEAR_FENCE
    ljvDB_USER_FENCE,
#endif

    ljvDB_NUM_TYPES,
} ljeDATABASE_TYPE;

#define ljdDATABASE_TYPE_MASK           0x000000FF
#define ljdDB_VIDEO_MEMORY_TYPE_MASK    0x0000FF00
#define ljdDB_VIDEO_MEMORY_TYPE_SHIFT   8

#define ljdDB_VIDEO_MEMORY_POOL_MASK    0x00FF0000
#define ljdDB_VIDEO_MEMORY_POOL_SHIFT   16

#define ljdDB_VIDEO_MEMORY_DBTYPE_MASK  0xFF000000
#define ljdDB_VIDEO_MEMORY_DBTYPE_SHIFT 24

typedef struct _ljsDATABASE_RECORD *ljsDATABASE_RECORD_PTR;
typedef struct _ljsDATABASE_RECORD {
    /* Pointer to kernel. */
    ljkKERNEL                           kernel;

    /* Pointer to next database record. */
    ljsDATABASE_RECORD_PTR              next;

    /* Type of record. */
    ljeDATABASE_TYPE                    type;

    /* Data for record. */
    ljtPOINTER                          data;
    ljtPHYS_ADDR                        physical;
    ljtSIZE_T                           bytes;
} ljsDATABASE_RECORD;

#if ljdENABLE_PERF_DISPATCH
typedef struct _ljsDISPATCH_PERF_RECORD {
    ljtUINT64 count;
    ljtUINT64 failed;

    ljtUINT64 cost;
    ljtUINT64 maximum;

    ljtPOINTER mutex;
} ljsDISPATCH_PERF_RECORD;
#endif

typedef struct _ljsDATABASE            *ljsDATABASE_PTR;
typedef struct _ljsDATABASE {
    /* Pointer to next entry is hash list. */
    ljsDATABASE_PTR                     next;
    ljtSIZE_T                           slot;

    /* Process ID. */
    ljtUINT32                           processID;

    /* Open-Close ref count */
    ljtPOINTER                          refs;

    /* Already mark for delete and cannot reenter */
    ljtBOOL                             deleted;

    /* Sizes to query. */
    ljsDATABASE_COUNTERS                vidMem;
    ljsDATABASE_COUNTERS                nonPaged;
    ljsDATABASE_COUNTERS                contiguous;
    ljsDATABASE_COUNTERS                mapMemory;

    ljsDATABASE_COUNTERS                vidMemType[ljvVIDMEM_TYPE_COUNT];
    /* Counter for each video memory pool. */
    ljsDATABASE_COUNTERS                vidMemPool[ljvPOOL_NUMBER_OF_POOLS];
    ljtPOINTER                          counterMutex;

    /* Idle time management. */
    ljtUINT64                           lastIdle;
    ljtUINT64                           idle;

    /* Pointer to database. */
    ljsDATABASE_RECORD_PTR              list[48];

    ljtPOINTER                          handleDatabase;
    ljtPOINTER                          handleDatabaseMutex;

    /* Per process mmu. */
    ljkMMU                              mmu;

#if ljdENABLE_PERF_DISPATCH
    ljsDISPATCH_PERF_RECORD     dispatchPerfRecords[ljvHAL_NUM_COMMAND_CODES];
#endif
} ljsDATABASE;

typedef struct _ljsFDPRIVATE *ljsFDPRIVATE_PTR;
typedef struct _ljsFDPRIVATE {
    ljtINT (*release)(ljsFDPRIVATE_PTR Private);
} ljsFDPRIVATE;

typedef struct _ljsRECORDER *ljkRECORDER;

typedef enum _ljeEVENT_FAULT {
    ljvEVENT_NO_FAULT,
    ljvEVENT_BUS_ERROR_FAULT,
} ljeEVENT_FAULT;

/* Create a process database that will contain all its allocations. */
ljeSTATUS
ljkKERNEL_CreateProcessDB(ljkKERNEL Kernel, ljtUINT32 ProcessID);

/* Add a record to the process database. */
ljeSTATUS
ljkKERNEL_AddProcessDB(ljkKERNEL Kernel,
                       ljtUINT32 ProcessID,
                       ljeDATABASE_TYPE Type,
                       ljtPOINTER Pointer,
                       ljtPHYS_ADDR Physical,
                       ljtSIZE_T Size);

/* Remove a record to the process database. */
ljeSTATUS
ljkKERNEL_RemoveProcessDB(ljkKERNEL Kernel,
                          ljtUINT32 ProcessID,
                          ljeDATABASE_TYPE Type,
                          ljtPOINTER Pointer);

ljeSTATUS
ljkKERNEL_RemoveDatabaseFromList(ljkKERNEL Kernel, ljsDATABASE_PTR Database, ljtUINT32 ProcessID);

/* Destroy the process database. */
ljeSTATUS
ljkKERNEL_DestroyProcessDB(ljkKERNEL Kernel, ljtUINT32 ProcessID);

/* Find a record to the process database. */
ljeSTATUS
ljkKERNEL_FindProcessDB(ljkKERNEL Kernel,
                        ljtUINT32 ProcessID,
                        ljtUINT32 ThreadID,
                        ljeDATABASE_TYPE Type,
                        ljtPOINTER Pointer,
                        ljsDATABASE_RECORD_PTR Record);

/* Query the process database. */
ljeSTATUS
ljkKERNEL_QueryProcessDB(ljkKERNEL Kernel,
                         ljtUINT32 ProcessID,
                         ljtBOOL LastProcessID,
                         ljeDATABASE_TYPE Type,
                         ljuDATABASE_INFO *Info);

/* Dump the process database. */
ljeSTATUS
ljkKERNEL_DumpProcessDB(ljkKERNEL Kernel);

/* Dump the video memory usage for process specified. */
ljeSTATUS
ljkKERNEL_DumpVidMemUsage(ljkKERNEL Kernel, ljtINT32 ProcessID);

ljeSTATUS
ljkKERNEL_DeleteRecord(ljkKERNEL Kernel, ljsDATABASE_PTR Database,
                       ljeDATABASE_TYPE Type, ljtPOINTER Data,
                       ljtSIZE_T_PTR Bytes OPTIONAL);

ljeSTATUS
ljkKERNEL_FindDatabase(ljkKERNEL Kernel,
                       ljtUINT32 ProcessID,
                       ljtBOOL LastProcessID,
                       ljsDATABASE_PTR *Database);

ljeSTATUS
ljkKERNEL_FindHandleDatabase(ljkKERNEL Kernel,
                             ljtUINT32 ProcessID,
                             ljtPOINTER *HandleDatabase,
                             ljtPOINTER *HandleDatabaseMutex);

ljeSTATUS
ljkKERNEL_CreateIntegerDatabase(ljkKERNEL Kernel,
                                ljtUINT32 Capacity,
                                ljtPOINTER *Database);

ljeSTATUS
ljkKERNEL_DestroyIntegerDatabase(ljkKERNEL Kernel, ljtPOINTER Database);

ljeSTATUS
ljkKERNEL_AllocateIntegerId(ljtPOINTER Database, ljtPOINTER Pointer, ljtUINT32 *Id);

ljeSTATUS
ljkKERNEL_FreeIntegerId(ljtPOINTER Database, ljtUINT32 Id);

ljeSTATUS
ljkKERNEL_QueryIntegerId(ljtPOINTER Database, ljtUINT32 Id, ljtPOINTER *Pointer);

/* Pointer rename  */
ljtUINT32
ljkKERNEL_AllocateNameFromPointer(ljkKERNEL Kernel, ljtPOINTER Pointer);

ljtPOINTER
ljkKERNEL_QueryPointerFromName(ljkKERNEL Kernel, ljtUINT32 Name);

ljeSTATUS
ljkKERNEL_DeleteName(ljkKERNEL Kernel, ljtUINT32 Name);

/*******************************************************************************
 ********* Timer Management ****************************************************/
typedef struct _ljsTIMER *ljsTIMER_PTR;
typedef struct _ljsTIMER {
    /* Start and Stop time holders. */
    ljtUINT64       startTime;
    ljtUINT64       stopTime;
}
ljsTIMER;

/******************************************************************************
 ********************************** Structures ********************************
 ******************************************************************************/

/* ljkDB object. */
struct _ljkDB {
    /* Database management. */
    ljsDATABASE_PTR             db[16];
    ljtPOINTER                  dbMutex;
    ljsDATABASE_PTR             freeDatabase;
    ljsDATABASE_RECORD_PTR      freeRecord;
    ljsDATABASE_PTR             lastDatabase;
    ljtUINT32                   lastProcessID;
    ljtUINT64                   lastIdle;
    ljtUINT64                   idleTime;
    ljtUINT64                   lastSlowdown;
    ljtUINT64                   lastSlowdownIdle;

    ljtPOINTER                  nameDatabase;
    ljtPOINTER                  nameDatabaseMutex;

    ljtPOINTER                  pointerDatabase;

    ljsLISTHEAD                 videoMemList;
    ljtPOINTER                  videoMemListMutex;

    ljtPOINTER                  refcnt;
};

#if ljdENABLE_TTM
typedef struct _ljsBUFSYNC {
    ljkVIDMEM_NODE nodeObj;
    ljtSIZE_T size;
    struct _ljsBUFSYNC *next;
} ljsBUFSYNC;

typedef struct _ljsBUFSYNC *ljtBUFSYNC;
#endif

/* ljkKERNEL object. */
struct _ljkKERNEL {
    /* Object. */
    ljsOBJECT                   object;

    /* Pointer to ljkOS object. */
    ljkOS                       os;

    /* Pointer to ljkDEVICE object. */
    ljkDEVICE                   device;

    /* Pointer to ljkHARDWARE object. */
    ljkHARDWARE                 hardware;

    /* Hardware type. */
    ljeHARDWARE_TYPE            type;

    /* Core */
    ljeCORE                     core;
    ljtUINT                     chipID;

    /* Brothers */
    ljtPOINTER                  atomBroCoreMask;

    /* Main command module, event and context. */
    ljkCOMMAND                  command;
    ljkCOMMAND                  sharedCommand;
    ljkEVENT                    eventObj;
    ljtPOINTER                  context;

    /* Async FE command and event modules. */
    ljkCOMMAND                  asyncCommand;
    ljkEVENT                    asyncEvent;

    /* Parity event. */
    ljkEVENT                    parityEvent;

    /* Pointer to ljkMMU object. */
    ljkMMU                      mmu;
    ljkMMU                      mmuCopy;

    /* If this core use shared MMU page table. */
    ljtBOOL                     sharedPageTable;

    /* If this page table is per-process. */
    ljtBOOL                     processPageTable;

    /* If use SW copy for per-process page table. */
    ljtBOOL                     switchMmuByCopy;

    /* If this page table enables flat mapping. */
    ljtBOOL                     flatMapping;

    /* Arom holding number of clients. */
    ljtPOINTER                  atomClients;

#if LJMICRO_PROFILER
    ljkPROFILER                 profiler;
#endif

#ifdef QNX_SINGLE_THREADED_DEBUGGING
    ljtPOINTER                  debugMutex;
#endif

    /* Database management. */
    ljkDB                       db;
    ljtBOOL                     dbCreated;

    ljtUINT64                   resetTimeStamp;

    /* Pointer to ljkEVENT object. */
    ljsTIMER                    timers[8];
    ljtUINT32                   timeOut;

#if ljdENABLE_VG
    ljkVGKERNEL                 vg;
#endif

#if ljdDVFS
    ljkDVFS                     dvfs;
#endif

#if ljdLINUX_SYNC_FILE
    ljtHANDLE                   timeline;
#endif

    /* Enable recovery. */
    ljtBOOL                     recovery;

    /* Level of dump information after stuck. */
    ljtUINT                     stuckDump;

#if ljdSECURITY || ljdENABLE_TRUST_APPLICATION
    ljtUINT32                   securityChannel;
#endif

    /* Timer to monitor GPU stuck. */
    ljtPOINTER                  monitorTimer;

    /* Flag to quit monitor timer. */
    ljtBOOL                     monitorTimerStop;

    /* Monitor states. */
    ljtBOOL                     monitoring;
    ljtUINT32                   lastCommitStamp;
    ljtUINT32                   timer;
    ljtADDRESS                  restoreAddress;
    ljtINT32                    restoreMask;

    ljkVIDMEM_BLOCK             vidMemBlock;
    ljtPOINTER                  vidMemBlockMutex;

#if ljdENABLE_TTM
    ljtBUFSYNC                  bufSyncList;
    ljtPOINTER                  bufSyncListMutex;
#endif

    ljtADDRESS                  contiguousBaseAddresses[ljdSYSTEM_RESERVE_COUNT];
    ljtADDRESS                  lowContiguousBaseAddress;
    ljtADDRESS                  externalBaseAddress;
    ljtADDRESS                  internalBaseAddress;
    ljtADDRESS                  exclusiveBaseAddress;

    /* External shared SRAM. */
    ljtADDRESS                  extSRAMBaseAddresses[ljvSRAM_EXT_COUNT];

    /* Per core SRAM description. */
    ljtUINT32                   sRAMIndex;
    ljkVIDMEM                   sRAMVidMem[ljvSRAM_INTER_COUNT];
    ljtPHYS_ADDR                sRAMPhysical[ljvSRAM_INTER_COUNT];
    ljtADDRESS                  sRAMBaseAddresses[ljvSRAM_INTER_COUNT];
    ljtUINT32                   sRAMSizes[ljvSRAM_INTER_COUNT];
    ljtBOOL                     sRAMPhysFaked[ljvSRAM_INTER_COUNT];
    ljtUINT64                   sRAMLoopMode;

    ljtUINT32                   timeoutPID;
    ljtBOOL                     threadInitialized;
    ljtBOOL                     killThread;
    ljtPOINTER                  resetStatus;
    ljtSEMAPHORE                sema;

    /* Device ID. */
    ljtUINT32                   devID;

    /* In multi-core context or not. */
    ljtBOOL                     inShared;

    ljtUINT32                   totalMmuDescNum;
    ljtUINT32                   nextMmuDescId;
    ljtUINT32                   *mmuDescMap;
    ljtPOINTER                  mmuDescMutex;

    /* vGPU type and id. */
    ljeVGPU_TYPE            vGPUType;
    ljtUINT32               vGPUId;

#if ljdENABLE_SW_PREEMPTION
    ljtPOINTER                  priorityQueueMutex[ljdMAX_PRIORITY_QUEUE_NUM];
    ljsPRIORITY_QUEUE_PTR       priorityQueues[ljdMAX_PRIORITY_QUEUE_NUM];
    ljtBOOL                     priorityDBCreated[ljdMAX_PRIORITY_QUEUE_NUM];
    ljtSEMAPHORE                preemptSema;
    ljePREEMPTION_MODE          preemptionMode;
    ljtBOOL                     killPreemptThread;
#endif
};

struct _FrequencyHistory {
    ljtUINT32 frequency;
    ljtUINT32 count;
};

/* ljkDVFS object. */
struct _ljkDVFS {
    ljkOS                    os;
    ljkHARDWARE              hardware;
    ljtPOINTER               timer;
    ljtUINT32                pollingTime;
    ljtBOOL                  stop;
    ljtUINT32                totalConfig;
    ljtUINT32                loads[8];
    ljtUINT8                 currentScale;
    struct _FrequencyHistory frequencyHistory[16];
};

typedef struct _ljsFENCE *ljkFENCE;
typedef struct _ljsFENCE {
    /* Pointer to required object. */
    ljkKERNEL                   kernel;

    /* Fence location. */
    ljkVIDMEM_NODE              videoMem;
    ljtPOINTER                  logical;
    ljtADDRESS                  address;

    ljsLISTHEAD                 waitingList;
    ljtPOINTER                  mutex;
} ljsFENCE;

/* A sync point attached to fence. */
typedef struct _ljsFENCE_SYNC *ljkFENCE_SYNC;
typedef struct _ljsFENCE_SYNC {
    /* Stamp of commit access this node. */
    ljtUINT64                   commitStamp;

    /* Attach to waiting list. */
    ljsLISTHEAD                 head;

    ljtPOINTER                  signal;

    ljtBOOL                     inList;
} ljsFENCE_SYNC;

typedef struct _ljsCOMMAND_QUEUE {
    ljtSIGNAL      signal;
    ljkVIDMEM_NODE videoMem;
    ljtPOINTER     logical;
    ljtADDRESS     address;
    ljePOOL        pool;
} ljsCOMMAND_QUEUE;

/* ljkCOMMAND object. */
struct _ljkCOMMAND {
    /* Object. */
    ljsOBJECT                   object;

    /* Pointer to required object. */
    ljkKERNEL                   kernel;
    ljkOS                       os;

    ljeHW_FE_TYPE               feType;

    /* Number of bytes per page. */
    ljtUINT32                   pageSize;

    /* Current pipe select. */
    ljePIPE_SELECT              pipeSelect;

    /* Command queue running flag. */
    ljtBOOL                     running;

    /* Idle flag and commit stamp. */
    ljtBOOL                     idle;
    ljtUINT64                   commitStamp;

    /* Command queue mutex. */
    ljtPOINTER                  mutexQueue;

    /* Context switching mutex. */
    ljtPOINTER                  mutexContext;

    /* Context sequence mutex. */
    ljtPOINTER                  mutexContextSeq;

    /* Command queue power semaphore. */
    ljtPOINTER                  powerSemaphore;

    /* Command queues. */
    ljsCOMMAND_QUEUE            queues[ljdCOMMAND_QUEUES];

    /* Current queue. */
    ljkVIDMEM_NODE              videoMem;
    ljtPOINTER                  logical;
    ljtADDRESS                  address;
    ljePOOL                     pool;

    ljtUINT32                   offset;
    ljtINT                      index;
#if ljmIS_DEBUG(ljdDEBUG_TRACE)
    ljtUINT                     wrapCount;
#endif

    /* The command queue is new. */
    ljtBOOL                     newQueue;

    /* Context management. */
    ljkCONTEXT                  currContext;
    ljtPOINTER                  stateMap;

    /* Pointer to last WAIT command. */
    /* Wait-Link FE only. */
    struct {
        ljkVIDMEM_NODE          videoMem;
        ljtUINT32               offset;
        ljtPOINTER              logical;
        ljtADDRESS              address;
        ljtUINT32               size;
    }
    waitPos;

    /* MCFE. */
    ljtUINT32                   totalSemaId;

    /*
     * freeSemaId   ------>  [pendingSema]
     * ^                    freePos -->+ +
     * |                    ^          | |
     * |                    |          | |
     * |                    nextPos ---+ |
     * |                                 |
     * |                                 |
     * |                                 v
     * nextSemaId <----------------------+
     *
     * Terminology:
     * 'freeSemaId': the top (final) free sema id can be used, signaled already.
     * 'nextSemaId': the next sema id to be used.
     * 'pendingSema': the used semaphores which are tracked in the ring.
     *
     * 3 Id sections:
     * Id(s) between 'nextSemaId' and 'freeSemaId':
     * Available for immediate use.
     *
     * Id(s) between 'freeSemaId' (exclusive) and 'pendingSema':
     * To be signed, will be available to use later.
     *
     * Id(s) between 'pendingSema' and 'nextSemaId':
     * The semaphores just used, not tracking in pendingSema ring.
     *
     * Conditions:
     * 'nextSemaId' = 'freeSemaId':
     * Using the final free semaphore, means the ring is full of used ones.
     *
     * 'freeSemaId' + 1 = 'nextSemaId':
     * Can use 'freeSema' + 1 to 'nextSema'(loop back), means empty ring,
     * ie, no used one.
     */

    /* MCFE semaphore id tracking ring. */
    ljtUINT32                   nextSemaId;
    ljtUINT32                   freeSemaId;

    ljtUINT32                   semaMinThreshhold;

    /* pending semaphore id tracking ring. */
    struct {
        ljtUINT32               semaId;
        ljtSIGNAL               signal;
    } pendingSema[8];

    ljtUINT32                   nextPendingPos;
    ljtUINT32                   freePendingPos;

    /* semaphore id (array index) to handle value map. */
    ljtUINT32                  *semaHandleMap;

    /*
     * Dirty channels, ie channels ever submitted commands.
     * Need sync to (send semaphore to) system channel.
     */
    ljtUINT64                   dirtyChannel[2];

    /*
     * Sync channels, need sync from (wait semaphore from) the system
     * channel before its own jobs.
     */
    ljtUINT64                   syncChannel[2];

    /* Command buffer alignment. */
    ljtUINT32                   alignment;

    /* Commit counter. */
    ljtPOINTER                  atomCommit;

    /* Kernel process ID. */
    ljtUINT32                   kernelProcessID;

    ljtUINT32                   kernelProcessAttached;

#if ljdRECORD_COMMAND
    ljkRECORDER                 recorder;
#endif

    ljkFENCE                    fence;

    ljtBOOL                     dummyDraw;
};

typedef struct _ljsEVENT       *ljsEVENT_PTR;

/* Structure holding one event to be processed. */
typedef struct _ljsEVENT {
    /* Pointer to next event in queue. */
    ljsEVENT_PTR                next;

    /* Event information. */
    ljsHAL_INTERFACE            info;

    /* Process ID owning the event. */
    ljtUINT32                   processID;

#ifdef __QNXNTO__
    /* Kernel. */
    ljkKERNEL                   kernel;
#endif

    ljtBOOL                     fromKernel;

    ljkMMU                      mmu;
} ljsEVENT;

/* Structure holding a list of events to be processed by an interrupt. */
typedef struct _ljsEVENT_QUEUE *ljsEVENT_QUEUE_PTR;
typedef struct _ljsEVENT_QUEUE {
    /* Time stamp. */
    ljtUINT64                   stamp;

    /* Source of the event. */
    ljeKERNEL_WHERE             source;

    /* Pointer to head of event queue. */
    ljsEVENT_PTR                head;

    /* Pointer to tail of event queue. */
    ljsEVENT_PTR                tail;

    /* Next list of events. */
    ljsEVENT_QUEUE_PTR          next;

    /* Current commit stamp. */
    ljtUINT64                   commitStamp;
} ljsEVENT_QUEUE;

/*
 *   ljdREPO_LIST_COUNT defines the maximum number of event queues with different
 *   hardware module sources that may coexist at the same time. Only two sources
 *   are supported - ljvKERNEL_COMMAND and ljvKERNEL_PIXEL. ljvKERNEL_COMMAND
 *   source is used only for managing the kernel command queue and is only issued
 *   when the current command queue gets full. Since we commit event queues every
 *   time we commit command buffers, in the worst case we can have up to three
 *   pending event queues:
 *       - ljvKERNEL_PIXEL
 *       - ljvKERNEL_COMMAND (queue overflow)
 *       - ljvKERNEL_PIXEL
 */
#define ljdREPO_LIST_COUNT      3

#define ljdEVENT_QUEUE_COUNT    29

/* ljkEVENT object. */
struct _ljkEVENT {
    /* The object. */
    ljsOBJECT                   object;

    /* Pointer to required objects. */
    ljkOS                       os;
    ljkKERNEL                   kernel;

    /* Pointer to COMMAND object, either one of the 3. */
    ljkCOMMAND                  command;

    /* Submit function pointer, different for different command module. */
    ljeSTATUS                   (*submitEvent)(ljkEVENT, ljtBOOL, ljtBOOL);

    /* Time stamp. */
    ljtUINT64                   stamp;
    ljtUINT32                   lastCommitStamp;

    /* Queue mutex. */
    ljtPOINTER                  eventQueueMutex;

    /* Array of event queues. */
    ljsEVENT_QUEUE              queues[ljdEVENT_QUEUE_COUNT];
    ljtINT32                    totalQueueCount;
    ljtINT32                    freeQueueCount;
    ljtUINT8                    lastID;

    /* Pending events. */
    ljtPOINTER                  pending;

    /* List of free event structures and its mutex. */
    ljsEVENT_PTR                freeEventList;
    ljtSIZE_T                   freeEventCount;
    ljtPOINTER                  freeEventMutex;

    /* Event queues. */
    ljsEVENT_QUEUE_PTR          queueHead;
    ljsEVENT_QUEUE_PTR          queueTail;
    ljsEVENT_QUEUE_PTR          freeList;
    ljsEVENT_QUEUE              repoList[ljdREPO_LIST_COUNT];
    ljtPOINTER                  eventListMutex;

    ljtPOINTER                  submitTimer;

#if ljdINTERRUPT_STATISTIC
    ljtPOINTER                  interruptCount;
#endif

    ljtINT                      notifyState;
};

/* Event Attribute. */
typedef struct _ljkEVENT_ATTR *ljkEVENT_ATTR;
typedef struct _ljkEVENT_ATTR {
    /* Submit requires one vacant event or not. */
    ljtBOOL wait;
    /* It is in multi-core programming or not. */
    ljtBOOL shared;
    /* It is called by power management or not.*/
    ljtBOOL fromPower;
    /* It is broadcast the new commit arrives or not.*/
    ljtBOOL broadcast;
} ljsEVENT_ATTR;

/* Construct a new ljkEVENT object. */
ljeSTATUS
ljkEVENT_Construct(ljkKERNEL Kernel, ljkCOMMAND Command, ljkEVENT *Event);

/* Destroy an ljkEVENT object. */
ljeSTATUS
ljkEVENT_Destroy(ljkEVENT Event);

/* Reserve the next available hardware event. */
ljeSTATUS
ljkEVENT_GetEvent(ljkEVENT Event,
                  ljtBOOL Wait,
                  ljtUINT8 *EventID,
                  ljeKERNEL_WHERE Source);

/* Add a new event to the list of events. */
ljeSTATUS
ljkEVENT_AddListEx(ljkEVENT Event,
                   ljsHAL_INTERFACE_PTR Interface,
                   ljeKERNEL_WHERE FromWhere,
                   ljtBOOL AllocateAllowed,
                   ljtBOOL FromKernel,
                   ljtUINT32 ProcessID);

/* Add a new event to the list of events. */
ljeSTATUS
ljkEVENT_AddList(ljkEVENT Event,
                 ljsHAL_INTERFACE_PTR Interface,
                 ljeKERNEL_WHERE FromWhere,
                 ljtBOOL AllocateAllowed,
                 ljtBOOL FromKernel);

/* Schedule a FreeVideoMemory event. */
ljeSTATUS
ljkEVENT_FreeVideoMemory(ljkEVENT Event,
                         ljuVIDMEM_NODE_PTR VideoMemory,
                         ljeKERNEL_WHERE FromWhere);

/* Schedule a signal event. */
ljeSTATUS
ljkEVENT_Signal(ljkEVENT Event, ljtSIGNAL Signal, ljeKERNEL_WHERE FromWhere);

/* Schedule an Unlock event. */
ljeSTATUS
ljkEVENT_Unlock(ljkEVENT Event, ljeKERNEL_WHERE FromWhere, ljkMMU Mmu, ljtPOINTER Node);

ljeSTATUS
ljkEVENT_CommitDone(ljkEVENT Event,
                    ljeKERNEL_WHERE FromWhere,
                    ljkCONTEXT Context);

ljeSTATUS
ljkEVENT_Submit(ljkEVENT Event, ljkEVENT_ATTR EventAttr);

ljeSTATUS
ljkEVENT_Commit(ljkEVENT Event,
                ljsQUEUE_PTR Queue,
                ljtBOOL Forced,
                ljtBOOL Submit,
                ljtBOOL Shared);

/* Event callback routine. */
ljeSTATUS
ljkEVENT_Notify(ljkEVENT Event, ljtUINT32 IDs, ljeEVENT_FAULT *Fault);

/* Event callback routine. */
ljeSTATUS
ljkEVENT_Interrupt(ljkEVENT Event, ljtUINT32 IDs);

ljeSTATUS
ljkEVENT_ParityNotify(ljkEVENT Event);

ljeSTATUS
ljkEVENT_Dump(ljkEVENT Event);

/* Free all events belonging to a process. */
ljeSTATUS
ljkEVENT_FreeProcess(ljkEVENT Event, ljtUINT32 ProcessID);

#if ljdENABLE_VIDEO_MEMORY_MIRROR
typedef struct _ljsVIDMEM_NODE_MIRROR {
    ljeMIRROR_TYPE  type;
    ljkVIDMEM_NODE  mirrorNode;
#if !ljdSTATIC_VIDEO_MEMORY_MIRROR
    ljtUINT32       refCount;
# endif
} ljsVIDMEM_NODE_MIRROR;
#endif

/* ljuVIDMEM_NODE structure. */
typedef union _ljuVIDMEM_NODE {
    /* Allocated from ljkVIDMEM. */
    struct _ljsVIDMEM_NODE_VIDMEM {
        /* Owner of this node. */
        ljkVIDMEM               parent;

        /* Dual-linked list of nodes. */
        ljuVIDMEM_NODE_PTR      next;
        ljuVIDMEM_NODE_PTR      prev;

        /* Dual linked list of free nodes. */
        ljuVIDMEM_NODE_PTR      nextFree;
        ljuVIDMEM_NODE_PTR      prevFree;

        /* Information for this node. */
        ljtSIZE_T               offset;

        ljtADDRESS              address;
        ljtSIZE_T               bytes;
        ljtUINT32               alignment;

        /* Client virtual address. */
        ljtPOINTER              logical;

        /* Process ID owning this memory. */
        ljtUINT32               processID;

        /* Locked counter. */
        ljtINT32                locked;

        /* Memory pool. */
        ljePOOL                 pool;

        /* Kernel virtual address. */
        ljtPOINTER              kvaddr;

        /* mdl record pointer. */
        ljtPHYS_ADDR            physical;

        /* Pointer to ljkKERNEL object. */
        ljkKERNEL               kernel;

        /* Information for this node. */
        /* Contiguously allocated? */
        ljtBOOL                 contiguous;

        /* Customer private handle */
        ljtUINT32               gid;

        /* Page table information. */
        ljtSIZE_T               pageCount;

        ljtPOINTER              pageTables[ljvCORE_COUNT];

        /* Virtual address. */
        ljtADDRESS              addresses[ljvCORE_COUNT];

        /* Locked counter. */
        ljtINT32                lockeds[ljvCORE_COUNT];

        /* If the virtual address needs to be within 32 bit. */
        ljtBOOL                 lowVA;

        /* Video memory allocation type. */
        ljeVIDMEM_TYPE          type;

        /* Secure GPU virtual address. */
        ljtBOOL                 secure;

        ljtBOOL                 onFault;

        /* If the node is allocated for user. */
        ljtBOOL                 fromUser;

        /* Identifier. */
        ljtINT                  id;

        /* Link in _ljkMMU::nodeList. */
        ljsLISTHEAD             lockLink;

#if ljdENABLE_VG
        /* Physical address of this node, only meaningful when it is contiguous. */
        ljtPHYS_ADDR_T          physicalAddress;

        ljtPOINTER              kernelVirtual;
#endif

#if ljdVMEM_ALIGN_WITH_PAD
        /* padding bytes for alignment by huym */
        ljtUINT32               pad_bytes;
#endif
    } VidMem;

    struct _ljsVIDMEM_NODE_VIRTUAL_CHUNK {
        /* Owner of this chunk */
        ljkVIDMEM_BLOCK         parent;

        /* Pointer to ljkKERNEL object. */
        ljkKERNEL               kernel;

        /* Dual-linked list of chunk. */
        ljuVIDMEM_NODE_PTR      next;
        ljuVIDMEM_NODE_PTR      prev;

        /* Dual linked list of free chunk. */
        ljuVIDMEM_NODE_PTR      nextFree;
        ljuVIDMEM_NODE_PTR      prevFree;

        /* Information for this chunk. */
        ljtSIZE_T               offset;
        ljtADDRESS              addresses[ljvCORE_COUNT];
        ljtINT32                lockeds[ljvCORE_COUNT];

        ljtSIZE_T               bytes;

        /* Mapped user logical */
        ljtPOINTER              logical;

        /* Kernel virtual address. */
        ljtPOINTER              kvaddr;

        ljtUINT32               processID;

        /* If the node is allocated for user. */
        ljtBOOL                 fromUser;

        ljtINT                  id;

        /* Link in _ljkMMU::nodeList. */
        ljsLISTHEAD             lockLink;
    } VirtualChunk;

} ljuVIDMEM_NODE;

/* ljkVIDMEM object. */
struct _ljkVIDMEM {
    /* Object. */
    ljsOBJECT                   object;

    /* Pointer to ljkOS object. */
    ljkOS                       os;

    /* mdl record pointer... a kmalloc address. Process agnostic. */
    ljtPHYS_ADDR                physical;

    /* Information for this video memory heap. */
    ljtPHYS_ADDR_T              physicalBase;
    ljtSIZE_T                   bytes;
    ljtSIZE_T                   freeBytes;
    ljtSIZE_T                   minFreeBytes;

    /* caps inherit from its allocator, ~0u if allocator was not applicable. */
    ljtUINT32                   capability;

    /* Mapping for each type of surface. */
    ljtINT                      mapping[ljvVIDMEM_TYPE_COUNT];

    /* Sentinel nodes for up to 8 banks. */
    ljuVIDMEM_NODE              sentinel[8];

    /* Allocation threshold. */
    ljtSIZE_T                   threshold;

    /* The heap mutex. */
    ljtPOINTER                  mutex;
};

/* ljkVIDMEM_BLOCK object. */
typedef struct _ljsVIDMEM_BLOCK {
    /* Object. */
    ljsOBJECT                   object;

    /* Pointer to ljkOS object. */
    ljkOS                       os;

    /* linked list of nodes. */
    ljkVIDMEM_BLOCK             next;

    /* Contiguously allocated? */
    ljtBOOL                     contiguous;

    /* Customer private handle */
    ljtUINT32                   gid;

    /* mdl record pointer... a kmalloc address. Process agnostic. */
    ljtPHYS_ADDR                physical;

    /* Information for this video memory virtual block. */
    ljtSIZE_T                   bytes;
    ljtSIZE_T                   freeBytes;

    /* 1M page count. */
    ljtUINT32                   pageCount;
    ljtUINT32                   fixedPageCount;

    /* Gpu virtual base of this video memory heap. */
    ljtADDRESS                  addresses[ljvCORE_COUNT];
    ljtPOINTER                  pageTables[ljvCORE_COUNT];

    ljeVIDMEM_TYPE              type;

    /* Virtual chunk. */
    ljuVIDMEM_NODE              node;

    ljtPOINTER                  mutex;

    ljtBOOL                     secure;
    ljtBOOL                     onFault;

    /* If the virtual address needs to be within 32 bit. */
    ljtBOOL                     lowVA;

    /* If the video memory block is allocated for user space. */
    /* Only used when pre-process page table is enabled. */
    ljtBOOL                     fromUser;

    /* Only used when pre-process page table is enabled. */
    ljtUINT32                   processID;

    ljtBOOL                     cacheable;
} ljsVIDMEM_BLOCK;

typedef struct _ljsVIDMEM_PIDINFO {
    ljtUINT32 pid;
    ljtPOINTER ref;
    struct _ljsVIDMEM_PIDINFO *next;
} ljsVIDMEM_PIDINFO;

typedef struct _ljsVIDMEM_PIDINFO *ljkVIDMEM_PIDINFO;

typedef struct _ljsVIDMEM_NODE {
    _VIV_VIDMEM_METADATA metadata;

    /* Pointer to ljuVIDMEM_NODE. */
    ljuVIDMEM_NODE_PTR          node;

    /* Pointer to ljkKERNEL object. */
    ljkKERNEL                   kernel;

    /* Mutex to protect node. */
    ljtPOINTER                  mutex;

    /* Reference count. */
    ljtPOINTER                  reference;

    /* Name for client to import. */
    ljtUINT32                   name;

    /* Link in _ljkDB::videoMemList. */
    ljsLISTHEAD                 link;

    /* dma_buf */
    ljtPOINTER                  dmabuf;

    /* fd */
    ljtINT                      fd;

    /* Video memory allocation type. */
    ljeVIDMEM_TYPE              type;

    /* Pool from which node is allocated. */
    ljePOOL                     pool;

    ljsFENCE_SYNC               sync[ljvENGINE_GPU_ENGINE_COUNT];

    /* For DRM usage */
    ljtUINT64                   timeStamp;
    ljkVIDMEM_NODE              tsNode;
    ljtUINT32                   tilingMode;
    ljtUINT32                   tsMode;
    ljtUINT32                   tsCacheMode;
    ljtUINT64                   clearValue;

    ljtBOOL                     dirty;
    ljtBOOL                     cached;
    ljtBOOL                     needUnmap;
    ljtINT32                    mappedOffset;

#if ljdCAPTURE_ONLY_MODE
    ljtSIZE_T                   captureSize;
    ljtPOINTER                  captureLogical;
#endif

    /* Allocation flag */
    ljtUINT32                   flag;

#if ljdENABLE_VIDEO_MEMORY_MIRROR
    ljsVIDMEM_NODE_MIRROR       mirror;
#endif

    ljtPOINTER                  bo;
    ljkVIDMEM_PIDINFO           pidInfo;

} ljsVIDMEM_NODE;

typedef struct _ljsVIDMEM_HANDLE *ljkVIDMEM_HANDLE;
typedef struct _ljsVIDMEM_HANDLE {
    /* Pointer to ljkVIDMEM_NODE. */
    ljkVIDMEM_NODE              node;

    /* Handle for current process. */
    ljtUINT32                   handle;

    /* Reference count for this handle. */
    ljtPOINTER                  reference;
} ljsVIDMEM_HANDLE;

typedef struct _ljsSHBUF *ljsSHBUF_PTR;
typedef struct _ljsSHBUF {
    /* ID. */
    ljtUINT32                   id;

    /* Reference count. */
    ljtPOINTER                  reference;

    /* Data size. */
    ljtUINT32                   size;

    /* Data. */
    ljtPOINTER                  data;
} ljsSHBUF;

typedef struct _ljsCORE_INFO {
    ljeHARDWARE_TYPE type;
    ljeCORE          core;
    ljkKERNEL        kernel;
    ljtUINT32        chipID;
} ljsCORE_INFO;

typedef struct _ljsCORE_LIST {
    ljkKERNEL   kernels[ljvCORE_COUNT];
    ljtUINT32   num;
} ljsCORE_LIST;

#if ljdENABLE_CLEAR_FENCE
typedef struct _ljsUSER_FENCE_INFO {
    ljtADDRESS addr;
    ljtUINT64  fenceValue;
    ljtUINT64  recordId;

    /* Link in _ljsDEVICE::fenceList */
    ljsLISTHEAD fenceLink;
    ljtINT32   tid;
    ljtBOOL    use64BitFence;
} ljsUSER_FENCE_INFO;
#endif

/* A ljkDEVICE is a group of cores (ljkKERNEL in software). */
typedef struct _ljsDEVICE {
    /* Global device id in all the platforms. */
    ljtUINT                     id;
    /* Platform id that the device locate. */
    ljtUINT                     platformIndex;

    ljsCORE_INFO                coreInfoArray[ljvCORE_COUNT];
    ljtUINT32                   coreNum;
    ljsCORE_LIST                map[ljvHARDWARE_NUM_TYPES];
    ljeHARDWARE_TYPE            defaultHwType;

    ljkOS                       os;

    /* Process resource database. */
    ljkDB                       database;

    /* Same hardware type of one platform device shares one MMU. */
    ljkMMU                      mmus[ljvHARDWARE_NUM_TYPES];

    ljkCOMMAND                  commands[ljvHARDWARE_NUM_TYPES];

    ljkKERNEL                   kernels[ljvCORE_COUNT];

    ljtUINT32                   extSRAMIndex;

    /* Physical address of internal SRAMs. */
    ljtPHYS_ADDR_T              sRAMBases[ljvCORE_COUNT][ljvSRAM_INTER_COUNT];
    /* Internal SRAMs' size. */
    ljtUINT32                   sRAMSizes[ljvCORE_COUNT][ljvSRAM_INTER_COUNT];
    /* GPU/VIP virtual address of internal SRAMs. */
    ljtADDRESS                  sRAMBaseAddresses[ljvCORE_COUNT][ljvSRAM_INTER_COUNT];
    ljtBOOL                     sRAMPhysFaked[ljvCORE_COUNT][ljvSRAM_INTER_COUNT];

    /* CPU physical address of external SRAMs. */
    ljtPHYS_ADDR_T              extSRAMBases[ljvSRAM_EXT_COUNT];
    /* GPU physical address of external SRAMs. */
    ljtPHYS_ADDR_T              extSRAMGPUBases[ljvSRAM_EXT_COUNT];
    /* External SRAMs' size. */
    ljtUINT32                   extSRAMSizes[ljvSRAM_EXT_COUNT];
    /* GPU/VIP virtual address of external SRAMs. */
    ljtADDRESS                  extSRAMBaseAddresses[ljvSRAM_EXT_COUNT];
    /* MDL. */
    ljtPHYS_ADDR                extSRAMPhysicals[ljvSRAM_EXT_COUNT];
    /* IntegerId. */
    ljtUINT32                   extSRAMGPUPhysNames[ljvSRAM_EXT_COUNT];
    /* Vidmem. */
    ljkVIDMEM                   extSRAMVidMems[ljvSRAM_EXT_COUNT];

    /* Show SRAM mapping info or not. */
    ljtUINT                     showSRAMMapInfo;

    /* Mutex to make sure stuck dump for multiple cores doesn't interleave. */
    ljtPOINTER                  stuckDumpMutex;

    /* Mutex for multi-core combine mode command submission */
    ljtPOINTER                  commitMutex;

    /* Mutex for per-device power management. */
    ljtPOINTER                  powerMutex;

    /* Mutex for recovery all core */
    ljtPOINTER                  recoveryMutex;

    /* Which memory this core uses. */
    ljtUINT32                   memIndex;

    ljtPHYS_ADDR_T              contiguousBases[ljdSYSTEM_RESERVE_COUNT];
    ljtSIZE_T                   contiguousSizes[ljdSYSTEM_RESERVE_COUNT];
    ljtPHYS_ADDR                contiguousPhysicals[ljdSYSTEM_RESERVE_COUNT];
    ljtUINT32                   contiguousPhysNames[ljdSYSTEM_RESERVE_COUNT];
    ljkVIDMEM                   contiguousVidMems[ljdSYSTEM_RESERVE_COUNT];

    ljtPHYS_ADDR_T              lowContiguousBase;
    ljtSIZE_T                   lowContiguousSize;

    ljtPHYS_ADDR_T              exclusiveBase;
    ljtSIZE_T                   exclusiveSize;
    ljtPHYS_ADDR                exclusivePhysical;
    ljtUINT32                   exclusivePhysName;
    ljtPOINTER                  exclusiveLogical;
    ljkVIDMEM                   exclusiveVidMem;

    ljtPHYS_ADDR_T              externalBase;
    ljtSIZE_T                   externalSize;
    ljtPHYS_ADDR                externalPhysical;
    ljtUINT32                   externalPhysName;
    ljtPOINTER                  externalLogical;
    ljkVIDMEM                   externalVidMem;

    /* IRQ management. */
    ljtINT                      irqLines[ljvCORE_COUNT];
    ljtBOOL                     isrInitializeds[ljvCORE_COUNT];
    ljtPOINTER                  isrThread[ljvCORE_COUNT];
    ljtBOOL                     killIsrThread;

    /* Register memory. */
    ljtPOINTER                  registerBases[ljvCORE_COUNT];
    ljtSIZE_T                   registerSizes[ljvCORE_COUNT];
    ljtPOINTER                  registerBasesMapped[ljvCORE_COUNT];

    /* By request_mem_region. */
    ljtUINT64                   requestedRegisterMemBases[ljvCORE_COUNT];
    ljtSIZE_T                   requestedRegisterMemSizes[ljvCORE_COUNT];

    ljtUINT32                   baseAddress;
    ljtUINT32                   physBase;
    ljtUINT32                   physSize;

    /* Thread management. */
    ljtPOINTER                  threadCtxts[ljvCORE_COUNT];
    ljtBOOL                     threadInitializeds[ljvCORE_COUNT];

    /* States before suspend. */
    ljeCHIPPOWERSTATE           statesStored[ljvCORE_COUNT];

    /* Which video memory is for device. */
    ljtUINT32                   vidMemIndex;

    /* Which SRAM is for device. */
    ljtUINT32                   sRAMIndex;

    /* Core count configuration. */
    ljtUINT32                   configCoreCount;
    ljtUINT32                   config2DCoreCount;

    /* Platform device structure. */
    ljtPOINTER                  dev;

    /* Print the memory info or not. */
    ljtBOOL                     showMemInfo;

    /* Process page table enable or not. */
    ljtBOOL                     processPageTable;

#if ljdENABLE_SW_PREEMPTION
    ljtPOINTER                  atomPriorityID;
    ljtPOINTER                  preemptThread[ljvCORE_COUNT];
    ljtBOOL                     preemptThreadInits[ljvCORE_COUNT];
#endif

#if ljdENABLE_CLEAR_FENCE
    ljsLISTHEAD                 fenceList;
    ljtPOINTER                  fenceListMutex;
    ljtPOINTER                  fenceIdr;
    spinlock_t                  fenceIdrLock;
#endif
} ljsDEVICE;

/* video memory pool functions. */
/* Construct a new ljkVIDMEM object. */
ljeSTATUS
ljkVIDMEM_Construct(ljkOS Os,
                    ljtPHYS_ADDR_T PhysicalBase,
                    ljtSIZE_T Bytes,
                    ljtSIZE_T Threshold,
                    ljtSIZE_T Banking,
                    ljkVIDMEM *Memory);

/* Destroy an ljkVDIMEM object. */
ljeSTATUS
ljkVIDMEM_Destroy(ljkVIDMEM Memory);

/* video memory node handle functions. */
ljeSTATUS
ljkVIDMEM_HANDLE_Allocate(ljkKERNEL Kernel, ljtUINT32 ProcessID, ljkVIDMEM_NODE Node, ljtUINT32 *Handle);

ljeSTATUS
ljkVIDMEM_HANDLE_Reference(ljkKERNEL Kernel, ljtUINT32 ProcessID, ljtUINT32 Handle);

ljeSTATUS
ljkVIDMEM_HANDLE_Dereference(ljkKERNEL Kernel, ljtUINT32 ProcessID, ljtUINT32 Handle);

ljeSTATUS
ljkVIDMEM_HANDLE_Lookup(ljkKERNEL Kernel,
                        ljtUINT32 ProcessID,
                        ljtUINT32 Handle,
                        ljkVIDMEM_NODE *Node);

ljeSTATUS
ljkVIDMEM_HANDLE_Lookup2(ljkKERNEL Kernel,
                         ljsDATABASE_PTR Database,
                         ljtUINT32 Handle,
                         ljkVIDMEM_NODE *Node);

#if ljdENABLE_VIDEO_MEMORY_MIRROR
ljeSTATUS
ljkVIDMEM_NODE_AllocateMirrorBuf(ljkKERNEL Kernel, ljkVIDMEM_NODE NodeObject,
                           ljeMIRROR_TYPE MirrorType);
ljeSTATUS
ljkVIDMEM_NODE_FreeMirrorBuf(ljkKERNEL Kernel, ljkVIDMEM_NODE NodeObject);
#endif

/* video memory node functions. */
ljeSTATUS
ljkVIDMEM_NODE_AllocateLinear(ljkKERNEL Kernel,
                              ljkVIDMEM VideoMemory,
                              ljePOOL Pool,
                              ljeVIDMEM_TYPE Type,
                              ljtUINT32 Flag,
                              ljtUINT32 Alignment,
                              ljtBOOL Specified,
                              ljtSIZE_T *Bytes,
                              ljkVIDMEM_NODE *NodeObject);

ljeSTATUS
ljkVIDMEM_NODE_AllocateVirtual(ljkKERNEL Kernel,
                               ljePOOL Pool,
                               ljeVIDMEM_TYPE Type,
                               ljtUINT32 Flag,
                               ljtSIZE_T *Bytes,
                               ljkVIDMEM_NODE *NodeObject);

ljeSTATUS
ljkVIDMEM_NODE_AllocateVirtualChunk(ljkKERNEL Kernel,
                                    ljePOOL Pool,
                                    ljeVIDMEM_TYPE Type,
                                    ljtUINT32 Flag,
                                    ljtSIZE_T *Bytes,
                                    ljkVIDMEM_NODE *NodeObject);

ljeSTATUS
ljkVIDMEM_NODE_Reference(ljkKERNEL Kernel, ljkVIDMEM_NODE Node);

ljeSTATUS
ljkVIDMEM_NODE_Dereference(ljkKERNEL Kernel, ljkVIDMEM_NODE Node);

ljeSTATUS
ljkVIDMEM_NODE_DereferenceEx(ljkKERNEL Kernel, ljkVIDMEM_NODE Node, ljtUINT32 ProcessID);

ljeSTATUS
ljkVIDMEM_NODE_GetReference(ljkKERNEL Kernel,
                            ljkVIDMEM_NODE NodeObject,
                            ljtINT32 *ReferenceCount);

ljeSTATUS
ljkVIDMEM_NODE_Lock(ljkKERNEL Kernel,
                    ljkVIDMEM_NODE NodeObject,
                    ljtADDRESS *Address);

ljeSTATUS
ljkVIDMEM_NODE_Unlock(ljkKERNEL Kernel,
                      ljkVIDMEM_NODE NodeObject,
                      ljkMMU Mmu,
                      ljtBOOL *Asynchroneous);

ljeSTATUS
ljkVIDMEM_NODE_CleanCache(ljkKERNEL Kernel,
                          ljkVIDMEM_NODE NodeObject,
                          ljtSIZE_T Offset,
                          ljtPOINTER Logical,
                          ljtSIZE_T Bytes);

ljeSTATUS
ljkVIDMEM_NODE_InvalidateCache(ljkKERNEL Kernel,
                               ljkVIDMEM_NODE NodeObject,
                               ljtSIZE_T Offset,
                               ljtPOINTER Logical,
                               ljtSIZE_T Bytes);

ljeSTATUS
ljkVIDMEM_NODE_GetLockCount(ljkKERNEL Kernel,
                            ljkVIDMEM_NODE NodeObject,
                            ljtINT32 *LockCount);

ljeSTATUS
ljkVIDMEM_NODE_LockCPU(ljkKERNEL Kernel,
                       ljkVIDMEM_NODE NodeObject,
                       ljtBOOL Cacheable,
                       ljtBOOL FromUser,
                       ljtPOINTER *Logical);

ljeSTATUS
ljkVIDMEM_NODE_UnlockCPU(ljkKERNEL Kernel,
                         ljkVIDMEM_NODE NodeObject,
                         ljtUINT32 ProcessID,
                         ljtBOOL FromUser,
                         ljtBOOL Defer);

ljeSTATUS
ljkVIDMEM_NODE_GetCPUPhysical(ljkKERNEL Kernel,
                              ljkVIDMEM_NODE NodeObject,
                              ljtSIZE_T Offset,
                              ljtPHYS_ADDR_T *PhysicalAddress);

ljeSTATUS
ljkVIDMEM_NODE_GetGPUPhysical(ljkKERNEL Kernel,
                              ljkVIDMEM_NODE NodeObject,
                              ljtUINT32 Offset,
                              ljtPHYS_ADDR_T *PhysicalAddress);

ljeSTATUS
ljkVIDMEM_NODE_GetGid(ljkKERNEL Kernel,
                      ljkVIDMEM_NODE NodeObject,
                      ljtUINT32 *Gid);

ljeSTATUS
ljkVIDMEM_NODE_GetSize(ljkKERNEL Kernel,
                       ljkVIDMEM_NODE NodeObject,
                       ljtSIZE_T *Size);

ljeSTATUS
ljkVIDMEM_NODE_GetType(ljkKERNEL Kernel,
                       ljkVIDMEM_NODE NodeObject,
                       ljeVIDMEM_TYPE *Type,
                       ljePOOL *Pool);

ljeSTATUS
ljkVIDMEM_NODE_Export(ljkKERNEL Kernel,
                      ljkVIDMEM_NODE NodeObject,
                      ljtINT32 Flags,
                      ljtPOINTER *DmaBuf,
                      ljtINT32 *FD);

ljeSTATUS
ljkVIDMEM_NODE_Name(ljkKERNEL Kernel,
                    ljkVIDMEM_NODE NodeObject,
                    ljtUINT32 *Name);

ljeSTATUS
ljkVIDMEM_NODE_Import(ljkKERNEL Kernel,
                      ljtUINT32 Name,
                      ljkVIDMEM_NODE *NodeObject);

ljeSTATUS
ljkVIDMEM_NODE_GetFd(ljkKERNEL Kernel,
                     ljkVIDMEM_NODE NodeObject,
                     ljtBOOL Exported,
                     ljtINT *Fd);

ljeSTATUS
ljkVIDMEM_NODE_WrapUserMemory(ljkKERNEL Kernel,
                              ljsUSER_MEMORY_DESC_PTR Desc,
                              ljeVIDMEM_TYPE Type,
                              ljkVIDMEM_NODE *NodeObject,
                              ljtUINT64 *Bytes);

ljeSTATUS
ljkVIDMEM_NODE_SetCommitStamp(ljkKERNEL Kernel,
                              ljeENGINE Engine,
                              ljkVIDMEM_NODE NodeObject,
                              ljtUINT64 CommitStamp);

ljeSTATUS
ljkVIDMEM_NODE_GetCommitStamp(ljkKERNEL Kernel,
                              ljeENGINE Engine,
                              ljkVIDMEM_NODE NodeObject,
                              ljtUINT64_PTR CommitStamp);

ljeSTATUS
ljkVIDMEM_NODE_Find(ljkKERNEL Kernel,
                    ljtADDRESS Address,
                    ljkVIDMEM_NODE *NodeObject,
                    ljtSIZE_T *Offset);

ljeSTATUS
ljkVIDMEM_NODE_IsContiguous(ljkKERNEL Kernel,
                            ljkVIDMEM_NODE NodeObject,
                            ljtBOOL *Contiguous);

ljeSTATUS
ljkVIDMEM_NODE_GetMemoryHandle(ljkKERNEL Kernel,
                               ljkVIDMEM_NODE NodeObject,
                               ljtPOINTER *MemoryHandle);

ljeSTATUS
ljkVIDMEM_NODE_GetMapKernel(ljkKERNEL Kernel,
                            ljkVIDMEM_NODE NodeObject,
                            ljtPOINTER *KernelMap);

ljeSTATUS
ljkVIDMEM_NODE_GetOffset(ljkKERNEL Kernel,
                         ljkVIDMEM_NODE NodeObject,
                         ljtSIZE_T *Offset);

#if ljdENABLE_TTM
ljeSTATUS
ljkVIDMEM_Validate(ljkKERNEL Kernel,
                        ljtUINT32 NodeHandle, ljtUINT32 ProcessID,
                        ljtUINT32 Pool);

ljeSTATUS
ljkVIDMEM_MoveClean(ljkKERNEL kernel, ljkVIDMEM_NODE newObject, ljtUINT32 newHandle);

ljeSTATUS
ljkVIDMEM_Exchange(ljkKERNEL Kernel, ljkVIDMEM_NODE oldObject, ljkVIDMEM_NODE newObject);

void
ljkVIDMEM_RevokeMap(ljkVIDMEM_NODE nodeObj);
#endif

ljeSTATUS
ljkOS_CreateKernelMapping(ljkOS Os,
                          ljtPHYS_ADDR Physical,
                          ljtSIZE_T Offset,
                          ljtSIZE_T Bytes,
                          ljtPOINTER *Logical);

ljeSTATUS
ljkOS_DestroyKernelMapping(ljkOS Os, ljtPHYS_ADDR Physical, ljtPOINTER Logical);

ljeSTATUS
ljkOS_GetFd(ljtSTRING Name, ljsFDPRIVATE_PTR Private, ljtINT *Fd);

/*******************************************************************************
 **
 **  ljkOS_ReadMappedPointer
 **
 **  Read pointer mapped from user pointer which returned by ljkOS_MapUserPointer.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtPOINTER Address
 **          Pointer returned by ljkOS_MapUserPointer.
 **
 **      ljtUINT32_PTR Data
 **          Pointer to hold 32 bits data.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_ReadMappedPointer(ljkOS Os, ljtPOINTER Address, ljtUINT32_PTR Data);

ljeSTATUS
ljkKERNEL_AttachProcess(ljkKERNEL Kernel, ljtBOOL Attach);

ljeSTATUS
ljkKERNEL_AttachProcessEx(ljkKERNEL Kernel, ljtBOOL Attach, ljtUINT32 PID);

ljeSTATUS
ljkKERNEL_AllocateVideoMemory(ljkKERNEL Kernel,
                              ljtUINT32 Alignment,
                              ljeVIDMEM_TYPE Type,
                              ljtUINT32 Flag,
                              ljtSIZE_T *Bytes,
                              ljePOOL *Pool,
                              ljkVIDMEM_NODE *NodeObject);

ljeSTATUS
ljkHARDWARE_QchannelPowerControl(ljkHARDWARE Hardware,
                                 ljtBOOL ClockState,
                                 ljtBOOL PowerState);

ljeSTATUS
ljkHARDWARE_QchannelBypass(ljkHARDWARE Hardware, ljtBOOL Enable);

ljeSTATUS
ljkHARDWARE_QueryIdle(ljkHARDWARE Hardware, ljtBOOL_PTR IsIdle);

ljeSTATUS
ljkHARDWARE_WaitFence(ljkHARDWARE Hardware,
                      ljtPOINTER Logical,
                      ljtUINT64 FenceData,
                      ljtADDRESS FenceAddress,
                      ljtUINT32 *Bytes);

ljeSTATUS
ljkHARDWARE_UpdateContextID(ljkHARDWARE Hardware);

ljeSTATUS
ljkHARDWARE_QueryMcfe(ljkHARDWARE Hardware,
                      const ljeMCFE_CHANNEL_TYPE * Channels[],
                      ljtUINT32 *Count);

#if ljdSECURITY
ljeSTATUS
ljkKERNEL_SecurityOpen(ljkKERNEL Kernel, ljtUINT32 GPU, ljtUINT32 *Channel);

/*
 ** Close a security service channel
 */
ljeSTATUS
ljkKERNEL_SecurityClose(ljtUINT32 Channel);

/*
 ** Security service interface.
 */
ljeSTATUS
ljkKERNEL_SecurityCallService(ljtUINT32 Channel, ljsTA_INTERFACE *Interface);

ljeSTATUS
ljkKERNEL_SecurityStartCommand(ljkKERNEL Kernel);

ljeSTATUS
ljkKERNEL_SecurityAllocateSecurityMemory(ljkKERNEL Kernel,
                                         ljtUINT32 Bytes,
                                         ljtUINT32 *Handle);

ljeSTATUS
ljkKERNEL_SecurityExecute(ljkKERNEL Kernel, ljtPOINTER Buffer, ljtUINT32 Bytes);

ljeSTATUS
ljkKERNEL_SecurityMapMemory(ljkKERNEL Kernel,
                            ljtUINT32 *PhysicalArray,
                            ljtUINT32 PageCount,
                            ljtADDRESS *GPUAddress);

ljeSTATUS
ljkKERNEL_SecurityUnmapMemory(ljkKERNEL Kernel,
                              ljtADDRESS GPUAddress,
                              ljtUINT32 PageCount);

#endif

#if ljdENABLE_TRUST_APPLICATION
ljeSTATUS
ljkKERNEL_SecurityOpen(ljkKERNEL Kernel, ljtUINT32 GPU, ljtUINT32 *Channel);

/*
 ** Close a security service channel
 */
ljeSTATUS
ljkKERNEL_SecurityClose(ljtUINT32 Channel);

/*
 ** Security service interface.
 */
ljeSTATUS
ljkKERNEL_SecurityCallService(ljtUINT32 Channel,
                              ljsTA_INTERFACE *Interface);

ljeSTATUS
ljkKERNEL_SecurityStartCommand(ljkKERNEL Kernel,
                               ljtADDRESS Address,
                               ljtUINT32 Bytes);

ljeSTATUS
ljkKERNEL_SecurityMapMemory(ljkKERNEL Kernel,
                            ljtUINT32 *PhysicalArray,
                            ljtPHYS_ADDR_T Physical,
                            ljtUINT32 PageCount,
                            ljtADDRESS *GPUAddress);

ljeSTATUS
ljkKERNEL_SecurityUnmapMemory(ljkKERNEL Kernel,
                              ljtADDRESS GPUAddress,
                              ljtUINT32 PageCount);

ljeSTATUS
ljkKERNEL_SecurityDumpMMUException(ljkKERNEL Kernel);

ljeSTATUS
ljkKERNEL_ReadMMUException(ljkKERNEL Kernel,
                           ljtUINT32_PTR MMUStatus,
                           ljtUINT32_PTR MMUException);

ljeSTATUS
ljkKERNEL_HandleMMUException(ljkKERNEL Kernel,
                             ljtUINT32 MMUStatus,
                             ljtPHYS_ADDR_T Physical,
                             ljtADDRESS GPUAddres);
#endif

ljeSTATUS
ljkKERNEL_CreateShBuffer(ljkKERNEL Kernel,
                         ljtUINT32 Size,
                         ljtSHBUF *ShBuf);

ljeSTATUS
ljkKERNEL_DestroyShBuffer(ljkKERNEL Kernel, ljtSHBUF ShBuf);

ljeSTATUS
ljkKERNEL_MapShBuffer(ljkKERNEL Kernel, ljtSHBUF ShBuf);

ljeSTATUS
ljkKERNEL_WriteShBuffer(ljkKERNEL Kernel,
                        ljtSHBUF ShBuf,
                        ljtPOINTER UserData,
                        ljtUINT32 ByteCount);

ljeSTATUS
ljkKERNEL_ReadShBuffer(ljkKERNEL Kernel,
                       ljtSHBUF ShBuf,
                       ljtPOINTER UserData,
                       ljtUINT32 ByteCount,
                       ljtUINT32 *BytesRead);

ljeSTATUS
ljkKERNEL_GetHardwareType(ljkKERNEL Kernel, ljeHARDWARE_TYPE *Type);

#if ljdENABLE_MP_SWITCH
ljeSTATUS
ljkKERNEL_DetectMpModeSwitch(ljkKERNEL Kernel,
                             ljeMULTI_PROCESSOR_MODE Mode,
                             ljtUINT32 *SwitchMpMode);
#endif

ljeSTATUS
ljkKERNEL_CancelJob(ljkKERNEL Kernel);

ljeSTATUS
ljkKERNEL_FindProcessMMU(ljkKERNEL Kernel, ljtUINT32 ProcessID, ljkMMU *Mmu);

ljeSTATUS
ljkKERNEL_GetCurrentMMU(ljkKERNEL Kernel, ljtBOOL FromUser,
                        ljtUINT32 ProcessID, ljkMMU *Mmu);

ljeSTATUS
ljkKERNEL_SwitchMMU(ljkKERNEL Kernel, ljtBOOL Shared, ljkMMU Mmu);

#if ljdENABLE_TTM
ljeSTATUS
ljkKERNEL_SyncBufList(ljkKERNEL Kernel);
#endif

/*******************************************************************************
 ******************************* ljkCONTEXT Object *****************************
 ******************************************************************************/

ljeSTATUS
ljkCONTEXT_Construct(ljkOS Os,
                     ljkHARDWARE Hardware,
                     ljtUINT32 ProcessID,
                     ljtBOOL Shared,
                     ljkCONTEXT *Context);

ljeSTATUS
ljkCONTEXT_Destroy(ljkCONTEXT Context);

ljeSTATUS
ljkCONTEXT_Update(ljkCONTEXT Context,
                  ljtUINT32 ProcessID,
                  ljsSTATE_DELTA_PTR StateDelta);

ljeSTATUS
ljkCONTEXT_MapBuffer(ljkCONTEXT Context, ljtUINT64 *Logicals, ljtUINT32 *Bytes);

void
ljkQUEUE_Enqueue(ljkQUEUE LinkQueue, ljuQUEUEDATA *Data);

void
ljkQUEUE_GetData(ljkQUEUE LinkQueue, ljtUINT32 Index, ljuQUEUEDATA **Data);

ljeSTATUS
ljkQUEUE_Allocate(ljkOS Os, ljkQUEUE Queue, ljtUINT32 Size);

ljeSTATUS
ljkQUEUE_Free(ljkOS Os, ljkQUEUE Queue);

/*******************************************************************************
 ****************************** ljkRECORDER Object *****************************
 ******************************************************************************/
ljeSTATUS
ljkRECORDER_Construct(ljkOS Os,
                      ljkHARDWARE Hardware,
                      ljkRECORDER *Recorder);

ljeSTATUS
ljkRECORDER_Destory(ljkOS Os, ljkRECORDER Recorder);

void
ljkRECORDER_AdvanceIndex(ljkRECORDER Recorder, ljtUINT64 CommitStamp);

void
ljkRECORDER_Record(ljkRECORDER Recorder,
                   ljtUINT8_PTR CommandBuffer,
                   ljtUINT32 CommandBytes,
                   ljtUINT8_PTR ContextBuffer,
                   ljtUINT32 ContextBytes);

void
ljkRECORDER_Dump(ljkRECORDER Recorder);

ljeSTATUS
ljkRECORDER_UpdateMirror(ljkRECORDER Recorder, ljtUINT32 State, ljtUINT32 Data);

/******************************************************************************
 ****************************** ljkCOMMAND Object *****************************
 ******************************************************************************/

/* Construct a new ljkCOMMAND object. */
ljeSTATUS
ljkCOMMAND_Construct(ljkKERNEL Kernel,
                     ljeHW_FE_TYPE FeType,
                     ljkCOMMAND *Command);

/* Destroy an ljkCOMMAND object. */
ljeSTATUS
ljkCOMMAND_Destroy(ljkCOMMAND Command);

/* Acquire command queue synchronization objects. */
ljeSTATUS
ljkCOMMAND_EnterCommit(ljkCOMMAND Command, ljtBOOL FromPower);

/* Release command queue synchronization objects. */
ljeSTATUS
ljkCOMMAND_ExitCommit(ljkCOMMAND Command, ljtBOOL FromPower);

/* Start the command queue. */
ljeSTATUS
ljkCOMMAND_Start(ljkCOMMAND Command);

/* Stop the command queue. */
ljeSTATUS
ljkCOMMAND_Stop(ljkCOMMAND Command);

/* Commit command buffers. */
ljeSTATUS
ljkCOMMAND_Commit(ljkCOMMAND Command,
                  ljsHAL_SUBCOMMIT *SubCommit,
                  ljtUINT32 ProcessId,
                  ljtBOOL Shared,
                  ljtUINT64_PTR CommitStamp,
                  ljtBOOL *contextSwitched);

/* Reserve space in the command buffer. */
ljeSTATUS
ljkCOMMAND_Reserve(ljkCOMMAND Command,
                   ljtUINT32 RequestedBytes,
                   ljtPOINTER *Buffer,
                   ljtUINT32 *BufferSize);

/*
 * Execute reserved space in the command buffer.
 * Wait link FE version.
 */
ljeSTATUS
ljkCOMMAND_Execute(ljkCOMMAND Command, ljtUINT32 RequstedBytes);

/*
 * Execute reserved space in the command buffer.
 * End command version.
 */
ljeSTATUS
ljkCOMMAND_ExecuteEnd(ljkCOMMAND Command, ljtUINT32 RequstedBytes);

/*
 * Execute reserved space in the command buffer.
 * Async FE version.
 */
ljeSTATUS
ljkCOMMAND_ExecuteAsync(ljkCOMMAND Command, ljtUINT32 RequestedBytes);

/*
 * Execute reserved space in the command buffer.
 * MC-FE version.
 */
ljeSTATUS
ljkCOMMAND_ExecuteMultiChannel(ljkCOMMAND Command,
                               ljtBOOL Priority,
                               ljtUINT32 ChannelId,
                               ljtUINT32 RequstedBytes);

/* Stall the command queue. */
ljeSTATUS
ljkCOMMAND_Stall(ljkCOMMAND Command, ljtBOOL FromPower);

/* Attach user process. */
ljeSTATUS
ljkCOMMAND_Attach(ljkCOMMAND Command,
                  ljkCONTEXT *Context,
                  ljtSIZE_T *MaxState,
                  ljtUINT32 *NumStates,
                  ljtUINT32 ProcessID,
                  ljtBOOL Shared);

/* Dump command buffer being executed by GPU. */
ljeSTATUS
ljkCOMMAND_DumpExecutingBuffer(ljkCOMMAND Command);

/* Detach user process. */
ljeSTATUS
ljkCOMMAND_Detach(ljkCOMMAND Command, ljkCONTEXT Context);

ljeSTATUS
ljkCOMMAND_CheckFlushMMU(ljkCOMMAND Command, ljkHARDWARE Hardware);

void
ljsLIST_Init(ljsLISTHEAD_PTR Node);

void
ljsLIST_Add(ljsLISTHEAD_PTR New, ljsLISTHEAD_PTR Head);

void
ljsLIST_AddTail(ljsLISTHEAD_PTR New, ljsLISTHEAD_PTR Head);

void
ljsLIST_Del(ljsLISTHEAD_PTR Node);

ljtBOOL
ljsLIST_Empty(ljsLISTHEAD_PTR Head);

#define ljmkLIST_FOR_EACH(pos, head) for (pos = (head)->next; pos != (head); pos = pos->next)

#define ljmkLIST_FOR_EACH_SAFE(pos, n, head)                                                       \
    for (pos = (head)->next, n = pos->next; pos != (head); pos = n, n = pos->next)

ljeSTATUS
ljkFENCE_Create(ljkOS Os, ljkKERNEL Kernel, ljkFENCE *Fence);

ljeSTATUS
ljkFENCE_Destory(ljkOS Os, ljkFENCE Fence);

ljeSTATUS
ljkFENCE_Signal(ljkOS Os, ljkFENCE Fence);

ljeSTATUS
ljkDEVICE_Construct(ljkOS Os, ljkDEVICE *Device);

ljeSTATUS
ljkDEVICE_DestroyCores(ljkDEVICE Device);

ljeSTATUS
ljkDEVICE_AddCore(ljkDEVICE Device,
                  ljeCORE Core,
                  ljtUINT chipID,
                  ljtPOINTER Context,
                  ljkKERNEL *Kernel);

ljeSTATUS
ljkDEVICE_Destroy(ljkOS Os, ljkDEVICE Device);

ljeSTATUS
ljkDEVICE_Dispatch(ljkDEVICE Device, ljsHAL_INTERFACE_PTR Interface);

ljeSTATUS
ljkDEVICE_ChipInfo(ljkDEVICE Device,
                   ljsHAL_INTERFACE_PTR Interface,
                   ljtUINT32 *CoreCount);

#if LJMICRO_PROFILER
ljeSTATUS
ljkDEVICE_Profiler_Dispatch(ljkDEVICE Device,
                            ljsHAL_PROFILER_INTERFACE_PTR Interface);
#endif

ljeSTATUS
ljkDEVICE_GetMMU(ljkDEVICE Device,
                 ljeHARDWARE_TYPE Type,
                 ljkMMU *Mmu);

ljeSTATUS
ljkDEVICE_SetMMU(ljkDEVICE Device,
                 ljeHARDWARE_TYPE Type,
                 ljkMMU Mmu);

ljeSTATUS
ljkDEVICE_GetCommandQueue(ljkDEVICE Device,
                          ljeHARDWARE_TYPE Type,
                          ljkCOMMAND *Command);

ljeSTATUS
ljkDEVICE_SetCommandQueue(ljkDEVICE Device,
                          ljeHARDWARE_TYPE Type,
                          ljkCOMMAND Command);

#if ljdENABLE_TRUST_APPLICATION
ljeSTATUS
ljkKERNEL_MapInTrustApplicaiton(ljkKERNEL Kernel,
                                ljtPOINTER Logical,
                                ljtPHYS_ADDR Physical,
                                ljtADDRESS GPUAddress,
                                ljtSIZE_T PageCount);
#endif

#if ljdSECURITY || ljdENABLE_TRUST_APPLICATION
ljeSTATUS
ljkOS_OpenSecurityChannel(ljkOS Os, ljeCORE Core, ljtUINT32 *Channel);

ljeSTATUS
ljkOS_CloseSecurityChannel(ljtUINT32 Channel);

ljeSTATUS
ljkOS_CallSecurityService(ljtUINT32 Channel, ljsTA_INTERFACE *Interface);

ljeSTATUS
ljkOS_InitSecurityChannel(ljtUINT32 Channel);

ljeSTATUS
ljkOS_AllocatePageArray(ljkOS Os,
                        ljtPHYS_ADDR Physical,
                        ljtSIZE_T PageCount,
                        ljtPOINTER *PageArrayLogical,
                        ljtPHYS_ADDR *PageArrayPhysical);
#endif

#ifdef __cplusplus
}
#endif

#endif /* __lj_hal_kernel_h_ */
