/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#include "lj_hal_kernel_linux.h"
#include "lj_hal_kernel_allocator.h"
#include <linux/pagemap.h>
#include <linux/seq_file.h>
#include <linux/mman.h>
#include <linux/atomic.h>
#include <linux/dma-mapping.h>
#include <linux/slab.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 32)
# include <linux/anon_inodes.h>
#endif
#include <linux/file.h>

#include "lj_hal_kernel_allocator_array.h"
#include "lj_hal_kernel_platform.h"

#define _LJ_OBJ_ZONE ljvZONE_OS

/*******************************************************************************
 ******************************* Debugfs Support *******************************
 *******************************************************************************/
static ljeSTATUS
_AllocatorDebugfsInit(ljkOS Os)
{
    ljeSTATUS status;
    ljkGALDEVICE device = Os->device;

    ljkDEBUGFS_DIR dir = &Os->allocatorDebugfsDir;

    ljmkONERROR(ljkDEBUGFS_DIR_Init(dir, device->debugfsDir.root, "allocators"));

    return ljvSTATUS_OK;

OnError:
    return status;
}

static void
_AllocatorDebugfsCleanup(ljkOS Os)
{
    ljkDEBUGFS_DIR dir = &Os->allocatorDebugfsDir;

    ljkDEBUGFS_DIR_Deinit(dir);
}

/****************************************************************************
 *********************** Allocator management *******************************
 ****************************************************************************/
#if ljdANON_FILE_FOR_ALLOCATOR
static int
tmp_mmap(struct file *fp, struct vm_area_struct *vma)
{
    return 0;
}

static const struct file_operations tmp_fops = {
    .mmap = tmp_mmap,
};

#endif

ljeSTATUS
ljkOS_ImportAllocators(ljkOS Os)
{
    ljeSTATUS status;
    ljtUINT i;
    ljkALLOCATOR allocator;

#if ljdANON_FILE_FOR_ALLOCATOR
    struct file *anon_file = ljvNULL;
    ljtINT32 ufd = 0;

    ufd = anon_inode_getfd("[ljmcore]", &tmp_fops, ljvNULL, O_RDWR);
    anon_file = fget(ufd);
#endif

    _AllocatorDebugfsInit(Os);

    INIT_LIST_HEAD(&Os->allocatorList);

    for (i = 0; i < ljmCOUNTOF(allocatorArray); i++) {
        if (allocatorArray[i].construct) {
            /* Construct allocator. */
            status = allocatorArray[i].construct(Os, &Os->allocatorDebugfsDir, &allocator);

            if (ljmIS_ERROR(status)) {
                ljmkPRINT("[" DEVICE_NAME "]: Can't construct allocator(%s)",
                          allocatorArray[i].name);

                continue;
            }

            allocator->name = allocatorArray[i].name;
#if ljdANON_FILE_FOR_ALLOCATOR
            allocator->anon_file = anon_file;
#endif
            list_add_tail(&allocator->link, &Os->allocatorList);
        }
    }

#if ljdDEBUG
    list_for_each_entry(allocator, &Os->allocatorList, link) {
        ljmkTRACE_ZONE(ljvLEVEL_WARNING, ljvZONE_OS,
                       "%s(%d) Allocator: %s",
                       __func__, __LINE__, allocator->name);
    }
#endif

    return ljvSTATUS_OK;
}

ljeSTATUS
ljkOS_FreeAllocators(ljkOS Os)
{
    ljkALLOCATOR allocator;
    ljkALLOCATOR temp;

    list_for_each_entry_safe(allocator, temp, &Os->allocatorList, link) {
        list_del(&allocator->link);

        /* Destroy allocator. */
        allocator->destructor(allocator);
    }

    _AllocatorDebugfsCleanup(Os);

    return ljvSTATUS_OK;
}

#if !ljdUSE_LINUX_SG_TABLE_API

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2, 6, 23)
static inline void
sg_set_page(struct scatterlist *sg, struct page *page,
            unsigned int len, unsigned int offset)
{
    sg->page = page;
    sg->offset = offset;
    sg->length = len;
}

static inline void
sg_mark_end(struct scatterlist *sg)
{
    (void)sg;
}
# endif

int
alloc_sg_list_from_pages(struct scatterlist **sgl, struct page **pages,
                         unsigned int n_pages, unsigned long offset,
                         unsigned long size, unsigned int *nents)
{
    unsigned int chunks;
    unsigned int i;
    unsigned int cur_page;
    struct scatterlist *s;

    chunks = 1;

    for (i = 1; i < n_pages; ++i) {
        if (page_to_pfn(pages[i]) != page_to_pfn(pages[i - 1]) + 1)
            ++chunks;
    }

    s = kzalloc(sizeof(*s) * chunks, GFP_KERNEL);
    if (unlikely(!s))
        return -ENOMEM;

    *sgl = s;
    *nents = chunks;

    cur_page = 0;

    for (i = 0; i < chunks; i++, s++) {
        unsigned long chunk_size;
        unsigned int j;

        for (j = cur_page + 1; j < n_pages; j++) {
            if (page_to_pfn(pages[j]) != page_to_pfn(pages[j - 1]) + 1)
                break;
        }

        chunk_size = ((j - cur_page) << PAGE_SHIFT) - offset;
        sg_set_page(s, pages[cur_page], min(size, chunk_size), offset);
        size -= chunk_size;
        offset = 0;
        cur_page = j;
    }

    sg_mark_end(s - 1);

    return 0;
}
#endif  /* #if !ljdUSE_LINUX_SG_TABLE_API */
