/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#include "lj_hal_kernel_linux.h"
#include "lj_hal_kernel_allocator.h"

#include <linux/pagemap.h>
#include <linux/seq_file.h>
#include <linux/mman.h>
#include <linux/slab.h>
#include <linux/mutex.h>
#include <linux/atomic.h>
#include <linux/dma-mapping.h>

#include <linux/dma-buf.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 2, 0)
#include <linux/dma-resv.h>
#endif
#include <linux/platform_device.h>

#define _LJ_OBJ_ZONE ljvZONE_OS

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 16, 0)
MODULE_IMPORT_NS(DMA_BUF);
#endif

/* Descriptor of a dma_buf imported. */
typedef struct _ljsDMABUF {
    struct dma_buf             *dmabuf;
    struct dma_buf_attachment  *attachment;
    struct sg_table            *sgt;
    unsigned long              *pagearray;

    int                         npages;
    int                         pid;
    struct list_head            list;
} ljsDMABUF;

struct allocator_priv {
    struct mutex     lock;
    struct list_head buf_list;
};

/*
 * Debugfs support.
 */
static int
dma_buf_info_show(struct seq_file *m, void *data)
{
    int ret;
    ljsDMABUF *buf_desc;
    struct dma_buf_attachment *attach_obj;
    int count = 0;
    size_t size   = 0;
    int npages = 0;
    const char *exp_name;

    ljsINFO_NODE *node = m->private;
    ljkALLOCATOR allocator = node->device;
    struct allocator_priv *priv = allocator->privateData;

    ret = mutex_lock_interruptible(&priv->lock);

    if (ret)
        return ret;

    seq_puts(m, "Attached dma-buf objects:\n");
    seq_puts(m, "   pid     fd    pages     size   exporter attached-devices\n");

    list_for_each_entry(buf_desc, &priv->buf_list, list) {
        struct dma_buf *buf_obj = buf_desc->dmabuf;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 2, 0)
        ret = dma_resv_lock_interruptible(buf_obj->resv, NULL);
#else
        ret = mutex_lock_interruptible(&buf_obj->lock);
#endif

        if (ret) {
            seq_puts(m, "ERROR locking buffer object: skipping\n");
            continue;
        }

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 10, 0)
        exp_name = buf_obj->exp_name;
#else
        exp_name = "unknown";
#endif

        seq_printf(m, "%6d %p %8d %8zu %10s",
                   buf_desc->pid, buf_desc->dmabuf, buf_desc->npages, buf_obj->size, exp_name);

        list_for_each_entry(attach_obj, &buf_obj->attachments, node) {
            seq_printf(m, " %s", dev_name(attach_obj->dev));
        }
        seq_puts(m, "\n");

        count++;
        size += buf_obj->size;
        npages += buf_desc->npages;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 2, 0)
        dma_resv_unlock(buf_obj->resv);
#else
        mutex_unlock(&buf_obj->lock);
#endif
    }

    seq_printf(m, "\nTotal %d objects, %d pages, %zu bytes\n", count, npages, size);

    mutex_unlock(&priv->lock);
    return 0;
}

static ljsINFO _InfoList[] = {
    { "bufinfo", dma_buf_info_show },
};

static void
_DebugfsInit(ljkALLOCATOR Allocator, ljkDEBUGFS_DIR Root)
{
    ljmkVERIFY_OK(ljkDEBUGFS_DIR_Init(&Allocator->debugfsDir,
                                      Root->root, "dma_buf"));

    ljmkVERIFY_OK(ljkDEBUGFS_DIR_CreateFiles(&Allocator->debugfsDir, _InfoList,
                                             ljmCOUNTOF(_InfoList), Allocator));
}

static void
_DebugfsCleanup(ljkALLOCATOR Allocator)
{
    ljmkVERIFY_OK(ljkDEBUGFS_DIR_RemoveFiles(&Allocator->debugfsDir, _InfoList, ljmCOUNTOF(_InfoList)));

    ljkDEBUGFS_DIR_Deinit(&Allocator->debugfsDir);
}

static ljeSTATUS
_DmabufAttach(ljkALLOCATOR Allocator, ljsATTACH_DESC_PTR Desc, PLINUX_MDL Mdl)
{
    ljeSTATUS status;

    ljkOS os = Allocator->os;

    struct dma_buf *dmabuf = Desc->dmaBuf.dmabuf;
    struct sg_table *sgt = NULL;
    struct dma_buf_attachment *attachment = NULL;
    int npages = 0;
    unsigned long *pagearray = NULL;
    int i, j, k = 0;
    struct scatterlist *s;
    struct allocator_priv *priv = Allocator->privateData;
    ljsDMABUF *buf_desc = NULL;
    ljtBOOL acquiredMutex = ljvFALSE;

    ljmkHEADER();

    ljmkVERIFY_OBJECT(os, ljvOBJ_OS);

    if (!dmabuf)
        ljmkONERROR(ljvSTATUS_NOT_SUPPORTED);

    get_dma_buf(dmabuf);
    attachment = dma_buf_attach(dmabuf, /* &os->device->platform->device->dev */(struct device *)Mdl->device);

    if (!attachment)
        ljmkONERROR(ljvSTATUS_NOT_SUPPORTED);

    sgt = dma_buf_map_attachment(attachment, DMA_BIDIRECTIONAL);

    if (!sgt)
        ljmkONERROR(ljvSTATUS_NOT_SUPPORTED);

    if (os->device->args.enableMmu == 0 &&
        os->iommu == ljvNULL &&
        sgt->nents != 1)
        ljmkONERROR(ljvSTATUS_NOT_SUPPORTED);

    /* Prepare page array. */
    /* Get number of pages. */
    for_each_sg(sgt->sgl, s, sgt->orig_nents, i)
        npages += (sg_dma_len(s) + PAGE_SIZE - 1) / PAGE_SIZE;

    /* Allocate page array. */
    ljmkONERROR(ljkOS_Allocate(os, npages * ljmSIZEOF(*pagearray), (ljtPOINTER *)&pagearray));

    /* Fill page array. */
    for_each_sg(sgt->sgl, s, sgt->orig_nents, i) {
        for (j = 0; j < (sg_dma_len(s) + PAGE_SIZE - 1) / PAGE_SIZE; j++)
            pagearray[k++] = sg_dma_address(s) + j * PAGE_SIZE;
    }

    /* Prepare descriptor. */
    ljmkONERROR(ljkOS_Allocate(os, sizeof(ljsDMABUF), (ljtPOINTER *)&buf_desc));

    buf_desc->dmabuf = dmabuf;
    buf_desc->pagearray = pagearray;
    buf_desc->attachment = attachment;
    buf_desc->sgt = sgt;

    /* Record in buffer list to support debugfs. */
    buf_desc->npages = npages;
    buf_desc->pid = _GetProcessID();

    ljmkONERROR(ljkOS_AcquireMutex(os, &priv->lock, ljvINFINITE));
    acquiredMutex = ljvTRUE;

    list_add(&buf_desc->list, &priv->buf_list);

    ljmkONERROR(ljkOS_ReleaseMutex(os, &priv->lock));
    acquiredMutex = ljvFALSE;

    /* Record page number. */
    Mdl->numPages = npages;

    Mdl->priv = buf_desc;

    Mdl->contiguous = (sgt->nents == 1) ? ljvTRUE : ljvFALSE;

    ljmkFOOTER_NO();
    return ljvSTATUS_OK;

OnError:
    if (pagearray) {
        ljmkOS_SAFE_FREE(os, pagearray);
    }

    if (acquiredMutex)
        ljkOS_ReleaseMutex(os, &priv->lock);

    if (sgt)
        dma_buf_unmap_attachment(attachment, sgt, DMA_BIDIRECTIONAL);

    ljmkFOOTER();
    return status;
}

static void
_DmabufFree(ljkALLOCATOR Allocator, PLINUX_MDL Mdl)
{
    ljsDMABUF *buf_desc = Mdl->priv;
    ljkOS os = Allocator->os;
    struct allocator_priv *priv = Allocator->privateData;
    ljtBOOL acquiredMutex = ljvFALSE;
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkONERROR(ljkOS_AcquireMutex(os, &priv->lock, ljvINFINITE));
    acquiredMutex = ljvTRUE;

    list_del(&buf_desc->list);

    ljmkONERROR(ljkOS_ReleaseMutex(os, &priv->lock));
    acquiredMutex = ljvFALSE;

    dma_buf_unmap_attachment(buf_desc->attachment, buf_desc->sgt, DMA_BIDIRECTIONAL);

    dma_buf_detach(buf_desc->dmabuf, buf_desc->attachment);

    dma_buf_put(buf_desc->dmabuf);

    ljkOS_Free(os, buf_desc->pagearray);

    ljkOS_Free(os, buf_desc);

OnError:
    if (acquiredMutex)
        ljkOS_ReleaseMutex(os, &priv->lock);

    return;
}

static void
_DmabufUnmapUser(ljkALLOCATOR Allocator, PLINUX_MDL Mdl, PLINUX_MDL_MAP MdlMap, ljtUINT32 Size)
{
    ljsDMABUF *buf_desc = Mdl->priv;
    ljtINT8_PTR userLogical = MdlMap->vmaAddr;

    if (unlikely(current->mm == ljvNULL))
        /* Do nothing if process is exiting. */
        return;

    userLogical -= buf_desc->sgt->sgl->offset;
    vm_munmap((unsigned long)userLogical, Mdl->numPages << PAGE_SHIFT);
}

static ljeSTATUS
_DmabufMapUser(ljkALLOCATOR Allocator, PLINUX_MDL Mdl, PLINUX_MDL_MAP MdlMap, ljtBOOL Cacheable)
{
    ljsDMABUF *buf_desc = Mdl->priv;
    ljtINT8_PTR userLogical = ljvNULL;
    ljeSTATUS status = ljvSTATUS_OK;
    struct file *fd = buf_desc->dmabuf->file;
    unsigned long flag = 0;

    flag |= (fd->f_mode & FMODE_READ ? PROT_READ : 0);
    flag |= (fd->f_mode & FMODE_WRITE ? PROT_WRITE : 0);

    userLogical = (ljtINT8_PTR)vm_mmap(fd, 0L,
                                       Mdl->numPages << PAGE_SHIFT,
                                       flag,
                                       MAP_SHARED | MAP_NORESERVE,
                                       0);

    if (IS_ERR(userLogical))
        ljmkONERROR(ljvSTATUS_OUT_OF_RESOURCES);
    userLogical += buf_desc->sgt->sgl->offset;

    MdlMap->vmaAddr = (ljtPOINTER)userLogical;
    MdlMap->cacheable = Cacheable;

OnError:
    if (ljmIS_ERROR(status) && MdlMap->vmaAddr)
        _DmabufUnmapUser(Allocator, Mdl, MdlMap, Mdl->numPages << PAGE_SHIFT);
    return status;
}

static ljeSTATUS
_DmabufMapKernel(ljkALLOCATOR Allocator, PLINUX_MDL Mdl,
                 ljtSIZE_T Offset, ljtSIZE_T Bytes, ljtPOINTER *Logical)
{
    /* Kernel doesn't access video memory. */
    return ljvSTATUS_NOT_SUPPORTED;
}

static ljeSTATUS
_DmabufUnmapKernel(ljkALLOCATOR Allocator, PLINUX_MDL Mdl, ljtPOINTER Logical)
{
    /* Kernel doesn't access video memory. */
    return ljvSTATUS_NOT_SUPPORTED;
}

static ljeSTATUS
_DmabufCache(ljkALLOCATOR Allocator, PLINUX_MDL Mdl, ljtSIZE_T Offset,
             ljtPOINTER Logical, ljtSIZE_T Bytes, ljeCACHEOPERATION Operation)
{
    ljsDMABUF *buf_desc = Mdl->priv;
    struct sg_table *sgt = buf_desc->sgt;
    enum dma_data_direction dir;
    struct device *dev = (struct device *)Mdl->device;

    switch (Operation) {
    case ljvCACHE_CLEAN:
        dir = DMA_TO_DEVICE;
        dma_sync_sg_for_device(dev, sgt->sgl, sgt->nents, dir);
        break;
    case ljvCACHE_FLUSH:
        dir = DMA_TO_DEVICE;
        dma_sync_sg_for_device(dev, sgt->sgl, sgt->nents, dir);
        dir = DMA_FROM_DEVICE;
        dma_sync_sg_for_cpu(dev, sgt->sgl, sgt->nents, dir);
        break;
    case ljvCACHE_INVALIDATE:
        dir = DMA_FROM_DEVICE;
        dma_sync_sg_for_cpu(dev, sgt->sgl, sgt->nents, dir);
        break;
    default:
        return ljvSTATUS_INVALID_ARGUMENT;
    }

    return ljvSTATUS_OK;
}

static ljeSTATUS
_DmabufPhysical(ljkALLOCATOR Allocator, PLINUX_MDL Mdl,
                unsigned long Offset, ljtPHYS_ADDR_T *Physical)
{
    ljsDMABUF *buf_desc = Mdl->priv;
    ljtUINT32 offsetInPage = Offset & ~PAGE_MASK;
    ljtUINT32 index = (ljtUINT32)(Offset / PAGE_SIZE);

    /* buf_desc->pagearray stores sg_dma_address */
    *Physical = buf_desc->pagearray[index] + offsetInPage;

    return ljvSTATUS_OK;
}

/* Default allocator operations. */
static ljsALLOCATOR_OPERATIONS DmabufAllocatorOperations = {
    .Attach      = _DmabufAttach,
    .Free        = _DmabufFree,
    .MapUser     = _DmabufMapUser,
    .UnmapUser   = _DmabufUnmapUser,
    .MapKernel   = _DmabufMapKernel,
    .UnmapKernel = _DmabufUnmapKernel,
    .Cache       = _DmabufCache,
    .Physical    = _DmabufPhysical,
    .GetDmaAddress = _DmabufPhysical,
};

static void
_DmabufAllocatorDestructor(ljsALLOCATOR *Allocator)
{
    _DebugfsCleanup(Allocator);

    ljkOS_ZeroMemory(Allocator->privateData, sizeof(struct allocator_priv));
    kfree(Allocator->privateData);

    kfree(Allocator);
}

/* Default allocator entry. */
ljeSTATUS
_DmabufAlloctorInit(ljkOS Os, ljsDEBUGFS_DIR *Parent, ljkALLOCATOR *Allocator)
{
    ljeSTATUS status;
    ljkALLOCATOR allocator;
    struct allocator_priv *priv = NULL;

    priv = kmalloc(sizeof(*priv), GFP_KERNEL | ljdNOWARN);

    if (!priv)
        ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);

    mutex_init(&priv->lock);
    INIT_LIST_HEAD(&priv->buf_list);

    ljmkONERROR(ljkALLOCATOR_Construct(Os, &DmabufAllocatorOperations, &allocator));

    allocator->capability = ljvALLOC_FLAG_DMABUF | ljvALLOC_FLAG_DMABUF_EXPORTABLE;

    /* Register private data. */
    allocator->privateData = priv;
    allocator->destructor = _DmabufAllocatorDestructor;

    _DebugfsInit(allocator, Parent);

    *Allocator = allocator;

    return ljvSTATUS_OK;

OnError:
    kfree(priv);

    return status;
}
