/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#include "lj_hal_kernel_linux.h"
#include "lj_hal_kernel_allocator.h"
#include <linux/pagemap.h>
#include <linux/seq_file.h>
#include <linux/mman.h>
#include <linux/atomic.h>
#include <linux/dma-mapping.h>
#include <linux/slab.h>
#if defined(CONFIG_X86) && (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 12, 0))
# include <asm/set_memory.h>
#endif
#include "lj_hal_kernel_platform.h"

#define _LJ_OBJ_ZONE ljvZONE_OS

#define ljdDISCRETE_PAGES 0

#ifndef MAX_ORDER
#define MAX_ORDER MAX_PAGE_ORDER
#endif

struct gfp_alloc {
    atomic_t low;
    atomic_t high;
};

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 24)
struct sg_table {
    struct scatterlist       *sgl;
    unsigned int              nents;
    unsigned int              orig_nents;
};
#endif

struct gfp_mdl_priv {
    int contiguous;

    union {
        /* Pointer to a array of pages. */
        struct {
            struct page      *contiguousPages;
            dma_addr_t        dma_addr;
            int               exact;
        };

        struct {
            /* Pointer to a array of pointers to page. */
            struct page     **nonContiguousPages;

            struct page     **Pages1M;
            int               numPages1M;
            int              *isExact;
            struct sg_table   sgt;
        };
    };

    ljsPLATFORM *platform;
};

/*******************************************************************************
 ************************* GFP Allocator Debugfs *******************************
 *******************************************************************************/

static int lj_usage_show(struct seq_file *m, void *data)
{
    ljsINFO_NODE *node = m->private;
    ljkALLOCATOR Allocator = node->device;
    struct gfp_alloc *priv = Allocator->privateData;
    long long low = (long long)atomic_read(&priv->low);
    long long high = (long long)atomic_read(&priv->high);

    seq_puts(m, "type        n pages        bytes\n");
    seq_printf(m, "normal   %10llu %12llu\n", low, low * PAGE_SIZE);
    seq_printf(m, "HighMem  %10llu %12llu\n", high, high * PAGE_SIZE);

    return 0;
}

static ljsINFO InfoList[] = {
    { "usage", lj_usage_show },
};

static void
_GFPAllocatorDebugfsInit(ljkALLOCATOR Allocator, ljkDEBUGFS_DIR Root)
{
    ljmkVERIFY_OK(ljkDEBUGFS_DIR_Init(&Allocator->debugfsDir, Root->root, "gfp"));

    ljmkVERIFY_OK(ljkDEBUGFS_DIR_CreateFiles(&Allocator->debugfsDir, InfoList,
                                             ljmCOUNTOF(InfoList), Allocator));
}

static void
_GFPAllocatorDebugfsCleanup(ljkALLOCATOR Allocator)
{
    ljmkVERIFY_OK(ljkDEBUGFS_DIR_RemoveFiles(&Allocator->debugfsDir,
                                             InfoList, ljmCOUNTOF(InfoList)));

    ljkDEBUGFS_DIR_Deinit(&Allocator->debugfsDir);
}

static void
_NonContiguousFree(struct page **Pages, ljtSIZE_T NumPages)
{
    ljtSIZE_T i;

    ljmkHEADER_ARG("Pages=%p, NumPages=%zx", Pages, NumPages);

    ljmkASSERT(Pages != ljvNULL);

    for (i = 0; i < NumPages; i++)
        __free_page(Pages[i]);

    if (is_vmalloc_addr(Pages))
        vfree(Pages);
    else
        kfree(Pages);

    ljmkFOOTER_NO();
}

static ljeSTATUS
_NonContiguousAlloc(struct gfp_mdl_priv *MdlPriv, ljtSIZE_T NumPages, ljtUINT32 Gfp)
{
    struct page **pages;
    struct page *p;
    ljtSIZE_T i, size;

    ljmkHEADER_ARG("NumPages=%zx", NumPages);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0)
    if (NumPages > totalram_pages()) {
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 32)
    if (NumPages > totalram_pages) {
#else
    if (NumPages > num_physpages) {
#endif
        ljmkFOOTER_NO();
        return ljvSTATUS_INVALID_ARGUMENT;
    }

    size = NumPages * sizeof(struct page *);

    pages = kmalloc(size, GFP_KERNEL | ljdNOWARN);

    if (!pages) {
        pages = vmalloc(size);

        if (!pages) {
            ljmkFOOTER_NO();
            return ljvSTATUS_OUT_OF_MEMORY;
        }
    }

    for (i = 0; i < NumPages; i++) {
        p = alloc_page(Gfp);

        if (!p) {
            _NonContiguousFree(pages, i);
            ljmkFOOTER_NO();
            return ljvSTATUS_OUT_OF_MEMORY;
        }

#if ljdDISCRETE_PAGES
        if (i != 0) {
            if (page_to_pfn(pages[i - 1]) == page_to_pfn(p) - 1) {
                /* Replaced page. */
                struct page *l = p;

                /* Allocate a page which is not contiguous to previous one. */
                p = alloc_page(Gfp);

                /* Give replaced page back. */
                __free_page(l);

                if (!p) {
                    _NonContiguousFree(pages, i);
                    ljmkFOOTER_NO();
                    return ljvSTATUS_OUT_OF_MEMORY;
                }
            }
        }
#endif

        pages[i] = p;
    }

    MdlPriv->nonContiguousPages = pages;

    ljmkFOOTER_ARG("pages=%p", pages);
    return ljvSTATUS_OK;
}

static void
_NonContiguous1MPagesFree(struct gfp_mdl_priv *MdlPriv, ljtUINT32 NumPages1M)
{
    ljtINT i;

    if (MdlPriv->Pages1M && MdlPriv->isExact) {
        for (i = 0; i < NumPages1M && MdlPriv->Pages1M[i]; i++) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 27)
            if (MdlPriv->isExact[i] == ljvTRUE)
                free_pages_exact(page_address(MdlPriv->Pages1M[i]), ljd1M_PAGE_SIZE);
            else
#endif
                __free_pages(MdlPriv->Pages1M[i], get_order(ljd1M_PAGE_SIZE));
        }
    }

    if (MdlPriv->Pages1M) {
        if (is_vmalloc_addr(MdlPriv->Pages1M))
            vfree(MdlPriv->Pages1M);
        else
            kfree(MdlPriv->Pages1M);
        MdlPriv->Pages1M = ljvNULL;
    }

    if (MdlPriv->isExact) {
        if (is_vmalloc_addr(MdlPriv->isExact))
            vfree(MdlPriv->isExact);
        else
            kfree(MdlPriv->isExact);
    }

    if (MdlPriv->nonContiguousPages) {
        if (is_vmalloc_addr(MdlPriv->nonContiguousPages))
            vfree(MdlPriv->nonContiguousPages);
        else
            kfree(MdlPriv->nonContiguousPages);
        MdlPriv->nonContiguousPages = ljvNULL;
    }
}

static ljeSTATUS
_NonContiguous1MPagesAlloc(struct gfp_mdl_priv *MdlPriv, ljtSIZE_T *NumPages, ljtUINT32 Gfp)
{
    ljeSTATUS status;
    size_t numPages1M, num, size;
    struct page **pages;
    struct page *page;
    void *addr = NULL;
    ljtINT i, j;

    MdlPriv->numPages1M = 0;

    numPages1M = ((*NumPages << PAGE_SHIFT) + (ljd1M_PAGE_SIZE - 1)) >> ljd1M_PAGE_SHIFT;

    *NumPages = (numPages1M << ljd1M_PAGE_SHIFT) >> PAGE_SHIFT;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0)
    if (*NumPages > totalram_pages())
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 32)
    if (*NumPages > totalram_pages)
#else
    if (*NumPages > num_physpages)
#endif
        return ljvSTATUS_INVALID_ARGUMENT;

    num = ljd1M_PAGE_SIZE / PAGE_SIZE;

    size = numPages1M * sizeof(struct page *);
    MdlPriv->Pages1M = kmalloc(size, GFP_KERNEL | ljdNOWARN);
    if (!MdlPriv->Pages1M) {
        MdlPriv->Pages1M = vmalloc(size);

        if (!MdlPriv->Pages1M)
            ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);
    }

    size = numPages1M * sizeof(int);
    MdlPriv->isExact = kmalloc(size, GFP_KERNEL | ljdNOWARN);
    if (!MdlPriv->isExact) {
        MdlPriv->isExact = vmalloc(size);
        if (!MdlPriv->isExact)
            ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);
    }
    memset(MdlPriv->isExact, 0, size);

    size = *NumPages * sizeof(struct page *);
    pages = kmalloc(size, GFP_KERNEL | ljdNOWARN);
    if (!pages) {
        pages = vmalloc(size);
        if (!pages)
            ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);
    }
    MdlPriv->nonContiguousPages = pages;

    for (i = 0; i < numPages1M; i++) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 27)
        addr = alloc_pages_exact(ljd1M_PAGE_SIZE, (Gfp & ~__GFP_HIGHMEM) | __GFP_NORETRY);

        MdlPriv->Pages1M[i] = addr ? virt_to_page(addr) : ljvNULL;
        if (MdlPriv->Pages1M[i])
            MdlPriv->isExact[i] = ljvTRUE;
#endif

        if (MdlPriv->Pages1M[i] == ljvNULL) {
            int order = get_order(ljd1M_PAGE_SIZE);

            if (order >= MAX_ORDER)
                ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);

            MdlPriv->Pages1M[i] = alloc_pages(Gfp, order);
        }

        if (MdlPriv->Pages1M[i] == ljvNULL)
            ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);

        MdlPriv->numPages1M += 1;

        for (j = 0; j < num; j++) {
            page = nth_page(MdlPriv->Pages1M[i], j);
            pages[i * num + j] = page;
        }
    }

    return ljvSTATUS_OK;
OnError:
    _NonContiguous1MPagesFree(MdlPriv, MdlPriv->numPages1M);

    return status;
}

/****************************************************************************
 *********************** GFP Allocator **************************************
 ****************************************************************************/

static ljeSTATUS
_GFPAlloc(ljkALLOCATOR Allocator, PLINUX_MDL Mdl, ljtSIZE_T NumPages, ljtUINT32 Flags)
{
    ljeSTATUS status;
    ljtSIZE_T i = 0;
    ljtBOOL contiguous = Flags & ljvALLOC_FLAG_CONTIGUOUS;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 6, 0)
    u32 normal_gfp = __GFP_HIGH | GFP_ATOMIC | __GFP_NORETRY | ljdNOWARN;
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(4, 4, 0)
    u32 normal_gfp = __GFP_HIGH | __GFP_ATOMIC | __GFP_NORETRY | ljdNOWARN;
#else
    u32 normal_gfp = __GFP_HIGH | __GFP_NORETRY | ljdNOWARN;
#endif
    u32 gfp = (contiguous ? normal_gfp : GFP_KERNEL) | __GFP_HIGHMEM | ljdNOWARN;

    struct gfp_alloc *priv = (struct gfp_alloc *)Allocator->privateData;
    struct device *dev = (struct device *)Mdl->device;
    struct gfp_mdl_priv *mdlPriv = ljvNULL;
    struct page **pages;
    int result;
    int low  = 0;
    int high = 0;

    ljmkHEADER_ARG("Allocator=%p Mdl=%p NumPages=%zu Flags=0x%x",
                   Allocator, Mdl, NumPages, Flags);

#ifdef ljdSYS_FREE_MEMORY_LIMIT
    if ((Flags & ljvALLOC_FLAG_MEMLIMIT) && !contiguous) {
        long freeram = 0;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 14, 0)
        freeram = global_zone_page_state(NR_FREE_PAGES);
#ifdef CONFIG_CMA
        freeram -= global_zone_page_state(NR_FREE_CMA_PAGES);
#  endif
# else
        struct sysinfo temsysinfo;

        si_meminfo(&temsysinfo);
        freeram = temsysinfo.freeram;
# endif
        if (freeram < NumPages || ((freeram - NumPages) < ljdSYS_FREE_MEMORY_LIMIT))
            ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);
    }
#endif

    mdlPriv = kzalloc(sizeof(*mdlPriv), GFP_KERNEL | __GFP_NORETRY);

    if (!mdlPriv)
        ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);

#if defined(CONFIG_ZONE_DMA32) && LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 37)
    if ((Flags & ljvALLOC_FLAG_4GB_ADDR) ||
        (Allocator->os->device->platform->flagBits & ljvPLATFORM_FLAG_LIMIT_4G_ADDRESS)) {
        /* remove __GFP_HIGHMEM bit, add __GFP_DMA32 bit */
        gfp &= ~__GFP_HIGHMEM;
        gfp |= __GFP_DMA32;
    }
#else
    if (Flags & ljvALLOC_FLAG_4GB_ADDR ||
        (Allocator->os->device->platform->flagBits & ljvPLATFORM_FLAG_LIMIT_4G_ADDRESS)) {
        /* remove __GFP_HIGHMEM bit, add __GFP_DMA bit */
        gfp &= ~__GFP_HIGHMEM;
        gfp |= __GFP_DMA;
    }

#endif

    if ((Flags & ljvALLOC_FLAG_NON_CONTIGUOUS) && (Flags & ljvALLOC_FLAG_1M_PAGES))
        Mdl->pageUnit1M = ljvTRUE;
    else
        Mdl->pageUnit1M = ljvFALSE;

    if (contiguous) {
        size_t bytes = NumPages << PAGE_SHIFT;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 27)
        void *addr = NULL;

        addr = alloc_pages_exact(bytes, (gfp & ~__GFP_HIGHMEM) | __GFP_NORETRY);

        mdlPriv->contiguousPages = addr ? virt_to_page(addr) : ljvNULL;

        if (mdlPriv->contiguousPages)
            mdlPriv->exact = ljvTRUE;
#endif

        if (mdlPriv->contiguousPages == ljvNULL) {
            int order = get_order(bytes);

            if (order >= MAX_ORDER) {
                status = ljvSTATUS_OUT_OF_MEMORY;
                goto OnError;
            }

            mdlPriv->contiguousPages = alloc_pages(gfp, order);
        }

        if (mdlPriv->contiguousPages == ljvNULL)
            ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);

        mdlPriv->dma_addr = dma_map_page(dev, mdlPriv->contiguousPages,
                                         0, NumPages * PAGE_SIZE, DMA_BIDIRECTIONAL);

//        if (dma_mapping_error(dev, mdlPriv->dma_addr)) {
        if (!mdlPriv->dma_addr) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 27)
            if (mdlPriv->exact)
                free_pages_exact(page_address(mdlPriv->contiguousPages), bytes);
            else
#endif
                __free_pages(mdlPriv->contiguousPages, get_order(bytes));

            ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);
        }

#if defined(CONFIG_X86)
        if (!PageHighMem(mdlPriv->contiguousPages)) {
#if ljdENABLE_BUFFERABLE_VIDEO_MEMORY
#if ljdENABLE_NONCACHE_COMMANDBUF
            if (Mdl->type == ljvVIDMEM_TYPE_COMMAND)
                if (set_memory_uc((unsigned long)page_address(mdlPriv->contiguousPages), NumPages) != 0)
                    pr_warn("%s(%d): failed to set_memory_uc\n", __func__, __LINE__);
            else
#  endif
                if (set_memory_wc((unsigned long)page_address(mdlPriv->contiguousPages), NumPages) != 0)
                    pr_warn("%s(%d): failed to set_memory_wc\n", __func__, __LINE__);
# else
            if (set_memory_uc((unsigned long)page_address(mdlPriv->contiguousPages), NumPages) != 0)
                pr_warn("%s(%d): failed to set_memory_uc\n", __func__, __LINE__);
# endif
        }
#endif
    } else {
        if (Mdl->pageUnit1M) {
            ljmkONERROR(_NonContiguous1MPagesAlloc(mdlPriv, &NumPages, gfp));
        } else {
            if ((Flags & ljvALLOC_FLAG_4GB_ADDR) ||
                (Allocator->os->device->platform->flagBits & ljvPLATFORM_FLAG_LIMIT_4G_ADDRESS)) {

                ljmkONERROR(_NonContiguousAlloc(mdlPriv, NumPages, gfp));
            } else {
                status = _NonContiguousAlloc(mdlPriv, NumPages, normal_gfp);

                if (ljmIS_ERROR(status))
                    ljmkONERROR(_NonContiguousAlloc(mdlPriv, NumPages, gfp));
            }
        }

#if ljdUSE_LINUX_SG_TABLE_API
        result = sg_alloc_table_from_pages(&mdlPriv->sgt, mdlPriv->nonContiguousPages,
                                           NumPages, 0, NumPages << PAGE_SHIFT, normal_gfp);

#else
        result = alloc_sg_list_from_pages(&mdlPriv->sgt.sgl, mdlPriv->nonContiguousPages,
                                          NumPages, 0, NumPages << PAGE_SHIFT, &mdlPriv->sgt.nents);
#endif
        mdlPriv->sgt.orig_nents = mdlPriv->sgt.nents;
        if (result < 0) {
            if (Mdl->pageUnit1M)
                _NonContiguous1MPagesFree(mdlPriv, mdlPriv->numPages1M);
            else
                _NonContiguousFree(mdlPriv->nonContiguousPages, NumPages);

            ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);
        }

        result = dma_map_sg(dev, mdlPriv->sgt.sgl, mdlPriv->sgt.nents, DMA_BIDIRECTIONAL);
        if (unlikely(result <= 0)) {
            if (Mdl->pageUnit1M)
                _NonContiguous1MPagesFree(mdlPriv, mdlPriv->numPages1M);
            else
                _NonContiguousFree(mdlPriv->nonContiguousPages, NumPages);

#if ljdUSE_LINUX_SG_TABLE_API
            sg_free_table(&mdlPriv->sgt);
#else
            kfree(mdlPriv->sgt.sgl);
#endif
            ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);
        } else if (unlikely(result != mdlPriv->sgt.nents))
            mdlPriv->sgt.nents = result;

#if defined(CONFIG_X86)
#if ljdENABLE_BUFFERABLE_VIDEO_MEMORY
#if ljdENABLE_NONCACHE_COMMANDBUF
        if (Mdl->type == ljvVIDMEM_TYPE_COMMAND)
            if (set_pages_array_uc(mdlPriv->nonContiguousPages, NumPages))
                pr_warn("%s(%d): failed to set_pages_array_uc\n", __func__, __LINE__);
        else
#  endif
            if (set_pages_array_wc(mdlPriv->nonContiguousPages, NumPages))
                pr_warn("%s(%d): failed to set_pages_array_wc\n", __func__, __LINE__);
# else
        if (set_pages_array_uc(mdlPriv->nonContiguousPages, NumPages))
            pr_warn("%s(%d): failed to set_pages_array_uc\n", __func__, __LINE__);
# endif
#endif
    }

    pages = kmalloc_array(NumPages, sizeof(struct page *), GFP_KERNEL | ljdNOWARN);
    if (!pages)
        ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);

    for (i = 0; i < NumPages; i++) {
        struct page *page;

        if (contiguous)
            page = nth_page(mdlPriv->contiguousPages, i);
        else
            page = mdlPriv->nonContiguousPages[i];

        pages[i] = page;

        SetPageReserved(page);

        if (PageHighMem(page))
            high++;
        else
            low++;
    }

    mdlPriv->platform = Allocator->os->device->platform;
    mdlPriv->contiguous = contiguous;
    atomic_add(low, &priv->low);
    atomic_add(high, &priv->high);

    Mdl->priv = mdlPriv;
    Mdl->numPages = NumPages;
    Mdl->pages = pages;

    ljmkFOOTER_NO();
    return ljvSTATUS_OK;

OnError:
    kfree(mdlPriv);

    ljmkFOOTER();
    return status;
}

static ljeSTATUS
_GFPGetSGT(ljkALLOCATOR Allocator, PLINUX_MDL Mdl, ljtSIZE_T Offset, ljtSIZE_T Bytes, ljtPOINTER *SGT)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 6, 0)
    struct page **pages = ljvNULL;
    struct sg_table *sgt = NULL;
    struct gfp_mdl_priv *mdlPriv = (struct gfp_mdl_priv *)Mdl->priv;

    ljeSTATUS status = ljvSTATUS_OK;
    ljtSIZE_T offset = Offset & ~PAGE_MASK;  /* Offset to the first page */
    ljtSIZE_T skipPages = Offset >> PAGE_SHIFT; /* skipped pages */
    ljtSIZE_T numPages = (PAGE_ALIGN(Offset + Bytes) >> PAGE_SHIFT) - skipPages;
    ljtSIZE_T i;

    ljmkASSERT(Offset + Bytes <= Mdl->numPages << PAGE_SHIFT);

    /* TODO: now alloc func will save the page array, here is not necessary */
    if (mdlPriv->contiguous) {
        pages = kmalloc_array(numPages, sizeof(struct page *), GFP_KERNEL | ljdNOWARN);
        if (!pages)
            ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);

        for (i = 0; i < numPages; ++i)
            pages[i] = nth_page(mdlPriv->contiguousPages, i + skipPages);
    } else {
        pages = &mdlPriv->nonContiguousPages[skipPages];
    }

    sgt = kmalloc(sizeof(*sgt), GFP_KERNEL | ljdNOWARN);
    if (!sgt)
        ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);

    if (sg_alloc_table_from_pages(sgt, pages, numPages, offset, Bytes, GFP_KERNEL) < 0)
        ljmkONERROR(ljvSTATUS_GENERIC_IO);

    *SGT = (ljtPOINTER)sgt;

OnError:
    if (mdlPriv->contiguous && pages)
        kfree(pages);

    if (ljmIS_ERROR(status) && sgt)
        kfree(sgt);

    return status;
#else
    return ljvSTATUS_NOT_SUPPORTED;
#endif
}

static void
_GFPFree(ljkALLOCATOR Allocator, PLINUX_MDL Mdl)
{
    ljtSIZE_T i;
    struct page *page;
    struct gfp_alloc *priv = (struct gfp_alloc *)Allocator->privateData;
    struct gfp_mdl_priv *mdlPriv = Mdl->priv;
    struct device *dev = (struct device *)Mdl->device;
    int low = 0;
    int high = 0;

    if (mdlPriv->contiguous) {
        dma_unmap_page(dev, mdlPriv->dma_addr,
                       Mdl->numPages << PAGE_SHIFT, DMA_FROM_DEVICE);
    } else {
        dma_unmap_sg(dev, mdlPriv->sgt.sgl,
                     mdlPriv->sgt.orig_nents, DMA_FROM_DEVICE);

#if ljdUSE_LINUX_SG_TABLE_API
        sg_free_table(&mdlPriv->sgt);
#else
        kfree(mdlPriv->sgt.sgl);
#endif
    }

    kfree(Mdl->pages);

    for (i = 0; i < Mdl->numPages; i++) {
        if (mdlPriv->contiguous)
            page = nth_page(mdlPriv->contiguousPages, i);
        else
            page = mdlPriv->nonContiguousPages[i];

        ClearPageReserved(page);

        if (PageHighMem(page))
            high++;
        else
            low++;
    }

    atomic_sub(low, &priv->low);
    atomic_sub(high, &priv->high);

    if (mdlPriv->contiguous) {
#if defined(CONFIG_X86)
        if (!PageHighMem(mdlPriv->contiguousPages))
            set_memory_wb((unsigned long)page_address(mdlPriv->contiguousPages), Mdl->numPages);
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 27)
        if (mdlPriv->exact == ljvTRUE)
            free_pages_exact(page_address(mdlPriv->contiguousPages), Mdl->numPages * PAGE_SIZE);
        else
#endif
            __free_pages(mdlPriv->contiguousPages, get_order(Mdl->numPages * PAGE_SIZE));
    } else {
#if defined(CONFIG_X86)
        set_pages_array_wb(mdlPriv->nonContiguousPages, Mdl->numPages);
#endif

        if (Mdl->pageUnit1M)
            _NonContiguous1MPagesFree(mdlPriv, mdlPriv->numPages1M);
        else
            _NonContiguousFree(mdlPriv->nonContiguousPages, Mdl->numPages);
    }

    kfree(Mdl->priv);
}

static ljeSTATUS
_GFPMmap(ljkALLOCATOR Allocator, PLINUX_MDL Mdl, ljtBOOL Cacheable,
         ljtSIZE_T skipPages, ljtSIZE_T numPages, struct vm_area_struct *vma)
{
    struct gfp_mdl_priv *mdlPriv = (struct gfp_mdl_priv *)Mdl->priv;
    ljsPLATFORM *platform = mdlPriv->platform;
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkHEADER_ARG("Allocator=%p Mdl=%p vma=%p", Allocator, Mdl, vma);

#if defined(CONFIG_X86)
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0)
    vma->vm_flags |= ljdVM_FLAGS;
#else
    vm_flags_set(vma, ljdVM_FLAGS);
#endif
#endif
    if (Cacheable == ljvFALSE) {
        /* Make this mapping non-cached. */
#if ljdENABLE_BUFFERABLE_VIDEO_MEMORY
#if ljdENABLE_NONCACHE_COMMANDBUF
        if (Mdl->type == ljvVIDMEM_TYPE_COMMAND)
            vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
        else
# endif
            vma->vm_page_prot = pgprot_writecombine(vma->vm_page_prot);
#else
        vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
#endif
    }
#if defined(CONFIG_X86)
    else {
        if (mdlPriv->contiguous)
            set_memory_wb((unsigned long)page_address(mdlPriv->contiguousPages) + (skipPages << PAGE_SHIFT), numPages);
        else
            set_pages_array_wb(mdlPriv->nonContiguousPages + skipPages, numPages);
    }
#endif

    if (platform && platform->ops->adjustProt)
        platform->ops->adjustProt(vma);

    ljmkASSERT(skipPages + numPages <= Mdl->numPages);

    /* Now map all the vmalloc pages to this user address. */
    if (mdlPriv->contiguous) {
        /* map kernel memory to user space.. */
        if (remap_pfn_range(vma, vma->vm_start,
                            page_to_pfn(mdlPriv->contiguousPages) + skipPages,
                            numPages << PAGE_SHIFT,
#if defined(CONFIG_X86)
                            vma->vm_page_prot) < 0)
#else/* test ok on ft */
                            vm_get_page_prot(vma->vm_flags)) < 0)
#endif
        {
            ljmkTRACE_ZONE(ljvLEVEL_INFO, ljvZONE_OS,
                           "%s(%d): remap_pfn_range error.",
                           __func__, __LINE__);

            ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);
        }
    } else {
        ljtSIZE_T i;
        unsigned long start = vma->vm_start;

        for (i = 0; i < numPages; ++i) {
            unsigned long pfn = page_to_pfn(mdlPriv->nonContiguousPages[i + skipPages]);

            if (remap_pfn_range(vma,
                                start,
                                pfn,
                                PAGE_SIZE,
#if defined(CONFIG_X86)
                                vma->vm_page_prot) < 0)
#else/* test ok on ft */
                                vm_get_page_prot(vma->vm_flags)) < 0)
#endif
            {
                ljmkTRACE(ljvLEVEL_ERROR,
                          "%s(%d): remap_pfn_range error.",
                          __func__, __LINE__);

                ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);
            }

            start += PAGE_SIZE;
        }
    }

OnError:
    ljmkFOOTER();
    return status;
}

static void
_GFPUnmapUser(ljkALLOCATOR Allocator, PLINUX_MDL Mdl, PLINUX_MDL_MAP MdlMap, ljtUINT32 Size)
{
    MdlMap->cacheable = ljvFALSE;

    if (unlikely(current->mm == ljvNULL))
        /* Do nothing if process is exiting. */
        return;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 5, 0)
    if (vm_munmap((unsigned long)MdlMap->vmaAddr, Size) < 0) {
        ljmkTRACE_ZONE(ljvLEVEL_WARNING, ljvZONE_OS,
                       "%s(%d): vm_munmap failed",
                       __func__, __LINE__);
    }
#else
    down_write(&current_mm_mmap_sem);
    if (do_munmap(current->mm, (unsigned long)MdlMap->vmaAddr, Size) < 0) {
        ljmkTRACE_ZONE(ljvLEVEL_WARNING, ljvZONE_OS,
                       "%s(%d): do_munmap failed",
                       __func__, __LINE__);
    }
    up_write(&current_mm_mmap_sem);
#endif

    MdlMap->vma = NULL;
    MdlMap->vmaAddr = NULL;
}

static ljeSTATUS
_GFPMapUser(ljkALLOCATOR Allocator, PLINUX_MDL Mdl, PLINUX_MDL_MAP MdlMap, ljtBOOL Cacheable)
{
    ljtPOINTER userLogical = ljvNULL;
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkHEADER_ARG("Allocator=%p Mdl=%p Cacheable=%d", Allocator, Mdl, Cacheable);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 4, 0)
#if ljdANON_FILE_FOR_ALLOCATOR
    userLogical = (ljtPOINTER)vm_mmap(Allocator->anon_file,
# else
    userLogical = (ljtPOINTER)vm_mmap(NULL,
# endif
                                      0L,
                                      Mdl->numPages * PAGE_SIZE, PROT_READ | PROT_WRITE,
                                      MAP_SHARED | MAP_NORESERVE,
                                      0);
#else
    down_write(&current_mm_mmap_sem);
    userLogical = (ljtPOINTER)do_mmap_pgoff(NULL, 0L,
                                            Mdl->numPages * PAGE_SIZE,
                                            PROT_READ | PROT_WRITE, MAP_SHARED, 0);
    up_write(&current_mm_mmap_sem);
#endif

    ljmkTRACE_ZONE(ljvLEVEL_INFO, ljvZONE_OS,
                   "%s(%d): vmaAddr->%p for phys_addr->%p",
                   __func__, __LINE__, userLogical, Mdl);

    if (IS_ERR(userLogical)) {
        ljmkTRACE_ZONE(ljvLEVEL_INFO, ljvZONE_OS,
                       "%s(%d): do_mmap_pgoff error",
                       __func__, __LINE__);

        userLogical = ljvNULL;

        ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);
    }

    down_write(&current_mm_mmap_sem);

    do {
        struct vm_area_struct *vma = find_vma(current->mm, (unsigned long)userLogical);

        if (vma == ljvNULL) {
            ljmkTRACE_ZONE(ljvLEVEL_INFO, ljvZONE_OS,
                           "%s(%d): find_vma error",
                           __func__, __LINE__);

            ljmkERR_BREAK(ljvSTATUS_OUT_OF_RESOURCES);
        }

        ljmkERR_BREAK(_GFPMmap(Allocator, Mdl, Cacheable, 0, Mdl->numPages, vma));
        MdlMap->vma = vma;
    } while (ljvFALSE);

    up_write(&current_mm_mmap_sem);

    if (ljmIS_SUCCESS(status)) {
        MdlMap->vmaAddr = userLogical;
        MdlMap->cacheable = Cacheable;
    }

OnError:
    if (ljmIS_ERROR(status) && userLogical) {
        MdlMap->vmaAddr = userLogical;
        _GFPUnmapUser(Allocator, Mdl, MdlMap, Mdl->numPages * PAGE_SIZE);
    }
    ljmkFOOTER();
    return status;
}

static ljeSTATUS
_GFPMapKernel(ljkALLOCATOR Allocator, PLINUX_MDL Mdl,
              ljtSIZE_T Offset, ljtSIZE_T Bytes, ljtPOINTER *Logical)
{
    void *addr = 0;
    ljtSIZE_T numPages = Mdl->numPages;
    struct gfp_mdl_priv *mdlPriv = Mdl->priv;
    unsigned long pgoff = (Offset >> PAGE_SHIFT);
    struct page **pages;
    ljtBOOL free = ljvFALSE;
    pgprot_t pgprot;

    if (Offset + Bytes > (numPages << PAGE_SHIFT))
        return ljvSTATUS_INVALID_ARGUMENT;

    numPages = ((Offset & ~PAGE_MASK) + Bytes + (PAGE_SIZE - 1)) >> PAGE_SHIFT;

    if (mdlPriv->contiguous) {
        ljtSIZE_T i;

        pages = kmalloc_array(numPages, sizeof(struct page *), GFP_KERNEL | ljdNOWARN);

        if (!pages)
            return ljvSTATUS_OUT_OF_MEMORY;

        for (i = 0; i < numPages; i++)
            pages[i] = nth_page(mdlPriv->contiguousPages, i + pgoff);

        free = ljvTRUE;
    } else {
        pages = &mdlPriv->nonContiguousPages[pgoff];
    }

    /* ioremap() can't work on system memory since 2.6.38. */
    if (Mdl->cacheable) {
        pgprot = PAGE_KERNEL;
    } else {
#if ljdENABLE_BUFFERABLE_VIDEO_MEMORY
#if ljdENABLE_NONCACHE_COMMANDBUF
        if (Mdl->type == ljvVIDMEM_TYPE_COMMAND)
            pgprot = pgprot_noncached(PAGE_KERNEL);
        else
# endif
            pgprot = pgprot_writecombine(PAGE_KERNEL);
#else
        pgprot = pgprot_noncached(PAGE_KERNEL);
#endif
    }

    addr = vmap(pages, numPages, 0, pgprot);

    if (free)
        kfree(pages);

    if (addr) {
        /* Append offset in page. */
        *Logical = (uint8_t *)addr + (Offset & ~PAGE_MASK);
        return ljvSTATUS_OK;
    } else {
        return ljvSTATUS_OUT_OF_MEMORY;
    }
}

static ljeSTATUS
_GFPUnmapKernel(ljkALLOCATOR Allocator, PLINUX_MDL Mdl, ljtPOINTER Logical)
{
    vunmap((void *)((uintptr_t)Logical & PAGE_MASK));

    return ljvSTATUS_OK;
}

static ljeSTATUS
_GFPCache(ljkALLOCATOR Allocator, PLINUX_MDL Mdl, ljtSIZE_T Offset,
          ljtPOINTER Logical, ljtSIZE_T Bytes, ljeCACHEOPERATION Operation)
{
    struct gfp_mdl_priv *mdlPriv = Mdl->priv;
    struct device *dev = (struct device *)Mdl->device;
    enum dma_data_direction dir;
    dma_addr_t dma_addr = (mdlPriv->dma_addr + Offset) & PAGE_MASK;
    ljtSIZE_T bytes = (mdlPriv->dma_addr + Offset + Bytes) - dma_addr;
    ljtINT numPages = GetPageCount(bytes, 0);

    switch (Operation) {
    case ljvCACHE_CLEAN:
        dir = DMA_TO_DEVICE;

        if (mdlPriv->contiguous) {
            dma_sync_single_for_device(dev, dma_addr,
                                       numPages << PAGE_SHIFT, dir);
        } else {
            dma_sync_sg_for_device(dev, mdlPriv->sgt.sgl,
                                   mdlPriv->sgt.nents, dir);
        }

        break;
    case ljvCACHE_FLUSH:
        dir = DMA_TO_DEVICE;

        if (mdlPriv->contiguous) {
            dma_sync_single_for_device(dev, dma_addr,
                                       numPages << PAGE_SHIFT, dir);
        } else {
            dma_sync_sg_for_device(dev, mdlPriv->sgt.sgl,
                                   mdlPriv->sgt.nents, dir);
        }

        dir = DMA_FROM_DEVICE;

        if (mdlPriv->contiguous) {
            dma_sync_single_for_cpu(dev, dma_addr,
                                    numPages << PAGE_SHIFT, dir);
        } else {
            dma_sync_sg_for_cpu(dev, mdlPriv->sgt.sgl,
                                mdlPriv->sgt.nents, dir);
        }

        break;
    case ljvCACHE_INVALIDATE:
        dir = DMA_FROM_DEVICE;

        if (mdlPriv->contiguous) {
            dma_sync_single_for_cpu(dev, dma_addr,
                                    numPages << PAGE_SHIFT, dir);
        } else {
            dma_sync_sg_for_cpu(dev, mdlPriv->sgt.sgl,
                                mdlPriv->sgt.nents, dir);
        }

        break;
    default:
        return ljvSTATUS_INVALID_ARGUMENT;
    }

    return ljvSTATUS_OK;
}

static ljeSTATUS
_GFPPhysical(ljkALLOCATOR Allocator, PLINUX_MDL Mdl,
             unsigned long Offset, ljtPHYS_ADDR_T *Physical)
{
    struct gfp_mdl_priv *mdlPriv = Mdl->priv;
    ljtUINT32 offsetInPage = Offset & ~PAGE_MASK;
    ljtUINT32 index = (ljtUINT32)(Offset / PAGE_SIZE);

    if (mdlPriv->contiguous)
        *Physical = page_to_phys(nth_page(mdlPriv->contiguousPages, index));
    else
        *Physical = page_to_phys(mdlPriv->nonContiguousPages[index]);

    *Physical += offsetInPage;

    return ljvSTATUS_OK;
}

static void
_GFPAllocatorDestructor(ljsALLOCATOR *Allocator)
{
    _GFPAllocatorDebugfsCleanup(Allocator);

    kfree(Allocator->privateData);

    kfree(Allocator);
}

static ljeSTATUS
_GFPGetDmaAddress(ljkALLOCATOR Allocator, PLINUX_MDL Mdl,
                unsigned long Offset, ljtPHYS_ADDR_T *DmaAddress)
{
    struct gfp_mdl_priv *mdlPriv = Mdl->priv;

    if (mdlPriv->contiguous) {
        *DmaAddress = mdlPriv->dma_addr + Offset;
    } else {
        struct sg_table *sgt = &mdlPriv->sgt;
        struct scatterlist *sg;
        unsigned int i;

        for_each_sg(sgt->sgl, sg, sgt->nents, i) {
            if (Offset < sg_dma_len(sg)) {
                /* current sg includes target offset */
                *DmaAddress = sg_dma_address(sg) + Offset;
                break;
            }
            Offset -= sg_dma_len(sg);
        }

        /* check for invalid address */
        if (i >= sgt->nents) {
            pr_warn("Offset exceeds mapped DMA region\n");
            return ljvSTATUS_INVALID_ADDRESS;
        }
    }
    return ljvSTATUS_OK;
}

/* GFP allocator operations. */
static ljsALLOCATOR_OPERATIONS GFPAllocatorOperations = {
    .Alloc       = _GFPAlloc,
    .Free        = _GFPFree,
    .Mmap        = _GFPMmap,
    .MapUser     = _GFPMapUser,
    .UnmapUser   = _GFPUnmapUser,
    .MapKernel   = _GFPMapKernel,
    .UnmapKernel = _GFPUnmapKernel,
    .Cache       = _GFPCache,
    .Physical    = _GFPPhysical,
    .GetSGT      = _GFPGetSGT,
    .GetDmaAddress = _GFPGetDmaAddress,
};

/* GFP allocator entry. */
ljeSTATUS
_GFPAlloctorInit(ljkOS Os, ljsDEBUGFS_DIR *Parent, ljkALLOCATOR *Allocator)
{
    ljeSTATUS status;
    ljkALLOCATOR allocator = ljvNULL;
    struct gfp_alloc *priv = ljvNULL;
#if 0
    if (Os->iommu)
        ljmkONERROR(ljvSTATUS_NOT_SUPPORTED);
#endif
    ljmkONERROR(ljkALLOCATOR_Construct(Os, &GFPAllocatorOperations, &allocator));

    priv = kzalloc(sizeof(*priv), GFP_KERNEL | ljdNOWARN);

    if (!priv)
        ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);

    atomic_set(&priv->low, 0);
    atomic_set(&priv->high, 0);

    /* Register private data. */
    allocator->privateData = priv;
    allocator->destructor = _GFPAllocatorDestructor;

    _GFPAllocatorDebugfsInit(allocator, Parent);

    allocator->capability = ljvALLOC_FLAG_CONTIGUOUS
                          | ljvALLOC_FLAG_NON_CONTIGUOUS
                          | ljvALLOC_FLAG_CACHEABLE
                          | ljvALLOC_FLAG_MEMLIMIT
                          | ljvALLOC_FLAG_ALLOC_ON_FAULT
                          | ljvALLOC_FLAG_DMABUF_EXPORTABLE
#if (defined(CONFIG_ZONE_DMA32) || defined(CONFIG_ZONE_DMA)) && LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 37)
                          | ljvALLOC_FLAG_4GB_ADDR
#endif
                          | ljvALLOC_FLAG_1M_PAGES
                          | ljvALLOC_FLAG_CPU_ACCESS
                          | ljvALLOC_FLAG_FROM_USER
#if ljdENABLE_VIDEO_MEMORY_MIRROR
                          | ljvALLOC_FLAG_WITH_MIRROR
#endif
                          | ljvALLOC_FLAG_32BIT_VA
                          | ljvALLOC_FLAG_PRIOR_32BIT_VA
                          ;

#if defined(ljdEMULATE_SECURE_ALLOCATOR)
    allocator->capability |= ljvALLOC_FLAG_SECURITY;
#endif

    *Allocator = allocator;

    return ljvSTATUS_OK;

OnError:
    kfree(allocator);
    return status;
}
