/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#include "lj_hal_kernel_linux.h"
#include "lj_hal_kernel_allocator.h"
#include <linux/pagemap.h>
#include <linux/seq_file.h>
#include <linux/mman.h>
#include <linux/atomic.h>
#include <linux/dma-mapping.h>
#include <linux/slab.h>
#include <linux/io.h>
#include <linux/ioport.h>
#include <linux/swiotlb.h>

#define _LJ_OBJ_ZONE ljvZONE_OS

#if (defined CONFIG_MIPS || defined CONFIG_LOONGARCH)
#ifndef pgprot_writecombine_wc
static inline pgprot_t pgprot_writecombine_wc(pgprot_t _prot)
{
        unsigned long prot = pgprot_val(_prot);
#ifdef CONFIG_MIPS
       	prot = (prot & ~_CACHE_MASK) | _CACHE_UNCACHED_ACCELERATED;
#endif
#ifdef CONFIG_LOONGARCH
        prot = (prot & ~_CACHE_MASK) | _CACHE_WUC;
#endif
        return __pgprot(prot);
}
#endif
#endif


/*
 * reserved_mem is for contiguous pool, internal pool and external pool, etc.
 */

/* mdl private. */
struct reserved_mem {
    unsigned long   start;
    unsigned long   size;
    unsigned int    offset_in_page;
    char            name[32];
    int             release;

    /* Link together. */
    struct          list_head link;
    /* the mdl is root or not */
    ljtBOOL         root;
};

/* allocator info. */
struct reserved_mem_alloc {
    /* Record allocated reserved memory regions. */
    struct list_head region;
    /* protect reserved memory */
    struct mutex     lock;
};

static int reserved_mem_show(struct seq_file *m, void *data)
{
    struct list_head *pos;
    ljsINFO_NODE *node = m->private;
    ljkALLOCATOR Allocator = node->device;
    struct reserved_mem_alloc *alloc = Allocator->privateData;

    list_for_each(pos, &alloc->region) {
        struct reserved_mem *res = list_entry(pos, struct reserved_mem, link);

        seq_printf(m, "0x%08lx-0x%08lx : %s\n",
                   res->start, res->start + res->size - 1, res->name);
    }

    return 0;
}

static ljsINFO info_list[] = {
    { "reserved-mem", reserved_mem_show },
};

static void
reserved_mem_debugfs_init(ljkALLOCATOR Allocator, ljkDEBUGFS_DIR Root)
{
    ljmkVERIFY_OK(ljkDEBUGFS_DIR_Init(&Allocator->debugfsDir,
                                      Root->root, "reserved-mem"));

    ljmkVERIFY_OK(ljkDEBUGFS_DIR_CreateFiles(&Allocator->debugfsDir, info_list,
                                             ljmCOUNTOF(info_list), Allocator));
}

static void
reserved_mem_debugfs_cleanup(ljkALLOCATOR Allocator)
{
    ljmkVERIFY_OK(ljkDEBUGFS_DIR_RemoveFiles(&Allocator->debugfsDir,
                                             info_list, ljmCOUNTOF(info_list)));

    ljkDEBUGFS_DIR_Deinit(&Allocator->debugfsDir);
}

static ljeSTATUS
reserved_mem_alloc(ljkALLOCATOR Allocator, PLINUX_MDL Mdl, ljtSIZE_T NumPages, ljtUINT32 Flags)
{
    return ljvSTATUS_OUT_OF_MEMORY;
}

static ljeSTATUS
reserved_mem_attach(ljkALLOCATOR Allocator, ljsATTACH_DESC_PTR Desc, PLINUX_MDL Mdl)
{
    struct reserved_mem_alloc *alloc = Allocator->privateData;
    struct reserved_mem *res;
    struct resource *region  = NULL;
    ljtPHYS_ADDR_T gpu_end = 0;
    ljtBOOL acquiredMutex = ljvFALSE;
    ljeSTATUS status = ljvSTATUS_OK;

    if (Desc == ljvNULL)
        return ljvSTATUS_INVALID_ARGUMENT;

    res = kzalloc(sizeof(*res), GFP_KERNEL | ljdNOWARN);

    if (!res)
        return ljvSTATUS_OUT_OF_MEMORY;

    res->start = Desc->reservedMem.start;
    res->size = Desc->reservedMem.size;
    res->offset_in_page = Desc->reservedMem.start & (PAGE_SIZE - 1);
    strncpy(res->name, Desc->reservedMem.name, sizeof(res->name) - 1);
    res->release = 0;
    res->root = Desc->reservedMem.root;

    /* the region requierd is handed by root MDL */
    if (Desc->reservedMem.root) {
        if (!Desc->reservedMem.requested) {
            region = request_mem_region(res->start, res->size, res->name);

            if (!region) {
                pr_warn("request mem %s(0x%lx - 0x%lx) failed\n",
                        res->name, res->start, res->start + res->size - 1);

                kfree(res);
                return ljvSTATUS_OUT_OF_RESOURCES;
            }

            res->release = 1;
        }

        ljmkONERROR(ljkOS_AcquireMutex(Allocator->os, &alloc->lock, ljvINFINITE));
        acquiredMutex = ljvTRUE;

        list_add(&res->link, &alloc->region);

        ljmkONERROR(ljkOS_ReleaseMutex(Allocator->os, &alloc->lock));
        acquiredMutex = ljvFALSE;

    }

    Mdl->priv = res;

    ljkOS_CPUPhysicalToGPUPhysical(Allocator->os,
                                   (res->start + res->size - 1),
                                   &gpu_end);
    if (gpu_end <= 0xFFFFFFFF)
        Allocator->capability |= ljvALLOC_FLAG_4GB_ADDR;

    return ljvSTATUS_OK;

OnError:
    if (acquiredMutex)
        ljkOS_ReleaseMutex(Allocator->os, &alloc->lock);

    return status;
}

static void
reserved_mem_detach(ljkALLOCATOR Allocator, PLINUX_MDL Mdl)
{
    struct reserved_mem_alloc *alloc = Allocator->privateData;
    struct reserved_mem *res = Mdl->priv;
    ljeSTATUS status = ljvSTATUS_OK;
    ljtBOOL acquiredMutex = ljvFALSE;

    if (res->root) {
        /* unlink from region list. */
        ljmkONERROR(ljkOS_AcquireMutex(Allocator->os, &alloc->lock, ljvINFINITE));
        acquiredMutex = ljvTRUE;

        list_del_init(&res->link);

        ljmkONERROR(ljkOS_ReleaseMutex(Allocator->os, &alloc->lock));
        acquiredMutex = ljvFALSE;

        if (res->release)
            release_mem_region(res->start, res->size);
    }

    kfree(res);

OnError:
    if (acquiredMutex)
        ljkOS_ReleaseMutex(Allocator->os, &alloc->lock);

}

static ljeSTATUS
reserved_mem_mmap(ljkALLOCATOR Allocator, PLINUX_MDL Mdl, ljtBOOL Cacheable,
                  ljtSIZE_T skipPages, ljtSIZE_T numPages, struct vm_area_struct *vma)
{
    struct reserved_mem *res = (struct reserved_mem *)Mdl->priv;
    unsigned long pfn;
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkHEADER_ARG("Allocator=%p Mdl=%p vma=%p", Allocator, Mdl, vma);

    ljmkASSERT(skipPages + numPages <= Mdl->numPages);

    pfn = (res->start >> PAGE_SHIFT) + skipPages;

    /* Make this mapping non-cached. */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 3, 0)
    vm_flags_set(vma, ljdVM_FLAGS);
#else
    vma->vm_flags |= ljdVM_FLAGS;
#endif

#if ljdENABLE_BUFFERABLE_VIDEO_MEMORY
#if (defined CONFIG_MIPS || defined CONFIG_LOONGARCH ||defined CONFIG_SW64)
        vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);    /* 0x37 */
#else
        vma->vm_page_prot = pgprot_writecombine(vma->vm_page_prot); /* 0x2f */
#endif
#else
        vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);    /* 0x37 */
#endif

    if (remap_pfn_range(vma, vma->vm_start, pfn,
                        numPages << PAGE_SHIFT, vma->vm_page_prot) < 0) {
        ljmkTRACE(ljvLEVEL_ERROR,
                  "%s(%d): remap_pfn_range error.",
                  __func__, __LINE__);

        status = ljvSTATUS_OUT_OF_MEMORY;
    }

    ljmkFOOTER();
    return status;
}

static void
reserved_mem_unmap_user(ljkALLOCATOR Allocator, PLINUX_MDL Mdl,
                        PLINUX_MDL_MAP MdlMap, ljtUINT32 Size)
{
    struct reserved_mem *res = (struct reserved_mem *)Mdl->priv;

    if (unlikely(!current->mm))
        return;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 4, 0)
    if (vm_munmap((unsigned long)MdlMap->vmaAddr - res->offset_in_page, res->size) < 0)
        pr_warn("%s: vm_munmap failed\n", __func__);
#else
    down_write(&current_mm_mmap_sem);
    if (do_munmap(current->mm,
                  (unsigned long)MdlMap->vmaAddr - res->offset_in_page,
                  res->size) < 0)
        pr_warn("%s: do_munmap failed\n", __func__);

    up_write(&current_mm_mmap_sem);
#endif

    MdlMap->vma = NULL;
    MdlMap->vmaAddr = NULL;
}

static ljeSTATUS
reserved_mem_map_user(ljkALLOCATOR Allocator, PLINUX_MDL Mdl,
                      PLINUX_MDL_MAP MdlMap, ljtBOOL Cacheable)
{
    struct reserved_mem *res = (struct reserved_mem *)Mdl->priv;
    ljtPOINTER userLogical = ljvNULL;
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkHEADER_ARG("Allocator=%p Mdl=%p Cacheable=%d", Allocator, Mdl, Cacheable);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 4, 0)
#if ljdANON_FILE_FOR_ALLOCATOR
    userLogical = (ljtPOINTER)vm_mmap(Allocator->anon_file, 0L, res->size,
# else
    userLogical = (ljtPOINTER)vm_mmap(NULL, 0L, res->size,
# endif
                                      PROT_READ | PROT_WRITE, MAP_SHARED | MAP_NORESERVE, 0);
#else
    down_write(&current_mm_mmap_sem);
    userLogical = (ljtPOINTER)do_mmap_pgoff(NULL, 0L, res->size,
                                            PROT_READ | PROT_WRITE, MAP_SHARED, 0);
    up_write(&current_mm_mmap_sem);
#endif

    ljmkTRACE_ZONE(ljvLEVEL_INFO, ljvZONE_OS,
                   "%s(%d): vmaAddr->%p for phys_addr->%p",
                   __func__, __LINE__, userLogical, Mdl);

    if (IS_ERR(userLogical)) {
        ljmkTRACE_ZONE(ljvLEVEL_INFO, ljvZONE_OS,
                       "%s(%d): do_mmap_pgoff error",
                       __func__, __LINE__);

        userLogical = ljvNULL;

        ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);
    }

    down_write(&current_mm_mmap_sem);

    do {
        struct vm_area_struct *vma = find_vma(current->mm, (unsigned long)userLogical);

        if (vma == ljvNULL) {
            ljmkTRACE_ZONE(ljvLEVEL_INFO, ljvZONE_OS,
                           "%s(%d): find_vma error",
                           __func__, __LINE__);

            ljmkERR_BREAK(ljvSTATUS_OUT_OF_RESOURCES);
        }

        ljmkERR_BREAK(reserved_mem_mmap(Allocator, Mdl, ljvFALSE, 0, Mdl->numPages, vma));

        MdlMap->vmaAddr = userLogical + res->offset_in_page;
        MdlMap->cacheable = ljvFALSE;
        MdlMap->vma = vma;
    } while (ljvFALSE);

    up_write(&current_mm_mmap_sem);

OnError:
    if (ljmIS_ERROR(status) && userLogical) {
        MdlMap->vmaAddr = userLogical + res->offset_in_page;
        reserved_mem_unmap_user(Allocator, Mdl, MdlMap, res->size);
    }

    ljmkFOOTER();
    return status;
}

static ljeSTATUS
reserved_mem_map_kernel(ljkALLOCATOR Allocator, PLINUX_MDL Mdl,
                        ljtSIZE_T Offset, ljtSIZE_T Bytes, ljtPOINTER *Logical)
{
    struct reserved_mem *res = Mdl->priv;
    void *vaddr;

    if (Offset + Bytes > res->size)
        return ljvSTATUS_INVALID_ARGUMENT;

#if ljdENABLE_BUFFERABLE_VIDEO_MEMORY
#if (defined CONFIG_MIPS || defined CONFIG_LOONGARCH ||defined CONFIG_SW64)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 6, 0)
    vaddr = ioremap(res->start + Offset, Bytes);
#else
    vaddr = ioremap_nocache(res->start + Offset, Bytes);
#endif
#else
    vaddr = ioremap_wc(res->start + Offset, Bytes);
#endif
#else
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 6, 0)
    vaddr = ioremap(res->start + Offset, Bytes);
# else
    vaddr = ioremap_nocache(res->start + Offset, Bytes);
# endif
#endif

    if (!vaddr)
        return ljvSTATUS_OUT_OF_MEMORY;

    *Logical = vaddr;
    return ljvSTATUS_OK;
}

static ljeSTATUS
reserved_mem_unmap_kernel(ljkALLOCATOR Allocator, PLINUX_MDL Mdl, ljtPOINTER Logical)
{
    iounmap((void *)Logical);

    return ljvSTATUS_OK;
}

static ljeSTATUS
reserved_mem_cache_op(ljkALLOCATOR Allocator, PLINUX_MDL Mdl, ljtSIZE_T Offset,
                      ljtPOINTER Logical, ljtSIZE_T Bytes, ljeCACHEOPERATION Operation)
{
    /* Always WC or UC, safe to use mb. */
    switch (Operation) {
    case ljvCACHE_CLEAN:
    case ljvCACHE_FLUSH:
        _MemoryBarrier();
        break;
    case ljvCACHE_INVALIDATE:
        break;
    default:
        return ljvSTATUS_INVALID_ARGUMENT;
    }

    return ljvSTATUS_OK;
}

static ljeSTATUS
reserved_mem_get_physical(ljkALLOCATOR Allocator, PLINUX_MDL Mdl,
                          unsigned long Offset, ljtPHYS_ADDR_T *Physical)
{
    struct reserved_mem *res = Mdl->priv;
    *Physical = res->start + Offset;

    return ljvSTATUS_OK;
}

static ljeSTATUS
reserved_mem_GetSGT(ljkALLOCATOR Allocator, PLINUX_MDL Mdl,
                    ljtSIZE_T Offset, ljtSIZE_T Bytes, ljtPOINTER *SGT)
{
    struct page *page = ljvNULL;
    struct sg_table *sgt = NULL;
    struct reserved_mem *res = Mdl->priv;
    ljeSTATUS status = ljvSTATUS_OK;
    struct scatterlist *sg = NULL;
    ljtUINT32 offset_in_page = 0;
    unsigned long mem_base = res->start + Offset;
    ljtSIZE_T sz = 0;
    ljtSIZE_T left = Bytes;
    ljtSIZE_T chunk_size = IO_TLB_SEGSIZE * (1 << IO_TLB_SHIFT);

    ljmkASSERT(Offset + Bytes <= Mdl->numPages << PAGE_SHIFT);

    sgt = kmalloc(sizeof(*sgt), GFP_KERNEL);
    if (!sgt)
        ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);

    if (sg_alloc_table(sgt, (Bytes + chunk_size - 1) / chunk_size, GFP_KERNEL))
        ljmkONERROR(ljvSTATUS_GENERIC_IO);

    sg = sgt->sgl;
    while (left) {
        if (left > chunk_size)
            sz = chunk_size;
        else
            sz = left;

        left -= sz;

        page = pfn_to_page(mem_base >> PAGE_SHIFT);
        offset_in_page = (ljtUINT32)mem_base & (PAGE_SIZE - 1);

        sg_set_page(sg, page, sz, offset_in_page);

        mem_base += sz;
        sg = sg_next(sg);
        if (!sg)
            break;
    }

    if (left)
        pr_warn("Bytes is mismatch with sgt size, start: 0x%llx, left: 0x%llx, Offset: 0x%llx, Bytes: 0x%llx\n",
                (ljtUINT64)res->start,
                (ljtUINT64)left,
                (ljtUINT64)Offset,
                (ljtUINT64)Bytes);

    *SGT = (ljtPOINTER)sgt;

OnError:
    if (ljmIS_ERROR(status) && sgt)
        kfree(sgt);

    return status;
}

static void
reserved_mem_dtor(ljsALLOCATOR *Allocator)
{
    reserved_mem_debugfs_cleanup(Allocator);

    ljkOS_ZeroMemory(Allocator->privateData, sizeof(struct reserved_mem_alloc));
    kfree(Allocator->privateData);

    kfree(Allocator);
}

/* GFP allocator operations. */
static ljsALLOCATOR_OPERATIONS reserved_mem_ops = {
    .Alloc       = reserved_mem_alloc,
    .Attach      = reserved_mem_attach,
    .Free        = reserved_mem_detach,
    .Mmap        = reserved_mem_mmap,
    .MapUser     = reserved_mem_map_user,
    .UnmapUser   = reserved_mem_unmap_user,
    .MapKernel   = reserved_mem_map_kernel,
    .UnmapKernel = reserved_mem_unmap_kernel,
    .Cache       = reserved_mem_cache_op,
    .Physical    = reserved_mem_get_physical,
    .GetSGT      = reserved_mem_GetSGT,
    .GetDmaAddress = reserved_mem_get_physical,
};

/* GFP allocator entry. */
ljeSTATUS
_ReservedMemoryAllocatorInit(ljkOS Os, ljsDEBUGFS_DIR *Parent, ljkALLOCATOR *Allocator)
{
    ljeSTATUS status;
    ljkALLOCATOR allocator = ljvNULL;
    struct reserved_mem_alloc *alloc = NULL;

    ljmkONERROR(ljkALLOCATOR_Construct(Os, &reserved_mem_ops, &allocator));

    alloc = kzalloc(sizeof(*alloc), GFP_KERNEL | ljdNOWARN);

    if (!alloc)
        ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);

    INIT_LIST_HEAD(&alloc->region);
    mutex_init(&alloc->lock);

    /* Register private data. */
    allocator->privateData = alloc;
    allocator->destructor = reserved_mem_dtor;

    reserved_mem_debugfs_init(allocator, Parent);

    allocator->capability = ljvALLOC_FLAG_LINUX_RESERVED_MEM
                          | ljvALLOC_FLAG_CONTIGUOUS
                          | ljvALLOC_FLAG_MEMLIMIT
                          | ljvALLOC_FLAG_CPU_ACCESS
                          | ljvALLOC_FLAG_DMABUF_EXPORTABLE
                          | ljvALLOC_FLAG_NON_CPU_ACCESS
                          | ljvALLOC_FLAG_FROM_USER
#if ljdENABLE_VIDEO_MEMORY_MIRROR
                          | ljvALLOC_FLAG_WITH_MIRROR
#endif
                          | ljvALLOC_FLAG_32BIT_VA
                          | ljvALLOC_FLAG_PRIOR_32BIT_VA
                          ;

    *Allocator = allocator;

    return ljvSTATUS_OK;

OnError:
    kfree(allocator);
    return status;
}
