/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#include "lj_hal_kernel_linux.h"
#include "lj_hal_kernel_allocator.h"
#include <linux/scatterlist.h>

#include <linux/slab.h>
#include <linux/pagemap.h>
#include <linux/cache.h>

#define _LJ_OBJ_ZONE        ljvZONE_ALLOCATOR

#define ljdUSING_PFN_FOLLOW 0

enum um_desc_type {
    UM_PHYSICAL_MAP,
    UM_PAGE_MAP,
    UM_PFN_MAP,
};

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 24)
struct sg_table {
    struct scatterlist *sgl;
    unsigned int        nents;
    unsigned int        orig_nents;
};
#endif

/* Descriptor of a user memory imported. */
struct um_desc {
    int                     type;

    union {
        /* UM_PHYSICAL_MAP. */
        unsigned long       physical;

        /* UM_PAGE_MAP. */
        struct {
            struct page   **pages;
        };

        /* UM_PFN_MAP. */
        struct {
            unsigned long  *pfns;
            int            *refs;
        };
    };

    struct sg_table         sgt;

    /* contiguous chunks, does not include padding pages. */
    int                     chunk_count;

    unsigned long           vm_flags;
    unsigned long           user_vaddr;
    size_t                  size;
    unsigned long           offset;
    dma_addr_t              dmaHandle;

    size_t                  pageCount;
    size_t                  extraPage;
    unsigned int            alloc_from_res;
};

static void *alloc_memory(size_t bytes)
{
    if (bytes > PAGE_SIZE)
        return vzalloc(bytes);
    else
        return kzalloc(bytes, GFP_KERNEL | ljdNOWARN);
}

static void free_memory(void *ptr)
{
    if (is_vmalloc_addr(ptr))
        vfree(ptr);
    else
        kfree(ptr);
}

static int import_physical_map(ljkOS Os, struct device *dev,
                               struct um_desc *um, unsigned long phys)
{
    um->type = UM_PHYSICAL_MAP;
    um->physical = phys & PAGE_MASK;
    um->chunk_count = 1;

    if (Os->iommu) {
        dma_addr_t dmaHandle;
        size_t size = um->size + (phys & (~PAGE_MASK));
        unsigned long pfn = phys >> PAGE_SHIFT;

        if (pfn_valid(pfn))
            dmaHandle = dma_map_page(dev, pfn_to_page(pfn),
                                     0, size, DMA_BIDIRECTIONAL);
        else
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 9, 0)
            dmaHandle = dma_map_resource(dev, pfn << PAGE_SHIFT,
                                         size, DMA_BIDIRECTIONAL, 0);
#else
            dmaHandle = dma_map_page(dev, pfn_to_page(pfn),
                                     0, size, DMA_BIDIRECTIONAL);
#endif

        if (dma_mapping_error(dev, dmaHandle))
            return ljvSTATUS_OUT_OF_MEMORY;

        um->dmaHandle = dmaHandle;
    }

    return 0;
}

static ljeSTATUS
phy_is_from_reserved(ljkOS Os, ljtPHYS_ADDR_T Physical)
{
    ljeSTATUS status = ljvSTATUS_NOT_FOUND;
    ljkGALDEVICE device = Os->device;
    ljtUINT32 i = 0;

    if (device->lowContiguousSize) {
        if ((Physical >= device->lowContiguousBase) &&
            (Physical < device->lowContiguousBase + device->lowContiguousSize)) {
            status = ljvSTATUS_OK;
            return status;
        }
    }
    for (i = 0; i < ljdSYSTEM_RESERVE_COUNT; i++) {
        if (device->contiguousSizes[i]) {
            if ((Physical >= device->contiguousBases[i]) &&
                (Physical < device->contiguousBases[i] + device->contiguousSizes[i])) {
                status = ljvSTATUS_OK;
                return status;
            }
        }
    }
    for (i = 0; i < ljdDEVICE_COUNT; i++) {
        if (device->externalSize[i]) {
            if ((Physical >= device->externalBase[i]) &&
                (Physical < device->externalBase[i] + device->externalSize[i])) {
                status = ljvSTATUS_OK;
                return status;
            }
        }
    }
    for (i = 0; i < ljdDEVICE_COUNT; i++) {
        if (device->exclusiveSize[i]) {
            if ((Physical >= device->exclusiveBase[i]) &&
                (Physical < device->exclusiveBase[i] + device->exclusiveSize[i])) {
                status = ljvSTATUS_OK;
                return status;
            }
        }
    }
    return status;
}

static int
import_page_map(ljkOS Os, struct device *dev, struct um_desc *um,
                unsigned long addr, size_t page_count, size_t size, unsigned long flags)
{
    int i;
    int result;
    int ret;
    struct page **pages;

    if ((addr & (cache_line_size() - 1)) || (size & (cache_line_size() - 1)))
        /* Not cpu cacheline size aligned, can not support. */
        return -EINVAL;

    pages = alloc_memory(page_count * sizeof(void *));
    if (!pages)
        return -ENOMEM;

    down_read(&current_mm_mmap_sem);

#if LINUX_VERSION_CODE > KERNEL_VERSION(5, 6, 0)
    result = pin_user_pages(addr & PAGE_MASK, page_count,
#elif LINUX_VERSION_CODE < KERNEL_VERSION(4, 6, 0)
    result = get_user_pages(current, current->mm, addr & PAGE_MASK, page_count,
#else
    result = get_user_pages(addr & PAGE_MASK, page_count,
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 9, 0) || defined(CONFIG_PPC)
                            (flags & VM_WRITE) ? FOLL_WRITE : 0,
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(4, 5, 1)
                            (flags & VM_WRITE) ? 1 : 0, 0,
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(4, 4, 168)
                            (flags & VM_WRITE) ? FOLL_WRITE : 0,
#else
                            (flags & VM_WRITE) ? 1 : 0, 0,
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 6, 0)
                            pages);
#else
                            pages, NULL);
#endif

    up_read(&current_mm_mmap_sem);

    /* Checking whether Physical is greater than 4G. */
    if (Os->device->platform->flagBits & ljvPLATFORM_FLAG_LIMIT_4G_ADDRESS) {
        ljtPHYS_ADDR_T addr;

        for (i = 0; i < page_count; i++) {
            addr = page_to_phys(pages[i]);
            if (addr > 0xFFFFFFFFu) {
                ret = -EINVAL;
                goto error;
            }
        }
    }

    if (result < page_count) {
        for (i = 0; i < result; i++) {
            if (pages[i])
#if LINUX_VERSION_CODE > KERNEL_VERSION(5, 6, 0)
                unpin_user_page(pages[i]);
#else
                put_page(pages[i]);
#endif
        }

        free_memory(pages);
        return -ENODEV;
    }

    um->chunk_count = 1;
    for (i = 1; i < page_count; i++) {
        if (page_to_pfn(pages[i]) != page_to_pfn(pages[i - 1]) + 1)
            ++um->chunk_count;
    }

    if (ljmIS_SUCCESS(phy_is_from_reserved(Os, (ljtPHYS_ADDR_T)page_to_pfn(pages[0]) << PAGE_SHIFT)))
        um->alloc_from_res = 1;

    if (!um->alloc_from_res) {
#if ljdUSE_LINUX_SG_TABLE_API
        result = sg_alloc_table_from_pages(&um->sgt, pages, page_count,
                                           addr & ~PAGE_MASK, size, GFP_KERNEL | ljdNOWARN);

#else
        result = alloc_sg_list_from_pages(&um->sgt.sgl, pages, page_count,
                                          addr & ~PAGE_MASK, size, &um->sgt.nents);

        um->sgt.orig_nents = um->sgt.nents;
#endif
        if (unlikely(result < 0)) {
            pr_warn("[ljmcore]: %s: sg_alloc_table_from_pages failed\n", __func__);
            ret = -ENOMEM;
            goto error;
        }

        result = dma_map_sg(dev, um->sgt.sgl, um->sgt.nents, DMA_TO_DEVICE);
        if (unlikely(result != um->sgt.nents)) {
            pr_warn("[ljmcore]: %s: dma_map_sg failed\n", __func__);
            //goto error;
            um->sgt.nents = result;
        }

        if (Os->iommu)
            um->dmaHandle = sg_dma_address(um->sgt.sgl);

        dma_sync_sg_for_cpu(dev, um->sgt.sgl, um->sgt.nents, DMA_FROM_DEVICE);
    }

    um->type = UM_PAGE_MAP;
    um->pages = pages;

    return 0;

error:
#if ljdUSE_LINUX_SG_TABLE_API
    sg_free_table(&um->sgt);
#else
    kfree(um->sgt.sgl);
#endif

    free_memory(pages);

    return ret;
}

static int
import_pfn_map(ljkOS Os, struct device *dev, struct um_desc *um,
               unsigned long addr, size_t pfn_count)
{
    int i;
    struct vm_area_struct *vma;
    unsigned long *pfns;
    int *refs;
    struct page **pages = ljvNULL;
    int result = 0;
    int rets = -ENOTTY;
    size_t pageCount = 0;

    if (!current->mm)
        return -ENOTTY;

    down_read(&current_mm_mmap_sem);
    vma = find_vma(current->mm, addr);
#if !ljdUSING_PFN_FOLLOW
    up_read(&current_mm_mmap_sem);
#endif

    if (!vma)
        return -ENOTTY;

    pfns = (unsigned long *)alloc_memory(pfn_count * sizeof(unsigned long));

    if (!pfns)
        return -ENOMEM;

    refs = (int *)alloc_memory(pfn_count * sizeof(int));

    if (!refs) {
        free_memory(pfns);
        return -ENOMEM;
    }

    pages = alloc_memory(pfn_count * sizeof(void *));
    if (!pages) {
        free_memory(pfns);
        free_memory(refs);
        return -ENOMEM;
    }

    for (i = 0; i < pfn_count; i++) {
#if ljdUSING_PFN_FOLLOW
        int ret = 0;
        ret = follow_pfn(vma, addr, &pfns[i]);
        if (ret < 0) {
            up_read(&current_mm_mmap_sem);
            goto err;
        }
#else
        /* protect pfns[i] */
        spinlock_t  *ptl;
        pgd_t *pgd;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 9, 0)
        p4d_t *p4d;
# endif
        pud_t *pud;
        pmd_t *pmd;
        pte_t *pte;

        pgd = pgd_offset(current->mm, addr);
        if (pgd_none(*pgd) || pgd_bad(*pgd))
            goto err;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 9, 0)
        p4d = p4d_offset(pgd, addr);
        if (p4d_none(READ_ONCE(*p4d)))
            goto err;

        pud = pud_offset(p4d, addr);
# elif (defined(CONFIG_X86)) && LINUX_VERSION_CODE >= KERNEL_VERSION (4, 12, 0)
        pud = pud_offset((p4d_t *)pgd, addr);
# elif (defined(CONFIG_CPU_CSKYV2)) && LINUX_VERSION_CODE >= KERNEL_VERSION (4, 11, 0)
        pud = pud_offset((p4d_t *)pgd, addr);
# else
        pud = pud_offset(pgd, addr);
# endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(5, 9, 0) */
        if (pud_none(*pud) || pud_bad(*pud))
            goto err;

        pmd = pmd_offset(pud, addr);
        if (pmd_none(*pmd) || pmd_bad(*pmd))
            goto err;

#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0)
        pte = pte_offset_map_lock(current->mm, pmd, addr, &ptl);
#else
        pte = pte_offset_kernel(pmd, (unsigned long)addr);
#endif

        if (!pte_present(*pte)) {
            pte_unmap_unlock(pte, ptl);
            goto err;
        }

        pfns[i] = pte_pfn(*pte);
        pte_unmap_unlock(pte, ptl);
#endif
        /* Advance to next. */
        addr += PAGE_SIZE;
    }
#if ljdUSING_PFN_FOLLOW
    up_read(&current_mm_mmap_sem);
#endif

    /* Check whether Physical is greater than 4G */
    if (Os->device->platform->flagBits & ljvPLATFORM_FLAG_LIMIT_4G_ADDRESS) {
        ljtPHYS_ADDR_T addr;

        for (i = 0; i < pageCount; i++) {
            addr = pfns[i] << PAGE_SHIFT;
            if (addr > 0xFFFFFFFFu) {
                rets = -EINVAL;
                goto err;
            }
        }
    }

    for (i = 0; i < pfn_count; i++) {
        if (pfn_valid(pfns[i])) {
            struct page *page = pfn_to_page(pfns[i]);

            refs[i] = get_page_unless_zero(page);
            pages[i] = page;
            pageCount++;
        }
    }

    um->chunk_count = 1;
    for (i = 1; i < pfn_count; i++) {
        if (pfns[i] != pfns[i - 1] + 1)
            ++um->chunk_count;
    }

    if (ljmIS_SUCCESS(phy_is_from_reserved(Os, (ljtPHYS_ADDR_T)pfns[0] << PAGE_SHIFT)))
        um->alloc_from_res = 1;

    if (pageCount == pfn_count && !um->alloc_from_res) {
#if ljdUSE_LINUX_SG_TABLE_API
        result = sg_alloc_table_from_pages(&um->sgt, pages, pfn_count, addr & ~PAGE_MASK,
                                           pfn_count * PAGE_SIZE, GFP_KERNEL | ljdNOWARN);

#else
        result = alloc_sg_list_from_pages(&um->sgt.sgl, pages, pfn_count, addr & ~PAGE_MASK,
                                          pfn_count * PAGE_SIZE, &um->sgt.nents);

        um->sgt.orig_nents = um->sgt.nents;
#endif
        if (unlikely(result < 0)) {
            pr_warn("[ljmcore]: %s: sg_alloc_table_from_pages failed\n", __func__);
            rets = -ENOMEM;
            goto err;
        }

        result = dma_map_sg(dev, um->sgt.sgl, um->sgt.nents, DMA_TO_DEVICE);

        if (unlikely(result != um->sgt.nents)) {
            /*
#if ljdUSE_LINUX_SG_TABLE_API
            sg_free_table(&um->sgt);
#else
            kfree(um->sgt.sgl);
#endif
            pr_warn("[ljmcore]: %s: dma_map_sg failed\n", __func__);
            rets = -ENOMEM;
            goto err;
            */
            um->sgt.nents = result;
        }
        if (Os->iommu)
            um->dmaHandle = sg_dma_address(um->sgt.sgl);
    }

    free_memory(pages);
    pages = ljvNULL;

    um->type = UM_PFN_MAP;
    um->pfns = pfns;
    um->refs = refs;
    return 0;

err:
    free_memory(pfns);
    free_memory(refs);
    free_memory(pages);

    return rets;
}
static ljeSTATUS
_Import(ljkOS Os, PLINUX_MDL Mdl, ljtPOINTER Memory,
        ljtPHYS_ADDR_T Physical, ljtSIZE_T Size, struct um_desc *UserMemory)
{
    ljeSTATUS status = ljvSTATUS_OK;
    unsigned long vm_flags = 0;
    struct vm_area_struct *vma = NULL;
    struct device *dev = (struct device *)Mdl->device;
    ljtSIZE_T start, end, memory;
    int result = 0;

    ljtSIZE_T extraPage;
    ljtSIZE_T pageCount;

    ljmkHEADER_ARG("Os=%p Memory=%p Physical=0x%llx Size=%lu",
                   Os, Memory, Physical, Size);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Memory != ljvNULL || Physical != ~0ULL);
    ljmkVERIFY_ARGUMENT(Size > 0);

    memory = (Physical != ljvINVALID_PHYSICAL_ADDRESS) ?
             Physical : untagged_addr((ljtSIZE_T)Memory);

    /* Get the number of required pages. */
    end = (memory + Size + PAGE_SIZE - 1) >> PAGE_SHIFT;
    start = memory >> PAGE_SHIFT;
    pageCount = end - start;

    /* Allocate extra page to avoid cache overflow */
#if ljdENABLE_2D
    extraPage = 2;
#else
    extraPage = (((memory + ljmALIGN(Size + 64, 64) + PAGE_SIZE - 1) >> PAGE_SHIFT) > end) ?
                1 : 0;
#endif

    ljmkTRACE_ZONE(ljvLEVEL_INFO, _LJ_OBJ_ZONE,
                   "%s(%d): pageCount: %d. extraPage: %d",
                   __func__, __LINE__, pageCount, extraPage);

    /* Overflow. */
    if ((memory + Size) < memory) {
        ljmkFOOTER_ARG("status=%d", ljvSTATUS_INVALID_ARGUMENT);
        return ljvSTATUS_INVALID_ARGUMENT;
    }

    memory = untagged_addr((ljtSIZE_T)Memory);

    if (memory) {
        ljtSIZE_T vaddr = memory;

        down_read(&current_mm_mmap_sem);
        vma = find_vma(current->mm, memory);
        up_read(&current_mm_mmap_sem);

        if (!vma)
            /* No such memory, or across vmas. */
            ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);

#ifdef CONFIG_ARM
        /* coherent cache in case vivt or vipt-aliasing cache. */
        __cpuc_flush_user_range(memory, memory + Size, vma->vm_flags);
#endif

        vm_flags = vma->vm_flags;
        vaddr = vma->vm_end;

        down_read(&current_mm_mmap_sem);
        while (vaddr < memory + Size) {
            vma = find_vma(current->mm, vaddr);

            if (!vma) {
                /* No such memory. */
                up_read(&current_mm_mmap_sem);
                ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
            }

            if ((vma->vm_flags & VM_PFNMAP) != (vm_flags & VM_PFNMAP)) {
                /*
                 * Can not support different map type:
                 * both PFN and PAGE detected.
                 */
                up_read(&current_mm_mmap_sem);
                ljmkONERROR(ljvSTATUS_NOT_SUPPORTED);
            }

            vaddr = vma->vm_end;
        }
        up_read(&current_mm_mmap_sem);
    }

    if (Physical != ljvINVALID_PHYSICAL_ADDRESS) {
        UserMemory->size = Size;
        /* Checking whether Physical is greater than 4G. */
        if ((Os->device->platform->flagBits & ljvPLATFORM_FLAG_LIMIT_4G_ADDRESS)
              && pfn_valid(PHYS_PFN(Physical))) {
            if (Physical > 0xFFFFFFFFu || Physical + Size > 0xFFFFFFFFu)
                ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
        }
        result = import_physical_map(Os, dev, UserMemory, Physical);
    } else {
        if (vm_flags & VM_PFNMAP)
            result = import_pfn_map(Os, dev, UserMemory, memory, pageCount);
        else
            result = import_page_map(Os, dev, UserMemory, memory, pageCount, Size, vm_flags);
    }

    if (result == -EINVAL)
        ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
    else if (result == -ENOMEM)
        ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);
    else if (result < 0)
        ljmkONERROR(ljvSTATUS_OUT_OF_RESOURCES);

    UserMemory->vm_flags = vm_flags;
    UserMemory->user_vaddr = (unsigned long)Memory;
    UserMemory->size = Size;
    UserMemory->offset = (Physical != ljvINVALID_PHYSICAL_ADDRESS) ?
                         (Physical & ~PAGE_MASK) : (memory & ~PAGE_MASK);

    UserMemory->pageCount = pageCount;
    UserMemory->extraPage = extraPage;

    /* Success. */
    ljmkFOOTER();
    return ljvSTATUS_OK;

OnError:
    ljmkFOOTER();
    return status;
}

static ljeSTATUS
_UserMemoryAttach(ljkALLOCATOR Allocator, ljsATTACH_DESC_PTR Desc, PLINUX_MDL Mdl)
{
    ljeSTATUS status;
    struct um_desc *userMemory = ljvNULL;

    ljkOS os = Allocator->os;

    ljmkHEADER();

    /* Handle is meangless for this importer. */
    ljmkVERIFY_ARGUMENT(Desc != ljvNULL);

    ljmkONERROR(ljkOS_Allocate(os, ljmSIZEOF(struct um_desc), (ljtPOINTER *)&userMemory));

    ljkOS_ZeroMemory(userMemory, ljmSIZEOF(struct um_desc));

    ljmkONERROR(_Import(os, Mdl, Desc->userMem.memory,
                        Desc->userMem.physical, Desc->userMem.size, userMemory));

    Mdl->priv = userMemory;
    Mdl->numPages = userMemory->pageCount + userMemory->extraPage;
    Mdl->contiguous = (userMemory->chunk_count == 1);

    ljmkFOOTER_NO();
    return ljvSTATUS_OK;

OnError:
    if (userMemory != ljvNULL)
        ljkOS_Free(os, (ljtPOINTER)userMemory);

    ljmkFOOTER();
    return status;
}

static void
release_physical_map(ljkOS Os, struct device *dev, struct um_desc *um)
{
    if (Os->iommu) {
        unsigned long pfn = um->physical >> PAGE_SHIFT;
        size_t size = um->size + um->offset;

        if (pfn_valid(pfn))
            dma_unmap_page(dev, um->dmaHandle, size, DMA_BIDIRECTIONAL);
        else
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 9, 0)
            dma_unmap_resource(dev, um->dmaHandle, size, DMA_BIDIRECTIONAL, 0);
#else
            dma_unmap_page(dev, um->dmaHandle, size, DMA_BIDIRECTIONAL);
#endif

        um->dmaHandle = 0;
    }
}

static void
release_page_map(ljkOS Os, struct device *dev, struct um_desc *um)
{
    int i;

    if (um->sgt.nents > 0) {
        dma_sync_sg_for_device(dev, um->sgt.sgl, um->sgt.nents, DMA_TO_DEVICE);

        dma_sync_sg_for_cpu(dev, um->sgt.sgl, um->sgt.nents, DMA_FROM_DEVICE);

        dma_unmap_sg(dev, um->sgt.sgl, um->sgt.nents, DMA_FROM_DEVICE);

        um->dmaHandle = 0;

#if ljdUSE_LINUX_SG_TABLE_API
        sg_free_table(&um->sgt);
#else
        kfree(um->sgt.sgl);
#endif
    }

    for (i = 0; i < um->pageCount; i++) {
        if (!PageReserved(um->pages[i]))
            SetPageDirty(um->pages[i]);

#if LINUX_VERSION_CODE > KERNEL_VERSION(5, 6, 0)
        unpin_user_page(um->pages[i]);
#else
        put_page(um->pages[i]);
#endif
    }

    free_memory(um->pages);
}

static void
release_pfn_map(ljkOS Os, struct device *dev, struct um_desc *um)
{
    int i;

    if (um->sgt.nents > 0) {
        dma_unmap_sg(dev, um->sgt.sgl, um->sgt.nents, DMA_FROM_DEVICE);

#if ljdUSE_LINUX_SG_TABLE_API
        sg_free_table(&um->sgt);
#else
        kfree(um->sgt.sgl);
#endif
    }

    um->dmaHandle = 0;

    for (i = 0; i < um->pageCount; i++) {
        if (pfn_valid(um->pfns[i])) {
            struct page *page = pfn_to_page(um->pfns[i]);

            if (!PageReserved(page))
                SetPageDirty(page);

            if (um->refs[i])
                put_page(page);
        }
    }

    free_memory(um->pfns);
    free_memory(um->refs);
}

static void
_UserMemoryFree(ljkALLOCATOR Allocator, PLINUX_MDL Mdl)
{
    ljkOS os = Allocator->os;
    struct um_desc *userMemory = Mdl->priv;
    struct device *dev = (struct device *)Mdl->device;

    ljmkHEADER();

    if (userMemory) {
        switch (userMemory->type) {
        case UM_PHYSICAL_MAP:
            release_physical_map(os, dev, userMemory);
            break;
        case UM_PAGE_MAP:
            release_page_map(os, dev, userMemory);
            break;
        case UM_PFN_MAP:
            release_pfn_map(os, dev, userMemory);
            break;
        }

        ljmkOS_SAFE_FREE(os, userMemory);
    }

    ljmkFOOTER_NO();
}

static ljeSTATUS
_UserMemoryMapUser(ljkALLOCATOR Allocator, PLINUX_MDL Mdl, PLINUX_MDL_MAP MdlMap, ljtBOOL Cacheable)
{
    struct um_desc *userMemory = Mdl->priv;

    MdlMap->vmaAddr = (ljtPOINTER)userMemory->user_vaddr;
    MdlMap->cacheable = ljvTRUE;

    return ljvSTATUS_OK;
}

static void
_UserMemoryUnmapUser(ljkALLOCATOR Allocator, PLINUX_MDL Mdl, PLINUX_MDL_MAP MdlMap, ljtUINT32 Size)
{
}

static ljeSTATUS
_UserMemoryMapKernel(ljkALLOCATOR Allocator, PLINUX_MDL Mdl,
                     ljtSIZE_T Offset, ljtSIZE_T Bytes, ljtPOINTER *Logical)
{
    /* Kernel doesn't access video memory. */
    return ljvSTATUS_NOT_SUPPORTED;
}

static ljeSTATUS
_UserMemoryUnmapKernel(ljkALLOCATOR Allocator, PLINUX_MDL Mdl, ljtPOINTER Logical)
{
    /* Kernel doesn't access video memory. */
    return ljvSTATUS_NOT_SUPPORTED;
}

static ljeSTATUS
_UserMemoryCache(ljkALLOCATOR Allocator, PLINUX_MDL Mdl, ljtSIZE_T Offset,
                 ljtPOINTER Logical, ljtSIZE_T Bytes, ljeCACHEOPERATION Operation)
{
    struct um_desc *um  = Mdl->priv;
    enum dma_data_direction dir;
    struct device *dev = (struct device *)Mdl->device;

    if (um->type == UM_PHYSICAL_MAP || um->alloc_from_res) {
        _MemoryBarrier();
        return ljvSTATUS_OK;
    }

    if (um->type == UM_PFN_MAP && um->sgt.nents <= 0) {
        _MemoryBarrier();
        return ljvSTATUS_OK;
    }

#ifdef CONFIG_ARM
    /* coherent cache in case vivt or vipt-aliasing cache. */
    __cpuc_flush_user_range(um->user_vaddr, um->user_vaddr + um->size, um->vm_flags);
#endif

    switch (Operation) {
    case ljvCACHE_CLEAN:
        dir = DMA_TO_DEVICE;
        dma_sync_sg_for_device(dev, um->sgt.sgl, um->sgt.nents, dir);
        break;
    case ljvCACHE_FLUSH:
        dir = DMA_TO_DEVICE;
        dma_sync_sg_for_device(dev, um->sgt.sgl, um->sgt.nents, dir);
        dir = DMA_FROM_DEVICE;
        dma_sync_sg_for_cpu(dev, um->sgt.sgl, um->sgt.nents, dir);
        break;
    case ljvCACHE_INVALIDATE:
        dir = DMA_FROM_DEVICE;
        dma_sync_sg_for_cpu(dev, um->sgt.sgl, um->sgt.nents, dir);
        break;
    default:
        return ljvSTATUS_INVALID_ARGUMENT;
    }

    return ljvSTATUS_OK;
}

static ljeSTATUS
_UserMemoryPhysical(ljkALLOCATOR Allocator, PLINUX_MDL Mdl,
                    unsigned long Offset, ljtPHYS_ADDR_T *Physical)
{
    ljkOS os = Allocator->os;
    struct um_desc *userMemory = Mdl->priv;
    unsigned long offset = Offset + userMemory->offset;
    ljtUINT32 offsetInPage = offset & ~PAGE_MASK;
    ljtUINT32 index = (ljtUINT32)(offset / PAGE_SIZE);

    if (index >= userMemory->pageCount) {
        if (index < userMemory->pageCount + userMemory->extraPage) {
            if (os->iommu)
                *Physical = os->iommu->paddingPageDmaHandle;
            else
                *Physical = page_to_phys(os->paddingPage);
        } else {
            return ljvSTATUS_INVALID_ARGUMENT;
        }
    } else {
        if (os->iommu) {
            *Physical = userMemory->dmaHandle + (ljtPHYS_ADDR_T)index * PAGE_SIZE;
        } else {
            switch (userMemory->type) {
            case UM_PHYSICAL_MAP:
                *Physical = userMemory->physical + (ljtPHYS_ADDR_T)index * PAGE_SIZE;
                break;
            case UM_PAGE_MAP:
                *Physical = page_to_phys(userMemory->pages[index]);
                break;
            case UM_PFN_MAP:
                *Physical = (ljtPHYS_ADDR_T)userMemory->pfns[index] << PAGE_SHIFT;
                break;
            }
        }
    }

    *Physical += offsetInPage;

    return ljvSTATUS_OK;
}

static ljeSTATUS
_UserMemoryGetSGT(
    IN ljkALLOCATOR Allocator,
    IN PLINUX_MDL Mdl,
    IN ljtSIZE_T Offset,
    IN ljtSIZE_T Bytes,
    OUT ljtPOINTER *SGT
    )
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,6,0)
    struct page ** pages = ljvNULL;
    struct sg_table *sgt = NULL;
    struct um_desc *um = Mdl->priv;

    ljeSTATUS status = ljvSTATUS_OK;
    ljtSIZE_T offset = 0;
    ljtSIZE_T numPages = 0;

    ljmkASSERT(Offset + Bytes <= Mdl->numPages << PAGE_SHIFT);

    pages = um->pages;
    numPages = um->pageCount;
    offset = um->offset;

    sgt = kmalloc(sizeof(struct sg_table), GFP_KERNEL | ljdNOWARN);
    if (!sgt)
    {
        ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);
    }

    if (sg_alloc_table_from_pages(sgt, pages, numPages, offset, Bytes, GFP_KERNEL) < 0)
    {
        ljmkONERROR(ljvSTATUS_GENERIC_IO);
    }

    *SGT = (ljtPOINTER)sgt;

OnError:
    if (ljmIS_ERROR(status) && sgt)
    {
        kfree(sgt);
    }

    return status;
#else
    return ljvSTATUS_NOT_SUPPORTED;
#endif
}

static void
_UserMemoryAllocatorDestructor(ljsALLOCATOR *Allocator)
{
    if (Allocator->privateData)
        kfree(Allocator->privateData);

    kfree(Allocator);
}

#if 0
static ljeSTATUS
_UserMemoryGetDmaAddress(ljkALLOCATOR Allocator, PLINUX_MDL Mdl,
                ljtUINT32 Offset, ljtPHYS_ADDR_T *DmaAddress)
{
    struct um_desc *userMemory = Mdl->priv;
    unsigned long offset = Offset + userMemory->offset;
    ljtUINT32 offsetInPage = offset & ~PAGE_MASK;
    ljtUINT32 index = offset / PAGE_SIZE;
    ljeSTATUS status = ljvSTATUS_OK;

    if (userMemory->dmaHandle) {
        *DmaAddress = userMemory->dmaHandle + index * PAGE_SIZE + offsetInPage;
    } else {
        status = _UserMemoryPhysical(Allocator, Mdl, Offset, DmaAddress);
        if (ljmIS_ERROR(status))
            return status;
    }

    if (Mdl->contiguous)
        return ljvSTATUS_OK;
    else {
        // pr_debug("[ljmcore]: %s: memory is not contiguous\n", __func__);
        return ljvSTATUS_INVALID_ADDRESS;
    }
}
#endif

/* User memory allocator (importer) operations. */
static ljsALLOCATOR_OPERATIONS UserMemoryAllocatorOperations = {
    .Attach      = _UserMemoryAttach,
    .Free        = _UserMemoryFree,
    .MapUser     = _UserMemoryMapUser,
    .UnmapUser   = _UserMemoryUnmapUser,
    .MapKernel   = _UserMemoryMapKernel,
    .UnmapKernel = _UserMemoryUnmapKernel,
    .Cache       = _UserMemoryCache,
    .Physical    = _UserMemoryPhysical,
    .GetSGT      = _UserMemoryGetSGT,
    .GetSGT      = _UserMemoryGetSGT,
    .GetDmaAddress = _UserMemoryPhysical,
};

/* Default allocator entry. */
ljeSTATUS
_UserMemoryAlloctorInit(ljkOS Os, ljsDEBUGFS_DIR *Parent, ljkALLOCATOR *Allocator)
{
    ljeSTATUS status;
    ljkALLOCATOR allocator;

    ljmkONERROR(ljkALLOCATOR_Construct(Os, &UserMemoryAllocatorOperations, &allocator));

    allocator->destructor = _UserMemoryAllocatorDestructor;

    allocator->capability = ljvALLOC_FLAG_USERMEMORY
                          | ljvALLOC_FLAG_32BIT_VA
                          | ljvALLOC_FLAG_PRIOR_32BIT_VA;

    *Allocator = allocator;

    return ljvSTATUS_OK;

OnError:
    return status;
}
