/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#include "lj_hal_kernel_precomp.h"
#include "lj_hal_kernel_debug.h"

/******************************************************************************
 ******************************* Debug Variables ******************************
 ******************************************************************************/

static ljeSTATUS _lastError = ljvSTATUS_OK;
//static ljtUINT32 _debugLevel = /*ljvLEVEL_ERROR*/ljvLEVEL_VERBOSE;
static ljtUINT32 _debugLevel = ljvLEVEL_ERROR;

/*
 * _debugZones config value
 * Please Reference define in lj_hal_base.h
 */
static ljtUINT32 _debugZones = ljdZONE_NONE;

/******************************************************************************
 ******************************** Debug Switches ******************************
 ******************************************************************************/

/*
 *   ljdTHREAD_BUFFERS
 *
 *   When greater then one, will accumulate messages from the specified number
 *   of threads in separate output buffers.
 */
#define ljdTHREAD_BUFFERS       1

/*
 *   ljdSHOW_LINE_NUMBER
 *
 *   When enabledm each print statement will be preceded with the current
 *   line number.
 */
#define ljdSHOW_LINE_NUMBER     0

/*
 *   ljdSHOW_PROCESS_ID
 *
 *   When enabledm each print statement will be preceded with the current
 *   process ID.
 */
#define ljdSHOW_PROCESS_ID      0

/*
 *   ljdSHOW_THREAD_ID
 *
 *   When enabledm each print statement will be preceded with the current
 *   thread ID.
 */
#define ljdSHOW_THREAD_ID       0

/*
 *   ljdSHOW_TIME
 *
 *   When enabled each print statement will be preceded with the current
 *   high-resolution time.
 */
#define ljdSHOW_TIME            0

/******************************************************************************
 ***************************** Private Structures *****************************
 ******************************************************************************/

typedef struct _ljsBUFFERED_OUTPUT *ljsBUFFERED_OUTPUT_PTR;
typedef struct _ljsBUFFERED_OUTPUT {
#if ljdTHREAD_BUFFERS > 1
    ljtUINT32 threadID;
#endif

#if ljdSHOW_LINE_NUMBER
    ljtUINT lineNumber;
#endif

    ljtINT indent;

    ljsBUFFERED_OUTPUT_PTR prev;
    ljsBUFFERED_OUTPUT_PTR next;
} ljsBUFFERED_OUTPUT;

static ljsBUFFERED_OUTPUT _outputBuffer[ljdTHREAD_BUFFERS];
static ljsBUFFERED_OUTPUT_PTR _outputBufferHead = ljvNULL;
static ljsBUFFERED_OUTPUT_PTR _outputBufferTail = ljvNULL;

/******************************************************************************
 ****************************** Printing Functions ****************************
 ******************************************************************************/

static int
_AppendIndent(ljtINT Indent, char *Buffer, int BufferSize)
{
    ljtINT i;

    ljtINT len = 0;
    ljtINT indent = Indent % 40;

    for (i = 0; i < indent; i += 1)
        Buffer[len++] = ' ';

    if (indent != Indent) {
        len += ljmkSPRINTF(Buffer + len, BufferSize - len, " <%d> ", Indent);

        Buffer[len] = '\0';
    }

    return len;
}

static ljtUINT
_PrintString(ljsBUFFERED_OUTPUT_PTR OutputBuffer, ljtINT Indent,
             ljtCONST_STRING Message, ljtPOINTER Data, char Buffer[], ljtUINT Size)
{
    ljtINT len;

    /* Append the indent string. */
    len = _AppendIndent(Indent, Buffer, Size);

    /* Format the string. */
    len += ljmkVSPRINTF(Buffer + len, Size - len, Message, Data);
    Buffer[len] = '\0';

    /* Add end-of-line if missing. */
    if (Buffer[len - 1] != '\n') {
        Buffer[len++] = '\n';
        Buffer[len] = '\0';
    }

    return (ljtUINT)len;
}

/******************************************************************************
 ****************************** Private Functions *****************************
 ******************************************************************************/

static ljmkINLINE void
_InitBuffers(void)
{
    int i;

    if (_outputBufferHead == ljvNULL) {
        for (i = 0; i < ljdTHREAD_BUFFERS; i += 1) {
            if (_outputBufferTail == ljvNULL)
                _outputBufferHead = &_outputBuffer[i];
            else
                _outputBufferTail->next = &_outputBuffer[i];

#if ljdTHREAD_BUFFERS > 1
            _outputBuffer[i].threadID = ~0U;
#endif

            _outputBuffer[i].prev = _outputBufferTail;
            _outputBuffer[i].next = ljvNULL;

            _outputBufferTail = &_outputBuffer[i];
        }
    }
}

static ljmkINLINE ljsBUFFERED_OUTPUT_PTR
_GetOutputBuffer(void)
{
    ljsBUFFERED_OUTPUT_PTR outputBuffer;

#if ljdTHREAD_BUFFERS > 1
    /* Get the current thread ID. */
    ljtUINT32 ThreadID = ljmkGETTHREADID();

    /* Locate the output buffer for the thread. */
    outputBuffer = _outputBufferHead;

    while (outputBuffer != ljvNULL) {
        if (outputBuffer->threadID == ThreadID)
            break;

        outputBuffer = outputBuffer->next;
    }

    /* No matching buffer found? */
    if (outputBuffer == ljvNULL) {
        /* Get the tail for the buffer. */
        outputBuffer = _outputBufferTail;

        /* Move it to the head. */
        _outputBufferTail = _outputBufferTail->prev;
        _outputBufferTail->next = ljvNULL;

        outputBuffer->prev = ljvNULL;
        outputBuffer->next = _outputBufferHead;

        _outputBufferHead->prev = outputBuffer;
        _outputBufferHead = outputBuffer;

        /* Reset the buffer. */
        outputBuffer->threadID = ThreadID;
#if ljdSHOW_LINE_NUMBER
        outputBuffer->lineNumber = 0;
# endif
    }
#else
    outputBuffer = _outputBufferHead;
#endif  /* #if ljdTHREAD_BUFFERS > 1 */

    return outputBuffer;
}

static void
_Print(ljtCONST_STRING Message, ljtARGUMENTS *Arguments)
{
    ljsBUFFERED_OUTPUT_PTR outputBuffer;
    char buffer[256];
    char *ptr = buffer;
    ljtINT len = 0;

    static ljmkDECLARE_MUTEX(printMutex);

    ljmkMUTEX_LOCK(printMutex);

    /* Initialize output buffer list. */
    _InitBuffers();

    /* Locate the proper output buffer. */
    outputBuffer = _GetOutputBuffer();

    /* Form the indent string. */
    if (Message[0] == '-' && Message[1] == '-')
        outputBuffer->indent -= 2;

    /* Print the message. */
    len += _PrintString(outputBuffer,
                        outputBuffer->indent,
                        Message, ((ljtPOINTER)Arguments), ptr,
                        ljmSIZEOF(buffer) - outputBuffer->indent - len);

    ljmkOUTPUT_STRING(buffer);

    /* Check increasing indent. */
    if (Message[0] == '+' && Message[1] == '+')
        outputBuffer->indent += 2;

    ljmkMUTEX_UNLOCK(printMutex);
}

/******************************************************************************
 ******************************** Debug Macros ********************************
 ******************************************************************************/

#ifdef __QNXNTO__

extern volatile unsigned int g_nQnxInIsrs;

# define ljmDEBUGPRINT(Message)                               \
    {                                                         \
        if (atomic_add_value(&g_nQnxInIsrs, 1) == 0) {        \
            ljtARGUMENTS __arguments__;                       \
            ljmkARGUMENTS_START(__arguments__, Message);      \
            _Print(Message, &__arguments__);                  \
            ljmkARGUMENTS_END(__arguments__);                 \
        }                                                     \
        atomic_sub(&g_nQnxInIsrs, 1);                         \
    }

#elif defined(__VXWORKS__)
# define ljmDEBUGPRINT(Message)                               \
    {                                                         \
        printf(Message);                                      \
    }

#else

# define ljmDEBUGPRINT(Message)                               \
    {                                                         \
        ljtARGUMENTS __arguments__;                           \
        ljmkARGUMENTS_START(__arguments__, Message);          \
        _Print(Message, &__arguments__);                      \
        ljmkARGUMENTS_END(__arguments__);                     \
    }

#endif

/******************************************************************************
 ********************************* Debug Code *********************************
 ******************************************************************************/

/*******************************************************************************
 **
 **  ljkOS_Print
 **
 **  Send a message to the debugger.
 **
 **  INPUT:
 **
 **      ljtCONST_STRING Message
 **          Pointer to message.
 **
 **      ...
 **          Optional arguments.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */

void
ljkOS_Print(ljtCONST_STRING Message, ...)
{
    ljmDEBUGPRINT(Message);
}

/*******************************************************************************
 **
 **  ljkOS_DebugTrace
 **
 **  Send a leveled message to the debugger.
 **
 **  INPUT:
 **
 **      ljtUINT32 Level
 **          Debug level of message.
 **
 **      ljtCONST_STRING Message
 **          Pointer to message.
 **
 **      ...
 **          Optional arguments.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */

void
ljkOS_DebugTrace(ljtUINT32 Level, ljtCONST_STRING Message, ...)
{
    if (Level > _debugLevel)
        return;

    ljmDEBUGPRINT(Message);
}

/*******************************************************************************
 **
 **  ljkOS_DebugTraceZone
 **
 **  Send a leveled and zoned message to the debugger.
 **
 **  INPUT:
 **
 **      ljtUINT32 Level
 **          Debug level for message.
 **
 **      ljtUINT32 Zone
 **          Debug zone for message.
 **
 **      ljtCONST_STRING Message
 **          Pointer to message.
 **
 **      ...
 **          Optional arguments.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */

void
ljkOS_DebugTraceZone(ljtUINT32 Level, ljtUINT32 Zone, ljtCONST_STRING Message, ...)
{
    if (Level > _debugLevel || !(Zone & _debugZones))
        return;

    ljmDEBUGPRINT(Message);
}

/*******************************************************************************
 **
 **  ljkOS_DebugBreak
 **
 **  Break into the debugger.
 **
 **  INPUT:
 **
 **      Nothing.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
void
ljkOS_DebugBreak(void)
{
    ljkOS_DebugTrace(ljvLEVEL_ERROR, "%s(%d)", __func__, __LINE__);
}

/*******************************************************************************
 **
 **  ljkOS_DebugFatal
 **
 **  Send a message to the debugger and break into the debugger.
 **
 **  INPUT:
 **
 **      ljtCONST_STRING Message
 **          Pointer to message.
 **
 **      ...
 **          Optional arguments.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
void
ljkOS_DebugFatal(ljtCONST_STRING Message, ...)
{
    ljmkPRINT_VERSION();
    ljmDEBUGPRINT(Message);

    /* Break into the debugger. */
    ljkOS_DebugBreak();
}

/*******************************************************************************
 **
 **  ljkOS_SetDebugLevel
 **
 **  Set the debug level.
 **
 **  INPUT:
 **
 **      ljtUINT32 Level
 **          New debug level.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */

void
ljkOS_SetDebugLevel(ljtUINT32 Level)
{
    _debugLevel = Level;
}

/*******************************************************************************
 **
 **  ljkOS_SetDebugZone
 **
 **  Set the debug zone.
 **
 **  INPUT:
 **
 **      ljtUINT32 Zone
 **          New debug zone.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
void
ljkOS_SetDebugZone(ljtUINT32 Zone)
{
    _debugZones = Zone;
}

/*******************************************************************************
 **
 **  ljkOS_SetDebugLevelZone
 **
 **  Set the debug level and zone.
 **
 **  INPUT:
 **
 **      ljtUINT32 Level
 **          New debug level.
 **
 **      ljtUINT32 Zone
 **          New debug zone.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */

void
ljkOS_SetDebugLevelZone(ljtUINT32 Level, ljtUINT32 Zone)
{
    _debugLevel = Level;
    _debugZones = Zone;
}

/*******************************************************************************
 **
 **  ljkOS_SetDebugZones
 **
 **  Enable or disable debug zones.
 **
 **  INPUT:
 **
 **      ljtUINT32 Zones
 **          Debug zones to enable or disable.
 **
 **      ljtBOOL Enable
 **          Set to ljvTRUE to enable the zones (or the Zones with the current
 **          zones) or ljvFALSE to disable the specified Zones.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */

void
ljkOS_SetDebugZones(ljtUINT32 Zones, ljtBOOL Enable)
{
    if (Enable) {
        /* Enable the zones. */
        _debugZones |= Zones;
    } else {
        /* Disable the zones. */
        _debugZones &= ~Zones;
    }
}

/*******************************************************************************
 **
 **  ljkOS_Verify
 **
 **  Called to verify the result of a function call.
 **
 **  INPUT:
 **
 **      ljeSTATUS Status
 **          Function call result.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */

void
ljkOS_Verify(ljeSTATUS status)
{
    _lastError = status;
}

ljtCONST_STRING
ljkOS_DebugStatus2Name(ljeSTATUS status)
{
    switch (status) {
    case ljvSTATUS_OK:
        return "ljvSTATUS_OK";
    case ljvSTATUS_TRUE:
        return "ljvSTATUS_TRUE";
    case ljvSTATUS_NO_MORE_DATA:
        return "ljvSTATUS_NO_MORE_DATA";
    case ljvSTATUS_CACHED:
        return "ljvSTATUS_CACHED";
    case ljvSTATUS_MIPMAP_TOO_LARGE:
        return "ljvSTATUS_MIPMAP_TOO_LARGE";
    case ljvSTATUS_NAME_NOT_FOUND:
        return "ljvSTATUS_NAME_NOT_FOUND";
    case ljvSTATUS_NOT_OUR_INTERRUPT:
        return "ljvSTATUS_NOT_OUR_INTERRUPT";
    case ljvSTATUS_MISMATCH:
        return "ljvSTATUS_MISMATCH";
    case ljvSTATUS_MIPMAP_TOO_SMALL:
        return "ljvSTATUS_MIPMAP_TOO_SMALL";
    case ljvSTATUS_LARGER:
        return "ljvSTATUS_LARGER";
    case ljvSTATUS_SMALLER:
        return "ljvSTATUS_SMALLER";
    case ljvSTATUS_CHIP_NOT_READY:
        return "ljvSTATUS_CHIP_NOT_READY";
    case ljvSTATUS_NEED_CONVERSION:
        return "ljvSTATUS_NEED_CONVERSION";
    case ljvSTATUS_SKIP:
        return "ljvSTATUS_SKIP";
    case ljvSTATUS_DATA_TOO_LARGE:
        return "ljvSTATUS_DATA_TOO_LARGE";
    case ljvSTATUS_INVALID_CONFIG:
        return "ljvSTATUS_INVALID_CONFIG";
    case ljvSTATUS_CHANGED:
        return "ljvSTATUS_CHANGED";
    case ljvSTATUS_NOT_SUPPORT_DITHER:
        return "ljvSTATUS_NOT_SUPPORT_DITHER";

    case ljvSTATUS_INVALID_ARGUMENT:
        return "ljvSTATUS_INVALID_ARGUMENT";
    case ljvSTATUS_INVALID_OBJECT:
        return "ljvSTATUS_INVALID_OBJECT";
    case ljvSTATUS_OUT_OF_MEMORY:
        return "ljvSTATUS_OUT_OF_MEMORY";
    case ljvSTATUS_MEMORY_LOCKED:
        return "ljvSTATUS_MEMORY_LOCKED";
    case ljvSTATUS_MEMORY_UNLOCKED:
        return "ljvSTATUS_MEMORY_UNLOCKED";
    case ljvSTATUS_HEAP_CORRUPTED:
        return "ljvSTATUS_HEAP_CORRUPTED";
    case ljvSTATUS_GENERIC_IO:
        return "ljvSTATUS_GENERIC_IO";
    case ljvSTATUS_INVALID_ADDRESS:
        return "ljvSTATUS_INVALID_ADDRESS";
    case ljvSTATUS_CONTEXT_LOSSED:
        return "ljvSTATUS_CONTEXT_LOSSED";
    case ljvSTATUS_TOO_COMPLEX:
        return "ljvSTATUS_TOO_COMPLEX";
    case ljvSTATUS_BUFFER_TOO_SMALL:
        return "ljvSTATUS_BUFFER_TOO_SMALL";
    case ljvSTATUS_INTERFACE_ERROR:
        return "ljvSTATUS_INTERFACE_ERROR";
    case ljvSTATUS_NOT_SUPPORTED:
        return "ljvSTATUS_NOT_SUPPORTED";
    case ljvSTATUS_MORE_DATA:
        return "ljvSTATUS_MORE_DATA";
    case ljvSTATUS_TIMEOUT:
        return "ljvSTATUS_TIMEOUT";
    case ljvSTATUS_OUT_OF_RESOURCES:
        return "ljvSTATUS_OUT_OF_RESOURCES";
    case ljvSTATUS_INVALID_DATA:
        return "ljvSTATUS_INVALID_DATA";
    case ljvSTATUS_INVALID_MIPMAP:
        return "ljvSTATUS_INVALID_MIPMAP";
    case ljvSTATUS_NOT_FOUND:
        return "ljvSTATUS_NOT_FOUND";
    case ljvSTATUS_NOT_ALIGNED:
        return "ljvSTATUS_NOT_ALIGNED";
    case ljvSTATUS_INVALID_REQUEST:
        return "ljvSTATUS_INVALID_REQUEST";
    case ljvSTATUS_GPU_NOT_RESPONDING:
        return "ljvSTATUS_GPU_NOT_RESPONDING";
    case ljvSTATUS_TIMER_OVERFLOW:
        return "ljvSTATUS_TIMER_OVERFLOW";
    case ljvSTATUS_VERSION_MISMATCH:
        return "ljvSTATUS_VERSION_MISMATCH";
    case ljvSTATUS_LOCKED:
        return "ljvSTATUS_LOCKED";
    case ljvSTATUS_INTERRUPTED:
        return "ljvSTATUS_INTERRUPTED";
    case ljvSTATUS_DEVICE:
        return "ljvSTATUS_DEVICE";
    case ljvSTATUS_NOT_MULTI_PIPE_ALIGNED:
        return "ljvSTATUS_NOT_MULTI_PIPE_ALIGNED";

    /* Linker errors. */
    case ljvSTATUS_GLOBAL_TYPE_MISMATCH:
        return "ljvSTATUS_GLOBAL_TYPE_MISMATCH";
    case ljvSTATUS_TOO_MANY_ATTRIBUTES:
        return "ljvSTATUS_TOO_MANY_ATTRIBUTES";
    case ljvSTATUS_TOO_MANY_UNIFORMS:
        return "ljvSTATUS_TOO_MANY_UNIFORMS";
    case ljvSTATUS_TOO_MANY_VARYINGS:
        return "ljvSTATUS_TOO_MANY_VARYINGS";
    case ljvSTATUS_UNDECLARED_VARYING:
        return "ljvSTATUS_UNDECLARED_VARYING";
    case ljvSTATUS_VARYING_TYPE_MISMATCH:
        return "ljvSTATUS_VARYING_TYPE_MISMATCH";
    case ljvSTATUS_MISSING_MAIN:
        return "ljvSTATUS_MISSING_MAIN";
    case ljvSTATUS_NAME_MISMATCH:
        return "ljvSTATUS_NAME_MISMATCH";
    case ljvSTATUS_INVALID_INDEX:
        return "ljvSTATUS_INVALID_INDEX";
    case ljvSTATUS_UNIFORM_MISMATCH:
        return "ljvSTATUS_UNIFORM_MISMATCH";
    case ljvSTATUS_UNSAT_LIB_SYMBOL:
        return "ljvSTATUS_UNSAT_LIB_SYMBOL";
    case ljvSTATUS_TOO_MANY_SHADERS:
        return "ljvSTATUS_TOO_MANY_SHADERS";
    case ljvSTATUS_LINK_INVALID_SHADERS:
        return "ljvSTATUS_LINK_INVALID_SHADERS";
    case ljvSTATUS_CS_NO_WORKGROUP_SIZE:
        return "ljvSTATUS_CS_NO_WORKGROUP_SIZE";
    case ljvSTATUS_LINK_LIB_ERROR:
        return "ljvSTATUS_LINK_LIB_ERROR";
    case ljvSTATUS_SHADER_VERSION_MISMATCH:
        return "ljvSTATUS_SHADER_VERSION_MISMATCH";
    case ljvSTATUS_TOO_MANY_INSTRUCTION:
        return "ljvSTATUS_TOO_MANY_INSTRUCTION";
    case ljvSTATUS_SSBO_MISMATCH:
        return "ljvSTATUS_SSBO_MISMATCH";
    case ljvSTATUS_TOO_MANY_OUTPUT:
        return "ljvSTATUS_TOO_MANY_OUTPUT";
    case ljvSTATUS_TOO_MANY_INPUT:
        return "ljvSTATUS_TOO_MANY_INPUT";
    case ljvSTATUS_NOT_SUPPORT_CL:
        return "ljvSTATUS_NOT_SUPPORT_CL";
    case ljvSTATUS_NOT_SUPPORT_INTEGER:
        return "ljvSTATUS_NOT_SUPPORT_INTEGER";
    case ljvSTATUS_UNIFORM_TYPE_MISMATCH:
        return "ljvSTATUS_UNIFORM_TYPE_MISMATCH";
    case ljvSTATUS_MISSING_PRIMITIVE_TYPE:
        return "ljvSTATUS_MISSING_PRIMITIVE_TYPE";
    case ljvSTATUS_MISSING_OUTPUT_VERTEX_COUNT:
        return "ljvSTATUS_MISSING_OUTPUT_VERTEX_COUNT";
    case ljvSTATUS_NON_INVOCATION_ID_AS_INDEX:
        return "ljvSTATUS_NON_INVOCATION_ID_AS_INDEX";
    case ljvSTATUS_INPUT_ARRAY_SIZE_MISMATCH:
        return "ljvSTATUS_INPUT_ARRAY_SIZE_MISMATCH";
    case ljvSTATUS_OUTPUT_ARRAY_SIZE_MISMATCH:
        return "ljvSTATUS_OUTPUT_ARRAY_SIZE_MISMATCH";

    /* Compiler errors. */
    case ljvSTATUS_COMPILER_FE_PREPROCESSOR_ERROR:
        return "ljvSTATUS_COMPILER_FE_PREPROCESSOR_ERROR";
    case ljvSTATUS_COMPILER_FE_PARSER_ERROR:
        return "ljvSTATUS_COMPILER_FE_PARSER_ERROR";
    default:
        return "nil";
    }
}

/******************************************************************************
 *******************************Kernel Dump************************************
 ******************************************************************************/

#ifndef ljmkDUMP_STRING
# define ljmkDUMP_STRING(os, s) ljmkOUTPUT_STRING((s))
#endif

static ljmkDECLARE_MUTEX(_dumpMutex);
static ljtCHAR _dumpStorage[512];

/*******************************************************************************
 **
 **  ljkOS_Dump
 **
 **  Formated print string to dump pool.
 **
 **  INPUT:
 **
 **      ljtCONST_STRING Format
 **          String format.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
void
ljkOS_Dump(ljkOS Os, ljtCONST_STRING Format, ...)
{
    char buffer[256];
    ljtINT len;
    ljtARGUMENTS args;

    ljmkARGUMENTS_START(args, Format);
    len = ljmkVSPRINTF(buffer, ljmSIZEOF(buffer) - 2, Format, &args);
    ljmkARGUMENTS_END(args);

    if (len > 0) {
        if (buffer[len - 1] != '\n') {
            buffer[len] = '\n';
            buffer[len + 1] = '\0';
        }

        ljmkMUTEX_LOCK(_dumpMutex);
        ljmkDUMP_STRING(Os, buffer);
        ljmkMUTEX_UNLOCK(_dumpMutex);
    }
}

static void
_DumpUserString(ljkOS Os, ljtPOINTER UserStr, ljtSIZE_T Size)
{
    ljeSTATUS status = ljvSTATUS_OK;
    ljtSIZE_T offset = 0;
    ljtSIZE_T length = 0;
    ljtBOOL needCopy = ljvTRUE;
    const ljtSIZE_T maxLength = ljmSIZEOF(_dumpStorage) - 1;

    ljmkVERIFY_OK(ljkOS_QueryNeedCopy(Os, 0, &needCopy));

    ljmkMUTEX_LOCK(_dumpMutex);

    while (offset < Size) {
        length = maxLength < (Size - offset) ? maxLength : (Size - offset);

        /* Copy or map from user. */
        if (needCopy) {
            ljmkONERROR(ljkOS_CopyFromUserData(Os, _dumpStorage, UserStr, length));
        } else {
            ljtPOINTER ptr = ljvNULL;

            ljmkONERROR(ljkOS_MapUserPointer(Os, UserStr, length, (ljtPOINTER *)&ptr));

            ljkOS_MemCopy(_dumpStorage, ptr, length);
            ljkOS_UnmapUserPointer(Os, UserStr, length, ptr);
        }

        _dumpStorage[length] = '\0';
        ljmkDUMP_STRING(Os, _dumpStorage);

        UserStr = (ljtUINT8_PTR)UserStr + length;
        offset += length;
    }

    ljmkDUMP_STRING(Os, "\n");

OnError:
    ljmkMUTEX_UNLOCK(_dumpMutex);
}

static void
_DumpDataBuffer(ljkOS Os, ljeDUMP_BUFFER_TYPE Type,
                ljtPOINTER Data, ljtUINT64 Address, ljtSIZE_T Size)
{
    ljeSTATUS status = ljvSTATUS_OK;
    ljtSIZE_T offset = 0;
    ljtSIZE_T length = 0;
    ljtBOOL needCopy = ljvTRUE;
    ljtCONST_STRING dumpTag;

    char buffer[256];
    const ljtSIZE_T maxLength = ljmSIZEOF(_dumpStorage);

    switch (Type) {
    case ljvDUMP_BUFFER_VERIFY:
        dumpTag = "verify";
        break;
    case ljvDUMP_BUFFER_PHYSICAL_MEMORY:
        dumpTag = "physical";
        break;
    default:
        dumpTag = "memory";
        break;
    }

    if (Type <= ljvDUMP_BUFFER_USER_TYPE_LAST)
        ljmkVERIFY_OK(ljkOS_QueryNeedCopy(Os, 0, &needCopy));

    ljmkMUTEX_LOCK(_dumpMutex);

    /* Form and print the opening string. */
    if (Type == ljvDUMP_BUFFER_PHYSICAL_MEMORY) {
        ljmkSPRINTF(buffer, ljmSIZEOF(buffer) - 1,
                    "@[%s 0x%llx 0x%08X\n",
                    dumpTag, Address, (ljtUINT32)Size);
    } else {
        ljmkSPRINTF(buffer, ljmSIZEOF(buffer) - 1,
                    "@[%s 0x%llx 0x%08X\n",
                    dumpTag, Address, (ljtUINT32)Size);
    }

    ljmkDUMP_STRING(Os, buffer);

    while (offset < Size) {
        ljtPOINTER data = ljvNULL;
        ljtUINT32_PTR ptr;
        ljtUINT8_PTR bytePtr;
        ljtSIZE_T count, tailByteCount;

        length = maxLength < (Size - offset) ? maxLength : (Size - offset);

        count = length / 4;
        tailByteCount = length % 4;

        ptr = (ljtUINT32_PTR)Data;

        if (Type <= ljvDUMP_BUFFER_USER_TYPE_LAST) {
            /* Copy or map from user. */
            if (needCopy) {
                ljmkONERROR(ljkOS_CopyFromUserData(Os, _dumpStorage, Data, length));

                ptr = (ljtUINT32_PTR)_dumpStorage;
            } else {
                ljmkONERROR(ljkOS_MapUserPointer(Os, Data, length, (ljtPOINTER *)&data));

                ptr = (ljtUINT32_PTR)data;
            }
        }

        while (count >= 4) {
            ljmkSPRINTF(buffer, ljmSIZEOF(buffer) - 1,
                        "  0x%08X 0x%08X 0x%08X 0x%08X\n",
                        ptr[0], ptr[1], ptr[2], ptr[3]);

            ptr += 4;
            count -= 4;

            ljmkDUMP_STRING(Os, buffer);
        }

        switch (count) {
        case 3:
            ljmkSPRINTF(buffer, ljmSIZEOF(buffer) - 1,
                        "  0x%08X 0x%08X 0x%08X",
                        ptr[0], ptr[1], ptr[2]);
            break;
        case 2:
            ljmkSPRINTF(buffer, ljmSIZEOF(buffer) - 1,
                        "  0x%08X 0x%08X", ptr[0], ptr[1]);
            break;
        case 1:
            ljmkSPRINTF(buffer, ljmSIZEOF(buffer) - 1,
                        "  0x%08X", ptr[0]);
            break;
        }

        if (count > 0)
            ljmkDUMP_STRING(Os, buffer);

        bytePtr = (ljtUINT8_PTR)(ptr + count);

        if (!count && tailByteCount) {
            /* There is an extra space for the new line. */
            ljmkDUMP_STRING(Os, " ");
        }

        switch (tailByteCount) {
        case 3:
            ljmkSPRINTF(buffer, ljmSIZEOF(buffer) - 1,
                        " 0x00%02X%02X%02X",
                        bytePtr[2], bytePtr[1], bytePtr[0]);
            break;
        case 2:
            ljmkSPRINTF(buffer, ljmSIZEOF(buffer) - 1,
                        " 0x0000%02X%02X", bytePtr[1], bytePtr[0]);
            break;
        case 1:
            ljmkSPRINTF(buffer, ljmSIZEOF(buffer) - 1,
                        " 0x000000%02X", bytePtr[0]);
            break;
        }

        if (tailByteCount)
            ljmkDUMP_STRING(Os, buffer);

        if (count || tailByteCount)
            ljmkDUMP_STRING(Os, "\n");

        if (Type <= ljvDUMP_BUFFER_USER_TYPE_LAST && !needCopy)
            ljkOS_UnmapUserPointer(Os, Data, length, data);

        /* advance to next batch. */
        Data = (ljtUINT8_PTR)Data + length;
        offset += length;
    }

OnError:
    ljmkSPRINTF(buffer, ljmSIZEOF(buffer) - 1, "] -- %s\n", dumpTag);
    ljmkDUMP_STRING(Os, buffer);

    ljmkMUTEX_UNLOCK(_dumpMutex);
}

/*******************************************************************************
 **
 **  ljkOS_DumpBuffer
 **
 **  Print the contents of the specified buffer.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to ljkOS object.
 **
 **      ljeDUMP_BUFFER_TYPE Type
 **          Buffer type.
 **
 **      ljtPOINTER Buffer
 **          Pointer to the buffer to print.
 **
 **      ljtUINT64 Address
 **          Address.
 **
 **      ljtUINT Size
 **          Size of the buffer.
 **
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
void
ljkOS_DumpBuffer(ljkOS Os, ljeDUMP_BUFFER_TYPE Type,
                 ljtPOINTER Buffer, ljtUINT64 Address, ljtSIZE_T Size)
{
    if (!Buffer)
        return;

    /* memory dump below. */
    if (Type >= ljvDUMP_BUFFER_TYPE_COUNT) {
        ljmkPRINT("#[ERROR: invalid buffer type]\n");
        return;
    }

    if (Type == ljvDUMP_BUFFER_USER_STRING)
        _DumpUserString(Os, Buffer, Size);
    else
        _DumpDataBuffer(Os, Type, Buffer, Address, Size);
}
