/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#include <linux/init.h>
#include <linux/hardirq.h>
#include <linux/spinlock.h>
#include <linux/rcupdate.h>
#include <linux/mutex.h>
#include <linux/jiffies.h>
#include <linux/rculist.h>
#include <linux/idr.h>
#include "lj_hal_kernel_dma_engine.h"
#include "lj_hal_kernel_debug.h"

#define _LJ_OBJ_ZONE    ljvZONE_DEVICE

static DEFINE_MUTEX(dmaListMutex);
static DEFINE_IDA(dmaIda);
static LIST_HEAD(dmaDeviceList);
static ljtUINT32 dmaEngineRef = 0;

static ljeSTATUS _DmaChannelGet(ljsDMA_CHANNEL *Chan)
{
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkHEADER_ARG("chan=%p", Chan);

    if (!Chan)
    {
        ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
    }

    if (Chan->clientCount)
    {
        goto OnError;
    }

    if (Chan->device->ljkDMADEV_AllocateChannelResource)
    {
        ljmkONERROR(Chan->device->ljkDMADEV_AllocateChannelResource(Chan));
    }
    Chan->clientCount++;
    ljmkFOOTER();
    return ljvSTATUS_OK;
OnError:
    ljmkFOOTER();
    return status;
}

static ljeSTATUS _DmaChannelPut(ljsDMA_CHANNEL *Chan)
{
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkHEADER_ARG("chan=%p", Chan);

    if (!Chan)
    {
        ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
    }

    if (!Chan->clientCount)
    {
        ljmkONERROR(ljvSTATUS_INVALID_DATA);
    }

    Chan->clientCount--;
    if (!Chan->clientCount && Chan->device->ljkDMADEV_FreeChannelResource)
    {
        ljkDMAENGINE_Synchronize(Chan);
        Chan->device->ljkDMADEV_FreeChannelResource(Chan);
    }
OnError:
    ljmkFOOTER();
    return ljvSTATUS_OK;
}

static ljeDMA_STATUS _DmaIsTxComplete(
    ljsDMA_CHANNEL *Chan,
    ljdDMA_TX_IDENTIFIER Identifier,
    ljdDMA_TX_IDENTIFIER *lastCompleted,
    ljdDMA_TX_IDENTIFIER *lastIssued
    )
{
    ljsDMA_TX_STATE state;
    ljeDMA_STATUS dmaStatus;

    dmaStatus = ljkDMAENGINE_TxStatus(Chan, Identifier, &state);
    if (lastCompleted)
    {
        *lastCompleted = state.lastCompleted;
    }

    if (lastIssued)
    {
        *lastIssued = state.lastIssued;
    }

    return dmaStatus;
}

ljeDMA_STATUS ljkDMA_SyncWait(
    ljsDMA_CHANNEL *Chan,
    ljdDMA_TX_IDENTIFIER Identifier,
    ljtSIZE_T Timeout
    )
{
    ljeDMA_STATUS dmaStatus;
    ljtSIZE_T waitTimeout = jiffies + msecs_to_jiffies(Timeout);

    ljkDMAENGINE_IssuePending(Chan);

    do {
        dmaStatus = _DmaIsTxComplete(Chan, Identifier, NULL, NULL);
        if (Timeout && time_after_eq(jiffies, waitTimeout))
        {
            return ljeDMA_STATUS_ERROR;
        }
        if (dmaStatus != ljeDMA_STATUS_IN_PROGRESS)
        {
            break;
        }
        cpu_relax();
    } while (1);

    return dmaStatus;
}

ljeSTATUS ljkDMA_IssuePendingAll(void)
{
    ljsDMA_DEVICE *dmaDev;
    ljsDMA_CHANNEL *dmaChan;

    rcu_read_lock();
    list_for_each_entry_rcu(dmaDev, &dmaDeviceList, glabolDevNode)
    {
        if (ljkDMA_HasCapMask(&dmaDev->devCapMasks, ljeDMA_TX_TYPE_PRIVATE))
        {
            continue;
        }
        list_for_each_entry(dmaChan, &dmaDev->chanHead, devNode)
        {
            if (dmaChan->clientCount)
            {
                dmaDev->ljkDMADEV_IssuePending(dmaChan);
            }
        }
    }
    rcu_read_unlock();
    return ljvSTATUS_OK;
}

ljeSTATUS ljkDMA_GetSlaveCaps(ljsDMA_CHANNEL *Chan, ljsDMA_SLAVE_CAPS *Caps)
{
    ljsDMA_DEVICE *dmaDev;
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkHEADER_ARG("Chan=%p Caps=%p", Chan, Caps);

    if (!Chan || !Caps)
    {
        ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
    }

    dmaDev = Chan->device;
    if (!(ljkDMA_HasCapMask(&dmaDev->devCapMasks, ljeDMA_TX_TYPE_SLAVE)
        || ljkDMA_HasCapMask(&dmaDev->devCapMasks, ljeDMA_TX_TYPE_CYCLIC)))
    {
        ljmkONERROR(ljvSTATUS_NOT_SUPPORTED);
    }

    if (!dmaDev->caps.directions)
    {
        ljmkONERROR(ljvSTATUS_NOT_SUPPORTED);
    }

    *Caps = dmaDev->caps;
    ljmkFOOTER();
    return ljvSTATUS_OK;
OnError:
    ljmkFOOTER();
    return status;
}

static ljsDMA_CHANNEL *_DmaPrivateCandidate(
    ljsDMA_DEVICE *dmaDev,
    const ljuDMA_CAP_MASK *mask,
    ljkDMAENGINE_FilterFunc func,
    void *funcParam
    )
{
    ljsDMA_CHANNEL *chan;

    if (mask && ((dmaDev->devCapMasks.dmaCapMasks | mask->dmaCapMasks)
        != dmaDev->devCapMasks.dmaCapMasks))
    {
        ljmkPRINT("dma_engine: %s caps mask 0x%08x, but request 0x%08x.\n",
                dmaDev->name,
                dmaDev->devCapMasks.dmaCapMasks,
                mask->dmaCapMasks);
        return ljvNULL;
    }

    if (dmaDev->chanCount > 1 && !ljkDMA_HasCapMask(&dmaDev->devCapMasks, ljeDMA_TX_TYPE_PRIVATE))
    {
        list_for_each_entry(chan, &dmaDev->chanHead, devNode)
        {
            /* the public channel has already used */
            if (chan->clientCount)
            {
                return ljvNULL;
            }
        }
    }

    list_for_each_entry(chan, &dmaDev->chanHead, devNode)
    {
        if (chan->clientCount)
        {
            continue;
        }

        if (func && !func(chan, funcParam))
        {
            continue;
        }
        return chan;
    }

    return ljvNULL;
}

static ljeSTATUS _DmaFindCandidate(
    ljsDMA_DEVICE *dmaDev,
    const ljuDMA_CAP_MASK *mask,
    ljkDMAENGINE_FilterFunc func,
    void *funcParam,
    ljsDMA_CHANNEL **chan
    )
{
    ljeSTATUS status;

    ljmkHEADER_ARG("dmaDev=%p mask=%p func=%p funcParam=%p chan=%p",
                    dmaDev, mask, func, funcParam, chan);

    *chan = _DmaPrivateCandidate(dmaDev, mask, func, funcParam);

    if (*chan)
    {
        status = _DmaChannelGet(*chan);
        if (ljmIS_ERROR(status))
        {
            ljmkPRINT("dma_engine: fail to get chan %d.\n", (*chan)->id);

            if (status == ljvSTATUS_DEVICE)
            {
                list_del_rcu(&dmaDev->glabolDevNode);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,19,0)
                ida_free(&dmaIda, dmaDev->id);
#else
                ida_remove(&dmaIda, dmaDev->id);
#endif
            }
            *chan = ljvNULL;
        }
    }
    else
    {
        status = ljvSTATUS_PROBE_LATER;
    }
    ljmkFOOTER();
    return status;
}

ljeSTATUS ljkDMA_RequestSpecialSlaveChannel(ljsDMA_CHANNEL *Chan)
{
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkHEADER_ARG("Chan=%p", Chan);

    ljmkMUTEX_LOCK(dmaListMutex);
    if (Chan->clientCount == 0)
    {
        status = _DmaChannelGet(Chan);
        if (ljmIS_ERROR(status))
        {
            ljmkPRINT("dma_engine: fail to get chan %d.\n", Chan->id);
            Chan = ljvNULL;
        }
    }
    else
    {
        ljmkPRINT("dma_engine: request chan %d is busy.\n", Chan->id);
        Chan = ljvNULL;
    }
    ljmkMUTEX_UNLOCK(dmaListMutex);
    ljmkFOOTER();

    return status;
}

ljeSTATUS ljkDMA_RequestAnySlaveChannel(ljsDMA_DEVICE *dmaDev, ljsDMA_CHANNEL **chan)
{
    ljuDMA_CAP_MASK masks;
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkHEADER_ARG("dmaDev=%p chan=%p", dmaDev, chan);

    ljkDMA_ZeroCapMask(&masks);
    ljkDMA_SetCapMasks(&masks, ljeDMA_TX_TYPE_PRIVATE);

    ljmkMUTEX_LOCK(dmaListMutex);
    status = _DmaFindCandidate(dmaDev, &masks, NULL, NULL, chan);
    ljmkMUTEX_UNLOCK(dmaListMutex);

    ljmkFOOTER();
    return status;
}

ljeSTATUS ljkDMA_RequestChannel(
    const ljuDMA_CAP_MASK *mask,
    ljkDMAENGINE_FilterFunc func,
    void *funcParam,
    ljsDMA_CHANNEL **chan
    )
{
    ljsDMA_DEVICE *dmaDev = ljvNULL, *_dma = ljvNULL;
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkHEADER_ARG("dmaDev=%p func=%p funcParam=%p", dmaDev, func, funcParam);

    ljmkMUTEX_LOCK(dmaListMutex);
    list_for_each_entry_safe(dmaDev, _dma,&dmaDeviceList, glabolDevNode)
    {
        status = _DmaFindCandidate(dmaDev, mask, func, funcParam, chan);
        if (ljmIS_ERROR(status))
        {
            break;
        }
    }
    ljmkMUTEX_UNLOCK(dmaListMutex);

    ljmkFOOTER();
    return status;
}

static const ljsDMA_SLAVE_MAP *_DmaFilterMatch(
    ljsDMA_DEVICE *dmaDev,
    const ljtSTRING devName,
    const ljtSTRING slave
    )
{
    ljtUINT i;

    if (dmaDev->filter.mapCount == 0)
    {
        return ljvNULL;
    }

    for (i = 0; i < dmaDev->filter.mapCount; i++)
    {
        const ljsDMA_SLAVE_MAP *map = &dmaDev->filter.maps[i];

        if (!strcmp(map->devName, devName) &&
            !strcmp(map->slave, slave))
        {
            return map;
        }
    }

    return ljvNULL;
}

ljeSTATUS ljkDMA_RequestSlaveChannel(
    const ljtSTRING devName,
    const ljtSTRING Slave,
    ljsDMA_CHANNEL **chan
    )
{
    ljsDMA_DEVICE *dmaDev = ljvNULL, *_dma = ljvNULL;
    ljeSTATUS status = ljvSTATUS_DEVICE;

    ljmkHEADER_ARG("devName=%p slave=%p chan=%p", devName, Slave, chan);

    ljmkMUTEX_LOCK(dmaListMutex);
    list_for_each_entry_safe(dmaDev, _dma, &dmaDeviceList, glabolDevNode)
    {
        ljuDMA_CAP_MASK masks;
        const ljsDMA_SLAVE_MAP *map = _DmaFilterMatch(dmaDev, devName, Slave);

        if (map)
        {
            ljkDMA_ZeroCapMask(&masks);
            ljkDMA_SetCapMasks(&masks, ljeDMA_TX_TYPE_PRIVATE);
            status = _DmaFindCandidate(
                    dmaDev, &masks, dmaDev->filter.func, map->param, chan);
            if (ljmIS_ERROR(status))
            {
                break;
            }
        }
    }
    ljmkMUTEX_UNLOCK(dmaListMutex);

    ljmkFOOTER();
    return status;
}

ljeSTATUS ljkDMA_RequestChannelByMask(const ljuDMA_CAP_MASK *mask, ljsDMA_CHANNEL** chan)
{
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkHEADER_ARG("mask=%p chan=%p", mask, chan);

    if (!mask)
    {
        ljmkFOOTER();
        return ljvSTATUS_INVALID_ARGUMENT;
    }

    status = ljkDMA_RequestChannel(mask, NULL, NULL, chan);

    ljmkFOOTER();
    return status;
}

void ljkDMA_ReleaseChannel(ljsDMA_CHANNEL *Chan)
{
    ljmkMUTEX_LOCK(dmaListMutex);
    ljmkASSERT(Chan->clientCount== 1);
    _DmaChannelPut(Chan);
    ljmkMUTEX_UNLOCK(dmaListMutex);
}

static int _DmaGetId(ljsDMA_DEVICE *dmaDev)
{
    int rc = 0;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,19,0)
    ljmkMUTEX_LOCK(dmaListMutex);
    rc = ida_alloc(&dmaIda, GFP_KERNEL);
    ljmkMUTEX_UNLOCK(dmaListMutex);
    if (rc < 0)
		return rc;
    dmaDev->id = rc;
#else
    do
    {
        if (!ida_pre_get(&dmaIda, GFP_KERNEL))
        {
            return -ENOMEM;
        }
        ljmkMUTEX_LOCK(dmaListMutex);
        rc = ida_get_new(&dmaIda, &dmaDev->id);
        ljmkMUTEX_UNLOCK(dmaListMutex);
    } while (rc == -EAGAIN);
#endif

    return rc;
}

void ljkDMAENGINE_Get(void)
{
    ljsDMA_DEVICE *dmaDev, *_dma;
    ljsDMA_CHANNEL *chan;
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkHEADER();

    ljmkMUTEX_LOCK(dmaListMutex);
    dmaEngineRef++;
    list_for_each_entry_safe(dmaDev, _dma, &dmaDeviceList, glabolDevNode)
    {
        if (ljkDMA_HasCapMask(&dmaDev->devCapMasks, ljeDMA_TX_TYPE_PRIVATE))
        {
            continue;
        }
        list_for_each_entry(chan, &dmaDev->chanHead, devNode)
        {
            status = _DmaChannelGet(chan);
            if (status == ljvSTATUS_DEVICE)
            {
                list_del_rcu(&dmaDev->glabolDevNode);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,19,0)
                ida_free(&dmaIda, dmaDev->id);
#else
                ida_remove(&dmaIda, dmaDev->id);
#endif
                break;
            }
            else if (ljmIS_ERROR(status))
            {
                ljmkPRINT("dma_engine: fail to get chan %d.\n", chan->id);
            }
        }
    }
    ljmkMUTEX_UNLOCK(dmaListMutex);

    ljmkFOOTER();
}

void ljkDMAENGINE_Put(void)
{
    ljsDMA_DEVICE *dmaDev;
    ljsDMA_CHANNEL *chan;

    ljmkMUTEX_LOCK(dmaListMutex);
    dmaEngineRef--;
    ljmkASSERT(dmaEngineRef >= 0);
    list_for_each_entry(dmaDev, &dmaDeviceList, glabolDevNode)
    {
        if (ljkDMA_HasCapMask(&dmaDev->devCapMasks, ljeDMA_TX_TYPE_PRIVATE))
        {
            continue;
        }

        list_for_each_entry(chan, &dmaDev->chanHead, devNode)
        {
            _DmaChannelPut(chan);
        }
    }
    ljmkMUTEX_UNLOCK(dmaListMutex);

}

ljeSTATUS ljkDMA_DeviceRegister(ljsDMA_DEVICE *dmaDev)
{
    ljtINT chanCount = 0, rc;
    ljsDMA_CHANNEL *chan;
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkHEADER_ARG("dmaDev=%p", dmaDev);

    if (!dmaDev)
    {
        ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
    }

    if (!dmaDev->ljkDMADEV_TxStatus)
    {
        ljmkPRINT("dma_engine:  ljkDMADEV_TxStatus function is not defined\n");
        ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
    }

    if (!dmaDev->ljkDMADEV_IssuePending)
    {
        ljmkPRINT("dma_engine: ljkDMADEV_IssueTransaction function is not defined\n");
        ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
    }

    if (ljkDMA_HasCapMask(&dmaDev->devCapMasks, ljeDMA_TX_TYPE_MEMCPY))
    {
        if (dmaDev->ljkDMADEV_PrepareDmaMemcpy)
        {
            ljmkPRINT("dma_engine: ljkDMADEV_PrepareMemCpy function is not defined\n");
            ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
        }
    }

    rc = _DmaGetId(dmaDev);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,19,0)
    if (rc < 0)
#else
    if (rc != 0)
#endif
    {
        ljmkONERROR(ljvSTATUS_OUT_OF_RESOURCES);
    }

    list_for_each_entry(chan, &dmaDev->chanHead, devNode)
    {
        chan->id = chanCount++;
        chan->clientCount = 0;
    }

    if (!chanCount)
    {
        ljmkPRINT("dma_engine: %s has no channels.\n", dmaDev->name);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,19,0)
        ida_free(&dmaIda, dmaDev->id);
#else
        ida_remove(&dmaIda, dmaDev->id);
#endif
        ljmkONERROR(ljvSTATUS_INVALID_DATA);
    }

    dmaDev->chanCount = chanCount;

    ljmkMUTEX_LOCK(dmaListMutex);
    if (dmaEngineRef && !ljkDMA_HasCapMask(&dmaDev->devCapMasks, ljeDMA_TX_TYPE_PRIVATE))
    {
        list_for_each_entry(chan, &dmaDev->chanHead, devNode)
        {
            if (ljmIS_ERROR(_DmaChannelGet(chan)))
            {
                dmaDev->chanCount = 0;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,19,0)
                ida_free(&dmaIda, dmaDev->id);
#else
                ida_remove(&dmaIda, dmaDev->id);
#endif
                ljmkMUTEX_UNLOCK(dmaListMutex);
                ljmkONERROR(ljvSTATUS_OUT_OF_RESOURCES);
            }
        }
    }

    list_add_tail_rcu(&dmaDev->glabolDevNode, &dmaDeviceList);
    ljmkMUTEX_UNLOCK(dmaListMutex);

    ljmkFOOTER();
    return ljvSTATUS_OK;
OnError:
    ljmkFOOTER();
    return status;
}

void ljkDMA_DeviceUnregister(ljsDMA_DEVICE *dmaDev)
{
    ljsDMA_CHANNEL *chan;

    list_for_each_entry(chan, &dmaDev->chanHead, devNode)
    {
        if (chan->clientCount)
        {
            ljmkPRINT("unregister dma%dchann%d while %d clients take a reference\n.",
                dmaDev->id, chan->id, chan->clientCount);
        }
    }

    ljmkMUTEX_LOCK(dmaListMutex);
    list_del_rcu(&dmaDev->glabolDevNode);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,19,0)
    ida_free(&dmaIda, dmaDev->id);
#else
    ida_remove(&dmaIda, dmaDev->id);
#endif
    ljmkMUTEX_UNLOCK(dmaListMutex);
}

void ljkDMA_TxDescriptorInit(ljsDMA_TX_DESCRIPTOR *Desc, ljsDMA_CHANNEL *Chan)
{
    Desc->chan = Chan;
}

ljeDMA_STATUS ljkDMA_WaitForTransaction(ljsDMA_TX_DESCRIPTOR *Desc)
{
    ljtSIZE_T waitTimeout = jiffies + msecs_to_jiffies(10000);
    ljeDMA_STATUS dmaStatus;

    if (!Desc)
    {
        return ljeDMA_STATUS_COMPLETE;
    }

    while(Desc->identifier == -EBUSY)
    {
        if (time_after_eq(jiffies, waitTimeout))
        {
            ljmkPRINT("dma_engine: waiting for descriptor submisson timeout.\n");
            return ljeDMA_STATUS_ERROR;
        }
        cpu_relax();
    }

    dmaStatus = ljkDMA_SyncWait(Desc->chan, Desc->identifier, 10000);

    return dmaStatus;

}

void ljkDMA_IdentifierInit(ljsDMA_CHANNEL *Chan)
{
    Chan->identifier = ljdDMA_MINI_MARK;
    Chan->completedIdentifier = ljdDMA_MINI_MARK;
}

ljdDMA_TX_IDENTIFIER ljkDMA_IdentifierAssign(ljsDMA_TX_DESCRIPTOR *Desc)
{
    ljsDMA_CHANNEL *chan = Desc->chan;
    ljdDMA_TX_IDENTIFIER identifier;

    identifier = chan->identifier + 1;
    if (identifier < ljdDMA_MINI_MARK)
    {
        identifier = ljdDMA_MINI_MARK;
    }

    Desc->identifier = chan->identifier = identifier;

    return identifier;
}

void ljkDMA_IdentifierComplete(ljsDMA_TX_DESCRIPTOR *Desc)
{
    Desc->chan->completedIdentifier = Desc->identifier;
    Desc->identifier = 0;
}

ljeDMA_STATUS ljkDMA_TxIsComplete(ljdDMA_TX_IDENTIFIER Identifier,
    ljdDMA_TX_IDENTIFIER lastCompleted, ljdDMA_TX_IDENTIFIER lastIssued)
{
    if (lastCompleted <= lastIssued)
    {
        if (Identifier <= lastCompleted || Identifier > lastIssued)
        {
            return ljeDMA_STATUS_COMPLETE;
        }
    }
    else
    {
        if (Identifier <= lastCompleted && Identifier > lastIssued)
        {
            return ljeDMA_STATUS_COMPLETE;
        }
    }

    return ljeDMA_STATUS_IN_PROGRESS;
}

ljeDMA_STATUS ljkDMA_IdentifierStatus(ljsDMA_CHANNEL *Chan,
    ljdDMA_TX_IDENTIFIER Identifier, ljsDMA_TX_STATE *txState)
{
    ljdDMA_TX_IDENTIFIER issued, completed;

    issued = Chan->identifier;
    completed = Chan->completedIdentifier;
    if (txState)
    {
        txState->lastIssued = issued;
        txState->lastCompleted = completed;
        txState->residue = 0;
    }

    return ljkDMA_TxIsComplete(Identifier, completed, issued);
}

ljeDMA_STATUS ljkDMA_GetStatus(ljsDMA_CHANNEL *Chan,
    ljdDMA_TX_IDENTIFIER Identifier,
    ljdDMA_TX_IDENTIFIER *Completed,
    ljdDMA_TX_IDENTIFIER *Issued
    )
{
    ljsDMA_TX_STATE state;
    ljeDMA_STATUS status;

    status = ljkDMAENGINE_TxStatus(Chan, Identifier, &state);

    if (Completed)
    {
        *Completed = state.lastCompleted;
    }
    if (Issued)
    {
        *Issued = state.lastIssued;
    }

    return status;
}

void ljkDMA_SetTxState(ljsDMA_TX_STATE *txState,
    ljdDMA_TX_IDENTIFIER lastCompleted, ljdDMA_TX_IDENTIFIER lastIssued, ljtUINT32 Residue)
{
    if (txState)
    {
        txState->lastCompleted = lastCompleted;
        txState->lastIssued = lastIssued;
        txState->residue = Residue;
    }
}

ljtBOOL ljkDMA_SubmitError(ljdDMA_TX_IDENTIFIER Identifier)
{
    return Identifier < 0 ? (ljtBOOL)Identifier : (ljtBOOL)0;
}

void ljkDMA_SetResidue(ljsDMA_TX_STATE *txState, ljtUINT32 Residue)
{
    if (txState)
    {
        txState->residue = Residue;
    }
}

ljtCHAR* ljkDMA_GetDeviceName(ljsDMA_DEVICE *Dev)
{
    if (Dev)
    {
        return Dev->name;
    }

    return ljvNULL;
}

ljtINT ljkDMA_GetDeviceId(ljsDMA_DEVICE *Dev)
{
    if (Dev)
    {
        return Dev->id;
    }

    return -1;
}

ljtINT ljkDMA_GetChannelId(ljsDMA_CHANNEL *Chan)
{
    if (Chan)
    {
        return Chan->id;
    }

    return -1;
}

void ljkDMA_ZeroCapMask(ljuDMA_CAP_MASK *Caps)
{
    Caps->dmaCapMasks = 0;
}

void ljkDMA_SetCapMasks(ljuDMA_CAP_MASK *Caps, ljeDMA_TX_TYPE Type)
{
    Caps->dmaCapMasks |= (0x1 << Type);
}

void ljkDMA_ClearCapMasks(ljuDMA_CAP_MASK *Caps, ljeDMA_TX_TYPE Type)
{
    Caps->dmaCapMasks &= ~(0x1 << Type);
}

ljtUINT32 ljkDMA_HasCapMask(ljuDMA_CAP_MASK *Caps, ljeDMA_TX_TYPE Type)
{
    return Caps->dmaCapMasks & (0x1 << Type);
}

ljeSTATUS ljkDMAENGINE_SlaveConfig(
    ljsDMA_CHANNEL *Chan,
    ljsDMA_SLAVE_CONFIG *Config
    )
{
    if (Chan->device->ljkDMADEV_Config)
    {
        return Chan->device->ljkDMADEV_Config(Chan, Config);
    }

    return ljvSTATUS_NOT_SUPPORTED;
}

ljtBOOL ljkDMAENGINE_IsSlaveDirection(
    ljeDMA_TX_DIRECTION direction
    )
{
    if (direction == ljeDMA_TX_DIRECTION_MEM_TO_DEV
        || direction == ljeDMA_TX_DIRECTION_DEV_TO_MEM)
    {
        return ljvTRUE;
    }
    else
    {
        return ljvFALSE;
    }
}

ljsDMA_TX_DESCRIPTOR *ljkDMAENGINE_PrepareSlaveSingle(
    ljsDMA_CHANNEL *Chan,
    dma_addr_t Addr,
    ljtSIZE_T Len,
    ljeDMA_TX_DIRECTION Dir,
    ljtUINT64 Flags
    )
{
    struct scatterlist sg;

    sg_init_table(&sg, 1);
    sg_dma_address(&sg) = Addr;
    sg_dma_len(&sg) = Len;

    if (!Chan || !Chan->device || !Chan->device->ljkDMADEV_PrepareSlaveSG)
    {
        return ljvNULL;
    }

    return Chan->device->ljkDMADEV_PrepareSlaveSG(Chan, &sg, 1, Dir, Flags, NULL);
}

ljsDMA_TX_DESCRIPTOR *ljkDMAENGINE_PrepareSlaveSG(
    ljsDMA_CHANNEL *Chan,
    struct scatterlist *SGList,
    ljtSIZE_T SGLen,
    ljeDMA_TX_DIRECTION Dir,
    ljtUINT64 Flags
    )
{
    if (!SGList || !SGLen)
    {
        return ljvNULL;
    }

    if (!Chan || !Chan->device || !Chan->device->ljkDMADEV_PrepareSlaveSG)
    {
        return ljvNULL;
    }

    return Chan->device->ljkDMADEV_PrepareSlaveSG(Chan, SGList, SGLen, Dir, Flags, NULL);
}

ljsDMA_TX_DESCRIPTOR *ljkDMAENGINE_PrepareMemcpy(
    ljsDMA_CHANNEL *Chan,
    dma_addr_t Src,
    dma_addr_t Dst,
    ljtSIZE_T Len,
    ljtUINT64 Flag
    )
{
    if (!Chan || !Chan->device || !Chan->device->ljkDMADEV_PrepareDmaMemcpy)
    {
        return ljvNULL;
    }

    return Chan->device->ljkDMADEV_PrepareDmaMemcpy(Chan, Src, Dst, Len, Flag);
}

ljeSTATUS ljkDMAENGINE_TerminateAll(ljsDMA_CHANNEL *Chan)
{
    if (Chan->device->ljkDMADEV_TerminateAll)
    {
        return Chan->device->ljkDMADEV_TerminateAll(Chan);
    }

    return ljvSTATUS_NOT_SUPPORTED;
}

ljeSTATUS ljkDMAENGINE_Synchronize(ljsDMA_CHANNEL *Chan)
{
    if (Chan->device->ljkDMADEV_Synchronize)
    {
        Chan->device->ljkDMADEV_Synchronize(Chan);
    }

    return ljvSTATUS_NOT_SUPPORTED;
}

ljeSTATUS ljkDMAENGINE_TerminateSync(ljsDMA_CHANNEL *Chan)
{
    ljeSTATUS status;

    status = ljkDMAENGINE_TerminateAll(Chan);
    if (status == ljvSTATUS_OK)
    {
        status = ljkDMAENGINE_Synchronize(Chan);
    }

    return status;
}

ljeSTATUS ljkDMAENGINE_Pause(ljsDMA_CHANNEL *Chan)
{
    if (Chan->device->ljkDMADEV_Pause)
    {
        Chan->device->ljkDMADEV_Pause(Chan);
    }

    return ljvSTATUS_NOT_SUPPORTED;
}

ljeSTATUS ljkDMAENGINE_Resume(ljsDMA_CHANNEL *Chan)
{
    if (Chan->device->ljkDAMDEV_Resume)
    {
        Chan->device->ljkDAMDEV_Resume(Chan);
    }

    return ljvSTATUS_NOT_SUPPORTED;
}

ljeDMA_STATUS ljkDMAENGINE_TxStatus(
    ljsDMA_CHANNEL *Chan,
    ljdDMA_TX_IDENTIFIER Identifier,
    ljsDMA_TX_STATE *TxState
    )
{
    return Chan->device->ljkDMADEV_TxStatus(Chan, Identifier, TxState);
}

void ljkDMAENGINE_IssuePending(
    ljsDMA_CHANNEL *Chan
    )
{
    Chan->device->ljkDMADEV_IssuePending(Chan);
}

ljdDMA_TX_IDENTIFIER ljkDMAENGINE_Submit(ljsDMA_TX_DESCRIPTOR *Desc)
{
    return Desc->ljkDMA_TxSubmit(Desc);
}

void ljkDMAINGINE_DescGetCallback(ljsDMA_TX_DESCRIPTOR *Desc,
   ljsDMAENGINE_DESC_CALLBACK *Cb
    )
{
    Cb->ljkDMA_TxCallback = Desc->ljkDMA_TxCallback;
    Cb->ljkDMA_TxCallbackResult = Desc->ljkDMA_TxCallbackResult;
    Cb->txCallbackParam = Desc->txCallbackParam;
}

void ljkDMAENGINE_DescCallbackInvoke(
    ljsDMAENGINE_DESC_CALLBACK *Cb,
    ljsDMA_RESULT *Result
    )
{
    ljsDMA_RESULT dummy = {ljeDMA_TX_RESULT_OK, 0};

    if (Cb->ljkDMA_TxCallbackResult)
    {
        if (!Result)
        {
            Result = &dummy;
        }
        Cb->ljkDMA_TxCallbackResult(Cb->txCallbackParam, Result);
    }
    else if (Cb->ljkDMA_TxCallback)
    {
        Cb->ljkDMA_TxCallback(Cb->txCallbackParam);
    }
}

void ljkDMAENGINE_DescGetCallbackAndInvoke(
    ljsDMA_TX_DESCRIPTOR *Desc,
    ljsDMA_RESULT *Result
    )
{
    ljsDMAENGINE_DESC_CALLBACK cb;

    ljkDMAINGINE_DescGetCallback(Desc, &cb);
    ljkDMAENGINE_DescCallbackInvoke(&cb, Result);
}

ljtBOOL ljkDMAENGINE_DescCallbackValidate(
    ljsDMAENGINE_DESC_CALLBACK *Cb
    )
{
    return (Cb->ljkDMA_TxCallback)? ljvTRUE : ljvFALSE;
}

