/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#include "lj_hal_kernel_linux.h"

#define _LJ_OBJ_ZONE ljvZONE_KERNEL

/******************************************************************************
 ******************************* ljkKERNEL API Code ***************************
 *****************************************************************************/

/******************************************************************************
 **
 **  ljkKERNEL_QueryVideoMemory
 **
 **  Query the amount of video memory.
 **
 **  INPUT:
 **
 **      ljkKERNEL Kernel
 **          Pointer to an ljkKERNEL object.
 **
 **  OUTPUT:
 **
 **      ljsHAL_INTERFACE *Interface
 **          Pointer to an ljsHAL_INTERFACE structure that will be filled in
 **          with the memory information.
 */
ljeSTATUS
ljkKERNEL_QueryVideoMemory(ljkKERNEL Kernel, ljsHAL_INTERFACE *Interface)
{
    ljkGALDEVICE galDevice = ljvNULL;
    ljkDEVICE device = ljvNULL;

    ljmkHEADER_ARG("Kernel=%p", Kernel);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Kernel, ljvOBJ_KERNEL);
    ljmkVERIFY_ARGUMENT(Interface);

    /* Extract the pointer to the ljkGALDEVICE class. */
    galDevice = (ljkGALDEVICE)Kernel->context;
    device = Kernel->device;

    /* Get internal memory size and physical address. */
    Interface->u.QueryVideoMemory.internalSize = galDevice->internalSize;
    Interface->u.QueryVideoMemory.internalPhysName = galDevice->internalPhysName;

    /* Get external memory size and physical address. */
    Interface->u.QueryVideoMemory.externalSize = device->externalSize;
    Interface->u.QueryVideoMemory.externalPhysName = device->externalPhysName;

    /* Get contiguous memory size and physical address. */
    Interface->u.QueryVideoMemory.contiguousSize = 0;   //device->contiguousSizes[device->memIndex];
    Interface->u.QueryVideoMemory.contiguousPhysName = 0;   //device->contiguousPhysNames[device->memIndex];

    /* Get exclusive memory size and physical address. */
    Interface->u.QueryVideoMemory.exclusiveSize = device->exclusiveSize;
    Interface->u.QueryVideoMemory.exclusivePhysName = device->exclusivePhysName;

    /* If the virtual pool can be an available video memory pool. */
#if ljdENABLE_VIRTUAL_POOL
    Interface->u.QueryVideoMemory.virtualPoolEnabled = ljvTRUE;
#else
    Interface->u.QueryVideoMemory.virtualPoolEnabled = ljvFALSE;
#endif

    /* Success. */
    ljmkFOOTER_NO();
    return ljvSTATUS_OK;
}

/******************************************************************************
 **
 **  ljkKERNEL_GetVideoMemoryPool
 **
 **  Get the ljkVIDMEM object belonging to the specified pool.
 **
 **  INPUT:
 **
 **      ljkKERNEL Kernel
 **          Pointer to an ljkKERNEL object.
 **
 **      ljePOOL Pool
 **          Pool to query ljkVIDMEM object for.
 **
 **  OUTPUT:
 **
 **      ljkVIDMEM *VideoMemory
 **          Pointer to a variable that will hold the pointer to the ljkVIDMEM
 **          object belonging to the requested pool.
 */
ljeSTATUS
ljkKERNEL_GetVideoMemoryPool(ljkKERNEL Kernel, ljePOOL Pool, ljkVIDMEM *VideoMemory)
{
    ljkGALDEVICE galDevice = ljvNULL;
    ljkDEVICE device = ljvNULL;
    ljkVIDMEM videoMemory;

    ljmkHEADER_ARG("Kernel=%p Pool=%d", Kernel, Pool);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Kernel, ljvOBJ_KERNEL);
    ljmkVERIFY_ARGUMENT(VideoMemory);

    /* Extract the pointer to the ljkGALDEVICE class. */
    galDevice = (ljkGALDEVICE)Kernel->context;
    device = Kernel->device;

    /* Dispatch on pool. */
    switch (Pool) {
    case ljvPOOL_LOCAL_INTERNAL:
        /* Internal memory. */
        videoMemory = galDevice->internalVidMem;
        break;

    case ljvPOOL_LOCAL_EXTERNAL:
        /* External memory. */
        videoMemory = device->externalVidMem;
        break;

    case ljvPOOL_SYSTEM:
        /* System memory. */
        videoMemory = device->contiguousVidMems[device->memIndex];
        break;

    case ljvPOOL_LOCAL_EXCLUSIVE:
        /* gpu exclusive memory. */
        videoMemory = device->exclusiveVidMem;
        break;

    case ljvPOOL_INTERNAL_SRAM:
        /* Internal SRAM memory. */
        videoMemory = Kernel->sRAMVidMem[Kernel->sRAMIndex];
        break;

    case ljvPOOL_EXTERNAL_SRAM:
        /* External SRAM memory. */
        videoMemory = device->extSRAMVidMems[device->extSRAMIndex];
        break;


#if ljdENABLE_40BIT_VA
    case ljvPOOL_SYSTEM_32BIT_VA:
        /* System memory. */
        videoMemory = galDevice->lowContiguousVidMem;
        break;
#endif

    default:
        /* Unknown pool. */
        videoMemory = NULL;
    }

    /* Return pointer to the ljkVIDMEM object. */
    *VideoMemory = videoMemory;

    /* Return status. */
    ljmkFOOTER_ARG("*VideoMemory=%p", *VideoMemory);
    return (!videoMemory) ? ljvSTATUS_OUT_OF_MEMORY : ljvSTATUS_OK;
}

/******************************************************************************
 **
 **  ljkKERNEL_MapMemory
 **
 **  Map video memory into the current process space.
 **
 **  INPUT:
 **
 **      ljkKERNEL Kernel
 **          Pointer to an ljkKERNEL object.
 **
 **      ljtPHYS_ADDR Physical
 **          Physical address of video memory to map.
 **
 **      ljtSIZE_T Bytes
 **          Number of bytes to map.
 **
 **  OUTPUT:
 **
 **      ljtPOINTER *Logical
 **          Pointer to a variable that will hold the base address of the
 **          mapped memory region.
 */
ljeSTATUS
ljkKERNEL_MapMemory(ljkKERNEL Kernel, ljtPHYS_ADDR Physical,
                    ljtSIZE_T Bytes, ljtPOINTER *Logical)
{
    ljkKERNEL kernel = Kernel;
    ljtPHYS_ADDR physical = ljmNAME_TO_PTR(Physical);

    return ljkOS_MapMemory(Kernel->os, physical, Bytes, Logical);
}

/******************************************************************************
 **
 **  ljkKERNEL_UnmapMemory
 **
 **  Unmap video memory from the current process space.
 **
 **  INPUT:
 **
 **      ljkKERNEL Kernel
 **          Pointer to an ljkKERNEL object.
 **
 **      ljtPHYS_ADDR Physical
 **          Physical address of video memory to map.
 **
 **      ljtSIZE_T Bytes
 **          Number of bytes to map.
 **
 **      ljtPOINTER Logical
 **          Base address of the mapped memory region.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkKERNEL_UnmapMemory(ljkKERNEL Kernel, ljtPHYS_ADDR Physical,
                      ljtSIZE_T Bytes, ljtPOINTER Logical, ljtUINT32 ProcessID)
{
    ljeSTATUS status = ljvSTATUS_OK;
    ljkKERNEL kernel = Kernel;
    ljtPHYS_ADDR physical = ljmNAME_TO_PTR(Physical);

    ljmkONERROR(ljkOS_UnmapMemoryEx(Kernel->os, physical, Bytes, Logical, ProcessID));

OnError:
    /* Retunn the status. */
    return status;
}

/******************************************************************************
 **
 **  ljkKERNEL_DestroyProcessReservedUserMap
 **
 **  Destroy process reserved memory
 **
 **  INPUT:
 **
 **      ljtPHYS_ADDR Physical
 **          Physical address of video memory to map.
 **
 **      ljtUINT32 Pid
 **          Process ID.
 */
ljeSTATUS
ljkKERNEL_DestroyProcessReservedUserMap(ljkKERNEL Kernel, ljtUINT32 Pid)
{
    ljeSTATUS status = ljvSTATUS_OK;
    ljkGALDEVICE galDevice = ljvNULL;
    ljkDEVICE device = ljvNULL;
    ljtSIZE_T bytes = 0;
    ljtPHYS_ADDR physHandle = ljvNULL;
    /* when unmap reserved memory, we don't need real logical*/
    ljtPOINTER Logical = (ljtPOINTER)0xFFFFFFFF;
    ljtINT i;
    PLINUX_MDL mdl;
    PLINUX_MDL_MAP mdlMap = ljvNULL;
    ljtBOOL acquiredMutex = ljvFALSE;

    ljmkHEADER_ARG("Logical=%p pid=%u", Logical, Pid);
    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Kernel, ljvOBJ_KERNEL);

    /* Extract the pointer to the ljkGALDEVICE class. */
    galDevice = (ljkGALDEVICE)Kernel->context;
    device = Kernel->device;

    physHandle = (PLINUX_MDL)galDevice->internalPhysical;
    bytes = galDevice->internalSize;

    if (bytes && physHandle) {
        mdl = physHandle;
        ljmkONERROR(ljkOS_AcquireMutex(Kernel->os, &mdl->mapsMutex, ljvINFINITE));
        acquiredMutex = ljvTRUE;
        mdlMap = FindMdlMap(mdl, Pid);
        ljmkONERROR(ljkOS_ReleaseMutex(Kernel->os, &mdl->mapsMutex));
        acquiredMutex = ljvFALSE;
        if (mdlMap)
            ljmkONERROR(ljkOS_UnmapMemoryEx(Kernel->os, physHandle, bytes, Logical, Pid));
    }

    for (i = 0; i < ljdLOCAL_MEMORY_COUNT; i++) {
        physHandle = (PLINUX_MDL)galDevice->externalPhysical[i];
        bytes = galDevice->externalSize[i];

        if (bytes && physHandle) {
            mdl = physHandle;
            ljmkONERROR(ljkOS_AcquireMutex(Kernel->os, &mdl->mapsMutex, ljvINFINITE));
            acquiredMutex = ljvTRUE;
            mdlMap = FindMdlMap(mdl, Pid);
            ljmkONERROR(ljkOS_ReleaseMutex(Kernel->os, &mdl->mapsMutex));
            acquiredMutex = ljvFALSE;
            if (mdlMap)
                ljmkONERROR(ljkOS_UnmapMemoryEx(Kernel->os, physHandle, bytes, Logical, Pid));
        }
    }

#if !ljdCAPTURE_ONLY_MODE
    /* System memory. */
    for (i = 0; i < ljdSYSTEM_RESERVE_COUNT; i++) {
        physHandle = (PLINUX_MDL)device->contiguousPhysicals[i];
        bytes = device->contiguousSizes[i];

        if (bytes && physHandle) {
            mdl = physHandle;
            ljmkONERROR(ljkOS_AcquireMutex(Kernel->os, &mdl->mapsMutex, ljvINFINITE));
            acquiredMutex = ljvTRUE;
            mdlMap = FindMdlMap(mdl, Pid);
            ljmkONERROR(ljkOS_ReleaseMutex(Kernel->os, &mdl->mapsMutex));
            acquiredMutex = ljvFALSE;
            if (mdlMap)
                ljmkONERROR(ljkOS_UnmapMemoryEx(Kernel->os, physHandle, bytes, Logical, Pid));
        }
    }
#endif

#if ljdENABLE_40BIT_VA
    physHandle = (PLINUX_MDL)galDevice->lowContiguousPhysical;
    bytes = galDevice->lowContiguousSize;

    if (bytes && physHandle) {
        mdl = physHandle;
        ljmkONERROR(ljkOS_AcquireMutex(Kernel->os, &mdl->mapsMutex, ljvINFINITE));
        acquiredMutex = ljvTRUE;
        mdlMap = FindMdlMap(mdl, Pid);
        ljmkONERROR(ljkOS_ReleaseMutex(Kernel->os, &mdl->mapsMutex));
        acquiredMutex = ljvFALSE;
        if (mdlMap)
            ljmkONERROR(ljkOS_UnmapMemoryEx(Kernel->os, physHandle, bytes, Logical, Pid));
    }
#endif

    /* External shared SRAM memory. */
    for (i = 0; i < ljvSRAM_EXT_COUNT; i++) {
        physHandle = (PLINUX_MDL)device->extSRAMPhysicals[i];
        bytes = device->extSRAMSizes[i];
        if (bytes && physHandle) {
            mdl = physHandle;
            ljmkONERROR(ljkOS_AcquireMutex(Kernel->os, &mdl->mapsMutex, ljvINFINITE));
            acquiredMutex = ljvTRUE;
            mdlMap = FindMdlMap(mdl, Pid);
            ljmkONERROR(ljkOS_ReleaseMutex(Kernel->os, &mdl->mapsMutex));
            acquiredMutex = ljvFALSE;
            if (mdlMap)
                ljmkONERROR(ljkOS_UnmapMemoryEx(Kernel->os, physHandle, bytes, Logical, Pid));
        }
    }

    /* Per core SRAM reserved usage. */
    for (i = 0; i < ljvSRAM_INTER_COUNT; i++) {
        if (!Kernel->sRAMPhysFaked[i]) {
            physHandle = (PLINUX_MDL)Kernel->sRAMPhysical[i];
            bytes = Kernel->sRAMSizes[i];
            if (bytes && physHandle) {
                mdl = physHandle;
                ljmkONERROR(ljkOS_AcquireMutex(Kernel->os, &mdl->mapsMutex, ljvINFINITE));
                acquiredMutex = ljvTRUE;
                mdlMap = FindMdlMap(mdl, Pid);
                ljmkONERROR(ljkOS_ReleaseMutex(Kernel->os, &mdl->mapsMutex));
                acquiredMutex = ljvFALSE;
                if (mdlMap)
                    ljmkONERROR(ljkOS_UnmapMemoryEx(Kernel->os, physHandle, bytes, Logical, Pid));
            }
        }
    }

OnError:
    /* Retunn the status. */
    if (acquiredMutex)
        ljkOS_ReleaseMutex(Kernel->os, &mdl->mapsMutex);

    ljmkFOOTER_NO();
    return status;
}

/******************************************************************************
 **
 **  ljkKERNEL_MapVideoMemory
 **
 **  Get the logical address for a hardware specific memory address for the
 **  current process.
 **
 **  INPUT:
 **
 **      ljkKERNEL Kernel
 **          Pointer to an ljkKERNEL object.
 **
 **      ljtBOOL InUserSpace
 **          ljvTRUE to map the memory into the user space.
 **
 **      ljePOOL Pool
 **          Specify pool type.
 **
 **      ljtUINT32 Offset
 **          Offset to pool start.
 **
 **      ljtUINT32 Bytes
 **          Number of bytes to map.
 **
 **  OUTPUT:
 **
 **      ljtPOINTER *Logical
 **          Pointer to a variable that will hold the logical address of the
 **          specified memory address.
 */
ljeSTATUS
ljkKERNEL_MapVideoMemory(ljkKERNEL Kernel, ljtBOOL InUserSpace, ljePOOL Pool,
                         ljtPHYS_ADDR Physical, ljtSIZE_T Offset, ljtSIZE_T Bytes, ljtPOINTER *Logical)
{
    ljkGALDEVICE galDevice = ljvNULL;
    ljkDEVICE device = ljvNULL;
    ljtSIZE_T bytes = 0;
    ljtPHYS_ADDR physHandle = ljvNULL;
    ljeSTATUS status = ljvSTATUS_OK;
    ljtPOINTER logical = ljvNULL;
    ljtUINT64 mappingInOne = 1;

    ljmkHEADER_ARG("Kernel=%p InUserSpace=%d Pool=%d Offset=%X Bytes=%X",
                   Kernel, InUserSpace, Pool, Offset, Bytes);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Kernel, ljvOBJ_KERNEL);
    ljmkVERIFY_ARGUMENT(Logical);

    if (Physical)
        ljmkONERROR(ljkOS_QueryOption(Kernel->os, "allMapInOne", &mappingInOne));

    if (mappingInOne) {
        /* Extract the pointer to the ljkGALDEVICE class. */
        galDevice = (ljkGALDEVICE)Kernel->context;
        device = Kernel->device;

        /* Dispatch on pool. */
        switch (Pool) {
        case ljvPOOL_LOCAL_INTERNAL:
            physHandle = (PLINUX_MDL)galDevice->internalPhysical;
            bytes = galDevice->internalSize;
            break;

        case ljvPOOL_LOCAL_EXTERNAL:
            physHandle = (PLINUX_MDL)device->externalPhysical;
            bytes = device->externalSize;
            break;

        case ljvPOOL_LOCAL_EXCLUSIVE:
            physHandle = (PLINUX_MDL)device->exclusivePhysical;
            bytes = device->exclusiveSize;
            break;

        case ljvPOOL_SYSTEM:
            /* System memory. */
            physHandle = (PLINUX_MDL)device->contiguousPhysicals[device->memIndex];
            bytes = device->contiguousSizes[device->memIndex];
            break;

#if ljdENABLE_40BIT_VA
        case ljvPOOL_SYSTEM_32BIT_VA:
            /* System memory. */
            physHandle = (PLINUX_MDL)galDevice->lowContiguousPhysical;
            bytes = galDevice->lowContiguousSize;
            break;
#endif

        case ljvPOOL_EXTERNAL_SRAM:
            /* External shared SRAM memory. */
            physHandle = (PLINUX_MDL)device->extSRAMPhysicals[device->extSRAMIndex];
            bytes = device->extSRAMSizes[device->extSRAMIndex];
            break;

        case ljvPOOL_INTERNAL_SRAM:
            /* Per core SRAM reserved usage. */
            if (Kernel->sRAMPhysFaked[Kernel->sRAMIndex]) {
                *Logical = ljvNULL;

                ljmkFOOTER_NO();
                return ljvSTATUS_OK;
            }
            /* Per core SRAM memory block. */
            physHandle = (PLINUX_MDL)Kernel->sRAMPhysical[Kernel->sRAMIndex];
            bytes = Kernel->sRAMSizes[Kernel->sRAMIndex];
            break;

        default:
            /* Invalid memory pool. */
            ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
        }

    } else {
        physHandle = (PLINUX_MDL)Physical;
        bytes = Bytes;
        Offset = 0;
    }

#if ljdENABLE_TTM
    if (*Logical) {
        logical = *Logical;
        ljmkONERROR(ljkOS_LockPages(Kernel->os, physHandle, bytes, ljvFALSE, &logical));
        goto OnError;
    }
#endif

    ljmkONERROR(ljkOS_LockPages(Kernel->os, physHandle, bytes, ljvFALSE, &logical));
    /* Build logical address of specified address. */
    *Logical = (ljtPOINTER)((ljtUINT8_PTR)logical + Offset);
OnError:
    /* Retunn the status. */
    ljmkFOOTER_ARG("*Logical=%p", ljmOPT_POINTER(Logical));
    return status;
}

/******************************************************************************
 **
 **  ljkKERNEL_UnmapVideoMemory
 **
 **  Unmap video memory for the current process.
 **
 **  INPUT:
 **
 **      ljkKERNEL Kernel
 **          Pointer to an ljkKERNEL object.
 **
 **      ljePOOL Pool
 **          Specify pool type.

 **      ljtPHYS_ADDR Physical
 **          Node ptr.
 **
 **      ljtUINT32 Pid
 **          Process ID of the current process.
 **
 **      ljtSIZE_T Bytes
 **          Number of bytes to map.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkKERNEL_UnmapVideoMemory(ljkKERNEL Kernel, ljePOOL Pool, ljtPHYS_ADDR Physical,
                           ljtPOINTER Logical, ljtUINT32 Pid, ljtSIZE_T Bytes)
{
    ljeSTATUS status = ljvSTATUS_OK;
    ljkGALDEVICE galDevice = ljvNULL;
    ljkDEVICE device = ljvNULL;
    ljtSIZE_T bytes = 0;
    ljtPHYS_ADDR physHandle = ljvNULL;
    ljtUINT64 mappingInOne = 1;

    ljmkHEADER_ARG("Logical=%p Bytes=%zu", Logical, Bytes);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Kernel, ljvOBJ_KERNEL);

    if (Logical == ljvNULL)
        return ljvSTATUS_OK;

    if (Physical)
        ljmkONERROR(ljkOS_QueryOption(Kernel->os, "allMapInOne", &mappingInOne));

    if (mappingInOne) {
        /* Extract the pointer to the ljkGALDEVICE class. */
        galDevice = (ljkGALDEVICE)Kernel->context;
        device = Kernel->device;

        /* Dispatch on pool. */
        switch (Pool) {
        case ljvPOOL_LOCAL_INTERNAL:
            physHandle = (PLINUX_MDL)galDevice->internalPhysical;
            bytes = galDevice->internalSize;
            break;

        case ljvPOOL_LOCAL_EXTERNAL:
            physHandle = (PLINUX_MDL)device->externalPhysical;
            bytes = device->externalSize;
            break;

        case ljvPOOL_LOCAL_EXCLUSIVE:
            physHandle = (PLINUX_MDL)device->exclusivePhysical;
            bytes = device->exclusiveSize;
            break;

        case ljvPOOL_SYSTEM:
            /* System memory. */
            physHandle = (PLINUX_MDL)device->contiguousPhysicals[device->memIndex];
            bytes = device->contiguousSizes[device->memIndex];
            break;

#if ljdENABLE_40BIT_VA
        case ljvPOOL_SYSTEM_32BIT_VA:
            /* System memory. */
            physHandle = (PLINUX_MDL)galDevice->lowContiguousPhysical;
            bytes = galDevice->lowContiguousSize;
            break;
#endif

        case ljvPOOL_EXTERNAL_SRAM:
            /* External shared SRAM memory. */
            physHandle = (PLINUX_MDL)device->extSRAMPhysicals[device->extSRAMIndex];
            bytes = device->extSRAMSizes[device->extSRAMIndex];
            break;

        case ljvPOOL_INTERNAL_SRAM:
            /* Per core SRAM reserved usage. */
            if (Kernel->sRAMPhysFaked[Kernel->sRAMIndex]) {
                ljmkFOOTER_NO();
                return ljvSTATUS_OK;
            }
            /* Per core SRAM memory block. */
            physHandle = (PLINUX_MDL)Kernel->sRAMPhysical[Kernel->sRAMIndex];
            bytes = Kernel->sRAMSizes[Kernel->sRAMIndex];
            break;

        default:
            /* Invalid memory pool. */
            WARN_ON(1);
            ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
        }
    } else {
        physHandle = (PLINUX_MDL)Physical;
        bytes = Bytes;
    }

    ljmkONERROR(ljkOS_UnlockPages(Kernel->os, physHandle, bytes, Logical));

OnError:
    /* Retunn the status. */
    ljmkFOOTER_NO();
    return status;
}

/******************************************************************************
 **
 **  ljkKERNEL_Notify
 **
 **  This function iscalled by clients to notify the ljkKERNRL object of an
 **  event.
 **
 **  INPUT:
 **
 **      ljkKERNEL Kernel
 **          Pointer to an ljkKERNEL object.
 **
 **      ljeNOTIFY Notification
 **          Notification event.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkKERNEL_Notify(ljkKERNEL Kernel, ljeNOTIFY Notification)
{
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkHEADER_ARG("Kernel=%p Notification=%d", Kernel, Notification);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Kernel, ljvOBJ_KERNEL);

    /* Dispatch on notifcation. */
    switch (Notification) {
    case ljvNOTIFY_INTERRUPT:
        /* Process the interrupt. */
#if COMMAND_PROCESSOR_VERSION > 1
        status = ljkINTERRUPT_Notify(Kernel->interrupt, 0);
#else
        status = ljkHARDWARE_Notify(Kernel->hardware);
#endif
        break;

    default:
        break;
    }

    /* Success. */
    ljmkFOOTER();
    return status;
}

#if ljdENABLE_VIDEO_MEMORY_MIRROR
ljeSTATUS
ljkKERNEL_SyncVideoMemoryMirror(ljkKERNEL Kernel,
                                ljkVIDMEM_NODE Node,
                                ljtSIZE_T Offset,
                                ljtSIZE_T Bytes,
                                ljtUINT32 Reason)
{
    ljeSTATUS status = ljvSTATUS_NOT_SUPPORTED;
    ljsPLATFORM *platform;
    ljtSIZE_T size;

    ljmkHEADER();

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Kernel, ljvOBJ_KERNEL);

    if (!Node->mirror.mirrorNode) {
        ljmkFOOTER();
        return ljvSTATUS_INVALID_ARGUMENT;
    }

    ljkVIDMEM_NODE_GetSize(Kernel, Node, &size);
    if (Offset + Bytes > size) {
        ljmkFOOTER();
        return ljvSTATUS_INVALID_ARGUMENT;
    }

    platform = Kernel->os->device->platform;

    if (platform && platform->ops->dmaCopy) {
        ljkVIDMEM_NODE srcNode = ljvNULL, dstNode = ljvNULL;
        ljsDMA_TRANS_INFO info;

        /* Mirror node is local memory */
        if (Node->mirror.type == ljvMIRROR_TYPE_LOCAL_MEMORY_MIRROR) {
            if (Reason == ljvSYNC_MEMORY_DIRECTION_LOCAL_TO_SYSTEM) {
                /* local -> system */
                srcNode = Node->mirror.mirrorNode;
                dstNode = Node;
            } else if (Reason == ljvSYNC_MEMORY_DIRECTION_SYSTEM_TO_LOCAL) {
                /* system -> local */
                srcNode = Node;
                dstNode = Node->mirror.mirrorNode;
            }
            /* Mirror node is system memory */
        } else if (Node->mirror.type == ljvMIRROR_TYPE_SYSTEM_MEMORY_MIRROR) {
            if (Reason == ljvSYNC_MEMORY_DIRECTION_LOCAL_TO_SYSTEM) {
                /* local -> system */
                srcNode = Node;
                dstNode = Node->mirror.mirrorNode;
            } else if (Reason == ljvSYNC_MEMORY_DIRECTION_SYSTEM_TO_LOCAL) {
                /* system -> local */
                srcNode = Node->mirror.mirrorNode;
                dstNode = Node;
            }
        } else {
            ljmkFOOTER();
            return ljvSTATUS_INVALID_ARGUMENT;
        }

        info.src_node = srcNode;
        info.dst_node = dstNode;
        info.offset = Offset;
        info.bytes = Bytes;
        info.reason = Reason;
//        printk("%s(%d): Bytes = 0x%lx Offset = 0x%lx; \n", __FUNCTION__, __LINE__, Bytes, Offset);
#if 0
        status = platform->ops->syncMemory(Kernel, dstNode, srcNode, Reason, Bytes, Offset);
#else
        status = platform->ops->dmaCopy(Kernel, &info);
#endif
    }

    ljmkFOOTER();
    return status;
}
#endif


ljeSTATUS
ljkKERNEL_DmaOperation(
    IN ljkKERNEL Kernel,
    IN ljkVIDMEM_NODE Node_Src,
    IN ljkVIDMEM_NODE Node_Dst,
    IN ljtUINT32 Dir,
    IN ljtUINT64 TransLen
    )
{
    ljeSTATUS status = ljvSTATUS_NOT_SUPPORTED;
    ljsPLATFORM * platform;

    ljmkHEADER();

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Kernel, ljvOBJ_KERNEL);

    platform = Kernel->os->device->platform;

    if (platform && platform->ops->dmaOperation)
    {
        status = platform->ops->dmaOperation(Kernel, (ljtPOINTER)Node_Src, (ljtPOINTER)Node_Dst, Dir, TransLen);
    }

    ljmkFOOTER();
    return status;
}

