/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#include <linux/dmapool.h>
#include <linux/err.h>
#include <linux/init.h>
#include <linux/interrupt.h>
#include <linux/slab.h>
#include <linux/pci.h>
#include <linux/list.h>
#include <linux/delay.h>
#include <linux/time.h>
#include <linux/semaphore.h>

#include "lj_hal_kernel_dma_engine.h"
#include "lj_hal_kernel_platform_hdma.h"
#include "ljmicro_family.h"

static void edma_complete_all(edma_channel *edma_chan);
static void edma_do_first_queued(edma_channel *edma_chan);
static void edma_dev_handle_done(edma_channel *edma_chan);
static void edma_dev_handle_error(edma_channel *edma_chan);

struct semaphore    sema_edma;
ljtBOOL             EdmaKillThread = ljvFALSE;

static int delay_time = 1;

#define EDMA_BUS_WIDTHS     (1 << ljeDMA_SLAVE_BUS_WIDTH_4BYTES)| \
                            (1 << ljeDMA_SLAVE_BUS_WIDTH_8BYTES)

edma_channel *to_edma_chan(ljsDMA_CHANNEL *chan)
{
    return container_of(chan, edma_channel, chan);
}

/**
 * to_edma_dev - convert dma desc to edma dev
 *
 * @ddma: dma_device structure pointer
 *
 * Return: pcie_edma pointer
 *
 */
edma_device *to_edma_dev(ljsDMA_DEVICE *ddma)
{
    edma_device *edma_dev = ljvNULL;

    if (ddma->caps.directions & (1 << ljeDMA_TX_DIRECTION_MEM_TO_DEV))
    {
        edma_dev = container_of(ddma, edma_device, slave_rdma);
    }
    else if (ddma->caps.directions & (1 << ljeDMA_TX_DIRECTION_DEV_TO_MEM))
    {
        edma_dev = container_of(ddma, edma_device, slave_wdma);
    }
    else
    {
        printk("glcore: invaild direction parameter.\n");
    }

    return edma_dev;
}

/*
 * edma_dev_check_status
 * Description:
 *         pcie check staus function, it main use to check some bit value change
 * Parameter:
 *         iomem_t reg_base -- check register address
 *         uint32_t flag -- check bit field,0x1<<x
 *         uint32_t level -- check bit value,
 *         uint32_t timeou t-- check time limit
 * Return:
 *         pcie_status
 */
static edma_status edma_dev_check_status(
#if DMA_NEW
iomem_t reg_base,
#else
struct pci_dev *reg_base,
#endif
uint32_t reg, uint32_t flag,uint8_t level,uint32_t timeout)
{
    uint32_t count = timeout;

    do
    {
        if (level)
        {
            if(edma_read(reg_base, reg) & flag)
            {
                return EDMA_OK;
            }
        }
        else
        {
            if ((edma_read(reg_base, reg) & flag) == 0)
            {
                return EDMA_OK;
            }
        }
        count--;
        cpu_relax();
    } while((count != 0) || (timeout == 0));

    return EDMA_TIMEOUT;
}

static void edma_dev_write_engine_off(
#if DMA_NEW
iomem_t reg_base
#else
struct pci_dev *reg_base
#endif
)
{
    uint32_t reg;

    edma_write(reg_base, DMA_WRITE_ENGINE_EN_REG, 0x0);
    do
    {
        reg = edma_read(reg_base, DMA_WRITE_ENGINE_EN_REG);
        if (reg == 0)
        {
            break;
        }
        cpu_relax();
    } while(1);
}

static void edma_dev_write_engine_on(
#if DMA_NEW
	iomem_t reg_base
#else
	struct pci_dev *reg_base
#endif
)
{
    uint32_t reg;

    reg = edma_read(reg_base, DMA_WRITE_ENGINE_EN_REG);
    if (reg == 0)
    {
        edma_write(reg_base, DMA_WRITE_ENGINE_EN_REG, 0x1);
    }
}

static void edma_dev_read_engine_off(
#if DMA_NEW
	iomem_t reg_base
#else
	struct pci_dev *reg_base
#endif
)
{
    uint32_t reg;

    edma_write(reg_base, DMA_READ_ENGINE_EN_REG, 0x0);
    do
    {
        reg = edma_read(reg_base, DMA_READ_ENGINE_EN_REG);
        if (reg == 0)
        {
            break;
        }
        cpu_relax();
    } while(1);
}

static void edma_dev_read_engine_on(
#if DMA_NEW
	iomem_t reg_base
#else
	struct pci_dev *reg_base
#endif
)
{
    uint32_t reg;

    reg = edma_read(reg_base, DMA_READ_ENGINE_EN_REG);
    if (reg == 0)
    {
        edma_write(reg_base, DMA_READ_ENGINE_EN_REG, 0x1);
    }
}

/*
 * Function:
 *         void edma_dev_init(iomem_t reg_base)
 * Description:
 *         pcie dma initial function
 * Parameter:
 *         uint32_t reg_base--pcie dma config register base address;
 * Return:
 *         void
 */
static void edma_dev_init(
#if DMA_NEW
	iomem_t reg_base
#else
	struct pci_dev *reg_base
#endif
)
{
#if DMA_NEW
    edma_write(reg_base, DMA_WRITE_CHANNEL_ARB_WEIGHT_LOW_REG, 0x00008421);
    edma_write(reg_base, DMA_WRITE_CHANNEL_ARB_WEIGHT_HIGH_REG, 0x00008421);
    edma_write(reg_base, DMA_READ_CHANNEL_ARB_WEIGHT_LOW_REG, 0x00008421);
    edma_write(reg_base, DMA_READ_CHANNEL_ARB_WEIGHT_HIGH_REG, 0x00008421);
#endif
    /* init all write channel interrupt register */
    /* mask all write interrupt */

#if 1   //fq
    edma_write(reg_base, DMA_WRITE_INT_MASK_REG, 0x0);
    /* clear all write interrupt */
    edma_write(reg_base, DMA_WRITE_INT_CLEAR_REG, 0x0);
#else
    edma_write(reg_base, DMA_WRITE_INT_MASK_REG, 0x00FF00FF);
    /* clear all write interrupt */
    edma_write(reg_base, DMA_WRITE_INT_CLEAR_REG, 0x00FF00FF);
#endif


    edma_write(reg_base, DMA_WRITE_DONE_IMWR_LOW_REG, 0x00000000);
    edma_write(reg_base, DMA_WRITE_DONE_IMWR_HIGH_REG, 0x00000000);
    edma_write(reg_base, DMA_WRITE_ABORT_IMWR_LOW_REG, 0x00000000);
    edma_write(reg_base, DMA_WRITE_ABORT_IMWR_HIGH_REG, 0x00000000);
#if DMA_NEW
	edma_write(reg_base, DMA_WRITE_CH01_IMWR_DATA_REG, 0x00000000);
	edma_write(reg_base, DMA_WRITE_CH23_IMWR_DATA_REG, 0x00000000);
	edma_write(reg_base, DMA_WRITE_CH45_IMWR_DATA_REG, 0x00000000);
	edma_write(reg_base, DMA_WRITE_CH67_IMWR_DATA_REG, 0x00000000);
#else
    edma_write(reg_base, DMA_WRITE_CH0_IMWR_DATA_REG, 0x00000000);
#endif
    /* unable all write link abort error */
    edma_write(reg_base, DMA_WRITE_LINKED_LIST_ERR_EN_REG, 0x00000000);

    /* init all read channel interrupt register */
    /* mask all read interrupt */

#if 1   //fq
        edma_write(reg_base, DMA_READ_INT_MASK_REG, 0x0);
        /* clear all write interrupt */
        edma_write(reg_base, DMA_READ_INT_CLEAR_REG, 0x0);
#else
        edma_write(reg_base, DMA_READ_INT_MASK_REG, 0x00FF00FF);
        /* clear all write interrupt */
        edma_write(reg_base, DMA_READ_INT_CLEAR_REG, 0x00FF00FF);
#endif


    edma_write(reg_base, DMA_READ_DONE_IMWR_LOW_REG, 0x00000000);
    edma_write(reg_base, DMA_READ_DONE_IMWR_HIGH_REG, 0x00000000);
    edma_write(reg_base, DMA_READ_ABORT_IMWR_LOW_REG, 0x00000000);
    edma_write(reg_base, DMA_READ_ABORT_IMWR_HIGH_REG, 0x00000000);
#if DMA_NEW
    edma_write(reg_base/*pdev*/, DMA_READ_CH01_IMWR_DATA_REG, 0x00000000);
    edma_write(reg_base/*pdev*/, DMA_READ_CH23_IMWR_DATA_REG, 0x00000000);
    edma_write(reg_base/*pdev*/, DMA_READ_CH45_IMWR_DATA_REG, 0x00000000);
    edma_write(reg_base/*pdev*/, DMA_READ_CH67_IMWR_DATA_REG, 0x00000000);
#endif
    /* unable all read link abort error */
    edma_write(reg_base, DMA_READ_LINKED_LIST_ERR_EN_REG, 0x00000000);
}

/*
 * Function:
 *         void pcie_edma_arbi_weight_set(uint32_t channel, uint32_t direction, uint32_t reg_base)
 * Description:
 *         set transfer channel arbi weight
 * Parameter:
 *             uint32_t reg_base-- pcie dma config register  base address
 *             uint8_t channel--dma transfer channel,
 *             uint8_t direction--dma transfer dirction, write or read,
 *             uint32_t weight--set weight value,
 * Return:
 *         void
 */
#if DMA_NEW
static void edma_dev_arbi_weight_set(
    iomem_t reg_base,
    uint8_t channel,
    uint8_t direction,
    uint32_t weight
    )
{
    uint32_t reg;

    weight &= 0x0000001F;
    if (direction == ljeDMA_TX_DIRECTION_DEV_TO_MEM)
    {
        if (channel < 4)
        {
            reg = edma_read(reg_base, DMA_WRITE_CHANNEL_ARB_WEIGHT_LOW_REG);
            edma_write(reg_base, DMA_WRITE_CHANNEL_ARB_WEIGHT_LOW_REG,
                reg | (weight << (channel * 5)));
        }
        else if (channel < 8)
        {
            reg = edma_read(reg_base, DMA_WRITE_CHANNEL_ARB_WEIGHT_HIGH_REG);
            edma_write(reg_base, DMA_WRITE_CHANNEL_ARB_WEIGHT_HIGH_REG,
                reg | (weight << (( channel - 4) * 5)));
        }
    }
    else
    {
        if( channel < 4)
        {
            reg = edma_read(reg_base, DMA_READ_CHANNEL_ARB_WEIGHT_LOW_REG);
            edma_write(reg_base, DMA_READ_CHANNEL_ARB_WEIGHT_LOW_REG,
                reg | (weight << (channel * 5)));
        }
        else if(channel < 8)
        {
            reg = edma_read(reg_base, DMA_READ_CHANNEL_ARB_WEIGHT_HIGH_REG);
            edma_write(reg_base, DMA_READ_CHANNEL_ARB_WEIGHT_HIGH_REG,
                reg | (weight << (( channel - 4) * 5)));
        }
    }
}
#endif
static void edma_dev_transfer_start(
#if DMA_NEW
	iomem_t reg_base,
#else
	struct pci_dev *reg_base,
#endif
uint8_t channel, uint8_t direction)
{
    switch (direction)
    {
        case ljeDMA_TX_DIRECTION_DEV_TO_MEM:
            edma_dev_write_engine_on(reg_base);
            /* enable the write doorbell */
            edma_write(reg_base, DMA_WRITE_DOORBELL_REG, channel);
            break;
        case ljeDMA_TX_DIRECTION_MEM_TO_DEV:
            edma_dev_read_engine_on(reg_base);
            /*  enable the read doorbell */
            edma_write(reg_base, DMA_READ_DOORBELL_REG, channel);
            break;
        default:
            printk("glcore: invaild pcie dDMA parameter.\n");
            break;
    }
}

/*
 * Function:
 *         uint32_t edma_dev_get_channel_status
 * Description:
 *         get channel hw status
 * Parameter:
 *         iomem_t reg_base-- pcie dma config register  base address
 *         uint8_t channel--dma transfer channel,
 *         uint8_t direction--dma transfer dirction, write or read,
 * Return:
 *         DMA_CS_RUNNING, DMA_CS_HALTED, DMA_CS_STOPPED
 */
static uint32_t edma_dev_get_channel_status(
#if DMA_NEW
    iomem_t reg_base,
#else
	struct pci_dev *reg_base,
#endif
    uint8_t channel,
    uint8_t direction
    )
{

    uint32_t reg = 0;

#if DMA_NEW
    /* read comtrol1 register to check CS field */
    if (direction == ljeDMA_TX_DIRECTION_DEV_TO_MEM)
    {
        reg = edma_read(reg_base, DMA_CH_CONTROL1_WRCH_REG(channel)) & (0x3L << 5);
    }
    else if ( direction == ljeDMA_TX_DIRECTION_MEM_TO_DEV)
    {
        reg = edma_read(reg_base, DMA_CH_CONTROL1_RDCH_REG(channel)) & (0x3L << 5);
    }
#else
	if (direction == ljeDMA_TX_DIRECTION_DEV_TO_MEM)
	{
		edma_write(reg_base, DMA_CHANNEL_CTX_INDEX_REG, 0x0);
		reg = edma_read(reg_base, DMA_CHANNEL_CTRL_REG) & (0x3L << 5);
	}
    else if ( direction == ljeDMA_TX_DIRECTION_MEM_TO_DEV)
    {
		edma_write(reg_base, DMA_CHANNEL_CTX_INDEX_REG, 0x80000000);
		reg = edma_read(reg_base, DMA_CHANNEL_CTRL_REG) & (0x3L << 5);
    }
#endif

    return reg;
}

/*
 * Function:
 *         edma_status edma_dev_transfer_stop(iomem_t reg_base, uint32_t channel, uint32_t direction)
 * Description:
 *         pcie dma stop current transfer
 * Parameter:
 *         iomem_t reg_base-- pcie dma config register  base address
 *         uint8_t channel--dma transfer channel,
 *         uint8_t direction--dma transfer dirction, write or read,
 * Return:
 *         edma_status
 */
static edma_status edma_dev_transfer_stop(
#if DMA_NEW
	iomem_t reg_base,
#else
	struct pci_dev *reg_base,
#endif
    uint8_t channel,
    uint8_t direction
    )
{
    uint32_t reg;

    if (direction == ljeDMA_TX_DIRECTION_DEV_TO_MEM)
    {
        /* ensure is runing */
        if (edma_dev_get_channel_status(reg_base, channel, direction) == DMA_CS_RUNNING)
        {
            /* stop the channel */
            edma_write(reg_base, DMA_WRITE_DOORBELL_REG,((uint32_t)channel) | DMA_TRANSFER_STOP);
            if (edma_dev_check_status(reg_base, DMA_WRITE_INT_STATUS_REG,
                DMA_ABORT_INT(channel), 1, 5000) == EDMA_OK)
            {
                /* clear the status */
                edma_write(reg_base, DMA_WRITE_INT_STATUS_REG, DMA_ABORT_INT(channel));
                /* read comtrol1 register to check CS field */
				#if DMA_NEW
                reg = edma_read(reg_base, DMA_CH_CONTROL1_WRCH_REG(channel)) & 0x00000060;
				#else
                reg = edma_read(reg_base, DMA_CHANNEL_CTRL_REG) & 0x00000060;
				#endif
                /* ensure is sttopped */
                if (reg == DMA_CS_STOPPED)
                {
                    return EDMA_ABORT;
                }
            }
        }
    }
    else if ( direction == ljeDMA_TX_DIRECTION_MEM_TO_DEV)
    {
        /* ensure is runing */
        if (edma_dev_get_channel_status(reg_base, channel, direction) == DMA_CS_RUNNING)
        {
            /* stop the channel */
            edma_write(reg_base, DMA_READ_DOORBELL_REG,((uint32_t)channel) | DMA_TRANSFER_STOP);
            if (edma_dev_check_status(reg_base, DMA_READ_INT_STATUS_REG,
                DMA_ABORT_INT(channel), 1, 5000)== EDMA_OK)
            {
                /* read comtrol1 register to check CS field */
				#if DMA_NEW
                reg = edma_read(reg_base, DMA_CH_CONTROL1_RDCH_REG(channel)) & 0x00000060;
				#else
                reg = edma_read(reg_base, DMA_CHANNEL_CTRL_REG) & 0x00000060;
				#endif
                /* ensure is runing */
                if(reg == DMA_CS_STOPPED)
                {
                    return EDMA_ABORT;
                }
            }
        }
    }

    return EDMA_ERROR;
}


/*
 * Function:
 *         edma_status edma_dev_write_channel_config(edma_hw_config *config)
 * Description:
 *         pcie dma write channel configure function
 * Parameter:
 *         edma_hw_config *config--pcie dma write channel configure information;
 * Return:
 *         edma_status
 */
static edma_status edma_dev_write_channel_config(
    edma_hw_config *config
    )
{
#if DMA_NEW
    iomem_t reg_base = config->reg_base;
    uint32_t reg = edma_read(reg_base, DMA_CHANNEL_CTRL_REG) & 0x0000000F;
#else
	struct pci_dev *reg_base = config->reg_base;
    uint32_t reg = 1;
#endif
    uint32_t imwr_data_reg;

    /* if channel does not exist, return error */
    if (config->channel >= reg)
    {
        return EDMA_ERROR;
    }

//	edma_write(reg_base, DMA_CHANNEL_CTX_INDEX_REG, config->channel);   fq

    /*  enable the write engine */
    edma_dev_write_engine_on(reg_base);
#if DMA_NEW
    /* set channel arbi weight */
    edma_dev_arbi_weight_set(reg_base, config->channel, ljeDMA_TX_DIRECTION_DEV_TO_MEM, config->arbi_weight);
#endif
    /* set write interrupt mask */
    //edma_write(reg_base, DMA_WRITE_INT_MASK_REG, config->interrupt_mask);
    /* set link err en */
    edma_write(reg_base, DMA_WRITE_LINKED_LIST_ERR_EN_REG, config->link_err_en);
    reg = DMA_DONE_INT(config->channel) | DMA_ABORT_INT(config->channel);
    /* clear the interrupt */
    edma_write(reg_base, DMA_WRITE_INT_CLEAR_REG, reg);
    /* set write done imwr address low */
    edma_write(reg_base, DMA_WRITE_DONE_IMWR_LOW_REG, config->done_imwr_addr_low);
    /* set write done imwr address high */
    edma_write(reg_base, DMA_WRITE_DONE_IMWR_HIGH_REG, config->done_imwr_addr_high);
    /* set write abort imwr address low */
    edma_write(reg_base, DMA_WRITE_ABORT_IMWR_LOW_REG, config->abort_imwr_addr_low);
    /* set write abort imwr address high */
    edma_write(reg_base, DMA_WRITE_ABORT_IMWR_HIGH_REG, config->abort_imwr_addr_high);
    /* set imwr data */
#if DMA_NEW
    switch (config->channel)
    {
        case 0:
        case 1:
            imwr_data_reg = DMA_WRITE_CH01_IMWR_DATA_REG;
            break;
        case 2:
        case 3:
            imwr_data_reg = DMA_WRITE_CH23_IMWR_DATA_REG;
            break;
        case 4:
        case 5:
            imwr_data_reg = DMA_WRITE_CH45_IMWR_DATA_REG;
            break;
        case 6:
        case 7:
            imwr_data_reg = DMA_WRITE_CH67_IMWR_DATA_REG;
            break;
        default:
            return EDMA_ERROR;
    }
#else
	imwr_data_reg = DMA_WRITE_CH0_IMWR_DATA_REG;
#endif

    reg = edma_read(reg_base, imwr_data_reg);

    switch (config->channel)
    {
        case 0:
        case 2:
        case 4:
        case 6:
            reg &= ~0x0000ffff;
            edma_write(reg_base, imwr_data_reg,
                        reg | (uint32_t)(config->imwr_data & 0x0000ffff));
            break;
        case 1:
        case 3:
        case 5:
        case 7:
            reg &= ~0xffff0000;
            edma_write(reg_base, imwr_data_reg,
                    reg | ((uint32_t)(config->imwr_data & 0x0000ffff) << 16));
            break;
        default:
            return EDMA_ERROR;
    }
#if DMA_NEW
    edma_write(reg_base, DMA_CH_CONTROL1_WRCH_REG(config->channel), config->control1);
    edma_write(reg_base, DMA_CH_CONTROL2_WRCH_REG(config->channel), config->control2);
    /* set the transfer size */
    edma_write(reg_base, DMA_TRANSFER_SIZE_WRCH_REG(config->channel), config->transfer_size);
    /* set the transfer source address */
    edma_write(reg_base, DMA_SAR_LOW_WRCH_REG(config->channel), config->sar_low);
    edma_write(reg_base, DMA_SAR_HIGH_WRCH_REG(config->channel), config->sar_high);
    /* set the transfer dest address */
    edma_write(reg_base, DMA_DAR_LOW_WRCH_REG(config->channel), config->dar_low);
    edma_write(reg_base, DMA_DAR_HIGH_WRCH_REG(config->channel), config->dar_high);
    /* set LLP */
    edma_write(reg_base, DMA_LLP_LOW_WRCH_REG(config->channel), config->llp_low);
    edma_write(reg_base, DMA_LLP_HIGH_WRCH_REG(config->channel), config->llp_high);
#else
	edma_write(reg_base, DMA_CHANNEL_CTRL_REG, config->control1);
	//edma_write(reg_base, DMA_CH_CONTROL2_WRCH_REG(config->channel), config->control2);
	/* set the transfer size */
	edma_write(reg_base, DMA_TRANSFER_SIZE_REG, config->transfer_size);
	/* set the transfer source address */
	edma_write(reg_base, DMA_SAR_LOW_REG, config->sar_low);
	edma_write(reg_base, DMA_SAR_HIGH_REG, config->sar_high);
	/* set the transfer dest address */
	edma_write(reg_base, DMA_DAR_LOW_REG, config->dar_low);
	edma_write(reg_base, DMA_DAR_HIGH_REG, config->dar_high);

	/* set LLP */
	edma_write(reg_base, DMA_LLP_LOW_REG, config->llp_low);
	edma_write(reg_base, DMA_LLP_HIGH_REG, config->llp_high);
#endif
#if EDMA_DEBUG
#if DMA_NEW
    printk("write back:channel=0x%x. \r\n", config->channel);
    printk("write back:enable=0x%x. \r\n", edma_read(reg_base, DMA_WRITE_ENGINE_EN_REG));
    printk("write back:sar_low=0x%x. \r\n", edma_read(reg_base, DMA_SAR_LOW_WRCH_REG(config->channel)));
    printk("write back:sar_high=0x%x. \r\n", edma_read(reg_base, DMA_SAR_HIGH_WRCH_REG(config->channel)));
    printk("write back:dar_low=0x%x. \r\n", edma_read(reg_base, DMA_DAR_LOW_WRCH_REG(config->channel)));
    printk("write back:dar_high=0x%x. \r\n", edma_read(reg_base, DMA_DAR_HIGH_WRCH_REG(config->channel)));
    printk("write back:length=0x%x. \r\n", edma_read(reg_base, DMA_TRANSFER_SIZE_WRCH_REG(config->channel)));
    printk("write back:done_imwr_addr_low=0x%x. \r\n", edma_read(reg_base, DMA_WRITE_DONE_IMWR_LOW_REG));
    printk("write back:done_imwr_addr_high=0x%x. \r\n",edma_read(reg_base, DMA_WRITE_DONE_IMWR_HIGH_REG));
    printk("write back:abort_imwr_addr_low=0x%x. \r\n", edma_read(reg_base, DMA_WRITE_ABORT_IMWR_LOW_REG));
    printk("write back:abort_imwr_addr_high=0x%x. \r\n", edma_read(reg_base, DMA_WRITE_ABORT_IMWR_HIGH_REG));
    printk("write back:imwr_data=0x%x. \r\n", edma_read(reg_base, imwr_data_reg));
    printk("write back:interrupt_mask=0x%x. \r\n", edma_read(reg_base, DMA_WRITE_INT_MASK_REG));
    printk("write back:link_err_en=0x%x. \r\n", edma_read(reg_base, DMA_WRITE_LINKED_LIST_ERR_EN_REG));
    printk("write back:control1=0x%x. \r\n", edma_read(reg_base, DMA_CH_CONTROL1_WRCH_REG(config->channel)));
    printk("write back:control2=0x%x. \r\n", edma_read(reg_base, DMA_CH_CONTROL2_WRCH_REG(config->channel)));
    printk("write back:llp_low=0x%x. \r\n", edma_read(reg_base, DMA_LLP_LOW_WRCH_REG(config->channel)));
    printk("write back:llp_high=0x%x. \r\n", edma_read(reg_base, DMA_LLP_HIGH_WRCH_REG(config->channel)));
#else
	printk("write back:channel=0x%x. \r\n", config->channel);
	printk("write back:enable=0x%x. \r\n", edma_read(reg_base, DMA_WRITE_ENGINE_EN_REG));
	printk("write back:sar_low=0x%x. \r\n", edma_read(reg_base, DMA_SAR_LOW_REG));
	printk("write back:sar_high=0x%x. \r\n", edma_read(reg_base, DMA_SAR_HIGH_REG));
	printk("write back:dar_low=0x%x. \r\n", edma_read(reg_base, DMA_DAR_LOW_REG));
	printk("write back:dar_high=0x%x. \r\n", edma_read(reg_base, DMA_DAR_HIGH_REG));
	printk("write back:length=0x%x. \r\n", edma_read(reg_base, DMA_TRANSFER_SIZE_REG));
	printk("write back:done_imwr_addr_low=0x%x. \r\n", edma_read(reg_base, DMA_WRITE_DONE_IMWR_LOW_REG));
	printk("write back:done_imwr_addr_high=0x%x. \r\n",edma_read(reg_base, DMA_WRITE_DONE_IMWR_HIGH_REG));
	printk("write back:abort_imwr_addr_low=0x%x. \r\n", edma_read(reg_base, DMA_WRITE_ABORT_IMWR_LOW_REG));
	printk("write back:abort_imwr_addr_high=0x%x. \r\n", edma_read(reg_base, DMA_WRITE_ABORT_IMWR_HIGH_REG));
	printk("write back:imwr_data=0x%x. \r\n", edma_read(reg_base, imwr_data_reg));
	printk("write back:interrupt_mask=0x%x. \r\n", edma_read(reg_base, DMA_WRITE_INT_MASK_REG));
	printk("write back:link_err_en=0x%x. \r\n", edma_read(reg_base, DMA_WRITE_LINKED_LIST_ERR_EN_REG));
	printk("write back:control1=0x%x. \r\n", edma_read(reg_base, DMA_CHANNEL_CTRL_REG));
	//printk("write back:control2=0x%x. \r\n", edma_read(reg_base, DMA_CH_CONTROL2_WRCH_REG));
	printk("write back:llp_low=0x%x. \r\n", edma_read(reg_base, DMA_LLP_LOW_REG));
	printk("write back:llp_high=0x%x. \r\n", edma_read(reg_base, DMA_LLP_HIGH_REG));
#endif
#endif

    return EDMA_OK;

}

/*
 * Function:
 *         edma_status edma_dev_read_channel_config(edma_hw_config *config)
 * Description:
 *         pcie dma read channel configure function
 * Parameter:
 *         edma_hw_config *config--pcie dma read channel configure information;
 * Return:
 *         edma_status
 */
static edma_status edma_dev_read_channel_config(
    edma_hw_config *config
    )
{
#if DMA_NEW
    iomem_t reg_base = config->reg_base;
    uint32_t reg = (edma_read(reg_base, DMA_CHANNEL_CTRL_REG) & 0x000F0000) >> 16;
#else
	struct pci_dev *reg_base = config->reg_base;
	uint32_t reg = 1;
#endif
    uint32_t imwr_data_reg;

    /* if channel does not exist, return error */
    if (config->channel >= reg)
    {
        return EDMA_ERROR;
    }

//	edma_write(reg_base, DMA_CHANNEL_CTX_INDEX_REG, 0x80000000);  fq

    /*  enable the read engine */
    edma_dev_read_engine_on(reg_base);
#if DMA_NEW
    /* set channel arbi weight */
    edma_dev_arbi_weight_set(reg_base, config->channel, ljeDMA_TX_DIRECTION_MEM_TO_DEV, config->arbi_weight);
#endif
    /* set write interrupt mask */
    //edma_write(reg_base, DMA_READ_INT_MASK_REG, config->interrupt_mask);
    /* set link err en */
    edma_write(reg_base, DMA_READ_LINKED_LIST_ERR_EN_REG, config->link_err_en);
    reg = DMA_DONE_INT(config->channel) | DMA_ABORT_INT(config->channel);
    /* clear the interrupt */
    edma_write(reg_base, DMA_READ_INT_CLEAR_REG, reg);
    /* set write done imwr address */
    edma_write(reg_base, DMA_READ_DONE_IMWR_LOW_REG, config->done_imwr_addr_low);
    edma_write(reg_base, DMA_READ_DONE_IMWR_HIGH_REG, config->done_imwr_addr_high);
    /* set write abort imwr address */
    edma_write(reg_base, DMA_READ_ABORT_IMWR_LOW_REG, config->abort_imwr_addr_low);
    edma_write(reg_base, DMA_READ_ABORT_IMWR_HIGH_REG, config->abort_imwr_addr_high);
    /* set imwr data */
#if DMA_NEW
    switch (config->channel)
    {
        case 0:
        case 1:
            imwr_data_reg = DMA_READ_CH01_IMWR_DATA_REG;
            break;
        case 2:
        case 3:
            imwr_data_reg = DMA_READ_CH23_IMWR_DATA_REG;
            break;
        case 4:
        case 5:
            imwr_data_reg = DMA_READ_CH45_IMWR_DATA_REG;
            break;
        case 6:
        case 7:
            imwr_data_reg = DMA_READ_CH67_IMWR_DATA_REG;
            break;
        default:
            return EDMA_ERROR;
    }
#else
	imwr_data_reg = DMA_WRITE_CH0_IMWR_DATA_REG;//abc: is it ok ???
#endif

    reg = edma_read(reg_base, imwr_data_reg);

    switch (config->channel)
    {
        case 0:
        case 2:
        case 4:
        case 6:
            reg &= ~0x0000ffff;
            edma_write(reg_base, imwr_data_reg,
                        reg | (uint32_t)(config->imwr_data & 0x0000ffff));
            break;
        case 1:
        case 3:
        case 5:
        case 7:
            reg &= ~0xffff0000;
            edma_write(reg_base, imwr_data_reg,
                    reg | ((uint32_t)(config->imwr_data & 0x0000ffff) << 16));
            break;
        default:
            return EDMA_ERROR;
    }

#if DMA_NEW
    /* config control1 */
    edma_write(reg_base, DMA_CH_CONTROL1_RDCH_REG(config->channel), config->control1);
    edma_write(reg_base, DMA_CH_CONTROL2_RDCH_REG(config->channel), config->control2);
    /* set the transfer size */
    edma_write(reg_base, DMA_TRANSFER_SIZE_RDCH_REG(config->channel), config->transfer_size);
    /* set the transfer source address */
    edma_write(reg_base, DMA_SAR_LOW_RDCH_REG(config->channel), config->sar_low);
    edma_write(reg_base, DMA_SAR_HIGH_RDCH_REG(config->channel), config->sar_high);
    /* set the transfer dest address */
    edma_write(reg_base, DMA_DAR_LOW_RDCH_REG(config->channel), config->dar_low);
    edma_write(reg_base, DMA_DAR_HIGH_RDCH_REG(config->channel), config->dar_high);
    /* set LLP */
    edma_write(reg_base, DMA_LLP_LOW_RDCH_REG(config->channel), config->llp_low);
    edma_write(reg_base, DMA_LLP_HIGH_RDCH_REG(config->channel), config->llp_high);
#else
	/* config control1 */
	edma_write(reg_base, DMA_CHANNEL_CTRL_REG, config->control1);
//	edma_write(reg_base, DMA_CH_CONTROL2_RDCH_REG, config->control2);
	/* set the transfer size */
	edma_write(reg_base, DMA_TRANSFER_SIZE_REG, config->transfer_size);
	/* set the transfer source address */
	edma_write(reg_base, DMA_SAR_LOW_REG, config->sar_low);
	edma_write(reg_base, DMA_SAR_HIGH_REG, config->sar_high);
	/* set the transfer dest address */
	edma_write(reg_base, DMA_DAR_LOW_REG, config->dar_low);
	edma_write(reg_base, DMA_DAR_HIGH_REG, config->dar_high);
	/* set LLP */
	edma_write(reg_base, DMA_LLP_LOW_REG, config->llp_low);
	edma_write(reg_base, DMA_LLP_HIGH_REG, config->llp_high);
#endif

#if EDMA_DEBUG
#if DMA_NEW
    printk("read back:channel=0x%x. \r\n", config->channel);
    printk("read back:enable=0x%x. \r\n", edma_read(reg_base, DMA_READ_ENGINE_EN_REG));
    printk("read back:sar_low=0x%x. \r\n", edma_read(reg_base, DMA_SAR_LOW_RDCH_REG(config->channel)));
    printk("read back:sar_high=0x%x. \r\n", edma_read(reg_base, DMA_SAR_HIGH_RDCH_REG(config->channel)));
    printk("read back:dar_low=0x%x. \r\n", edma_read(reg_base, DMA_DAR_LOW_RDCH_REG(config->channel)));
    printk("read back:dar_high=0x%x. \r\n", edma_read(reg_base, DMA_DAR_HIGH_RDCH_REG(config->channel)));
    printk("read back:length=0x%x. \r\n", edma_read(reg_base, DMA_TRANSFER_SIZE_RDCH_REG(config->channel)));
    printk("read back:done_imwr_addr_low=0x%x. \r\n", edma_read(reg_base, DMA_READ_DONE_IMWR_LOW_REG));
    printk("read back:done_imwr_addr_high=0x%x. \r\n",edma_read(reg_base, DMA_READ_DONE_IMWR_HIGH_REG));
    printk("read back:abort_imwr_addr_low=0x%x. \r\n", edma_read(reg_base, DMA_READ_ABORT_IMWR_LOW_REG));
    printk("read back:abort_imwr_addr_high=0x%x. \r\n", edma_read(reg_base, DMA_READ_ABORT_IMWR_HIGH_REG));
    printk("read back:imwr_data=0x%x. \r\n", edma_read(reg_base, imwr_data_reg));
    printk("read back:interrupt_mask=0x%x. \r\n", edma_read(reg_base, DMA_READ_INT_MASK_REG));
    printk("read back:link_err_en=0x%x. \r\n", edma_read(reg_base, DMA_READ_LINKED_LIST_ERR_EN_REG));
    printk("read back:control1=0x%x. \r\n", edma_read(reg_base, DMA_CH_CONTROL1_RDCH_REG(config->channel)));
    printk("read back:control2=0x%x. \r\n", edma_read(reg_base, DMA_CH_CONTROL2_RDCH_REG(config->channel)));
    printk("read back:llp_low=0x%x. \r\n", edma_read(reg_base, DMA_LLP_LOW_RDCH_REG(config->channel)));
    printk("read back:llp_high=0x%x. \r\n", edma_read(reg_base, DMA_LLP_HIGH_RDCH_REG(config->channel)));
#else
	printk("read back:channel=0x%x. \r\n", config->channel);
	printk("read back:enable=0x%x. \r\n", edma_read(reg_base, DMA_READ_ENGINE_EN_REG));
	printk("read back:sar_low=0x%x. \r\n", edma_read(reg_base, DMA_SAR_LOW_REG));
	printk("read back:sar_high=0x%x. \r\n", edma_read(reg_base, DMA_SAR_HIGH_REG));
	printk("read back:dar_low=0x%x. \r\n", edma_read(reg_base, DMA_DAR_LOW_REG));
	printk("read back:dar_high=0x%x. \r\n", edma_read(reg_base, DMA_DAR_HIGH_REG));
	printk("read back:length=0x%x. \r\n", edma_read(reg_base, DMA_TRANSFER_SIZE_REG));
	printk("read back:done_imwr_addr_low=0x%x. \r\n", edma_read(reg_base, DMA_READ_DONE_IMWR_LOW_REG));
	printk("read back:done_imwr_addr_high=0x%x. \r\n",edma_read(reg_base, DMA_READ_DONE_IMWR_HIGH_REG));
	printk("read back:abort_imwr_addr_low=0x%x. \r\n", edma_read(reg_base, DMA_READ_ABORT_IMWR_LOW_REG));
	printk("read back:abort_imwr_addr_high=0x%x. \r\n", edma_read(reg_base, DMA_READ_ABORT_IMWR_HIGH_REG));
	printk("read back:imwr_data=0x%x. \r\n", edma_read(reg_base, imwr_data_reg));
	printk("read back:interrupt_mask=0x%x. \r\n", edma_read(reg_base, DMA_READ_INT_MASK_REG));
	printk("read back:link_err_en=0x%x. \r\n", edma_read(reg_base, DMA_READ_LINKED_LIST_ERR_EN_REG));
	printk("read back:control1=0x%x. \r\n", edma_read(reg_base, DMA_CHANNEL_CTRL_REG));
//	printk("read back:control2=0x%x. \r\n", edma_read(reg_base, DMA_CH_CONTROL2_RDCH_REG(config->channel)));
	printk("read back:llp_low=0x%x. \r\n", edma_read(reg_base, DMA_LLP_LOW_REG));
	printk("read back:llp_high=0x%x. \r\n", edma_read(reg_base, DMA_LLP_HIGH_REG));
#endif
#endif
    return EDMA_OK;
}


 /*
 * Function:
 *         edma_status edma_dev_write_transfer(edma_transfer_info *info)
 * Description:
 *         pcie dma write transfer function
 * Parameter:
 *         edma_transfer_info *info--pcie dma write transfer handle, it contains all necessary parameters;
 * Return:
 *         edma_status
 */
static edma_status edma_dev_write_transfer(
    edma_transfer_info *info
    )
{
    uint32_t reg;
    edma_hw_config config;

    /* double config the info according to transfer mode */
    config.channel = info->channel;
    config.reg_base = info->reg_base;
    config.arbi_weight = 0xf;
    if (info->transfer_mode == EDMA_TRANSFER_MODE_NORMAL)
    {
        config.transfer_size = info->transfer_size;
        config.sar_low = QWORD_LO(info->src_addr);
        config.sar_high = QWORD_HI(info->src_addr);
        config.dar_low = QWORD_LO(info->dst_addr);
        config.dar_high = QWORD_HI(info->dst_addr);
        config.link_err_en = 0;
        config.llp_high = 0;
        config.llp_low = 0;
    }
    else
    {
        config.transfer_size = 0;
        config.sar_low = 0;
        config.sar_high = 0;
        config.dar_low = 0;
        config.dar_high = 0;
        config.llp_high = QWORD_HI(info->lle);
        config.llp_low = QWORD_LO(info->lle);
    }
    /* cpu config local dma */
    if (info->cpu_mode == EDMA_CPU_MODE_LOCAL)
    {
        config.done_imwr_addr_low = 0;
        config.done_imwr_addr_high = 0;
        config.abort_imwr_addr_low = 0;
        config.abort_imwr_addr_high = 0;
        config.imwr_data = 0;
        /* interrupt mode */
        if (info->check_mode == EDMA_CHECK_MODE_INTERRUPT)
        {
            reg = edma_read(info->reg_base, DMA_WRITE_INT_MASK_REG);
            /* unmask the interrupt */
            reg &= (~(DMA_DONE_INT(info->channel) | DMA_ABORT_INT(info->channel)));
            config.interrupt_mask = reg;

            if (info->transfer_mode == EDMA_TRANSFER_MODE_NORMAL)
            {
                /* open local cpu interrupt */
                config.control1 = DMA_LIE | DMA_TD;
            }
            else if (info->transfer_mode == EDMA_TRANSFER_MODE_LINK)
            {
                /* LL mode */
                reg = edma_read(info->reg_base, DMA_WRITE_LINKED_LIST_ERR_EN_REG);
                reg |= (DMA_LLLAIE(info->channel));
                config.link_err_en = reg;
                /* open local cpu interrupt */
                config.control1=DMA_TD|DMA_LIE|DMA_LLE|DMA_CCS;
            }
            else
            {
                printk("invalid transfer mode\n");
                return EDMA_ERROR;
            }
        }
        /* polling mode */
        else if (info->check_mode == EDMA_CHECK_MODE_POLLING)
        {
            reg = edma_read(info->reg_base, DMA_WRITE_INT_MASK_REG);
            /* mask the interrupt */
            reg |= (DMA_DONE_INT(info->channel) | DMA_ABORT_INT(info->channel));
            config.interrupt_mask = reg;

            if (info->transfer_mode == EDMA_TRANSFER_MODE_NORMAL)
            {
                /* close local cpu interrupt */
                config.control1 = DMA_TD;
            }
            else if (info->transfer_mode == EDMA_TRANSFER_MODE_LINK)
            {
                /* LL mode */
                reg = edma_read(info->reg_base, DMA_WRITE_LINKED_LIST_ERR_EN_REG);
                reg |= (DMA_LLLAIE(info->channel));
                config.link_err_en = reg;
                /* open local cpu interrupt */
                config.control1=DMA_TD|DMA_LLE|DMA_CCS;
            }
            else
            {
                printk("invalid transfer mode\n");
                return EDMA_ERROR;
            }
        }
        else
        {
            printk("invalid check mode\n");
            return EDMA_ERROR;
        }
        config.control2 = 0;
    }
    /* cpu config remote dma */
    else if (info->cpu_mode == EDMA_CPU_MODE_REMOTE)
    {
        /* interrupt mode */
        if (info->check_mode == EDMA_CHECK_MODE_INTERRUPT)
        {
            config.done_imwr_addr_low = QWORD_LO(info->msi_addr);
            config.done_imwr_addr_high = QWORD_HI(info->msi_addr);
            config.abort_imwr_addr_low = QWORD_LO(info->msi_addr);
            config.abort_imwr_addr_high = QWORD_HI(info->msi_addr);
            config.imwr_data = info->msi_data;
            reg = edma_read(info->reg_base, DMA_WRITE_INT_MASK_REG);
            /* unmask the interrupt */
            reg &= (~(DMA_DONE_INT(info->channel) | DMA_ABORT_INT(info->channel)));
            config.interrupt_mask = reg;

            if (info->transfer_mode == EDMA_TRANSFER_MODE_NORMAL)
            {
                /* open local cpu interrupt */
                config.control1 = DMA_RIE | DMA_TD | DMA_LIE;
            }
            else if (info->transfer_mode == EDMA_TRANSFER_MODE_LINK)
            {
                /* LL mode */
                reg = edma_read(info->reg_base, DMA_WRITE_LINKED_LIST_ERR_EN_REG);
                reg |= (DMA_LLRAIE(info->channel));
                config.link_err_en = reg;
                /* open local cpu interrupt */
                config.control1=DMA_TD|DMA_LIE|DMA_RIE|DMA_LLE|DMA_CCS;
            }
            else
            {
                printk("invalid transfer mode\n");
                return EDMA_ERROR;
            }
        }
        /* polling mode */
        else if (info->check_mode == EDMA_CHECK_MODE_POLLING)
        {
            config.done_imwr_addr_low = 0;
            config.done_imwr_addr_high = 0;
            config.abort_imwr_addr_low = 0;
            config.abort_imwr_addr_high = 0;
            config.imwr_data = 0;
            reg = edma_read(info->reg_base, DMA_WRITE_INT_MASK_REG);
            /* mask the interrupt */
            reg |= (DMA_DONE_INT(info->channel) | DMA_ABORT_INT(info->channel));
            config.interrupt_mask = reg;
            /* close local cpu interrupt */
            config.control1 = DMA_LIE | DMA_TD;

            if (info->transfer_mode == EDMA_TRANSFER_MODE_NORMAL)
            {
                /* open local cpu interrupt */
                config.control1 = DMA_LIE | DMA_TD;
            }
            else if (info->transfer_mode == EDMA_TRANSFER_MODE_LINK)
            {
                /* LL mode */
                reg = edma_read(info->reg_base, DMA_WRITE_LINKED_LIST_ERR_EN_REG);
                reg |= (DMA_LLRAIE(info->channel));
                config.link_err_en = reg;
                /* open local cpu interrupt */
                config.control1=DMA_TD|DMA_LLE|DMA_CCS;
            }
            else
            {
                printk("invalid transfer mode\n");
                return EDMA_ERROR;
            }
        }
        else
        {
            printk("invalid check mode\n");
            return EDMA_ERROR;
        }
        config.control2 = 0;
    }
    else
    {
        printk("invalid cpu mode\n");
        return EDMA_ERROR;
    }

    /* config the write channel */
    if (edma_dev_write_channel_config(&config) == EDMA_ERROR)
    {
        return EDMA_ERROR;
    }
    /* open the door bell, start transfer */
    edma_dev_transfer_start(info->reg_base, info->channel, ljeDMA_TX_DIRECTION_DEV_TO_MEM);

    return EDMA_OK;
}

/*
 * edma_read_transfer
 */
static edma_status edma_dev_read_transfer(
    edma_transfer_info *info
    )
{
    uint32_t reg;
    edma_hw_config config;

    /* double config the info according to transfer mode */
    config.channel = info->channel;
    config.reg_base = info->reg_base;
    config.arbi_weight = 0xf;
    if (info->transfer_mode == EDMA_TRANSFER_MODE_NORMAL)
    {
        config.transfer_size = info->transfer_size;
        config.sar_low = QWORD_LO(info->src_addr);
        config.sar_high = QWORD_HI(info->src_addr);
        config.dar_low = QWORD_LO(info->dst_addr);
        config.dar_high = QWORD_HI(info->dst_addr);
        config.link_err_en = 0;
        config.llp_high = 0;
        config.llp_low = 0;
    }
    else
    {
        config.transfer_size = 0;
        config.sar_low = 0;
        config.sar_high = 0;
        config.dar_low = 0;
        config.dar_high = 0;
        config.llp_high = QWORD_HI(info->lle);
        config.llp_low = QWORD_LO(info->lle);
    }
    /* cpu config local dma */
    if (info->cpu_mode == EDMA_CPU_MODE_LOCAL)
    {
        config.done_imwr_addr_low = 0;
        config.done_imwr_addr_high = 0;
        config.abort_imwr_addr_low = 0;
        config.abort_imwr_addr_high = 0;
        config.imwr_data = 0;
        /* interrupt mode */
        if(info->check_mode == EDMA_CHECK_MODE_INTERRUPT)
        {
            reg = edma_read(info->reg_base, DMA_READ_INT_MASK_REG);
            /* unmask the interrupt */
            reg &= (~(DMA_DONE_INT(info->channel) | DMA_ABORT_INT(info->channel)));
            config.interrupt_mask = reg;

            if (info->transfer_mode == EDMA_TRANSFER_MODE_NORMAL)
            {
                /* open local cpu interrupt */
                config.control1 = DMA_LIE | DMA_TD;
            }
            else if (info->transfer_mode == EDMA_TRANSFER_MODE_LINK)
            {
                reg = edma_read(info->reg_base, DMA_READ_LINKED_LIST_ERR_EN_REG);
                /* unmask the interrupt */
                reg |= (DMA_LLLAIE(info->channel));
                config.link_err_en = reg;

                /* open local cpu interrupt */
                config.control1=DMA_TD|DMA_LIE|DMA_LLE|DMA_CCS;
            }
            else
            {
                printk("invalid transfer mode\n");
                return EDMA_ERROR;
            }
        }
        /* polling mode */
        else if (info->check_mode == EDMA_CHECK_MODE_POLLING)
        {
            reg = edma_read(info->reg_base, DMA_READ_INT_MASK_REG);
            /* mask the interrupt */
            reg |= (DMA_DONE_INT(info->channel) | DMA_ABORT_INT(info->channel));
            config.interrupt_mask = reg;

            if (info->transfer_mode == EDMA_TRANSFER_MODE_NORMAL)
            {
                /* close local cpu interrupt */
                config.control1 = DMA_LIE | DMA_TD;
            }
            else if (info->transfer_mode == EDMA_TRANSFER_MODE_LINK)
            {
                /* LL mode */
                reg = edma_read(info->reg_base, DMA_READ_LINKED_LIST_ERR_EN_REG);
                reg |= (DMA_LLLAIE(info->channel));
                config.link_err_en = reg;
                /* open local cpu interrupt */
                config.control1=DMA_TD|DMA_LLE|DMA_CCS;
            }
            else
            {
                printk("invalid transfer mode\n");
                return EDMA_ERROR;
            }
        }
        else
        {
            printk("invalid check mode\n");
            return EDMA_ERROR;
        }
        config.control2 = 0;
    }
    /* cpu config remote dma */
    else if (info->cpu_mode == EDMA_CPU_MODE_REMOTE)
    {
        /* interrupt mode */
        if (info->check_mode == EDMA_CHECK_MODE_INTERRUPT)
        {
            config.done_imwr_addr_low = QWORD_LO(info->msi_addr);
            config.done_imwr_addr_high = QWORD_HI(info->msi_addr);
            config.abort_imwr_addr_low = QWORD_LO(info->msi_addr);
            config.abort_imwr_addr_high = QWORD_HI(info->msi_addr);
            config.imwr_data = info->msi_data;
            reg = edma_read(info->reg_base, DMA_READ_INT_MASK_REG);
            /* unmask the interrupt */
            reg &= (~(DMA_DONE_INT(info->channel) | DMA_ABORT_INT(info->channel)));
            config.interrupt_mask = reg;

            if (info->transfer_mode == EDMA_TRANSFER_MODE_NORMAL)
            {
                /* open remote cpu interrupt */
                config.control1 = DMA_RIE | DMA_TD | DMA_LIE;
            }
            else if (info->transfer_mode == EDMA_TRANSFER_MODE_LINK)
            {
                /* LL mode */
                reg = edma_read(info->reg_base, DMA_READ_LINKED_LIST_ERR_EN_REG);
                reg |= (DMA_LLRAIE(info->channel));
                config.link_err_en = reg;
                /* open remote cpu interrupt */
                config.control1=DMA_TD|DMA_RIE|DMA_LLE|DMA_CCS;
            }
            else
            {
                printk("invalid transfer mode\n");
                return EDMA_ERROR;
            }
        }
        /* polling mode */
        else if (info->check_mode == EDMA_CHECK_MODE_POLLING)
        {
            config.done_imwr_addr_low = 0;
            config.done_imwr_addr_high = 0;
            config.abort_imwr_addr_low = 0;
            config.abort_imwr_addr_high = 0;
            config.imwr_data = 0;
            reg = edma_read(info->reg_base, DMA_READ_INT_MASK_REG);
            /* mask the interrupt */
            reg |=(DMA_DONE_INT(info->channel) | DMA_ABORT_INT(info->channel));
            config.interrupt_mask = reg;

            if (info->transfer_mode == EDMA_TRANSFER_MODE_NORMAL)
            {
                /* close local cpu interrupt */
                config.control1 = DMA_LIE | DMA_TD;
            }
            else if (info->transfer_mode == EDMA_TRANSFER_MODE_LINK)
            {
                /* LL mode */
                reg = edma_read(info->reg_base, DMA_READ_LINKED_LIST_ERR_EN_REG);
                reg |= (DMA_LLRAIE(info->channel));
                config.link_err_en = reg;
                /* close local cpu interrupt */
                config.control1=DMA_TD|DMA_LLE|DMA_CCS;
            }
            else
            {
                printk("invalid transfer mode\n");
                return EDMA_ERROR;
            }
        }
        else
        {
            printk("invalid check mode\n");
            return EDMA_ERROR;
        }
        config.control2 = 0;
    }
    else
    {
        printk("invalid cpu mode\n");
        return EDMA_ERROR;
    }
    /* config the write channel */
    if (edma_dev_read_channel_config(&config) == EDMA_ERROR)
    {
        return EDMA_ERROR;
    }
    /* open the door bell, start transfer */
    edma_dev_transfer_start(info->reg_base, info->channel, ljeDMA_TX_DIRECTION_MEM_TO_DEV);

    return EDMA_OK;
}

static void edma_dev_fill_link_table(
    ljsDMA_CHANNEL *chan,
    edma_descriptor *desc,
    struct scatterlist *sgl,
    size_t sg_len)
{
    edma_channel *edma_chan = to_edma_chan(chan);
    edma_device *edma_dev = to_edma_dev(chan->device);
    edma_desc_ll *link_table = ljvNULL;
    edma_desc_llp *edma_llp = ljvNULL;
    struct scatterlist *sg;
    dma_addr_t addr, src = 0, dst = 0;
    size_t avail, total_len = 0;
    uint32_t ctl = 0;
    int i = 0, i_v = 0;

    if (edma_chan->direction == ljeDMA_TX_DIRECTION_MEM_TO_DEV)
    {
        desc->lle = (uint64_t)LINKTABLE_OFFSET(
                        edma_dev->plink_table_base, edma_chan->chan.id);

        link_table = (edma_desc_ll*)LINKTABLE_OFFSET(
                        edma_dev->vlink_table_base, edma_chan->chan.id);
    }
    else if (edma_chan->direction == ljeDMA_TX_DIRECTION_DEV_TO_MEM)
    {
        desc->lle = (uint64_t)LINKTABLE_OFFSET(
                        edma_dev->plink_table_base,
                        edma_dev->rchan_count + edma_chan->chan.id);

        link_table = (edma_desc_ll*)LINKTABLE_OFFSET(
                        edma_dev->vlink_table_base,
                        edma_dev->rchan_count + edma_chan->chan.id);
    }

    src = edma_chan->sconfig.srcAddr;
    dst = edma_chan->sconfig.dstAddr;
    ctl = DMA_CB;
    for_each_sg(sgl, sg, sg_len, i)
    {
        ljmkASSERT(i * ljmSIZEOF(edma_desc_ll) < EDMA_LT_SIZE);

        avail = sg_dma_len(sg);
        addr = sg_dma_address(sg);

        link_table[i].control = ctl;
        link_table[i].size = avail;

        if (edma_chan->direction == ljeDMA_TX_DIRECTION_MEM_TO_DEV)
        {
            /* mem -> dev */
            link_table[i].sar_low = QWORD_LO(addr);
            link_table[i].sar_high = QWORD_HI(addr);
            link_table[i].dar_low = QWORD_LO(dst);
            link_table[i].dar_high = QWORD_HI(dst);
            dst += avail;
        }
        else
        {
            /* dev -> mem */
            link_table[i].sar_low = QWORD_LO(src);
            link_table[i].sar_high = QWORD_HI(src);
            link_table[i].dar_low = QWORD_LO(addr);
            link_table[i].dar_high = QWORD_HI(addr);
            src += avail;
        }
        total_len += avail;
    }

    if (edma_dev->interrupt_enable)
    {
        ctl |= (DMA_LIE | DMA_RIE);
        link_table[sg_len - 1].control = ctl;
    }
    else
    {
        ctl |= DMA_LIE;
        link_table[sg_len - 1].control = ctl;
    }

    edma_llp = (edma_desc_llp*)(&link_table[sg_len]);
    edma_llp->control = DMA_LLP | DMA_TCB;
    edma_llp->reserved = 0;
    edma_llp->llp_high = QWORD_HI(desc->lle);
    edma_llp->llp_low = QWORD_LO(desc->lle);

    /* insure the config has written into link table */
    //link_table[sg_len + 1].control = 0x1;
    //while(!link_table[sg_len + 1].control);
    //memset(&link_table[sg_len + 1], 0, sizeof(link_table[sg_len]));
    //while(link_table[sg_len + 1].control);

    desc->total_len = total_len;
    desc->residue = total_len;

#if EDMA_DEBUG
    for (i = 0; i < sg_len; i++)
    {
        printk(" fill link table [%d]@0x%08llx:\n", i, desc->lle + sizeof(edma_desc_ll)*i);
        printk("    control = 0x%08x\n", link_table[i].control);
        printk("    size    = 0x%08x\n", link_table[i].size);
        printk("    sar_lo  = 0x%08x\n", link_table[i].sar_low);
        printk("    sar_hi  = 0x%08x\n", link_table[i].sar_high);
        printk("    dar_lo  = 0x%08x\n", link_table[i].dar_low);
        printk("    dar_hi  = 0x%08x\n", link_table[i].dar_high);
    }
    printk("    llp_control 	= 0x%08x\n", link_table[i].control);
    printk("    llp_reserve    	= 0x%08x\n", link_table[i].size);
    printk("    llp_high  		= 0x%08x\n", link_table[i].sar_low);
    printk("    llp_low  		= 0x%08x\n", link_table[i].sar_high);
#endif

#if (defined CONFIG_LOONGARCH ||defined CONFIG_SW64)
    udelay(delay_time);
#endif
    if (edma_chan->direction == ljeDMA_TX_DIRECTION_MEM_TO_DEV)
    {
    dst -= avail;
    while((link_table[sg_len-1].size != avail) || (link_table[sg_len-1].sar_high != QWORD_HI(addr)) || (link_table[sg_len-1].dar_high != QWORD_HI(dst)) )
     {
#if 0
          printk("wait DDR SG0    control = 0x%08x\n", link_table[sg_len-1].control);
          printk("    size    = 0x%08x\n", link_table[sg_len-1].size);
          printk("    sar_lo  = 0x%08x\n", link_table[sg_len-1].sar_low);
          printk("    sar_hi  = 0x%08x\n", link_table[sg_len-1].sar_high);
          printk("    dar_lo  = 0x%08x\n", link_table[sg_len-1].dar_low);
          printk("    dar_hi  = 0x%08x\n", link_table[sg_len-1].dar_high);
#endif
          printk("---DDR SG0 delay time: %dus-----\n", delay_time);
          udelay(delay_time);
          delay_time *= 2;
          if(i_v++ == 32)
          {
              printk("------------DDR SG0 wrong 32 times!!!--------------\n");
              break;
          }

          src = edma_chan->sconfig.srcAddr;
          dst = edma_chan->sconfig.dstAddr;
          ctl = DMA_CB;
          for_each_sg(sgl, sg, sg_len, i)
          {
              ljmkASSERT(i * ljmSIZEOF(edma_desc_ll) < EDMA_LT_SIZE);

              avail = sg_dma_len(sg);
              addr = sg_dma_address(sg);

              link_table[i].control = ctl;
              link_table[i].size = avail;

              if (edma_chan->direction == ljeDMA_TX_DIRECTION_MEM_TO_DEV)
              {
                  /* mem -> dev */
                  link_table[i].sar_low = QWORD_LO(addr);
                  link_table[i].sar_high = QWORD_HI(addr);
                  link_table[i].dar_low = QWORD_LO(dst);
                  link_table[i].dar_high = QWORD_HI(dst);
                  dst += avail;
              }
              else
              {
                  /* dev -> mem */
                  link_table[i].sar_low = QWORD_LO(src);
                  link_table[i].sar_high = QWORD_HI(src);
                  link_table[i].dar_low = QWORD_LO(addr);
                  link_table[i].dar_high = QWORD_HI(addr);
                  src += avail;
              }
              total_len += avail;
          }

          if (edma_dev->interrupt_enable)
          {
              ctl |= (DMA_LIE | DMA_RIE);
              link_table[sg_len - 1].control = ctl;
          }
          else
          {
              ctl |= DMA_LIE;
              link_table[sg_len - 1].control = ctl;
          }

          edma_llp = (edma_desc_llp*)(&link_table[sg_len]);
          edma_llp->control = DMA_LLP | DMA_TCB;
          edma_llp->reserved = 0;
          edma_llp->llp_high = QWORD_HI(desc->lle);
          edma_llp->llp_low = QWORD_LO(desc->lle);

          dst -= avail;
     }
    }else{
    src -= avail;
    while((link_table[sg_len-1].size != avail) || (link_table[sg_len-1].sar_high != QWORD_HI(src)) || (link_table[sg_len-1].dar_high != QWORD_HI(addr)) )
    {
#if 0  /* LOONGARCH printk too much */
         printk("wait DDR SG1    control = 0x%08x\n", link_table[sg_len-1].control);
         printk("    size    = 0x%08x\n", link_table[sg_len-1].size);
         printk("    sar_lo  = 0x%08x\n", link_table[sg_len-1].sar_low);
         printk("    sar_hi  = 0x%08x\n", link_table[sg_len-1].sar_high);
         printk("    dar_lo  = 0x%08x\n", link_table[sg_len-1].dar_low);
         printk("    dar_hi  = 0x%08x\n", link_table[sg_len-1].dar_high);
#endif
         printk("---DDR SG1 delay time: %dus-----\n", delay_time);
         udelay(delay_time);
         delay_time *= 2;
         if(i_v++ == 32)
         {
             printk("------------DDR SG1 wrong 32 times!!!--------------\n");
             break;
         }

         src = edma_chan->sconfig.srcAddr;
         dst = edma_chan->sconfig.dstAddr;
         ctl = DMA_CB;
         for_each_sg(sgl, sg, sg_len, i)
         {
             ljmkASSERT(i * ljmSIZEOF(edma_desc_ll) < EDMA_LT_SIZE);

             avail = sg_dma_len(sg);
             addr = sg_dma_address(sg);

             link_table[i].control = ctl;
             link_table[i].size = avail;

             if (edma_chan->direction == ljeDMA_TX_DIRECTION_MEM_TO_DEV)
             {
                 /* mem -> dev */
                 link_table[i].sar_low = QWORD_LO(addr);
                 link_table[i].sar_high = QWORD_HI(addr);
                 link_table[i].dar_low = QWORD_LO(dst);
                 link_table[i].dar_high = QWORD_HI(dst);
                 dst += avail;
             }
             else
             {
                 /* dev -> mem */
                 link_table[i].sar_low = QWORD_LO(src);
                 link_table[i].sar_high = QWORD_HI(src);
                 link_table[i].dar_low = QWORD_LO(addr);
                 link_table[i].dar_high = QWORD_HI(addr);
                 src += avail;
             }
             total_len += avail;
         }

         if (edma_dev->interrupt_enable)
         {
             ctl |= (DMA_LIE | DMA_RIE);
             link_table[sg_len - 1].control = ctl;
         }
         else
         {
             ctl |= DMA_LIE;
             link_table[sg_len - 1].control = ctl;
         }

         edma_llp = (edma_desc_llp*)(&link_table[sg_len]);
         edma_llp->control = DMA_LLP | DMA_TCB;
         edma_llp->reserved = 0;
         edma_llp->llp_high = QWORD_HI(desc->lle);
         edma_llp->llp_low = QWORD_LO(desc->lle);

         src -= avail;
    }
    }

}

static ljeSTATUS edma_dev_channels_init(edma_device *edma_dev)
{
    int i;
#if DMA_NEW
    uint32_t reg;
#endif

    if (!edma_dev)
    {
        return ljvSTATUS_INVALID_ARGUMENT;
    }

    /* get read/write channels count from reg */
#if DMA_NEW
    if (edma_dev->single_chan) {
        edma_dev->rchan_count = 1;
        edma_dev->wchan_count = 1;
    } else {
        reg = edma_read(edma_dev->reg_base, DMA_CHANNEL_CTRL_REG);
        edma_dev->rchan_count = (reg >> 16) & 0xf;
        edma_dev->wchan_count = reg & 0xf;
    }
#else
    edma_dev->rchan_count = 1;
    edma_dev->wchan_count = 1;
#endif
    edma_dev->wchan = kcalloc(edma_dev->wchan_count, sizeof(edma_channel), GFP_KERNEL);
    if (!edma_dev->wchan)
    {
        return ljvSTATUS_OUT_OF_MEMORY;
    }

    edma_dev->rchan = kcalloc(edma_dev->rchan_count, sizeof(edma_channel), GFP_KERNEL);
    if (!edma_dev->rchan)
    {
        kfree(edma_dev->wchan);
        return ljvSTATUS_OUT_OF_MEMORY;
    }

    /* write channels */
    for (i = 0; i < edma_dev->wchan_count; i++)
    {
        edma_channel *chan = &edma_dev->wchan[i];

        chan->chan.device = &edma_dev->slave_wdma;
        ljkDMA_IdentifierInit(&chan->chan);

        list_add_tail(&chan->chan.devNode, &edma_dev->slave_wdma.chanHead);

        spin_lock_init(&chan->lock);
        INIT_LIST_HEAD(&chan->active_list);
        INIT_LIST_HEAD(&chan->pending_list);

        chan->reg_base = edma_dev->reg_base;
        chan->status = ljeDMA_STATUS_COMPLETE;
        chan->direction = ljeDMA_TX_DIRECTION_DEV_TO_MEM;

	chan->desc = kzalloc(sizeof(edma_descriptor), GFP_NOWAIT);
    }

    /* read channels */
    for (i = 0; i < edma_dev->rchan_count; i++)
    {
        edma_channel *chan = &edma_dev->rchan[i];

        chan->chan.device = &edma_dev->slave_rdma;
        ljkDMA_IdentifierInit(&chan->chan);

        list_add_tail(&chan->chan.devNode, &edma_dev->slave_rdma.chanHead);

        spin_lock_init(&chan->lock);
        INIT_LIST_HEAD(&chan->active_list);
        INIT_LIST_HEAD(&chan->pending_list);

        chan->reg_base = edma_dev->reg_base;
        chan->status = ljeDMA_STATUS_COMPLETE;
        chan->direction = ljeDMA_TX_DIRECTION_MEM_TO_DEV;

	chan->desc = kzalloc(sizeof(edma_descriptor), GFP_NOWAIT);
    }

    return ljvSTATUS_OK;
}

static ljeSTATUS edma_dev_msi_addr_and_data(edma_device *edma_dev)
{
    struct pci_dev *pdev = edma_dev->pdev;
    uint8_t cap_off;
    uint32_t addr_hi, addr_lo;
    uint16_t flags;
    uint64_t msi_addr;
    uint32_t msi_data;

    if (!(pdev->msi_cap && pdev->msi_enabled))
    {
        return ljvSTATUS_NOT_SUPPORTED;
    }
    cap_off = pdev->msi_cap + PCI_MSI_FLAGS;
    pci_read_config_word(pdev, cap_off, &flags);

    if (!(flags & PCI_MSI_FLAGS_ENABLE)) {
        return ljvSTATUS_NOT_SUPPORTED;
    }
    cap_off = pdev->msi_cap + PCI_MSI_ADDRESS_LO;
    pci_read_config_dword(pdev, cap_off, &addr_lo);

    if (flags & PCI_MSI_FLAGS_64BIT) {
        cap_off = pdev->msi_cap + PCI_MSI_ADDRESS_HI;
        pci_read_config_dword(pdev, cap_off, &addr_hi);
        cap_off = pdev->msi_cap + PCI_MSI_DATA_64;
    }
    else
    {
        addr_hi = 0;
        cap_off = pdev->msi_cap + PCI_MSI_DATA_32;
    }

    msi_addr = addr_hi;
    msi_addr <<= 32;
    msi_addr |= addr_lo;

    pci_read_config_dword(pdev, cap_off, &msi_data);
    msi_data &= 0xffff;

    edma_dev->msi_addr = msi_addr;
    edma_dev->msi_data = msi_data;

    return ljvSTATUS_OK;
}

static int threadEdma(void *ctxt)
{
    edma_device *edma_dev = (edma_device *)ctxt;
    uint32_t wr_status, rd_status;
    int i;

  for (;;)
  {
        int down;

        down = down_interruptible(&sema_edma);
        if (down && down != -EINTR)
        {
            return down;
        }

        if (unlikely(EdmaKillThread))
        {
            /* The daemon exits. */
            while (!kthread_should_stop())
            {
                ljkOS_Delay(NULL, 1);
            }

            return 0;
        }


    /* get rd and wr int status */
    wr_status = edma_read(edma_dev->reg_base, DMA_WRITE_INT_STATUS_REG);
    rd_status = edma_read(edma_dev->reg_base, DMA_READ_INT_STATUS_REG);

    /* it is read channel int */
    if (rd_status)
    {
        for (i = 0; i < edma_dev->rchan_count; i++)
        {
            if (rd_status & DMA_ABORT_INT(i))
            {
                edma_dev_handle_error(&edma_dev->rchan[i]);
            }
            if(rd_status & DMA_DONE_INT(i))
            {
//              printk("\n--------edma_dev_tasklet----edma_dev_handle_done_rd---\n");
                edma_dev_handle_done(&edma_dev->rchan[i]);
            }
        }
    }

    /* it is write channel int */
    if (wr_status)
    {
        for (i = 0; i < edma_dev->wchan_count; i++)
        {
            if (wr_status & DMA_ABORT_INT(i))
            {
                edma_dev_handle_error(&edma_dev->wchan[i]);
            }
            if(wr_status & DMA_DONE_INT(i))
            {
//              printk("\n--------edma_dev_tasklet----edma_dev_handle_done_wr---\n");
                edma_dev_handle_done(&edma_dev->wchan[i]);
            }
        }
    }

    /* unmask all read interrupt */
    //edma_write(edma_dev->reg_base, DMA_READ_INT_MASK_REG, 0);
    /* unmask all write interrupt */
    //edma_write(edma_dev->reg_base, DMA_WRITE_INT_MASK_REG, 0);
    /*DMA_SINGLE_CHANNEL ok*/
    LJM_MISC_W32_BASE(edma_dev->reg_base_2, MISC_INTR_EN,
                      LJM_MISC_R32_BASE(edma_dev->reg_base_2, MISC_INTR_EN) | m_INTR_EN_EDMA);
    }
}

/**
 * edma_dev_interrupt_handler - dma interrupt handler.
 *
 * @irq: irq number
 * @dev_id: driver private at int registry
 *
 * Return: always 0.
 *
 */
static irqreturn_t edma_dev_interrupt_handler(int irq, void *dev_id)
{
    edma_device *edma_dev = dev_id;
    uint32_t wr_status, rd_status;
    //unsigned long flags;

    //spin_lock_irqsave(&edma_dev->irq_lock, flags);
    spin_lock(&edma_dev->irq_lock);

    /* read int status */
    wr_status = edma_read(edma_dev->reg_base, DMA_WRITE_INT_STATUS_REG);
    rd_status = edma_read(edma_dev->reg_base, DMA_READ_INT_STATUS_REG);

    if (!(wr_status | rd_status))
    {
        //spin_unlock_irqrestore(&edma_dev->irq_lock, flags);
        spin_unlock(&edma_dev->irq_lock);
        return IRQ_NONE;
    }

    /* mask all read interrupt */
    //edma_write(edma_dev->reg_base, DMA_READ_INT_MASK_REG, 0x00FF00FF);
    /* mask all write interrupt */
    //edma_write(edma_dev->reg_base, DMA_WRITE_INT_MASK_REG, 0x00FF00FF);
    /*DMA_SINGLE_CHANNEL ok*/
    LJM_MISC_W32_BASE(edma_dev->reg_base_2, MISC_INTR_EN,
                      LJM_MISC_R32_BASE(edma_dev->reg_base_2, MISC_INTR_EN) & (~m_INTR_EN_EDMA));

    //spin_unlock_irqrestore(&edma_dev->irq_lock, flags);
    spin_unlock(&edma_dev->irq_lock);

    //tasklet_schedule(&edma_dev->tasklet);
    up(&sema_edma);

    return IRQ_HANDLED;
}

/**
 * edma_dev_alloc_chan_resources - init the channel resource
 *
 * @chann: dma_chann structure pointer
 *
 * Return: none
 *
 */
static ljeSTATUS edma_dev_alloc_chan_resources(ljsDMA_CHANNEL *chan)
{
    edma_channel *edma_chan = to_edma_chan(chan);
    edma_device *edma_dev = to_edma_dev(chan->device);
    unsigned long flags;

    if (edma_dev_get_channel_status(
            edma_dev->reg_base,
            edma_chan->chan.id,
            edma_chan->direction) == DMA_CS_RUNNING)
    {
        printk("dma %s chan %d is busy.\n", chan->device->name, edma_chan->chan.id);
        return ljvSTATUS_DEVICE;
    }

    ljkDMA_IdentifierInit(chan);

    spin_lock_irqsave(&edma_chan->lock, flags);
    if (edma_chan->direction == ljeDMA_TX_DIRECTION_DEV_TO_MEM)
    {
        if (!edma_dev->wchan_in_use)
        {
            edma_dev_write_engine_on(edma_dev->reg_base);
        }
        edma_dev->wchan_in_use |= (0x1 << edma_chan->chan.id);
    }
    else if (edma_chan->direction == ljeDMA_TX_DIRECTION_MEM_TO_DEV)
    {
        if(!edma_dev->rchan_in_use)
        {
            edma_dev_read_engine_on(edma_dev->reg_base);
        }
        edma_dev->rchan_in_use |= (0x1 << edma_chan->chan.id);
    }
    spin_unlock_irqrestore(&edma_chan->lock, flags);

    return ljvSTATUS_OK;
}

/**
 * edma_dev_free_chan_resources - just disable channel irq, the channel resource
 * will destoried in complete functions.
 *
 * @chan: ljsDMA_CHANNEL structure pointer
 *
 * Return: none
 *
 */
static void edma_dev_free_chan_resources(ljsDMA_CHANNEL *chan)
{
    edma_channel *edma_chan = to_edma_chan(chan);
    edma_device *edma_dev = to_edma_dev(chan->device);
    unsigned long flags;
    uint32_t reg;

    BUG_ON(!list_empty(&edma_chan->active_list));
    BUG_ON(!list_empty(&edma_chan->pending_list));
    BUG_ON(edma_chan->descs_allocated);

    spin_lock_irqsave(&edma_chan->lock, flags);
    if (edma_chan->direction == ljeDMA_TX_DIRECTION_MEM_TO_DEV)
    {
        /* disable channel interrupt */
        reg = edma_read(edma_dev->reg_base, DMA_READ_INT_MASK_REG);
        //edma_write(edma_dev->reg_base, DMA_READ_INT_MASK_REG, reg | DMA_DONE_INT(edma_chan->chan.id));
        //edma_write(edma_dev->reg_base, DMA_READ_INT_MASK_REG, reg | DMA_ABORT_INT(edma_chan->chan.id));

        edma_dev->rchan_in_use &= ~(0x1 << edma_chan->chan.id);
        if (!edma_dev->rchan_in_use)
        {
            edma_dev_read_engine_off(edma_dev->reg_base);
        }
    }

    if (edma_chan->direction == ljeDMA_TX_DIRECTION_DEV_TO_MEM)
    {
        /* disable channel interrupt */
        reg = edma_read(edma_dev->reg_base, DMA_WRITE_INT_MASK_REG);
        //edma_write(edma_dev->reg_base, DMA_WRITE_INT_MASK_REG, reg | DMA_DONE_INT(edma_chan->chan.id));
        //edma_write(edma_dev->reg_base, DMA_WRITE_INT_MASK_REG, reg | DMA_ABORT_INT(edma_chan->chan.id));

        edma_dev->wchan_in_use &= ~(0x1 << edma_chan->chan.id);
        if (!edma_dev->wchan_in_use)
        {
            edma_dev_write_engine_off(edma_dev->reg_base);
        }
    }
    spin_unlock_irqrestore(&edma_chan->lock, flags);

}

/**
 * edma_dev_prep_slave_sg - alloc dma desctriptors for every sg node.
 *
 * @chann: dma_chann structure pointer
 * @sgl: scatterlist structure pointer
 * @sg_len: the dma data len
 * @direction: transfer direction
 * @flags: other info
 *
 * Return: ljsDMA_TX_DESCRIPTOR pointer.
 *
 */
static ljsDMA_TX_DESCRIPTOR* edma_dev_prep_slave_sg(
    ljsDMA_CHANNEL *chan,
    struct scatterlist *sgl,
    size_t sg_len,
    ljeDMA_TX_DIRECTION direction,
    unsigned int flags,
    void *context
    )
{
    edma_channel *edma_chan;
    edma_device *edma_dev;
    edma_descriptor *desc;

    if(!chan || !sgl || sg_len == 0)
    {
        return NULL;
    }

    if(sg_len > EDMA_LT_SIZE/ljmSIZEOF(edma_desc_ll))
    {
        printk("edma failed: sg_len(%ld) over the max size(%ld).\n", sg_len, EDMA_LT_SIZE/ljmSIZEOF(edma_desc_ll));
        return NULL;
    }

    if (!ljkDMAENGINE_IsSlaveDirection(direction))
    {
        printk("invalid dma direction.\n");
        return NULL;
    }

    edma_chan = to_edma_chan(chan);
    edma_dev = to_edma_dev(chan->device);

    if (edma_chan->direction != direction)
    {
        printk("chan %d do not support direction %d\n", edma_chan->chan.id , direction);
        return NULL;
    }

    desc = edma_descriptor_alloc(edma_chan);
    if (!desc)
    {
        printk("fail to alloc descriptor.\n");
        return NULL;
    }

    desc->cpu_mode = edma_dev->interrupt_enable ? EDMA_CPU_MODE_REMOTE : EDMA_CPU_MODE_LOCAL;
    desc->check_mode = edma_dev->interrupt_enable ? EDMA_CHECK_MODE_INTERRUPT : EDMA_CHECK_MODE_POLLING;
    desc->transfer_mode = EDMA_TRANSFER_MODE_LINK;

    desc->src_addr = 0;
    desc->dst_addr = 0;
    desc->total_len = 0;
    desc->residue = 0;
    desc->timeout = 5000;
    desc->is_done = false;

    edma_dev_fill_link_table(chan, desc, sgl, sg_len);

    return &desc->txd;
}

/**
 * edma_dev_prep_memcpy - alloc dma desctriptors for normal mode.
 *
 * @chan: ljsDMA_CHANNEL structure pointer
 * @Src: the src bus address
 * @Dst: the dst bus address
 * @Len: the dma data len
 * @flags: other info
 *
 * Return: ljsDMA_TX_DESCRIPTOR pointer.
 *
 */
static ljsDMA_TX_DESCRIPTOR* edma_dev_prep_dma_memcpy(
    ljsDMA_CHANNEL *Chan,
    dma_addr_t Src,
    dma_addr_t Dst,
    ljtSIZE_T Len,
    ljtUINT32 Flag
    )
{
    edma_channel *edma_chan;
    edma_device *edma_dev;
    edma_descriptor *desc;
    edma_desc_ll *link_table = ljvNULL;
    //edma_desc_llp *edma_llp = ljvNULL;
    uint32_t ctl;
    uint32_t link_t[16];
    uint32_t i_v = 0;

    if(!Chan || Len == 0)
    {
        return NULL;
    }

    edma_chan = to_edma_chan(Chan);
    edma_dev = to_edma_dev(Chan->device);

    desc = edma_descriptor_alloc(edma_chan);
    if (!desc)
    {
        printk("fail to alloc descriptor.\n");
        return NULL;
    }

    desc->cpu_mode = edma_dev->interrupt_enable ? EDMA_CPU_MODE_REMOTE : EDMA_CPU_MODE_LOCAL;
    desc->check_mode = edma_dev->interrupt_enable ? EDMA_CHECK_MODE_INTERRUPT : EDMA_CHECK_MODE_POLLING;
    desc->transfer_mode = EDMA_TRANSFER_MODE_LINK;

    desc->src_addr = 0;
    desc->dst_addr = 0;
    desc->total_len = 0;
    desc->residue = 0;
    desc->timeout = 5000;
    desc->is_done = false;

    /* use linked mode to replace normal mode */
    if (edma_chan->direction == ljeDMA_TX_DIRECTION_MEM_TO_DEV)
    {
        desc->lle = (uint64_t)LINKTABLE_OFFSET(
                        edma_dev->plink_table_base, edma_chan->chan.id);

        link_table = (edma_desc_ll*)LINKTABLE_OFFSET(
                        edma_dev->vlink_table_base, edma_chan->chan.id);
    }
    else if (edma_chan->direction == ljeDMA_TX_DIRECTION_DEV_TO_MEM)
    {
        desc->lle = (uint64_t)LINKTABLE_OFFSET(
                        edma_dev->plink_table_base,
                        edma_dev->rchan_count + edma_chan->chan.id);

        link_table = (edma_desc_ll*)LINKTABLE_OFFSET(
                        edma_dev->vlink_table_base,
                        edma_dev->rchan_count + edma_chan->chan.id);
    }
    else
    {
        return ljvNULL;
    }

    ctl = DMA_CB;
#if 0/* abc: 256bit align write ddr */
    link_table[0].control = ctl;
    link_table[0].size = Len;
    link_table[0].sar_low = QWORD_LO(Src);
    link_table[0].sar_high = QWORD_HI(Src);
    link_table[0].dar_low = QWORD_LO(Dst);
    link_table[0].dar_high = QWORD_HI(Dst);

    if (edma_dev->interrupt_enable)
    {
        ctl |= (DMA_LIE | DMA_RIE);
        link_table[0].control = ctl;
    }
    else
    {
        ctl |= DMA_LIE;
        link_table[0].control = ctl;
    }

    edma_llp = (edma_desc_llp*)(&link_table[1]);
    edma_llp->control = DMA_LLP | DMA_TCB;
    edma_llp->reserved = 0;
    edma_llp->llp_high = QWORD_HI(desc->lle);
    edma_llp->llp_low = QWORD_LO(desc->lle);
#else
    link_t[0] = ctl;
    link_t[1] = Len;
    link_t[2] = QWORD_LO(Src);
    link_t[3] = QWORD_HI(Src);
    link_t[4] = QWORD_LO(Dst);
    link_t[5] = QWORD_HI(Dst);

    if (edma_dev->interrupt_enable)
    {
        ctl |= (DMA_LIE | DMA_RIE);
        link_t[0] = ctl;
    }
    else
    {
        ctl |= DMA_LIE;
        link_t[0] = ctl;
    }

    link_t[6] = DMA_LLP | DMA_TCB;
    link_t[7] = 0;
    link_t[8] = QWORD_LO(desc->lle);
    link_t[9] = QWORD_HI(desc->lle);
    link_t[10] = 0;
    link_t[11] = 0;
    link_t[12] = 0;
    link_t[13] = 0;
    link_t[14] = 0;
    link_t[15] = 0;

    memcpy(&link_table[0], &link_t[0], 64);
    //memcpy((uint32_t *)(&link_table[0])+8, &link_t[8], 32);
#endif
    /* insure the config has written into link table */
    //link_table[2].control = 0x1;
    //while(!link_table[2].control);
    //memset(&link_table[2], 0, sizeof(link_table[2]));
    //while(link_table[2].control);

    desc->total_len = Len;
    desc->residue = Len;

#if EDMA_DEBUG
    printk("edma_dev_prep_dma_memcpy : link table [%d]@0x%08llx:\n", 0, desc->lle);
    printk("    control = 0x%08x\n", link_table[0].control);
    printk("    size    = 0x%08x\n", link_table[0].size);
    printk("    sar_lo  = 0x%08x\n", link_table[0].sar_low);
    printk("    sar_hi  = 0x%08x\n", link_table[0].sar_high);
    printk("    dar_lo  = 0x%08x\n", link_table[0].dar_low);
    printk("    dar_hi  = 0x%08x\n", link_table[0].dar_high);

    printk("    llp_control = 0x%08x\n", link_table[1].control);
    printk("    llp_reserved    = 0x%08x\n", link_table[1].size);
    printk("    llp_high  = 0x%08x\n", link_table[1].sar_low);
    printk("    llp_low  = 0x%08x\n", link_table[1].sar_high);
#endif

#if (defined CONFIG_LOONGARCH ||defined CONFIG_SW64)
    udelay(delay_time);
#endif
   while((link_table[0].size != Len) || (link_table[0].sar_high != QWORD_HI(Src)) || (link_table[0].dar_high != QWORD_HI(Dst)) )
   {
    ctl = DMA_CB;
    link_table[0].control = ctl;
    link_table[0].size = Len;
    link_table[0].sar_low = QWORD_LO(Src);
    link_table[0].sar_high = QWORD_HI(Src);
    link_table[0].dar_low = QWORD_LO(Dst);
    link_table[0].dar_high = QWORD_HI(Dst);

    if (edma_dev->interrupt_enable)
    {
        ctl |= (DMA_LIE | DMA_RIE);
        link_table[0].control = ctl;
    }
    else
    {
        ctl |= DMA_LIE;
        link_table[0].control = ctl;
    }
    printk("---DDR 0 delay time: %dus-----\n", delay_time);
    udelay(delay_time);
    delay_time *= 2;
#if 0  /* LOONGARCH printk too much */
     printk(" WAIT DDR 0   control = 0x%08x\n", link_table[0].control);
     printk("    size    = 0x%08x\n", link_table[0].size);
     printk("    sar_lo  = 0x%08x\n", link_table[0].sar_low);
     printk("    sar_hi  = 0x%08x\n", link_table[0].sar_high);
     printk("    dar_lo  = 0x%08x\n", link_table[0].dar_low);
     printk("    dar_hi  = 0x%08x\n", link_table[0].dar_high);
#endif
    if(i_v++ == 32)
    {
        printk("------------DDR 0 wrong 32 times!!!--------------\n");
        break;
    }
   }


    return &desc->txd;
}

/**
 * edma_dev_handle_error - dma abort interrupt handler
 *
 * @edma_chann: pcie_edma_chann structure pointer
 *
 * Return: none
 *
 */
#if 0
static void edma_dev_handle_error(edma_channel *edma_chan)
{
    edma_descriptor *bad_desc;
    unsigned long flags;

    spin_lock_irqsave(&edma_chan->lock, flags);
    /* delete the bad desc */
    bad_desc = edma_get_first_active(edma_chan);
    list_del_init(&bad_desc->desc_node);

    /* clear abort bit */
    if (edma_chan->direction == ljeDMA_TX_DIRECTION_MEM_TO_DEV)
    {
        edma_write(edma_chan->reg_base, DMA_READ_INT_CLEAR_REG, DMA_ABORT_INT(edma_chan->chan.id));
        edma_dev_read_engine_off(edma_chan->reg_base);
        edma_dev_read_engine_on(edma_chan->reg_base);
    }
    else if (edma_chan->direction == ljeDMA_TX_DIRECTION_DEV_TO_MEM)
    {
        edma_write(edma_chan->reg_base, DMA_WRITE_INT_CLEAR_REG, DMA_ABORT_INT(edma_chan->chan.id));
        edma_dev_write_engine_off(edma_chan->reg_base);
        edma_dev_write_engine_on(edma_chan->reg_base);
    }

    /* going to do the work */
    edma_do_first_queued(edma_chan);

    spin_unlock_irqrestore(&edma_chan->lock, flags);

    printk("edma channel %d identifier %d error from 0x%08llx to 0x%08llx\n",
        edma_chan->chan.id, bad_desc->txd.identifier, bad_desc->src_addr, bad_desc->src_addr);

    /* free resource about bad desc */
    edma_descriptor_complete(edma_chan, bad_desc, true);
}
#else
static void edma_dev_handle_error(edma_channel *edma_chan)
{
//    edma_descriptor *bad_desc;
    unsigned long flags;

    printk(KERN_ERR "edma(%px): meets with edma_dev_handle_error\n", edma_chan);

    spin_lock_irqsave(&edma_chan->lock, flags);
    /* delete the bad desc */
//    bad_desc = edma_get_first_active(edma_chan);
//    list_del_init(&bad_desc->desc_node);

    /* clear abort bit */
    if (edma_chan->direction == ljeDMA_TX_DIRECTION_MEM_TO_DEV)
    {
        edma_write(edma_chan->reg_base, DMA_READ_INT_CLEAR_REG, DMA_ABORT_INT(edma_chan->chan.id));
        edma_dev_read_engine_off(edma_chan->reg_base);
        edma_dev_read_engine_on(edma_chan->reg_base);
    }
    else if (edma_chan->direction == ljeDMA_TX_DIRECTION_DEV_TO_MEM)
    {
        edma_write(edma_chan->reg_base, DMA_WRITE_INT_CLEAR_REG, DMA_ABORT_INT(edma_chan->chan.id));
        edma_dev_write_engine_off(edma_chan->reg_base);
        edma_dev_write_engine_on(edma_chan->reg_base);
    }

    /* going to do the work */
    edma_do_first_queued(edma_chan);

    spin_unlock_irqrestore(&edma_chan->lock, flags);

//    printk("edma channel %d identifier %d error from 0x%08llx to 0x%08llx\n",
//        edma_chan->chan.id, bad_desc->txd.identifier, bad_desc->src_addr, bad_desc->src_addr);

//    /* free resource about bad desc */
//    edma_descriptor_complete(edma_chan, bad_desc, true);
}
#endif

/**
 * edma_dev_handle_done - dma done interrupt handler
 *
 * @edma_chan: edma_channel structure pointer
 *
 * Return: none
 *
 */
static void edma_dev_handle_done(edma_channel *edma_chan)
{
    edma_descriptor *desc;
    struct list_head *active;
    unsigned long flags;

    spin_lock_irqsave(&edma_chan->lock, flags);
    /* clear done bit */
    if (edma_chan->direction == ljeDMA_TX_DIRECTION_MEM_TO_DEV)
    {
        edma_write(edma_chan->reg_base, DMA_READ_INT_CLEAR_REG, DMA_DONE_INT(edma_chan->chan.id));
    }
    else if (edma_chan->direction == ljeDMA_TX_DIRECTION_DEV_TO_MEM)
    {
        edma_write(edma_chan->reg_base, DMA_WRITE_INT_CLEAR_REG, DMA_DONE_INT(edma_chan->chan.id));
    }

    /* get current active */
    desc = edma_get_first_active(edma_chan);
    active = edma_chan->current_active;
    if (edma_desc_entry(active) != desc)
    {
        printk("edma: it shouldn't be here, interrupt lost?\n");
    }

    spin_unlock_irqrestore(&edma_chan->lock, flags);
    /* desc has done */
    edma_complete_all(edma_chan);
}

#if 0
/**
 * pcie_edma_tasklet - dma interrupt tasklet.
 *
 * @irq: irq number
 * @dev_id: driver private at int registry
 *
 * Return: always 0.
 *
 */
static void edma_dev_tasklet(unsigned long data)
{
    edma_device *edma_dev = (edma_device *)data;
    uint32_t wr_status, rd_status;
    int i;
    //unsigned int *tmp_r;
    //unsigned int tmp;

    /* get rd and wr int status */
    wr_status = edma_read(edma_dev->reg_base, DMA_WRITE_INT_STATUS_REG);
    rd_status = edma_read(edma_dev->reg_base, DMA_READ_INT_STATUS_REG);

    /* it is read channel int */
    if (rd_status)
    {
        for (i = 0; i < edma_dev->rchan_count; i++)
        {
            if (rd_status & DMA_ABORT_INT(i))
            {
                edma_dev_handle_error(&edma_dev->rchan[i]);
            }
            if(rd_status & DMA_DONE_INT(i))
            {
                edma_dev_handle_done(&edma_dev->rchan[i]);
            }
        }
    }

    /* it is write channel int */
    if (wr_status)
    {
        for (i = 0; i < edma_dev->wchan_count; i++)
        {
            if (wr_status & DMA_ABORT_INT(i))
            {
                edma_dev_handle_error(&edma_dev->wchan[i]);
            }
            if(wr_status & DMA_DONE_INT(i))
            {
                edma_dev_handle_done(&edma_dev->wchan[i]);
            }
        }
    }

    /* unmask all read interrupt */
    edma_write(edma_dev->reg_base, DMA_READ_INT_MASK_REG, 0);
    /* unmask all write interrupt */
    edma_write(edma_dev->reg_base, DMA_WRITE_INT_MASK_REG, 0);
    LJM_MISC_W32_BASE(edma_dev->reg_base_2, MISC_INTR_EN,
                 LJM_MISC_R32_BASE(edma_dev->reg_base_2, MISC_INTR_EN) | (m_INTR_EN_EDMAW_C0 | m_INTR_EN_EDMAR_C0));

    /* insure the master write on pcie has done */
    //tmp_r = (unsigned int *)(edma_dev->vlink_table_base + 0x80000);
    //tmp = ~tmp_r[0];
    //memset(&tmp_r[0], tmp, sizeof(unsigned int));
    //while(tmp_r[0] != tmp);
}
#endif

/**
 * edma_dev_do_transaction - issue dma into hw
 *
 * @edma_chann: pcie_edma_chann structure pointer
 * @desc: dma pcie_edma_descriptor structure pointer
 *
 * Return: none
 *
 */
static void edma_dev_do_transaction(edma_channel *edma_chan, edma_descriptor *desc)
{
    edma_device *edma_dev = to_edma_dev(edma_chan->chan.device);
    edma_transfer_info info;
    edma_status status;

    if (edma_dev_get_channel_status(
            edma_chan->reg_base,
            edma_chan->chan.id,
            edma_chan->direction) == DMA_CS_RUNNING)
    {
        printk("dma %s chan %d is busy, do transaction fail\n",
                    edma_chan->chan.device->name, edma_chan->chan.id);
        return;
    }

    info.reg_base = edma_chan->reg_base;
    info.channel = edma_chan->chan.id;
    info.cpu_mode = desc->cpu_mode;
    info.check_mode = desc->check_mode;
    info.transfer_mode = desc->transfer_mode;
    info.src_addr = desc->src_addr;
    info.dst_addr = desc->dst_addr;
    info.transfer_size = desc->total_len;
    info.lle = desc->lle;
    info.timeout = desc->timeout;
    info.msi_addr = edma_dev->msi_addr;
    info.msi_data = edma_dev->msi_data;

    if (edma_chan->direction == ljeDMA_TX_DIRECTION_DEV_TO_MEM)
    {
        /* write */
        status = edma_dev_write_transfer(&info);
    }
    else if (edma_chan->direction == ljeDMA_TX_DIRECTION_MEM_TO_DEV)
    {
        /* read */
        status = edma_dev_read_transfer(&info);
    }
}

/**
 * to_pcie_edma_descriptor - convert dma desc to edma desc
 *
 * @txd: edma_chann structure pointer
 *
 * Return: pcie_edma_descriptor pointer
 *
 */
static edma_descriptor *to_edma_descriptor(ljsDMA_TX_DESCRIPTOR *txd)
{
    return container_of(txd, edma_descriptor, txd);
}

/**
 * edma_get_first_active - get first desc on the active list of a channel
 *
 * @edma_chan: edma_chann structure pointer
 *
 * Return: pcie_edma_descriptor pointer
 *
 */
edma_descriptor *edma_get_first_active(edma_channel *edma_chan)
{
#if 0  //fq
    if (list_empty(&edma_chan->active_list))
    {
        return NULL;
    }
#endif
    return edma_desc_entry(edma_chan->active_list.next);
}

/**
 * edma_txd_submit - submit the desc chain to pending list of channel
 *
 * @txd: ljsDMA_TX_DESCRIPTOR structure pointer
 *
 * Return: ljdDMA_TX_IDENTIFIER
 *
 */
static ljdDMA_TX_IDENTIFIER edma_txd_submit(ljsDMA_TX_DESCRIPTOR *txd)
{
    edma_descriptor *edma_desc = to_edma_descriptor(txd);
    edma_channel *edma_chan = to_edma_chan(txd->chan);
    ljdDMA_TX_IDENTIFIER identifier;
    unsigned long flags;

    spin_lock_irqsave(&edma_chan->lock, flags);
    identifier = ljkDMA_IdentifierAssign(txd);
    /* add the parent descriptor into pending list */
    list_add_tail(&edma_desc->desc_node, &edma_chan->pending_list);
    spin_unlock_irqrestore(&edma_chan->lock, flags);
    return identifier;
}

/**
 * edma_descriptor_alloc - alloc channel descriptor resource
 * the function need to be called under lock
 *
 * @edma_chann: pcie_edma_chann structure pointer
 *
 * Return: pcie_edma_descriptor pointer
 *
 */
edma_descriptor *edma_descriptor_alloc(edma_channel *edma_chan)
{
    edma_descriptor *edma_desc;

    //edma_desc = kzalloc(sizeof(*edma_desc), GFP_NOWAIT);
    edma_desc = edma_chan->desc;
    if (!edma_desc)
    {
        return NULL;
    }
    edma_chan->descs_allocated++;
    ljkDMA_TxDescriptorInit(&edma_desc->txd, &edma_chan->chan);

    edma_desc->txd.ljkDMA_TxSubmit = edma_txd_submit;

    return edma_desc;
}

/**
 * edma_descriptor_free - free channel descriptor resource
 * the function need to be called under lock
 *
 * @edma_chan: pcie_edma_chann structure pointer
 * @edma_desc: desc pointer which will be freed.
 *
 * Return: none
 *
 */
void edma_descriptor_free(edma_channel *edma_chan,
    edma_descriptor *edma_desc)
{
    if (!edma_desc)
    {
        return;
    }
    list_del(&edma_desc->desc_node);
    //kfree(edma_desc);
    edma_chan->descs_allocated--;
}

/**
 * edma_descriptor_complete - free a desc resource and invoke the callback
 * to the dma client driver
 *
 * @edma_chann: pcie_edma_chann structure pointer
 * @desc: edma_descriptor pointer
 * @invoke_callback: invoke the callback or not
 *
 * Return: none
 *
 */
void edma_descriptor_complete(
    edma_channel *edma_chan,
    edma_descriptor *desc,
    bool invoke_callback
    )
{
    ljsDMA_TX_DESCRIPTOR *txd = &desc->txd;
    unsigned long flags;
    ljsDMAENGINE_DESC_CALLBACK cb;

    spin_lock_irqsave(&edma_chan->lock, flags);
    ljkDMA_IdentifierComplete(txd);
    if (invoke_callback)
    {
        ljkDMAINGINE_DescGetCallback(txd, &cb);
    }
    else
    {
        memset(&cb, 0, sizeof(cb));
    }

    edma_descriptor_free(edma_chan, desc);
    spin_unlock_irqrestore(&edma_chan->lock, flags);

    ljkDMAENGINE_DescCallbackInvoke(&cb, NULL);
}

/**
 * edma_free_descriptor_list - free all desc on the list
 * resource
 *
 * @edma_chann: edma_channel structure pointer
 * @list: list header which will free the desc
 *
 * Return: none
 *
 */
void edma_free_descriptor_list(
    edma_channel *edma_chan,
    struct list_head *list
    )
{
    edma_descriptor *desc, *tmp;

    list_for_each_entry_safe(desc, tmp, list, desc_node)
    {
        edma_descriptor_complete(edma_chan, desc, true);
    }
}

/**
 * edma_do_first_queued - queue the next pending dma task, get the first
 * dma desc list in the active list.
 *
 * @edma_chann: pcie_edma_chann structure pointer
 *
 * Return: none
 *
 */
static void edma_do_first_queued(edma_channel *edma_chan)
{
    edma_descriptor *desc;

    if (list_empty(&edma_chan->pending_list))
    {
        return;
    }

    /* move next pending to active */
    list_move_tail(edma_chan->pending_list.next, &edma_chan->active_list);

    desc = edma_get_first_active(edma_chan);
    if (!desc)
    {
        return;
    }
    edma_chan->current_active = &desc->desc_node;

    edma_dev_do_transaction(edma_chan, desc);
}

/**
 * edma_complete_all - queue next pending transaction and free completed
 * resource
 *
 * @edma_chan: pcie_edma_chann structure pointer
 *
 * Return: none
 *
 */
static void edma_complete_all(edma_channel *edma_chan)
{
    edma_status status;
    unsigned long flags;
    LIST_HEAD(list);

    spin_lock_irqsave(&edma_chan->lock, flags);

    if (edma_dev_get_channel_status(
        edma_chan->reg_base, edma_chan->chan.id, edma_chan->direction) == DMA_CS_RUNNING)
    {
        printk("chan %d is no idle\n", edma_chan->chan.id);
        status = edma_dev_transfer_stop(edma_chan->reg_base, edma_chan->chan.id, edma_chan->direction);
        if (status == EDMA_ERROR)
        {
            printk("stop chan %d fail\n", edma_chan->chan.id);
        }
    }

    if (list_empty(&edma_chan->active_list))
    {
        spin_unlock_irqrestore(&edma_chan->lock, flags);
        return;
    }

    /* remove finished transaction desc from active list */
    list_move(edma_chan->active_list.next, &list);

    edma_do_first_queued(edma_chan);

    spin_unlock_irqrestore(&edma_chan->lock, flags);

    edma_free_descriptor_list(edma_chan, &list);
}

static int edma_polling(void *ctxt)
{
    edma_device *edma_dev = (edma_device*)ctxt;

    while (1)
    {
        if (unlikely(edma_dev->kill_thread))
        {
            /* The daemon exits. */
            while (!kthread_should_stop())
            {
                msleep(10);
            }

            return 0;
        }

        edma_dev_interrupt_handler(-1, edma_dev);

        msleep(1);
    }

    return 0;
}

/**
 * pcie_edma_terminate_all - stop dma transaction on the channel.
 *
 * @chann: dma_chann structure pointer
 *
 * Return: always 0.
 *
 */
static ljeSTATUS edma_terminate_all(ljsDMA_CHANNEL *chan)
{
    edma_channel *edma_chan = to_edma_chan(chan);
    unsigned long flags;
    LIST_HEAD(list);

    int size = 0;

    if(edma_chan->direction == ljeDMA_TX_DIRECTION_MEM_TO_DEV)
        size = edma_read(edma_chan->reg_base, DMA_TRANSFER_SIZE_RDCH_REG(chan->id));
    else if(edma_chan->direction == ljeDMA_TX_DIRECTION_DEV_TO_MEM)
        size = edma_read(edma_chan->reg_base, DMA_TRANSFER_SIZE_WRCH_REG(chan->id));

    if( size !=0 )
        printk("------size = %d--maybe we need to add edma wait time!!!-----\n", size);

    edma_dev_transfer_stop(edma_chan->reg_base, edma_chan->chan.id, edma_chan->direction);

    spin_lock_irqsave(&edma_chan->lock, flags);
    list_splice_init(&edma_chan->pending_list, &list);
    list_splice_init(&edma_chan->active_list, &list);
    spin_unlock_irqrestore(&edma_chan->lock, flags);

    if (!list_empty(&list))
    {
        edma_free_descriptor_list(edma_chan, &list);
    }

    return ljvSTATUS_OK;
}

/**
 * edma_find_desc - get descriptor based on identifier in the active list.
 *
 * @chan: edma_channel structure pointer
 * @identifier: ljdDMA_TX_IDENTIFIER retured from tx_submit()
 *
 * Return: edma_descriptor pointer if found, otherwise NULL.
 *
 */
edma_descriptor *edma_find_desc(
    edma_channel *edma_chan,
    ljdDMA_TX_IDENTIFIER identifier
    )
{
    edma_descriptor *desc;

    /* find in active */
    list_for_each_entry(desc, &edma_chan->active_list, desc_node)
    {
        if (desc->txd.identifier == identifier)
        {
            return desc;
        }
    }
    /* find in pending */
    list_for_each_entry(desc, &edma_chan->pending_list, desc_node)
    {
        if (desc->txd.identifier == identifier)
        {
            return desc;
        }
    }

    return NULL;
}

/**
 * edma_get_residue - get residue in the descriptor list.
 *
 * @chan: dma_chann structure pointer
 * @identifier: dma_identifier_t retured from tx_submit()
 *
 * Return: number of residue data.
 *
 */
uint32_t edma_get_residue(
    edma_channel *edma_chan,
    ljdDMA_TX_IDENTIFIER identifier
    )
{
    edma_descriptor *desc;
    unsigned long flags;
    uint32_t residue;

    spin_lock_irqsave(&edma_chan->lock, flags);
    desc = edma_find_desc(edma_chan, identifier);

    if (desc)
    {
        residue = desc->residue;
    }
    else
    {
        residue = 0;
    }
    spin_unlock_irqrestore(&edma_chan->lock, flags);

    return residue;
}


/**
 * edma_tx_status - get current channel status.
 *
 * @chan: dma_chann structure pointer
 * @identifier: dma_identifier_t retured from tx_submit()
 * @state: ljsDMA_TX_STATE structure pointer
 *
 * Return: ljeDMA_STATUS.
 *
 */
static ljeDMA_STATUS edma_tx_status(
    ljsDMA_CHANNEL *chan,
    ljdDMA_TX_IDENTIFIER identifier,
    ljsDMA_TX_STATE *state
    )
{
    edma_channel *edma_chan = to_edma_chan(chan);
    unsigned long flags;
    ljeDMA_STATUS ret;
    ljtUINT32 pid;
    ljkOS_GetProcessID(&pid);


    ret = ljkDMA_IdentifierStatus(chan, identifier, state);
    if (ret == ljeDMA_STATUS_COMPLETE)
    {

        return ret;
    }

    spin_lock_irqsave(&edma_chan->lock, flags);
    edma_do_first_queued(edma_chan);
    spin_unlock_irqrestore(&edma_chan->lock, flags);

    ret = ljkDMA_IdentifierStatus(chan, identifier, state);
    if (ret == ljeDMA_STATUS_COMPLETE)
    {

        return ret;
    }

    ljkDMA_SetResidue(state, edma_get_residue(edma_chan, identifier));

    return ret;
}

static ljeSTATUS edma_config(ljsDMA_CHANNEL *chan, ljsDMA_SLAVE_CONFIG *sconfig)
{
    edma_channel *edma_chan;

    if (!chan || !sconfig)
    {
        return ljvSTATUS_INVALID_ARGUMENT;
    }

    edma_chan = to_edma_chan(chan);

    if (((sconfig->srcAddrBusWidth | (EDMA_BUS_WIDTHS)) != (EDMA_BUS_WIDTHS)) ||
        (sconfig->dstAddrBusWidth | (EDMA_BUS_WIDTHS)) != (EDMA_BUS_WIDTHS))
    {
        return ljvSTATUS_INVALID_ARGUMENT;
    }
    if (edma_chan->direction != sconfig->direction)
    {
        return ljvSTATUS_INVALID_ARGUMENT;
    }

    memcpy(&edma_chan->sconfig, sconfig, sizeof(*sconfig));

    return ljvSTATUS_OK;
}

/**
 * edma_issue_pending - issue channel transction from active_list
 *
 * @chann: dma_chann structure pointer
 *
 * Return: none
 *
 */
static void edma_issue_pending(ljsDMA_CHANNEL *chan)
{
    edma_channel *edma_chan = to_edma_chan(chan);
    unsigned long flags;

    spin_lock_irqsave(&edma_chan->lock, flags);
    if (list_empty(&edma_chan->active_list))
    {
        edma_do_first_queued(edma_chan);
    }
    spin_unlock_irqrestore(&edma_chan->lock, flags);
}

int edma_device_probe(edma_device *edma_dev)
{
#if LEGACY_INT
    int ret;
#endif
    ljeSTATUS status;
    struct task_struct * task;

    if (!edma_dev)
    {
        return ljvSTATUS_INVALID_ARGUMENT;
    }

    if (!edma_dev->pdev || !edma_dev->reg_base)
    {
        return ljvSTATUS_INVALID_ARGUMENT;
    }

    if(edma_dev->hdma_flag)
        hdma_dev_init(edma_dev->reg_base_2);
    else
        edma_dev_init(edma_dev->reg_base);

    INIT_LIST_HEAD(&edma_dev->slave_rdma.chanHead);
    INIT_LIST_HEAD(&edma_dev->slave_wdma.chanHead);
    /* init edma channels */

    if(edma_dev->hdma_flag)
        status = hdma_dev_channels_init(edma_dev);
    else
        status = edma_dev_channels_init(edma_dev);
    if (status)
    {
        return status;
    }

    /* get msi address and data */
    if(edma_dev->msi_flag)
    {
        status = edma_dev_msi_addr_and_data(edma_dev);
        if (status && DMA_NEW)
        {
            goto err_msi_addr;
        }
    }

    edma_dev->interrupt_enable = false;

    if (edma_dev->irq >= 0)
    {
#if LEGACY_INT
        if(edma_dev->hdma_flag)
        {
            ret = request_irq(edma_dev->irq, hdma_dev_interrupt_handler, IRQF_SHARED, "ljmhdma", edma_dev);
            if (ret)
            {
                    goto err_request_irq;
            }
        }else{
            ret = request_irq(edma_dev->irq, edma_dev_interrupt_handler, IRQF_SHARED, "ljmedma", edma_dev);
            if (ret)
            {
                    goto err_request_irq;
            }
        }
#endif
        edma_dev->interrupt_enable = true;
    }
    else
    {
        edma_dev->isr_thread = kthread_run(edma_polling, (void*)edma_dev, "galedma");

        if (IS_ERR(edma_dev->isr_thread))
        {
            printk("fail to start kthread for edma\n");
            goto err_request_irq;
        }
    }

    sprintf(edma_dev->slave_rdma.name, "galedma-r");
    sprintf(edma_dev->slave_wdma.name, "galedma-w");

    spin_lock_init(&edma_dev->irq_lock);
//  tasklet_init(&edma_dev->tasklet, edma_dev_tasklet, (unsigned long)edma_dev);
    sema_init(&sema_edma, 0);

    if(edma_dev->hdma_flag)
    {
        task = kthread_run(threadHdma, (void *)edma_dev,
                "hdma_deamon");
    }
    else
    {
        task = kthread_run(threadEdma, (void *)edma_dev,
                "edma_deamon");
    }

    ljkDMA_ZeroCapMask(&edma_dev->slave_rdma.devCapMasks);
    ljkDMA_SetCapMasks(&edma_dev->slave_rdma.devCapMasks, ljeDMA_TX_TYPE_PRIVATE);
    ljkDMA_SetCapMasks(&edma_dev->slave_rdma.devCapMasks, ljeDMA_TX_TYPE_SLAVE);

    edma_dev->slave_rdma.dev = &edma_dev->pdev->dev;
    edma_dev->slave_rdma.ljkDMADEV_Config = edma_config;

    if(edma_dev->hdma_flag)
    {
        edma_dev->slave_rdma.ljkDMADEV_AllocateChannelResource = hdma_dev_alloc_chan_resources;
        edma_dev->slave_rdma.ljkDMADEV_FreeChannelResource = hdma_dev_free_chan_resources;
        edma_dev->slave_rdma.ljkDMADEV_PrepareSlaveSG = hdma_dev_prep_slave_sg;
        edma_dev->slave_rdma.ljkDMADEV_PrepareDmaMemcpy = hdma_dev_prep_dma_memcpy;
        edma_dev->slave_rdma.ljkDMADEV_TerminateAll = hdma_terminate_all;
        edma_dev->slave_rdma.ljkDMADEV_TxStatus = hdma_tx_status;
        edma_dev->slave_rdma.ljkDMADEV_IssuePending = hdma_issue_pending;
    }
    else
    {
        edma_dev->slave_rdma.ljkDMADEV_AllocateChannelResource = edma_dev_alloc_chan_resources;
        edma_dev->slave_rdma.ljkDMADEV_FreeChannelResource = edma_dev_free_chan_resources;
        edma_dev->slave_rdma.ljkDMADEV_PrepareSlaveSG = edma_dev_prep_slave_sg;
        edma_dev->slave_rdma.ljkDMADEV_PrepareDmaMemcpy = edma_dev_prep_dma_memcpy;
        edma_dev->slave_rdma.ljkDMADEV_TerminateAll = edma_terminate_all;
        edma_dev->slave_rdma.ljkDMADEV_TxStatus = edma_tx_status;
        edma_dev->slave_rdma.ljkDMADEV_IssuePending = edma_issue_pending;
    }

    edma_dev->slave_rdma.caps.srcAddrWidth = EDMA_BUS_WIDTHS;
    edma_dev->slave_rdma.caps.dstAddrWidth = EDMA_BUS_WIDTHS;
    edma_dev->slave_rdma.caps.copyAlign = ljeDMA_ALIGNMENT_1_BYTE;
    edma_dev->slave_rdma.caps.directions = 1 << ljeDMA_TX_DIRECTION_MEM_TO_DEV;
    edma_dev->slave_rdma.caps.residueGranularity = ljeDMA_RESIDUE_GRANULARITY_DESCRIPTOR;

    ljkDMA_ZeroCapMask(&edma_dev->slave_wdma.devCapMasks);
    ljkDMA_SetCapMasks(&edma_dev->slave_wdma.devCapMasks, ljeDMA_TX_TYPE_PRIVATE);
    ljkDMA_SetCapMasks(&edma_dev->slave_wdma.devCapMasks, ljeDMA_TX_TYPE_SLAVE);

    edma_dev->slave_wdma.dev = &edma_dev->pdev->dev;
    edma_dev->slave_wdma.ljkDMADEV_Config = edma_config;

    if(edma_dev->hdma_flag)
    {
        edma_dev->slave_wdma.ljkDMADEV_AllocateChannelResource = hdma_dev_alloc_chan_resources;
        edma_dev->slave_wdma.ljkDMADEV_FreeChannelResource = hdma_dev_free_chan_resources;
        edma_dev->slave_wdma.ljkDMADEV_PrepareSlaveSG = hdma_dev_prep_slave_sg;
        edma_dev->slave_wdma.ljkDMADEV_PrepareDmaMemcpy = hdma_dev_prep_dma_memcpy;
        edma_dev->slave_wdma.ljkDMADEV_TerminateAll = hdma_terminate_all;
        edma_dev->slave_wdma.ljkDMADEV_TxStatus = hdma_tx_status;
        edma_dev->slave_wdma.ljkDMADEV_IssuePending = hdma_issue_pending;
    }
    else
    {
        edma_dev->slave_wdma.ljkDMADEV_AllocateChannelResource = edma_dev_alloc_chan_resources;
        edma_dev->slave_wdma.ljkDMADEV_FreeChannelResource = edma_dev_free_chan_resources;
        edma_dev->slave_wdma.ljkDMADEV_PrepareSlaveSG = edma_dev_prep_slave_sg;
        edma_dev->slave_wdma.ljkDMADEV_PrepareDmaMemcpy = edma_dev_prep_dma_memcpy;
        edma_dev->slave_wdma.ljkDMADEV_TerminateAll = edma_terminate_all;
        edma_dev->slave_wdma.ljkDMADEV_TxStatus = edma_tx_status;
        edma_dev->slave_wdma.ljkDMADEV_IssuePending = edma_issue_pending;
    }

    edma_dev->slave_wdma.caps.srcAddrWidth = EDMA_BUS_WIDTHS;
    edma_dev->slave_wdma.caps.dstAddrWidth = EDMA_BUS_WIDTHS;
    edma_dev->slave_wdma.caps.copyAlign = ljeDMA_ALIGNMENT_1_BYTE;
    edma_dev->slave_wdma.caps.directions = 1 << ljeDMA_TX_DIRECTION_DEV_TO_MEM;
    edma_dev->slave_wdma.caps.residueGranularity = ljeDMA_RESIDUE_GRANULARITY_DESCRIPTOR;

    /* register read channel device */
    status = ljkDMA_DeviceRegister(&edma_dev->slave_rdma);
    if (status)
    {
        goto err_rdma_register;
    }
    /* register write channel device */
    status = ljkDMA_DeviceRegister(&edma_dev->slave_wdma);

    if (status)
    {
        goto err_wdma_register;
    }

    return 0;
err_wdma_register:
    ljkDMA_DeviceUnregister(&edma_dev->slave_rdma);
err_rdma_register:
    if (edma_dev->interrupt_enable)
    {
        free_irq(edma_dev->irq, edma_dev);
    }
    else
    {
        edma_dev->kill_thread = true;
        kthread_stop(edma_dev->isr_thread);
        edma_dev->isr_thread = NULL;
    }
//  tasklet_kill(&edma_dev->tasklet);
    EdmaKillThread = ljvTRUE;
    up(&sema_edma);

err_request_irq:
err_msi_addr:

    if (edma_dev->rchan)
    {
        kfree(edma_dev->rchan);
    }
    if (edma_dev->wchan)
    {
        kfree(edma_dev->wchan);
    }

    return -1;
}
/**
 * edma_device_remove - remove edma resource
 *
 * @edma_device: pcie_edma_soc structure pointer
 *
 * Return: none
 *
 */
void edma_device_remove(edma_device *edma_dev)
{
    edma_channel *chan, *tmp;
    int i = 0;

    ljkDMA_DeviceUnregister(&edma_dev->slave_rdma);
    ljkDMA_DeviceUnregister(&edma_dev->slave_wdma);

    if (edma_dev->interrupt_enable)
    {
        free_irq(edma_dev->irq, edma_dev);
    }
    else
    {
        edma_dev->kill_thread = true;
        kthread_stop(edma_dev->isr_thread);
        edma_dev->isr_thread = NULL;
    }
//  tasklet_kill(&edma_dev->tasklet);
    EdmaKillThread = ljvTRUE;
    up(&sema_edma);

    /* free write channels desc */
    for (i = 0; i < edma_dev->wchan_count; i++)
    {
               edma_channel *chan = &edma_dev->wchan[i];
               kfree(chan->desc);
    }
    /* free read channels desc */
    for (i = 0; i < edma_dev->rchan_count; i++)
    {
               edma_channel *chan = &edma_dev->rchan[i];
               kfree(chan->desc);
    }

    list_for_each_entry_safe(chan, tmp, &edma_dev->slave_rdma.chanHead, chan.devNode)
    {
        list_del(&chan->chan.devNode);
    }
    list_for_each_entry_safe(chan, tmp, &edma_dev->slave_wdma.chanHead, chan.devNode)
    {
        list_del(&chan->chan.devNode);
    }
    kfree(edma_dev->rchan);
    kfree(edma_dev->wchan);

}

