/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

/******************************************************************************
*
*  版权所有(c)，2020-2099，中船第709研究所
*
*******************************************************************************
*
*  文件名称: lj_hal_kernel_platform_hdma.c
*  版 本 号: V1.0
*  作    者: cbb
*  生成日期: 2023年5月15日
*
*  功能描述: HDMA
*
*  修改历史:
*         1. 日    期: 2023年5月15日
*            作    者: cbb
*            修改内容: 创建文件
*
******************************************************************************/
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/semaphore.h>

#include "lj_hal_kernel_platform_hdma.h"
#include "lj_hal_kernel_dma_engine.h"
#include "ljmicro_family.h"

extern    struct semaphore    sema_edma;
extern    ljtBOOL             EdmaKillThread;
static int delay_time = 1;

/*
 * edma_dev_check_status
 * Description:
 *         pcie check staus function, it main use to check some bit value change
 * Parameter:
 *         iomem_t reg_base -- check register address
 *         uint32_t flag -- check bit field,0x1<<x
 *         uint32_t level -- check bit value,
 *         uint32_t timeou t-- check time limit
 * Return:
 *         pcie_status
 */
edma_status hdma_dev_check_status(
    iomem_t reg_base,
    uint32_t reg, uint32_t flag,uint8_t level,uint32_t timeout)
{
    uint32_t count = timeout;

    do
    {
        if (level)
        {
            if(edma_read(reg_base, reg) & flag)
            {
                return EDMA_OK;
            }
        }
        else
        {
            if ((edma_read(reg_base, reg) & flag) == 0)
            {
                return EDMA_OK;
            }
        }
        count--;
        cpu_relax();
    } while((count != 0) || (timeout == 0));

    return EDMA_TIMEOUT;
}

static void hdma_dev_write_engine_off(
    iomem_t reg_base,
    uint8_t channel
)
{
    uint32_t reg;

    edma_write(reg_base, HDMA_EN_OFF_WRCH(channel), 0x0);
    do
    {
        reg = edma_read(reg_base, HDMA_EN_OFF_WRCH(channel));
        if (reg == 0)
        {
            break;
        }
        cpu_relax();
    } while(1);
}

static void hdma_dev_write_engine_on(
    iomem_t reg_base,
    uint8_t channel
)
{
    uint32_t reg;

    reg = edma_read(reg_base, HDMA_EN_OFF_WRCH(channel));
    if (reg == 0)
    {
        edma_write(reg_base, HDMA_EN_OFF_WRCH(channel), 0x1);
    }
}

static void hdma_dev_read_engine_off(
    iomem_t reg_base,
    uint8_t channel
)
{
    uint32_t reg;

    edma_write(reg_base, HDMA_EN_OFF_RDCH(channel), 0x0);
    do
    {
        reg = edma_read(reg_base, HDMA_EN_OFF_RDCH(channel));
        if (reg == 0)
        {
            break;
        }
        cpu_relax();
    } while(1);
}

static void hdma_dev_read_engine_on(
    iomem_t reg_base,
    uint8_t channel
)
{
    uint32_t reg;

    reg = edma_read(reg_base, HDMA_EN_OFF_RDCH(channel));
    if (reg == 0)
    {
        edma_write(reg_base, HDMA_EN_OFF_RDCH(channel), 0x1);
    }
}

static void hdma_dev_transfer_start(
    iomem_t reg_base,
    uint8_t channel, uint8_t direction)
{
    switch (direction)
    {
        case ljeDMA_TX_DIRECTION_DEV_TO_MEM:
            hdma_dev_write_engine_on(reg_base, channel);
            /* enable the write doorbell */
            edma_write(reg_base, HDMA_DOORBELL_OFF_WRCH(channel), 0x1);
            break;
        case ljeDMA_TX_DIRECTION_MEM_TO_DEV:
            hdma_dev_read_engine_on(reg_base, channel);
            /*  enable the read doorbell */
            edma_write(reg_base, HDMA_DOORBELL_OFF_RDCH(channel), 0x1);
            break;
        default:
            printk("glcore: invaild pcie dDMA parameter.\n");
            break;
    }
#if EDMA_DEBUG
    printk("hdma_dev_transfer_start: dir: 0x%d.; channel: 0x%d\n", direction, channel);
#endif
}

/*
 * Function:
 *         uint32_t hdma_dev_get_channel_status
 * Description:
 *         get channel hw status
 * Parameter:
 *         iomem_t reg_base-- pcie dma config register  base address
 *         uint8_t channel--dma transfer channel,
 *         uint8_t direction--dma transfer dirction, write or read,
 * Return:
 *         DMA_CS_RUNNING, DMA_CS_HALTED, DMA_CS_STOPPED
 */
static uint32_t hdma_dev_get_channel_status(
    iomem_t reg_base,
    uint8_t channel,
    uint8_t direction
    )
{

    uint32_t reg = 0;

    /* read control1 register to check CS field */
    if (direction == ljeDMA_TX_DIRECTION_DEV_TO_MEM)
    {
        reg = edma_read(reg_base, HDMA_STATUS_OFF_WRCH(channel)) & (0x7L);
    }
    else if ( direction == ljeDMA_TX_DIRECTION_MEM_TO_DEV)
    {
        reg = edma_read(reg_base, HDMA_STATUS_OFF_RDCH(channel)) & (0x7L);
    }

    return reg;
}

/*
 * Function:
 *         edma_status edma_dev_transfer_stop(iomem_t reg_base, uint32_t channel, uint32_t direction)
 * Description:
 *         pcie dma stop current transfer
 * Parameter:
 *         iomem_t reg_base-- pcie dma config register  base address
 *         uint8_t channel--dma transfer channel,
 *         uint8_t direction--dma transfer dirction, write or read,
 * Return:
 *         edma_status
 */
static edma_status hdma_dev_transfer_stop(
    iomem_t reg_base,
    uint8_t channel,
    uint8_t direction
    )
{
    uint32_t reg;

    if (direction == ljeDMA_TX_DIRECTION_DEV_TO_MEM)
    {
        /* ensure is runing */
        if (hdma_dev_get_channel_status(reg_base, channel, direction) == HDMA_CS_RUNNING)
        {
            /* stop the channel */
            edma_write(reg_base, HDMA_DOORBELL_OFF_WRCH(channel), HDMA_TRANSFER_STOP);
            if (hdma_dev_check_status(reg_base, HDMA_INT_STATUS_OFF_WRCH(channel),
                HDMA_ABORT_INT, 1, 5000) == EDMA_OK)
            {
                /* clear the status */
                edma_write(reg_base, HDMA_INT_STATUS_OFF_WRCH(channel), HDMA_ABORT_INT);
                /* read control1 register to check CS field */
                reg = edma_read(reg_base, HDMA_STATUS_OFF_WRCH(channel)) & 0x7;
                /* ensure is sttopped */
                if (reg == HDMA_CS_STOPPED)
                {
                    return EDMA_ABORT;
                }
            }
        }
    }
    else if ( direction == ljeDMA_TX_DIRECTION_MEM_TO_DEV)
    {
        /* ensure is runing */
        if (hdma_dev_get_channel_status(reg_base, channel, direction) == HDMA_CS_RUNNING)
        {
            /* stop the channel */
            edma_write(reg_base, HDMA_DOORBELL_OFF_RDCH(channel), HDMA_TRANSFER_STOP);
            if (hdma_dev_check_status(reg_base, HDMA_INT_STATUS_OFF_RDCH(channel),
                HDMA_ABORT_INT, 1, 5000)== EDMA_OK)
            {
                /* clear the status */
                edma_write(reg_base, HDMA_INT_STATUS_OFF_RDCH(channel), HDMA_ABORT_INT);
                /* read control1 register to check CS field */
                reg = edma_read(reg_base, HDMA_STATUS_OFF_RDCH(channel)) & 0x7;
                /* ensure is runing */
                if(reg == DMA_CS_STOPPED)
                {
                    return EDMA_ABORT;
                }
            }
        }
    }

    return EDMA_ERROR;
}

/*
 * Function:
 *         edma_status edma_dev_write_channel_config(edma_hw_config *config)
 * Description:
 *         pcie dma write channel configure function
 * Parameter:
 *         edma_hw_config *config--pcie dma write channel configure information;
 * Return:
 *         edma_status
 */
static edma_status hdma_dev_write_channel_config(
    edma_hw_config *config
    )
{
    iomem_t reg_base = config->reg_base;
    uint32_t reg = HDMA_V0_MAX_NR_CH;

    /* if channel does not exist, return error */
    if (config->channel >= reg)
    {
        return EDMA_ERROR;
    }

    /*  enable the write engine */
    hdma_dev_write_engine_on(reg_base, config->channel);

    /* set write interrupt mask */
    edma_write(reg_base, HDMA_INT_SETUP_OFF_WRCH(config->channel), config->interrupt_mask);

    /* Channel control */
    edma_write(reg_base, HDMA_CONTROL1_OFF_WRCH(config->channel), HDMA_V0_LINKLIST_EN);

    /* set LLP */
    edma_write(reg_base, HDMA_LLP_LOW_OFF_WRCH(config->channel), config->llp_low);
    edma_write(reg_base, HDMA_LLP_HIGH_OFF_WRCH(config->channel), config->llp_high);
#if EDMA_DEBUG
    printk("write back:channel=0x%x. \r\n", config->channel);
    printk("write back:enable=0x%x. \r\n", edma_read(reg_base, HDMA_EN_OFF_WRCH(config->channel)));
    printk("write back:sar_low=0x%x. \r\n", edma_read(reg_base, HDMA_SAR_LOW_OFF_WRCH(config->channel)));
    printk("write back:sar_high=0x%x. \r\n", edma_read(reg_base, HDMA_SAR_HIGH_OFF_WRCH(config->channel)));
    printk("write back:dar_low=0x%x. \r\n", edma_read(reg_base, HDMA_DAR_LOW_OFF_WRCH(config->channel)));
    printk("write back:dar_high=0x%x. \r\n", edma_read(reg_base, HDMA_DAR_HIGH_OFF_WRCH(config->channel)));
    printk("write back:length=0x%x. \r\n", edma_read(reg_base, HDMA_XFERSIZE_OFF_WRCH(config->channel)));
    printk("write back:interrupt_mask=0x%x. \r\n", edma_read(reg_base, HDMA_INT_SETUP_OFF_WRCH(config->channel)));
    printk("write back:control1=0x%x. \r\n", edma_read(reg_base, HDMA_CONTROL1_OFF_WRCH(config->channel)));
    printk("write back:llp_low=0x%x. \r\n", edma_read(reg_base, HDMA_LLP_LOW_OFF_WRCH(config->channel)));
    printk("write back:llp_high=0x%x. \r\n", edma_read(reg_base, HDMA_LLP_HIGH_OFF_WRCH(config->channel)));
#endif

    return EDMA_OK;

}

/*
 * Function:
 *         edma_status edma_dev_read_channel_config(edma_hw_config *config)
 * Description:
 *         pcie dma read channel configure function
 * Parameter:
 *         edma_hw_config *config--pcie dma read channel configure information;
 * Return:
 *         edma_status
 */
static edma_status hdma_dev_read_channel_config(
    edma_hw_config *config
    )
{
    iomem_t reg_base = config->reg_base;
    uint32_t reg = HDMA_V0_MAX_NR_CH;

    /* if channel does not exist, return error */
    if (config->channel >= reg)
    {
        return EDMA_ERROR;
    }

    /*  enable the read engine */
    hdma_dev_read_engine_on(reg_base, config->channel);

    /* set write interrupt mask */
    edma_write(reg_base, HDMA_INT_SETUP_OFF_RDCH(config->channel), config->interrupt_mask);

    /* Channel control */
    edma_write(reg_base, HDMA_CONTROL1_OFF_RDCH(config->channel), HDMA_V0_LINKLIST_EN);

    /* set LLP */
    edma_write(reg_base, HDMA_LLP_LOW_OFF_RDCH(config->channel), config->llp_low);
    edma_write(reg_base, HDMA_LLP_HIGH_OFF_RDCH(config->channel), config->llp_high);

#if EDMA_DEBUG
    printk("read back:channel=0x%x. \r\n", config->channel);
    printk("read back:enable=0x%x. \r\n", edma_read(reg_base, HDMA_EN_OFF_RDCH(config->channel)));
    printk("read back:sar_low=0x%x. \r\n", edma_read(reg_base, HDMA_SAR_LOW_OFF_RDCH(config->channel)));
    printk("read back:sar_high=0x%x. \r\n", edma_read(reg_base, HDMA_SAR_HIGH_OFF_RDCH(config->channel)));
    printk("read back:dar_low=0x%x. \r\n", edma_read(reg_base, HDMA_DAR_LOW_OFF_RDCH(config->channel)));
    printk("read back:dar_high=0x%x. \r\n", edma_read(reg_base, HDMA_DAR_HIGH_OFF_RDCH(config->channel)));
    printk("read back:length=0x%x. \r\n", edma_read(reg_base, HDMA_XFERSIZE_OFF_RDCH(config->channel)));
    printk("read back:interrupt_mask=0x%x. \r\n", edma_read(reg_base, HDMA_INT_SETUP_OFF_RDCH(config->channel)));
    printk("read back:control1=0x%x. \r\n", edma_read(reg_base, HDMA_CONTROL1_OFF_RDCH(config->channel)));
    printk("read back:llp_low=0x%x. \r\n", edma_read(reg_base, HDMA_LLP_LOW_OFF_RDCH(config->channel)));
    printk("read back:llp_high=0x%x. \r\n", edma_read(reg_base, HDMA_LLP_HIGH_OFF_RDCH(config->channel)));
#endif
    return EDMA_OK;
}


 /*
 * Function:
 *         edma_status edma_dev_write_transfer(edma_transfer_info *info)
 * Description:
 *         pcie dma write transfer function
 * Parameter:
 *         edma_transfer_info *info--pcie dma write transfer handle, it contains all necessary parameters;
 * Return:
 *         edma_status
 */
static edma_status hdma_dev_write_transfer(
    edma_transfer_info *info
    )
{
    uint32_t reg;
    edma_hw_config config;

    /* double config the info according to transfer mode */
    config.channel = info->channel;
    config.reg_base = info->reg_base;
    config.arbi_weight = 0xf;
    if (info->transfer_mode == EDMA_TRANSFER_MODE_NORMAL)
    {
        config.transfer_size = info->transfer_size;
        config.sar_low = QWORD_LO(info->src_addr);
        config.sar_high = QWORD_HI(info->src_addr);
        config.dar_low = QWORD_LO(info->dst_addr);
        config.dar_high = QWORD_HI(info->dst_addr);
        config.link_err_en = 0;
        config.llp_high = 0;
        config.llp_low = 0;
    }
    else
    {
        config.transfer_size = 0;
        config.sar_low = 0;
        config.sar_high = 0;
        config.dar_low = 0;
        config.dar_high = 0;
        config.llp_high = QWORD_HI(info->lle);
        config.llp_low = QWORD_LO(info->lle);
    }
    /* cpu config local dma */
    if (info->cpu_mode == EDMA_CPU_MODE_LOCAL)
    {
        config.done_imwr_addr_low = 0;
        config.done_imwr_addr_high = 0;
        config.abort_imwr_addr_low = 0;
        config.abort_imwr_addr_high = 0;
        config.imwr_data = 0;
        /* interrupt mode */
        if (info->check_mode == EDMA_CHECK_MODE_INTERRUPT)
        {
/*            reg = edma_read(info->reg_base, DMA_WRITE_INT_MASK_REG);
            reg &= (~(DMA_DONE_INT(info->channel) | DMA_ABORT_INT(info->channel)));*/
            config.interrupt_mask = HDMA_V0_LOCAL_STOP_INT_EN;

            if (info->transfer_mode == EDMA_TRANSFER_MODE_NORMAL)
            {
                /* open local cpu interrupt */
                config.control1 = DMA_LIE | DMA_TD;
            }
            else if (info->transfer_mode == EDMA_TRANSFER_MODE_LINK)
            {
                /* LL mode */
                //reg = edma_read(info->reg_base, DMA_WRITE_LINKED_LIST_ERR_EN_REG);
                reg |= (DMA_LLLAIE(info->channel));
                config.link_err_en = reg;
                /* open local cpu interrupt */
                config.control1=DMA_TD|DMA_LIE|DMA_LLE|DMA_CCS;
            }
            else
            {
                printk("invalid transfer mode\n");
                return EDMA_ERROR;
            }
        }
        /* polling mode */
        else if (info->check_mode == EDMA_CHECK_MODE_POLLING)
        {
/*            reg = edma_read(info->reg_base, DMA_WRITE_INT_MASK_REG);
            reg |= (DMA_DONE_INT(info->channel) | DMA_ABORT_INT(info->channel));*/
            config.interrupt_mask = HDMA_V0_LOCAL_STOP_INT_EN;

            if (info->transfer_mode == EDMA_TRANSFER_MODE_NORMAL)
            {
                /* close local cpu interrupt */
                config.control1 = DMA_TD;
            }
            else if (info->transfer_mode == EDMA_TRANSFER_MODE_LINK)
            {
                /* LL mode */
                //reg = edma_read(info->reg_base, DMA_WRITE_LINKED_LIST_ERR_EN_REG);
                reg |= (DMA_LLLAIE(info->channel));
                config.link_err_en = reg;
                /* open local cpu interrupt */
                config.control1=DMA_TD|DMA_LLE|DMA_CCS;
            }
            else
            {
                printk("invalid transfer mode\n");
                return EDMA_ERROR;
            }
        }
        else
        {
            printk("invalid check mode\n");
            return EDMA_ERROR;
        }
        config.control2 = 0;
    }
    /* cpu config remote dma */
    else if (info->cpu_mode == EDMA_CPU_MODE_REMOTE)
    {
        /* interrupt mode */
        if (info->check_mode == EDMA_CHECK_MODE_INTERRUPT)
        {
            config.done_imwr_addr_low = QWORD_LO(info->msi_addr);
            config.done_imwr_addr_high = QWORD_HI(info->msi_addr);
            config.abort_imwr_addr_low = QWORD_LO(info->msi_addr);
            config.abort_imwr_addr_high = QWORD_HI(info->msi_addr);
            config.imwr_data = info->msi_data;
/*            reg = edma_read(info->reg_base, DMA_WRITE_INT_MASK_REG);
            reg &= (~(DMA_DONE_INT(info->channel) | DMA_ABORT_INT(info->channel)));*/
            config.interrupt_mask = HDMA_V0_LOCAL_STOP_INT_EN;

            if (info->transfer_mode == EDMA_TRANSFER_MODE_NORMAL)
            {
                /* open local cpu interrupt */
                config.control1 = DMA_RIE | DMA_TD | DMA_LIE;
            }
            else if (info->transfer_mode == EDMA_TRANSFER_MODE_LINK)
            {
                /* LL mode */
                //reg = edma_read(info->reg_base, DMA_WRITE_LINKED_LIST_ERR_EN_REG);
                reg |= (DMA_LLRAIE(info->channel));
                config.link_err_en = reg;
                /* open local cpu interrupt */
                config.control1=DMA_TD|DMA_LIE|DMA_RIE|DMA_LLE|DMA_CCS;
            }
            else
            {
                printk("invalid transfer mode\n");
                return EDMA_ERROR;
            }
        }
        /* polling mode */
        else if (info->check_mode == EDMA_CHECK_MODE_POLLING)
        {
            config.done_imwr_addr_low = 0;
            config.done_imwr_addr_high = 0;
            config.abort_imwr_addr_low = 0;
            config.abort_imwr_addr_high = 0;
            config.imwr_data = 0;
/*            reg = edma_read(info->reg_base, DMA_WRITE_INT_MASK_REG);
            reg |= (DMA_DONE_INT(info->channel) | DMA_ABORT_INT(info->channel));*/
            config.interrupt_mask = HDMA_V0_LOCAL_STOP_INT_EN;
            /* close local cpu interrupt */
            config.control1 = DMA_LIE | DMA_TD;

            if (info->transfer_mode == EDMA_TRANSFER_MODE_NORMAL)
            {
                /* open local cpu interrupt */
                config.control1 = DMA_LIE | DMA_TD;
            }
            else if (info->transfer_mode == EDMA_TRANSFER_MODE_LINK)
            {
                /* LL mode */
                //reg = edma_read(info->reg_base, DMA_WRITE_LINKED_LIST_ERR_EN_REG);
                reg |= (DMA_LLRAIE(info->channel));
                config.link_err_en = reg;
                /* open local cpu interrupt */
                config.control1=DMA_TD|DMA_LLE|DMA_CCS;
            }
            else
            {
                printk("invalid transfer mode\n");
                return EDMA_ERROR;
            }
        }
        else
        {
            printk("invalid check mode\n");
            return EDMA_ERROR;
        }
        config.control2 = 0;
    }
    else
    {
        printk("invalid cpu mode\n");
        return EDMA_ERROR;
    }

    /* config the write channel */
    if (hdma_dev_write_channel_config(&config) == EDMA_ERROR)
    {
        return EDMA_ERROR;
    }
    /* open the door bell, start transfer */
    hdma_dev_transfer_start(info->reg_base, info->channel, ljeDMA_TX_DIRECTION_DEV_TO_MEM);

    return EDMA_OK;
}

/*
 * edma_read_transfer
 */
static edma_status hdma_dev_read_transfer(
    edma_transfer_info *info
    )
{
    uint32_t reg;
    edma_hw_config config;

    /* double config the info according to transfer mode */
    config.channel = info->channel;
    config.reg_base = info->reg_base;
    config.arbi_weight = 0xf;
    if (info->transfer_mode == EDMA_TRANSFER_MODE_NORMAL)
    {
        config.transfer_size = info->transfer_size;
        config.sar_low = QWORD_LO(info->src_addr);
        config.sar_high = QWORD_HI(info->src_addr);
        config.dar_low = QWORD_LO(info->dst_addr);
        config.dar_high = QWORD_HI(info->dst_addr);
        config.link_err_en = 0;
        config.llp_high = 0;
        config.llp_low = 0;
    }
    else
    {
        config.transfer_size = 0;
        config.sar_low = 0;
        config.sar_high = 0;
        config.dar_low = 0;
        config.dar_high = 0;
        config.llp_high = QWORD_HI(info->lle);
        config.llp_low = QWORD_LO(info->lle);
    }
    /* cpu config local dma */
    if (info->cpu_mode == EDMA_CPU_MODE_LOCAL)
    {
        config.done_imwr_addr_low = 0;
        config.done_imwr_addr_high = 0;
        config.abort_imwr_addr_low = 0;
        config.abort_imwr_addr_high = 0;
        config.imwr_data = 0;
        /* interrupt mode */
        if(info->check_mode == EDMA_CHECK_MODE_INTERRUPT)
        {
            /*reg = edma_read(info->reg_base, DMA_READ_INT_MASK_REG);
            reg &= (~(DMA_DONE_INT(info->channel) | DMA_ABORT_INT(info->channel)));*/
            config.interrupt_mask = HDMA_V0_LOCAL_STOP_INT_EN;

            if (info->transfer_mode == EDMA_TRANSFER_MODE_NORMAL)
            {
                /* open local cpu interrupt */
                config.control1 = DMA_LIE | DMA_TD;
            }
            else if (info->transfer_mode == EDMA_TRANSFER_MODE_LINK)
            {
                //reg = edma_read(info->reg_base, DMA_READ_LINKED_LIST_ERR_EN_REG);
                /* unmask the interrupt */
                reg |= (DMA_LLLAIE(info->channel));
                config.link_err_en = reg;

                /* open local cpu interrupt */
                config.control1=DMA_TD|DMA_LIE|DMA_LLE|DMA_CCS;
            }
            else
            {
                printk("invalid transfer mode\n");
                return EDMA_ERROR;
            }
        }
        /* polling mode */
        else if (info->check_mode == EDMA_CHECK_MODE_POLLING)
        {
            //reg = edma_read(info->reg_base, DMA_READ_INT_MASK_REG);
            /* mask the interrupt */
            //reg |= (DMA_DONE_INT(info->channel) | DMA_ABORT_INT(info->channel));
            config.interrupt_mask = HDMA_V0_LOCAL_STOP_INT_EN;

            if (info->transfer_mode == EDMA_TRANSFER_MODE_NORMAL)
            {
                /* close local cpu interrupt */
                config.control1 = DMA_LIE | DMA_TD;
            }
            else if (info->transfer_mode == EDMA_TRANSFER_MODE_LINK)
            {
                /* LL mode */
                //reg = edma_read(info->reg_base, DMA_READ_LINKED_LIST_ERR_EN_REG);
                reg |= (DMA_LLLAIE(info->channel));
                config.link_err_en = reg;
                /* open local cpu interrupt */
                config.control1=DMA_TD|DMA_LLE|DMA_CCS;
            }
            else
            {
                printk("invalid transfer mode\n");
                return EDMA_ERROR;
            }
        }
        else
        {
            printk("invalid check mode\n");
            return EDMA_ERROR;
        }
        config.control2 = 0;
    }
    /* cpu config remote dma */
    else if (info->cpu_mode == EDMA_CPU_MODE_REMOTE)
    {
        /* interrupt mode */
        if (info->check_mode == EDMA_CHECK_MODE_INTERRUPT)
        {
            config.done_imwr_addr_low = QWORD_LO(info->msi_addr);
            config.done_imwr_addr_high = QWORD_HI(info->msi_addr);
            config.abort_imwr_addr_low = QWORD_LO(info->msi_addr);
            config.abort_imwr_addr_high = QWORD_HI(info->msi_addr);
            config.imwr_data = info->msi_data;
            /*reg = edma_read(info->reg_base, DMA_READ_INT_MASK_REG);
            reg &= (~(DMA_DONE_INT(info->channel) | DMA_ABORT_INT(info->channel)));*/
            config.interrupt_mask = HDMA_V0_LOCAL_STOP_INT_EN;

            if (info->transfer_mode == EDMA_TRANSFER_MODE_NORMAL)
            {
                /* open remote cpu interrupt */
                config.control1 = DMA_RIE | DMA_TD | DMA_LIE;
            }
            else if (info->transfer_mode == EDMA_TRANSFER_MODE_LINK)
            {
                /* LL mode */
                //reg = edma_read(info->reg_base, DMA_READ_LINKED_LIST_ERR_EN_REG);
                reg |= (DMA_LLRAIE(info->channel));
                config.link_err_en = reg;
                /* open remote cpu interrupt */
                config.control1=DMA_TD|DMA_RIE|DMA_LLE|DMA_CCS;
            }
            else
            {
                printk("invalid transfer mode\n");
                return EDMA_ERROR;
            }
        }
        /* polling mode */
        else if (info->check_mode == EDMA_CHECK_MODE_POLLING)
        {
            config.done_imwr_addr_low = 0;
            config.done_imwr_addr_high = 0;
            config.abort_imwr_addr_low = 0;
            config.abort_imwr_addr_high = 0;
            config.imwr_data = 0;
            //reg = edma_read(info->reg_base, DMA_READ_INT_MASK_REG);
            /* mask the interrupt */
            //reg |=(DMA_DONE_INT(info->channel) | DMA_ABORT_INT(info->channel));
            config.interrupt_mask = HDMA_V0_LOCAL_STOP_INT_EN;

            if (info->transfer_mode == EDMA_TRANSFER_MODE_NORMAL)
            {
                /* close local cpu interrupt */
                config.control1 = DMA_LIE | DMA_TD;
            }
            else if (info->transfer_mode == EDMA_TRANSFER_MODE_LINK)
            {
                /* LL mode */
                //reg = edma_read(info->reg_base, DMA_READ_LINKED_LIST_ERR_EN_REG);
                reg |= (DMA_LLRAIE(info->channel));
                config.link_err_en = reg;
                /* close local cpu interrupt */
                config.control1=DMA_TD|DMA_LLE|DMA_CCS;
            }
            else
            {
                printk("invalid transfer mode\n");
                return EDMA_ERROR;
            }
        }
        else
        {
            printk("invalid check mode\n");
            return EDMA_ERROR;
        }
        config.control2 = 0;
    }
    else
    {
        printk("invalid cpu mode\n");
        return EDMA_ERROR;
    }
    /* config the write channel */
    if (hdma_dev_read_channel_config(&config) == EDMA_ERROR)
    {
        return EDMA_ERROR;
    }
    /* open the door bell, start transfer */
    hdma_dev_transfer_start(info->reg_base, info->channel, ljeDMA_TX_DIRECTION_MEM_TO_DEV);

    return EDMA_OK;
}

static void hdma_dev_fill_link_table(
    ljsDMA_CHANNEL *chan,
    edma_descriptor *desc,
    struct scatterlist *sgl,
    size_t sg_len)
{
    edma_channel *edma_chan = to_edma_chan(chan);
    edma_device *edma_dev = to_edma_dev(chan->device);
    edma_desc_ll *link_table = ljvNULL;
    edma_desc_llp *edma_llp = ljvNULL;
    struct scatterlist *sg;
    dma_addr_t addr, src = 0, dst = 0;
    size_t avail, total_len = 0;
    uint32_t ctl = 0;
    int i = 0, i_v = 0;

    if (edma_chan->direction == ljeDMA_TX_DIRECTION_MEM_TO_DEV)
    {
        desc->lle = (uint64_t)LINKTABLE_OFFSET(
                        edma_dev->plink_table_base, edma_chan->chan.id);

        link_table = (edma_desc_ll*)LINKTABLE_OFFSET(
                        edma_dev->vlink_table_base, edma_chan->chan.id);
    }
    else if (edma_chan->direction == ljeDMA_TX_DIRECTION_DEV_TO_MEM)
    {
        desc->lle = (uint64_t)LINKTABLE_OFFSET(
                        edma_dev->plink_table_base,
                        edma_dev->rchan_count + edma_chan->chan.id);

        link_table = (edma_desc_ll*)LINKTABLE_OFFSET(
                        edma_dev->vlink_table_base,
                        edma_dev->rchan_count + edma_chan->chan.id);
    }

    src = edma_chan->sconfig.srcAddr;
    dst = edma_chan->sconfig.dstAddr;
    //ctl = DMA_CB;
    ctl = 0;
    for_each_sg(sgl, sg, sg_len, i)
    {
        ljmkASSERT(i * ljmSIZEOF(edma_desc_ll) < (edma_dev->single_chan ? EDMA_LT_SIZE*4 : EDMA_LT_SIZE));

        avail = sg_dma_len(sg);
        addr = sg_dma_address(sg);

        link_table[i].control = ctl;
        link_table[i].size = avail;

        if (edma_chan->direction == ljeDMA_TX_DIRECTION_MEM_TO_DEV)
        {
            /* mem -> dev */
            link_table[i].sar_low = QWORD_LO(addr);
            link_table[i].sar_high = QWORD_HI(addr);
            link_table[i].dar_low = QWORD_LO(dst);
            link_table[i].dar_high = QWORD_HI(dst);
            dst += avail;
        }
        else
        {
            /* dev -> mem */
            link_table[i].sar_low = QWORD_LO(src);
            link_table[i].sar_high = QWORD_HI(src);
            link_table[i].dar_low = QWORD_LO(addr);
            link_table[i].dar_high = QWORD_HI(addr);
            src += avail;
        }
        total_len += avail;
    }

    if (edma_dev->interrupt_enable)
    {
        //ctl |= (DMA_LIE | DMA_RIE);
        link_table[sg_len - 1].control = ctl;
    }
    else
    {
        ctl |= DMA_LIE;
        link_table[sg_len - 1].control = ctl;
    }

    edma_llp = (edma_desc_llp*)(&link_table[sg_len]);
    edma_llp->control = DMA_LLP | DMA_CB;
    edma_llp->reserved = 0;
    edma_llp->llp_high = QWORD_HI(desc->lle);
    edma_llp->llp_low = QWORD_LO(desc->lle);

    /* insure the config has written into link table */
    //link_table[sg_len + 1].control = 0x1;
    //while(!link_table[sg_len + 1].control);
    //memset(&link_table[sg_len + 1], 0, sizeof(link_table[sg_len]));
    //while(link_table[sg_len + 1].control);

    desc->total_len = total_len;
    desc->residue = total_len;

#if EDMA_DEBUG
    for (i = 0; i < sg_len; i++)
    {
        printk(" fill link table [%d]@0x%08llx:\n", i, desc->lle + sizeof(edma_desc_ll)*i);
        printk("    control = 0x%08x\n", link_table[i].control);
        printk("    size    = 0x%08x\n", link_table[i].size);
        printk("    sar_lo  = 0x%08x\n", link_table[i].sar_low);
        printk("    sar_hi  = 0x%08x\n", link_table[i].sar_high);
        printk("    dar_lo  = 0x%08x\n", link_table[i].dar_low);
        printk("    dar_hi  = 0x%08x\n", link_table[i].dar_high);
    }
    printk("    llp_control     = 0x%08x\n", link_table[i].control);
    printk("    llp_reserve     = 0x%08x\n", link_table[i].size);
    printk("    llp_low        = 0x%08x\n", link_table[i].sar_low);
    printk("    llp_high         = 0x%08x\n", link_table[i].sar_high);
#endif

#if (defined CONFIG_LOONGARCH ||defined CONFIG_SW64)
    udelay(delay_time);
#endif
    if (edma_chan->direction == ljeDMA_TX_DIRECTION_MEM_TO_DEV)
    {
    dst -= avail;
    while((link_table[sg_len-1].size != avail) || (link_table[sg_len-1].sar_high != QWORD_HI(addr)) || (link_table[sg_len-1].dar_high != QWORD_HI(dst)) )
     {

          /*printk("wait DDR SG0    control = 0x%08x\n", link_table[sg_len-1].control);
          printk("    size    = 0x%08x\n", link_table[sg_len-1].size);
          printk("    sar_lo  = 0x%08x\n", link_table[sg_len-1].sar_low);
          printk("    sar_hi  = 0x%08x\n", link_table[sg_len-1].sar_high);
          printk("    dar_lo  = 0x%08x\n", link_table[sg_len-1].dar_low);
          printk("    dar_hi  = 0x%08x\n", link_table[sg_len-1].dar_high);*/
          printk("---DDR SG0 delay time: %dus-----\n", delay_time);
          udelay(delay_time);
          delay_time *= 2;
          if(i_v++ == 32)
          {
               printk("------------DDR SG0 wrong 32 times!!!--------------\n");
               break;
          }

          src = edma_chan->sconfig.srcAddr;
          dst = edma_chan->sconfig.dstAddr;
          ctl = DMA_CB;
          for_each_sg(sgl, sg, sg_len, i)
          {
              ljmkASSERT(i * ljmSIZEOF(edma_desc_ll) < (edma_dev->single_chan ? EDMA_LT_SIZE*4 : EDMA_LT_SIZE));

              avail = sg_dma_len(sg);
              addr = sg_dma_address(sg);

              link_table[i].control = ctl;
              link_table[i].size = avail;

              if (edma_chan->direction == ljeDMA_TX_DIRECTION_MEM_TO_DEV)
              {
                  /* mem -> dev */
                  link_table[i].sar_low = QWORD_LO(addr);
                  link_table[i].sar_high = QWORD_HI(addr);
                  link_table[i].dar_low = QWORD_LO(dst);
                  link_table[i].dar_high = QWORD_HI(dst);
                  dst += avail;
              }
              else
              {
                  /* dev -> mem */
                  link_table[i].sar_low = QWORD_LO(src);
                  link_table[i].sar_high = QWORD_HI(src);
                  link_table[i].dar_low = QWORD_LO(addr);
                  link_table[i].dar_high = QWORD_HI(addr);
                  src += avail;
              }
              total_len += avail;
          }

          if (edma_dev->interrupt_enable)
          {
              ctl |= (DMA_LIE | DMA_RIE);
              link_table[sg_len - 1].control = ctl;
          }
          else
          {
              ctl |= DMA_LIE;
              link_table[sg_len - 1].control = ctl;
          }

          edma_llp = (edma_desc_llp*)(&link_table[sg_len]);
          edma_llp->control = DMA_LLP | DMA_TCB;
          edma_llp->reserved = 0;
          edma_llp->llp_high = QWORD_HI(desc->lle);
          edma_llp->llp_low = QWORD_LO(desc->lle);

          dst -= avail;
     }
    }else{
    src -= avail;
    while((link_table[sg_len-1].size != avail) || (link_table[sg_len-1].sar_high != QWORD_HI(src)) || (link_table[sg_len-1].dar_high != QWORD_HI(addr)) )
    {

         /*printk("wait DDR SG1    control = 0x%08x\n", link_table[sg_len-1].control);
         printk("    size    = 0x%08x\n", link_table[sg_len-1].size);
         printk("    sar_lo  = 0x%08x\n", link_table[sg_len-1].sar_low);
         printk("    sar_hi  = 0x%08x\n", link_table[sg_len-1].sar_high);
         printk("    dar_lo  = 0x%08x\n", link_table[sg_len-1].dar_low);
         printk("    dar_hi  = 0x%08x\n", link_table[sg_len-1].dar_high);*/
         printk("---DDR SG1 delay time: %dus-----\n", delay_time);
         udelay(delay_time);
         delay_time *= 2;
         if(i_v++ == 32)
         {
              printk("------------DDR SG1 wrong 32 times!!!--------------\n");
              break;
         }

         src = edma_chan->sconfig.srcAddr;
         dst = edma_chan->sconfig.dstAddr;
         ctl = DMA_CB;
         for_each_sg(sgl, sg, sg_len, i)
         {
             ljmkASSERT(i * ljmSIZEOF(edma_desc_ll) < (edma_dev->single_chan ? EDMA_LT_SIZE*4 : EDMA_LT_SIZE));

             avail = sg_dma_len(sg);
             addr = sg_dma_address(sg);

             link_table[i].control = ctl;
             link_table[i].size = avail;

             if (edma_chan->direction == ljeDMA_TX_DIRECTION_MEM_TO_DEV)
             {
                 /* mem -> dev */
                 link_table[i].sar_low = QWORD_LO(addr);
                 link_table[i].sar_high = QWORD_HI(addr);
                 link_table[i].dar_low = QWORD_LO(dst);
                 link_table[i].dar_high = QWORD_HI(dst);
                 dst += avail;
             }
             else
             {
                 /* dev -> mem */
                 link_table[i].sar_low = QWORD_LO(src);
                 link_table[i].sar_high = QWORD_HI(src);
                 link_table[i].dar_low = QWORD_LO(addr);
                 link_table[i].dar_high = QWORD_HI(addr);
                 src += avail;
             }
             total_len += avail;
         }

         if (edma_dev->interrupt_enable)
         {
             ctl |= (DMA_LIE | DMA_RIE);
             link_table[sg_len - 1].control = ctl;
         }
         else
         {
             ctl |= DMA_LIE;
             link_table[sg_len - 1].control = ctl;
         }

         edma_llp = (edma_desc_llp*)(&link_table[sg_len]);
         edma_llp->control = DMA_LLP | DMA_TCB;
         edma_llp->reserved = 0;
         edma_llp->llp_high = QWORD_HI(desc->lle);
         edma_llp->llp_low = QWORD_LO(desc->lle);

         src -= avail;
    }
    }

}

/*******todo: how to get hdma channel count???********/
ljeSTATUS hdma_dev_channels_init(edma_device *edma_dev)
{
    int i;

    if (!edma_dev)
    {
        return ljvSTATUS_INVALID_ARGUMENT;
    }

    /* get read/write channels count from reg */
    if (edma_dev->single_chan) {
        edma_dev->rchan_count = 2;
        edma_dev->wchan_count = 2;
    } else {
        //reg = edma_read(edma_dev->reg_base, DMA_CHANNEL_CTRL_REG);
        edma_dev->rchan_count = HDMA_V0_MAX_NR_CH;
        edma_dev->wchan_count = HDMA_V0_MAX_NR_CH;
    }

    edma_dev->wchan = kcalloc(edma_dev->wchan_count, sizeof(edma_channel), GFP_KERNEL);
    if (!edma_dev->wchan)
    {
        return ljvSTATUS_OUT_OF_MEMORY;
    }

    edma_dev->rchan = kcalloc(edma_dev->rchan_count, sizeof(edma_channel), GFP_KERNEL);
    if (!edma_dev->rchan)
    {
        kfree(edma_dev->wchan);
        return ljvSTATUS_OUT_OF_MEMORY;
    }

    /* write channels */
    for (i = 0; i < edma_dev->wchan_count; i++)
    {
        edma_channel *chan = &edma_dev->wchan[i];

        chan->chan.device = &edma_dev->slave_wdma;
        ljkDMA_IdentifierInit(&chan->chan);

        list_add_tail(&chan->chan.devNode, &edma_dev->slave_wdma.chanHead);

        spin_lock_init(&chan->lock);
        INIT_LIST_HEAD(&chan->active_list);
        INIT_LIST_HEAD(&chan->pending_list);

        chan->reg_base = edma_dev->reg_base_2;
        chan->status = ljeDMA_STATUS_COMPLETE;
        chan->direction = ljeDMA_TX_DIRECTION_DEV_TO_MEM;

        chan->desc = kzalloc(sizeof(edma_descriptor), GFP_NOWAIT);
    }

    /* read channels */
    for (i = 0; i < edma_dev->rchan_count; i++)
    {
        edma_channel *chan = &edma_dev->rchan[i];

        chan->chan.device = &edma_dev->slave_rdma;
        ljkDMA_IdentifierInit(&chan->chan);

        list_add_tail(&chan->chan.devNode, &edma_dev->slave_rdma.chanHead);

        spin_lock_init(&chan->lock);
        INIT_LIST_HEAD(&chan->active_list);
        INIT_LIST_HEAD(&chan->pending_list);

        chan->reg_base = edma_dev->reg_base_2;
        chan->status = ljeDMA_STATUS_COMPLETE;
        chan->direction = ljeDMA_TX_DIRECTION_MEM_TO_DEV;

        chan->desc = kzalloc(sizeof(edma_descriptor), GFP_NOWAIT);
    }

    return ljvSTATUS_OK;
}




void hdma_dev_init(iomem_t reg_base)
{
#if 0
    edma_write(reg_base, DMA_WRITE_CHANNEL_ARB_WEIGHT_LOW_REG, 0x00008421);
    edma_write(reg_base, DMA_WRITE_CHANNEL_ARB_WEIGHT_HIGH_REG, 0x00008421);
    edma_write(reg_base, DMA_READ_CHANNEL_ARB_WEIGHT_LOW_REG, 0x00008421);
    edma_write(reg_base, DMA_READ_CHANNEL_ARB_WEIGHT_HIGH_REG, 0x00008421);
#endif
    /* init all write channel interrupt register */
    /* mask all write interrupt */

#if 0   //fq
    edma_write(reg_base, DMA_WRITE_INT_MASK_REG, 0x0);
    /* clear all write interrupt */
    edma_write(reg_base, DMA_WRITE_INT_CLEAR_REG, 0x0);
#endif

#if 0
    edma_write(reg_base, DMA_WRITE_DONE_IMWR_LOW_REG, 0x00000000);
    edma_write(reg_base, DMA_WRITE_DONE_IMWR_HIGH_REG, 0x00000000);
    edma_write(reg_base, DMA_WRITE_ABORT_IMWR_LOW_REG, 0x00000000);
    edma_write(reg_base, DMA_WRITE_ABORT_IMWR_HIGH_REG, 0x00000000);

    edma_write(reg_base, DMA_WRITE_CH01_IMWR_DATA_REG, 0x00000000);
    edma_write(reg_base, DMA_WRITE_CH23_IMWR_DATA_REG, 0x00000000);
    edma_write(reg_base, DMA_WRITE_CH45_IMWR_DATA_REG, 0x00000000);
    edma_write(reg_base, DMA_WRITE_CH67_IMWR_DATA_REG, 0x00000000);

    /* unable all write link abort error */
    edma_write(reg_base, DMA_WRITE_LINKED_LIST_ERR_EN_REG, 0x00000000);

    /* init all read channel interrupt register */
    /* mask all read interrupt */

    edma_write(reg_base, DMA_READ_INT_MASK_REG, 0x0);
    /* clear all write interrupt */
    edma_write(reg_base, DMA_READ_INT_CLEAR_REG, 0x0);
#endif

#if 0
    edma_write(reg_base, DMA_READ_DONE_IMWR_LOW_REG, 0x00000000);
    edma_write(reg_base, DMA_READ_DONE_IMWR_HIGH_REG, 0x00000000);
    edma_write(reg_base, DMA_READ_ABORT_IMWR_LOW_REG, 0x00000000);
    edma_write(reg_base, DMA_READ_ABORT_IMWR_HIGH_REG, 0x00000000);

    edma_write(reg_base/*pdev*/, DMA_READ_CH01_IMWR_DATA_REG, 0x00000000);
    edma_write(reg_base/*pdev*/, DMA_READ_CH23_IMWR_DATA_REG, 0x00000000);
    edma_write(reg_base/*pdev*/, DMA_READ_CH45_IMWR_DATA_REG, 0x00000000);
    edma_write(reg_base/*pdev*/, DMA_READ_CH67_IMWR_DATA_REG, 0x00000000);
    /* unable all read link abort error */
    edma_write(reg_base, DMA_READ_LINKED_LIST_ERR_EN_REG, 0x00000000);
#endif
}

int threadHdma(void *ctxt)
{
    edma_device *edma_dev = (edma_device *)ctxt;
    uint32_t wr_status, rd_status;
    int i;
    int int_status = 0;

  for (;;)
  {
        int down;

        down = down_interruptible(&sema_edma);
        if (down && down != -EINTR)
        {
            return down;
        }

        if (unlikely(EdmaKillThread))
        {
            /* The daemon exits. */
            while (!kthread_should_stop())
            {
                ljkOS_Delay(NULL, 1);
            }

            return 0;
        }

        /* get rd and wr int status */
    int_status = LJM_MISC_R32_BASE(edma_dev->reg_base_2, MISC_INTR_SRC);
    //wr_status = int_status&((1<<edma_dev->wchan_count) - 1);
    //rd_status = (int_status&(((1<<edma_dev->wchan_count) - 1)<<edma_dev->wchan_count))>>edma_dev->wchan_count;
    wr_status = int_status&0xff;
    rd_status = (int_status&0xff00)>>8;
#if EDMA_DEBUG
    printk("-------ws:0x%x; rs:0x%x-------\n", wr_status, rd_status);
#endif
    /* it is write channel int */
    if (wr_status)
    {
        for (i = 0; i < edma_dev->wchan_count; i++)
        {
            /*if (wr_status & DMA_ABORT_INT(i))
            {
                edma_dev_handle_error(&edma_dev->wchan[i]);
            }*/
            if(wr_status & (0x1<<i))
            {
              //printk("\n--------hdma_dev_tasklet----hdma_dev_handle_done_wr---\n");
                hdma_dev_handle_done(&edma_dev->wchan[i]);
            }
        }
    }

    /* it is read channel int */
    if (rd_status)
    {
        for (i = 0; i < edma_dev->rchan_count; i++)
        {
            /*if (rd_status & DMA_ABORT_INT(i))
            {
                edma_dev_handle_error(&edma_dev->rchan[i]);
            }*/
            if(rd_status & (0x1<<i))
            {
              //printk("\n--------hdma_dev_tasklet----hdma_dev_handle_done_rd---\n");
                hdma_dev_handle_done(&edma_dev->rchan[i]);
            }
        }
    }

    /* unmask all read interrupt */
    //edma_write(edma_dev->reg_base, DMA_READ_INT_MASK_REG, 0);
    /* unmask all write interrupt */
    //edma_write(edma_dev->reg_base, DMA_WRITE_INT_MASK_REG, 0);
    /*DMA_SINGLE_CHANNEL ok*/
    LJM_MISC_W32_BASE(edma_dev->reg_base_2, MISC_INTR_EN, LJM_MISC_R32_BASE(edma_dev->reg_base_2, MISC_INTR_EN) | m_INTR_EN_EDMA);
  }

}

/**
 * hdma_dev_interrupt_handler - dma interrupt handler.
 *
 * @irq: irq number
 * @dev_id: driver private at int registry
 *
 * Return: always 0.
 *
 */
irqreturn_t hdma_dev_interrupt_handler(int irq, void *dev_id)
{
    edma_device *edma_dev = dev_id;
    //uint32_t wr_status, rd_status;
    //unsigned long flags;
    //int int_status = 0;
    //printk("%s(%d)\n", __FUNCTION__, __LINE__);
    //spin_lock_irqsave(&edma_dev->irq_lock, flags);
    spin_lock(&edma_dev->irq_lock);
#if 0
    /* get rd and wr int status */
    int_status = LJM_MISC_R32_BASE(edma_dev->reg_base_2, MISC_INTR_SRC);
    wr_status = int_status&((1<<edma_dev->wchan_count) - 1);
    rd_status = (int_status&(((1<<edma_dev->wchan_count) - 1)<<edma_dev->wchan_count))>>edma_dev->wchan_count;

    if (!(wr_status | rd_status))
    {
        //spin_unlock_irqrestore(&edma_dev->irq_lock, flags);
        spin_unlock(&edma_dev->irq_lock);
        return IRQ_NONE;
    }
#endif
    /* mask all read interrupt */
    //edma_write(edma_dev->reg_base, DMA_READ_INT_MASK_REG, 0x00FF00FF);
    /* mask all write interrupt */
    //edma_write(edma_dev->reg_base, DMA_WRITE_INT_MASK_REG, 0x00FF00FF);
    /*DMA_SINGLE_CHANNEL ok*/
    LJM_MISC_W32_BASE(edma_dev->reg_base_2, MISC_INTR_EN, LJM_MISC_R32_BASE(edma_dev->reg_base_2, MISC_INTR_EN) & ~m_INTR_EN_EDMA);

    //spin_unlock_irqrestore(&edma_dev->irq_lock, flags);
    spin_unlock(&edma_dev->irq_lock);

    //tasklet_schedule(&edma_dev->tasklet);
    up(&sema_edma);

    return IRQ_HANDLED;
}

/**
 * hdma_dev_handle_done - dma done interrupt handler
 *
 * @edma_chan: edma_channel structure pointer
 *
 * Return: none
 *
 */
void hdma_dev_handle_done(edma_channel *edma_chan)
{
    edma_descriptor *desc;
    struct list_head *active;
    unsigned long flags;

    spin_lock_irqsave(&edma_chan->lock, flags);
    /* clear done bit */
    if (edma_chan->direction == ljeDMA_TX_DIRECTION_MEM_TO_DEV)
    {
        edma_write(edma_chan->reg_base, HDMA_INT_CLEAR_OFF_RDCH(edma_chan->chan.id), 0x1);
    }
    else if (edma_chan->direction == ljeDMA_TX_DIRECTION_DEV_TO_MEM)
    {
        edma_write(edma_chan->reg_base, HDMA_INT_CLEAR_OFF_WRCH(edma_chan->chan.id), 0x1);
    }

    /* get current active */
    desc = edma_get_first_active(edma_chan);
    active = edma_chan->current_active;
    if (edma_desc_entry(active) != desc)
    {
        printk("hdma: it shouldn't be here, interrupt lost?\n");
    }

    spin_unlock_irqrestore(&edma_chan->lock, flags);
    /* desc has done */
    hdma_complete_all(edma_chan);
}

/**
 * edma_dev_do_transaction - issue dma into hw
 *
 * @edma_chann: pcie_edma_chann structure pointer
 * @desc: dma pcie_edma_descriptor structure pointer
 *
 * Return: none
 *
 */
static void hdma_dev_do_transaction(edma_channel *edma_chan, edma_descriptor *desc)
{
    edma_device *edma_dev = to_edma_dev(edma_chan->chan.device);
    edma_transfer_info info;
    edma_status status;

    if (hdma_dev_get_channel_status(
            edma_chan->reg_base,
            edma_chan->chan.id,
            edma_chan->direction) == HDMA_CS_RUNNING)
    {
        printk("dma %s chan %d is busy, do transaction fail\n",
                    edma_chan->chan.device->name, edma_chan->chan.id);
        return;
    }

    info.reg_base = edma_chan->reg_base;
    info.channel = edma_chan->chan.id;
    info.cpu_mode = desc->cpu_mode;
    info.check_mode = desc->check_mode;
    info.transfer_mode = desc->transfer_mode;
    info.src_addr = desc->src_addr;
    info.dst_addr = desc->dst_addr;
    info.transfer_size = desc->total_len;
    info.lle = desc->lle;
    info.timeout = desc->timeout;
    info.msi_addr = edma_dev->msi_addr;
    info.msi_data = edma_dev->msi_data;

    if (edma_chan->direction == ljeDMA_TX_DIRECTION_DEV_TO_MEM)
    {
        /* write */
        status = hdma_dev_write_transfer(&info);
    }
    else if (edma_chan->direction == ljeDMA_TX_DIRECTION_MEM_TO_DEV)
    {
        /* read */
        status = hdma_dev_read_transfer(&info);
    }
}

/**
 * edma_dev_alloc_chan_resources - init the channel resource
 *
 * @chann: dma_chann structure pointer
 *
 * Return: none
 *
 */
ljeSTATUS hdma_dev_alloc_chan_resources(ljsDMA_CHANNEL *chan)
{
    edma_channel *edma_chan = to_edma_chan(chan);
    edma_device *edma_dev = to_edma_dev(chan->device);
    unsigned long flags;

    if (hdma_dev_get_channel_status(
            edma_dev->reg_base_2,
            edma_chan->chan.id,
            edma_chan->direction) == DMA_CS_RUNNING)
    {
        printk("dma %s chan %d is busy.\n", chan->device->name, edma_chan->chan.id);
        return ljvSTATUS_DEVICE;
    }

    ljkDMA_IdentifierInit(chan);

    spin_lock_irqsave(&edma_chan->lock, flags);
    if (edma_chan->direction == ljeDMA_TX_DIRECTION_DEV_TO_MEM)
    {
        if (!edma_dev->wchan_in_use)
        {
            hdma_dev_write_engine_on(edma_dev->reg_base_2, edma_chan->chan.id);
        }
        edma_dev->wchan_in_use |= (0x1 << edma_chan->chan.id);
    }
    else if (edma_chan->direction == ljeDMA_TX_DIRECTION_MEM_TO_DEV)
    {
        if(!edma_dev->rchan_in_use)
        {
            hdma_dev_read_engine_on(edma_dev->reg_base_2, edma_chan->chan.id);
        }
        edma_dev->rchan_in_use |= (0x1 << edma_chan->chan.id);
    }
    spin_unlock_irqrestore(&edma_chan->lock, flags);

    return ljvSTATUS_OK;
}

/**
 * edma_dev_free_chan_resources - just disable channel irq, the channel resource
 * will destoried in complete functions.
 *
 * @chan: ljsDMA_CHANNEL structure pointer
 *
 * Return: none
 *
 */
void hdma_dev_free_chan_resources(ljsDMA_CHANNEL *chan)
{
    edma_channel *edma_chan = to_edma_chan(chan);
    edma_device *edma_dev = to_edma_dev(chan->device);
    unsigned long flags;
    //uint32_t reg;

    BUG_ON(!list_empty(&edma_chan->active_list));
    BUG_ON(!list_empty(&edma_chan->pending_list));
    BUG_ON(edma_chan->descs_allocated);

    spin_lock_irqsave(&edma_chan->lock, flags);
    if (edma_chan->direction == ljeDMA_TX_DIRECTION_MEM_TO_DEV)
    {
        /* disable channel interrupt */
        //reg = edma_read(edma_dev->reg_base, DMA_READ_INT_MASK_REG);
        //edma_write(edma_dev->reg_base, DMA_READ_INT_MASK_REG, reg | DMA_DONE_INT(edma_chan->chan.id));
        //edma_write(edma_dev->reg_base, DMA_READ_INT_MASK_REG, reg | DMA_ABORT_INT(edma_chan->chan.id));

        edma_dev->rchan_in_use &= ~(0x1 << edma_chan->chan.id);
        if (!edma_dev->rchan_in_use)
        {
            hdma_dev_read_engine_off(edma_dev->reg_base_2, edma_chan->chan.id);
        }
    }

    if (edma_chan->direction == ljeDMA_TX_DIRECTION_DEV_TO_MEM)
    {
        /* disable channel interrupt */
        //reg = edma_read(edma_dev->reg_base, DMA_WRITE_INT_MASK_REG);
        //edma_write(edma_dev->reg_base, DMA_WRITE_INT_MASK_REG, reg | DMA_DONE_INT(edma_chan->chan.id));
        //edma_write(edma_dev->reg_base, DMA_WRITE_INT_MASK_REG, reg | DMA_ABORT_INT(edma_chan->chan.id));

        edma_dev->wchan_in_use &= ~(0x1 << edma_chan->chan.id);
        if (!edma_dev->wchan_in_use)
        {
            hdma_dev_write_engine_off(edma_dev->reg_base_2, edma_chan->chan.id);
        }
    }
    spin_unlock_irqrestore(&edma_chan->lock, flags);

}

/**
 * edma_dev_prep_slave_sg - alloc dma desctriptors for every sg node.
 *
 * @chann: dma_chann structure pointer
 * @sgl: scatterlist structure pointer
 * @sg_len: the dma data len
 * @direction: transfer direction
 * @flags: other info
 *
 * Return: ljsDMA_TX_DESCRIPTOR pointer.
 *
 */
ljsDMA_TX_DESCRIPTOR* hdma_dev_prep_slave_sg(
    ljsDMA_CHANNEL *chan,
    struct scatterlist *sgl,
    size_t sg_len,
    ljeDMA_TX_DIRECTION direction,
    unsigned int flags,
    void *context
    )
{
    edma_channel *edma_chan;
    edma_device *edma_dev;
    edma_descriptor *desc;

    if(!chan || !sgl || sg_len == 0)
    {
        return NULL;
    }

    if (!ljkDMAENGINE_IsSlaveDirection(direction))
    {
        printk("invalid dma direction.\n");
        return NULL;
    }

    edma_chan = to_edma_chan(chan);
    edma_dev = to_edma_dev(chan->device);

    if(sg_len > (edma_dev->single_chan ? EDMA_LT_SIZE*4 : EDMA_LT_SIZE)/ljmSIZEOF(edma_desc_ll))
    {
        printk("hdma failed: sg_len(%ld) over the max size(%ld).\n", sg_len, (edma_dev->single_chan ? EDMA_LT_SIZE*4 : EDMA_LT_SIZE)/ljmSIZEOF(edma_desc_ll));
        return NULL;
    }

    if (edma_chan->direction != direction)
    {
        printk("chan %d do not support direction %d\n", edma_chan->chan.id , direction);
        return NULL;
    }

    desc = edma_descriptor_alloc(edma_chan);
    if (!desc)
    {
        printk("fail to alloc descriptor.\n");
        return NULL;
    }

    desc->cpu_mode = edma_dev->interrupt_enable ? EDMA_CPU_MODE_REMOTE : EDMA_CPU_MODE_LOCAL;
    desc->check_mode = edma_dev->interrupt_enable ? EDMA_CHECK_MODE_INTERRUPT : EDMA_CHECK_MODE_POLLING;
    desc->transfer_mode = EDMA_TRANSFER_MODE_LINK;

    desc->src_addr = 0;
    desc->dst_addr = 0;
    desc->total_len = 0;
    desc->residue = 0;
    desc->timeout = 5000;
    desc->is_done = false;

    hdma_dev_fill_link_table(chan, desc, sgl, sg_len);

    return &desc->txd;
}

/**
 * edma_dev_prep_memcpy - alloc dma desctriptors for normal mode.
 *
 * @chan: ljsDMA_CHANNEL structure pointer
 * @Src: the src bus address
 * @Dst: the dst bus address
 * @Len: the dma data len
 * @flags: other info
 *
 * Return: ljsDMA_TX_DESCRIPTOR pointer.
 *
 */
ljsDMA_TX_DESCRIPTOR* hdma_dev_prep_dma_memcpy(
    ljsDMA_CHANNEL *Chan,
    dma_addr_t Src,
    dma_addr_t Dst,
    ljtSIZE_T Len,
    ljtUINT32 Flag
    )
{
    edma_channel *edma_chan;
    edma_device *edma_dev;
    edma_descriptor *desc;
    edma_desc_ll *link_table = ljvNULL;
    //edma_desc_llp *edma_llp = ljvNULL;
    uint32_t ctl;
    uint32_t link_t[16];
    uint32_t i_v = 0;

    if(!Chan || Len == 0)
    {
        return NULL;
    }

    edma_chan = to_edma_chan(Chan);
    edma_dev = to_edma_dev(Chan->device);

    desc = edma_descriptor_alloc(edma_chan);
    if (!desc)
    {
        printk("fail to alloc descriptor.\n");
        return NULL;
    }

    desc->cpu_mode = edma_dev->interrupt_enable ? EDMA_CPU_MODE_REMOTE : EDMA_CPU_MODE_LOCAL;
    desc->check_mode = edma_dev->interrupt_enable ? EDMA_CHECK_MODE_INTERRUPT : EDMA_CHECK_MODE_POLLING;
    desc->transfer_mode = EDMA_TRANSFER_MODE_LINK;

    desc->src_addr = 0;
    desc->dst_addr = 0;
    desc->total_len = 0;
    desc->residue = 0;
    desc->timeout = 5000;
    desc->is_done = false;

    /* use linked mode to replace normal mode */
    if (edma_chan->direction == ljeDMA_TX_DIRECTION_MEM_TO_DEV)
    {
        desc->lle = (uint64_t)LINKTABLE_OFFSET(
                        edma_dev->plink_table_base, edma_chan->chan.id);

        link_table = (edma_desc_ll*)LINKTABLE_OFFSET(
                        edma_dev->vlink_table_base, edma_chan->chan.id);
    }
    else if (edma_chan->direction == ljeDMA_TX_DIRECTION_DEV_TO_MEM)
    {
        desc->lle = (uint64_t)LINKTABLE_OFFSET(
                        edma_dev->plink_table_base,
                        edma_dev->rchan_count + edma_chan->chan.id);

        link_table = (edma_desc_ll*)LINKTABLE_OFFSET(
                        edma_dev->vlink_table_base,
                        edma_dev->rchan_count + edma_chan->chan.id);
    }

    //ctl = DMA_CB;
    ctl = 0;
#if 0/* abc: 256bit align write ddr */
    link_table[0].control = ctl;
    link_table[0].size = Len;
    link_table[0].sar_low = QWORD_LO(Src);
    link_table[0].sar_high = QWORD_HI(Src);
    link_table[0].dar_low = QWORD_LO(Dst);
    link_table[0].dar_high = QWORD_HI(Dst);

    if (edma_dev->interrupt_enable)
    {
        ctl |= (DMA_LIE | DMA_RIE);
        link_table[0].control = ctl;
    }
    else
    {
        ctl |= DMA_LIE;
        link_table[0].control = ctl;
    }

    edma_llp = (edma_desc_llp*)(&link_table[1]);
    edma_llp->control = DMA_LLP | DMA_TCB;
    edma_llp->reserved = 0;
    edma_llp->llp_high = QWORD_HI(desc->lle);
    edma_llp->llp_low = QWORD_LO(desc->lle);
#else
    link_t[0] = ctl;
    link_t[1] = Len;
    link_t[2] = QWORD_LO(Src);
    link_t[3] = QWORD_HI(Src);
    link_t[4] = QWORD_LO(Dst);
    link_t[5] = QWORD_HI(Dst);

    if (edma_dev->interrupt_enable)
    {
//        ctl |= (DMA_LIE | DMA_RIE);
//        link_t[0] = ctl;
    }
    else
    {
        ctl |= DMA_LIE;
        link_t[0] = ctl;
    }

    //link_t[6] = DMA_LLP | DMA_TCB;
    link_t[6] = DMA_LLP | DMA_CB;
    link_t[7] = 0;
    link_t[8] = QWORD_LO(desc->lle);
    link_t[9] = QWORD_HI(desc->lle);
    link_t[10] = 0;
    link_t[11] = 0;
    link_t[12] = 0;
    link_t[13] = 0;
    link_t[14] = 0;
    link_t[15] = 0;

    memcpy(&link_table[0], &link_t[0], 64);
    //memcpy((uint32_t *)(&link_table[0])+8, &link_t[8], 32);
#endif
    /* insure the config has written into link table */
    //link_table[2].control = 0x1;
    //while(!link_table[2].control);
    //memset(&link_table[2], 0, sizeof(link_table[2]));
    //while(link_table[2].control);

    desc->total_len = Len;
    desc->residue = Len;

#if EDMA_DEBUG
    printk("hdma_dev_prep_dma_memcpy : link table [%d]@0x%08llx:\n", 0, desc->lle);
    printk("    control = 0x%08x\n", link_table[0].control);
    printk("    size    = 0x%08x\n", link_table[0].size);
    printk("    sar_lo  = 0x%08x\n", link_table[0].sar_low);
    printk("    sar_hi  = 0x%08x\n", link_table[0].sar_high);
    printk("    dar_lo  = 0x%08x\n", link_table[0].dar_low);
    printk("    dar_hi  = 0x%08x\n", link_table[0].dar_high);

    printk("    llp_control = 0x%08x\n", link_table[1].control);
    printk("    llp_reserved    = 0x%08x\n", link_table[1].size);
    printk("    llp_low  = 0x%08x\n", link_table[1].sar_low);
    printk("    llp_high  = 0x%08x\n", link_table[1].sar_high);
#endif

#if (defined CONFIG_LOONGARCH ||defined CONFIG_SW64)
    udelay(delay_time);
#endif
   while((link_table[0].size != Len) || (link_table[0].sar_high != QWORD_HI(Src)) || (link_table[0].dar_high != QWORD_HI(Dst)) )
   {
    ctl = DMA_CB;
    link_table[0].control = ctl;
    link_table[0].size = Len;
    link_table[0].sar_low = QWORD_LO(Src);
    link_table[0].sar_high = QWORD_HI(Src);
    link_table[0].dar_low = QWORD_LO(Dst);
    link_table[0].dar_high = QWORD_HI(Dst);

    if (edma_dev->interrupt_enable)
    {
        ctl |= (DMA_LIE | DMA_RIE);
        link_table[0].control = ctl;
    }
    else
    {
        ctl |= DMA_LIE;
        link_table[0].control = ctl;
    }
    printk("---DDR 0 delay time: %dus-----\n", delay_time);
    udelay(delay_time);
    delay_time *= 2;
     /*printk(" WAIT DDR 0   control = 0x%08x\n", link_table[0].control);
     printk("    size    = 0x%08x\n", link_table[0].size);
     printk("    sar_lo  = 0x%08x\n", link_table[0].sar_low);
     printk("    sar_hi  = 0x%08x\n", link_table[0].sar_high);
     printk("    dar_lo  = 0x%08x\n", link_table[0].dar_low);
     printk("    dar_hi  = 0x%08x\n", link_table[0].dar_high);*/

    if(i_v++ == 32)
    {
         printk("------------DDR 0 wrong 32 times!!!--------------\n");
         break;
    }
   }


    return &desc->txd;
}

/**
 * hdma_do_first_queued - queue the next pending dma task, get the first
 * dma desc list in the active list.
 *
 * @edma_chann: pcie_edma_chann structure pointer
 *
 * Return: none
 *
 */
void hdma_do_first_queued(edma_channel *edma_chan)
{
    edma_descriptor *desc;

    if (list_empty(&edma_chan->pending_list))
    {
        return;
    }

    /* move next pending to active */
    list_move_tail(edma_chan->pending_list.next, &edma_chan->active_list);

    desc = edma_get_first_active(edma_chan);
    if (!desc)
    {
        return;
    }
    edma_chan->current_active = &desc->desc_node;

    hdma_dev_do_transaction(edma_chan, desc);
}

/**
 * hdma_complete_all - queue next pending transaction and free completed
 * resource
 *
 * @edma_chan: pcie_edma_chann structure pointer
 *
 * Return: none
 *
 */
void hdma_complete_all(edma_channel *edma_chan)
{
    edma_status status;
    unsigned long flags;
    LIST_HEAD(list);

    spin_lock_irqsave(&edma_chan->lock, flags);

    if (hdma_dev_get_channel_status(
        edma_chan->reg_base, edma_chan->chan.id, edma_chan->direction) == HDMA_CS_RUNNING)
    {
        printk("chan %d is no idle\n", edma_chan->chan.id);
        status = hdma_dev_transfer_stop(edma_chan->reg_base, edma_chan->chan.id, edma_chan->direction);
        if (status == EDMA_ERROR)
        {
            printk("stop chan %d fail\n", edma_chan->chan.id);
        }
    }

    if (list_empty(&edma_chan->active_list))
    {
        spin_unlock_irqrestore(&edma_chan->lock, flags);
        return;
    }

    /* remove finished transaction desc from active list */
    list_move(edma_chan->active_list.next, &list);

    hdma_do_first_queued(edma_chan);

    spin_unlock_irqrestore(&edma_chan->lock, flags);

    edma_free_descriptor_list(edma_chan, &list);
}

/**
 * pcie_hdma_terminate_all - stop dma transaction on the channel.
 *
 * @chann: dma_chann structure pointer
 *
 * Return: always 0.
 *
 */
ljeSTATUS hdma_terminate_all(ljsDMA_CHANNEL *chan)
{
    edma_channel *edma_chan = to_edma_chan(chan);
    unsigned long flags;
    LIST_HEAD(list);

    int size = 0;

    if(edma_chan->direction == ljeDMA_TX_DIRECTION_MEM_TO_DEV)
        size = edma_read(edma_chan->reg_base, HDMA_XFERSIZE_OFF_RDCH(chan->id));
    else if(edma_chan->direction == ljeDMA_TX_DIRECTION_DEV_TO_MEM)
        size = edma_read(edma_chan->reg_base, HDMA_XFERSIZE_OFF_WRCH(chan->id));

    if( size !=0 )
        printk("------size = %d--maybe we need to add hdma wait time!!!-----\n", size);

    hdma_dev_transfer_stop(edma_chan->reg_base, edma_chan->chan.id, edma_chan->direction);

    spin_lock_irqsave(&edma_chan->lock, flags);
    list_splice_init(&edma_chan->pending_list, &list);
    list_splice_init(&edma_chan->active_list, &list);
    spin_unlock_irqrestore(&edma_chan->lock, flags);

    if (!list_empty(&list))
    {
        edma_free_descriptor_list(edma_chan, &list);
    }

    return ljvSTATUS_OK;
}

/**
 * edma_tx_status - get current channel status.
 *
 * @chan: dma_chann structure pointer
 * @identifier: dma_identifier_t retured from tx_submit()
 * @state: ljsDMA_TX_STATE structure pointer
 *
 * Return: ljeDMA_STATUS.
 *
 */
ljeDMA_STATUS hdma_tx_status(
    ljsDMA_CHANNEL *chan,
    ljdDMA_TX_IDENTIFIER identifier,
    ljsDMA_TX_STATE *state
    )
{
    edma_channel *edma_chan = to_edma_chan(chan);
    unsigned long flags;
    ljeDMA_STATUS ret;
    ljtUINT32 pid;
    ljkOS_GetProcessID(&pid);


    ret = ljkDMA_IdentifierStatus(chan, identifier, state);
    if (ret == ljeDMA_STATUS_COMPLETE)
    {

        return ret;
    }

    spin_lock_irqsave(&edma_chan->lock, flags);
    hdma_do_first_queued(edma_chan);
    spin_unlock_irqrestore(&edma_chan->lock, flags);

    ret = ljkDMA_IdentifierStatus(chan, identifier, state);
    if (ret == ljeDMA_STATUS_COMPLETE)
    {

        return ret;
    }

    ljkDMA_SetResidue(state, edma_get_residue(edma_chan, identifier));

    return ret;
}

/**
 * edma_issue_pending - issue channel transction from active_list
 *
 * @chann: dma_chann structure pointer
 *
 * Return: none
 *
 */
void hdma_issue_pending(ljsDMA_CHANNEL *chan)
{
    edma_channel *edma_chan = to_edma_chan(chan);
    unsigned long flags;

    spin_lock_irqsave(&edma_chan->lock, flags);
    if (list_empty(&edma_chan->active_list))
    {
        hdma_do_first_queued(edma_chan);
    }
    spin_unlock_irqrestore(&edma_chan->lock, flags);
}

