/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#include "lj_hal.h"
#include "lj_hal_base.h"

#if ljdLINUX_SYNC_FILE

#include <linux/kernel.h>
#include <linux/file.h>
#include <linux/fs.h>
#include <linux/miscdevice.h>
#include <linux/module.h>
#include <linux/syscalls.h>
#include <linux/uaccess.h>
#include <linux/slab.h>

#include "lj_hal_kernel_sync.h"
#include "lj_hal_kernel_linux.h"

#ifndef CONFIG_SYNC_FILE

static struct sync_pt *ljm_sync_pt_dup(struct sync_pt *sync_pt)
{
    ljeSTATUS status;
    struct ljm_sync_pt *pt;
    struct ljm_sync_pt *src;
    struct ljm_sync_timeline *obj;

    src = (struct ljm_sync_pt *)sync_pt;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 17, 0)
    obj = (struct ljm_sync_timeline *)sync_pt_parent(sync_pt);
# else
    obj = (struct ljm_sync_timeline *)sync_pt->parent;
# endif

    /* Create the new sync_pt. */
    pt = (struct ljm_sync_pt *)sync_pt_create(&obj->obj, sizeof(struct ljm_sync_pt));

    pt->stamp = src->stamp;

    /* Reference signal. */
    status = ljkOS_MapSignal(obj->os,
                             src->signal,
                             ljvNULL /* (ljtHANDLE) _GetProcessID() */,
                             &pt->signal);

    if (ljmIS_ERROR(status)) {
        sync_pt_free((struct sync_pt *)pt);
        return NULL;
    }

    return (struct sync_pt *)pt;
}

static int ljm_sync_pt_has_signaled(struct sync_pt *sync_pt)
{
    ljeSTATUS status;
    struct ljm_sync_pt *pt;
    struct ljm_sync_timeline *obj;

    pt  = (struct ljm_sync_pt *)sync_pt;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 17, 0)
    obj = (struct ljm_sync_timeline *)sync_pt_parent(sync_pt);
# else
    obj = (struct ljm_sync_timeline *)sync_pt->parent;
# endif

    status = _QuerySignal(obj->os, pt->signal);

    if (ljmIS_ERROR(status)) {
        /* Error. */
        return -1;
    }

    return (int)status;
}

static int ljm_sync_pt_compare(struct sync_pt *a, struct sync_pt *b)
{
    int ret;
    struct ljm_sync_pt *pt1 = (struct ljm_sync_pt *)a;
    struct ljm_sync_pt *pt2 = (struct ljm_sync_pt *)b;

    ret = (pt1->stamp < pt2->stamp) ?
          -1 : (pt1->stamp == pt2->stamp) ? 0 : 1;

    return ret;
}

static void ljm_sync_pt_free(struct sync_pt *sync_pt)
{
    struct ljm_sync_pt *pt;
    struct ljm_sync_timeline *obj;

    pt = (struct ljm_sync_pt *)sync_pt;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 17, 0)
    obj = (struct ljm_sync_timeline *)sync_pt_parent(sync_pt);
# else
    obj = (struct ljm_sync_timeline *)sync_pt->parent;
# endif

    ljkOS_DestroySignal(obj->os, pt->signal);
}

static void ljm_timeline_value_str(struct sync_timeline *timeline, char *str, int size)
{
    struct ljm_sync_timeline *obj;

    obj = (struct ljm_sync_timeline *)timeline;
    snprintf(str, size, "stamp_%llu", obj->stamp);
}

static void ljm_pt_value_str(struct sync_pt *sync_pt, char *str, int size)
{
    struct ljm_sync_pt *pt;

    pt = (struct ljm_sync_pt *)sync_pt;
    snprintf(str, size, "signal_%lu@stamp_%llu", (unsigned long)pt->signal, pt->stamp);
}

static struct sync_timeline_ops ljm_timeline_ops = {
    .driver_name        = "ljm_gpu_sync",
    .dup                = ljm_sync_pt_dup,
    .has_signaled       = ljm_sync_pt_has_signaled,
    .compare            = ljm_sync_pt_compare,
    .free_pt            = ljm_sync_pt_free,
    .timeline_value_str = ljm_timeline_value_str,
    .pt_value_str       = ljm_pt_value_str,
};

struct ljm_sync_timeline *ljm_sync_timeline_create(const char *name, ljkOS os)
{
    struct ljm_sync_timeline *obj;

    obj = (struct ljm_sync_timeline *)sync_timeline_create(&ljm_timeline_ops,
                                                           sizeof(struct ljm_sync_timeline),
                                                           name);

    obj->os = os;
    obj->stamp = 0;

    return obj;
}

struct sync_pt *ljm_sync_pt_create(struct ljm_sync_timeline *obj, ljtSIGNAL Signal)
{
    ljeSTATUS status;
    struct ljm_sync_pt *pt;

    pt = (struct ljm_sync_pt *)sync_pt_create(&obj->obj, sizeof(struct ljm_sync_pt));

    pt->stamp = obj->stamp++;

    /* Dup signal. */
    status = ljkOS_MapSignal(obj->os, Signal,
                             ljvNULL /* (ljtHANDLE) _GetProcessID() */,
                             &pt->signal);

    if (ljmIS_ERROR(status)) {
        sync_pt_free((struct sync_pt *)pt);
        return NULL;
    }

    return (struct sync_pt *)pt;
}

# else

struct ljm_sync_timeline *ljm_sync_timeline_create(const char *name, ljkOS Os)
{
    struct ljm_sync_timeline *timeline;

    timeline = kmalloc(sizeof(*timeline), ljdNOWARN | GFP_KERNEL);

    if (!timeline)
        return NULL;

    strncpy(timeline->name, name, sizeof(timeline->name) - 1);
    timeline->context = dma_fence_context_alloc(1);
    atomic64_set(&timeline->seqno, 0);
    timeline->os = Os;

    return timeline;
}

void ljm_sync_timeline_destroy(struct ljm_sync_timeline *timeline)
{
    kfree(timeline);
}

static const char *ljm_fence_get_driver_name(struct dma_fence *fence)
{
    return "ljm_gpu_sync";
}

static const char *ljm_fence_get_timeline_name(struct dma_fence *fence)
{
    struct ljm_fence *f = (struct ljm_fence *)fence;

    return f->parent->name;
}

/* Same as fence_signaled. */
static inline bool __ljm_fence_signaled(struct dma_fence *fence)
{
    struct ljm_fence *f = (struct ljm_fence *)fence;
    struct ljm_sync_timeline *timeline = f->parent;
    ljeSTATUS status;

    status = _QuerySignal(timeline->os, f->signal);

    return (status == ljvSTATUS_TRUE) ? true : false;
}

static bool ljm_fence_enable_signaling(struct dma_fence *fence)
{
    /* fence is locked already. */
    return !__ljm_fence_signaled(fence);
}

static bool ljm_fence_signaled(struct dma_fence *fence)
{
    /* fence could be locked, could be not. */
    return __ljm_fence_signaled(fence);
}

static void ljm_fence_release(struct dma_fence *fence)
{
    struct ljm_fence *f = (struct ljm_fence *)fence;
    struct ljm_sync_timeline *timeline = f->parent;

    if (f->signal)
        ljkOS_DestroySignal(timeline->os, f->signal);

    kfree(fence);
}

static struct dma_fence_ops ljm_fence_ops = {
    .get_driver_name   = ljm_fence_get_driver_name,
    .get_timeline_name = ljm_fence_get_timeline_name,
    .enable_signaling  = ljm_fence_enable_signaling,
    .signaled          = ljm_fence_signaled,
    .wait              = dma_fence_default_wait,
    .release           = ljm_fence_release,
};

struct dma_fence *ljm_fence_create(struct ljm_sync_timeline *timeline, ljsSIGNAL *signal)
{
    ljeSTATUS status;
    struct ljm_fence *fence;
    struct dma_fence *old_fence = NULL;
    unsigned int seqno;

    fence = kzalloc(sizeof(*fence), ljdNOWARN | GFP_KERNEL);

    if (!fence)
        return NULL;

    /* Reference signal in fence. */
    status = ljkOS_MapSignal(timeline->os,
                             (ljtSIGNAL)(uintptr_t)signal->id,
                             NULL,
                             &fence->signal);

    if (ljmIS_ERROR(status)) {
        kfree(fence);
        return NULL;
    }

    spin_lock_init(&fence->lock);

    fence->parent = timeline;

    seqno = (unsigned int)atomic64_inc_return(&timeline->seqno);

    dma_fence_init((struct dma_fence *)fence, &ljm_fence_ops,
                   &fence->lock, timeline->context, seqno);

    /*
     * Reference fence in signal.
     * Be aware of recursive reference!!
     */
    spin_lock(&signal->lock);

    if (signal->fence) {
        old_fence = signal->fence;
        signal->fence = NULL;
    }

    if (!signal->done) {
        signal->fence = (struct dma_fence *)fence;

#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 9, 68)
        dma_fence_get((struct dma_fence *)fence);
#  endif
    }

    spin_unlock(&signal->lock);

    if (old_fence)
        dma_fence_put(old_fence);

    if (!signal->fence) {
        /* Fence already signaled. */
        ljkOS_DestroySignal(timeline->os, fence->signal);
        fence->signal = NULL;

        dma_fence_signal_locked((struct dma_fence *)fence);
    }

    return (struct dma_fence *)fence;
}

# endif /* CONFIG_SYNC_FILE */

#endif /* ljdLINUX_SYNC_FILE */
