/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

/******************************************************************************
 *
 *  版权所有(c)，2020-2099，中船第709研究所
 *
 *******************************************************************************
 *
 *  文件名称: ljmicro_ddc.c
 *  版 本 号: V1.0
 *  作    者: 胡艳明
 *  生成日期: 2022年4月15日
 *
 *  功能描述: DDC协议驱动
 *
 *  修改历史:
 *         1. 日    期: 2022年4月15日
 *            作    者: 胡艳明
 *            修改内容: 创建文件
 *
 ******************************************************************************/

/*******************************************************************************
 *                                   非标准库头文件                                *
 ******************************************************************************/

#include "ljmicro_base.h"

#define _LJ_OBJ_ZONE ljvZONE_KMS

/******************************************************************************
 *                                    函数实现                                    *
 ******************************************************************************/

/*****************************************************************************
*  函 数 名: ljm_ddc_get_edid
*  功能描述: 通过调用I2C硬件驱动函数获取显示设备的EDI-
                  D数据
*  输入参数: void *hw_dev                           I2C设备结构体指针
*            int (*i2c_read_req)(void *hw_dev       I2C设备读取函数指针
*            unsigned char target
*            unsigned char addr
*            unsigned char *buf
*            int size)
*            int (*i2c_raw_write_req)(void *hw_dev  I2C设备写数据函数指针
*            unsigned char target
*            unsigned char data)
*            int *psize
*  输出参数: 无
*  返 回 值: 获取的EDID数据，错误时返回NULL
*
*  修改历史:
*         1. 日    期: 2022年4月15日
*            作    者: 胡艳明
*            修改内容: 新生成函数
*
*****************************************************************************/
struct edid *ljm_ddc_get_edid(void *hw_dev,
                              int (*i2c_read_req)(void *hw_dev, unsigned char target, unsigned char addr,
                                                  unsigned char *buf, int size),
                              int (*i2c_raw_write_req)(void *hw_dev, unsigned char target, unsigned char data),
                              int *psize)
{
    int r;
    int i = 0;
    u8 *block, *new;
    int size;

    HEADER_ARG("hw_dev=%lx psize=%lx", (unsigned long)hw_dev, (unsigned long)psize);

    if (i2c_read_req == NULL)
    {
        ERR_PRINT("ljm_ddc_get_edid: invalid callback function\n");
        FOOTER_ARG("r=%d", -EINVAL);
        return NULL;
    }

    size = DDC_EDID_SIZE;
    if ((block = kzalloc(DDC_EDID_SIZE, GFP_KERNEL)) == NULL)
    {
        if (psize)
            *psize = 0;
        FOOTER_ARG("ERR(%d): Out of memory", -ENOMEM);
        return NULL;
    }

    /*为解决读取不到VGA的edid问题，在此处添加延时*/
    msleep(50);
    /*ERR_PRINT("-------sleep--50---\n");*/
    /* read 0 block */
    r = i2c_read_req(hw_dev, DDC_EDID_ADDR, 0, block, DDC_EDID_SIZE);
    if (r)
    {
        DEBUG_PRINT("ljm_ddc_get_edid: i2c read request fails (%d)\n", r);
        goto OnError;
    }

    /* if there's no extensions, we have done */
    if (block[DDC_EDID_SIZE - 2] == 0)
    {
        goto OnSuccess;
    }

    size = (block[DDC_EDID_SIZE - 2] + 1) * DDC_EDID_SIZE;
#if 0
    new = krealloc(block, size, GFP_KERNEL);
    if (!new) {
        goto OnError;
    }
#else
    new = kzalloc(size, GFP_KERNEL);
    if (!new)
    {
        goto OnError;
    }
    memcpy(new, block, DDC_EDID_SIZE);
    kfree(block);
#endif
    block = new;

    /* read another block in segment pointer 0 */
    r = i2c_read_req(hw_dev, DDC_EDID_ADDR, DDC_EDID_SIZE, block + DDC_EDID_SIZE, DDC_EDID_SIZE);
    if (r)
    {
        ERR_PRINT("ljm_ddc_get_edid: i2c read request fails (%d)\n", r);
        goto OnError;
    }

    if ((size - DDC_EDID_SIZE * 2) > 0 && i2c_raw_write_req == NULL)
    {
        ERR_PRINT("ljm_ddc_get_edid: invalid i2c_raw_write_req\n");
        goto OnError;
    }

    /* read other blocks with segment pointer register */
    for (i = 2; i < block[DDC_EDID_SIZE - 2] + 1; i++)
    {
        /* set segment pointer register */
        if (i % 2 == 0)
        {
            r = i2c_raw_write_req(hw_dev, DDC_SEG_PTR_ADDR, i / 2);
            if (r)
            {
                ERR_PRINT("ljm_ddc_get_edid: i2c write segment pointer register fails (%d)\n", r);
                goto OnError;
            }
        }

        /* read two blocks in a segment */
        if (block[DDC_EDID_SIZE - 2] + 1 - i > 1)
        {
            r = i2c_read_req(hw_dev, DDC_EDID_ADDR, 0, block + DDC_EDID_SIZE * i, DDC_EDID_SIZE * 2);
            i++;
            /* read a block in a segment */
        }
        else
        {
            r = i2c_read_req(hw_dev, DDC_EDID_ADDR, DDC_EDID_SIZE * (i % 2), block + DDC_EDID_SIZE * i, DDC_EDID_SIZE);
        }
        if (r)
        {
            ERR_PRINT("ljm_ddc_get_edid: i2c read request fails (%d)\n", r);
            goto OnError;
        }
    }

    /* FIXME: need to reset segment pointer register to 0 ? */
    if (i > 2)
    {
        r = i2c_raw_write_req(hw_dev, DDC_SEG_PTR_ADDR, 0);
        if (r)
        {
            ERR_PRINT("ljm_ddc_get_edid: i2c writing 0 to segment pointer register fails (%d)\n", r);
            /* still continue */
        }
    }

OnSuccess:
#if DEBUG_DUMP_EDID
    dump_bytes(block, (block[DDC_EDID_SIZE - 2] + 1) * DDC_EDID_SIZE, "EDID");
#endif

    if (psize)
        *psize = size;
    FOOTER_ARG("new EDID=0x%lx", (unsigned long)block);
    return (struct edid *)block;

OnError:
    kfree(block);
    if (psize)
        *psize = 0;
    FOOTER();
    return NULL;
}
