/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

/*******************************************************************************
 * * 版权所有 (C)2024, 武汉凌久微电子有限公司。
 * * 
 * * 文件名称： ljmicro_dma_pcm.c
 * * 文件标识： 
 * * 内容摘要： 本文件定义pcm驱动，pcm通过dma将buffer中的音频数据传递给IIS的FIFO
 * * 其它说明： 
 * * 当前版本： v1.0
 * * 作    者： xusn, 武汉凌久微电子有限公司
 * * 完成日期： 2024年12月31日
 * ***************************************************************************/
#if IS_ENABLED(CONFIG_SND_SOC) && defined(ljdENABLE_AUDIO)

#include <linux/dmaengine.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include <linux/interrupt.h>
#include <linux/fs.h>

#include "ljmicro_adma_reg.h"
#include "ljmicro_device.h"
#include "ljmicro_object.h"

#define DMAC_STOP 0
#define DMAC_START 1
#define DMAC_IRQ_ALL GENMASK(31, 0)
#define DMAC_IRQ_NONE 0
#define DMAC_MAX_CHAN_NR 8

#define ST_RUNNING (1<<0)
#define ST_OPENED (1<<1)

#define BUFFER_BYTES_MAX (2 * 2 * 8 * PERIOD_BYTES_MIN) //64kb
#define PERIOD_BYTES_MIN 4096
#define PERIODS_MIN 2

struct ljm_device *get_ljm_dev(void);

struct ljm_adma_ctrl_block {
    void __iomem *dma_reg_base;   //DMA寄存器起始地址
    unsigned int irq_nr;          //DMA中断号
    spinlock_t sp_lock;         //自旋锁
    unsigned int chan_nr[2];      //DMA传输通道号
    int state;           //保存当前substream的运行状态
    dma_addr_t sar_start;       //DMA传输起始地址
    dma_addr_t pos;             //DMA传输当前地址
    dma_addr_t sar_end;         //DMA传输结束地址
    dma_addr_t period;          //音频Buffer中period数
    dma_addr_t periodsz;        //一个periodsz占用的字节数  
    dma_addr_t dar[2];          //DMA传输目标地址
    dma_addr_t cur_pos;         //DMA当前传输位置
    dma_addr_t interr_num;      //DMA传输中断数量
    unsigned int bytes_per_frame; //每一帧传输的字节数
    void *ss;                     //保存当前substream地址
    void (*callback)(void *ss, int bytes_xfer);    //完成一个period时的回调函数
};

static const struct snd_pcm_hardware ljm_pcm_hardware = {
    .info = SNDRV_PCM_INFO_INTERLEAVED |
            SNDRV_PCM_INFO_MMAP |
            SNDRV_PCM_INFO_MMAP_VALID |
            SNDRV_PCM_INFO_BLOCK_TRANSFER,
    .rates = SNDRV_PCM_RATE_32000 |
             SNDRV_PCM_RATE_44100 |
             SNDRV_PCM_RATE_48000,
    .rate_min = 32000,
    .rate_max = 48000,
    .formats = SNDRV_PCM_FMTBIT_S16_LE |
               SNDRV_PCM_FMTBIT_S24_LE |
               SNDRV_PCM_FMTBIT_S32_LE,
    .channels_min = 2,
    .channels_max = 2,
    .buffer_bytes_max = BUFFER_BYTES_MAX,  
    .period_bytes_min = PERIOD_BYTES_MIN,  //4kb
    .period_bytes_max = BUFFER_BYTES_MAX / PERIODS_MIN, //64kb
    .periods_min = PERIODS_MIN,
    .periods_max = BUFFER_BYTES_MAX / PERIOD_BYTES_MIN,
    .fifo_size = 16,
};

/* 用于将读取到的数据分配至显存空间 */
struct dma_buf {
    dma_addr_t offset;    /* 用于储存分配显存的偏移地址 */  
    size_t bytes;	  /* buffer size in bytes */
    unsigned char *area;  /* virtual pointer */
    void *private_data;	  /* private for allocator; don't touch */
};

/******************* 单块传输的DMA编程流 ************************************/
/* DMAC公共 */
static void program_dma_en(struct ljm_adma_ctrl_block *dma_cb)
{
    u32 val;

    val = readl(dma_cb->dma_reg_base + DMAC_CFGREG);
    val |= BIT(0);
    writel(val, dma_cb->dma_reg_base + DMAC_CFGREG);
}

static void program_dma_dis(struct ljm_adma_ctrl_block *dma_cb)
{
    u32 val;

    val = readl(dma_cb->dma_reg_base + DMAC_CFGREG);
    val &= ~BIT(0);
    writel(val, dma_cb->dma_reg_base + DMAC_CFGREG);
}

static void program_dma_irq_en(struct ljm_adma_ctrl_block *dma_cb)
{
    u32 val;

    val = readl(dma_cb->dma_reg_base + DMAC_CFGREG);
    val |= BIT(1);
    writel(val, dma_cb->dma_reg_base + DMAC_CFGREG);
}

#if 0
static void program_dma_irq_dis(struct ljm_adma_ctrl_block *dma_cb)
{
    u32 val;

    val = readl(dma_cb->dma_reg_base + DMAC_CFGREG);
    val &= ~BIT(1);
    writel(val, dma_cb->dma_reg_base + DMAC_CFGREG);
}
#endif

/* DMAC通道 */
static void program_chan_irq_set(struct ljm_adma_ctrl_block *dma_cb, u32 irq_mask, u32 i2s_num)
{
    u32 val = irq_mask;
    writel(val, dma_cb->dma_reg_base + CHx_INTSTATUS_ENABLEREG + 0x100 * (dma_cb->chan_nr[i2s_num] - 1));
}

#if 0
static void program_chan_irq_dis(struct ljm_adma_ctrl_block *dma_cb, u32 irq_mask)
{
    u32 val;

    if (likely(irq_mask == DMAC_IRQ_ALL)) {
        writel(DMAC_IRQ_NONE, dma_cb->dma_reg_base + CHx_INTSTATUS_ENABLEREG + 0x100 * (dma_cb->chan_nr - 1));
    } else {
        val = readl(dma_cb->dma_reg_base + CHx_INTSTATUS_ENABLEREG + 0x100 * (dma_cb->chan_nr - 1));
        val &= ~irq_mask;
        writel(DMAC_IRQ_NONE, dma_cb->dma_reg_base + CHx_INTSTATUS_ENABLEREG + 0x100 * (dma_cb->chan_nr - 1));
    }
}
#endif

static void program_chan_irq_sig_set(struct ljm_adma_ctrl_block *dma_cb, u32 irq_mask, u32 i2s_num)
{
    u32 val = irq_mask;
    writel(val, dma_cb->dma_reg_base + CHx_INTSIGNAL_ENABLEREG + 0x100 * (dma_cb->chan_nr[i2s_num] - 1));
}

static void program_chan_irq_clear(struct ljm_adma_ctrl_block *dma_cb, u32 irq_mask, u32 i2s_num)
{
    u32 val = irq_mask;
    writel(val, dma_cb->dma_reg_base + CHx_INTCLEARREG + 0x100 * (dma_cb->chan_nr[i2s_num] - 1));
}

/* 流程 */
/* 1、通过DMAC_ChEnReg，选择一个可用的通道 */
static int program_sel_free_chan(struct ljm_adma_ctrl_block *dma_cb)
{
    int id;
    u32 val = readl(dma_cb->dma_reg_base + DMAC_CHENREG);

    for (id = 1; id <= DMAC_CHAN_MAX; id++) {
        if ((val & BIT(id - 1)) || (val & BIT(id))) {
            printk(KERN_ERR "LJM PCM: no dma chan is free.\n");  //只有channel(1)和channel(2) fifo为（3fff + 1） * 4
            return -1;
        }
        else {
            dma_cb->chan_nr[0] = id;
            dma_cb->chan_nr[1] = id + 1;
            return 0;
        }
    }

    printk(KERN_ERR "LJM PCM: no dma chan is free.\n");
    return -1;
}

static void program_free_sel_chan(struct ljm_adma_ctrl_block *dma_cb, u32 i2s_num)
{
    u32 val = readl(dma_cb->dma_reg_base + DMAC_CHENREG);
    val &= ~BIT(dma_cb->chan_nr[i2s_num] - 1);
    writel(val, dma_cb->dma_reg_base + DMAC_CHENREG);
}

static void program_free_chan(struct ljm_adma_ctrl_block *dma_cb)
{
    u32 val = 0;
    writel(val, dma_cb->dma_reg_base + DMAC_CHENREG);
}

/* 在DMAC_Common_interrupt_reg中使能相应的中断寄存器 */
static void program_common_interrupt_reg(struct ljm_adma_ctrl_block *dma_cb)
{
    u32 val = 0x1ff;
    writel(val, dma_cb->dma_reg_base + DMAC_COMMONREG_INTSTATUS_ENABLEREG);

    val = 0x128;
    writel(val, dma_cb->dma_reg_base + DMAC_COMMONREG_INTSIGNAL_ENABLEREG);
}

/* 2、配置CH_CFG的源和目的的multi-block类型 */
static void program_chan_cfg(struct ljm_adma_ctrl_block *dma_cb, u32 i2s_num)
{
    u32 val;

    val = 0x0;
    writel(val, dma_cb->dma_reg_base + CHx_CFG_L + 0x100 * (dma_cb->chan_nr[i2s_num] - 1));

    val = 0x7f800009;
    writel(val, dma_cb->dma_reg_base + CHx_CFG_H + 0x100 * (dma_cb->chan_nr[0] - 1)); //HDMI
    val = 0x7f801009;
    writel(val, dma_cb->dma_reg_base + CHx_CFG_H + 0x100 * (dma_cb->chan_nr[1] - 1)); //DP
}

static uint32_t SRC_single_trans_bytes_count(u32 val)
{
    u32 width;
    switch ((val >> 8) & 0x7) {
        case 0:
            width = 1;
            break;
        case 1:
            width = 2;
            break;
        case 2:
            width = 4;
            break;
        case 3:
            width = 8;
            break;
        case 4:
            width = 16;
            break;
        case 5:
            width = 24;
            break;
        case 6:
            width = 32;
            break;
        default:
            width = 0;
            break;
    }

    return width;
}

/* 3、配置CH_SAR\CH_DAR\CH_BLOCK_TS\CH_CTL为适当的值 */
static void program_other_chcfg_reg(struct ljm_adma_ctrl_block *dma_cb, u32 i2s_num)
{
    u32 val = 0, src_tr_width; //dst_tr_width;
    u32 period_size = dma_cb->periodsz; //以字节为单位

    val = (dma_cb->sar_start & 0xffffffff);
    writel(val, dma_cb->dma_reg_base + CHx_SAR_L + 0x100 * (dma_cb->chan_nr[i2s_num] - 1));
    val = (dma_cb->sar_start >> 32) & 0xffffffff;
    writel(val, dma_cb->dma_reg_base + CHx_SAR_H + 0x100 * (dma_cb->chan_nr[i2s_num] - 1));

    val = dma_cb->dar[i2s_num] & 0xffffffff;
    writel(val, dma_cb->dma_reg_base + CHx_DAR_L + 0x100 * (dma_cb->chan_nr[i2s_num] - 1));
    val = (dma_cb->dar[i2s_num] >> 32) & 0xffffffff;
    writel(val, dma_cb->dma_reg_base + CHx_DAR_H + 0x100 * (dma_cb->chan_nr[i2s_num] - 1));

    val = 0x60005;
    writel(val, dma_cb->dma_reg_base + CHx_AXI_IDREG + 0x100 * (dma_cb->chan_nr[i2s_num] - 1));

    val = 0xf87d2;
    writel(val, dma_cb->dma_reg_base + CHx_CTL_H + 0x100 * (dma_cb->chan_nr[i2s_num] - 1));

    val = 0xcd240; //SRC_WIDTH = 32bits, DST_WIDTH = 32bits

    writel(val, dma_cb->dma_reg_base + CHx_CTL_L + 0x100 * (dma_cb->chan_nr[i2s_num] - 1));
    src_tr_width = SRC_single_trans_bytes_count(val);

    val = period_size / src_tr_width - 1;
    writel(val, dma_cb->dma_reg_base + CHx_BLOCK_TS + 0x100 * (dma_cb->chan_nr[i2s_num] - 1)); //DMA传输以字节为单位
}

/* 4、写1到DMAC_ChEnReg寄存器对应的位，使能通道 */
static void program_chan_en(struct ljm_adma_ctrl_block *dma_cb, u32 i2s_num)
{
    u32 val = readl(dma_cb->dma_reg_base + DMAC_CHENREG);

    val |= BIT(dma_cb->chan_nr[i2s_num] - 1) << DMAC_CHAN_EN_SHIFT | BIT(dma_cb->chan_nr[i2s_num] - 1) << DMAC_CHAN_EN_WE_SHIFT;
    writel(val, dma_cb->dma_reg_base + DMAC_CHENREG);
}

static void program_chan_dis(struct ljm_adma_ctrl_block *dma_cb, u32 i2s_num)
{
    u32 val = readl(dma_cb->dma_reg_base + DMAC_CHENREG);

    val &= ~BIT(dma_cb->chan_nr[i2s_num] - 1) << DMAC_CHAN_EN_SHIFT; 
    val |= BIT(dma_cb->chan_nr[i2s_num] - 1) << DMAC_CHAN_EN_WE_SHIFT;
    writel(val, dma_cb->dma_reg_base + DMAC_CHENREG);
}
/*****************************************************************************/

static int ljm_adma_getpos(struct ljm_adma_ctrl_block *dma_cb)
{
    dma_addr_t cur_abs_pos = 0, tmp;

    tmp = readl(dma_cb->dma_reg_base + CHx_SAR_L);
    cur_abs_pos |= tmp;
    tmp = readl(dma_cb->dma_reg_base + CHx_SAR_H);
    cur_abs_pos |= tmp << 32;

    dma_cb->cur_pos = (cur_abs_pos - dma_cb->sar_start) / (dma_cb->bytes_per_frame);  //bytes to frames

    return 0;
}

static int ljm_adma_control(struct ljm_adma_ctrl_block *dma_cb, u32 flag)
{
    u32 i2s_num;

    for (i2s_num = 0; i2s_num < LJM_MAX_I2SS; i2s_num++) {
        if (flag == DMAC_START) {
            program_chan_en(dma_cb, i2s_num);
        }
        else {
            program_chan_dis(dma_cb, i2s_num);
        }
    }

    return 0;
}

static void ljm_adma_period_done(void *ss, int bytes_xfer)
{
    struct snd_pcm_substream *substream = ss;
    struct ljm_adma_ctrl_block *dma_cb = substream->runtime->private_data;

    if (dma_cb && (dma_cb->state & ST_RUNNING)) {
        snd_pcm_period_elapsed(substream);
    }
}

/* 设置一个period传输完成时回调函數 */
static void ljm_adma_setcallback(struct snd_pcm_substream *substream, void (*cb)(void *, int))
{
    struct ljm_adma_ctrl_block *dma_cb = substream->runtime->private_data;

    spin_lock(&dma_cb->sp_lock);
    dma_cb->callback = cb;
    spin_unlock(&dma_cb->sp_lock);
}

static irqreturn_t ljm_adma_irq_handler(int irq_nr, void *dev_id)
{
    u32 val, misc_intr_mask, misc_intr_src;
    int i2s_num;

    struct ljm_adma_ctrl_block *dma_cb = dev_id;
    struct snd_pcm_substream *substream = dma_cb->ss;
    struct snd_dma_buffer *buf = &substream->dma_buffer;
    struct ljm_gem_object *bo = buf->private_data;

    misc_intr_src = LJM_MISC_R32(0, MISC_INTR_SRC);
    misc_intr_mask = LJM_MISC_R32(0, MISC_INTR_EN);

    if ((readl(dma_cb->dma_reg_base + DMAC_INTSTATUSREG) & 3) == 0x3) {
        dma_cb->interr_num ++;
        dma_cb->callback(dma_cb->ss, dma_cb->period);

        for (i2s_num = 0; i2s_num < LJM_MAX_I2SS; i2s_num++) {
            val = readl(dma_cb->dma_reg_base + CHx_INTSTATUS_L + 0x100 * (dma_cb->chan_nr[i2s_num] - 1));
            program_chan_irq_clear(dma_cb, val, i2s_num);

            if ((dma_cb->interr_num % dma_cb->period) == 0) {
                dma_cb->sar_start = VRAM_START_AMBA + get_offset_from_bo(bo);
                /* 将DMA搬移的起始地址返回 */
                val = (dma_cb->sar_start & 0xffffffff);
                writel(val, dma_cb->dma_reg_base + CHx_SAR_L + 0x100 * (dma_cb->chan_nr[i2s_num] - 1));
                val = (dma_cb->sar_start >> 32) & 0xffffffff;
                writel(val, dma_cb->dma_reg_base + CHx_SAR_H + 0x100 * (dma_cb->chan_nr[i2s_num] - 1));
            }

            /* 重启DMA操作 */ /* 只要period传输完成，重启DMA */
            program_chan_en(dma_cb, i2s_num);
        }
    }    
    else {
        return IRQ_NONE;
    }

    return IRQ_HANDLED;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 10, 0)
static int ljm_dma_pcm_open(struct snd_soc_component *component, struct snd_pcm_substream *substream)
#else
static int ljm_dma_pcm_open(struct snd_pcm_substream *substream)
#endif
{
    struct snd_pcm_runtime *runtime = substream->runtime;
    struct ljm_device *ljm_dev = get_ljm_dev();
    struct ljm_adma_ctrl_block *dma_cb;
    int ret;

    snd_soc_set_runtime_hwparams(substream, &ljm_pcm_hardware);
    snd_pcm_hw_constraint_integer(runtime, SNDRV_PCM_HW_PARAM_PERIODS);

    /* 分配DMA控制数据结构体 */
    dma_cb = kzalloc(sizeof(struct ljm_adma_ctrl_block), GFP_KERNEL);
    if (!dma_cb) {    /* 当返回值为0时，返回 */
        return -ENOMEM;
    }

    dma_cb->dma_reg_base = ljm_dev->mmio_base + REG_DMAC_OFFSET;
    dma_cb->irq_nr = ljm_dev->pdev->irq;
    dma_cb->ss = substream;
    dma_cb->interr_num = 0;

    program_dma_en(dma_cb);          //读 可用于检查dma传输是否结束
    program_dma_irq_en(dma_cb);      //0x10写3 enable dmac

    program_sel_free_chan(dma_cb);   //选择1个free的通道, 这行需有，给dma_cb->chan赋值

    spin_lock_init(&dma_cb->sp_lock);
    runtime->private_data = dma_cb;

    /* 注册DMA中断 */
    LJM_MISC_W32(0, MISC_INTR_EN, LJM_MISC_R32(0, MISC_INTR_EN) | 0x1000000);  //使能adma mask中断
    ret = request_irq(dma_cb->irq_nr, ljm_adma_irq_handler, IRQF_SHARED, "ljm-adma", dma_cb);

    return 0;
}

#define CRU_SOFTRST 0x64
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 10, 0)
static int ljm_dma_pcm_close(struct snd_soc_component *component, struct snd_pcm_substream *substream)
#else
static int ljm_dma_pcm_close(struct snd_pcm_substream *substream)
#endif
{
    struct snd_pcm_runtime *runtime = substream->runtime;
    struct ljm_adma_ctrl_block *dma_cb = runtime->private_data;

    free_irq(dma_cb->irq_nr, dma_cb);
    program_dma_dis(dma_cb);
    kfree(dma_cb);

    /* reset dmac */
    LJM_CRU_W32(0, CRU_SOFTRST, 1 << 28);
    LJM_CRU_W32(0, CRU_SOFTRST, 0);

    return 0;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 10, 0)
static int ljm_dma_pcm_hw_params(struct snd_soc_component *component, struct snd_pcm_substream *substream,
                                 struct snd_pcm_hw_params *hw_params)
#else
static int ljm_dma_pcm_hw_params(struct snd_pcm_substream *substream, struct snd_pcm_hw_params *hw_params)
#endif
{
    struct snd_pcm_runtime *runtime = substream->runtime;
    struct ljm_adma_ctrl_block *dma_cb = runtime->private_data;

    struct snd_dma_buffer *buf = &substream->dma_buffer;
    struct ljm_gem_object *bo = buf->private_data;
    int i;

    snd_pcm_set_runtime_buffer(substream, &substream->dma_buffer);
    runtime->dma_bytes = params_buffer_bytes(hw_params); //以字节为单位
    //此时runtime里面的buf值还没传递过来

    dma_cb->sar_start = VRAM_START_AMBA + get_offset_from_bo(bo); //该处应是GPU侧的地址
    dma_cb->period = params_periods(hw_params); //Get the num of periods from the hw period, 与runtime中获取的一致
    dma_cb->periodsz = params_period_bytes(hw_params); //与从runtime->period_size(以帧为单位)中获取的不一致，此处以字节为单位；
    dma_cb->sar_end = dma_cb->sar_start + runtime->dma_bytes;

    for (i = 0; i < LJM_MAX_I2SS; i++) {
        dma_cb->dar[i] = MEM_IO_START_AMBA + REG_I2S_OFFSET + REG_I2S_SIZE * i + I2S_AUDIO_DATA_OFFSET;   
    }

    runtime->format = params_format(hw_params);

    /* 通过获取的字节数计算bytes  1表示单声道、2表示立体声；format用于设置采样率和位深度 */
    dma_cb->bytes_per_frame = snd_pcm_format_physical_width(runtime->format) * params_channels(hw_params) / 8;
    ljm_adma_setcallback(substream, ljm_adma_period_done);

    return 0;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 10, 0)
static int ljm_dma_pcm_hw_free(struct snd_soc_component *component, struct snd_pcm_substream *substream)
#else
static int ljm_dma_pcm_hw_free(struct snd_pcm_substream *substream)
#endif
{
    struct snd_pcm_runtime *runtime = substream->runtime;
    struct ljm_adma_ctrl_block *dma_cb = runtime->private_data;
    int i2s_num;

    snd_pcm_set_runtime_buffer(substream, NULL);

    for (i2s_num = 0; i2s_num < LJM_MAX_I2SS; i2s_num++) {
        program_free_sel_chan(dma_cb, i2s_num);   //将0x18写0
    }

    program_free_chan(dma_cb);

#if 0
    program_dma_dis(dma_cb);
    program_dma_irq_dis(dma_cb);
#endif
    return 0;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 10, 0)
static int ljm_dma_pcm_prepare(struct snd_soc_component *component, struct snd_pcm_substream *substream)
#else
static int ljm_dma_pcm_prepare(struct snd_pcm_substream *substream)
#endif
{
    struct ljm_adma_ctrl_block *dma_cb = substream->runtime->private_data;
    int i2s_num;

    /* 刷新DMA通道 */
    program_common_interrupt_reg(dma_cb);          //0x40、0x48

    for (i2s_num = 0; i2s_num < LJM_MAX_I2SS; i2s_num++) {
        program_chan_irq_set(dma_cb, 0xffffffef, i2s_num);      //0x180  关闭0x180(0x10)小中断
        program_chan_irq_sig_set(dma_cb, 0xffffffff, i2s_num);  //0x190
        program_chan_cfg(dma_cb, i2s_num);                      //0x120、  0x124？
        program_other_chcfg_reg(dma_cb, i2s_num);               //0x100、0x104、0x108、0x10c、0x110、0x150、0x11c、0x118
    }
    return 0;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 10, 0)
static int ljm_dma_pcm_trigger(struct snd_soc_component *component, struct snd_pcm_substream *substream, int cmd)
#else
static int ljm_dma_pcm_trigger(struct snd_pcm_substream *substream, int cmd)
#endif
{
    struct ljm_adma_ctrl_block *dma_cb = substream->runtime->private_data;
    int ret = 0;

    spin_lock(&dma_cb->sp_lock);

    switch (cmd) {
        case SNDRV_PCM_TRIGGER_RESUME:
        case SNDRV_PCM_TRIGGER_START:
        case SNDRV_PCM_TRIGGER_PAUSE_RELEASE:
            dma_cb->state |= ST_RUNNING;
            ljm_adma_control(dma_cb, DMAC_START);
            break;
        case SNDRV_PCM_TRIGGER_SUSPEND:
        case SNDRV_PCM_TRIGGER_STOP:
        case SNDRV_PCM_TRIGGER_PAUSE_PUSH:
            dma_cb->state &= ~ST_RUNNING;
            ljm_adma_control(dma_cb, DMAC_STOP);
            break;
        default:
            ret = -EINVAL;
            break;
    }

    spin_unlock(&dma_cb->sp_lock);

    return 0;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 10, 0)
static snd_pcm_uframes_t ljm_dma_pcm_pointer(struct snd_soc_component *component, struct snd_pcm_substream *substream)
#else
static snd_pcm_uframes_t ljm_dma_pcm_pointer(struct snd_pcm_substream *substream)
#endif
{
    struct snd_pcm_runtime *runtime = substream->runtime;
    struct ljm_adma_ctrl_block *dma_cb  = runtime->private_data;

    spin_lock(&dma_cb->sp_lock);

    ljm_adma_getpos(dma_cb); //获取当前传输位置

    spin_unlock(&dma_cb->sp_lock);

    if (dma_cb->cur_pos >= runtime->buffer_size) {
        dma_cb->cur_pos = 0;
    }

    spin_unlock(&dma_cb->sp_lock);

    return dma_cb->cur_pos; //bytes_to_frames
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 10, 0)
static int ljm_dma_pcm_mmap(struct snd_soc_component *component, struct snd_pcm_substream *substream,
                            struct vm_area_struct *vma)
#else
static int ljm_dma_pcm_mmap(struct snd_pcm_substream *substream, struct vm_area_struct *vma)
#endif
{
    unsigned long size, offset;
    int ret;

    vma->vm_page_prot = pgprot_writecombine(vma->vm_page_prot);  //设置页属性，允许内存区域进行写合并；
    size = vma->vm_end - vma->vm_start;
    offset = vma->vm_pgoff << PAGE_SHIFT;
    ret = io_remap_pfn_range(vma, vma->vm_start, (substream->dma_buffer.addr + offset) >> PAGE_SHIFT,
                             size, vma->vm_page_prot);

    return ret;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 10, 0)
static const struct snd_pcm_ops ljm_dma_pcm_ops = {
    .open      = ljm_dma_pcm_open,       // 1
    .close     = ljm_dma_pcm_close,      // 5
    .ioctl     = snd_pcm_lib_ioctl,
    .hw_params = ljm_dma_pcm_hw_params,  // 2
    .hw_free   = ljm_dma_pcm_hw_free,    // 4
    .prepare   = ljm_dma_pcm_prepare,    // 3
    .trigger   = ljm_dma_pcm_trigger,
    .pointer   = ljm_dma_pcm_pointer,
    .mmap      = ljm_dma_pcm_mmap,
};
#endif

/*
 * 预分配显存空间，地址保存在substream->dma_buffer
 *
 *
 *
 *
 */
static int ljm_audio_dma_buffer_preallocate(struct snd_pcm *pcm, int stream)
{
    struct snd_pcm_substream *substream = pcm->streams[stream].substream;
    struct snd_dma_buffer *buf = &substream->dma_buffer;
    size_t size = ljm_pcm_hardware.buffer_bytes_max;
    int    ret;
    void  *kptr = NULL;
    struct drm_gem_object *obj = NULL;
    struct ljm_gem_object *bo = NULL;
    struct ljm_device *ljm_dev = get_ljm_dev();

    ret = ljm_gem_object_create(ljm_dev, size, &obj);
    if (ret) {
        printk(KERN_ERR "Failed to alloc vram audio buffer.\n");
        return ret;
    }

    bo = to_ljm_gem(obj);
    ret = ljm_bo_pin(bo, NULL);
    if (ret) {
        printk(KERN_ERR "Ljm: Failed to pin bo.\n");
        return ret;
    }
    ljm_bo_kmap(bo, &kptr);

    buf->dev.dev = pcm->card->dev;
    buf->private_data = bo;

    buf->dev.type = SNDRV_DMA_TYPE_CONTINUOUS;
    buf->addr = get_cpu_addr_from_bo(bo); //CPU phys_addr
    buf->bytes = size;
    buf->area = (unsigned char * __force)kptr; //CPU virt_addr

    return 0;
}

static void ljm_audio_dma_buffer_preallocate_free(struct snd_pcm *pcm)
{
    struct ljm_gem_object *bo = pcm->streams[SNDRV_PCM_STREAM_PLAYBACK].substream->dma_buffer.private_data;

    ljm_bo_kunmap(bo);
    ljm_put_vram(bo->base.dev, bo);
    drm_gem_object_unreference_unlocked(&bo->base);

    pcm->streams[SNDRV_PCM_STREAM_PLAYBACK].substream->dma_buffer.area = NULL;
    pcm->streams[SNDRV_PCM_STREAM_PLAYBACK].substream->dma_buffer.addr = 0;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 10, 0)
static int ljm_dma_pcm_new(struct snd_soc_component *component, struct snd_soc_pcm_runtime *rtd)
#else
static int ljm_dma_pcm_new(struct snd_soc_pcm_runtime *rtd)
#endif
{
    return ljm_audio_dma_buffer_preallocate(rtd->pcm, SNDRV_PCM_STREAM_PLAYBACK); //分配显存空间
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 10, 0)
static void ljm_dma_pcm_free(struct snd_soc_component *component, struct snd_pcm *pcm)
#else
static void ljm_dma_pcm_free(struct snd_pcm *pcm)
#endif
{
    ljm_audio_dma_buffer_preallocate_free(pcm);
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 10, 0)
static const struct snd_soc_component_driver ljm_dma_pcm_component_drv = {
    .ops = &ljm_dma_pcm_ops,
    .pcm_new = ljm_dma_pcm_new,    /* 用于分配PCM数据显存空间 */
    .pcm_free = ljm_dma_pcm_free,
};
#else
static const struct snd_soc_component_driver ljm_dma_pcm_component_drv = {
    .open = ljm_dma_pcm_open,
    .close = ljm_dma_pcm_close,
    .trigger = ljm_dma_pcm_trigger,
    .pointer = ljm_dma_pcm_pointer,
    .mmap = ljm_dma_pcm_mmap,
    .hw_params = ljm_dma_pcm_hw_params,
    .hw_free = ljm_dma_pcm_hw_free,
    .prepare = ljm_dma_pcm_prepare,
    .pcm_construct = ljm_dma_pcm_new,
    .pcm_destruct = ljm_dma_pcm_free,
};
#endif

int ljm_dma_pcm_register(struct device *dev)
{
    int ret = 0;
    struct snd_soc_component *ljm_component;

    ljm_component = devm_kzalloc(dev, sizeof(*ljm_component), GFP_KERNEL);

#ifdef CONFIG_DEBUG_FS
    ljm_component->debugfs_prefix = "ljm_dma";
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 6, 0)
    ret = devm_snd_soc_register_component(dev, &ljm_dma_pcm_component_drv, NULL, 0);
#else
    ret = snd_soc_add_component(dev, ljm_component, &ljm_dma_pcm_component_drv, NULL, 0);
#endif

    if (ret) {
        dev_err(dev, "LJM DMA PCM: could not register pcm: %d.\n", ret);
    }

    return ret;
}

#endif
