/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

/******************************************************************************
 *
 *  版权所有(c)，2020-2099，武汉凌久微电子有限公司
 *
 ******************************************************************************
 *
 *  文件名称: ljmicro_edid.c
 *  版 本 号: V1.0
 *  作    者: zhaiyf
 *  生成日期: 2022年4月25日
 *
 *  功能描述: edid解析功能，依赖内核drm接口。
 *
 *  修改历史:
 *         1. 日    期: 2022年4月25日
 *            作    者: zhaiyf
 *            修改内容: 创建文件
 *
 *****************************************************************************/

/*******************************************************************************
 *                                   非标准库头文件                                *
 ******************************************************************************/
#include "ljmicro_hdmi.h"
#include "ljmicro_edid.h"


/*****************************************************************************
 *  函 数 名: ljm_edid_find_extension
 *  功能描述: 搜索edid中extension block
 *  输入参数: u8 *edid
 *            int ext_id
 *  输出参数: 无
 *  返 回 值: static u8 *
 *
 *  修改历史:
 *         1. 日    期: 2022年4月29日
 *            作    者: zhaiyf
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
static u8 *ljm_edid_find_extension(u8 *edid, int ext_id)
{
    u8 *edid_ext = NULL;
    int i;

    /* No EDID or EDID extensions */
    if (edid == NULL || edid[LJM_EDID_EXTENSIONS] == 0)
        return NULL;

    /* Find CEA extension */
    for (i = 0; i < edid[LJM_EDID_EXTENSIONS]; i++)
    {
        edid_ext = edid + LJM_EDID_LENGTH * (i + 1);
        if (edid_ext[0] == ext_id)
            break;
    }

    if (i == edid[LJM_EDID_EXTENSIONS])
        return NULL;

    return edid_ext;
}

static int ljm_edid_cea_db_payload_len(const u8 *db)
{
    return db[0] & 0x1f;
}

static int ljm_edid_cea_db_offsets(const u8 *cea, int *start, int *end)
{
    /* Data block offset in CEA extension block */
    *start = 4;
    *end   = cea[2];
    if (*end == 0)
        *end = 127;
    if (*end < 4 || *end > 127)
        return -ERANGE;

    /*
     * XXX: cea[2] is equal to the real value minus one in some sink edid.
     */
    if (*end != 4)
    {
        int i;

        i = *start;
        while (i < (*end) && i + ljm_edid_cea_db_payload_len(&(cea)[i]) < (*end))
            i += ljm_edid_cea_db_payload_len(&(cea)[i]) + 1;

        if (ljm_edid_cea_db_payload_len(&(cea)[i]) && i + ljm_edid_cea_db_payload_len(&(cea)[i]) == (*end))
            (*end)++;
    }

    return 0;
}

static int ljm_edid_cea_db_tag(const u8 *db)
{
    return db[0] >> 5;
}

/*****************************************************************************
 *  函 数 名: ljm_edid_cea_db_is_hdmi_vsdb
 *  功能描述: 确认edid中cea datablock是否为VSDB(Vendor-Specific
                   Data Block)
 *  输入参数: const u8 *db
 *  输出参数: 无
 *  返 回 值: static bool
 *
 *  修改历史:
 *         1. 日    期: 2022年4月29日
 *            作    者: zhaiyf
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
static bool ljm_edid_cea_db_is_hdmi_vsdb(const u8 *db)
{
    int hdmi_id;

    if (ljm_edid_cea_db_tag(db) != LJM_EDID_VENDOR_BLOCK)
        return false;

    if (ljm_edid_cea_db_payload_len(db) < 5)
        return false;

    hdmi_id = db[1] | (db[2] << 8) | (db[3] << 16);

    return hdmi_id == LJM_HDMI_IEEE_OUI;
}

/*****************************************************************************
 *  函 数 名: ljm_edid_is_hdmi
 *  功能描述: 根据edid判断设备是hdmi还是dvi
 *  输入参数: u8 *edid
 *  输出参数: 无
 *  返 回 值: int
 *            DVI: 0 HDMI: 1
 *  修改历史:
 *         1. 日    期: 2022年4月29日
 *            作    者: zhaiyf
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
bool ljm_edid_is_hdmi(u8 *edid)
{
    u8 *edid_ext;
    int i;
    int start_offset, end_offset;

    edid_ext = ljm_edid_find_extension(edid, LJM_EDID_CEA_EXT);
    if (!edid_ext)
        goto no_hdmi;

    if (ljm_edid_cea_db_offsets(edid_ext, &start_offset, &end_offset))
        goto no_hdmi;

    /*
     * Because HDMI identifier is in Vendor Specific Block,
     * search it from all data blocks of CEA extension.
     */
    ljm_edid_for_each_cea_db(edid_ext, i, start_offset, end_offset)
    {
        if (ljm_edid_cea_db_is_hdmi_vsdb(&edid_ext[i]))
        {
            DEBUG_PRINT("[%s]EDID is HDMI\n", __FUNCTION__);
            return true;
        }
    }

no_hdmi:
    DEBUG_PRINT("[%s]EDID is NOT HDMI\n", __FUNCTION__);
    return false;
}

/*****************************************************************************
 *  函 数 名: ljm_mode_vrefresh
 *  功能描述: 计算刷新值
 *  输入参数: struct ljm_display_mode
 *  输出参数: 无
 *  返 回 值: int
 *
 *  修改历史:
 *         1. 日    期: 2023年10月16日
 *            作    者: zhaiyf
 *            修改内容: 新生成函数,内核5.10.0以上版本去掉vrefresh,需要计算得出
 *
 *****************************************************************************/
int ljm_mode_vrefresh(const struct drm_display_mode *mode)
{
    int refresh = 0;
    unsigned int calc_val;

    if (mode->htotal > 0 && mode->vtotal > 0) {
        int vtotal;
        vtotal = mode->vtotal;
        /* work out vrefresh the value will be x1000 */
        calc_val = (mode->clock * 1000);
        calc_val /= mode->htotal;
        refresh = (calc_val + vtotal / 2) / vtotal;

        if (mode->flags & DRM_MODE_FLAG_INTERLACE)
            refresh *= 2;
        if (mode->flags & DRM_MODE_FLAG_DBLSCAN)
            refresh /= 2;
        if (mode->vscan > 1)
            refresh /= mode->vscan;
    }
    return refresh;
}
