/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#include <drm/drm_gem_framebuffer_helper.h>

#include "ljmicro_base.h"
#include "ljmicro_device.h"
#include "ljmicro_mode.h"
#include "ljmicro_object.h"
#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 6, 0)
#include <drm/drm_modeset_helper.h>
#endif

#define _LJ_OBJ_ZONE    ljvZONE_KMS

#if 0
static int
ljm_user_framebuffer_create_handle(
    struct drm_framebuffer *fb,
    struct drm_file *file_priv,
    unsigned int *handle
    )
{
    struct ljm_framebuffer *ljm_fb = to_ljm_framebuffer(fb);
    int r;

    HEADER_ARG("Fb=0x%p file_priv=0x%p handle=0x%p", fb, file_priv, handle);

    r = drm_gem_handle_create(file_priv, ljm_fb->obj, handle);

    FOOTER_ARG("r=%d", r);
    return r;
}

/* Assign to ljm_fb_funcs.destroy. Destroy ljm framebuffer object. */
static void ljm_user_framebuffer_destroy(struct drm_framebuffer *fb)
{
    struct ljm_framebuffer *ljm_fb = to_ljm_framebuffer(fb);

    HEADER_ARG("fb=%p", fb);

    if (ljm_fb->obj)
        drm_gem_object_put_unlocked(ljm_fb->obj);
    drm_framebuffer_cleanup(fb);
    kfree(ljm_fb);

    FOOTER();
}
#endif

/* framebuffer operation functions */
static const struct drm_framebuffer_funcs ljm_fb_funcs = {
    .create_handle = drm_gem_fb_create_handle, /* ljm_user_framebuffer_create_handle */
    .destroy = drm_gem_fb_destroy, /* ljm_user_framebuffer_destroy */
};

/* initialize ljm framebuffer */
int ljm_framebuffer_init(struct drm_device *ddev, struct drm_framebuffer *fb, CMD_ARGS *mode_cmd, struct drm_gem_object *obj)
{
    int r;

    HEADER_ARG("ddev=0x%p fb=0x%p mode_cmd=0x%p obj=0x%p", ddev, fb, mode_cmd, obj);

    TRACE("mode_cmd=%p (%dx%d@%4.4s, pitch[0]=%d)",
        mode_cmd, mode_cmd->width, mode_cmd->height,
        (char *)&mode_cmd->pixel_format, mode_cmd->pitches[0]);

    fb->obj[0] = obj;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,11,0)
    drm_helper_mode_fill_fb_struct(ddev, fb, mode_cmd);
#else
    drm_helper_mode_fill_fb_struct(fb, mode_cmd);
#endif

    r = drm_framebuffer_init(ddev, fb, &ljm_fb_funcs);
    if (r) {
        fb->obj[0] = NULL;
        DRM_ERROR("drm_framebuffer_init failed: %d\n", r);

        FOOTER_ARG("r=%d", r);
        return r;
    }

    FOOTER_ARG("fb(%d): (0x%px)", fb->base.id, fb);
    return 0;
}

/* Create GP framebuffer */
struct drm_framebuffer *
ljm_user_framebuffer_create(
    IN struct drm_device *ddev,
    IN struct drm_file *file_priv,
    IN const struct drm_mode_fb_cmd2 *mode_cmd
    )
{
    struct drm_framebuffer *fb;
    struct drm_gem_object *obj;
    int r;

    HEADER_ARG("ddev=%p file_priv=%p mode_cmd=%p mode_cmd->handles[0]=%d", ddev, file_priv, mode_cmd, mode_cmd->handles[0]);

    obj = DRM_GEM_OBJECT_LOOKUP(ddev, file_priv, mode_cmd->handles[0]);
    if (obj == NULL) {
        DRM_ERROR("No GEM object associated to handle 0x%08X, "
                  "can't create framebuffer\n", mode_cmd->handles[0]);

        FOOTER_ARG("r=%d", -ENOENT);
        return ERR_PTR(-ENOENT);
    }

    /* Handle is imported dma-buf, so cannot be migrated to VRAM for scanout */
    if (obj->import_attach) {
        DRM_DEBUG_KMS("Cannot create framebuffer from imported dma_buf\n");
        FOOTER_ARG("r=%d", -EINVAL);
        return ERR_PTR(-EINVAL);
    }

    fb = kzalloc(sizeof(*fb), GFP_KERNEL);
    if (fb == NULL) {
        drm_gem_object_unreference_unlocked(obj);

        DRM_ERROR("Can not alloc framebuffer\n");

        FOOTER_ARG("r=%d", -ENOMEM);
        return ERR_PTR(-ENOMEM);
    }

    r = ljm_framebuffer_init(ddev, fb, mode_cmd, obj);
    if (r) {
        kfree(fb);
        drm_gem_object_unreference_unlocked(obj);

        FOOTER_ARG("r=%d", r);
        return ERR_PTR(r);
    }

    DRM_DEBUG_KMS("%s: new drm_framebuffer=%px %d\n",
        __FUNCTION__, fb, fb->base.id);

    FOOTER_ARG("%px", fb);
    return fb;
}

static int ljm_fb_helper_set_par(struct fb_info *Info)
{
    int r;

    HEADER_ARG("Info=0x%p", Info);

    r = drm_fb_helper_set_par(Info);

    /* XXX: with universal plane support fbdev will automatically disable
     * all non-primary planes (including the cursor)
     */
    if (r == 0) {
        struct drm_fb_helper *fb_helper = Info->par;
        struct ljm_device *ljm_dev = GetLjmDevFrmDrm(fb_helper->dev);
        int i;

        for (i = 0; i < ljm_dev->num_crtcs; i++) {
            struct drm_crtc *crtc = ljm_dev->crtcs[i];

            ljm_crtc_cursor_set2(crtc, NULL, 0, 0, 0, 0, 0);
        }
    }

    FOOTER();
    return r;
}

static struct fb_ops ljmfb_ops = {
    .owner = THIS_MODULE,
    .fb_check_var = drm_fb_helper_check_var,
    .fb_set_par = ljm_fb_helper_set_par,
    .fb_fillrect = cfb_fillrect,
    .fb_copyarea = cfb_copyarea,
    .fb_imageblit = cfb_imageblit,
    .fb_pan_display = drm_fb_helper_pan_display,
    .fb_blank = drm_fb_helper_blank,   //dpms
    .fb_setcmap = drm_fb_helper_setcmap,
};

/* Destory a pinned GP Gem object for framebuffer */
static void ljmfb_destroy_pinned_object(struct drm_gem_object *gobj)
{
    struct ljm_gem_object *bo = to_ljm_gem(gobj);

    HEADER_ARG("gobj=0x%p", gobj);

    ljm_bo_kunmap(bo);
    ljm_put_vram(gobj->dev, bo);    // ljm_bo_unpin(bo), replaced by hym on 20220923
    drm_gem_object_unreference_unlocked(gobj);

    FOOTER();
}

/* Create a pinned LJM Gem object for framebuffer */
static int ljmfb_create_pinned_object(
    IN struct ljm_device *ljm_dev,          /* ljm device */
    IN struct drm_mode_fb_cmd2 *modecmd,    /* framebuffer mode command */
    OUT struct drm_gem_object **pobj        /* pointer of Gem object */
    )
{
    struct drm_gem_object *obj = NULL;
    struct ljm_gem_object *bo = NULL;
    int r;
    int size;

    HEADER_ARG("ljm_dev=0x%p modecmd=0x%p pobj=0x%p", ljm_dev, modecmd, pobj);

    size = modecmd->pitches[0] * modecmd->height;
    size = ALIGN(size, PAGE_SIZE);  /* align to PAGE_SIZE */

    // create and pin
    r = ljm_gem_object_create(ljm_dev, size, &obj);
    if (r) {
        DRM_ERROR("Failed to allocate framebuffer (%d)\n", size);
        FOOTER_ARG("r=%d", -ENOMEM);
        return -ENOMEM;
    }
    bo = to_ljm_gem(obj);

    //r = ljm_get_vram(ddev, bo); // ljm_bo_pin(bo, NULL), replaced by hym on 20220923
    r = ljm_bo_pin(bo, NULL);
    if (r) {
        DRM_ERROR("Failed to pin bo\n");
        goto out_unref;
    }

//  r = ljm_bo_kmap(bo, NULL);
//  if (r) {
//      DRM_ERROR("Failed to ioremap ljm gem\n");
//      goto out_unref;
//  }

    *pobj = obj;
    FOOTER_ARG("*pobj=0x%p", *pobj);
    return 0;

out_unref:
    ljmfb_destroy_pinned_object(obj);
    *pobj = NULL;
    FOOTER_ARG("r=%d", r);
    return r;
}

static int
ljm_fb_create(
    struct drm_fb_helper *Helper,
    struct drm_fb_helper_surface_size *Size
    )
{
    struct ljm_fbdev *fbdev = to_ljm_fbdev(Helper);
    struct drm_device *ddev = fbdev->helper.dev;
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(ddev);
    struct fb_info *info = NULL;
    struct drm_framebuffer *fb = NULL;
    struct drm_mode_fb_cmd2 mode_cmd;
    struct drm_gem_object *obj = NULL;
    struct ljm_gem_object *bo = NULL;
    int r;
    void *kptr;

    HEADER_ARG("Helper=0x%p Size=0x%p w=%d h=%d bbp=%d",
                   Helper, Size, Size->surface_width, Size->surface_height, Size->surface_bpp);

    DRM_INFO("Create framebuffer with size %dx%d %dBPP\n", Size->surface_width, Size->surface_height, Size->surface_bpp);

    /* we don't do packed 24bpp */
    if (Size->surface_bpp == 24)
        Size->surface_bpp = 32;

    mode_cmd.width = Size->surface_width;
    mode_cmd.height = Size->surface_height;
    /* Packed formats: only pitches[0] and offsets[0] entries are used. RGB is packed formats. */
    mode_cmd.pitches[0] = ALIGN(mode_cmd.width * ((Size->surface_bpp + 7) / 8), 8);     /* 8 bytes alignment for pitch by hym */
    mode_cmd.pitches[1] =
    mode_cmd.pitches[2] =
    mode_cmd.pitches[3] = 0;
    mode_cmd.offsets[0] =
    mode_cmd.offsets[1] =
    mode_cmd.offsets[2] =
    mode_cmd.offsets[3] = 0;
    mode_cmd.pixel_format = drm_mode_legacy_fb_format(Size->surface_bpp, Size->surface_depth);
    /* Semiplanar formats: the first two entries are used.
     * Planar: the first 3 entries are used.
     * The fourth entry is alpha plane?
     */

    r = ljmfb_create_pinned_object(ljm_dev, &mode_cmd, &obj);
    if (r) {
        DRM_ERROR("Failed to create fbcon object %d\n", r);
        FOOTER_ARG("Failed to create DRM GEM object (r=%d)", r);
        return r;
    }

    bo = to_ljm_gem(obj);

#if 0
    info = framebuffer_alloc(0, ddev->dev);

    if (!info) {
        DRM_ERROR("failed to alloc fb\n");
        r = -ENOMEM;
        goto out_unref;
    }

    r = fb_alloc_cmap(&info->cmap, 256, 0);
    if (r) {
        DRM_ERROR("%s: can't allocate color map\n", info->fix.id);
        framebuffer_release(info);
        goto out_unref;
    }

#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0)
    /* setup aperture base/size for vesafb takeover */
    info->apertures = alloc_apertures(1);
    if (!info->apertures) {
        fb_dealloc_cmap(&info->cmap);
        framebuffer_release(info);
        r = -ENOMEM;
        goto out_unref;
    }
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 2, 0)
    fbdev->helper.fbdev = info;
#else
    fbdev->helper.info = info;
#endif

#else
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0)
    /* okay we have an object now allocate the framebuffer */
    info = drm_fb_helper_alloc_fbi(Helper);
#else
    info = drm_fb_helper_alloc_info(Helper);
#endif
    /* drm_fb_helper_alloc_fbi函数会将info->skip_vt_switch变量修改为true，
       这导致s3唤醒后帧缓冲处理异常。skip_vt_switch变量在内核中用于控制
       resume/suspend过程中是否跳过虚拟终端（VT）切换，
       其引起帧缓冲异常的原因尚不明确。因此此处修改只是规避该问题。*/
    info->skip_vt_switch = false;
    if (IS_ERR(info)) {
        r = PTR_ERR(info);
        goto out_unref;
    }
#endif

    info->par = fbdev;

    r = ljm_framebuffer_init(ddev, &fbdev->fb, &mode_cmd, obj);
    if (r) {
        DRM_ERROR("failed to initialize framebuffer %d\n", r);
        goto out_destroy_fbi;
    }

    fb = &fbdev->fb;

    /* setup helper */
    fbdev->helper.fb = fb;

    ljm_bo_kmap(bo, &kptr);
    memset_io(kptr, 0x0, bo->base.size);

    strcpy(info->fix.id, FB_INFO_NAME);

    info->screen_base = kptr;   //kernel virtual address
    info->screen_size = bo->base.size;
    info->fix.smem_len = bo->base.size;
    //info->fix.smem_start = ljm_dev->smem_start + get_offset_from_bo(bo);  //physical address
    info->fix.smem_start = get_cpu_addr_from_bo(bo);    //physical address

#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 2, 0)
    ddev->mode_config.fb_base = info->fix.smem_start;
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0)
    info->flags = FBINFO_DEFAULT;
#endif
    info->fbops = &ljmfb_ops;
#if LINUX_VERSION_CODE > KERNEL_VERSION (5, 4, 0)
    drm_fb_helper_fill_info(info, &fbdev->helper, Size);
#else
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,11,0)
    drm_fb_helper_fill_fix(info, fb->pitches[0], fb->format->depth);
#else
    drm_fb_helper_fill_fix(info, fb->pitches[0], fb->depth);
#endif
    drm_fb_helper_fill_var(info, &fbdev->helper, Size->fb_width, Size->fb_height);
#endif

    DRM_INFO("fb mappable at 0x%lX\n", info->fix.smem_start);
    DRM_INFO("vram apper at 0x%lX\n", (unsigned long)info->screen_base);
    DRM_INFO("size %lu\n", (unsigned long)info->screen_size);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,11,0)
    DRM_INFO("fb depth is %d\n", fb->format->depth);
#else
    DRM_INFO("fb depth is %d\n", fb->depth);
#endif
    DRM_INFO("   pitch is %d\n", fb->pitches[0]);

    FOOTER();
    return 0;

out_destroy_fbi:
#if 1
    if (info) {
        if (info->cmap.len)
            fb_dealloc_cmap(&info->cmap);
        framebuffer_release(info);
    }

#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 2, 0)
    fbdev->helper.fbdev = NULL;
#else
    fbdev->helper.info = NULL;
#endif
#else
    drm_fb_helper_release_fbi(Helper);
#endif

out_unref:
    if (obj) {
        drm_gem_object_unreference_unlocked(obj);
    }
    if (fb) {
        drm_framebuffer_unregister_private(fb);
        drm_framebuffer_cleanup(fb);
        kfree(fb);
    }
    FOOTER_ARG("r=%d", r);
    return r;
}

static struct drm_fb_helper_funcs ljm_fb_helper_funcs = {
    .fb_probe = ljm_fb_create,
};

struct ljm_fbdev *ljm_fbdev_init(struct drm_device *ddev)
{
    struct ljm_fbdev *fbdev;
#if LINUX_VERSION_CODE < KERNEL_VERSION(5,10,0)
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(ddev);
#endif
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0)
    int bpp_sel = 24;
#endif
    int r;

    HEADER_ARG("ddev=0x%p", ddev);

    fbdev = kzalloc(sizeof(struct ljm_fbdev), GFP_KERNEL);
    if (!fbdev) {
        DRM_ERROR("ljm_fbdev_init: Out of memory!\n");
        FOOTER_ARG("ERR(%d): Out of memory", -ENOMEM);
        return NULL;
    }


#if LINUX_VERSION_CODE < KERNEL_VERSION(6,6,0)
    drm_fb_helper_prepare(ddev, &fbdev->helper, &ljm_fb_helper_funcs); //fodora 19->21
#else
    drm_fb_helper_prepare(ddev, &fbdev->helper, 0, &ljm_fb_helper_funcs);
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 10, 0)
    r = drm_fb_helper_init(ddev, &fbdev->helper);
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(4,11,0)
    r = drm_fb_helper_init(ddev, &fbdev->helper, ljm_dev->num_ports);
#else
    r = drm_fb_helper_init(ddev, &fbdev->helper, ljm_dev->num_crtcs, ljm_dev->num_ports);
#endif
    if (r) {
        DRM_ERROR("drm init fb helper: %d\n", r);
        kfree(fbdev);
        FOOTER();
        return NULL;
    }

#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 10, 0)
    r = drm_fb_helper_single_add_all_connectors(&fbdev->helper);
    if (r) {
        DRM_ERROR("drm add all connectors: %d\n", r);
        drm_fb_helper_fini(&fbdev->helper);
        kfree(fbdev);
        FOOTER();
        return NULL;
    }
#endif

    /* disable all the possible outputs/crtcs before entering KMS mode */
    drm_helper_disable_unused_functions(ddev);  /* disable encoder and crtc by calling their dpms functions */

#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0)
    r = drm_fb_helper_initial_config(&fbdev->helper, bpp_sel);    /* call ljm_fb_helper_funcs.fb_probe (ljmfb_create) */
#else
    r = drm_fb_helper_initial_config(&fbdev->helper);
#endif
    if (r) {
        DRM_ERROR("drm initial config: %d\n", r);
        drm_fb_helper_fini(&fbdev->helper);
        kfree(fbdev);
        FOOTER();
        return NULL;
    }

    DRM_DEBUG_KMS("%s: new drm_framebuffer=%px (%px)\n",
        __FUNCTION__, &fbdev->fb, fbdev->helper.fb);

    FOOTER_ARG("new fbdev(%p)", fbdev);
    return fbdev;
}

static int ljm_fbdev_destroy(struct drm_device *ddev, struct ljm_fbdev *fbdev)
{
    struct drm_framebuffer *fb = &fbdev->fb;

    HEADER_ARG("ddev=0x%p fbdev=0x%p", ddev, fbdev);

#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 2, 0)
    drm_fb_helper_unregister_fbi(&fbdev->helper);
#else
    drm_fb_helper_unregister_info(&fbdev->helper);
#endif

    if (fb->obj[0]) {
        ljmfb_destroy_pinned_object(fb->obj[0]);
        fb->obj[0] = NULL;
        // remove ?
        drm_framebuffer_unregister_private(fb);
        drm_framebuffer_cleanup(fb);
    }

    drm_fb_helper_fini(&fbdev->helper);

    FOOTER();
    return 0;
}

/* finish and free framebuffer device */
void ljm_fbdev_fini(IN struct ljm_device *ljm_dev)
{
    struct ljm_fbdev *fbdev = ljm_dev->fbdev;

    HEADER_ARG("LjmDev=0x%p", ljm_dev);

    if (!fbdev) {
        FOOTER();
        return;
    }

    ljm_fbdev_destroy(ljm_dev->ddev, fbdev);
    kfree(fbdev);
    ljm_dev->fbdev = NULL;

    FOOTER();
}

#if 0
void ljm_fb_clear(struct ljm_device *ljm_dev)
{
    resource_size_t     smem_start;
    void *kptr;

    HEADER_ARG("ljm_dev=%p", ljm_dev);

    //printk("ljm_fb_clear: ljm_dev=%p\n", ljm_dev);

    /*********** set GART table **************/
    /* get GP video memory */
    smem_start = pci_resource_start(ljm_dev->ddev->pdev, 0);

    /* initialize all the page tables */
    kptr = ioremap_wc(smem_start + ljm_dev->fb_bo_offset, ljm_dev->fb_bo_size);
    if (kptr == NULL) {
        printk(KERN_ERR "----ioremap error: smem_start=%llx fb_bo_offset=%lx fb_bo_size=%lx\n",
            smem_start, ljm_dev->fb_bo_offset, ljm_dev->fb_bo_size);
        FOOTER();
        return;
    }
    TRACE("Map smem_start=%llx+fb_bo_offset=%lx@fb_bo_size=%lx to %lx\n",
        smem_start, ljm_dev->fb_bo_offset, ljm_dev->fb_bo_size, (unsigned long)kptr);

    memset_io(kptr, 0, ljm_dev->fb_bo_size);

    iounmap(kptr);
    FOOTER();
}
#endif

bool ljm_fbdev_bo_is_fb(struct ljm_device *ljm_dev, struct ljm_gem_object *ljm_obj)
{
    if (!ljm_dev->fbdev)
        return false;

    if (ljm_obj == to_ljm_gem(drm_gem_fb_get_obj(&ljm_dev->fbdev->fb, 0)))
        return true;
    return false;
}

