/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#include "ljmicro_base.h"
#include "ljmicro_gpio.h"
#include "ljmicro_device.h"
#include "ljmicro_vgav2.h"
#include "gp200.h"

/**
 * @brief 设置 GPIO MASK 位值
 * @param int card      : 使用的 DEVICE ID
 * @param uint8_t gpiox : 需要设置的引脚号
 * @param int mode      : 为 0 时设置对应位为（无效）关闭，为 1 时设置对应位为（有效）开启
 * @return 空
*/
void ljm_gpio_set_mask(int card, uint8_t gpiox, int mode)
{
    uint8_t gpio_mask;
    struct ljm_device *ljm_dev = get_ljm_devs(0);

    /* 读取GPIO MASK寄存器 */
    gpio_mask = LJM_IOMUX_R8(card, REG_GPIO_MASK);
    if ((ljm_dev->flags & LJMICRO_CHIP_MASK) == CHIP_GP202)
    {
        /* GP202 MASK 为 ff 时禁止数据写入，为 00 时允许数据写入 */
        gpio_mask = mode ? (gpio_mask & ~(0x1 << gpiox)) : (gpio_mask | (0x1 << gpiox));
        LJM_IOMUX_W8(card, REG_GPIO_MASK, gpio_mask);
    }
    else
    {
        /* GP201 MASK 为 00 时禁止数据写入，为 ff 时允许数据写入 */
        gpio_mask = mode ? (gpio_mask | (0x1 << gpiox)) : (gpio_mask & ~(0x1 << gpiox));
        LJM_IOMUX_W8(card, REG_GPIO_MASK, gpio_mask);
    }
}

/**
 * @brief 设置 GPIO OEN 位值
 * @param int card      : 使用的 DEVICE ID
 * @param uint8_t gpiox : 需要设置的引脚号
 * @param int mode      : 为 0 时设置对应位为（无效）关闭，为 1 时设置对应位为（有效）开启
 * @return 空
*/
void ljm_gpio_set_oen(int card, uint8_t gpiox, int mode)
{
    uint8_t gpio_oen;
    struct ljm_device *ljm_dev = get_ljm_devs(0);

    if ((ljm_dev->flags & LJMICRO_CHIP_MASK) == CHIP_GP202)
    {
        gpio_oen = LJM_IOMUX_R8(card, REG_GP202_GPIO_PAD_OEN);
        gpio_oen = mode ? (gpio_oen | (0x1 << gpiox)) : (gpio_oen & ~(0x1 << gpiox));

        LJM_IOMUX_W8(card, REG_GP202_GPIO_PAD_OEN, gpio_oen);
    }
    else
    {
        gpio_oen = LJM_IOMUX_R8(card, REG_GP201_GPIO_PAD_OEN);
        gpio_oen = mode ? (gpio_oen | (0x1 << gpiox)) : (gpio_oen & ~(0x1 << gpiox));

        LJM_IOMUX_W8(card, REG_GP201_GPIO_PAD_OEN, gpio_oen);
    }
}

/**
 * @brief 设置 GPIO OUTPUT 位值
 * @param int card      : 使用的 DEVICE ID
 * @param uint8_t gpiox : 需要设置的引脚号
 * @param int mode      : 为 0 时设置对应位为（无效）关闭，为 1 时设置对应位为（有效）开启
 * @return 空
*/
void ljm_gpio_set_output(int card, uint8_t gpiox, int mode)
{
    uint8_t gpio_out;
    gpio_out = LJM_IOMUX_R8(card, REG_GPIO_DATA_O);
    gpio_out = mode ? (gpio_out | (0x1 << gpiox)) : (gpio_out & ~(0x1 << gpiox));

    LJM_IOMUX_W8(card, REG_GPIO_DATA_O, gpio_out);
}

/**
 * @brief 获取 GPIO MASK 位值
 * @param int card    : 使用的 DEVICE ID
 * @return uint8_t    : 读取到的GPIO MASK寄存器状态值
 *           GP202 :  返回值为 00 时，寄存器实际值为 ff (GP202 Mask 寄存器为 ff 时禁止数据写入)
 *                    返回值为 ff 时，寄存器实际值为 00 (GP202 Mask 寄存器为 00 时允许数据写入)
 *           GP201 :  返回值为 00 时，寄存器实际值为 00 (GP201 Mask 寄存器为 00 时禁止数据写入)
 *                    返回值为 ff 时，寄存器实际值为 ff (GP201 Mask 寄存器为 ff 时允许数据写入)
 */
uint8_t ljm_gpio_get_mask(int card)
{
    struct ljm_device *ljm_dev = get_ljm_devs(0);
    if ((ljm_dev->flags & LJMICRO_CHIP_MASK) == CHIP_GP202)
    {
        return ~(LJM_IOMUX_R8(card, REG_GPIO_MASK));
    }
    else
    {
        return LJM_IOMUX_R8(card, REG_GPIO_MASK);
    }
}

/**
 * @brief 获取 GPIO OEN 位值
 * @param int card    : 使用的 DEVICE ID
 * @return uint8_t    : 读取到的GPIO OEN寄存器状态值
*/
uint8_t ljm_gpio_get_oen(int card)
{
    struct ljm_device *ljm_dev = get_ljm_devs(0);
    if ((ljm_dev->flags & LJMICRO_CHIP_MASK) == CHIP_GP202)
    {
        return LJM_IOMUX_R8(card, REG_GP202_GPIO_PAD_OEN);
    }
    else
    {
        return LJM_IOMUX_R8(card, REG_GP201_GPIO_PAD_OEN);
    }
}

/**
 * @brief 获取 GPIO INPUT 位值
 * @param int card    : 使用的 DEVICE ID
 * @return uint8_t    : 读取到的GPIO INPUT寄存器状态值
*/
uint8_t ljm_gpio_get_input(int card)
{
    return LJM_IOMUX_R8(card, REG_GPIO_DATA_I);
}

/**
 * @brief 获取 GPIO OUTPUT 位值
 * @param int card    : 使用的 DEVICE ID
 * @return uint8_t    : 读取到的GPIO OUTPUT寄存器状态值
*/
uint8_t ljm_gpio_get_output(int card)
{
    return LJM_IOMUX_R8(card, REG_GPIO_DATA_O);
}
