/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

/**************************************************************************
* 版权所有 (C)2022, 武汉凌久微电子有限公司。
*
* 文件名称： ljmicro_hdmi_codec.c
* 文件标识：
* 内容摘要： 本文件定义hdmi codec驱动函数
* 其它说明：
* 当前版本： v1.0
* 作    者： zhoupw, 武汉凌久微电子有限公司
* 完成日期： 2022年4月27日
***************************************************************************/

#if IS_ENABLED(CONFIG_SND_SOC) && defined(ljdENABLE_AUDIO)

#include <linux/dma-mapping.h>
#include <drm/drm_crtc.h>
#include <linux/err.h>

#include "ljmicro_device.h"
#include "ljmicro_hdmi.h"
#include "ljmicro_hdmiv2.h"
#include "ljmicro_connector_codec.h"

static inline void hdmi_writeb(struct hdmi_dev *hdmi, u8 val, int offset)
{
    writeb(val, hdmi->regbase + hdmi->offset + offset);
}

static inline u8 hdmi_readb(struct hdmi_dev *hdmi, int offset)
{
    u8 val = 0;

    val = readb(hdmi->regbase + hdmi->offset + offset);

    return val;
}

static void hdmi_modb(struct hdmi_dev *hdmi, u8 data, u8 mask, unsigned reg)
{
    u8 val = 0;
    u8 tmp = 0;

    val = readb(hdmi->regbase + hdmi->offset + reg);
    tmp = val & ~mask;
    tmp |= data & mask;

    writeb(tmp, hdmi->regbase + hdmi->offset + reg);
}

#if LINUX_VERSION_CODE > KERNEL_VERSION(5, 0, 0)
static void handle_plugged_change(struct hdmi_dev *hdmi, bool plugged)
{
    if (hdmi->plugged_cb && hdmi->codec_dev)
        hdmi->plugged_cb(hdmi->codec_dev, plugged);
}

int ljm_hdmi_set_plugged_cb(struct hdmi_dev *hdmi, hdmi_codec_plugged_cb fn, struct device *codec_dev)
{
    bool plugged;

    mutex_lock(&hdmi->mutex);
    hdmi->plugged_cb = fn;
    hdmi->codec_dev = codec_dev;
    plugged = hdmi->last_connector_result == connector_status_connected;
    handle_plugged_change(hdmi, plugged);
    mutex_unlock(&hdmi->mutex);

    return 0;
}
#endif

static void hdmi_set_cts_n(struct hdmi_dev *hdmi, unsigned int cts, unsigned int n)
{
    hdmi_modb(hdmi, 0, HDMI_AUD_CTS3_CTS_MANUAL, HDMI_AUD_CTS3);

    hdmi_modb(hdmi, 0, HDMI_AUD_CTS3_N_SHIFT_MASK, HDMI_AUD_CTS3);

    if (cts)
        hdmi_writeb(hdmi, ((cts >> 16) & HDMI_AUD_CTS3_AUDCTS19_16_MASK) | HDMI_AUD_CTS3_CTS_MANUAL, HDMI_AUD_CTS3);
    else
        hdmi_writeb(hdmi, 0, HDMI_AUD_CTS3);
    hdmi_writeb(hdmi, (cts >> 8) & 0xff, HDMI_AUD_CTS2);
    hdmi_writeb(hdmi, cts & 0xff, HDMI_AUD_CTS1);

    hdmi_writeb(hdmi, (n >> 16) & 0x0f, HDMI_AUD_N3);
    hdmi_writeb(hdmi, (n >> 8) & 0xff, HDMI_AUD_N2);
    hdmi_writeb(hdmi, n & 0xff, HDMI_AUD_N1);
}

static unsigned int hdmi_compute_n(unsigned int freq, unsigned long pixel_clk)
{
    unsigned int n = (128 * freq) / 1000;
    unsigned int mult = 1;

    while (freq > 48000) {
        mult *= 2;
        freq /= 2;
    }

    switch (freq) {
        case 32000:
            if (pixel_clk == 25175000)
                n = 4576;
            else if (pixel_clk == 27027000)
                n = 4096;
            else if (pixel_clk == 74176000 || pixel_clk == 148352000)
                n = 11648;
            else
                n = 4096;
            n *= mult;
            break;

        case 44100:
            if (pixel_clk == 25175000)
                n = 7007;
            else if (pixel_clk == 74176000)
                n = 17836;
            else if (pixel_clk == 148352000)
                n = 8918;
            else
                n = 6272;
            n *= mult;
            break;

        case 48000:
            if (pixel_clk == 25175000)
                n = 6864;
            else if (pixel_clk == 27027000)
                n = 6144;
            else if (pixel_clk == 74176000)
                n = 11648;
            else if (pixel_clk == 148352000)
                n = 5824;
            else
                n = 6144;
            n *= mult;
            break;

        default:
            break;
    }

    return n;
}

static void hdmi_set_clk_regenerator(struct hdmi_dev *hdmi, unsigned long pixel_clk, unsigned int sample_rate)
{
    unsigned long ftdms = pixel_clk;
    unsigned int n, cts;
    u8 config3;
    u64 tmp;

    n = hdmi_compute_n(sample_rate, pixel_clk);

    config3 = hdmi_readb(hdmi, CONFIG3_ID);

    if (config3 & HDMI_CONFIG3_AHBAUDDMA) {
        tmp = (u64)ftdms * n;
        do_div(tmp, 128 * sample_rate);
        cts = tmp;
    }
    else {
        cts = 0;
    }

    spin_lock_irq(&hdmi->audio_lock);
    hdmi->audio_n = n;
    hdmi->audio_cts = cts;
    hdmi_set_cts_n(hdmi, cts, hdmi->audio_enable ? n : 0);
    spin_unlock_irq(&hdmi->audio_lock);
}

static void ljm_hdmi_set_channel_status(struct hdmi_dev *hdmi, u8 *channel_status)
{
    hdmi_writeb(hdmi, channel_status[3], HDMI_FC_AUDSCHNLS7);
    hdmi_writeb(hdmi, channel_status[4], HDMI_FC_AUDSCHNLS8);
}

static void ljm_hdmi_set_sample_rate(struct hdmi_dev *hdmi, unsigned int rate)
{
    mutex_lock(&hdmi->audio_mutex);
    hdmi->sample_rate = rate;
    hdmi_set_clk_regenerator(hdmi, hdmi->pixelclk, hdmi->sample_rate);
    mutex_unlock(&hdmi->audio_mutex);
}

static void ljm_hdmi_set_channel_count(struct hdmi_dev *hdmi, unsigned int cnt)
{
    u8 layout;

    mutex_lock(&hdmi->audio_mutex);

    if (cnt > 2)
        layout = HDMI_FC_AUDSCONF_AUD_PACKET_LAYOUT_LAYOUT1;
    else
        layout = HDMI_FC_AUDSCONF_AUD_PACKET_LAYOUT_LAYOUT0;

    hdmi_modb(hdmi, layout, HDMI_FC_AUDSCONF_AUD_PACKET_LAYOUT_MASK, FC_AUDSCONF);

    hdmi_modb(hdmi, (cnt - 1) << HDMI_FC_AUDICONF0_CC_OFFSET, HDMI_FC_AUDICONF0_CC_MASK, FC_AUDICONF0);

    mutex_unlock(&hdmi->audio_mutex);
}

static void ljm_hdmi_set_channel_allocation(struct hdmi_dev *hdmi, unsigned int ca)
{
    mutex_lock(&hdmi->audio_mutex);

    hdmi_writeb(hdmi, ca, FC_AUDICONF2);

    mutex_unlock(&hdmi->audio_mutex);
}

static void ljm_hdmi_audio_enable_w(struct hdmi_dev *hdmi)
{
    unsigned long flags;

    spin_lock_irqsave(&hdmi->audio_lock, flags);
    hdmi->audio_enable = true;
    if (hdmi->enable_audio)
        hdmi->enable_audio(hdmi);
    spin_unlock_irqrestore(&hdmi->audio_lock, flags);
}

void ljm_hdmi_audio_disable_w(struct hdmi_dev *hdmi)
{
    unsigned long flags;

    spin_lock_irqsave(&hdmi->audio_lock, flags);
    hdmi->audio_enable = false;
    if (hdmi->disable_audio)
        hdmi->disable_audio(hdmi);
    spin_unlock_irqrestore(&hdmi->audio_lock, flags);
}

static void hdmi_enable_audio_clk(struct hdmi_dev *hdmi, bool enable)
{
    if (enable)
        hdmi->mc_clkdis &= ~HDMI_MC_CLKDIS_AUDCLK_DISABLE;
    else
        hdmi->mc_clkdis |= HDMI_MC_CLKDIS_AUDCLK_DISABLE;
    hdmi_writeb(hdmi, hdmi->mc_clkdis, MC_CLKDIS);
}

static void ljm_hdmi_audio_enable(struct hdmi_dev *hdmi)
{
    hdmi_set_cts_n(hdmi, hdmi->audio_cts, hdmi->audio_n);
    hdmi_enable_audio_clk(hdmi, true);
}

static void ljm_hdmi_audio_disable(struct hdmi_dev *hdmi)
{
    hdmi_enable_audio_clk(hdmi, false);
}

static inline void hdmi_write(struct ljm_hdmi_codec_data *audio, u8 val, int offset)
{
    struct hdmi_dev *hdmi = audio->hdmi;

    audio->write(hdmi, val, offset);
}

static inline u8 hdmi_read(struct ljm_hdmi_codec_data *audio, int offset)
{
    struct hdmi_dev *hdmi = audio->hdmi;

    return audio->read(hdmi, offset);
}

static int ljm_hdmi_codec_hw_params(struct device *dev, void *data,
        struct hdmi_codec_daifmt *fmt, struct hdmi_codec_params *hparams)
{
    struct ljm_hdmi_codec_data *audio = data;
    struct hdmi_dev *hdmi = audio->hdmi;
    u8 conf0 = 0;
    u8 conf1 = 0;
    u8 inputclkfs = 0;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 6, 0)
    if (fmt->bit_clk_provider | fmt->frame_clk_provider) {
#else
    if (fmt->bit_clk_master | fmt->frame_clk_master) {
#endif
        dev_err(dev, "LJM HDMI CODEC: unsupported clock settings\n");
        return -EINVAL;
    }

    hdmi_write(audio, HDMI_AUD_CONF0_SW_RESET, HDMI_AUD_CONF0);
    hdmi_write(audio, (u8)~HDMI_MC_SWRSTZ_I2SSWRST_REQ, MC_SWRSTZREQ);

    inputclkfs = HDMI_AUD_INPUTCLKFS_64FS;
    conf0 = (HDMI_AUD_CONF0_I2S_SELECT | HDMI_AUD_CONF0_I2S_EN0);

    switch (hparams->channels) {
        case 7 ... 8:
            conf0 |= HDMI_AUD_CONF0_I2S_EN3;
            /* fallthrough; */
        case 5 ... 6:
            conf0 |= HDMI_AUD_CONF0_I2S_EN2;
            /* fallthrough; */
        case 3 ... 4:
            conf0 |= HDMI_AUD_CONF0_I2S_EN1;
    }

    switch (hparams->sample_width) {
        case 16:
            conf1 = HDMI_AUD_CONF1_WIDTH_16;
            break;
        case 24:
        case 32:
            conf1 = HDMI_AUD_CONF1_WIDTH_24;
            break;
    }

    switch (fmt->fmt) {
        case HDMI_I2S:
            conf1 |= HDMI_AUD_CONF1_MODE_I2S;
            break;
        case HDMI_RIGHT_J:
            conf1 |= HDMI_AUD_CONF1_MODE_RIGHT_J;
            break;
        case HDMI_LEFT_J:
            conf1 |= HDMI_AUD_CONF1_MODE_LEFT_J;
            break;
        case HDMI_DSP_A:
            conf1 |= HDMI_AUD_CONF1_MODE_BURST_1;
            break;
        case HDMI_DSP_B:
            conf1 |= HDMI_AUD_CONF1_MODE_BURST_2;
            break;
        default:
            dev_err(dev, "LJM HDMI CODEC: unsupported format\n");
            return -EINVAL;
    }

    ljm_hdmi_set_sample_rate(hdmi, hparams->sample_rate);
    ljm_hdmi_set_channel_status(hdmi, hparams->iec.status);
    ljm_hdmi_set_channel_count(hdmi, hparams->channels);
    ljm_hdmi_set_channel_allocation(hdmi, hparams->cea.channel_allocation);

    hdmi_write(audio, inputclkfs, HDMI_AUD_INPUTCLKFS);
    hdmi_write(audio, conf0, HDMI_AUD_CONF0);
    hdmi_write(audio, conf1, HDMI_AUD_CONF1);

    return 0;
}

static int ljm_hdmi_codec_audio_startup(struct device *dev, void *data)
{
    struct ljm_hdmi_codec_data *audio = data;
    struct hdmi_dev *hdmi = audio->hdmi;

    ljm_hdmi_audio_enable_w(hdmi);

    return 0;
}

static void ljm_hdmi_codec_audio_shutdown(struct device *dev, void *data)
{
    struct ljm_hdmi_codec_data *audio = data;
    struct hdmi_dev *hdmi = audio->hdmi;

    ljm_hdmi_audio_disable_w(hdmi);
}

static int ljm_hdmi_codec_get_eld(struct device *dev, void *data, uint8_t *buf, size_t len)
{
    struct ljm_hdmi_codec_data *audio = data;

    memcpy(buf, audio->eld, min_t(size_t, MAX_ELD_BYTES, len));

    return 0;
}

static int ljm_hdmi_codec_get_dai_id(struct snd_soc_component *component, struct device_node *endpoint)
{
    struct of_endpoint of_ep;
    int ret;

    ret = of_graph_parse_endpoint(endpoint, &of_ep);
    if (ret < 0)
        return ret;

    if (of_ep.port == 2)
        return 0;

    return -EINVAL;
}

#if LINUX_VERSION_CODE > KERNEL_VERSION(5, 0, 0)
static int ljm_hdmi_codec_hook_plugged_cb(struct device *dev, void *data,
        hdmi_codec_plugged_cb fn,
        struct device *codec_dev)
{
    struct ljm_hdmi_codec_data *audio = data;
    struct hdmi_dev *hdmi = audio->hdmi;

    return ljm_hdmi_set_plugged_cb(hdmi, fn, codec_dev);
}
#endif

static struct hdmi_codec_ops ljm_hdmi_codec_ops = {
    .hw_params = ljm_hdmi_codec_hw_params,
    .audio_startup = ljm_hdmi_codec_audio_startup,
    .audio_shutdown = ljm_hdmi_codec_audio_shutdown,
    .get_eld = ljm_hdmi_codec_get_eld,
    .get_dai_id = ljm_hdmi_codec_get_dai_id,
#if LINUX_VERSION_CODE > KERNEL_VERSION(5, 0, 0)
    .hook_plugged_cb = ljm_hdmi_codec_hook_plugged_cb,
#endif
};

/*
static const struct regmap_config hdmi_regmap_8bit_config = {
    .reg_bits = 32,
    .val_bits = 8,
    .reg_stride = 1,
    .max_register = I2CM_FS_SCL_LCNT_0_ADDR,
};
*/

int ljm_hdmi_codec_init(struct ljm_connector_codec_data *audio_data, struct ljm_device *ljm_dev)
{
    int i;
    int j = 0;

    audio_data->ljm_hdmi_codec_ops = &ljm_hdmi_codec_ops;

    for (i = 0; i < LJM_MAX_PORTS; i++) {
        if (!ljm_dev->ports[i]) {
            continue;
        }
        /* 初始化与HDMI音频相关部分 */
        if (ljm_dev->ports[i]->type == LJM_PORT_HDMI) {
            /*
            ljm_dev->ports[i]->vdev.hdmi.regm = devm_regmap_init_mmio(
                ljm_dev->dev, ljm_dev->ports[i]->vdev.hdmi.regbase +
                ljm_dev->ports[i]->vdev.hdmi.offset,
                &hdmi_regmap_8bit_config);
            if (IS_ERR(ljm_dev->ports[i]->vdev.hdmi.regm)) {
                dev_err(ljm_dev->dev, "LJM HDMI CODEC: failed to configure hdmi regmap\n");
                return PTR_ERR(ljm_dev->ports[i]->vdev.hdmi.regm);
            }
	    */
            ljm_dev->ports[i]->vdev.hdmi.enable_audio = ljm_hdmi_audio_enable;
            ljm_dev->ports[i]->vdev.hdmi.disable_audio = ljm_hdmi_audio_disable;
            ljm_dev->ports[i]->vdev.hdmi.dev = ljm_dev->dev;

            audio_data->hdmi_devs[j] = &ljm_dev->ports[i]->vdev.hdmi;
            j++;
        }
        else {
            continue;
        }
    }
    audio_data->num_hdmis = j;
    audio_data->hdmi_audio_data->write = hdmi_writeb;
    audio_data->hdmi_audio_data->read = hdmi_readb;

    return 0;
}

#endif
