/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

/******************************************************************************
 *
 *  版权所有(c)，2020-2099，中船第709研究所
 *
 *******************************************************************************
 *
 *  文件名称: ljmicro_i2c.c
 *  版 本 号: V1.0
 *  作    者: 胡艳明
 *  生成日期: 2022年4月15日
 *
 *  功能描述: I2C控制器驱动
 *
 *  修改历史:
 *         1. 日    期: 2022年4月15日
 *            作    者: 胡艳明
 *            修改内容: 创建文件
 *
 ******************************************************************************/

/*******************************************************************************
 *                                   非标准库头文件                                *
 ******************************************************************************/

#include "ljmicro_i2c.h"

#define _LJ_OBJ_ZONE ljvZONE_KMS

/******************************************************************************
 *                                     宏定义                                     *
 ******************************************************************************/

/* DDC controller registers */
#define OC_RW_PRER_LOW_ADR  0
#define OC_RW_PRER_HIGH_ADR 1
#define OC_RW_CTR_ADR       2
#define OC_RW_CTR_EN_MASK   (1U << 7) /* EN, I2C core enable bit. When set to ‘1’, the core is enabled */
#define OC_RW_CTR_IEN_MASK  (1U << 6) /* IEN, I2C core interrupt enable bit. When set to ‘1’, interrupt is enabled */

#define OC_TXR_ADR      3
#define OC_CR_ADR       4         /* DDC CR (control register) */
#define OC_CR_IACK_MASK (1U << 0) /* IACK, Interrupt acknowledge. When set, clears a pending interrupt */
#define OC_CR_ACK_MASK  (1U << 3) /* ACK, when a receiver, sent ACK (ACK = ‘0’) or NACK (ACK = ‘1’) */
#define OC_CR_WR_MASK   (1U << 4) /* WR, write to slave */
#define OC_CR_RD_MASK   (1U << 5) /* RD, read from slave */
#define OC_CR_STO_MASK  (1U << 6) /* STO, generate stop condition */
#define OC_CR_STA_MASK  (1U << 7) /* STA, generate (repeated) start condition */

#define OC_RXR_ADR        3
#define OC_SR_ADR         4         /* DDC SR (state register) */
#define OC_SR_IRQ_MASK    (1U << 0) /* interrupt flag. This bit is set when an interrupt is pending */
#define OC_SR_TIP_MASK    (1U << 1) /* transfer is progress. '1' when transferring data, '0' when transfer complete */
#define OC_SR_RXNACK_MASK (1U << 7) /* received acknowledge from slave, 1=>NACK, 0=>ACK */

//#define OC_R_TXR_ADR            5
//#define OC_R_CR_ADR             6

#ifndef MAX_EDID
#define MAX_EDID 512
#endif

/******************************************************************************
 *                                    函数实现                                    *
 ******************************************************************************/

/*****************************************************************************
 *  函 数 名: ljm_i2c_oc_init
 *  功能描述: 初始化OC I2C控制器和结构体
 *  输入参数: struct i2c_dev *i2c_dev  I2C结构体指针
 *            void *base               寄存器空间虚拟基地址/UEFI GOP Private指针
 *            int offset               I2C控制器寄存器地址起始偏移
 *            int index                I2C控制器序号
 *  输出参数: 无
 *  返 回 值: int类型，0表示成功，否则返回错误代码
 *
 *  修改历史:
 *         1. 日    期: 2022年4月15日
 *            作    者: 刘天玥
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
int ljm_i2c_oc_init(struct i2c_dev *i2c_dev, void *base, int offset, int index)
{
    HEADER_ARG("i2c_dev=%lx base=0x%lx offset=%d", (unsigned long)i2c_dev, (unsigned long)base, offset);

#ifdef UEFI_GOP
    i2c_dev->Private = base;
#else
    i2c_dev->regbase = base;
#endif

    i2c_dev->offset = offset;
    i2c_dev->index  = index;

    /* initialize i2c controller, such as set SCL clock, etc. */
    TRACE("----i2c_oc %d: RW_CTR=0x%x----\n", i2c_dev->index, I2C_REG_R8(i2c_dev, OC_RW_CTR_ADR));
    //I2C_REG_W8(i2c_dev, OC_RW_CTR_ADR, OC_RW_CTR_EN_MASK);

    FOOTER();
    return 0;
}

#if 0
/*****************************************************************************
 *  函 数 名: _wait_SR
 *  功能描述: 等待，直到I2C命令传输完成
 *  输入参数: struct i2c_dev *i2c_dev  I2C结构体指针
 *  输出参数: unsigned char *pSR       命令传输状态
 *  返 回 值: int类型，返回0表示target设备正常应答，否则表示超时或设备无应答
 *
 *  修改历史:
 *         1. 日    期: 2022年4月15日
 *            作    者: 刘天玥
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
static int _wait_SR(struct i2c_dev *i2c_dev, unsigned char *pSR)
{
    unsigned char SR = 0;
    int           i;

    for (i = 0; i < DDC_TIMEOUT_CNT; i++)
    {
        SR = I2C_REG_R8(i2c_dev, OC_SR_ADR);
        if (!(SR & OC_SR_TIP_MASK))
            break;
        usleep_range(50, 100);
    }

    if (pSR)
    {
        *pSR = SR;
    }

    if (i == DDC_TIMEOUT_CNT)
        return DDC_TIMEOUT;

    if (SR & OC_SR_RXNACK_MASK)
    {
        return DDC_NACK;
    }

    return 0;
}
#else
#define CHECK_TRANSFER_COMPLETE(i2c_dev, sr) \
    do { \
        int __timeout_cnt = DDC_TIMEOUT_CNT;\
        sr = I2C_REG_R8(i2c_dev, OC_SR_ADR); \
        while ((sr & OC_SR_TIP_MASK) && __timeout_cnt > 0) { \
            sr = I2C_REG_R8(i2c_dev, OC_SR_ADR); \
            usleep_range(20, 80); \
            __timeout_cnt--;\
        } \
        if (__timeout_cnt <= 0) \
        { \
            ERR_PRINT("oc i2c %d check transfer complete timeout\n", i2c_dev->index); \
            goto OnError; \
        } \
    } \
    while (0)

static int ljm_i2c_oc_check_ack(struct i2c_dev *i2c_dev, unsigned char *pSR)
{
    unsigned char sr;

    CHECK_TRANSFER_COMPLETE(i2c_dev, sr);

    if (pSR)
        *pSR = sr;

    if (sr & OC_SR_RXNACK_MASK)
    {
        I2C_REG_W8(i2c_dev, OC_CR_ADR, OC_CR_STO_MASK);
        return DDC_NACK;
    }
    else
        return 0;
OnError:
    return DDC_TIMEOUT;
}
#endif

/*****************************************************************************
 *  函 数 名: ljm_i2c_oc_detect
 *  功能描述: 检测target设备是否存在
 *  输入参数: struct i2c_dev *i2c_dev  I2C结构体指针
 *  输出参数: 无
 *  返 回 值: bool类型，target设备存在，返回true，否则返回false
 *
 *  修改历史:
 *         1. 日    期: 2022年4月15日
 *            作    者: 刘天玥
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
bool ljm_i2c_oc_detect(struct i2c_dev *i2c_dev)
{
    unsigned char SR = 0;
    int           r;

    HEADER_ARG("i2c_dev(%d)=%lx", i2c_dev->index, (unsigned long)i2c_dev);

    I2C_REG_W8(i2c_dev, OC_TXR_ADR, DDC_EDID_WR_ADDR);              // slave address is EDID
    I2C_REG_W8(i2c_dev, OC_CR_ADR, OC_CR_WR_MASK | OC_CR_STA_MASK); // DDC start with write

#if 0
    r = _wait_SR(i2c_dev, &SR);
#else
    r = ljm_i2c_oc_check_ack(i2c_dev, &SR);
#endif
    TRACE("oc i2c %d gets DDC status %x", i2c_dev->index, SR);
    if (r == 0)
    {                                                   // not timeout and there is slave device
        I2C_REG_W8(i2c_dev, OC_CR_ADR, OC_CR_STO_MASK); // I2C stop

        FOOTER_ARG("%s", "true");
        return true;
    }

    FOOTER_ARG("%s", "false");
    return false;
}

/*****************************************************************************
 *  函 数 名: ljm_i2c_oc_read_req
 *  功能描述: OC I2C从target设备读取一块缓存
 *  输入参数: void *hw_dev          I2C结构体指针
 *            unsigned char target  target设备地址
 *            unsigned char addr    target设备内偏移地址
 *            unsigned char *buf    数据缓冲指针
 *            int size              数据字节数
 *  输出参数: 无
 *  返 回 值: int类型，0表示成功，否则返回错误代码
 *
 *  修改历史:
 *         1. 日    期: 2022年4月15日
 *            作    者: 刘天玥
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
int ljm_i2c_oc_read_req(void *hw_dev, unsigned char target, unsigned char addr, unsigned char *buf, int size)
{
    struct i2c_dev *i2c_dev = (struct i2c_dev *)hw_dev;
    unsigned char   SR;
    int             r = 0;
    int             i = 0;

    HEADER_ARG("i2c_dev=%p target=%x addr=%x buf=%p size=%x", i2c_dev, target, addr, buf, size);

    /* base block fetch */
    I2C_REG_W8(i2c_dev, OC_TXR_ADR, (target << 1) | DDC_WR_BIT);
    I2C_REG_W8(i2c_dev, OC_CR_ADR, OC_CR_WR_MASK | OC_CR_STA_MASK);
#if 0
    r = _wait_SR(i2c_dev, &SR);
#else
    r = ljm_i2c_oc_check_ack(i2c_dev, &SR);
#endif
    if (r)
    { // timeout or no slave device
        TRACE("oc i2c %d gets DDC status %x", i2c_dev->index, SR);

        goto OnError;
    }

    /* set eeprom start address */
    I2C_REG_W8(i2c_dev, OC_TXR_ADR, addr); // EDID start address is 0x00
    I2C_REG_W8(i2c_dev, OC_CR_ADR, OC_CR_WR_MASK);
#if 0
    r = _wait_SR(i2c_dev, &SR);
#else
    /*
    r = ljm_i2c_oc_check_ack(i2c_dev, &SR);

    if (r)
    {
        TRACE("oc i2c %d gets DDC status %x", i2c_dev->index, SR);

        goto OnError;
    }
    */
    CHECK_TRANSFER_COMPLETE(i2c_dev, SR);
#endif

    /* i2c switch to read */
    I2C_REG_W8(i2c_dev, OC_TXR_ADR, (target << 1) | DDC_RD_BIT);
    I2C_REG_W8(i2c_dev, OC_CR_ADR, OC_CR_WR_MASK | OC_CR_STA_MASK); /* FIXME: OC_CR_RD_MASK | OC_CR_STA_MASK ? */
#if 0
    r = _wait_SR(i2c_dev, &SR);
#else
    r = ljm_i2c_oc_check_ack(i2c_dev, &SR);
#endif
    if (r)
    {
        TRACE("oc i2c %d gets DDC status %x", i2c_dev->index, SR);

        goto OnError;
    }

    /* i2c read eeprom data to buf */
    for (i = 0; i < size - 1; i++)
    {
        I2C_REG_W8(i2c_dev, OC_CR_ADR, OC_CR_RD_MASK);
#if 0
        r = _wait_SR(i2c_dev, &SR);
#else
        /*
        r = ljm_i2c_oc_check_ack(i2c_dev, &SR);
        if (r)
        {
            TRACE("oc i2c %d gets DDC status %x", i2c_dev->index, SR);

            goto OnError;
        }
        */
        CHECK_TRANSFER_COMPLETE(i2c_dev, SR);
#endif
        buf[i] = I2C_REG_R8(i2c_dev, OC_RXR_ADR);
    }

    I2C_REG_W8(i2c_dev, OC_CR_ADR, OC_CR_STO_MASK | OC_CR_RD_MASK | OC_CR_ACK_MASK);
#if 0
    r = _wait_SR(i2c_dev, &SR);
    if (r)
    {
        TRACE("oc i2c %d gets DDC status %x", i2c_dev->index, SR);

        goto OnError;
    }
#else
    CHECK_TRANSFER_COMPLETE(i2c_dev, SR);
#endif
    buf[i] = I2C_REG_R8(i2c_dev, OC_RXR_ADR);

    FOOTER();
    return 0;

OnError:
    DEBUG_PRINT("oc i2c %d gets wrong device status\n", i2c_dev->index);

    FOOTER();

    return r;
}

/*****************************************************************************
 *  函 数 名: ljm_i2c_oc_write_req
 *  功能描述: I2C写缓冲到target设备的指定地址内
 *  输入参数: void *hw_dev          I2C结构体指针
 *            unsigned char target  target设备地址
 *            unsigned char addr    target设备内偏移地址
 *            unsigned char *buf    数据缓冲指针
 *            int size              数据字节数
 *  输出参数: 无
 *  返 回 值: int类型，0表示成功，否则返回错误代码
 *
 *  修改历史:
 *         1. 日    期: 2022年4月15日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
int ljm_i2c_oc_write_req(void *hw_dev, unsigned char target, unsigned char addr, unsigned char *buf, int size)
{
    struct i2c_dev *i2c_dev = (struct i2c_dev *)hw_dev;
    unsigned char   SR;
    int             r = DDC_TIMEOUT;
    int             i = 0;

    HEADER_ARG("i2c_dev=%p target=%d addr=%d buf=%p size=%d", i2c_dev, target, addr, buf, size);

    /* base block fetch */
    I2C_REG_W8(i2c_dev, OC_TXR_ADR, (target << 1) | DDC_WR_BIT);
    I2C_REG_W8(i2c_dev, OC_CR_ADR, OC_CR_WR_MASK | OC_CR_STA_MASK);

#if 0
    r = _wait_SR(i2c_dev, &SR);
#else
    r = ljm_i2c_oc_check_ack(i2c_dev, &SR);
#endif
    if (r)
    { // timeout or no slave device
        TRACE("oc i2c %d gets DDC status %x(%d)", i2c_dev->index, SR, r);

        goto OnError;
    }

    /* set eeprom start address */
    I2C_REG_W8(i2c_dev, OC_TXR_ADR, addr); // EDID start address is 0x00
    I2C_REG_W8(i2c_dev, OC_CR_ADR, OC_CR_WR_MASK);
#if 0
    r = _wait_SR(i2c_dev, &SR);
    if (r)
    {
        TRACE("oc i2c %d gets DDC status %x", i2c_dev->index, SR);

        goto OnError;
    }
#else
    CHECK_TRANSFER_COMPLETE(i2c_dev, SR);
#endif


    /* i2c write eeprom */
    for (i = 0; i < size - 1; i++)
    {
        I2C_REG_W8(i2c_dev, OC_TXR_ADR, buf[i]);
        I2C_REG_W8(i2c_dev, OC_CR_ADR, OC_CR_WR_MASK);
#if 0
        r = _wait_SR(i2c_dev, &SR);
        if (r)
        {
            TRACE("oc i2c %d gets DDC status %x", i2c_dev->index, SR);

            goto OnError;
        }
#else
        CHECK_TRANSFER_COMPLETE(i2c_dev, SR);
#endif
    }

    I2C_REG_W8(i2c_dev, OC_TXR_ADR, buf[i]);
    I2C_REG_W8(i2c_dev, OC_CR_ADR, OC_CR_STO_MASK | OC_CR_WR_MASK | OC_CR_ACK_MASK);
#if 0
    r = _wait_SR(i2c_dev, &SR);
    if (r)
    {
        TRACE("oc i2c %d gets DDC status %x", i2c_dev->index, SR);

        goto OnError;
    }
#else
    CHECK_TRANSFER_COMPLETE(i2c_dev, SR);
#endif

    FOOTER();
    return 0;

OnError:
    ERR_PRINT("oc i2c %d gets wrong device status\n", i2c_dev->index);

    FOOTER();
    return r;
}

/*****************************************************************************
 *  函 数 名: ljm_i2c_oc_raw_write_req
 *  功能描述: I2C原始写target设备
 *  输入参数: void *hw_dev          I2C结构体指针
 *            unsigned char target  target设备地址
 *            unsigned char data    写入数据
 *  输出参数: 无
 *  返 回 值: int类型，0表示成功，否则返回错误代码
 *
 *  修改历史:
 *         1. 日    期: 2022年4月15日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
int ljm_i2c_oc_raw_write_req(void *hw_dev, unsigned char target, unsigned char data)
{
    struct i2c_dev *i2c_dev = (struct i2c_dev *)hw_dev;
    unsigned char   SR;
    int             r = DDC_TIMEOUT;

    HEADER_ARG("i2c_dev=%p target=%d data=%d", i2c_dev, target, data);

    /* base block fetch */
    I2C_REG_W8(i2c_dev, OC_TXR_ADR, (target << 1) | DDC_WR_BIT);
    I2C_REG_W8(i2c_dev, OC_CR_ADR, OC_CR_WR_MASK | OC_CR_STA_MASK);

#if 0
    r = _wait_SR(i2c_dev, &SR);
#else
    r = ljm_i2c_oc_check_ack(i2c_dev, &SR);
#endif
    if (r)
    { // timeout or no slave device
        TRACE("oc i2c %d gets DDC status %x(%d)", i2c_dev->index, SR, r);
        goto OnError;
    }

    /* set eeprom start address */
    I2C_REG_W8(i2c_dev, OC_TXR_ADR, data); // EDID start address is 0x00
    I2C_REG_W8(i2c_dev, OC_CR_ADR, OC_CR_STO_MASK | OC_CR_WR_MASK | OC_CR_ACK_MASK);
#if 0
    r = _wait_SR(i2c_dev, &SR);
    if (r)
    {
        TRACE("oc i2c %d gets DDC status %x(%d)", i2c_dev->index, SR, r);
        goto OnError;
    }
#else
    CHECK_TRANSFER_COMPLETE(i2c_dev, SR);
#endif

    FOOTER();
    return 0;

OnError:
    ERR_PRINT("oc i2c %d gets wrong device status\n", i2c_dev->index);

    FOOTER();
    return r;
}

#if 0
/**
 * ljm_i2c_oc_get_edid - get EDID data
 * @ i2c_dev: pointer to i2c device
 *
 * Return: Pointer to valid EDID or NULL if we couldn't find any.
 */
struct edid *ljm_i2c_oc_get_edid(struct i2c_dev *i2c_dev, int *psize)
{
    unsigned char SR;
    int r;
    int i = 0;
    u8 *block, *new;
    int size;

    HEADER_ARG("i2c_dev=%lx psize=%lx", (unsigned long)i2c_dev, (unsigned long)psize);

    size = DDC_EDID_SIZE;
    if ((block = kzalloc(DDC_EDID_SIZE, GFP_KERNEL)) == NULL) {
        if (psize) *psize = 0;
        FOOTER_ARG("ERR(%d): Out of memory", -ENOMEM);
        return NULL;
    }

    /* base block fetch */
    I2C_REG_W8(i2c_dev, OC_TXR_ADR, DDC_EDID_WR_ADDR);
    I2C_REG_W8(i2c_dev, OC_CR_ADR,  OC_CR_WR_MASK | OC_CR_STA_MASK);

    r = _wait_SR(i2c_dev, &SR);
    if ((r == DDC_TIMEOUT) || (SR & OC_SR_RXACK_MASK)) {    // timeout or no slave device
        TRACE("oc i2c %d gets DDC status %x(%d)", i2c_dev->index, SR, r);
        ERR_PRINT("oc i2c %d gets wrong device status\n", i2c_dev->index);
        goto OnError;
    }

    /* set EDID start address */
    I2C_REG_W8(i2c_dev, OC_TXR_ADR, 0x00);   // EDID start address is 0x00
    I2C_REG_W8(i2c_dev, OC_CR_ADR,  OC_CR_WR_MASK);
    if(_wait_SR(i2c_dev, &SR) == DDC_TIMEOUT) {
        TRACE("oc i2c %d gets DDC timeout", i2c_dev->index);
        goto OnError;
    }

    /* DDC switch to read */
    I2C_REG_W8(i2c_dev, OC_TXR_ADR, DDC_EDID_RD_ADDR);
    I2C_REG_W8(i2c_dev, OC_CR_ADR, OC_CR_WR_MASK | OC_CR_STA_MASK);  /* FIXME: OC_CR_RD_MASK | OC_CR_STA_MASK ? */
    if (_wait_SR(i2c_dev, &SR) == DDC_TIMEOUT) {
        TRACE("oc i2c %d gets DDC status timeout", i2c_dev->index);
        goto OnError;
    }

    /* DDC read EDID data to block */
    for (i=0; i<DDC_EDID_SIZE-1; i++) {
        I2C_REG_W8(i2c_dev, OC_CR_ADR, OC_CR_RD_MASK);
        if (_wait_SR(i2c_dev, &SR) == DDC_TIMEOUT) {
            TRACE("oc i2c %d gets DDC status timeout", i2c_dev->index);
            goto OnError;
        }
        block[i] = I2C_REG_R8(i2c_dev, OC_RXR_ADR);
    }

    /* if there's no extensions, we have done */
    if (block[DDC_EDID_SIZE-2] == 0) {
        I2C_REG_W8(i2c_dev, OC_CR_ADR, OC_CR_STO_MASK | OC_CR_RD_MASK | OC_CR_ACK_MASK);
        if(_wait_SR(i2c_dev, &SR) == DDC_TIMEOUT)
            goto OnError;

        block[i] = I2C_REG_R8(i2c_dev, OC_RXR_ADR);

        if (psize) *psize = size;
        FOOTER_ARG("EDID=0x%lx", (unsigned long)block);
        return (struct edid *)block;
    }

    size = (block[DDC_EDID_SIZE-2] + 1) * DDC_EDID_SIZE;
#if 0
    new = krealloc(block, size, GFP_KERNEL);
    if (!new) {
        goto OnError;
    }
#else
    new = kzalloc(size, GFP_KERNEL);
    if (!new)
    {
        goto OnError;
    }
    memcpy(new, block, DDC_EDID_SIZE);
    kfree(block);
#endif
    block = new;

    for (; i<(block[DDC_EDID_SIZE-2] + 1) * DDC_EDID_SIZE - 1; i++) {
        I2C_REG_W8(i2c_dev, OC_CR_ADR, OC_CR_RD_MASK);
        if(_wait_SR(i2c_dev, &SR) == DDC_TIMEOUT) {
            TRACE("oc i2c %d gets DDC status timeout", i2c_dev->index);
            goto OnError;
        }
        block[i] = I2C_REG_R8(i2c_dev, OC_RXR_ADR);
    }

    I2C_REG_W8(i2c_dev, OC_CR_ADR, OC_CR_STO_MASK | OC_CR_RD_MASK | OC_CR_ACK_MASK);
    if(_wait_SR(i2c_dev, &SR) == DDC_TIMEOUT) {
        TRACE("oc i2c %d gets DDC status timeout", i2c_dev->index);
        goto OnError;
    }
    block[i] = I2C_REG_R8(i2c_dev, OC_RXR_ADR);

    if (psize)
        *psize = size;
    FOOTER_ARG("new EDID=0x%lx", (unsigned long)block);
    return (struct edid *)block;

OnError:
    kfree(block);
    if (psize) *psize = 0;
    FOOTER();
    return NULL;
}
#endif

/* I2C Control register offsets */
#define IC_CON            0x0000
#define IC_TAR            0x0004
#define IC_DATA_CMD       0x0010
#define IC_INTR_STAT      0x002c
#define IC_INTR_MASK      0x0030
#define IC_RAW_INTR_STAT  0x0034
#define IC_RX_TL          0x38
#define IC_TX_TL          0x3c
#define IC_CLR_INTR       0x0040
#define IC_CLR_TX_ABRT    0x0054
#define IC_CLR_STOP_DET   0x0060
#define IC_ENABLE         0x006c
#define IC_TXFLR          0x0074
#define IC_RXFLR          0x0078
#define IC_TX_ABRT_SOURCE 0x0080
#define IC_ENABLE_STATUS  0x009c
#define IC_COMP_PARAM_1   0x00f4

/* I2C Control register entry bit positions and sizes */
#define IC_COMP_PARAM_1_MAX_SPEED_MODE_INDEX  2
#define IC_COMP_PARAM_1_MAX_SPEED_MODE_WIDTH  2
#define IC_COMP_PARAM_1_RX_BUFFER_DEPTH_INDEX 8
#define IC_COMP_PARAM_1_RX_BUFFER_DEPTH_WIDTH 8
#define IC_COMP_PARAM_1_TX_BUFFER_DEPTH_INDEX 16
#define IC_COMP_PARAM_1_TX_BUFFER_DEPTH_WIDTH 8
#define IC_CON_MASTER_MODE_INDEX              0
#define IC_CON_MASTER_MODE_WIDTH              1
#define IC_CON_RESTART_EN_INDEX               5
#define IC_CON_RESTART_EN_WIDTH               1
#define IC_CON_RX_FIFO_FULL_HOLD_INDEX        9
#define IC_CON_RX_FIFO_FULL_HOLD_WIDTH        1
#define IC_CON_SLAVE_DISABLE_INDEX            6
#define IC_CON_SLAVE_DISABLE_WIDTH            1
#define IC_CON_10BITADDR_MASTER_INDEX         4
#define IC_CON_10BITADDR_MASTER_WIDTH         1
#define IC_CON_SPEED_INDEX                    1
#define IC_CON_SPEED_WIDTH                    2
#define IC_DATA_CMD_CMD_INDEX                 8
#define IC_DATA_CMD_CMD_WIDTH                 1
#define IC_DATA_CMD_STOP_INDEX                9
#define IC_DATA_CMD_STOP_WIDTH                1
#define IC_ENABLE_ABORT_INDEX                 1
#define IC_ENABLE_ABORT_WIDTH                 1
#define IC_ENABLE_EN_INDEX                    0
#define IC_ENABLE_EN_WIDTH                    1
#define IC_ENABLE_STATUS_EN_INDEX             0
#define IC_ENABLE_STATUS_EN_WIDTH             1
#define IC_INTR_MASK_TX_EMPTY_INDEX           4
#define IC_INTR_MASK_TX_EMPTY_WIDTH           1
#define IC_RAW_INTR_STAT_RX_FULL_INDEX        2
#define IC_RAW_INTR_STAT_RX_FULL_WIDTH        1
#define IC_RAW_INTR_STAT_STOP_DET_INDEX       9
#define IC_RAW_INTR_STAT_STOP_DET_WIDTH       1
#define IC_RAW_INTR_STAT_TX_ABRT_INDEX        6
#define IC_RAW_INTR_STAT_TX_ABRT_WIDTH        1
#define IC_RAW_INTR_STAT_TX_EMPTY_INDEX       4
#define IC_RAW_INTR_STAT_TX_EMPTY_WIDTH       1
#define IC_RAW_INTR_STAT_TX_OVER_INDEX        3

/* I2C Control register value */
#define IC_TX_ABRT_7B_ADDR_NOACK 0x0001
#define IC_TX_ABRT_ARB_LOST      0x1000

/* Bit setting and getting macros
 *  The get macro will extract the current bit field value from within
 *  the variable
 *
 *  The set macro will clear the current bit field value within the
 *  variable and then set the bit field of the variable to the
 *  specified value
 */
#define GET_BITS(_var, _index, _width) (((_var) >> (_index)) & ((0x1 << (_width)) - 1))

#define SET_BITS(_var, _index, _width, _val) \
    do \
    { \
        (_var) &= ~(((0x1 << (_width)) - 1) << (_index)); \
        (_var) |= (((_val) & ((0x1 << (_width)) - 1)) << (_index)); \
    } \
    while (0)

#if 0
#define GET_BITS_LE(_var, _index, _width) ((le32_to_cpu((_var)) >> (_index)) & ((0x1 << (_width)) - 1))

#define SET_BITS_LE(_var, _index, _width, _val) \
    do \
    { \
        (_var) &= cpu_to_le32(~(((0x1 << (_width)) - 1) << (_index))); \
        (_var) |= cpu_to_le32((((_val) & ((0x1 << (_width)) - 1)) << (_index))); \
    } \
    while (0)
#endif

/* Macros for building, reading or writing register values or bits
 * within the register values of I2C Control registers.
 */
#define XI2C_GET_BITS(_var, _prefix, _field) GET_BITS((_var), _prefix##_##_field##_INDEX, _prefix##_##_field##_WIDTH)

#define XI2C_SET_BITS(_var, _prefix, _field, _val) \
    SET_BITS((_var), _prefix##_##_field##_INDEX, _prefix##_##_field##_WIDTH, (_val))

#ifdef UEFI_GOP
#define XI2C_IOREAD(i2c_dev, _reg)        RREG32_BASE((i2c_dev)->Private, (i2c_dev)->offset + (_reg))
#define XI2C_IOWRITE(i2c_dev, _reg, _val) WREG32_BASE((i2c_dev)->Private, (i2c_dev)->offset + (_reg), _val)
#else
#define XI2C_IOREAD(i2c_dev, _reg)        RREG32_BASE((i2c_dev)->regbase, (i2c_dev)->offset + (_reg))
#define XI2C_IOWRITE(i2c_dev, _reg, _val) WREG32_BASE((i2c_dev)->regbase, (i2c_dev)->offset + (_reg), _val)
#endif

#define XI2C_IOREAD_BITS(i2c_dev, _reg, _field) \
    GET_BITS(XI2C_IOREAD((i2c_dev), (_reg)), _reg##_##_field##_INDEX, _reg##_##_field##_WIDTH)

#define XI2C_IOWRITE_BITS(i2c_dev, _reg, _field, _val) \
    do \
    { \
        u32 reg_val = XI2C_IOREAD((i2c_dev), (_reg)); \
        SET_BITS(reg_val, _reg##_##_field##_INDEX, _reg##_##_field##_WIDTH, (_val)); \
        XI2C_IOWRITE((i2c_dev), (_reg), reg_val); \
    } \
    while (0)

#define XGBE_ABORT_COUNT   500
#define XGBE_DISABLE_COUNT 1000

#define XGBE_STD_SPEED 1

#define XGBE_INTR_RX_FULL     BIT(IC_RAW_INTR_STAT_RX_FULL_INDEX)
#define XGBE_INTR_TX_EMPTY    BIT(IC_RAW_INTR_STAT_TX_EMPTY_INDEX)
#define XGBE_INTR_TX_ABRT     BIT(IC_RAW_INTR_STAT_TX_ABRT_INDEX)
#define XGBE_INTR_STOP_DET    BIT(IC_RAW_INTR_STAT_STOP_DET_INDEX)
#define XGBE_INTR_TX_OVER     BIT(IC_RAW_INTR_STAT_TX_OVER_INDEX)
#define XGBE_DEFAULT_INT_MASK (XGBE_INTR_RX_FULL | XGBE_INTR_TX_EMPTY | XGBE_INTR_TX_ABRT | XGBE_INTR_STOP_DET)

#define XGBE_I2C_RESTART BIT(10)
#define XGBE_I2C_READ    BIT(8)
#define XGBE_I2C_WRITE  ~(BIT(8))
#define XGBE_I2C_STOP    BIT(9)

/*****************************************************************************
 *  函 数 名: dw_i2c_abort
 *  功能描述: DW I2C传输取消
 *  输入参数: struct i2c_dev *i2c_dev  I2C结构体指针
 *  输出参数: 无
 *  返 回 值: int类型，0表示成功，否则返回错误代码
 *
 *  修改历史:
 *         1. 日    期: 2022年4月15日
 *            作    者: 刘天玥
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
static int dw_i2c_abort(struct i2c_dev *i2c_dev)
{
    unsigned int wait = XGBE_ABORT_COUNT;

    /* Must be enabled to recognize the abort request */
    XI2C_IOWRITE_BITS(i2c_dev, IC_ENABLE, EN, 1);

    /* Issue the abort */
    XI2C_IOWRITE_BITS(i2c_dev, IC_ENABLE, ABORT, 1);

    while (wait--)
    {
        if (!XI2C_IOREAD_BITS(i2c_dev, IC_ENABLE, ABORT))
            return 0;

        usleep_range(500, 600);
    }

    return -EBUSY;
}

/*****************************************************************************
 *  函 数 名: dw_i2c_set_enable
 *  功能描述: 使能DW I2C控制器
 *  输入参数: struct i2c_dev *i2c_dev  I2C结构体指针
 *            bool enable              使能控制参数，true表示使能，false表示禁用
 *  输出参数: 无
 *  返 回 值: int类型，0表示成功，否则返回错误代码
 *
 *  修改历史:
 *         1. 日    期: 2022年4月15日
 *            作    者: 刘天玥
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
static int dw_i2c_set_enable(struct i2c_dev *i2c_dev, bool enable)
{
    unsigned int wait = XGBE_DISABLE_COUNT;
    unsigned int mode = enable ? 1 : 0;

    while (wait--)
    {
        XI2C_IOWRITE_BITS(i2c_dev, IC_ENABLE, EN, mode);
        if (XI2C_IOREAD_BITS(i2c_dev, IC_ENABLE_STATUS, EN) == mode)
            return 0;

#ifdef CONFIG_PREEMPT
        udelay(105);
#else
        usleep_range(100, 110);
#endif
    }

    return -EBUSY;
}

/*****************************************************************************
 *  函 数 名: dw_i2c_disable
 *  功能描述: 禁用DW I2C控制器
 *  输入参数: struct i2c_dev *i2c_dev  I2C结构体指针
 *  输出参数: 无
 *  返 回 值: int类型，0表示成功，否则返回错误代码
 *
 *  修改历史:
 *         1. 日    期: 2022年4月15日
 *            作    者: 刘天玥
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
static int dw_i2c_disable(struct i2c_dev *i2c_dev)
{
    unsigned int ret;

    ret = dw_i2c_set_enable(i2c_dev, false);
    if (ret)
    {
        /* Disable failed, try an abort */
        ret = dw_i2c_abort(i2c_dev);
        if (ret)
            return ret;

        /* Abort succeeded, try to disable again */
        ret = dw_i2c_set_enable(i2c_dev, false);
    }

    return ret;
}

/*****************************************************************************
 *  函 数 名: dw_i2c_enable
 *  功能描述: 使能DW I2C控制器
 *  输入参数: struct i2c_dev *i2c_dev  I2C结构体指针
 *  输出参数: 无
 *  返 回 值: int类型，0表示成功，否则返回错误代码
 *
 *  修改历史:
 *         1. 日    期: 2022年4月15日
 *            作    者: 刘天玥
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
static int dw_i2c_enable(struct i2c_dev *i2c_dev)
{
    return dw_i2c_set_enable(i2c_dev, true);
}

/*****************************************************************************
 *  函 数 名: dw_i2c_get_features
 *  功能描述: 获取DW I2C控制器特性
 *  输入参数: struct i2c_dev *i2c_dev  I2C结构体指针
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2022年4月15日
 *            作    者: 刘天玥
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
static void dw_i2c_get_features(struct i2c_dev *i2c_dev)
{
    unsigned int reg;

    reg                     = XI2C_IOREAD(i2c_dev, IC_COMP_PARAM_1);
    i2c_dev->max_speed_mode = XI2C_GET_BITS(reg, IC_COMP_PARAM_1, MAX_SPEED_MODE);
    i2c_dev->rx_fifo_size   = XI2C_GET_BITS(reg, IC_COMP_PARAM_1, RX_BUFFER_DEPTH);
    i2c_dev->tx_fifo_size   = XI2C_GET_BITS(reg, IC_COMP_PARAM_1, TX_BUFFER_DEPTH);

    DEBUG_PRINT("I2C features: %s=%u, %s=%u, %s=%u\n", "MAX_SPEED_MODE", i2c_dev->max_speed_mode, "RX_BUFFER_DEPTH",
                i2c_dev->rx_fifo_size, "TX_BUFFER_DEPTH", i2c_dev->tx_fifo_size);
}

/*****************************************************************************
 *  函 数 名: dw_i2c_set_target
 *  功能描述: 设置target设备地址
 *  输入参数: struct i2c_dev *i2c_dev  I2C结构体指针
 *            unsigned int target      target设备地址
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2022年4月15日
 *            作    者: 刘天玥
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
static void dw_i2c_set_target(struct i2c_dev *i2c_dev, unsigned int target)
{
    XI2C_IOWRITE(i2c_dev, IC_TAR, target);
}

/*****************************************************************************
 *  函 数 名: dw_i2c_set_mode
 *  功能描述: 设置DW I2C工作模式
 *  输入参数: struct i2c_dev *i2c_dev  I2C结构体指针
 *  输出参数: 无
 *  返 回 值: 无
 *
 *  修改历史:
 *         1. 日    期: 2022年4月15日
 *            作    者: 刘天玥
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
static void dw_i2c_set_mode(struct i2c_dev *i2c_dev)
{
    unsigned int reg;

    reg = XI2C_IOREAD(i2c_dev, IC_CON);
    //DEBUG_PRINT("reg=0x%x\n", reg);
#if 1
    XI2C_SET_BITS(reg, IC_CON, MASTER_MODE, 1);
    XI2C_SET_BITS(reg, IC_CON, SLAVE_DISABLE, 1);
    XI2C_SET_BITS(reg, IC_CON, RESTART_EN, 1);
    XI2C_SET_BITS(reg, IC_CON, SPEED, XGBE_STD_SPEED);
    XI2C_SET_BITS(reg, IC_CON, RX_FIFO_FULL_HOLD, 0);
    XI2C_SET_BITS(reg, IC_CON, 10BITADDR_MASTER, 0);
    XI2C_IOWRITE(i2c_dev, IC_CON, reg);
#else
    XI2C_IOWRITE(i2c_dev, IC_CON, 0x00000063);
#endif
}

/*****************************************************************************
 *  函 数 名: ljm_i2c_dw_init
 *  功能描述: 初始化DW I2C控制器和结构体
 *  输入参数: struct i2c_dev *i2c_dev  I2C结构体指针
 *            void *base               寄存器空间虚拟基地址/UEFI GOP Private指针
 *            int offset               I2C控制器寄存器地址起始偏移
 *            int index                I2C控制器序号
 *  输出参数: 无
 *  返 回 值: int类型，0表示成功，否则返回错误代码
 *
 *  修改历史:
 *         1. 日    期: 2022年4月15日
 *            作    者: 刘天玥
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
int ljm_i2c_dw_init(struct i2c_dev *i2c_dev, void *base, int offset, int index)
{
    int ret;

    HEADER_ARG("i2c_dev=%px base=0x%px offset=%x index=%d", i2c_dev, base, offset, index);

#ifdef UEFI_GOP
    i2c_dev->Private = base;
#else
    i2c_dev->regbase = base;
#endif

    i2c_dev->offset = offset;
    i2c_dev->index  = index;

    /* disable interrupt */
    XI2C_IOWRITE(i2c_dev, IC_INTR_MASK, 0);

    ret = dw_i2c_disable(i2c_dev);
    if (ret)
    {
        DEBUG_PRINT("failed to disable i2c master\n");
        return ret;
    }

    dw_i2c_get_features(i2c_dev);

    dw_i2c_set_mode(i2c_dev);

    /* clear all interrupt */
    XI2C_IOREAD(i2c_dev, IC_CLR_INTR);
#if 0
    /* initialize i2c controller, such as set SCL clock, etc. */
    // set IC_ENABLE (disable i2c)
    ljm_i2c_dw_set_enable(dw_i2c_dev, false);

    // spike suppression
    I2C_REG_W32(dw_i2c_dev, IC_FS_SPKLEN, 0x00000005);
    // set IC_SDA_HOLD
    I2C_REG_W32(dw_i2c_dev, IC_SDA_HOLD, 0x0000001E);
    // set IC_SS_SCL_LCNT and IC_SS_SCL_HCNT
    // set scl clock low period count
    I2C_REG_W32(dw_i2c_dev, IC_SS_SCL_HCNT, 0x00000257);
    // set scl clock high period count
    I2C_REG_W32(dw_i2c_dev, IC_SS_SCL_LCNT, 0x00000184);
    // set IC_CON
    // master mode is enabled; slave 7bit addessing mode;
    //standard speed mode of operation
    // master restart enabled; master 7bit addressing mode;
    //slave mode is disabled
    I2C_REG_W32(dw_i2c_dev, IC_CON, 0x00000063);

    // set IC_INTR_MASK (this is the I2C interruot mask register)
    I2C_REG_W32(dw_i2c_dev, IC_INTR_MASK, 0x00000000);//I2C_REG_W32(dw_i2c_dev, IC_INTR_MASK, 0x00000204);
    // set IC_RX_TL (this is the Rx FIFO threshold level)
    I2C_REG_W32(dw_i2c_dev, IC_RX_TL, 255);// set Rx FIFO threshold level
    // set IC_TX_TL (this is the Rx FIFO threshold level)
    I2C_REG_W32(dw_i2c_dev, IC_TX_TL, 255);// set Rx FIFO threshold level

    // set IC_TAR (this is the target address for any mater transaction)
    //I2C_REG_W32(dw_i2c_dev, IC_TAR, (target | 0x00000400);

    // set IC_ENABLE (enable i2c)
    ljm_i2c_dw_set_enable(dw_i2c_dev, true);
#endif

    FOOTER();
    return 0;
}

/*****************************************************************************
 *  函 数 名: ljm_i2c_dw_detect
 *  功能描述: 检测target设备是否存在
 *  输入参数: struct i2c_dev *i2c_dev  I2C结构体指针
 *  输出参数: 无
 *  返 回 值: bool类型，target设备存在，返回true，否则返回false
 *
 *  修改历史:
 *         1. 日    期: 2022年4月15日
 *            作    者: 刘天玥
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
bool ljm_i2c_dw_detect(struct i2c_dev *i2c_dev, unsigned char target)
{
    int          ret;
    unsigned int isr;
    unsigned int timeout;

    ret = dw_i2c_disable(i2c_dev);
    if (ret)
    {
        ERR_PRINT("failed to disable i2c master\n");
        return false;
    }

    dw_i2c_set_target(i2c_dev, target);

    /* clear all interrupt */
    XI2C_IOREAD(i2c_dev, IC_CLR_INTR);

    ret = dw_i2c_enable(i2c_dev);
    if (ret)
    {
        ERR_PRINT("failed to enable i2c master\n");
        return false;
    }

    /* write 0 address */
    XI2C_IOWRITE(i2c_dev, IC_DATA_CMD, 0x200);

    /* 等待数据接受完成 */
    timeout = DDC_TIMEOUT_CNT;
    while (--timeout)
    {
        isr = XI2C_IOREAD(i2c_dev, IC_RAW_INTR_STAT);
        if (isr & XGBE_INTR_TX_ABRT)
        {
            DEBUG_PRINT("i2c %d: TX_ABRT received (%#010x) for target %#04x\n", i2c_dev->index,
                      XI2C_IOREAD(i2c_dev, IC_TX_ABRT_SOURCE), target);
            return false;
        }

        if (isr & XGBE_INTR_STOP_DET)
        {
            return true;
        }
        usleep_range(20, 80);
    }

    if (timeout == 0)
    {
        dw_i2c_disable(i2c_dev);
        return false;
    }

    return true;
}

/*****************************************************************************
 *  函 数 名: ljm_i2c_dw_read_req
 *  功能描述: OC I2C从target设备读取一块缓存
 *  输入参数: void *hw_dev          I2C结构体指针
 *            unsigned char target  target设备地址
 *            unsigned char addr    target设备内偏移地址
 *            unsigned char *buf    数据缓冲指针
 *            int size              数据字节数
 *  输出参数: 无
 *  返 回 值: int类型，0表示成功，否则返回错误代码
 *
 *  修改历史:
 *         1. 日    期: 2022年4月15日
 *            作    者: 刘天玥
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
int ljm_i2c_dw_read_req(void *hw_dev, unsigned char target, unsigned char addr, unsigned char *buf, int size)
{
    struct i2c_dev *i2c_dev = (struct i2c_dev *)hw_dev;
    int             ret;
    unsigned int    isr;
    unsigned int    cmd;
    unsigned int    i, cur_size;
    unsigned int    timeout;

    ret = dw_i2c_disable(i2c_dev);
    if (ret)
    {
        ERR_PRINT("failed to disable i2c master\n");
        return ret;
    }

    dw_i2c_set_target(i2c_dev, target);

    /* clear all interrupt */
    XI2C_IOREAD(i2c_dev, IC_CLR_INTR);

    ret = dw_i2c_enable(i2c_dev);
    if (ret)
    {
        ERR_PRINT("failed to enable i2c master\n");
        return ret;
    }

    isr = XI2C_IOREAD(i2c_dev, IC_RAW_INTR_STAT);
    // PRINT("----before int enable: IC_RAW_INTR_STAT=%x\n", isr);

    /* 设置I2C从设备数据起始地址 */
    cmd = addr;
    XI2C_IOWRITE(i2c_dev, IC_DATA_CMD, cmd);

    /* enable interrupts */
    XI2C_IOWRITE(i2c_dev, IC_INTR_MASK, XGBE_DEFAULT_INT_MASK);

    cmd = XGBE_I2C_RESTART;
    while (size > 0)
    {
        cur_size = ((i2c_dev->rx_fifo_size + 1) > size) ? size : (i2c_dev->rx_fifo_size + 1);

        /* 设置RX的门槛 */
        XI2C_IOWRITE(i2c_dev, IC_RX_TL, cur_size - 1);

        for (i = 0; i < cur_size; i++)
        {
            cmd |= XGBE_I2C_READ;
            if (size == cur_size && i == (cur_size - 1))
                XI2C_SET_BITS(cmd, IC_DATA_CMD, STOP, 1);
            XI2C_IOWRITE(i2c_dev, IC_DATA_CMD, cmd);
            cmd = 0;
        }

        /* 等待数据接受完成 */
        timeout = DDC_TIMEOUT_CNT;
        while (--timeout)
        {
            isr = XI2C_IOREAD(i2c_dev, IC_RAW_INTR_STAT);
            if (isr & XGBE_INTR_TX_ABRT)
            {
                DEBUG_PRINT("i2c %d: TX_ABRT received (%#010x) for target %#04x\n", i2c_dev->index,
                            XI2C_IOREAD(i2c_dev, IC_TX_ABRT_SOURCE), target);
                return -EIO;
            }

            if (isr & XGBE_INTR_RX_FULL)
            {
                /* 读取数据 */
                for (i = 0; i < cur_size; i++)
                {
                    *buf++ = (unsigned char)XI2C_IOREAD(i2c_dev, IC_DATA_CMD);
                }
                break;
            }

#ifdef CONFIG_PREEMPT
            udelay(40);
#else
            usleep_range(20, 80);
#endif
        }

        if (timeout == 0)
        {
            ERR_PRINT("i2c %d: read timeout\n", i2c_dev->index);
            return -2;
        }

        size -= cur_size;
    }

    return 0;
}

/*****************************************************************************
 *  函 数 名: ljm_i2c_dw_write_req
 *  功能描述: I2C写缓冲到target设备的指定地址内
 *  输入参数: void *hw_dev          I2C结构体指针
 *            unsigned char target  target设备地址
 *            unsigned char addr    target设备内偏移地址
 *            unsigned char *buf    数据缓冲指针
 *            int size              数据字节数
 *  输出参数: 无
 *  返 回 值: int类型，0表示成功，否则返回错误代码
 *
 *  修改历史:
 *         1. 日    期: 2022年4月15日
 *            作    者: 刘天玥
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
int ljm_i2c_dw_write_req(void *hw_dev, unsigned char target, unsigned char addr, unsigned char *buf, int size)
{
    struct i2c_dev *i2c_dev = (struct i2c_dev *)hw_dev;
    int             ret;
    unsigned int    isr;
    unsigned int    cmd;
    unsigned int    i, cur_size;
    unsigned int    timeout;

    ret = dw_i2c_disable(i2c_dev);
    if (ret)
    {
        ERR_PRINT("failed to disable i2c master\n");
        return ret;
    }

    dw_i2c_set_target(i2c_dev, target);

    /* clear all interrupt */
    XI2C_IOREAD(i2c_dev, IC_CLR_INTR);

    ret = dw_i2c_enable(i2c_dev);
    if (ret)
    {
        ERR_PRINT("failed to enable i2c master\n");
        return ret;
    }

    isr = XI2C_IOREAD(i2c_dev, IC_RAW_INTR_STAT);
    // PRINT("----before int enable: IC_RAW_INTR_STAT=%x\n", isr);

    /* 设置I2C从设备数据起始地址 */
    cmd = addr;
    XI2C_IOWRITE(i2c_dev, IC_DATA_CMD, cmd);

    /* enable interrupts */
    XI2C_IOWRITE(i2c_dev, IC_INTR_MASK, XGBE_DEFAULT_INT_MASK);

    //cmd = XGBE_I2C_RESTART;
    while (size > 0)
    {
        cur_size = ((i2c_dev->tx_fifo_size + 1) > size) ? size : (i2c_dev->tx_fifo_size + 1);

        /* 设置TX的门槛 */
        XI2C_IOWRITE(i2c_dev, IC_TX_TL, cur_size - 1);

        for (i = 0; i < cur_size; i++)
        {
            cmd =*buf++;
            cmd &= XGBE_I2C_WRITE;
            if (size == cur_size && i == (cur_size - 1))
                XI2C_SET_BITS(cmd, IC_DATA_CMD, STOP, 1);
            XI2C_IOWRITE(i2c_dev, IC_DATA_CMD, cmd);
            cmd = 0;
        }

        /* 等待数据接受完成 */
        timeout = DDC_TIMEOUT_CNT;
        while (--timeout)
        {
            isr = XI2C_IOREAD(i2c_dev, IC_RAW_INTR_STAT);
            if (isr & XGBE_INTR_TX_ABRT)
            {
                DEBUG_PRINT("i2c %d: TX_ABRT received (%#010x) for target %#04x\n", i2c_dev->index,
                            XI2C_IOREAD(i2c_dev, IC_TX_ABRT_SOURCE), target);
                return -EIO;
            }

            if (isr & XGBE_INTR_TX_EMPTY)
                break;

#ifdef CONFIG_PREEMPT
            udelay(40);
#else
            usleep_range(20, 80);
#endif
        }

        if (timeout == 0)
        {
            ERR_PRINT("i2c %d: read timeout\n", i2c_dev->index);
            return -2;
        }

        size -= cur_size;
    }

    return 0;
}

/*****************************************************************************
 *  函 数 名: ljm_i2c_dw_raw_write_req
 *  功能描述: I2C原始写target设备
 *  输入参数: void *hw_dev          I2C结构体指针
 *            unsigned char target  target设备地址
 *            unsigned char data    写入数据
 *  输出参数: 无
 *  返 回 值: int类型，0表示成功，否则返回错误代码
 *
 *  修改历史:
 *         1. 日    期: 2022年4月15日
 *            作    者: 胡艳明
 *            修改内容: 新生成函数
 *
 *****************************************************************************/
int ljm_i2c_dw_raw_write_req(void *hw_dev, unsigned char target, unsigned char data)
{

    struct i2c_dev *i2c_dev = (struct i2c_dev *)hw_dev;
    int             ret;
    unsigned int    isr;
    unsigned int    cmd;
    unsigned int    timeout;

    ret = dw_i2c_disable(i2c_dev);
    if (ret)
    {
        ERR_PRINT("failed to disable i2c master\n");
        return ret;
    }

    dw_i2c_set_target(i2c_dev, target);

    /* clear all interrupt */
    XI2C_IOREAD(i2c_dev, IC_CLR_INTR);

    ret = dw_i2c_enable(i2c_dev);
    if (ret)
    {
        ERR_PRINT("failed to enable i2c master\n");
        return ret;
    }

    isr = XI2C_IOREAD(i2c_dev, IC_RAW_INTR_STAT);
    // PRINT("----before int enable: IC_RAW_INTR_STAT=%x\n", isr);

    /* enable interrupts */
    XI2C_IOWRITE(i2c_dev, IC_INTR_MASK, XGBE_DEFAULT_INT_MASK);

    /*transfer data*/
    cmd = data;
    cmd &= XGBE_I2C_WRITE;
    XI2C_IOWRITE(i2c_dev, IC_DATA_CMD, cmd);


        /* 等待数据接受完成 */
        timeout = DDC_TIMEOUT_CNT;
        while (--timeout)
        {
            isr = XI2C_IOREAD(i2c_dev, IC_RAW_INTR_STAT);

            if (isr & XGBE_INTR_TX_ABRT)
            {
                DEBUG_PRINT("i2c %d: TX_ABRT received (%#010x) for target %#04x\n", i2c_dev->index,
                            XI2C_IOREAD(i2c_dev, IC_TX_ABRT_SOURCE), target);
                return -EIO;
            }

           if (isr & XGBE_INTR_TX_EMPTY)
                break;
            usleep_range(20, 80);
        }

        if (timeout == 0)
        {
            ERR_PRINT("i2c %d: read timeout\n", i2c_dev->index);
            return -2;
        }

    return 0;
}
