/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#include <linux/pm_runtime.h>
#include "ljmicro_device.h"

#define _LJ_OBJ_ZONE    ljvZONE_KMS

/**
 * ljm_msi_ok - asic specific msi checks
 *
 * @ljm_dev: ljm device pointer
 *
 * Handles asic specific MSI checks to determine if
 * MSIs should be enabled on a particular chip (all asics).
 * Returns true if MSIs should be enabled, false if MSIs
 * should not be enabled.
 */
static bool ljm_msi_ok(struct ljm_device *ljm_dev)
{
    if (ljm_dev->flags & LJMICRO_IS_MSI)
        return true;
    else
        return false;
}

/*
 * Handle hotplug events outside the interrupt handler proper.
 */
/**
 * ljm_hotplug_work_func - display hotplug work handler
 *
 * @work: work struct
 *
 * This is the hot plug event work handler (all asics).
 * The work gets scheduled from the irq handler if there
 * was a hot plug interrupt.  It walks the connector table
 * and calls the hotplug handler for each one, then sends
 * a drm hotplug event to alert userspace.
 */
static void ljm_hotplug_work_func(struct work_struct *work)
{
    struct ljm_device *ljm_dev = container_of(work, struct ljm_device,
                                              hotplug_work.work);
    struct drm_device *ddev = ljm_dev->ddev;
    struct drm_mode_config *mode_config = &ddev->mode_config;
    struct drm_connector *connector;
    u32 misc_intr_mask;
    uint8_t buf;

    misc_intr_mask = 0;
    buf = 0;
    /* we can race here at startup, some boards seem to trigger
     * hotplug irqs when they shouldn't. */
    if (!ljm_dev->mode_config_initialized) {
        return;
    }

    TRACE("hdmi hotplug process...");

    mutex_lock(&mode_config->mutex);
    list_for_each_entry(connector, &mode_config->connector_list, head) {
        ljm_connector_hotplug(connector);
    }
    mutex_unlock(&mode_config->mutex);

#if USE_DP_CONNECTOR
    /* DP */
    if (ljm_dev->dp_irq_is_extra) { /* enable external IRQ if needed */
        if (LJM_MISC_R32(ljm_dev->cur_card, MISC_INTR_SRC) & m_INTR_SRC_EIRQ0) {    /* clear external IRQ */
            ljm_i2c_dw_read_req((void *)ljm_dev->i2c_devs[5], DP_GP201_HPD_TARGET_ADR, DP_GP201_HPD_TARGET_INPUTPORT_SUBADR, &buf, 1);
        }
        misc_intr_mask = LJM_MISC_R32(ljm_dev->cur_card, MISC_INTR_EN);
        if (!(misc_intr_mask & m_INTR_EN_EIRQ0)) {
            LJM_MISC_W32(ljm_dev->cur_card, MISC_INTR_EN, misc_intr_mask | m_INTR_EN_EIRQ0);
        }
    }
#endif

    /* Just fire off a uevent and let userspace tell us what to do */
    drm_helper_hpd_irq_event(ddev);

    TRACE("hdmi hotplug process done");
}

/**
 * ljm_driver_irq_handler_kms - irq handler for KMS
 *
 * @int irq, void *arg: args
 *
 * This is the irq handler for the ljmicro KMS driver (all asics).
 * ljm_irq_process is a macro that points to the per-asic
 * irq handler callback.
 */
irqreturn_t ljm_driver_irq_handler_kms(int irq, void *arg)
{
    struct drm_device *ddev = (struct drm_device *) arg;
    struct ljm_device *ljm_dev;
    irqreturn_t r;

    /**
     * @brief 避免关机流程后出现设备结构体空指针问题
     */
    if (unlikely(!ddev || !ddev->dev_private)) {
        LJM_DISABLE_ALL_IRQ(0);
        printk(KERN_ERR "drm=%px dev_private=%px\n", ddev, ddev ? ddev->dev_private : NULL);
        return IRQ_HANDLED;
    }

#ifndef LJMICRO_DRV_ONLY
    if (unlikely(((ljkGALDEVICE)ddev->dev_private)->gotoShutdown)) {
        LJM_DISABLE_ALL_IRQ(0);
        printk(KERN_WARNING "drm is shutdown\n");
        return IRQ_HANDLED;
    }
#endif

    ljm_dev = GetLjmDevFrmDrm(ddev);
    if (unlikely(!ljm_dev)) {
        LJM_DISABLE_ALL_IRQ(0);
        printk(KERN_ERR "ljm_dev is NULL\n");
        return IRQ_HANDLED;
    }

    r = ljm_irq_process(ljm_dev);
//  if (r == IRQ_HANDLED)
//      pm_runtime_mark_last_busy(ddev->dev);

//  TRACE("irq=%d arg=%p r=%d", irq, arg, r);
    return r;
}


/**
 * ljm_driver_irq_preinstall_kms - drm irq preinstall callback
 *
 * @ddev: drm dev pointer
 *
 * Gets the hw ready to enable irqs (all asics).
 * This function disables all interrupt sources on the GPU.
 */
void ljm_driver_irq_preinstall_kms(struct drm_device *ddev)
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(ddev);
    unsigned long irqflags;
    unsigned i;
    u32 val;

    HEADER_ARG("dev=%p", ddev);

    spin_lock_irqsave(&ljm_dev->irq.lock, irqflags);
    if (ljm_dev->family == CHIP_GP200) {
        val = LJM_MISC_R32(ljm_dev->cur_card, MISC_INTR_EN);

        val &= ~m_INTR_EN_VGA;    /* disable VGA interrupt to pcie */

        if (ljm_dev->dp_irq_is_extra)
            val &= ~m_INTR_EN_EIRQ0;  /* disable external irq0 interrupt to pcie */

        LJM_MISC_W32(ljm_dev->cur_card, MISC_INTR_EN, val);
    }

    /* Disable *all* interrupts */
    for (i = 0; i < ljm_dev->num_ports; i++)
        ljm_dev->ports[i]->hpd_en = false;
    for (i = 0; i < ljm_dev->num_crtcs; i++) {
        ljm_dev->irq.crtc_vblank_int[i] = false;
        atomic_set(&ljm_dev->irq.pflip[i], 0);
    }
    ljm_irq_set(ljm_dev);
    spin_unlock_irqrestore(&ljm_dev->irq.lock, irqflags);

    /* Clear bits */
    ljm_irq_process(ljm_dev);

    FOOTER();
}

/**
 * ljm_driver_irq_postinstall_kms - drm irq preinstall callback
 *
 * @dev: drm dev pointer
 *
 * Handles stuff to be done after enabling irqs (all asics).
 * Returns 0 on success.
 */
int ljm_driver_irq_postinstall_kms(struct drm_device *ddev)
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(ddev);
    unsigned long irqflags;

    HEADER_ARG("ddev=%p", ddev);

    ddev->max_vblank_count = LJM_MAX_VBLANK_CNT;

    spin_lock_irqsave(&ljm_dev->irq.lock, irqflags);
    ljm_irq_postinstall(ljm_dev);
    spin_unlock_irqrestore(&ljm_dev->irq.lock, irqflags);

    FOOTER();

    return 0;
}

/**
 * ljm_driver_irq_uninstall_kms - drm irq uninstall callback
 *
 * @dev: drm dev pointer
 *
 * This function disables all interrupt sources on the GPU (all asics).
 */
void ljm_driver_irq_uninstall_kms(struct drm_device *ddev)
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(ddev);
    unsigned long irqflags;
    unsigned i;

    HEADER_ARG("ddev=%p", ddev);

    if (ljm_dev == NULL) {
        return;
    }

    spin_lock_irqsave(&ljm_dev->irq.lock, irqflags);
    /* Disable *all* interrupts */
    for (i = 0; i < ljm_dev->num_ports; i++)
        ljm_dev->ports[i]->hpd_en = false;
    for (i = 0; i < ljm_dev->num_crtcs; i++) {
        ljm_dev->irq.crtc_vblank_int[i] = false;
        atomic_set(&ljm_dev->irq.pflip[i], 0);
    }
    ljm_irq_set(ljm_dev);

    /* disable VGA interrupt to pcie */
    if (ljm_dev->family == CHIP_GP200) {
        //printk("----disable VGA interrupt----\n");
        LJM_MISC_W32(ljm_dev->cur_card, MISC_INTR_EN,
                     LJM_MISC_R32(ljm_dev->cur_card, MISC_INTR_EN) & ~m_INTR_EN_VGA);
    }
    spin_unlock_irqrestore(&ljm_dev->irq.lock, irqflags);

    FOOTER();
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 2, 0)
static int ljm_irq_install(struct ljm_device *rdev, int irq)
{
	struct drm_device *dev = rdev->ddev;
	int ret;

	if (irq == IRQ_NOTCONNECTED)
		return -ENOTCONN;

	ljm_driver_irq_preinstall_kms(dev);

	/* PCI devices require shared interrupts. */
	ret = request_irq(irq, ljm_driver_irq_handler_kms,
			  IRQF_SHARED, dev->driver->name, dev);
	if (ret)
		return ret;

	ljm_driver_irq_postinstall_kms(dev);

	return 0;
}

static void ljm_irq_uninstall(struct ljm_device *rdev)
{
	struct drm_device *dev = rdev->ddev;
	struct pci_dev *pdev = to_pci_dev(dev->dev);

	ljm_driver_irq_uninstall_kms(dev);
	free_irq(pdev->irq, dev);
}
#endif

/**
 * ljm_irq_kms_init - init driver interrupt info
 *
 * @ljm_dev: ljmicro device pointer
 *
 * Sets up the work irq handlers, vblank init, MSIs, etc. (all asics).
 * Returns 0 for success, error for failure.
 */
int ljm_irq_kms_init(struct ljm_device *ljm_dev)
{
    int r = 0;
    HEADER_ARG("ljm_dev=0x%p", ljm_dev);

    spin_lock_init(&ljm_dev->irq.lock);

#if USE_KMS_VBLANK
    DRM_DEBUG_KMS("init vblank...\n");
    /* init for vblank */
    init_waitqueue_head(&ljm_dev->irq.vblank_queue);
    //printk("----%s: ljm_dev->num_crtcs=%d----\n", __FUNCTION__, ljm_dev->num_crtcs);
    r = drm_vblank_init(ljm_dev->ddev, ljm_dev->num_crtcs);
    if (r) {
        FOOTER_ARG("r=%d", r);
        return r;
    }
#endif

    if (ljm_msi_ok(ljm_dev)) {
#ifdef LJMICRO_DRV_ONLY
        r = pci_enable_msi(ljm_dev->pdev);
        if (!r) {
            ljm_dev->msi_enabled = 1;
            dev_info(ljm_dev->dev, "ljmicro: using MSI.\n");
        }
#else
        /* pci msi has been enabled in csmcore */
        ljm_dev->msi_enabled = 1;
#endif
    }

    INIT_DELAYED_WORK(&ljm_dev->hotplug_work, ljm_hotplug_work_func);

    ljm_dev->irq.installed = true;
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 2, 0)
    r = drm_irq_install(ljm_dev->ddev, ljm_dev->ddev->pdev->irq);
#else
    r = ljm_irq_install(ljm_dev, ljm_dev->pdev->irq);
#endif
    if (r) {
        ljm_dev->irq.installed = false;
        flush_delayed_work(&ljm_dev->hotplug_work);
        FOOTER_ARG("r=%d", r);
        return r;
    }

    //DRM_INFO("%s: drm irq %d initialized.\n", ljm_dev->ddev->driver->name , ljm_dev->ddev->pdev->irq);
    DRM_INFO("%s: drm irq %d initialized.\n", ljm_dev->ddev->driver->name , ljm_dev->pdev->irq);
    FOOTER();
    return 0;
}

/**
 * ljm_irq_kms_fini - tear down driver interrupt info
 *
 * @ljm_dev: ljmicro device pointer
 *
 * Tears down the work irq handlers, vblank handlers, MSIs, etc. (all asics).
 */
void ljm_irq_kms_fini(struct ljm_device *ljm_dev)
{
#if USE_KMS_VBLANK
#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 12, 0)   /* kernel(91e976777d4ad068ef61b005a192457a879c81dc) drm/amd|radeon: Drop drm_vblank_cleanup */
        drm_vblank_cleanup(ljm_dev->ddev);
#endif
#endif
    if (ljm_dev->irq.installed) {
        //drm_irq_uninstall(ljm_dev->ddev);
#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 2, 0)
        ljm_irq_uninstall(ljm_dev);
#endif
        ljm_dev->irq.installed = false;
#ifdef LJMICRO_DRV_ONLY
        if (ljm_dev->msi_enabled)
            pci_disable_msi(ljm_dev->pdev);
#endif
        flush_delayed_work(&ljm_dev->hotplug_work);
    }
}

