/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#include "ljmicro_base.h"
#include "ljmicro_device.h"
#include <linux/vgaarb.h>
#include "ljmicro_connector_codec.h"
#ifndef LJMICRO_DRV_ONLY
#include "lj_hal_kernel_platform_gp.h"
#endif

#define _LJ_OBJ_ZONE    ljvZONE_KMS

int ljm_device_init(struct ljm_device *ljm_dev, struct drm_device *ddev, struct pci_dev *pdev, uint32_t flags);
void ljm_device_fini(struct ljm_device *ljm_dev);

int ljm_irq_kms_init(struct ljm_device *ljm_dev);
void ljm_irq_kms_fini(struct ljm_device *ljm_dev);

struct _gcsPLATFORM_PCIE;

/**
 * ljm_driver_unload_kms - Main unload function for KMS.
 *
 * @dev: drm dev pointer
 *
 * This is the main unload function for KMS (all asics).
 * It calls ljm_modeset_fini() to tear down the
 * displays, and ljm_device_fini() to tear down
 * the rest of the device (CP, writeback, etc.).
 * Returns 0 on success.
 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,11,0)
void ljm_driver_unload_kms(struct drm_device *ddev)
#else
int ljm_driver_unload_kms(struct drm_device *ddev)
#endif
{
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(ddev);

    HEADER_ARG("ddev=%p", ddev);

    /* TODO */
    if (ljm_dev == NULL) {
        FOOTER();
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,11,0)
        return;
#else
        return 0;
#endif
    }

    if (ljm_dev->mmio_base == NULL)
        goto do_free;

    ljm_dev->is_shutdown = true;

    ljm_modeset_fini(ljm_dev);

    /* free irq manager */
    ljm_irq_kms_fini(ljm_dev);

    /* free GP device */
    //ljm_device_fini(ljm_dev);

do_free:
    kfree(ljm_dev);
    ljm_dev = NULL;
    SetLjmDevFrmDrm(ddev, ljm_dev);
    FOOTER();
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4,11,0)
#else
    return 0;
#endif
}

/* true = enable decode, false = disable decoder */
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 2, 0)
static unsigned int ljm_vga_set_decode(void *cookie, bool state)
#else
static unsigned int ljm_vga_set_decode(struct pci_dev *pdev, bool state)
#endif
{
	//struct ljm_device *ljm_dev = cookie;

	//ljm_modeset_vga_set_state(dev_priv, state);

	//printk("ljm_vga_set_decode:%d\n", state);

	if (state)
		return VGA_RSRC_LEGACY_IO | VGA_RSRC_LEGACY_MEM |
		       VGA_RSRC_NORMAL_IO | VGA_RSRC_NORMAL_MEM;
	else
		return VGA_RSRC_NORMAL_IO | VGA_RSRC_NORMAL_MEM;
}

#ifdef LJMICRO_DRV_ONLY
extern int dbg_check(void);
#endif

#if IS_ENABLED(CONFIG_SND_SOC) && defined(ljdENABLE_AUDIO)
extern int ljm_audio_card_register(struct ljm_device *ljm_dev);
#endif

/**
 * ljm_driver_load_kms - Main load function for KMS.
 *
 * @dev: drm dev pointer
 * @flags: device flags
 *
 * This is the main load function for KMS (all asics).
 * It calls ljm_device_init() to set up the non-display
 * parts of the chip (asic init, CP, writeback, etc.), and
 * ljm_modeset_init() to set up the display parts
 * (crtcs, encoders, hotplug detect, etc.).
 * Returns 0 on success, error on failure.
 */
int ljm_driver_load_kms(struct drm_device *ddev, unsigned long flags)
{
    struct ljm_device *ljm_dev;
    int r;
    struct pci_dev *pdev;

    HEADER_ARG("ddev=%p flags=%lx", ddev, flags);

    ljm_dev = kzalloc(sizeof(struct ljm_device), GFP_KERNEL);
    if (ljm_dev == NULL) {
        FOOTER_ARG("r=%d", -ENOMEM);
        return -ENOMEM;
    }

    SetLjmDevFrmDrm(ddev, ljm_dev);
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 2, 0)
    pdev = ddev->pdev;
#else
    pdev = to_pci_dev(ddev->dev);
#endif

    //增加vgarb相关代码,解决集成显卡存在时DRI2无法加载的问题
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 2, 0)
    r= vga_client_register(pdev, ljm_dev, NULL, ljm_vga_set_decode);
#else
    r= vga_client_register(pdev, ljm_vga_set_decode);
#endif
    if(r&&r!=-ENODEV)
	return r;

    /* init ljm device
     * ljm_device_init should report only fatal error
     * like memory allocation failure or iomapping failure,
     * or memory manager initialization failure, it must
     * properly initialize the GPU controller and permit
     * VRAM allocation
     */
    r = ljm_device_init(ljm_dev, ddev, pdev, flags);
    if (r) {
#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 2, 0)
        dev_err(&ddev->pdev->dev, "Fatal error during GPU init\n");
        dev_err(&pdev->dev, "Fatal error during GPU init\n");
#else
        dev_err(ddev->dev, "Fatal error during GPU init\n");
#endif
        goto out;
    }

    /* init a drm interrupt */
    r = ljm_irq_kms_init(ljm_dev);

    /* init ljm modeset
     * Again modeset_init should fail only on fatal error
     * otherwise it should provide enough functionalities
     * for shadowfb to run
     */
    r = ljm_modeset_init(ljm_dev);

#if IS_ENABLED(CONFIG_SND_SOC) && defined(ljdENABLE_AUDIO)
    /* register audio codec and i2s */
    ljm_codec_register(ljm_dev);
    ljm_i2s_register(ljm_dev);
    if ((ljm_dev->flags & LJMICRO_CHIP_MASK) == CHIP_GP201) {
        ljm_pcm_register(ljm_dev->dev);
    }
    else {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 6, 0)
        ljm_pcm_register(ljm_dev->dev);
#else
        ljm_dma_pcm_register(ljm_dev->dev);
#endif
    }

    /* register audio card */
    ljm_audio_card_register(ljm_dev);
#endif

////////////////////////////////////////////////////////////
//  /* 当DRIVER_MODESET特征设置后，下面设置可以取消 */
//  pci_set_drvdata(ddev->pdev, ddev);
//  ljm_dev->ddev = ddev;
//  ljm_dev->pdev = ddev->pdev;
//  ljm_dev->dev = &ddev->pdev->dev;
///////////////////////////////////////////////////////////

#ifdef LJMICRO_DRV_ONLY
    if(dbg_check())
        TRACE("debug: headers and footers are not in pair!");
#endif

out:
    if (r) {
        DRM_ERROR("Fatal error during modeset init\n");
        ljm_driver_unload_kms(ddev);
    }

    FOOTER();
    return r;
}

#if USE_KMS_VBLANK
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 6, 0)
u32 ljm_get_vblank_counter_kms(struct drm_crtc *crtc)
{
    struct drm_device *ddev = crtc->dev;
    unsigned int pipe = crtc->index;
#else
u32 ljm_get_vblank_counter_kms(struct drm_device *ddev, unsigned int pipe)
{
#endif
    u32 count;
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(ddev);

    //HEADER_ARG("ddev=%p crtc=%d", ddev, pipe);

    if (pipe < 0 || pipe >= ljm_dev->num_crtcs) {
        DRM_ERROR("Invalid crtc %u\n", pipe);
        //FOOTER_ARG("r=%d", -EINVAL);
        return -EINVAL;
    }

    /* GP200: vblank interrupt generated at start of vsync, vblank counter increased at start of vsync */
    count = ljm_get_vblank_counter(ljm_dev, pipe);

    //FOOTER();
    return count;
}

/**
 * ljm_enable_vblank - enable vblank interrupt
 *
 * @ddev: drm dev pointer
 * @crtc: crtc to enable vblank interrupt for
 *
 * Enable the interrupt on the requested crtc (all asics).
 * Returns 0 on success, -EINVAL on failure.
 */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 6, 0)
int ljm_enable_vblank_kms(struct drm_crtc *crtc)
{
    struct drm_device *ddev = crtc->dev;
    unsigned int pipe = crtc->index;
#else
int ljm_enable_vblank_kms(struct drm_device *ddev, unsigned int pipe)
{
#endif
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(ddev);
    unsigned long irqflags;
    int r;

    HEADER_ARG("ddev=0x%p crtc=%d", ddev, pipe);

    DRM_DEBUG_VBL("crtc %d: enable vblank\n", pipe);

    if (pipe < 0 || pipe >= ljm_dev->num_crtcs) {
        DRM_ERROR("Invalid crtc %d\n", pipe);
        FOOTER_ARG("r=%d", -EINVAL);
        return -EINVAL;
    }

    spin_lock_irqsave(&ljm_dev->irq.lock, irqflags);
    ljm_dev->irq.crtc_vblank_int[pipe] = true;
    r = ljm_irq_vblank_set(ljm_dev);
    spin_unlock_irqrestore(&ljm_dev->irq.lock, irqflags);

    FOOTER();
    return r;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 6, 0)
void ljm_disable_vblank_kms(struct drm_crtc *crtc)
{
    struct drm_device *ddev = crtc->dev;
    unsigned int pipe = crtc->index;
#else
void ljm_disable_vblank_kms(struct drm_device *ddev, unsigned int pipe)
{
#endif
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(ddev);
    unsigned long irqflags;

    HEADER_ARG("ddev=0x%p crtc=%d", ddev, pipe);

    DRM_DEBUG_VBL("crtc %d: disable vblank\n", pipe);

    if (pipe < 0 || pipe >= ljm_dev->num_crtcs)
    {
        DRM_ERROR("Invalid crtc %d\n", pipe);

        FOOTER();
        return;
    }

    spin_lock_irqsave(&ljm_dev->irq.lock, irqflags);
    ljm_dev->irq.crtc_vblank_int[pipe] = false;
    ljm_irq_vblank_set(ljm_dev);
    spin_unlock_irqrestore(&ljm_dev->irq.lock, irqflags);

    FOOTER();
}

#if 0
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 6, 0)
bool ljm_get_vblank_timestamp_kms(struct drm_crtc *crtc, int *max_error, ktime_t *vblank_time, bool in_vblank_irq)
{
    struct drm_device *ddev = crtc->dev;
    unsigned int pipe = crtc->index;
#else
bool ljm_get_vblank_timestamp_kms(struct drm_device *ddev, unsigned int pipe, int *max_error, ktime_t *vblank_time, bool in_vblank_irq)
{
#endif
//TODO
}
#endif

#endif

