/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#include <linux/list.h>
#include <linux/slab.h>
#include "ljmicro_device.h"
#include "ljmicro_object.h"
#include "ljmicro_mode.h"

#define _LJ_OBJ_ZONE    ljvZONE_KMS

#ifdef LJMICRO_DRV_ONLY
void ljm_bo_force_delete(struct ljm_device *ljm_dev)
{
    struct ljm_gem_object *bo, *n;

    HEADER_ARG("ljm_dev=0x%p", ljm_dev);

    if (list_empty(&ljm_dev->gem.objects)) {
        FOOTER();
        return;
    }
    dev_err(ljm_dev->dev, "Userspace still has active objects !\n");

    list_for_each_entry_safe(bo, n, &ljm_dev->gem.objects, list) {
        dev_err(ljm_dev->dev, "%p %p %lu %lu force free\n",
            &bo->base, bo, (unsigned long)bo->base.size,
            *((unsigned long *)&bo->base.refcount));

        mutex_lock(&ljm_dev->gem.mutex);
        list_del_init(&bo->list);
        mutex_unlock(&ljm_dev->gem.mutex);

        /* this should unref the gem bo */
        drm_gem_object_unreference_unlocked(&bo->base);
    }

    FOOTER();
}
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,11,0) && !defined LJMICRO_DRV_ONLY
extern const struct drm_gem_object_funcs ljm_gem_object_funcs;
#endif

int ljm_bo_create(struct ljm_device *ljm_dev, unsigned long size, struct ljm_gem_object **bo_ptr)
{
    struct ljm_gem_object *bo;
    int r;

    HEADER_ARG("ljm_dev=0x%p size=%lu", ljm_dev, size);

    size = ALIGN(size, PAGE_SIZE);

    bo = kzalloc(sizeof(struct ljm_gem_object), GFP_KERNEL);
    if (bo == NULL) {
        FOOTER_ARG("r=%d", -ENOMEM);
        return -ENOMEM;
    }

    r = drm_gem_object_init(ljm_dev->ddev, &bo->base, size);
    if (unlikely(r)) {
        kfree(bo);
        FOOTER_ARG("r=%d", r);
        return r;
    }

    bo->node_handle = (uint32_t)(-1);

#ifdef LJMICRO_DRV_ONLY
    bo->ljm_dev = ljm_dev;
    INIT_LIST_HEAD(&bo->list);
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5,11,0)
    bo->base.funcs = &ljm_gem_object_funcs;
#endif

    *bo_ptr = bo;

    FOOTER_ARG("%p", *bo_ptr);
    return 0;
}

#if defined DEBUG && defined LJMICRO_DRV_ONLY
void ljm_dump_vram_list(struct ljm_device *ljm_dev, const char * title)
{
    struct ljm_mm_obj *mo, *mo_next;
    int i = 0;

    /* dump free VRAM */
    if (title) printk(KERN_INFO "---- dump vram list (%s) ------------\n", title);
    else printk(KERN_INFO "---- dump vram list ------------\n");
    list_for_each_entry_safe(mo, mo_next, &ljm_dev->vram_free.list, list) {
        printk(KERN_INFO "%d:\t%p <- &list 0x%p (0x%p: %zu@0x%lx) -> 0x%p\n",
            i++, mo->list.prev, &mo->list, mo, mo->size, mo->offset, mo->list.next);
    }
    printk(KERN_INFO "----^--^------------------------\n");
}
#define DUMP_VRAM_LIST(LJM_DEV, TITLE)  ljm_dump_vram_list(LJM_DEV, TITLE)
#else
#define DUMP_VRAM_LIST(LJM_DEV, TITLE)
#endif

int ljm_get_process_id(void)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 24)
    return task_tgid_vnr(current);
#else
    return current->tgid;
#endif
}

char *ljm_get_process_name(void)
{
    static char name[TASK_COMM_LEN];

    get_task_comm(name, current);
    return name;
}

/* allocate VRAM for Gem object */
int ljm_get_vram(struct drm_device *ddev, struct ljm_gem_object *bo)
{
#ifdef LJMICRO_DRV_ONLY
    struct ljm_mm_obj *mo, *mo_next, *mo_prev;
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(ddev);

    HEADER_ARG("ddev=0x%p bo=0x%p", ddev, bo);

    DUMP_VRAM_LIST(ljm_dev, "before ljm_get_vram");

    /* Last memory block is the max block. Check whether there is enough memory. */
    mo = list_last_entry(&ljm_dev->vram_free.list, typeof(*mo), list);
    if(bo->base.size > mo->size) {
        DRM_ERROR("failed to alloc vram bo (%dB), the max left block (%dB) is too small\n", (unsigned int)bo->base.size, (unsigned int)mo->size);

        FOOTER_ARG("r=%d", -ENOMEM);
        return -ENOMEM;
    }

    /* allocate a memory block for bo */
    list_for_each_entry_safe(mo, mo_next, &ljm_dev->vram_free.list, list) {
        if (mo->size >= bo->base.size) {    /* bo is enough */
            bo->size = bo->base.size;
            bo->offset = mo->offset;
            bo->gpu_addr = mo->offset + ljm_dev->vram_free.vram_start;
            bo->cpu_addr = mo->offset + ljm_dev->smem_start;

            TRACE("Get %zu bytes @ VRAM 0x%08lx (gpu_addr=0x%08lx cpu_addr=0x%08lx)",
                bo->size, bo->offset, (unsigned long)bo->gpu_addr, (unsigned long)bo->cpu_addr
                );

            mutex_lock(&ljm_dev->vram_free.mutex);
            mo->offset += bo->size;
            mo->size -= bo->size;
            ljm_dev->vram_free.vram_size -= bo->size;

            if (mo->size == 0) {    /* If bo is empty, just delete it. */
                list_del(&mo->list);
                kfree(mo);
                mutex_unlock(&ljm_dev->vram_free.mutex);
                break;  /* break list_for_each_entry_safe */
            }
            mutex_unlock(&ljm_dev->vram_free.mutex);

            /* sort the memory block */
            mo_prev = list_entry(mo->list.prev, typeof(*mo), list);
            while (&mo_prev->list != &ljm_dev->vram_free.list) {
                if (mo_prev->size > mo->size) {
                    mutex_lock(&ljm_dev->vram_free.mutex);
                    /* exchange prev and bo */
                    mo_prev->list.next = mo->list.next;
                    mo->list.next = &mo_prev->list;
                    mo->list.prev = mo_prev->list.prev;
                    mo_prev->list.prev = &mo->list;
                    mutex_unlock(&ljm_dev->vram_free.mutex);

                    /* prev move to previous of bo */
                    mo_prev = list_entry(mo->list.prev, typeof(*mo), list);
                }
                else {
                    /* sort end */
                    break;
                }
            }
            break;
        }
    }

    DUMP_VRAM_LIST(ljm_dev, "after ljm_get_vram");

    FOOTER_ARG("%zu bytes @0x%08lx", bo->size, bo->offset);
    return 0;
#else
    ljkGALDEVICE gal_dev;
    ljkDEVICE device;
    ljkKERNEL kernel;
    ljkVIDMEM videoMemory;
    ljtSIZE_T alignSize = bo->base.size;
    ljeSTATUS status = ljvSTATUS_OK;

    HEADER_ARG("ddev=%p bo=%p pid=%d(%s)", ddev, bo, ljm_get_process_id(), ljm_get_process_name());

    if (bo->node_object) {
        ljkVIDMEM_NODE_Reference(bo->node_object->kernel, bo->node_object);

        FOOTER_ARG("%zu bytes", BO_GET_SIZE(bo));
        return 0;
    }

    gal_dev = GetGalFrmDrm(ddev);
    if (!gal_dev)
    {
        status = ljvSTATUS_INVALID_ARGUMENT;
        goto OnError;
    }
    device = gal_dev->devices[0];
    kernel = device->kernels[0];

    /* Allocate video memory node in external. */
    videoMemory = gal_dev->externalVidMem[device->platformIndex];
    status = ljkVIDMEM_NODE_AllocateLinear(
        kernel,
        videoMemory,
        ljvPOOL_LOCAL_EXTERNAL,
        ljvVIDMEM_TYPE_GENERIC,
        ljvALLOC_FLAG_DMABUF_EXPORTABLE | ljvALLOC_FLAG_CONTIGUOUS,
        256,
        ljvFALSE,
        &alignSize,
        &bo->node_object
        );

    if (ljmIS_SUCCESS(status)) {
        DRM_DEBUG_KMS("get %zu bytes@0x%08llx from external pool\n", BO_GET_SIZE(bo), BO_GET_VIDMEM_OFFSET(bo));
    } else {
        /* Allocate video memory node in exclusive. */
        videoMemory = gal_dev->exclusiveVidMem[device->platformIndex];
        ljmkONERROR(ljkVIDMEM_NODE_AllocateLinear(
            kernel,
            videoMemory,
            ljvPOOL_LOCAL_EXCLUSIVE,
            ljvVIDMEM_TYPE_GENERIC,
            ljvALLOC_FLAG_DMABUF_EXPORTABLE | ljvALLOC_FLAG_CONTIGUOUS,
            256,
            ljvFALSE,
            &alignSize,
            &bo->node_object
            ));
        DRM_DEBUG_KMS("get %zu bytes@0x%08llx from exclusive pool", BO_GET_SIZE(bo), BO_GET_VIDMEM_OFFSET(bo));
    }

    DRM_INFO("allocate new vram: %zu\n", bo->base.size);

OnError:
    FOOTER_ARG("%zu bytes", BO_GET_SIZE(bo));

    return ljmIS_ERROR(status) ? -ENOTTY : 0;
#endif
}

/* release VRAM from Gem object */
int ljm_put_vram(struct drm_device *ddev, struct ljm_gem_object *bo)
{
#ifdef LJMICRO_DRV_ONLY
    struct ljm_mm_obj *mo, *mo_next;
    bool join = false;
    struct ljm_device *ljm_dev = GetLjmDevFrmDrm(ddev);

    HEADER_ARG("ddev=0x%p bo=0x%p bo->base.size=%zu", ddev, bo, bo->base.size);

    DUMP_VRAM_LIST(ljm_dev, "before ljm_put_vram");

    list_for_each_entry_safe(mo, mo_next, &ljm_dev->vram_free.list, list) {
        if (bo->offset == mo->offset + mo->size) {  /* GEM object connects at the end of mo. Join it to bo. */
            mutex_lock(&ljm_dev->vram_free.mutex);
            mo->size += bo->size;
            mutex_unlock(&ljm_dev->vram_free.mutex);

            {   /* check and merge ljm_mm_obj */
                struct ljm_mm_obj *pos, *n;
                list_for_each_entry_safe(pos, n, &ljm_dev->vram_free.list, list) {
                    if (mo->offset + mo->size == pos->offset) { /* as mo expands at end, merging only happens at end of mo */
                        TRACE("mm_obj 0x%p(%zu@0x%lx) can connect with pos 0x%p(%zu@%lx)",
                            mo, mo->size, mo->offset, pos, pos->size, pos->offset
                            );

                        DUMP_VRAM_LIST(ljm_dev, "before merge");

                        mutex_lock(&ljm_dev->vram_free.mutex);
                        mo->size += pos->size;
                        list_del(&pos->list);
                        mo_next = container_of(mo->list.next, typeof(*mo_next), list);  /* as mo_next pointed to pos, which was deleted, mo_next should update to mo->list.next */
                        mutex_unlock(&ljm_dev->vram_free.mutex);
                        kfree(pos);

                        DUMP_VRAM_LIST(ljm_dev, "after merge");
                        break;
                    }
                }
            }

            join = true;
        }
        else if (bo->offset + bo->size == mo->offset) { /* GEM object connects at the begin of bo. Join it to bo. */
            mutex_lock(&ljm_dev->vram_free.mutex);
            mo->offset = bo->offset;
            mo->size += bo->size;
            mutex_unlock(&ljm_dev->vram_free.mutex);

            {   /* check and merge ljm_mm_obj */
                struct ljm_mm_obj *pos, *n;

                list_for_each_entry_safe(pos, n, &ljm_dev->vram_free.list, list) {
                if (mo->offset == pos->offset + pos->size) {    /* as mo expands at begin, merging only happens at begin of mo */
                    TRACE("pos 0x%p(%zu@0x%lx) can connect with mm_obj 0x%p(%zu@%lx)",
                        pos, pos->size, pos->offset, mo, mo->size, mo->offset
                        );

                    DUMP_VRAM_LIST(ljm_dev, "before merge");

                    mutex_lock(&ljm_dev->vram_free.mutex);
                    mo->size += pos->size;
                    mo->offset = pos->offset;
                    list_del(&pos->list);
                    mo_next = container_of(mo->list.next, typeof(*mo_next), list);  /* as mo_next pointed to pos, which was deleted, mo_next should update to mo->list.next */
                    mutex_unlock(&ljm_dev->vram_free.mutex);
                    kfree(pos);

                    DUMP_VRAM_LIST(ljm_dev, "before merge");
                    break;
                    }
                }
            }

            join = true;
        }

        if (join) { /* sort the memory block */
            while (&mo_next->list != &ljm_dev->vram_free.list) {
                if (mo_next->size < mo->size) {
                    /* exchange next and bo */
                    mutex_lock(&ljm_dev->vram_free.mutex);
                    mo->list.next = mo_next->list.next;
                    mo_next->list.next = &mo->list;
                    mo_next->list.prev = mo->list.prev;
                    mo->list.prev = &mo_next->list;
                    mutex_unlock(&ljm_dev->vram_free.mutex);

                    /* prev move to previous of bo */
                    mo_next = list_entry(mo->list.next, typeof(*mo_next), list);
                }
                else {
                    /* sort end */
                    break;  /* break while */
                }
            }

            break;  /* break list_for_each_entry_safe */
        }
    }

    /* Can not find close-by VRAM block. Create a new one and sort it. */
    if (!join) {
        struct ljm_mm_obj *new_mo = kzalloc(sizeof(struct ljm_mm_obj), GFP_KERNEL);
        if (!new_mo) {
            DRM_ERROR("Out of memory\n");
            FOOTER_ARG("r=%d", -ENOMEM);
            return -ENOMEM;
        }

        new_mo->offset = bo->offset;
        new_mo->size = bo->size;
        INIT_LIST_HEAD(&new_mo->list);

        if (list_empty(&ljm_dev->vram_free.list)) {
            mutex_lock(&ljm_dev->vram_free.mutex);
            list_add(&new_mo->list, &ljm_dev->vram_free.list);
            mutex_unlock(&ljm_dev->vram_free.mutex);
        }
        else {
            list_for_each_entry_safe(mo, mo_next, &ljm_dev->vram_free.list, list) {
                if (new_mo->size <= mo->size) {      /* New bo is smaller than current bo, add the new on bo's tail. */
                    mutex_lock(&ljm_dev->vram_free.mutex);
                    list_add_tail(&new_mo->list, &mo->list);
                    mutex_unlock(&ljm_dev->vram_free.mutex);
                    break;
                }
                else if (&mo_next->list == &ljm_dev->vram_free.list) {   /* New bo is the largest, add the new on bo's head. */
                    mutex_lock(&ljm_dev->vram_free.mutex);
                    list_add(&new_mo->list, &mo->list);
                    mutex_unlock(&ljm_dev->vram_free.mutex);
                    break;
                }
            }
        }
    }

    /* update global VRAM size */
    mutex_lock(&ljm_dev->vram_free.mutex);
    ljm_dev->vram_free.vram_size += bo->size;
    mutex_unlock(&ljm_dev->vram_free.mutex);
    bo->size = 0;

    DUMP_VRAM_LIST(ljm_dev, "after ljm_put_vram");
#else

    HEADER_ARG("ddev=%p bo=%p(%px) %d bo->base.size=%zu pid=%d(%s)", ddev, bo, bo, bo->node_handle, bo->base.size, ljm_get_process_id(), ljm_get_process_name());

    DRM_DEBUG_KMS("put %zu bytes@0x%08llx\n", BO_GET_SIZE(bo), BO_GET_VIDMEM_OFFSET(bo));
    ljkVIDMEM_NODE_Dereference(bo->node_object->kernel, bo->node_object);
#endif

    FOOTER();
    return 0;
}

int ljm_bo_kmap(struct ljm_gem_object *bo, void **ptr)
{
    HEADER_ARG("bo=0x%p ptr=%p", bo, ptr);

#ifdef LJMICRO_DRV_ONLY
    if (bo->kptr == NULL) {
        bo->kptr = ioremap_wc(get_cpu_addr_from_bo(bo), bo->base.size);
        if (!bo->kptr) {
            FOOTER_ARG("r=%d", -ENOMEM);
            return -ENOMEM;
        }

        TRACE("Map 0x%zu bytes @ 0x%08lx to 0x%08lx",
            bo->base.size, (unsigned long)get_cpu_addr_from_bo(bo), (unsigned long)bo->kptr
            );
    }

    if (ptr) {
        *ptr = bo->kptr;
    }
#else
    if (ljkVIDMEM_NODE_LockCPU(bo->node_object->kernel, bo->node_object, ljvFALSE, ljvFALSE, (ljtPOINTER *)ptr) != ljvSTATUS_OK) {
        DRM_ERROR("bo(%lx): map kernel address failed\n", (unsigned long)bo);
        FOOTER_ARG("r=%d", -ENOMEM);
        return -ENOMEM;
    }
#endif

    FOOTER();
    return 0;
}

void ljm_bo_kunmap(struct ljm_gem_object *bo)
{
    HEADER();

#ifdef LJMICRO_DRV_ONLY
    if (bo->kptr) {
        iounmap(bo->kptr);
        bo->kptr = NULL;
    }
#elif 0
    if (bo->node_object->node->VidMem.logical) {
        ljkVIDMEM_NODE_UnlockCPU(bo->node_object->kernel, bo->node_object, 0, ljvFALSE, ljvTRUE);
    }
#endif

    FOOTER();
}

/* GEM pin a block of VRAM */
int ljm_bo_pin(struct ljm_gem_object *bo, u64 *vram_offset)
{
    struct ljm_device *ljm_dev;
    int r = 0;

#ifdef LJMICRO_DRV_ONLY
    HEADER_ARG("bo=%p pid=%d(%s)", bo, ljm_get_process_id(), ljm_get_process_name());

    ljm_dev = GetLjmDevFrmDrm(bo->base.dev);

    if (bo->pin_count) {    /* bo has been pinned */
        bo->pin_count++;
        DRM_INFO("warning: bo(%p)->pin_count=%d\n", bo, bo->pin_count);
    }
    else {
        /* pin a new vram block */
        bo->pin_count = 1;
        r = ljm_get_vram(ljm_dev->ddev, bo);
        if (r) {
            DRM_ERROR("Out of VRAM\n");
            FOOTER_ARG("r=%d", r);
            return r;
        }
    }

    if (vram_offset) {
        *vram_offset = get_offset_from_bo(bo);
        FOOTER_ARG("*gpu_offset=0x%08llx", *vram_offset);
    }
    else {
        FOOTER();
    }

    return 0;
#else   /* !LJMICRO_DRV_ONLY */
    HEADER_ARG("bo=%p(%px) pid=%d(%s)", bo, bo, ljm_get_process_id(), ljm_get_process_name());

    ljm_dev = GetLjmDevFrmDrm(bo->base.dev);

    if ((r = ljm_get_vram(ljm_dev->ddev, bo)) != 0) {
        DRM_ERROR("Out of VRAM\n");
        if (vram_offset)
            *vram_offset = 0;

        FOOTER_ARG("r=%d", r);
        return r;
    }

    if (vram_offset) {
        if (bo->node_object->pool == ljvPOOL_LOCAL_EXCLUSIVE) {
            *vram_offset = BO_GET_VIDMEM_OFFSET(bo);
        } else {
            *vram_offset = BO_GET_VIDMEM_OFFSET(bo)  + BO_GET_VIDMEM_START(bo) - ljm_dev->smem_start;
        }
        TRACE("*vram_offset=0x%08llx", *vram_offset);
    }

    FOOTER();
    return r;
#endif  /* ~LJMICRO_DRV_ONLY */
}

int ljm_bo_unpin(struct ljm_gem_object *bo)
{
    int r = 0;

    HEADER_ARG("bo=%p(%px) pid=%d(%s)", bo, bo, ljm_get_process_id(), ljm_get_process_name());

#ifdef LJMICRO_DRV_ONLY
    if (!bo->pin_count) {
        dev_warn(bo->base.dev->dev, "%p unpin not necessary\n", bo);
        FOOTER();
        return 0;
    }
    bo->pin_count--;
    if (bo->pin_count) {
        DRM_DEBUG_KMS("bo(%p)->pin_count=%d, skip put vram\n", bo, bo->pin_count);
        FOOTER();
        return 0;
    }

    if(bo->kptr) {
        iounmap(bo->kptr);
        bo->kptr = NULL;
    }
#else
    /* do nothing for unpin by hym on 20220922 */
    r = ljm_put_vram(bo->base.dev, bo);
#endif

    FOOTER();
    return r;
}

