/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

/**************************************************************************
* 版权所有 (C)2022, 武汉凌久微电子有限公司。
*
* 文件名称： ljmicro_pcm.c
* 文件标识：
* 内容摘要： 本文件定义pcm驱动，pcm将buffer中的音频数据传递给IIS的FIFO
* 其它说明：
* 当前版本： v1.0
* 作    者： zhoupw, 武汉凌久微电子有限公司
* 完成日期： 2022年4月27日
***************************************************************************/
#if IS_ENABLED(CONFIG_SND_SOC) && defined(ljdENABLE_AUDIO)

#include <linux/io.h>
#include <linux/rcupdate.h>
#include <sound/pcm.h>
#include <sound/pcm_params.h>
#include "ljmicro_i2s.h"
#include "ljmicro_device.h"
#include "lj_hal_kernel_linux.h"

#define BUFFER_BYTES_MAX    (3 * 2 * 8 * PERIOD_BYTES_MIN)
#define PERIOD_BYTES_MIN    4096
#define PERIODS_MIN         2

struct ljm_device *get_ljm_dev(void);

#define ljm_pcm_tx_fn(sample_bits) \
static unsigned int ljm_pcm_tx_##sample_bits(struct ljm_i2s_dev *dev, \
        struct snd_pcm_runtime *runtime, unsigned int tx_ptr, \
        bool *period_elapsed) \
{ \
    const u##sample_bits (*p)[2] = (void *)runtime->dma_area; \
    unsigned int period_pos = tx_ptr % runtime->period_size; \
    int i; \
\
    for (i = 0; i < dev->fifo_th; i++) { \
        iowrite32(p[tx_ptr][0], dev->i2s_base + LRBR_LTHR(0)); \
        iowrite32(p[tx_ptr][1], dev->i2s_base + RRBR_RTHR(0)); \
        period_pos++; \
        if (++tx_ptr >= runtime->buffer_size) { \
            tx_ptr = 0; \
        } \
    } \
    *period_elapsed = period_pos >= runtime->period_size; \
    return tx_ptr; \
}

#define ljm_pcm_rx_fn(sample_bits) \
static unsigned int ljm_pcm_rx_##sample_bits(struct ljm_i2s_dev *dev, \
        struct snd_pcm_runtime *runtime, unsigned int rx_ptr, \
        bool *period_elapsed) \
{ \
    u##sample_bits (*p)[2] = (void *)runtime->dma_area; \
    unsigned int period_pos = rx_ptr % runtime->period_size; \
    int i; \
\
    for (i = 0; i < dev->fifo_th; i++) { \
        p[rx_ptr][0] = ioread32(dev->i2s_base + LRBR_LTHR(0)); \
        p[rx_ptr][1] = ioread32(dev->i2s_base + RRBR_RTHR(0)); \
        period_pos++; \
        if (++rx_ptr >= runtime->buffer_size) \
            rx_ptr = 0; \
    } \
    *period_elapsed = period_pos >= runtime->period_size; \
    return rx_ptr; \
}

ljm_pcm_tx_fn(16);
ljm_pcm_tx_fn(32);
ljm_pcm_rx_fn(16);
ljm_pcm_rx_fn(32);

#undef ljm_pcm_tx_fn
#undef ljm_pcm_rx_fn

static const struct snd_pcm_hardware ljm_pcm_hardware = {
    .info = SNDRV_PCM_INFO_INTERLEAVED |
        SNDRV_PCM_INFO_MMAP |
        SNDRV_PCM_INFO_MMAP_VALID |
        SNDRV_PCM_INFO_BLOCK_TRANSFER,
    .rates = SNDRV_PCM_RATE_32000 |
        SNDRV_PCM_RATE_44100 |
        SNDRV_PCM_RATE_48000,
    .rate_min = 32000,
    .rate_max = 48000,
    .formats = SNDRV_PCM_FMTBIT_S16_LE |
        SNDRV_PCM_FMTBIT_S24_LE |
        SNDRV_PCM_FMTBIT_S32_LE,
    .channels_min = 2,
    .channels_max = 2,
    .buffer_bytes_max = BUFFER_BYTES_MAX,
    .period_bytes_min = PERIOD_BYTES_MIN,
    .period_bytes_max = BUFFER_BYTES_MAX / PERIODS_MIN,
    .periods_min = PERIODS_MIN,
    .periods_max = BUFFER_BYTES_MAX / PERIOD_BYTES_MIN,
    .fifo_size = 16,
};

static void ljm_pcm_transfer(struct ljm_i2s_dev *dev, bool push)
{
    struct snd_pcm_substream *substream;
    bool active, period_elapsed;

    rcu_read_lock();
    if (push)
        substream = rcu_dereference(dev->tx_substream);
    else
        substream = rcu_dereference(dev->rx_substream);
    active = substream && snd_pcm_running(substream);
    if (active) {
        unsigned int ptr;
        unsigned int new_ptr;

        if (push) {
            ptr = READ_ONCE(dev->tx_ptr);
            new_ptr = dev->tx_fn(dev, substream->runtime, ptr, &period_elapsed);
            cmpxchg(&dev->tx_ptr, ptr, new_ptr);
        }
        else {
            ptr = READ_ONCE(dev->rx_ptr);
            new_ptr = dev->rx_fn(dev, substream->runtime, ptr, &period_elapsed);
            cmpxchg(&dev->rx_ptr, ptr, new_ptr);
        }

        if (period_elapsed) {
            snd_pcm_period_elapsed(substream);
        }
    }
    rcu_read_unlock();
}

void ljm_pcm_push_tx(struct ljm_i2s_dev *dev)
{
    ljm_pcm_transfer(dev, true);
}

void ljm_pcm_pop_rx(struct ljm_i2s_dev *dev)
{
    ljm_pcm_transfer(dev, false);
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 10, 0)
static int ljm_pcm_open(struct snd_soc_component *component, struct snd_pcm_substream *substream)
#else
static int ljm_pcm_open(struct snd_pcm_substream *substream)
#endif
{
    struct snd_pcm_runtime *runtime = substream->runtime;
    struct ljm_device *ljm_dev = get_ljm_dev();

    snd_soc_set_runtime_hwparams(substream, &ljm_pcm_hardware);
    snd_pcm_hw_constraint_integer(runtime, SNDRV_PCM_HW_PARAM_PERIODS);
    runtime->private_data = ljm_dev;

    hrtimer_start(&ljm_dev->hrt, ns_to_ktime(ljm_dev->i2s_devs[0]->poll_time_ns), HRTIMER_MODE_REL);

    return 0;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 10, 0)
static int ljm_pcm_close(struct snd_soc_component *component, struct snd_pcm_substream *substream)
#else
static int ljm_pcm_close(struct snd_pcm_substream *substream)
#endif
{
    struct ljm_device *ljm_dev = substream->runtime->private_data;
    hrtimer_cancel(&ljm_dev->hrt);
    synchronize_rcu();
    return 0;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 10, 0)
static int ljm_pcm_hw_params(struct snd_soc_component *component,
        struct snd_pcm_substream *substream,
        struct snd_pcm_hw_params *hw_params)
#else
static int ljm_pcm_hw_params(struct snd_pcm_substream *substream,
        struct snd_pcm_hw_params *hw_params)
#endif
{
    struct snd_pcm_runtime *runtime = substream->runtime;
    struct ljm_device *ljm_dev = runtime->private_data;
    struct ljm_i2s_dev *dev = NULL;
    int i;
    int ret;

    for (i = 0; i < LJM_MAX_I2SS; i++) {
        dev = ljm_dev->i2s_devs[i];
        switch (params_channels(hw_params)) {
            case 2:
                break;
            default:
                dev_err(dev->dev, "LJM PCM: invalid channels number\n");
                return -EINVAL;
        }

        switch (params_format(hw_params)) {
            case SNDRV_PCM_FORMAT_S16_LE:
                dev->tx_fn = ljm_pcm_tx_16;
                dev->rx_fn = ljm_pcm_rx_16;
                break;
            case SNDRV_PCM_FORMAT_S24_LE:
            case SNDRV_PCM_FORMAT_S32_LE:
                dev->tx_fn = ljm_pcm_tx_32;
                dev->rx_fn = ljm_pcm_rx_32;
                break;
            default:
                dev_err(dev->dev, "LJM PCM: invalid format\n");
                return -EINVAL;
        }
    }

    ret = snd_pcm_lib_malloc_pages(substream, params_buffer_bytes(hw_params));
    if (ret < 0)
        return ret;
    else
        return 0;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 10, 0)
static int ljm_pcm_hw_free(struct snd_pcm_substream *substream)
{
    return snd_pcm_lib_free_pages(substream);
}
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 10, 0)
static int ljm_pcm_trigger(struct snd_soc_component *component, struct snd_pcm_substream *substream, int cmd)
#else
static int ljm_pcm_trigger(struct snd_pcm_substream *substream, int cmd)
#endif
{
    struct snd_pcm_runtime *runtime = substream->runtime;
    struct ljm_device *ljm_dev = runtime->private_data;
    struct ljm_i2s_dev *dev = NULL;
    int i;
    int ret = 0;

    for (i = 0; i < LJM_MAX_I2SS; i++) {
        dev = ljm_dev->i2s_devs[i];
        switch (cmd) {
            case SNDRV_PCM_TRIGGER_START:
            case SNDRV_PCM_TRIGGER_RESUME:
            case SNDRV_PCM_TRIGGER_PAUSE_RELEASE:
                if (substream->stream == SNDRV_PCM_STREAM_PLAYBACK) {
                    WRITE_ONCE(dev->tx_ptr, 0);
                    rcu_assign_pointer(dev->tx_substream, substream);
                }
                else {
                    WRITE_ONCE(dev->rx_ptr, 0);
                    rcu_assign_pointer(dev->rx_substream, substream);
                }
                break;
            case SNDRV_PCM_TRIGGER_STOP:
            case SNDRV_PCM_TRIGGER_SUSPEND:
            case SNDRV_PCM_TRIGGER_PAUSE_PUSH:
                if (substream->stream == SNDRV_PCM_STREAM_PLAYBACK) {
                    rcu_assign_pointer(dev->tx_substream, NULL);
                }
                else
                    rcu_assign_pointer(dev->rx_substream, NULL);
                break;
            default:
                ret = -EINVAL;
                break;
        }
    }

    return ret;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 10, 0)
static snd_pcm_uframes_t ljm_pcm_pointer(struct snd_soc_component *component, struct snd_pcm_substream *substream)
#else
static snd_pcm_uframes_t ljm_pcm_pointer(struct snd_pcm_substream *substream)
#endif
{
    struct snd_pcm_runtime *runtime = substream->runtime;
    struct ljm_device *ljm_dev = runtime->private_data;
    struct ljm_i2s_dev *dev = ljm_dev->i2s_devs[0];
    snd_pcm_uframes_t pos;

    if (substream->stream == SNDRV_PCM_STREAM_PLAYBACK)
        pos = READ_ONCE(dev->tx_ptr);
    else
        pos = READ_ONCE(dev->rx_ptr);

    return pos < runtime->buffer_size ? pos : 0;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 10, 0)
static int ljm_pcm_new(struct snd_soc_component *component, struct snd_soc_pcm_runtime *rtd)
#else
static int ljm_pcm_new(struct snd_soc_pcm_runtime *rtd)
#endif
{
    size_t size = ljm_pcm_hardware.buffer_bytes_max;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(6, 6, 0)
	snd_pcm_set_managed_buffer_all(rtd->pcm,
			SNDRV_DMA_TYPE_CONTINUOUS,
			NULL, size, size);
#else
    snd_pcm_lib_preallocate_pages_for_all(rtd->pcm,
        SNDRV_DMA_TYPE_CONTINUOUS,
        snd_dma_continuous_data(GFP_KERNEL), size, size);
#endif

    return 0;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(5, 10, 0)
static void ljm_pcm_free(struct snd_pcm *pcm)
{
    snd_pcm_lib_preallocate_free_for_all(pcm);
}
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 10, 0)
static const struct snd_soc_component_driver ljm_pcm_component = {
    .open = ljm_pcm_open,
    .close = ljm_pcm_close,
    .hw_params = ljm_pcm_hw_params,
    .trigger = ljm_pcm_trigger,
    .pointer = ljm_pcm_pointer,
    .pcm_construct = ljm_pcm_new,
};
#else
static const struct snd_pcm_ops ljm_pcm_ops = {
    .open = ljm_pcm_open,
    .close = ljm_pcm_close,
    .ioctl = snd_pcm_lib_ioctl,
    .hw_params = ljm_pcm_hw_params,
    .hw_free = ljm_pcm_hw_free,
    .trigger = ljm_pcm_trigger,
    .pointer = ljm_pcm_pointer,
};

static const struct snd_soc_component_driver ljm_pcm_component = {
    .pcm_new = ljm_pcm_new,
    .pcm_free = ljm_pcm_free,
    .ops = &ljm_pcm_ops,
};
#endif

static enum hrtimer_restart hrtimer_callback(struct hrtimer *hrt)
{
    int i;
    struct ljm_device *ljm_dev = container_of(hrt, struct ljm_device, hrt);
    struct ljm_i2s_dev *dev = NULL;
    bool irq_valid = false;
    u32 isr[4];
    int j;

    for (i = 0; i < LJM_MAX_I2SS; i++) {
        dev = ljm_dev->i2s_devs[i];
        for (j = 0; j < 4; j++) {
            isr[j] = i2s_read_reg(dev->i2s_base, ISR(j));
        }

        for (j = 0; j < 4; j++) {
            /*
             * Check if TX fifo is empty. If empty fill FIFO with samples
             * NOTE: Only two channels supported
             */
            if ((isr[j] & ISR_TXFE) && (j == 0) && dev->use_pio) {
                ljm_pcm_push_tx(dev);
                irq_valid = true;
            }

            /*
             * Data available. Retrieve samples from FIFO
             * NOTE: Only two channels supported
             */

            /* Error Handling: TX */
            if (isr[j] & ISR_TXFO) {
                dev_err(dev->dev, "LJM IIS: TX overrun (ch_id=%d).\n", j);
                irq_valid = true;
            }
        }
    }

    hrtimer_forward_now(hrt, ns_to_ktime(dev->poll_time_ns));
    return HRTIMER_RESTART;
}

int ljm_pcm_register(struct device *dev)
{
    int ret = 0;
    struct ljm_device *ljm_dev = get_ljm_dev();

    ret = devm_snd_soc_register_component(dev, &ljm_pcm_component, NULL, 0);

    if (!ret) {
        hrtimer_init(&ljm_dev->hrt, CLOCK_MONOTONIC, HRTIMER_MODE_REL);
        ljm_dev->hrt.function = hrtimer_callback;
    }
    else {
        dev_err(ljm_dev->dev, "LJM PCM: could not register pcm: %d.\n", ret);
    }

    return ret;
}

#endif
