/****************************************************************************
*
*    Copyright (c) 2016 - 2099 by LJW.  All rights reserved.
*
*    The material in this file is confidential and contains trade secrets
*    of LJW. This is proprietary information owned by LJW. No
*    part of this work may be disclosed, reproduced, copied, transmitted,
*    or used in any way for any purpose, without the express written
*    permission of LJW.
*
*****************************************************************************/

#include "lj_hal_kernel_linux.h"
#include "lj_hal_dump.h"

#include <linux/pagemap.h>
#include <linux/seq_file.h>
#include <linux/mman.h>
#include <linux/atomic.h>
#include <linux/dma-mapping.h>
#include <linux/slab.h>
#include <linux/workqueue.h>
#include <linux/irqflags.h>
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 23)
# include <linux/math64.h>
#endif
#include <linux/delay.h>
#include <linux/platform_device.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 27)
# include <linux/anon_inodes.h>
#endif

#include <linux/cpufreq.h>

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 5, 0)
# include <linux/io.h>
#endif

#if ljdLINUX_SYNC_FILE
# include <linux/file.h>
# include "lj_hal_kernel_sync.h"
#endif

#if defined(CONFIG_DMA_SHARED_BUFFER)
# include <linux/dma-buf.h>
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 7, 0)
#if defined(CONFIG_TRACE_GPU_MEM)
# include <trace/events/gpu_mem.h>
typedef struct trace_mem {
    ljtUINT32 pid;
    atomic64_t total;
    struct trace_mem *next;
} traceMem;

traceMem *memTraceList;
ljtBOOL memTraceFlag = 1;
#endif
#endif

#define _LJ_OBJ_ZONE        ljvZONE_OS
#define ljdUSING_PFN_FOLLOW    0

#include "lj_hal_kernel_allocator.h"

#define ljmkBUG_ON(x)                                                    \
    do {                                                                 \
        if (unlikely(!!(x))) {                                           \
            pr_warn("[ljmcore]: BUG ON @ %s(%d)\n", __func__, __LINE__); \
            dump_stack();                                                \
        }                                                                \
    } while (0)

/******************************************************************************
 ******************************* Private Functions ****************************
 *****************************************************************************/
static ljtINT
_GetThreadID(void)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 24)
    return task_pid_nr(current);
#else
    return current->pid;
#endif
}

/* Must hold Mdl->mpasMutex before call this function. */
static inline PLINUX_MDL_MAP
_CreateMdlMap(PLINUX_MDL Mdl, ljtINT ProcessID)
{
    PLINUX_MDL_MAP mdlMap = ljvNULL;
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkHEADER_ARG("Mdl=%p ProcessID=%d", Mdl, ProcessID);

    mdlMap = (PLINUX_MDL_MAP)kmalloc(sizeof(*mdlMap), GFP_KERNEL | ljdNOWARN);
    if (mdlMap == ljvNULL)
        ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);

    mdlMap->pid = ProcessID;
    mdlMap->vmaAddr = ljvNULL;
    mdlMap->count = 0;

    list_add(&mdlMap->link, &Mdl->mapsHead);

OnError:
    ljmkFOOTER_ARG("ret=%p", mdlMap);
    return mdlMap;
}

/* Must hold Mdl->mpasMutex before call this function. */
static inline ljeSTATUS
_DestroyMdlMap(PLINUX_MDL Mdl, PLINUX_MDL_MAP MdlMap)
{
    ljmkHEADER_ARG("Mdl=%p MdlMap=%p", Mdl, MdlMap);

    /* Verify the arguments. */
    ljmkVERIFY_ARGUMENT(MdlMap != ljvNULL);

    list_del(&MdlMap->link);
    kfree(MdlMap);

    ljmkFOOTER_NO();
    return ljvSTATUS_OK;
}

/* Must hold Mdl->mapsMutex before call this function. */
extern PLINUX_MDL_MAP
FindMdlMap(PLINUX_MDL Mdl, ljtINT ProcessID)
{
    PLINUX_MDL_MAP mdlMap = ljvNULL;

    ljmkHEADER_ARG("Mdl=%p ProcessID=%d", Mdl, ProcessID);

    if (Mdl) {
        PLINUX_MDL_MAP iter = ljvNULL;

        list_for_each_entry(iter, &Mdl->mapsHead, link) {
            if (iter->pid == ProcessID) {
                mdlMap = iter;
                break;
            }
        }
    }

    ljmkFOOTER_ARG("ret=%p", mdlMap);
    return mdlMap;
}

static PLINUX_MDL
_CreateMdl(ljkOS Os, ljkKERNEL Kernel)
{
    PLINUX_MDL mdl;

    ljmkHEADER();

    mdl = (PLINUX_MDL)kzalloc(sizeof(*mdl), GFP_KERNEL | ljdNOWARN);

    if (mdl) {
        mdl->os = Os;
        atomic_set(&mdl->refs, 1);
        mutex_init(&mdl->mapsMutex);
        INIT_LIST_HEAD(&mdl->mapsHead);
        INIT_LIST_HEAD(&mdl->rmaHead);

        if (Kernel && Kernel->device) {
            mdl->device = Kernel->device->dev;
        } else {
            /* Get the default device. */
            mdl->device = Os->device->devices[0]->dev;
        }

        mdl->pages = ljvNULL;
    }

    ljmkFOOTER_ARG("%p", mdl);
    return mdl;
}

static ljeSTATUS
_DestroyMdl(PLINUX_MDL Mdl)
{
    ljeSTATUS status = ljvSTATUS_OK;
    ljtBOOL acquiredMapsMutex = ljvFALSE;
    ljtBOOL acquiredMdlMutex = ljvFALSE;

    ljmkHEADER_ARG("Mdl=%p", Mdl);

    /* Verify the arguments. */
    ljmkVERIFY_ARGUMENT(Mdl != ljvNULL);

    if (atomic_dec_and_test(&Mdl->refs)) {
        ljkOS os = Mdl->os;
        ljkALLOCATOR allocator = Mdl->allocator;
        PLINUX_MDL_MAP mdlMap, next;

        /* Valid private means alloc/attach successfully */
        if (Mdl->priv) {
            if (Mdl->addr) {
                allocator->ops->UnmapKernel(allocator, Mdl, Mdl->addr);
                Mdl->addr = ljvNULL;
            }
            allocator->ops->Free(allocator, Mdl);
        }

        ljmkONERROR(ljkOS_AcquireMutex(os, &Mdl->mapsMutex, ljvINFINITE));
        acquiredMapsMutex = ljvTRUE;

        list_for_each_entry_safe(mdlMap, next, &Mdl->mapsHead, link) {
            ljmkVERIFY_OK(_DestroyMdlMap(Mdl, mdlMap));
        }

        ljmkONERROR(ljkOS_ReleaseMutex(os, &Mdl->mapsMutex));
        acquiredMapsMutex = ljvFALSE;

        if (Mdl->link.next) {
            /* Remove the node from global list.. */
            ljmkONERROR(ljkOS_AcquireMutex(os, &os->mdlMutex, ljvINFINITE));
            acquiredMdlMutex = ljvTRUE;

            list_del(&Mdl->link);

            ljmkONERROR(ljkOS_ReleaseMutex(os, &os->mdlMutex));
            acquiredMdlMutex = ljvFALSE;
        } else if (Mdl->rmaLink.next) {
            /* Remove the sub node from root mdl */
            ljmkONERROR(ljkOS_AcquireMutex(os, &os->mdlMutex, ljvINFINITE));
            acquiredMdlMutex = ljvTRUE;

            list_del(&Mdl->rmaLink);

            ljmkONERROR(ljkOS_ReleaseMutex(os, &os->mdlMutex));
            acquiredMdlMutex = ljvFALSE;
        }

        Mdl->device = ljvNULL;

        ljkOS_ZeroMemory(&Mdl->mapsMutex, sizeof(struct mutex));
        kfree(Mdl);
    }

    ljmkFOOTER_NO();
    return ljvSTATUS_OK;

OnError:
    if (acquiredMapsMutex)
        ljmkVERIFY_OK(ljkOS_ReleaseMutex(Mdl->os, &Mdl->mapsMutex));

    if (acquiredMdlMutex)
        ljmkVERIFY_OK(ljkOS_ReleaseMutex(Mdl->os, &Mdl->os->mdlMutex));

    ljmkFOOTER();
    return status;

}

/*******************************************************************************
 ** Integer Id Management.
 */
ljeSTATUS
_AllocateIntegerId(ljsINTEGER_DB_PTR Database, ljtPOINTER KernelPointer, ljtUINT32 *Id)
{
    int result;
    ljtINT next;
    unsigned long flags = 0;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 9, 0)
    idr_preload(GFP_KERNEL | ljdNOWARN);

    if (in_irq())
        spin_lock(&Database->lock);
    else
        spin_lock_irqsave(&Database->lock, flags);

    next = (Database->curr + 1 <= 0) ? 1 : Database->curr + 1;

    result = idr_alloc(&Database->idr, KernelPointer, next, 0, GFP_ATOMIC);

    /* ID allocated should not be 0. */
    ljmkASSERT(result != 0);

    if (result > 0)
        Database->curr = *Id = result;

    if (in_irq())
        spin_unlock(&Database->lock);
    else
        spin_unlock_irqrestore(&Database->lock, flags);

    idr_preload_end();

    if (result < 0)
        return ljvSTATUS_OUT_OF_RESOURCES;
#else  /* !(LINUX_VERSION_CODE >= KERNEL_VERSION(3, 9, 0)) */
again:
    if (idr_pre_get(&Database->idr, GFP_KERNEL | ljdNOWARN) == 0)
        return ljvSTATUS_OUT_OF_MEMORY;

    if (in_irq())
        spin_lock(&Database->lock);
    else
        spin_lock_irqsave(&Database->lock, flags);

    next = (Database->curr + 1 <= 0) ? 1 : Database->curr + 1;

    /* Try to get a id greater than 0. */
    result = idr_get_new_above(&Database->idr, KernelPointer, next, Id);

    if (!result)
        Database->curr = *Id;

    if (in_irq())
        spin_unlock(&Database->lock);
    else
        spin_unlock_irqrestore(&Database->lock, flags);

    if (result == -EAGAIN)
        goto again;

    if (result != 0)
        return ljvSTATUS_OUT_OF_RESOURCES;
#endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(3, 9, 0) */

    return ljvSTATUS_OK;
}

ljeSTATUS
_QueryIntegerId(ljsINTEGER_DB_PTR Database, ljtUINT32 Id, ljtPOINTER *KernelPointer)
{
    ljtPOINTER pointer;
    unsigned long flags = 0;

    if (in_irq())
        spin_lock(&Database->lock);
    else
        spin_lock_irqsave(&Database->lock, flags);

    pointer = idr_find(&Database->idr, Id);

    if (in_irq())
        spin_unlock(&Database->lock);
    else
        spin_unlock_irqrestore(&Database->lock, flags);

    if (pointer) {
        *KernelPointer = pointer;
        return ljvSTATUS_OK;
    }

    ljmkTRACE_ZONE(ljvLEVEL_ERROR, ljvZONE_OS,
                   "%s(%d) Id = %d is not found",
                   __func__, __LINE__, Id);

    return ljvSTATUS_NOT_FOUND;
}

ljeSTATUS
_DestroyIntegerId(ljsINTEGER_DB_PTR Database, ljtUINT32 Id)
{
    unsigned long flags = 0;

    if (in_irq())
        spin_lock(&Database->lock);
    else
        spin_lock_irqsave(&Database->lock, flags);

    idr_remove(&Database->idr, Id);

    if (in_irq())
        spin_unlock(&Database->lock);
    else
        spin_unlock_irqrestore(&Database->lock, flags);

    return ljvSTATUS_OK;
}

static inline ljeSTATUS
_QueryProcessPageTable(ljtPOINTER Logical, ljtPHYS_ADDR_T *Address)
{
    unsigned long logical = (unsigned long)Logical;
    unsigned long offset = logical & ~PAGE_MASK;

    if (is_vmalloc_addr(Logical)) {
        /* vmalloc area. */
        *Address = page_to_phys(vmalloc_to_page(Logical)) | offset;
        return ljvSTATUS_OK;
    } else if (virt_addr_valid((ljtPOINTER)logical)) {
        /* Kernel logical address. */
        *Address = virt_to_phys(Logical);
        return ljvSTATUS_OK;
#if ljdUSING_PFN_FOLLOW
    } else {
        struct vm_area_struct *vma;
        unsigned long pfn = 0;
        int ret = 0;

        down_read(&current_mm_mmap_sem);
        vma = find_vma(current->mm, logical);
        if (!vma) {
            up_read(&current_mm_mmap_sem);
            return ljvSTATUS_NOT_FOUND;
        }
        up_read(&current_mm_mmap_sem);
        ret = follow_pfn(vma, logical, &pfn);
        if (ret < 0) {
            return ljvSTATUS_NOT_FOUND;
        } else {
            *Address = (pfn << PAGE_SHIFT) | offset;
            return ljvSTATUS_OK;
        }
    }
#else
    } else {
        /* Try user VM area. */
        struct vm_area_struct *vma;
        spinlock_t *ptl;
        pgd_t *pgd;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 9, 0)
        p4d_t *p4d;
# endif
        pud_t *pud;
        pmd_t *pmd;
        pte_t *pte;

        if (!current->mm)
            return ljvSTATUS_NOT_FOUND;

        down_read(&current_mm_mmap_sem);
        vma = find_vma(current->mm, logical);
        up_read(&current_mm_mmap_sem);

        /* To check if mapped to user. */
        if (!vma)
            return ljvSTATUS_NOT_FOUND;

        pgd = pgd_offset(current->mm, logical);
        if (pgd_none(*pgd) || pgd_bad(*pgd))
            return ljvSTATUS_NOT_FOUND;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 9, 0)
        p4d = p4d_offset(pgd, logical);
        if (p4d_none(READ_ONCE(*p4d)))
            return ljvSTATUS_NOT_FOUND;

        pud = pud_offset(p4d, logical);
# elif (defined(CONFIG_X86)) && LINUX_VERSION_CODE >= KERNEL_VERSION (4, 12, 0)
        pud = pud_offset((p4d_t *)pgd, logical);
# elif (defined(CONFIG_CPU_CSKYV2)) && LINUX_VERSION_CODE >= KERNEL_VERSION (4, 11, 0)
        pud = pud_offset((p4d_t *)pgd, logical);
# else
        pud = pud_offset(pgd, logical);
# endif
        if (pud_none(*pud) || pud_bad(*pud))
            return ljvSTATUS_NOT_FOUND;

        pmd = pmd_offset(pud, logical);
        if (pmd_none(*pmd) || pmd_bad(*pmd))
            return ljvSTATUS_NOT_FOUND;

#if LINUX_VERSION_CODE < KERNEL_VERSION(6, 6, 0)
        pte = pte_offset_map_lock(current->mm, pmd, logical, &ptl);
#else
        pte = pte_offset_kernel(pmd, (unsigned long)logical);
#endif
        if (!pte_present(*pte)) {
            pte_unmap_unlock(pte, ptl);
            return ljvSTATUS_NOT_FOUND;
        }

        *Address = (pte_pfn(*pte) << PAGE_SHIFT) | offset;
        pte_unmap_unlock(pte, ptl);

        return ljvSTATUS_OK;
    }
#endif
}

static ljeSTATUS
_ShrinkMemory(ljkOS Os)
{
    ljsPLATFORM *platform;
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkHEADER_ARG("Os=%p", Os);
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);

    platform = Os->device->platform;

    if (platform && platform->ops->shrinkMemory)
        status = platform->ops->shrinkMemory(platform);
    else
        status = ljvSTATUS_NOT_SUPPORTED;

    ljmkFOOTER();
    return status;
}

#if ljdDUMP_IN_KERNEL

# define DUMP_TO_KERNEL_DMESG 0
# define DUMP_TO_FILE         1
# define DUMP_IGNORE          2

static void set_dump_file(ljkOS os, char *fname)
{
    if (os->dumpFilp) {
        /* close exist opened file. */
        pr_warn("ljmcore: end dump to file: %s\n", os->dumpFileName);

        filp_close(os->dumpFilp, NULL);
        os->dumpFilp = NULL;
    }

    if (fname[0] == '\0' || !strcmp(fname, "[ignored]")) {
        /* empty or [ignored] means ignored. */
        pr_warn("ljmcore: dump ignored\n");

        os->dumpTarget = DUMP_IGNORE;
        strcpy(os->dumpFileName, "[ignored]");
    } else if (!strcmp(fname, "[dmesg]")) {
        /* [dmesg] means dump to kernel dmesg. */
        pr_warn("ljmcore: dump to kernel dmesg\n");

        os->dumpTarget = DUMP_TO_KERNEL_DMESG;
        strcpy(os->dumpFileName, "[dmesg]");
    } else if (fname[0] != '/') {
        /* invalid path, switch to kernel dmesg. */
        pr_err("ljmcore: invalid path: %s\n", fname);
        pr_err("ljmcore: must be absolute path start with '/'\n");
        pr_warn("ljmcore: dump to kernel dmesg\n");

        os->dumpTarget = DUMP_TO_KERNEL_DMESG;
        strcpy(os->dumpFileName, "[dmesg]");
    } else {
        /* try open file. */
        os->dumpFilp = filp_open(fname, O_RDWR | O_CREAT, 0644);

        if (IS_ERR(os->dumpFilp)) {
            pr_err("ljmcore: failed to open file: %s\n", fname);
            pr_warn("ljmcore: dump to kernel dmesg\n");

            os->dumpFilp   = NULL;
            os->dumpTarget = DUMP_TO_KERNEL_DMESG;
            strcpy(os->dumpFileName, "[dmesg]");
        } else {
            pr_warn("ljmcore: start dump to file: %s\n", fname);

            os->dumpTarget = DUMP_TO_FILE;
            strcpy(os->dumpFileName, fname);
        }
    }
}

static int dump_file_show(struct seq_file *m, void *unused)
{
    ljsINFO_NODE *node = m->private;
    ljkOS os = node->device;

    seq_printf(m, "%s\n", os->dumpFileName);
    return 0;
}

static int dump_file_write(const char __user *buf, size_t count, void *data)
{
    ljsINFO_NODE *node = data;
    ljkOS os = node->device;
    char fname[256];
    size_t len = min(count, sizeof(fname) - 1);
    ljeSTATUS status = ljvSTATUS_OK;
    ljtBOOL acquiredMutex = ljvFALSE;

    if (copy_from_user(fname, buf, len))
        return -EFAULT;

    /* Remove tailing space. */
    while (len > 0 && (fname[len - 1] == '\n' || fname[len - 1] == ' '))
        fname[len - 1] = '\0';

    fname[len] = '\0';

    ljmkONERROR(ljkOS_AcquireMutex(os, &os->dumpFilpMutex, ljvINFINITE));
    acquiredMutex = ljvTRUE;

    set_dump_file(os, fname);

    ljmkONERROR(ljkOS_ReleaseMutex(os, &os->dumpFilpMutex));
    acquiredMutex = ljvFALSE;

    return count;

OnError:
    if (acquiredMutex)
        ljkOS_ReleaseMutex(os, &os->dumpFilpMutex);

    if (status == ljvSTATUS_INTERRUPTED)
        return -EINTR;

    return count;
}

static ljsINFO dumpDebugList[] = {
    { "dump_file", dump_file_show, dump_file_write },
};

static ljeSTATUS
_DumpDebugfsInit(ljkOS Os)
{
    ljeSTATUS status;
    ljkGALDEVICE device = Os->device;
    ljkDEBUGFS_DIR dir = &Os->dumpDebugfsDir;

    ljmkONERROR(ljkDEBUGFS_DIR_Init(dir, device->debugfsDir.root, "dump"));

    ljmkONERROR(ljkDEBUGFS_DIR_CreateFiles(dir, dumpDebugList,
                                           ljmCOUNTOF(dumpDebugList), Os));

OnError:
    return status;
}

static void
_DumpDebugfsCleanup(ljkOS Os)
{
    ljkDEBUGFS_DIR dir = &Os->dumpDebugfsDir;

    if (dir->root) {
        ljkDEBUGFS_DIR_RemoveFiles(dir, dumpDebugList,
                                   ljmCOUNTOF(dumpDebugList));
        ljkDEBUGFS_DIR_Deinit(dir);
    }
}
#endif /* ljdDUMP_IN_KERNEL */

/******************************************************************************
 **
 **  ljkOS_Construct
 **
 **  Construct a new ljkOS object.
 **
 **  INPUT:
 **
 **      ljtPOINTER Context
 **          Pointer to the ljkGALDEVICE class.
 **
 **  OUTPUT:
 **
 **      ljkOS *Os
 **          Pointer to a variable that will hold the pointer to the ljkOS
 **          object.
 */
ljeSTATUS
ljkOS_Construct(ljtPOINTER Context, ljkOS *Os)
{
    ljkOS os = ljvNULL;
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkHEADER_ARG("Context=%p", Context);

    /* Verify the arguments. */
    ljmkVERIFY_ARGUMENT(Os != ljvNULL);

    /* Allocate the ljkOS object. */
    os = (ljkOS)kmalloc(ljmSIZEOF(struct _ljkOS), GFP_KERNEL | ljdNOWARN);

    if (os == ljvNULL)
        ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);

    /* Zero the memory. */
    ljkOS_ZeroMemory(os, ljmSIZEOF(struct _ljkOS));

    /* Initialize the ljkOS object. */
    os->object.type = ljvOBJ_OS;

    /* Set device. */
    os->device = Context;

    /* Set allocateCount to 0, ljkOS_Allocate has not been used yet. */
    atomic_set(&os->allocateCount, 0);

    atomic_set(&os->nodeID, 0);

    /* Initialize the memory lock. */
    mutex_init(&os->mdlMutex);

    INIT_LIST_HEAD(&os->mdlHead);

    /* Get the kernel process ID. */
    os->kernelProcessID = _GetProcessID();

    /*
     * Initialize the signal manager.
     */

    /* Initialize spinlock. */
    spin_lock_init(&os->signalLock);

    /* Initialize signal id database lock. */
    spin_lock_init(&os->signalDB.lock);

    /* Initialize signal id database. */
    idr_init(&os->signalDB.idr);

    /* Create a workqueue for os timer. */
    os->workqueue = create_singlethread_workqueue("ljmcore workqueue");

    if (os->workqueue == ljvNULL)
        /* Out of memory. */
        ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);

#if defined(CONFIG_ZONE_DMA32) && LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 37)
    os->paddingPage = alloc_page(GFP_KERNEL | __GFP_DMA32 | ljdNOWARN);
#else
    os->paddingPage = alloc_page(GFP_KERNEL | __GFP_DMA | ljdNOWARN);
#endif

    if (os->paddingPage == ljvNULL)
        /* Out of memory. */
        ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);
    else
        SetPageReserved(os->paddingPage);

    spin_lock_init(&os->registerAccessLock);

    /* Check iommu. */
    if (ljmIS_ERROR(ljkIOMMU_Construct(os, &os->iommu))) {
        ljmkTRACE_ZONE(ljvLEVEL_INFO, ljvZONE_OS,
                       "%s(%d): Fail to setup IOMMU",
                       __func__, __LINE__);
    }

    ljkOS_ImportAllocators(os);

#if ljdDUMP_IN_KERNEL
    mutex_init(&os->dumpFilpMutex);

    /* Set default dump file. */
    set_dump_file(os, ljdDUMP_FILE_IN_KERNEL);

    /* Init debugfs for kernel dump feature. */
    _DumpDebugfsInit(os);
#endif /* ljdDUMP_IN_KERNEL */

    /* Return pointer to the ljkOS object. */
    *Os = os;

OnError:
    if (ljmIS_ERROR(status) && os) {
        if (os->workqueue != ljvNULL)
            destroy_workqueue(os->workqueue);

        kfree(os);
        os = ljvNULL;
    }

    /* Return the error. */
    ljmkFOOTER_ARG("*Os=%p", os);
    return status;
}

/******************************************************************************
 **
 **  ljkOS_Destroy
 **
 **  Destroy an ljkOS object.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object that needs to be destroyed.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_Destroy(ljkOS Os)
{
#if ljdDUMP_IN_KERNEL
    ljeSTATUS status = ljvSTATUS_OK;
    ljtBOOL acquiredMutex = ljvFALSE;
#endif

    ljmkHEADER_ARG("Os=%p", Os);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);

    if (Os->paddingPage != ljvNULL) {
        ClearPageReserved(Os->paddingPage);
        __free_page(Os->paddingPage);
        Os->paddingPage = ljvNULL;
    }

    /*
     * Destroy the signal manager.
     */

    /* Wait for all works done. */
    flush_workqueue(Os->workqueue);

    /* Destroy work queue. */
    destroy_workqueue(Os->workqueue);

    ljkOS_FreeAllocators(Os);

    if (Os->iommu)
        ljkIOMMU_Destory(Os, Os->iommu);

    /* Mark the ljkOS object as unknown. */
    Os->object.type = ljvOBJ_UNKNOWN;


#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 7, 0)
#if defined(CONFIG_TRACE_GPU_MEM)
    while (ljvTRUE) {
        if (memTraceList) {
            traceMem *p_node = memTraceList;

            memTraceList = memTraceList->next;
            ljkOS_Free(Os, p_node);
            p_node = NULL;
        } else {
            break;
        }
    }
#endif
#endif

#if ljdDUMP_IN_KERNEL
    ljmkONERROR(ljkOS_AcquireMutex(Os, &Os->dumpFilpMutex, ljvINFINITE));
    acquiredMutex = ljvTRUE;

    if (Os->dumpFilp) {
        filp_close(Os->dumpFilp, NULL);
        Os->dumpFilp   = NULL;
        Os->dumpTarget = DUMP_IGNORE;
    }

    ljmkONERROR(ljkOS_ReleaseMutex(Os, &Os->dumpFilpMutex));
    acquiredMutex = ljvFALSE;

    /* Cleanup debugfs for kernel dump feature. */
    _DumpDebugfsCleanup(Os);
    ljkOS_ZeroMemory(&Os->dumpFilpMutex, sizeof(struct mutex));
#endif /* ljdDUMP_IN_KERNEL */

    ljkOS_ZeroMemory(&Os->mdlMutex, sizeof(struct mutex));
    /* Free the ljkOS object. */
    kfree(Os);

    /* Success. */
    ljmkFOOTER_NO();
    return ljvSTATUS_OK;
#if ljdDUMP_IN_KERNEL
OnError:
    if (acquiredMutex)
        ljkOS_ReleaseMutex(Os, &Os->dumpFilpMutex);

    ljmkFOOTER();
    return status;
#endif
}

ljeSTATUS
ljkOS_CreateKernelMapping(ljkOS Os, ljtPHYS_ADDR Physical, ljtSIZE_T Offset,
                          ljtSIZE_T Bytes, ljtPOINTER *Logical)
{
    ljeSTATUS status = ljvSTATUS_OK;
    PLINUX_MDL mdl = (PLINUX_MDL)Physical;
    ljkALLOCATOR allocator = mdl->allocator;

    ljmkHEADER_ARG("Os=%p Physical=%p Offset=0x%zx Bytes=0x%zx",
                   Os, Physical, Offset, Bytes);

    if (mdl->addr) {
        /* Already mapped whole memory. */
        *Logical = (ljtUINT8_PTR)mdl->addr + Offset;
    } else {
        ljmkONERROR(allocator->ops->MapKernel(allocator, mdl, Offset, Bytes, Logical));
        if (Offset == 0 && Bytes == mdl->bytes)
            /* the whole mdl has mapped */
            mdl->addr = *Logical;
    }

OnError:
    ljmkFOOTER_ARG("*Logical=%p", ljmOPT_POINTER(Logical));
    return status;
}

ljeSTATUS
ljkOS_DestroyKernelMapping(ljkOS Os, ljtPHYS_ADDR Physical, ljtPOINTER Logical)
{
    PLINUX_MDL mdl = (PLINUX_MDL)Physical;
    ljkALLOCATOR allocator = mdl->allocator;

    ljmkHEADER_ARG("Os=%p Physical=%p Logical=%p", Os, Physical, Logical);

    if (mdl->addr)
        /* Nothing to do.
         * it will be unmpped in vidmem free
         */
        ;
    else
        allocator->ops->UnmapKernel(allocator, mdl, Logical);

    ljmkFOOTER_NO();
    return ljvSTATUS_OK;
}

/******************************************************************************
 **
 **  ljkOS_Allocate
 **
 **  Allocate memory.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtSIZE_T Bytes
 **          Number of bytes to allocate.
 **
 **  OUTPUT:
 **
 **      ljtPOINTER *Memory
 **          Pointer to a variable that will hold the allocated memory location.
 */
ljeSTATUS
ljkOS_Allocate(ljkOS Os, ljtSIZE_T Bytes, ljtPOINTER *Memory)
{
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkHEADER_ARG("Os=%p Bytes=0x%zx", Os, Bytes);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Bytes > 0);
    ljmkVERIFY_ARGUMENT(Memory != ljvNULL);

    ljmkONERROR(ljkOS_AllocateMemory(Os, Bytes, Memory));

OnError:
    /* Return the status. */
    ljmkFOOTER_ARG("*Memory=%p", ljmOPT_POINTER(Memory));
    return status;
}

/******************************************************************************
 **
 **  ljkOS_Free
 **
 **  Free allocated memory.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtPOINTER Memory
 **          Pointer to memory allocation to free.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_Free(ljkOS Os, ljtPOINTER Memory)
{
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkHEADER_ARG("Os=%p Memory=%p", Os, Memory);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Memory != ljvNULL);

    ljmkONERROR(ljkOS_FreeMemory(Os, Memory));

OnError:
    /* Return the status. */
    ljmkFOOTER();
    return status;
}

/******************************************************************************
 **
 **  ljkOS_AllocateMemory
 **
 **  Allocate memory wrapper.
 **
 **  INPUT:
 **
 **      ljtSIZE_T Bytes
 **          Number of bytes to allocate.
 **
 **  OUTPUT:
 **
 **      ljtPOINTER *Memory
 **          Pointer to a variable that will hold the allocated memory location.
 */
ljeSTATUS
ljkOS_AllocateMemory(ljkOS Os, ljtSIZE_T Bytes, ljtPOINTER *Memory)
{
    ljtPOINTER memory = ljvNULL;
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkHEADER_ARG("Os=%p Bytes=0x%zx", Os, Bytes);

    /* Verify the arguments. */
    ljmkVERIFY_ARGUMENT(Bytes > 0);
    ljmkVERIFY_ARGUMENT(Memory != ljvNULL);

    if (Bytes > PAGE_SIZE)
        memory = (ljtPOINTER)vmalloc(Bytes);
    else
        memory = (ljtPOINTER)kmalloc(Bytes, GFP_KERNEL | ljdNOWARN);

    if (memory == ljvNULL)
        /* Out of memory. */
        ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);

    /* Increase count. */
    atomic_inc(&Os->allocateCount);

    /* Return pointer to the memory allocation. */
    *Memory = memory;

OnError:
    /* Return the status. */
    ljmkFOOTER_ARG("*Memory=%p", ljmOPT_POINTER(Memory));
    return status;
}

/*******************************************************************************
 **
 **  ljkOS_FreeMemory
 **
 **  Free allocated memory wrapper.
 **
 **  INPUT:
 **
 **      ljtPOINTER Memory
 **          Pointer to memory allocation to free.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_FreeMemory(ljkOS Os, ljtPOINTER Memory)
{
    ljmkHEADER_ARG("Os=%p Memory=%p", Os, Memory);

    /* Verify the arguments. */
    ljmkVERIFY_ARGUMENT(Memory != ljvNULL);

    /* Free the memory from the OS pool. */
    if (is_vmalloc_addr(Memory))
        vfree(Memory);
    else
        kfree(Memory);

    /* Decrease count. */
    atomic_dec(&Os->allocateCount);

    /* Success. */
    ljmkFOOTER_NO();
    return ljvSTATUS_OK;
}

/*******************************************************************************
 **
 **  ljkOS_MapMemory
 **
 **  Map physical memory into the current process.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtPHYS_ADDR Physical
 **          Start of physical address memory.
 **
 **      ljtSIZE_T Bytes
 **          Number of bytes to map.
 **
 **  OUTPUT:
 **
 **      ljtPOINTER *Memory
 **          Pointer to a variable that will hold the logical address of the
 **          mapped memory.
 */
ljeSTATUS
ljkOS_MapMemory(ljkOS Os, ljtPHYS_ADDR Physical,
                ljtSIZE_T Bytes, ljtPOINTER *Logical)
{
    ljeSTATUS status = ljvSTATUS_OK;
    PLINUX_MDL_MAP mdlMap;
    PLINUX_MDL mdl = (PLINUX_MDL)Physical;
    ljkALLOCATOR allocator;
    ljtINT pid = _GetProcessID();
    ljtBOOL acquiredMutex = ljvFALSE;

    ljmkHEADER_ARG("Os=%p Physical=%p Bytes=0x%zx", Os, Physical, Bytes);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Physical != 0);
    ljmkVERIFY_ARGUMENT(Bytes > 0);
    ljmkVERIFY_ARGUMENT(Logical != ljvNULL);

    ljmkONERROR(ljkOS_AcquireMutex(Os, &mdl->mapsMutex, ljvINFINITE));
    acquiredMutex = ljvTRUE;

    mdlMap = FindMdlMap(mdl, pid);

    if (mdlMap == ljvNULL) {
        mdlMap = _CreateMdlMap(mdl, pid);
        if (mdlMap == ljvNULL)
            ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);
    }

    if (mdlMap->vmaAddr == ljvNULL) {
        allocator = mdl->allocator;

        ljmkONERROR(allocator->ops->MapUser(allocator, mdl, mdlMap, ljvFALSE));
    }

    ljmkONERROR(ljkOS_ReleaseMutex(Os, &mdl->mapsMutex));
    acquiredMutex = ljvFALSE;

    *Logical = mdlMap->vmaAddr;
    ljmkFOOTER_ARG("*Logical=%p", Logical);
    return ljvSTATUS_OK;

OnError:
    if (acquiredMutex)
        ljkOS_ReleaseMutex(Os, &mdl->mapsMutex);

    ljmkFOOTER();
    return status;
}

/******************************************************************************
 **
 **  ljkOS_UnmapMemory
 **
 **  Unmap physical memory out of the current process.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtPHYS_ADDR Physical
 **          Start of physical address memory.
 **
 **      ljtSIZE_T Bytes
 **          Number of bytes to unmap.
 **
 **      ljtPOINTER Memory
 **          Pointer to a previously mapped memory region.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_UnmapMemory(ljkOS Os, ljtPHYS_ADDR Physical,
                  ljtSIZE_T Bytes, ljtPOINTER Logical)
{
    ljeSTATUS status = ljvSTATUS_OK;
    ljmkHEADER_ARG("Os=%p Physical=0%p Bytes=0x%zx Logical=%p",
                   Os, Physical, Bytes, Logical);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Physical != 0);
    ljmkVERIFY_ARGUMENT(Bytes > 0);
    ljmkVERIFY_ARGUMENT(Logical != ljvNULL);

    ljmkONERROR(ljkOS_UnmapMemoryEx(Os, Physical, Bytes, Logical, _GetProcessID()));

OnError:
    ljmkFOOTER_NO();
    return status;
}

/******************************************************************************
 **
 **  ljkOS_UnmapMemoryEx
 **
 **  Unmap physical memory in the specified process.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtPHYS_ADDR Physical
 **          Start of physical address memory.
 **
 **      ljtSIZE_T Bytes
 **          Number of bytes to unmap.
 **
 **      ljtPOINTER Memory
 **          Pointer to a previously mapped memory region.
 **
 **      ljtUINT32 PID
 **          Pid of the process that opened the device and mapped this memory.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_UnmapMemoryEx(ljkOS Os, ljtPHYS_ADDR Physical,
                    ljtSIZE_T Bytes, ljtPOINTER Logical, ljtUINT32 PID)
{
    PLINUX_MDL_MAP mdlMap;
    PLINUX_MDL mdl = (PLINUX_MDL)Physical;
    ljeSTATUS status = ljvSTATUS_OK;
    ljtBOOL acquiredMutex = ljvFALSE;

    ljmkHEADER_ARG("Os=%p Physical=%p Bytes=0x%zx Logical=%p PID=%d",
                   Os, Physical, Bytes, Logical, PID);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Physical != 0);
    ljmkVERIFY_ARGUMENT(Bytes > 0);
    ljmkVERIFY_ARGUMENT(Logical != ljvNULL);
    ljmkVERIFY_ARGUMENT(PID != 0);

    if (Logical) {
        ljkALLOCATOR allocator = mdl->allocator;

        ljmkONERROR(ljkOS_AcquireMutex(Os, &mdl->mapsMutex, ljvINFINITE));
        acquiredMutex = ljvTRUE;

        mdlMap = FindMdlMap(mdl, PID);

        if (mdlMap == ljvNULL) {
            ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
        }

        if (mdlMap->vmaAddr != ljvNULL) {
            BUG_ON(!allocator || !allocator->ops->UnmapUser);
            allocator->ops->UnmapUser(allocator, mdl, mdlMap, mdl->bytes);
        }

        ljmkVERIFY_OK(_DestroyMdlMap(mdl, mdlMap));

        ljmkONERROR(ljkOS_ReleaseMutex(Os, &mdl->mapsMutex));
        acquiredMutex = ljvFALSE;
    }
    ljmkFOOTER_NO();
    return status;

OnError:
    if (acquiredMutex)
        ljkOS_ReleaseMutex(Os, &mdl->mapsMutex);

    ljmkFOOTER();
    return status;
}

/******************************************************************************
 **
 **  ljkOS_AllocateNonPagedMemory
 **
 **  Allocate a number of pages from non-paged memory.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljkKERNEL Kernel
 **          Which core allocates the memory.
 **
 **      ljtBOOL InUserSpace
 **          ljvTRUE if the pages need to be mapped into user space.
 **
 **      ljtUINT32 Flag
 **          Allocation attribute.
 **
 **      ljtSIZE_T *Bytes
 **          Pointer to a variable that holds the number of bytes to allocate.
 **
 **  OUTPUT:
 **
 **      ljtSIZE_T *Bytes
 **          Pointer to a variable that hold the number of bytes allocated.
 **
 **      ljtPHYS_ADDR *Physical
 **          Pointer to a variable that will hold the physical address of the
 **          allocation.
 **
 **      ljtPOINTER *Logical
 **          Pointer to a variable that will hold the logical address of the
 **          allocation.
 */
ljeSTATUS
ljkOS_AllocateNonPagedMemory(ljkOS Os,
                             ljkKERNEL Kernel,
                             ljtBOOL InUserSpace,
                             ljtUINT32 Flag,
                             ljtSIZE_T *Bytes,
                             ljtPHYS_ADDR *Physical,
                             ljtPOINTER *Logical)
{
    ljtSIZE_T bytes;
    ljtSIZE_T numPages;
    PLINUX_MDL mdl = ljvNULL;
    PLINUX_MDL_MAP mdlMap = ljvNULL;
    ljtPOINTER addr;
    ljeSTATUS status = ljvSTATUS_NOT_SUPPORTED;
    ljkALLOCATOR allocator;
    ljtBOOL zoneDMA32 = ljvFALSE;
    ljtBOOL acquiredMapsMutex = ljvFALSE;
    ljtBOOL acquiredMdlMutex = ljvFALSE;

    ljmkHEADER_ARG("Os=%p InUserSpace=%d *Bytes=0x%zx",
                   Os, InUserSpace, ljmOPT_VALUE(Bytes));

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Bytes != ljvNULL);
    ljmkVERIFY_ARGUMENT(*Bytes > 0);
    ljmkVERIFY_ARGUMENT(Physical != ljvNULL);
    ljmkVERIFY_ARGUMENT(Logical != ljvNULL);

    /* Align number of bytes to page size. */
    bytes = ljmALIGN(*Bytes, PAGE_SIZE);

    /* Get total number of pages.. */
    numPages = GetPageCount(bytes, 0);

    /* Allocate mdl structure */
    mdl = _CreateMdl(Os, Kernel);
    if (mdl == ljvNULL)
        ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);

    ljmkASSERT(Flag & ljvALLOC_FLAG_CONTIGUOUS);

#if defined(CONFIG_ZONE_DMA32) || defined(CONFIG_ZONE_DMA)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 37)
    zoneDMA32 = ljvTRUE;
# endif
#endif /* defined(CONFIG_ZONE_DMA32) || defined(CONFIG_ZONE_DMA) */

    if ((Flag & ljvALLOC_FLAG_4GB_ADDR) && !zoneDMA32)
        Flag &= ~ljvALLOC_FLAG_4GB_ADDR;

    mdl->fromUser = (Flag & ljvALLOC_FLAG_FROM_USER) != 0;

    /* Walk all allocators. */
    list_for_each_entry(allocator, &Os->allocatorList, link) {
        ljmkTRACE_ZONE(ljvLEVEL_INFO, ljvZONE_OS,
                       "%s(%d) flag = %x allocator->capability = %x",
                       __func__, __LINE__, Flag, allocator->capability);

#ifndef NO_DMA_COHERENT
        /* Point to dma coherent allocator. */
        if (!strcmp(allocator->name, "dma") ||
            ((Flag & allocator->capability) == Flag && numPages == 1)) {
            status = allocator->ops->Alloc(allocator, mdl, numPages, Flag);

            if (ljmIS_SUCCESS(status)) {
                mdl->allocator = allocator;
                break;
            }
        }
#else
        if ((Flag & allocator->capability) == Flag) {
            status = allocator->ops->Alloc(allocator, mdl, numPages, Flag);

            if (ljmIS_SUCCESS(status)) {
                mdl->allocator = allocator;
                break;
            }
        }
#endif /* NO_DMA_COHERENT */
    }

    /* Check status. */
    ljmkONERROR(status);

    mdl->cacheable = Flag & ljvALLOC_FLAG_CACHEABLE;

    mdl->bytes = bytes;
    mdl->numPages = numPages;

    mdl->contiguous = ljvTRUE;

    ljmkONERROR(allocator->ops->MapKernel(allocator, mdl, 0, bytes, &addr));

    if (!strcmp(allocator->name, "gfp"))
        /* Trigger a page fault. */
        memset(addr, 0, numPages * PAGE_SIZE);

    mdl->addr = addr;

    if (InUserSpace) {
        ljmkONERROR(ljkOS_AcquireMutex(Os, &mdl->mapsMutex, ljvINFINITE));
        acquiredMapsMutex = ljvTRUE;

        mdlMap = _CreateMdlMap(mdl, _GetProcessID());

        ljmkONERROR(ljkOS_ReleaseMutex(Os, &mdl->mapsMutex));
        acquiredMapsMutex = ljvFALSE;

        if (mdlMap == ljvNULL)
            ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);

        ljmkONERROR(allocator->ops->MapUser(allocator, mdl, mdlMap, ljvFALSE));

        *Logical = mdlMap->vmaAddr;
    } else {
        *Logical = addr;
    }

    /*
     * Add this to a global list.
     * Will be used by get physical address
     * and mapuser pointer functions.
     */
    ljmkONERROR(ljkOS_AcquireMutex(Os, &Os->mdlMutex, ljvINFINITE));
    acquiredMdlMutex = ljvTRUE;

    list_add_tail(&mdl->link, &Os->mdlHead);

    ljmkONERROR(ljkOS_ReleaseMutex(Os, &Os->mdlMutex));
    acquiredMdlMutex = ljvFALSE;

    /* Return allocated memory. */
    *Bytes = bytes;
    *Physical = (ljtPHYS_ADDR)mdl;

    /* Success. */
    status = ljvSTATUS_OK;

OnError:
    if (acquiredMdlMutex)
        ljmkVERIFY_OK(ljkOS_ReleaseMutex(Os, &Os->mdlMutex));
    if (acquiredMapsMutex)
        ljmkVERIFY_OK(ljkOS_ReleaseMutex(Os, &mdl->mapsMutex));
    if (ljmIS_ERROR(status)) {
        if (mdlMap) {
            ljmkVERIFY_OK(ljkOS_AcquireMutex(Os, &mdl->mapsMutex, ljvINFINITE));
            /* Free LINUX_MDL_MAP. */
            ljmkVERIFY_OK(_DestroyMdlMap(mdl, mdlMap));
            ljmkVERIFY_OK(ljkOS_ReleaseMutex(Os, &mdl->mapsMutex));
        }

        if (mdl) {
            /* Free LINUX_MDL. */
            ljmkVERIFY_OK(_DestroyMdl(mdl));
        }
    }
    /* Return the status. */
    ljmkFOOTER_ARG("*Bytes=0x%zx *Physical=%p *Logical=%p",
                   ljmOPT_VALUE(Bytes),
                   ljmOPT_POINTER(Physical),
                   ljmOPT_POINTER(Logical));
    return status;
}

/******************************************************************************
 **
 **  ljkOS_FreeNonPagedMemory
 **
 **  Free previously allocated and mapped pages from non-paged memory.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtPHYS_ADDR Physical
 **          Physical address of the allocated memory.
 **
 **      ljtPOINTER Logical
 **          Logical address of the allocated memory.
 **
 **      ljtSIZE_T Bytes
 **          Number of bytes allocated.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_FreeNonPagedMemory(ljkOS Os, ljtPHYS_ADDR Physical,
                         ljtPOINTER Logical, ljtSIZE_T Bytes)
{
    PLINUX_MDL mdl = (PLINUX_MDL)Physical;

    ljmkHEADER_ARG("Os=%p Bytes=0x%zx Physical=%p Logical=%p",
                   Os, Bytes, Physical, Logical);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Bytes > 0);
    ljmkVERIFY_ARGUMENT(Physical != 0);
    ljmkVERIFY_ARGUMENT(Logical != ljvNULL);

    ljmkVERIFY_OK(_DestroyMdl(mdl));

    /* Success. */
    ljmkFOOTER_NO();
    return ljvSTATUS_OK;
}

static inline ljkALLOCATOR
_FindAllocator(ljkOS Os, ljtUINT Flag)
{
    ljkALLOCATOR allocator;

    list_for_each_entry(allocator, &Os->allocatorList, link) {
        if ((allocator->capability & Flag) == Flag)
            return allocator;
    }

    return ljvNULL;
}

ljeSTATUS
ljkOS_RequestReservedMemory(ljkOS Os,
                            ljtPHYS_ADDR_T Start,
                            ljtSIZE_T Size,
                            const char *Name,
                            ljtBOOL Requested,
                            ljtPOINTER *MemoryHandle)
{
    PLINUX_MDL mdl = ljvNULL;
    ljeSTATUS status;
    ljkALLOCATOR allocator;
    ljsATTACH_DESC desc;
    ljtBOOL acquiredMutex = ljvFALSE;

    ljmkHEADER_ARG("start=0x%llx size=0x%lx name=%s", Start, Size, Name);

    /* Round up to page size. */
    Size = (Size + ~PAGE_MASK) & PAGE_MASK;

    mdl = _CreateMdl(Os, ljvNULL);
    if (!mdl)
        ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);

    desc.reservedMem.start = Start;
    desc.reservedMem.size = Size;
    desc.reservedMem.name = Name;
    desc.reservedMem.requested = Requested;
    desc.reservedMem.root = ljvTRUE;

    allocator = _FindAllocator(Os, ljvALLOC_FLAG_LINUX_RESERVED_MEM);
    if (!allocator) {
        ljmkPRINT("reserved-mem allocator not integrated!");
        ljmkONERROR(ljvSTATUS_GENERIC_IO);
    }

    /* Call attach. */
    ljmkONERROR(allocator->ops->Attach(allocator, &desc, mdl));

    /* Assign alloator. */
    mdl->allocator = allocator;
    mdl->bytes = Size;
    mdl->numPages = Size >> PAGE_SHIFT;
    mdl->contiguous = ljvTRUE;
    mdl->addr = ljvNULL;
    mdl->dmaHandle = Start;
    mdl->gid = 0;
    mdl->device = ljvNULL;

    /*
     * Add this to a global list.
     * Will be used by get physical address
     * and mapuser pointer functions.
     */
    ljmkONERROR(ljkOS_AcquireMutex(Os, &Os->mdlMutex, ljvINFINITE));
    acquiredMutex = ljvTRUE;

    list_add_tail(&mdl->link, &Os->mdlHead);

    ljmkONERROR(ljkOS_ReleaseMutex(Os, &Os->mdlMutex));
    acquiredMutex = ljvFALSE;

    *MemoryHandle = (void *)mdl;

    ljmkFOOTER_NO();
    return ljvSTATUS_OK;

OnError:
    if (mdl)
        ljmkVERIFY_OK(_DestroyMdl(mdl));

    if (acquiredMutex)
        ljkOS_ReleaseMutex(Os, &Os->mdlMutex);

    ljmkFOOTER();
    return status;
}

void
ljkOS_ReleaseReservedMemory(ljkOS Os, ljtPOINTER MemoryHandle)
{
    PLINUX_MDL mdl = (PLINUX_MDL)MemoryHandle;

    if (mdl)
        ljmkVERIFY_OK(_DestroyMdl(mdl));
}

/******************************************************************************
 **
 **  ljkOS_RequestReservedMemoryArea
 **
 **  request a reserved memory area. it is used for dynamic mapping for usr
 **
 **  INPUT:
 **
 **      ljtPOINTER MemoryHandle
 **          Pointer to the root MDL.
 **
 **      ljtSIZE_T Offset
 **          Offset from the root reserved memory.
 **
 **       ljtSIZE_T Size
 **          Area size.
 **
 **       ljtPOINTER *MemoryAreaHandle
 **          Sub MDL address to save
 **  OUTPUT:
 **
 **      ljtPOINTER *MemoryAreaHandle
 **          Pointer to sub MDL.
 */
ljeSTATUS
ljkOS_RequestReservedMemoryArea(ljkOS Os,
                                ljtPOINTER MemoryHandle,
                                ljtSIZE_T Offset,
                                ljtSIZE_T Size,
                                ljtPOINTER *MemoryAreaHandle)
{
    PLINUX_MDL rootMdl = (PLINUX_MDL)MemoryHandle;
    PLINUX_MDL subMdl = ljvNULL;
    ljeSTATUS status;
    ljsATTACH_DESC desc;
    ljtBOOL acquiredMutex = ljvFALSE;

    ljmkHEADER_ARG("MemoryHandle=%p Offset=0x%lx size=0x%lx",
                   MemoryHandle, Offset, Size);

    /* Round up to page size. */
    Size = (Size + ~PAGE_MASK) & PAGE_MASK;

    subMdl = _CreateMdl(rootMdl->os, ljvNULL);
    if (!subMdl)
        ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);

    desc.reservedMem.start = rootMdl->dmaHandle + Offset;
    desc.reservedMem.size = Size;
    desc.reservedMem.name = "subRMA";
    /* consider that the memory region has requested by the root mdl */
    desc.reservedMem.requested = ljvTRUE;
    desc.reservedMem.root = ljvFALSE;

    /* Call attach. */
    ljmkONERROR(((ljkALLOCATOR)rootMdl->allocator)->ops->Attach((ljkALLOCATOR)rootMdl->allocator,
                                                                &desc, subMdl));

    /* Assign alloator. */
    subMdl->allocator = rootMdl->allocator;
    subMdl->bytes = Size;
    subMdl->numPages = Size >> PAGE_SHIFT;
    subMdl->contiguous = ljvTRUE;
    subMdl->addr = ljvNULL;
    subMdl->dmaHandle  = rootMdl->dmaHandle + Offset;
    subMdl->gid = 0;

    ljmkONERROR(ljkOS_AcquireMutex(Os, &rootMdl->os->mdlMutex, ljvINFINITE));
    acquiredMutex = ljvTRUE;

    list_add_tail(&subMdl->rmaLink, &rootMdl->rmaHead);

    ljmkONERROR(ljkOS_ReleaseMutex(Os, &rootMdl->os->mdlMutex));
    acquiredMutex = ljvFALSE;

    *MemoryAreaHandle = (void *)subMdl;

    ljmkFOOTER_NO();
    return ljvSTATUS_OK;

OnError:
    if (subMdl)
        ljmkVERIFY_OK(_DestroyMdl(subMdl));

    if (acquiredMutex)
        ljkOS_ReleaseMutex(Os, &rootMdl->os->mdlMutex);

    ljmkFOOTER();
    return status;
}

void
ljkOS_ReleaseReservedMemoryArea(ljtPOINTER MemoryAreaHandle)
{
    PLINUX_MDL subMdl = (PLINUX_MDL)MemoryAreaHandle;

    if (subMdl)
        ljmkVERIFY_OK(_DestroyMdl(subMdl));
}

/******************************************************************************
 **
 **  ljkOS_ReadRegisterEx
 **
 **  Read data from a register.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljkKERNEL Kernel,
 **          Which core to be read.
 **
 **      ljtUINT32 Address
 **          Address of register.
 **
 **  OUTPUT:
 **
 **      ljtUINT32 *Data
 **          Pointer to a variable that receives the data read from the
 **          register.
 */

ljeSTATUS
ljkOS_ReadRegisterEx(ljkOS Os, ljkKERNEL Kernel,
                     ljtUINT32 Address, ljtUINT32 *Data)
{
    ljkDEVICE device = Kernel->device;

    if (Kernel->core >= ljvCORE_COUNT)
        return ljvSTATUS_INVALID_ARGUMENT;

    if (Address > device->registerSizes[Kernel->core] - 1)
        return ljvSTATUS_INVALID_ARGUMENT;

    if (in_irq()) {
        uint32_t data;

        spin_lock(&Os->registerAccessLock);

        if (unlikely(Os->clockStates[device->id][Kernel->core] == ljvFALSE)) {
            spin_unlock(&Os->registerAccessLock);

            /*
             * Read register when external clock off:
             * 1. In shared IRQ, read register may be called and
             *    that's not our irq.
             */
            return ljvSTATUS_GENERIC_IO;
        }

        data = readl(device->registerBases[Kernel->core]);

        if (unlikely((data & 0x3) == 0x3)) {
            spin_unlock(&Os->registerAccessLock);

            /*
             * Read register when internal clock off:
             * a. In shared IRQ, read register may be called and
             *    that's not our irq.
             * b. In some condition, when ISR handled normal FE/PE,
             *    PM thread could turn off internal clock before
             *    ISR read register of async FE. And then IRQ
             *    handler will call read register with internal
             *    clock off.So here we just skip for such case.
             */
            return ljvSTATUS_GENERIC_IO;
        }

        *Data = readl((ljtUINT8 *)device->registerBases[Kernel->core] + Address);
        spin_unlock(&Os->registerAccessLock);
    } else {
        unsigned long flags;

        spin_lock_irqsave(&Os->registerAccessLock, flags);

        if (unlikely(Os->clockStates[device->id][Kernel->core] == ljvFALSE)) {
            spin_unlock_irqrestore(&Os->registerAccessLock, flags);

            /*
             * Read register when external clock off:
             * 2. In non-irq context, register access should not
             *    be called, otherwise it's driver bug.
             */
            pr_err("[ljmcore]: %s(%d) GPU[%d] core%d external clock off",
                   __func__, __LINE__, device->id, Kernel->core);
            ljmkBUG_ON(1);
            return ljvSTATUS_GENERIC_IO;
        }

        *Data = readl((ljtUINT8 *)device->registerBases[Kernel->core] + Address);
        spin_unlock_irqrestore(&Os->registerAccessLock, flags);

#if ljdDUMP_AHB_ACCESS
        /* Dangerous to print in interrupt context, skip. */
        ljmkPRINT("@[RD %d] %08x %08x", Kernel->core, Address, *Data);
#endif
    }

    /* Success. */
    return ljvSTATUS_OK;
}

static ljeSTATUS
_WriteRegisterEx(ljkOS Os, ljkKERNEL Kernel,
                 ljtUINT32 Address, ljtUINT32 Data, ljtBOOL Dump)
{
    ljkDEVICE device = Kernel->device;

    if (Kernel->core > ljvCORE_COUNT)
        return ljvSTATUS_INVALID_ARGUMENT;

    if (Address > device->registerSizes[Kernel->core] - 1)
        return ljvSTATUS_INVALID_ARGUMENT;

    if (Kernel->vGPUType == ljvVGPU_MDEV) {
        if (Address == 0x0) {
            Data = Data | (Kernel->vGPUId << 20);
            if (((Data << 9) >> 29) != Kernel->vGPUId)
                return ljvSTATUS_OK;
        }
    }

    if (in_irq()) {
        spin_lock(&Os->registerAccessLock);

        if (unlikely(Os->clockStates[device->id][Kernel->core] == ljvFALSE)) {
            spin_unlock(&Os->registerAccessLock);

            return ljvSTATUS_GENERIC_IO;
        }

        writel(Data, (ljtUINT8 *)device->registerBases[Kernel->core] + Address);
        spin_unlock(&Os->registerAccessLock);
    } else {
        unsigned long flags;

        if (Dump)
            ljmkDUMP(Os, "@[register.write %u 0x%05X 0x%08X]", Kernel->core, Address, Data);

        spin_lock_irqsave(&Os->registerAccessLock, flags);

        if (unlikely(Os->clockStates[device->id][Kernel->core] == ljvFALSE)) {
            spin_unlock_irqrestore(&Os->registerAccessLock, flags);

            pr_err("[ljmcore]: %s(%d) GPU[%d] core%d external clock off",
                   __func__, __LINE__, device->id, Kernel->core);

            /* Driver bug: register write when clock off. */
            ljmkBUG_ON(1);
            return ljvSTATUS_GENERIC_IO;
        }

        writel(Data, (ljtUINT8 *)device->registerBases[Kernel->core] + Address);
        spin_unlock_irqrestore(&Os->registerAccessLock, flags);

#if ljdDUMP_AHB_ACCESS
        /* Dangerous to print in interrupt context, skip. */
        ljmkPRINT("@[WR %d] %08x %08x", Kernel->core, Address, Data);
#endif
    }

    /* Success. */
    return ljvSTATUS_OK;
}

/*******************************************************************************
 **
 **  ljkOS_WriteRegisterEx
 **
 **  Write data to a register.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljkKERNEL Kernel
 **          Which core to be written.
 **
 **      ljtUINT32 Address
 **          Address of register.
 **
 **      ljtUINT32 Data
 **          Data for register.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */

ljeSTATUS
ljkOS_WriteRegisterEx(ljkOS Os, ljkKERNEL Kernel,
                      ljtUINT32 Address, ljtUINT32 Data)
{
    return _WriteRegisterEx(Os, Kernel, Address, Data, ljvTRUE);
}

ljeSTATUS
ljkOS_WriteRegisterEx_NoDump(ljkOS Os, ljkKERNEL Kernel,
                             ljtUINT32 Address, ljtUINT32 Data)
{
    return _WriteRegisterEx(Os, Kernel, Address, Data, ljvFALSE);
}

/******************************************************************************
 **
 **  ljkOS_GetPageSize
 **
 **  Get the system's page size.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **  OUTPUT:
 **
 **      ljtSIZE_T *PageSize
 **          Pointer to a variable that will receive the system's page size.
 */
ljeSTATUS
ljkOS_GetPageSize(ljkOS Os, ljtSIZE_T *PageSize)
{
    ljmkHEADER_ARG("Os=%p", Os);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(PageSize != ljvNULL);

    /* Return the page size. */
    *PageSize = (ljtSIZE_T)PAGE_SIZE;

    /* Success. */
    ljmkFOOTER_ARG("*PageSize=0x%zx", *PageSize);
    return ljvSTATUS_OK;
}

/******************************************************************************
 **
 **  _GetPhysicalAddressProcess
 **
 **  Get the physical system address of a corresponding virtual address for a
 **  given process.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to ljkOS object.
 **
 **      ljtPOINTER Logical
 **          Logical address.
 **
 **      ljtUINT32 ProcessID
 **          Process ID.
 **
 **  OUTPUT:
 **
 **      ljtPHYS_ADDR_T *Address
 **          Poinetr to a variable that receives the physical address.
 */
static ljeSTATUS
_GetPhysicalAddressProcess(ljkOS Os, ljtPOINTER Logical,
                           ljtUINT32 ProcessID, ljtPHYS_ADDR_T *Address)
{
    PLINUX_MDL mdl;
    ljeSTATUS status = ljvSTATUS_INVALID_ADDRESS;
    ljtINT i;
    ljtBOOL acquiredMapsMutex = ljvFALSE;
    ljtBOOL acquiredMdlMutex = ljvFALSE;

    ljmkHEADER_ARG("Os=%p Logical=%p ProcessID=%d",
                   Os, Logical, ProcessID);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Address != ljvNULL);

    ljmkONERROR(ljkOS_AcquireMutex(Os, &Os->mdlMutex, ljvINFINITE));
    acquiredMdlMutex = ljvTRUE;

    for (i = 0; i < ljdSYSTEM_RESERVE_COUNT; i++) {
        if (Os->device->contiguousPhysicals[i]) {
            /* Try the contiguous memory pool. */
            mdl = (PLINUX_MDL) Os->device->contiguousPhysicals[i];

            ljmkONERROR(ljkOS_AcquireMutex(Os, &mdl->mapsMutex, ljvINFINITE));
            acquiredMapsMutex = ljvTRUE;

            status = _ConvertLogical2Physical(Os, Logical, ProcessID, mdl, Address);

            ljkOS_ReleaseMutex(Os, &mdl->mapsMutex);
            acquiredMapsMutex = ljvFALSE;

            if (status == ljvSTATUS_OK)
                break;
        }
    }

    if (ljmIS_ERROR(status)) {
        /* Walk all MDLs. */
        list_for_each_entry(mdl, &Os->mdlHead, link) {
            ljmkONERROR(ljkOS_AcquireMutex(Os, &mdl->mapsMutex, ljvINFINITE));
            acquiredMapsMutex = ljvTRUE;

            if (mdl->addr != ljvNULL) {
                status = _ConvertLogical2Physical(Os, Logical, ProcessID, mdl, Address);
            } else if (!list_empty(&mdl->rmaHead)) {
                PLINUX_MDL subMdl;

                list_for_each_entry(subMdl, &mdl->rmaHead, rmaLink) {
                    status = _ConvertLogical2Physical(Os, Logical, ProcessID, subMdl, Address);
                    if (ljmIS_SUCCESS(status))
                        break;
                }
            }

            ljkOS_ReleaseMutex(Os, &mdl->mapsMutex);
            acquiredMapsMutex = ljvFALSE;

            if (ljmIS_SUCCESS(status))
                break;
        }
    }

    ljkOS_ReleaseMutex(Os, &Os->mdlMutex);
    acquiredMdlMutex = ljvFALSE;

    ljmkONERROR(status);
    /* Success. */

OnError:
    if (acquiredMapsMutex)
        ljkOS_ReleaseMutex(Os, &mdl->mapsMutex);

    if (acquiredMdlMutex)
        ljkOS_ReleaseMutex(Os, &Os->mdlMutex);

    /* Return the status. */
    ljmkFOOTER_ARG("*Address=0x%llx", *Address);
    return status;
}

/******************************************************************************
 **
 **  ljkOS_GetPhysicalAddress
 **
 **  Get the physical system address of a corresponding virtual address.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtPOINTER Logical
 **          Logical address.
 **
 **  OUTPUT:
 **
 **      ljtPHYS_ADDR_T *Address
 **          Poinetr to a variable that receives the physical address.
 */
ljeSTATUS
ljkOS_GetPhysicalAddress(ljkOS Os, ljtPOINTER Logical, ljtPHYS_ADDR_T *Address)
{
    ljeSTATUS status;
    ljtUINT32 processID;

    ljmkHEADER_ARG("Os=%p Logical=%p", Os, Logical);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Address != ljvNULL);

    /* Query page table of current process first. */
    status = _QueryProcessPageTable(Logical, Address);

    if (ljmIS_ERROR(status)) {
        /* Get current process ID. */
        processID = _GetProcessID();

        /* Route through other function. */
        ljmkONERROR(_GetPhysicalAddressProcess(Os, Logical, processID, Address));
    }

    /* Success. */
    ljmkFOOTER_ARG("*Address=0x%llx", *Address);
    return ljvSTATUS_OK;

OnError:
    /* Return the status. */
    ljmkFOOTER();
    return status;
}

ljeSTATUS
ljkOS_GetPhysicalFromHandle(ljkOS Os, ljtPHYS_ADDR Physical,
                            ljtSIZE_T Offset, ljtPHYS_ADDR_T *PhysicalAddress)
{
    PLINUX_MDL mdl = (PLINUX_MDL)Physical;
    ljkALLOCATOR allocator = mdl->allocator;

    return allocator->ops->Physical(allocator, mdl, Offset, PhysicalAddress);
}

ljeSTATUS
ljkOS_GetDmaAddressFromHandle(ljkOS Os, ljtPHYS_ADDR Physical,
                            ljtSIZE_T Offset, ljtPHYS_ADDR_T *DmaAddress)
{
    PLINUX_MDL mdl = (PLINUX_MDL)Physical;
    ljkALLOCATOR allocator = mdl->allocator;

    return allocator->ops->GetDmaAddress(allocator, mdl, Offset, DmaAddress);
}

/******************************************************************************
 **
 **  ljkOS_UserLogicalToPhysical
 **
 **  Get the physical system address of a corresponding user virtual address.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtPOINTER Logical
 **          Logical address.
 **
 **  OUTPUT:
 **
 **      ljtPHYS_ADDR_T *Address
 **          Pointer to a variable that receives the physical address.
 */
ljeSTATUS
ljkOS_UserLogicalToPhysical(ljkOS Os, ljtPOINTER Logical, ljtPHYS_ADDR_T *Address)
{
    return ljkOS_GetPhysicalAddress(Os, Logical, Address);
}

ljeSTATUS
_ConvertLogical2Physical(ljkOS Os,
                         ljtPOINTER Logical,
                         ljtUINT32 ProcessID,
                         PLINUX_MDL Mdl,
                         ljtPHYS_ADDR_T *Physical)
{
    ljkALLOCATOR allocator = Mdl->allocator;
    ljtSIZE_T offset;
    ljeSTATUS status = ljvSTATUS_NOT_FOUND;
    ljtINT8_PTR vBase;

    /* TASK_SIZE is userspace - kernelspace virtual memory split. */
    if ((ljtUINTPTR_T)Logical >= TASK_SIZE) {
        /* Kernel virtual address. */
        vBase = Mdl->addr;
    } else {
        /* User virtual address. */
        PLINUX_MDL_MAP map;

        map = FindMdlMap(Mdl, (ljtINT)ProcessID);
        vBase = (map == ljvNULL) ? ljvNULL : (ljtINT8_PTR)map->vmaAddr;
    }

    /* Is the given address within that range. */
    if (vBase != ljvNULL &&
        ((ljtINT8_PTR)Logical >= vBase) &&
        ((ljtINT8_PTR)Logical < vBase + Mdl->bytes)) {
        offset = (ljtINT8_PTR)Logical - vBase;

        allocator->ops->GetDmaAddress(allocator, Mdl, offset, Physical);

        status = ljvSTATUS_OK;
    }

    return status;
}

/******************************************************************************
 **
 **  ljkOS_MapPhysical
 **
 **  Map a physical address into kernel space.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtPHYS_ADDR_T Physical
 **          Physical address of the memory to map.
 **
 **      ljtSIZE_T Bytes
 **          Number of bytes to map.
 **
 **  OUTPUT:
 **
 **      ljtPOINTER *Logical
 **          Pointer to a variable that receives the base address of the mapped
 **          memory.
 */
ljeSTATUS
ljkOS_MapPhysical(ljkOS Os, ljtPHYS_ADDR_T Physical,
                  ljtSIZE_T Bytes, ljtPOINTER *Logical)
{
    ljtPOINTER logical;
    PLINUX_MDL mdl;
    ljtBOOL found = ljvFALSE;
    dma_addr_t physical = Physical;
    ljeSTATUS status = ljvSTATUS_OK;
    ljtBOOL acquiredMutex = ljvFALSE;

    ljmkHEADER_ARG("Os=%p Physical=0x%llx Bytes=0x%zx", Os, Physical, Bytes);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Bytes > 0);
    ljmkVERIFY_ARGUMENT(Logical != ljvNULL);

    ljmkONERROR(ljkOS_AcquireMutex(Os, &Os->mdlMutex, ljvINFINITE));
    acquiredMutex = ljvTRUE;


    /*
     * Go through our mapping to see if we know this physical
     * address already.
     */
    list_for_each_entry(mdl, &Os->mdlHead, link) {
        if (mdl->dmaHandle != 0) {
            if (physical >= mdl->dmaHandle &&
                (physical < mdl->dmaHandle + mdl->bytes)) {
                if (mdl->addr != ljvNULL) {
                    *Logical = mdl->addr + (physical - mdl->dmaHandle);
                    found = ljvTRUE;
                } else if (!list_empty(&mdl->rmaHead)) {
                    PLINUX_MDL subMdl;

                    /*
                     * when enable dynamic mapping, MDL from the mdlHead is the root.
                     * the sub MDL should be looped from root MDL.
                     */
                    list_for_each_entry(subMdl, &mdl->rmaHead, rmaLink) {
                        if (physical >= subMdl->dmaHandle &&
                            (physical < subMdl->dmaHandle + subMdl->bytes) &&
                            subMdl->addr != 0) {
                            *Logical = subMdl->addr + (physical - subMdl->dmaHandle);
                            found = ljvTRUE;
                            break;
                        }
                    }
                }

                if (found)
                    break;
            }
        }
    }

    ljmkONERROR(ljkOS_ReleaseMutex(Os, &Os->mdlMutex));
    acquiredMutex = ljvFALSE;

    if (!found) {
        unsigned long pfn = physical >> PAGE_SHIFT;

        if (pfn_valid(pfn)) {
            ljtUINT32 offset = physical & ~PAGE_MASK;
            struct page **pages;
            struct page *page;
            ljtSIZE_T numPages;
            ljtSIZE_T i;
            pgprot_t pgprot;

            numPages = GetPageCount(PAGE_ALIGN(offset + Bytes), 0);

            pages = kmalloc_array(numPages, sizeof(struct page *), GFP_KERNEL | ljdNOWARN);
            if (!pages)
                ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);

            page = pfn_to_page(pfn);

            for (i = 0; i < numPages; i++)
                pages[i] = nth_page(page, i);

#if ljdENABLE_BUFFERABLE_VIDEO_MEMORY
            pgprot = pgprot_writecombine(PAGE_KERNEL);
#else
            pgprot = pgprot_noncached(PAGE_KERNEL);
#endif

            logical = vmap(pages, numPages, 0, pgprot);

            kfree(pages);

            if (logical == ljvNULL) {
                ljmkTRACE_ZONE(ljvLEVEL_INFO, ljvZONE_OS,
                               "%s(%d): Failed to vmap",
                               __func__, __LINE__);

                /* Out of resources. */
                ljmkONERROR(ljvSTATUS_OUT_OF_RESOURCES);
            }

            logical += offset;
        } else {
            /* Map memory as cached memory. */
            request_mem_region(physical, Bytes, "MapRegion");
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 6, 0)
            logical = (ljtPOINTER)ioremap(physical, Bytes);
#else
            logical = (ljtPOINTER)ioremap_nocache(physical, Bytes);
#endif
            if (logical == ljvNULL) {
                ljmkTRACE_ZONE(ljvLEVEL_INFO, ljvZONE_OS,
                               "%s(%d): Failed to ioremap",
                               __func__, __LINE__);

                /* Out of resources. */
                ljmkONERROR(ljvSTATUS_OUT_OF_RESOURCES);
            }
        }

        /* Return pointer to mapped memory. */
        *Logical = logical;
    }

OnError:
    if (acquiredMutex)
        ljkOS_ReleaseMutex(Os, &Os->mdlMutex);

    /* Success. */
    ljmkFOOTER_ARG("*Logical=%p", *Logical);
    return status;
}

/******************************************************************************
 **
 **  ljkOS_UnmapPhysical
 **
 **  Unmap a previously mapped memory region from kernel memory.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtPOINTER Logical
 **          Pointer to the base address of the memory to unmap.
 **
 **      ljtSIZE_T Bytes
 **          Number of bytes to unmap.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_UnmapPhysical(ljkOS Os, ljtPOINTER Logical, ljtSIZE_T Bytes)
{
    PLINUX_MDL mdl;
    ljtBOOL found = ljvFALSE;
    ljeSTATUS status = ljvSTATUS_OK;
    ljtBOOL acquiredMutex = ljvFALSE;

    ljmkHEADER_ARG("Os=%p Logical=%p Bytes=0x%zx", Os, Logical, Bytes);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Logical != ljvNULL);
    ljmkVERIFY_ARGUMENT(Bytes > 0);

    ljmkONERROR(ljkOS_AcquireMutex(Os, &Os->mdlMutex, ljvINFINITE));
    acquiredMutex = ljvTRUE;

    list_for_each_entry(mdl, &Os->mdlHead, link) {
        if (mdl->addr != ljvNULL) {
            if ((Logical >= (ljtPOINTER)mdl->addr) &&
                (Logical < (ljtPOINTER)((ljtSTRING)mdl->addr + mdl->bytes))) {
                found = ljvTRUE;
            }
        } else if (!list_empty(&mdl->rmaHead)) {
            PLINUX_MDL subMdl;
            /* Find the subMDL */
            list_for_each_entry(subMdl, &mdl->rmaHead, rmaLink) {
                if (subMdl->addr != ljvNULL &&
                    (Logical >= (ljtPOINTER)subMdl->addr) &&
                    (Logical < (ljtPOINTER)((ljtSTRING)subMdl->addr + subMdl->bytes))) {
                    found = ljvTRUE;
                    break;
                }
            }
        }

        if (found)
            break;
    }

    ljmkONERROR(ljkOS_ReleaseMutex(Os, &Os->mdlMutex));
    acquiredMutex = ljvFALSE;

    if (!found)
        /* Unmap the memory. */
        vunmap((void *)((unsigned long)Logical & PAGE_MASK));

    /* Success. */
    ljmkFOOTER_NO();
    return ljvSTATUS_OK;

OnError:
    if (acquiredMutex)
        ljkOS_ReleaseMutex(Os, &Os->mdlMutex);
    ljmkFOOTER();
    return status;
}

/******************************************************************************
 **
 **  ljkOS_DeleteMutex
 **
 **  Delete a mutex.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtPOINTER Mutex
 **          Pointer to the mute to be deleted.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
#if IS_ENABLED(CONFIG_PROVE_LOCKING) && (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 1, 0))
ljeSTATUS
ljkOS_DeleteMutex(ljkOS Os, ljtPOINTER Mutex)
{
    ljeSTATUS status = ljvSTATUS_OK;
    struct key_mutex *key_mut = (struct key_mutex *)Mutex;

    ljmkHEADER_ARG("Os=%p Mutex=%p", Os, Mutex);

    /* Validate the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Mutex != ljvNULL);

    /* Destroy the mutex. */

    mutex_destroy((struct mutex *)&key_mut->mut);
    lockdep_unregister_key(&key_mut->key);

    /* Free the mutex structure. */
    ljmkONERROR(ljkOS_Free(Os, Mutex));

OnError:
    /* Return status. */
    ljmkFOOTER();
    return status;
}


#else
ljeSTATUS
ljkOS_DeleteMutex(ljkOS Os, ljtPOINTER Mutex)
{
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkHEADER_ARG("Os=%p Mutex=%p", Os, Mutex);

    /* Validate the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Mutex != ljvNULL);

    /* Destroy the mutex. */
    mutex_destroy((struct mutex *)Mutex);

    /* Free the mutex structure. */
    ljmkONERROR(ljkOS_Free(Os, Mutex));

OnError:
    /* Return status. */
    ljmkFOOTER();
    return status;
}
#endif
/*******************************************************************************
 **
 **  ljkOS_AcquireMutex
 **
 **  Acquire a mutex.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtPOINTER Mutex
 **          Pointer to the mutex to be acquired.
 **
 **      ljtUINT32 Timeout
 **          Timeout value specified in milliseconds.
 **          Specify the value of ljvINFINITE to keep the thread suspended
 **          until the mutex has been acquired.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_AcquireMutex(ljkOS Os, ljtPOINTER Mutex, ljtUINT32 Timeout)
{
    ljeSTATUS status = ljvSTATUS_TIMEOUT;

    ljmkHEADER_ARG("Os=%p Mutex=%p Timeout=%u", Os, Mutex, Timeout);

    /* Validate the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Mutex != ljvNULL);

    if (Timeout == ljvINFINITE) {
        /* Lock the mutex. */
        mutex_lock(Mutex);

        /* Success. */
        status = ljvSTATUS_OK;
    } else {
        for (;;) {
            /* Try to acquire the mutex. */
            if (mutex_trylock(Mutex)) {
                /* Success. */
                status = ljvSTATUS_OK;
                break;
            }

            if (Timeout-- == 0)
                break;

            /* Wait for 1 millisecond. */
            ljmkVERIFY_OK(ljkOS_Delay(Os, 1));
        }
    }

    /* Timeout. */
    ljmkFOOTER();
    return status;
}

/*******************************************************************************
 **
 **  ljkOS_ReleaseMutex
 **
 **  Release an acquired mutex.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtPOINTER Mutex
 **          Pointer to the mutex to be released.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_ReleaseMutex(ljkOS Os, ljtPOINTER Mutex)
{
    ljmkHEADER_ARG("Os=%p Mutex=%p", Os, Mutex);

    /* Validate the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Mutex != ljvNULL);

    /* Release the mutex. */
    mutex_unlock(Mutex);

    /* Success. */
    ljmkFOOTER_NO();
    return ljvSTATUS_OK;
}

/*******************************************************************************
 **
 **  ljkOS_AtomicExchange
 **
 **  Atomically exchange a pair of 32-bit values.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtINT32_PTR Target
 **          Pointer to the 32-bit value to exchange.
 **
 **      ljtINT32 NewValue
 **          Specifies a new value for the 32-bit value pointed to by Target.
 **
 **      ljtINT32_PTR OldValue
 **          The old value of the 32-bit value pointed to by Target.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_AtomicExchange(ljkOS Os, ljtUINT32_PTR Target,
                     ljtUINT32 NewValue, ljtUINT32_PTR OldValue)
{
    /* Exchange the pair of 32-bit values. */
    *OldValue = (ljtUINT32)atomic_xchg((atomic_t *)Target, (int)NewValue);
    return ljvSTATUS_OK;
}

/*******************************************************************************
 **
 **  ljkOS_AtomicExchangePtr
 **
 **  Atomically exchange a pair of pointers.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtPOINTER *Target
 **          Pointer to the 32-bit value to exchange.
 **
 **      ljtPOINTER NewValue
 **          Specifies a new value for the pointer pointed to by Target.
 **
 **      ljtPOINTER *OldValue
 **          The old value of the pointer pointed to by Target.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_AtomicExchangePtr(ljkOS Os, ljtPOINTER *Target,
                        ljtPOINTER NewValue, ljtPOINTER *OldValue)
{
    /* Exchange the pair of pointers. */
    *OldValue = (ljtPOINTER)(ljtUINTPTR_T)atomic_xchg((atomic_t *)Target,
                                                      (int)(ljtUINTPTR_T)NewValue);
    return ljvSTATUS_OK;
}

/*******************************************************************************
 **
 **  ljkOS_AtomicSetMask
 **
 **  Atomically set mask to Atom
 **
 **  INPUT:
 **      ljtPOINTER Atom
 **          Pointer to the atom to set.
 **
 **      ljtUINT32 Mask
 **          Mask to set.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_AtomSetMask(ljtPOINTER Atom, ljtUINT32 Mask)
{
    ljtUINT32 oval, nval;

    do {
        oval = atomic_read((atomic_t *)Atom);
        nval = oval | Mask;
    } while (atomic_cmpxchg((atomic_t *)Atom, oval, nval) != oval);

    return ljvSTATUS_OK;
}

/*******************************************************************************
 **
 **  ljkOS_AtomClearMask
 **
 **  Atomically clear mask from Atom
 **
 **  INPUT:
 **      ljtPOINTER Atom
 **          Pointer to the atom to clear.
 **
 **      ljtUINT32 Mask
 **          Mask to clear.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_AtomClearMask(ljtPOINTER Atom, ljtUINT32 Mask)
{
    ljtUINT32 oval, nval;

    do {
        oval = atomic_read((atomic_t *)Atom);
        nval = oval & ~Mask;
    } while (atomic_cmpxchg((atomic_t *)Atom, oval, nval) != oval);

    return ljvSTATUS_OK;
}

/*******************************************************************************
 **
 **  ljkOS_AtomConstruct
 **
 **  Create an atom.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to a ljkOS object.
 **
 **  OUTPUT:
 **
 **      ljtPOINTER *Atom
 **          Pointer to a variable receiving the constructed atom.
 */
ljeSTATUS
ljkOS_AtomConstruct(ljkOS Os, ljtPOINTER *Atom)
{
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkHEADER_ARG("Os=%p", Os);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Atom != ljvNULL);

    /* Allocate the atom. */
    ljmkONERROR(ljkOS_Allocate(Os, ljmSIZEOF(atomic_t), Atom));

    /* Initialize the atom. */
    atomic_set((atomic_t *)*Atom, 0);

OnError:
    /* Return the status. */
    ljmkFOOTER_ARG("*Atom=%p", *Atom);
    return status;
}

/*******************************************************************************
 **
 **  ljkOS_AtomDestroy
 **
 **  Destroy an atom.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to a ljkOS object.
 **
 **      ljtPOINTER Atom
 **          Pointer to the atom to destroy.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_AtomDestroy(ljkOS Os, ljtPOINTER Atom)
{
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkHEADER_ARG("Os=%p Atom=%p", Os, Atom);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Atom != ljvNULL);

    /* Free the atom. */
    ljmkONERROR(ljmkOS_SAFE_FREE(Os, Atom));

OnError:
    /* Return the status. */
    ljmkFOOTER();
    return status;
}

/*******************************************************************************
 **
 **  ljkOS_AtomGet
 **
 **  Get the 32-bit value protected by an atom.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to a ljkOS object.
 **
 **      ljtPOINTER Atom
 **          Pointer to the atom.
 **
 **  OUTPUT:
 **
 **      ljtINT32_PTR Value
 **          Pointer to a variable the receives the value of the atom.
 */
ljeSTATUS
ljkOS_AtomGet(ljkOS Os, ljtPOINTER Atom, ljtINT32_PTR Value)
{
    /* Return the current value of atom. */
    *Value = atomic_read((atomic_t *)Atom);
    return ljvSTATUS_OK;
}

/*******************************************************************************
 **
 **  ljkOS_AtomSet
 **
 **  Set the 32-bit value protected by an atom.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to a ljkOS object.
 **
 **      ljtPOINTER Atom
 **          Pointer to the atom.
 **
 **      ljtINT32 Value
 **          The value of the atom.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_AtomSet(ljkOS Os, ljtPOINTER Atom, ljtINT32 Value)
{
    /* Set the current value of atom. */
    atomic_set((atomic_t *)Atom, Value);
    return ljvSTATUS_OK;
}

/*******************************************************************************
 **
 **  ljkOS_AtomIncrement
 **
 **  Atomically increment the 32-bit integer value inside an atom.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to a ljkOS object.
 **
 **      ljtPOINTER Atom
 **          Pointer to the atom.
 **
 **  OUTPUT:
 **
 **      ljtINT32_PTR Value
 **          Pointer to a variable that receives the original value of the atom.
 */
ljeSTATUS
ljkOS_AtomIncrement(ljkOS Os, ljtPOINTER Atom, ljtINT32_PTR Value)
{
    *Value = atomic_inc_return((atomic_t *)Atom) - 1;
    return ljvSTATUS_OK;
}

/*******************************************************************************
 **
 **  ljkOS_AtomDecrement
 **
 **  Atomically decrement the 32-bit integer value inside an atom.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to a ljkOS object.
 **
 **      ljtPOINTER Atom
 **          Pointer to the atom.
 **
 **  OUTPUT:
 **
 **      ljtINT32_PTR Value
 **          Pointer to a variable that receives the original value of the atom.
 */
ljeSTATUS
ljkOS_AtomDecrement(ljkOS Os, ljtPOINTER Atom, ljtINT32_PTR Value)
{
    /* Decrement the atom. */
    *Value = atomic_dec_return((atomic_t *)Atom) + 1;
    return ljvSTATUS_OK;
}

/*******************************************************************************
 **
 **  ljkOS_Delay
 **
 **  Delay execution of the current thread for a number of milliseconds.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtUINT32 Delay
 **          Delay to sleep, specified in milliseconds.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_Delay(ljkOS Os, ljtUINT32 Delay)
{
    ljmkHEADER_ARG("Os=%p Delay=%u", Os, Delay);

    if (Delay > 0) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 28)
        ktime_t delay = ktime_set((Delay / MSEC_PER_SEC),
                                  (Delay % MSEC_PER_SEC) * NSEC_PER_MSEC);
        __set_current_state(TASK_UNINTERRUPTIBLE);
        schedule_hrtimeout(&delay, HRTIMER_MODE_REL);
#else
        msleep(Delay);
#endif
    }

    /* Success. */
    ljmkFOOTER_NO();
    return ljvSTATUS_OK;
}

/******************************************************************************
 **
 **  ljkOS_Udelay
 **
 **  Delay execution of the current thread for a number of microseconds.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtUINT32 Delay
 **          Delay to sleep, specified in microseconds.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_Udelay(ljkOS Os, ljtUINT32 Delay)
{
    ljmkHEADER_ARG("Os=%p Delay=%u", Os, Delay);

    if (Delay > 0) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 28)
        ktime_t delay = ktime_set((Delay / USEC_PER_SEC),
                                  (Delay % USEC_PER_SEC) * NSEC_PER_USEC);
        __set_current_state(TASK_UNINTERRUPTIBLE);
        schedule_hrtimeout(&delay, HRTIMER_MODE_REL);
#else
        usleep_range((unsigned long)Delay, (unsigned long)Delay + 1);
#endif
    }

    /* Success. */
    ljmkFOOTER_NO();
    return ljvSTATUS_OK;
}

/******************************************************************************
 **
 **  ljkOS_GetTicks
 **
 **  Get the number of milliseconds since the system started.
 **
 **  INPUT:
 **
 **  OUTPUT:
 **
 **      ljtUINT32_PTR Time
 **          Pointer to a variable to get time.
 **
 */
ljeSTATUS
ljkOS_GetTicks(ljtUINT32_PTR Time)
{
    ljmkHEADER();

    *Time = jiffies_to_msecs(jiffies);

    ljmkFOOTER_NO();
    return ljvSTATUS_OK;
}

/******************************************************************************
 **
 **  ljkOS_TicksAfter
 **
 **  Compare time values got from ljkOS_GetTicks.
 **
 **  INPUT:
 **      ljtUINT32 Time1
 **          First time value to be compared.
 **
 **      ljtUINT32 Time2
 **          Second time value to be compared.
 **
 **  OUTPUT:
 **
 **      ljtBOOL_PTR IsAfter
 **          Pointer to a variable to result.
 **
 */
ljeSTATUS
ljkOS_TicksAfter(ljtUINT32 Time1, ljtUINT32 Time2, ljtBOOL_PTR IsAfter)
{
    ljmkHEADER();

    *IsAfter = time_after((unsigned long)Time1, (unsigned long)Time2);

    ljmkFOOTER_NO();
    return ljvSTATUS_OK;
}

/******************************************************************************
 **
 **  ljkOS_GetTime
 **
 **  Get the number of microseconds since the system started.
 **
 **  INPUT:
 **
 **  OUTPUT:
 **
 **      ljtUINT64_PTR Time
 **          Pointer to a variable to get time.
 **
 */
ljeSTATUS
ljkOS_GetTime(ljtUINT64_PTR Time)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 17, 0)
    struct timespec64 tv;

    ljmkHEADER();

    /* Return the time of day in microseconds. */
    ktime_get_real_ts64(&tv);
    *Time = (tv.tv_sec * 1000000ULL) + (tv.tv_nsec / 1000);
#else
    struct timeval tv;

    ljmkHEADER();

    /* Return the time of day in microseconds. */
    do_gettimeofday(&tv);
    *Time = (tv.tv_sec * 1000000ULL) + tv.tv_usec;
#endif

    ljmkFOOTER_NO();
    return ljvSTATUS_OK;
}

/******************************************************************************
 **
 **  _ExternalCacheOperation
 **
 **  External device cache operation, if support. If the core has any
 **  additional caches they must be invalidated after this function returns.
 **  If the core does not have any additional caches the externalCacheOperation
 **  in the platform->ops should remain NULL.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljeCACHEOPERATION Operation
 **          Cache Operation: ljvCACHE_FLUSH, ljvCACHE_CLEAN or
 **          ljvCACHE_INVALIDATE.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
static void
_ExternalCacheOperation(ljkOS Os, ljeCACHEOPERATION Operation)
{
    ljsPLATFORM *platform = Os->device->platform;

    if (platform && platform->ops->externalCacheOperation)
        platform->ops->externalCacheOperation(platform, Operation);
}

/******************************************************************************
 **
 **  ljkOS_MemoryBarrier
 **
 **  Make sure the CPU has executed everything up to this point and the data
 **  got written to the specified pointer.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtPOINTER Address
 **          Address of memory that needs to be barriered.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_MemoryBarrier(ljkOS Os, ljtPOINTER Address)
{
#if ljdWAR_PCIE_WC
    ljtUINT32 data;

    if (Address)
        ljkOS_ReadMappedPointer(Os, Address, &data);

    (void)data;
#endif

    _MemoryBarrier();

    _ExternalCacheOperation(Os, ljvCACHE_INVALIDATE);

    return ljvSTATUS_OK;
}

/******************************************************************************
 **
 **  ljkOS_AllocatePagedMemory
 **
 **  Allocate memory from the paged pool.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljkKERNEL Kernel
 **          Which core allocates the memory.
 **
 **      ljtUINT32 Flag
 **          Allocation attribute.
 **
 **      ljeVIDMEM_TYPE Type
 **          Allocation memory Type.
 **
 **      ljtSIZE_T *Bytes
 **          Number of bytes to allocate.
 **
 **  OUTPUT:
 **
 **      ljtSIZE_T *Bytes
 **          Return number of bytes actually allocated.
 **
 **      ljtUINT32 *Gid
 **          Save the global ID for the piece of allocated memory.
 **
 **      ljtPHYS_ADDR *Physical
 **          Pointer to a variable that receives the physical address of the
 **          memory allocation.
 */
ljeSTATUS
ljkOS_AllocatePagedMemory(ljkOS Os, ljkKERNEL Kernel, ljtUINT32 Flag, ljeVIDMEM_TYPE Type,
                          ljtSIZE_T *Bytes, ljtUINT32 *Gid, ljtPHYS_ADDR *Physical)
{
    ljtSIZE_T numPages;
    PLINUX_MDL mdl = ljvNULL;
    ljtSIZE_T bytes;
    ljeSTATUS status = ljvSTATUS_NOT_SUPPORTED;
    ljkALLOCATOR allocator;
    ljtBOOL zoneDMA32 = ljvFALSE;
    ljtBOOL acquiredMutex = ljvFALSE;

    ljmkHEADER_ARG("Os=%p Flag=%x *Bytes=0x%zx", Os, Flag, *Bytes);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(*Bytes > 0);
    ljmkVERIFY_ARGUMENT(Physical != ljvNULL);

    bytes = ljmALIGN(*Bytes, PAGE_SIZE);

    numPages = GetPageCount(bytes, 0);

    mdl = _CreateMdl(Os, Kernel);
    if (mdl == ljvNULL) {
        status = ljvSTATUS_OUT_OF_MEMORY;
        goto OnError;
    }

    mdl->type = Type;
#if defined(CONFIG_ZONE_DMA32) || defined(CONFIG_ZONE_DMA)
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 37)
    zoneDMA32 = ljvTRUE;
# endif
#endif

    if ((Flag & ljvALLOC_FLAG_4GB_ADDR) && !zoneDMA32)
        Flag &= ~ljvALLOC_FLAG_4GB_ADDR;

    mdl->fromUser = (Flag & ljvALLOC_FLAG_FROM_USER) != 0;

    /* Walk all allocators. */
    list_for_each_entry(allocator, &Os->allocatorList, link) {
        ljmkTRACE_ZONE(ljvLEVEL_INFO, ljvZONE_OS,
                       "%s(%d) flag = %x allocator->capability = %x",
                       __func__, __LINE__, Flag, allocator->capability);

        if ((Flag & allocator->capability) != Flag)
            continue;

        status = allocator->ops->Alloc(allocator, mdl, numPages, Flag);

        if (ljmIS_SUCCESS(status)) {
            mdl->allocator = allocator;
            break;
        }
    }

    /* Check status. */
    if (status == ljvSTATUS_OUT_OF_MEMORY)
        /*
         * Ignore error print in this function, leave it to
         * high level function.
         */
        goto OnError;
    else
        ljmkONERROR(status);

    mdl->dmaHandle = 0;
    mdl->addr = 0;
    mdl->bytes = bytes;
    mdl->numPages = numPages;
    mdl->contiguous = Flag & ljvALLOC_FLAG_CONTIGUOUS;
    mdl->cacheable = Flag & ljvALLOC_FLAG_CACHEABLE;

    /*
     * Add this to a global list.
     * Will be used by get physical address
     * and mapuser pointer functions.
     */
    ljmkONERROR(ljkOS_AcquireMutex(Os, &Os->mdlMutex, ljvINFINITE));
    acquiredMutex = ljvTRUE;

    list_add_tail(&mdl->link, &Os->mdlHead);

    ljmkONERROR(ljkOS_ReleaseMutex(Os, &Os->mdlMutex));
    acquiredMutex = ljvFALSE;

    /* Return allocated bytes. */
    *Bytes = bytes;

    if (Gid != ljvNULL)
        *Gid = mdl->gid;

    /* Return physical address. */
    *Physical = (ljtPHYS_ADDR)mdl;

    /* Success. */
    status = ljvSTATUS_OK;

OnError:
    if (ljmIS_ERROR(status) && mdl)
        /* Free the memory. */
        _DestroyMdl(mdl);

    if (acquiredMutex)
        ljkOS_ReleaseMutex(Os, &Os->mdlMutex);

    /* Return the status. */
    ljmkFOOTER_ARG("*Physical=%p", *Physical);
    return status;
}

/******************************************************************************
 **
 **  ljkOS_FreePagedMemory
 **
 **  Free memory allocated from the paged pool.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtPHYS_ADDR Physical
 **          Physical address of the allocation.
 **
 **      ljtSIZE_T Bytes
 **          Number of bytes of the allocation.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_FreePagedMemory(ljkOS Os, ljtPHYS_ADDR Physical, ljtSIZE_T Bytes)
{
    PLINUX_MDL mdl = (PLINUX_MDL)Physical;

    ljmkHEADER_ARG("Os=%p Physical=%p Bytes=0x%zx", Os, Physical, Bytes);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Physical != ljvNULL);
    ljmkVERIFY_ARGUMENT(Bytes > 0);

    /* Free the structure... */
    ljmkVERIFY_OK(_DestroyMdl(mdl));

    /* Success. */
    ljmkFOOTER_NO();
    return ljvSTATUS_OK;
}

/******************************************************************************
 **
 **  ljkOS_LockPages
 **
 **  Lock memory allocated from the paged pool.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtPHYS_ADDR Physical
 **          Physical address of the allocation.
 **
 **      ljtSIZE_T Bytes
 **          Number of bytes of the allocation.
 **
 **      ljtBOOL Cacheable
 **          Cache mode of mapping.
 **
 **  OUTPUT:
 **
 **      ljtPOINTER *Logical
 **          Pointer to a variable that receives the address of the mapped
 **          memory.
 */
ljeSTATUS
ljkOS_LockPages(ljkOS Os, ljtPHYS_ADDR Physical,
                ljtSIZE_T Bytes, ljtBOOL Cacheable, ljtPOINTER *Logical)
{
    ljeSTATUS status = ljvSTATUS_OK;
    PLINUX_MDL mdl;
    PLINUX_MDL_MAP mdlMap;
    ljkALLOCATOR allocator;
    ljtBOOL acquiredMutex = ljvFALSE;

    ljmkHEADER_ARG("Os=%p Physical=%p Bytes=0x%zx, Logical=%p",
                   Os, Physical, Bytes, Logical);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Physical != ljvNULL);
    ljmkVERIFY_ARGUMENT(Logical != ljvNULL);

    mdl = (PLINUX_MDL)Physical;
    allocator = mdl->allocator;

    ljmkONERROR(ljkOS_AcquireMutex(Os, &mdl->mapsMutex, ljvINFINITE));
    acquiredMutex = ljvTRUE;

    mdlMap = FindMdlMap(mdl, _GetProcessID());

    if (mdlMap == ljvNULL) {
        mdlMap = _CreateMdlMap(mdl, _GetProcessID());

        if (mdlMap == ljvNULL)
            ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);
    }

#if ljdENABLE_TTM
    if (mdlMap->vmaAddr == ljvNULL && !*Logical) {
        ljmkONERROR(allocator->ops->MapUser(allocator, mdl, mdlMap, Cacheable));
        mdlMap->fromDRM = 0;
    } else {
        mdlMap->fromDRM = 1;
        mdlMap->vmaAddr = *Logical;
    }
#else
    if (mdlMap->vmaAddr == ljvNULL)
        ljmkONERROR(allocator->ops->MapUser(allocator, mdl, mdlMap, Cacheable));
#endif

    mdlMap->count++;

    /* Convert pointer to MDL. */
    *Logical = mdlMap->vmaAddr;

OnError:
    if (acquiredMutex)
        ljkOS_ReleaseMutex(Os, &mdl->mapsMutex);

    /* Success. */
    ljmkFOOTER_ARG("*Logical=%p", *Logical);
    return status;
}

/* PageCount is GPU page count. */
ljeSTATUS
ljkOS_MapPagesEx(ljkOS Os,
                 ljkKERNEL Kernel,
                 ljkMMU Mmu,
                 ljtPHYS_ADDR Physical,
                 ljtSIZE_T Offset,
                 ljtSIZE_T PageCount,
                 ljtADDRESS Address,
                 ljtPOINTER PageTable,
                 ljtBOOL Writable,
                 ljeVIDMEM_TYPE Type)
{
    ljeSTATUS status = ljvSTATUS_OK;
    PLINUX_MDL mdl;
    ljtUINT32 *table;
    ljtSIZE_T offset = Offset;
    ljtUINT32 bytes = PageCount * 4;
    ljkALLOCATOR allocator;
    ljtUINT32 policyID  = 0;
    ljtUINT32 axiConfig = 0;
    ljsPLATFORM *platform = Os->device->platform;
    ljkMMU mmu = Mmu;
    ljsADDRESS_AREA *area;

    ljmkHEADER_ARG("Os=%p Kernel=%p Physical=%p PageCount=0x%zx Address=0x%llx PageTable=%p",
                   Os, Kernel, Physical, PageCount, Address, PageTable);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Physical != ljvNULL);
    ljmkVERIFY_ARGUMENT(PageCount > 0);
    ljmkVERIFY_ARGUMENT(PageTable != ljvNULL);
    ljmkVERIFY_ARGUMENT(Kernel != ljvNULL);

    /* Convert pointer to MDL. */
    mdl = (PLINUX_MDL)Physical;

    allocator = mdl->allocator;

    ljmkASSERT(allocator != ljvNULL);

    ljmkTRACE_ZONE(ljvLEVEL_INFO, ljvZONE_OS,
                   "%s(%d): Physical->0x%X PageCount->0x%X",
                   __func__, __LINE__,
                   (ljtUINT32)(ljtUINTPTR_T)Physical,
                   (ljtUINT32)(ljtUINTPTR_T)PageCount);

    area = &mmu->dynamicArea4K;

    table = (ljtUINT32 *)PageTable;

    if (platform && platform->ops->getPolicyID) {
        platform->ops->getPolicyID(platform, Type, &policyID, &axiConfig);

        ljmkBUG_ON(policyID > 0x1F);

        /* ID[3:0] is used in STLB. */
        policyID &= 0xF;
    }

    /* Get all the physical addresses and store them in the page table. */

    PageCount = PageCount / (PAGE_SIZE / 4096);

    /* Try to get the user pages so DMA can happen. */
    while (PageCount-- > 0) {
        ljtUINT i;
        ljtPHYS_ADDR_T phys = ~0ULL;

        allocator->ops->GetDmaAddress(allocator, mdl, offset, &phys);

        ljmkVERIFY_OK(ljkOS_CPUPhysicalToGPUPhysical(Os, phys, &phys));

        if (policyID) {
            /* AxUSER must not used for address currently. */
            ljmkBUG_ON((phys >> 32) & 0xF);

            /* Merge policyID to AxUSER[7:4]. */
            phys |= ((ljtPHYS_ADDR_T)policyID << 36);
        }

        /* remove LSB. */
        phys &= ~(4096ull - 1);

#if ljdENABLE_VG
        if (Kernel->core == ljvCORE_VG) {
            if ((phys + PAGE_SIZE) >> 32)
                ljmkONERROR(ljvSTATUS_NOT_SUPPORTED);

            for (i = 0; i < (PAGE_SIZE / 4096); i++) {
                ljmkONERROR(ljkVGMMU_SetPage(Kernel->vg->mmu,
                                             phys + (i * 4096),
                                             table++));
            }
        } else {
            for (i = 0; i < (PAGE_SIZE / 4096); i++) {
                ljmkONERROR(ljkMMU_SetPage(Kernel->mmu,
                                           phys + (i * 4096),
                                           ljvPAGE_TYPE_4K,
                                           (Address < ljd4G_SIZE), Writable,
                                           table++));
            }
        }
#else
        for (i = 0; i < (PAGE_SIZE / 4096); i++) {
            ljmkONERROR(ljkMMU_SetPage(mmu,
                                       phys + (i * 4096),
                                       ljvPAGE_TYPE_4K,
                                       (Address < ljd4G_SIZE),
                                       Writable,
                                       table++));
        }
#endif /* ljdENABLE_VG */

        offset += PAGE_SIZE;
    }

#if ljdENABLE_VG
    if (Kernel->core == ljvCORE_VG) {
        ljkVGMMU mmu_vg = Kernel->vg->mmu;
        ljtPHYS_ADDR mmuMdl = mmu_vg->pageTablePhysical;

        offset = (ljtUINT8_PTR)PageTable - (ljtUINT8_PTR)mmu_vg->pageTableLogical;

        ljmkVERIFY_OK(ljkOS_CacheClean(Os, _GetProcessID(), mmuMdl,
                                       offset, PageTable, bytes));
    } else {
        offset = (ljtUINT8_PTR)PageTable - (ljtUINT8_PTR)area->stlbLogical;

        /* must be in dynamic area. */
        ljmkASSERT(offset < area->stlbSize);

        ljmkVERIFY_OK(ljkVIDMEM_NODE_CleanCache(Kernel,
                                                area->stlbVideoMem,
                                                offset, PageTable, bytes));

        if (mmu->mtlbVideoMem)
            /* Flush MTLB table. */
            ljmkVERIFY_OK(ljkVIDMEM_NODE_CleanCache(Kernel,
                                                    mmu->mtlbVideoMem,
                                                    offset,
                                                    mmu->mtlbLogical,
                                                    mmu->mtlbSize));
    }
#else

    offset = (ljtUINT8_PTR)PageTable - (ljtUINT8_PTR)area->stlbLogical;

    /* must be in dynamic area. */
    ljmkASSERT(offset < area->stlbSize);

    ljmkVERIFY_OK(ljkVIDMEM_NODE_CleanCache(Kernel,
                                            area->stlbVideoMem,
                                            offset, PageTable, bytes));

    if (mmu->mtlbVideoMem) {
        /* Flush MTLB table. */
        ljmkVERIFY_OK(ljkVIDMEM_NODE_CleanCache(Kernel,
                                                mmu->mtlbVideoMem,
                                                offset,
                                                mmu->mtlbLogical,
                                                mmu->mtlbSize));
    }
#endif
OnError:
    /* Return the status. */
    ljmkFOOTER();
    return status;
}

/* PageCount is GPU page count. */
ljeSTATUS
ljkOS_UnmapPages(ljkOS Os, ljtSIZE_T PageCount, ljtADDRESS Address)
{
    return ljvSTATUS_OK;
}

/* Map 1M size GPU page */
ljeSTATUS
ljkOS_Map1MPages(ljkOS Os,
                 ljkKERNEL Kernel,
                 ljkMMU Mmu,
                 ljtPHYS_ADDR Physical,
                 ljtSIZE_T PageCount,
                 ljtADDRESS Address,
                 ljtPOINTER PageTable,
                 ljtBOOL Writable,
                 ljeVIDMEM_TYPE Type)
{
    ljeSTATUS status = ljvSTATUS_OK;
    PLINUX_MDL mdl;
    ljtUINT32 *table;
    ljtSIZE_T offset = 0;
    ljtSIZE_T bytes = PageCount * 4;
    ljkALLOCATOR allocator;
    ljtUINT32 policyID = 0;
    ljtUINT32 axiConfig = 0;
    ljsPLATFORM *platform = Os->device->platform;
    ljkMMU mmu = Mmu;
    ljsADDRESS_AREA *area;

    ljmkHEADER_ARG("Os=%p Kernel=%p Physical=%p PageCount=0x%zx Address=0x%llx PageTable=%p",
                   Os, Kernel, Physical, PageCount, Address, PageTable);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Physical != ljvNULL);
    ljmkVERIFY_ARGUMENT(PageCount > 0);
    ljmkVERIFY_ARGUMENT(Kernel != ljvNULL);
    ljmkVERIFY_ARGUMENT(PageTable != ljvNULL);

    /* Convert pointer to MDL. */
    mdl = (PLINUX_MDL)Physical;

    allocator = mdl->allocator;

    ljmkASSERT(allocator != ljvNULL);

    ljmkTRACE_ZONE(ljvLEVEL_INFO, ljvZONE_OS,
                   "%s(%d): Physical->0x%X PageCount->0x%X",
                   __func__, __LINE__,
                   (ljtUINT32)(ljtUINTPTR_T)Physical,
                   (ljtUINT32)(ljtUINTPTR_T)PageCount);

    area = &mmu->dynamicArea1M;

    table = (ljtUINT32 *)PageTable;

    if (platform && platform->ops->getPolicyID) {
        platform->ops->getPolicyID(platform, Type, &policyID, &axiConfig);

        ljmkBUG_ON(policyID > 0x1F);

        /* ID[3:0] is used in STLB. */
        policyID &= 0xF;
    }

    while (PageCount-- > 0) {
        ljtPHYS_ADDR_T phys = ~0ULL;

        allocator->ops->GetDmaAddress(allocator, mdl, offset, &phys);

        ljmkVERIFY_OK(ljkOS_CPUPhysicalToGPUPhysical(Os, phys, &phys));

        if (policyID) {
            /* AxUSER must not used for address currently. */
            ljmkBUG_ON((phys >> 32) & 0xF);

            /* Merge policyID to AxUSER[7:4]. */
            phys |= ((ljtPHYS_ADDR_T)policyID << 36);
        }

        /* Get the start physical of 1M page. */
        phys &= ~(ljd1M_PAGE_SIZE - 1);

        ljmkONERROR(ljkMMU_SetPage(mmu,
                                   phys,
                                   ljvPAGE_TYPE_1M,
                                   (Address < ljd4G_SIZE),
                                   Writable,
                                   table++));

        offset += ljd1M_PAGE_SIZE;
    }

    /* Flush the page table cache. */
    offset = (ljtUINT8_PTR)PageTable - (ljtUINT8_PTR)area->stlbLogical;

    /* must be in dynamic area. */
    ljmkASSERT(offset < area->stlbSize);

    ljmkVERIFY_OK(ljkVIDMEM_NODE_CleanCache(Kernel,
                                            area->stlbVideoMem,
                                            offset, PageTable, bytes));

    if (mmu->mtlbVideoMem) {
        /* Flush MTLB table. */
        ljmkVERIFY_OK(ljkVIDMEM_NODE_CleanCache(Kernel,
                                                mmu->mtlbVideoMem,
                                                offset,
                                                mmu->mtlbLogical,
                                                mmu->mtlbSize));
    }

OnError:
    /* Return the status. */
    ljmkFOOTER();
    return status;
}

/******************************************************************************
 **
 **  ljkOS_UnlockPages
 **
 **  Unlock memory allocated from the paged pool.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtPHYS_ADDR Physical
 **          Physical address of the allocation.
 **
 **      ljtSIZE_T Bytes
 **          Number of bytes of the allocation.
 **
 **      ljtPOINTER Logical
 **          Address of the mapped memory.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_UnlockPages(ljkOS Os,
                  ljtPHYS_ADDR Physical,
                  ljtSIZE_T Bytes,
                  ljtPOINTER Logical)
{
    PLINUX_MDL_MAP mdlMap;
    PLINUX_MDL mdl = (PLINUX_MDL)Physical;
    ljkALLOCATOR allocator = mdl->allocator;
    ljtINT pid = _GetProcessID();
    ljeSTATUS status = ljvSTATUS_OK;
    ljtBOOL acquiredMutex = ljvFALSE;
#if ljdENABLE_TTM
    ljtUINT32 ret = 0;
#endif

    ljmkHEADER_ARG("Os=%p Physical=%p Bytes=0x%zx Logical=%p",
                   Os, Physical, Bytes, Logical);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Physical != ljvNULL);

    ljmkONERROR(ljkOS_AcquireMutex(Os, &mdl->mapsMutex, ljvINFINITE));
    acquiredMutex = ljvTRUE;

/*
 *    for reserved memory(local memory), only need vm_munmap.
 */
    list_for_each_entry(mdlMap, &mdl->mapsHead, link) {
        if (mdlMap->vmaAddr != ljvNULL && mdlMap->pid == pid) {
            if (--mdlMap->count == 0) {
                #if ljdENABLE_TTM
                if (strcmp(allocator->name, "reserved-mem")) {
                    allocator->ops->UnmapUser(allocator, mdl, mdlMap, mdl->bytes);
                } else {
                    if (unlikely(!current->mm))
                        status = ljvSTATUS_INVALID_ARGUMENT;

                    if (mdlMap->vmaAddr != Logical)
                        ljmkPRINT("unlock page error");
                    ret = vm_munmap(ljmPTR_TO_UINT64(mdlMap->vmaAddr), Bytes);

                    if (ret == -EINVAL)
                        status = ljvSTATUS_INVALID_ARGUMENT;
                }
                #else
                allocator->ops->UnmapUser(allocator, mdl, mdlMap, mdl->bytes);
                #endif

                mdlMap->vmaAddr = ljvNULL;
            }
        }
    }

    ljmkONERROR(ljkOS_ReleaseMutex(Os, &mdl->mapsMutex));
    acquiredMutex = ljvFALSE;

    /* Success. */
    ljmkFOOTER_NO();
    return ljvSTATUS_OK;

OnError:
    if (acquiredMutex)
        ljkOS_ReleaseMutex(Os, &mdl->mapsMutex);
    ljmkFOOTER();
    return status;
}

/******************************************************************************
 **
 **  ljkOS_MapUserPointer
 **
 **  Map a pointer from the user process into the kernel address space.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtPOINTER Pointer
 **          Pointer in user process space that needs to be mapped.
 **
 **      ljtSIZE_T Size
 **          Number of bytes that need to be mapped.
 **
 **  OUTPUT:
 **
 **      ljtPOINTER *KernelPointer
 **          Pointer to a variable receiving the mapped pointer in kernel
 **          address pace.
 */
ljeSTATUS
ljkOS_MapUserPointer(ljkOS Os,
                     ljtPOINTER Pointer,
                     ljtSIZE_T Size,
                     ljtPOINTER *KernelPointer)
{
    ljmkHEADER_ARG("Os=%p Pointer=%p Size=0x%zx", Os, Pointer, Size);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Pointer != ljvNULL);
    ljmkVERIFY_ARGUMENT(Size > 0);
    ljmkVERIFY_ARGUMENT(KernelPointer != ljvNULL);

    *KernelPointer = Pointer;

    ljmkFOOTER_ARG("*KernelPointer=%p", *KernelPointer);
    return ljvSTATUS_OK;
}

/******************************************************************************
 **
 **  ljkOS_UnmapUserPointer
 **
 **  Unmap a user process pointer from the kernel address space.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtPOINTER Pointer
 **          Pointer in user process space that needs to be unmapped.
 **
 **      ljtSIZE_T Size
 **          Number of bytes that need to be unmapped.
 **
 **      ljtPOINTER KernelPointer
 **          Pointer in kernel address space that needs to be unmapped.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_UnmapUserPointer(ljkOS Os,
                       ljtPOINTER Pointer,
                       ljtSIZE_T Size,
                       ljtPOINTER KernelPointer)
{
    ljmkHEADER_ARG("Os=%p Pointer=%p Size=0x%zx KernelPointer=%p",
                   Os, Pointer, Size, KernelPointer);

    ljmkFOOTER_NO();
    return ljvSTATUS_OK;
}

/******************************************************************************
 **
 **  ljkOS_QueryNeedCopy
 **
 **  Query whether the memory can be accessed or mapped directly or it has to
 **  be copied.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtUINT32 ProcessID
 **          Process ID of the current process.
 **
 **  OUTPUT:
 **
 **      ljtBOOL_PTR NeedCopy
 **          Pointer to a boolean receiving ljvTRUE if the memory needs a copy
 **          or ljvFALSE if the memory can be accessed or mapped dircetly.
 */
ljeSTATUS
ljkOS_QueryNeedCopy(ljkOS Os, ljtUINT32 ProcessID, ljtBOOL_PTR NeedCopy)
{
    ljmkHEADER_ARG("Os=%p ProcessID=%d", Os, ProcessID);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(NeedCopy != ljvNULL);

    /* We need to copy data. */
    *NeedCopy = ljvTRUE;

    /* Success. */
    ljmkFOOTER_ARG("*NeedCopy=%d", *NeedCopy);
    return ljvSTATUS_OK;
}

/******************************************************************************
 **
 **  ljkOS_CopyFromUserData
 **
 **  Copy data from user to kernel memory.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtPOINTER KernelPointer
 **          Pointer to kernel memory.
 **
 **      ljtPOINTER Pointer
 **          Pointer to user memory.
 **
 **      ljtSIZE_T Size
 **          Number of bytes to copy.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_CopyFromUserData(ljkOS Os,
                       ljtPOINTER KernelPointer,
                       ljtPOINTER Pointer,
                       ljtSIZE_T Size)
{
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkHEADER_ARG("Os=%p KernelPointer=%p Pointer=%p Size=0x%zx",
                   Os, KernelPointer, Pointer, Size);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(KernelPointer != ljvNULL);
    ljmkVERIFY_ARGUMENT(Pointer != ljvNULL);
    ljmkVERIFY_ARGUMENT(Size > 0);

    /* Copy data from user. */
    if (copy_from_user(KernelPointer, Pointer, Size) != 0)
        /* Could not copy all the bytes. */
        ljmkONERROR(ljvSTATUS_OUT_OF_RESOURCES);

OnError:
    /* Return the status. */
    ljmkFOOTER();
    return status;
}

/******************************************************************************
 **
 **  ljkOS_CopyToUserData
 **
 **  Copy data from kernel to user memory.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtPOINTER KernelPointer
 **          Pointer to kernel memory.
 **
 **      ljtPOINTER Pointer
 **          Pointer to user memory.
 **
 **      ljtSIZE_T Size
 **          Number of bytes to copy.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_CopyToUserData(ljkOS Os,
                     ljtPOINTER KernelPointer,
                     ljtPOINTER Pointer,
                     ljtSIZE_T Size)
{
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkHEADER_ARG("Os=%p KernelPointer=%p Pointer=%p Size=0x%zx",
                   Os, KernelPointer, Pointer, Size);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(KernelPointer != ljvNULL);
    ljmkVERIFY_ARGUMENT(Pointer != ljvNULL);
    ljmkVERIFY_ARGUMENT(Size > 0);

    /* Copy data to user. */
    if (copy_to_user(Pointer, KernelPointer, Size) != 0)
        /* Could not copy all the bytes. */
        ljmkONERROR(ljvSTATUS_OUT_OF_RESOURCES);

OnError:
    /* Return the status. */
    ljmkFOOTER();
    return status;
}

/******************************************************************************
 **
 **  ljkOS_WriteMemory
 **
 **  Write data to a memory.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtPOINTER Address
 **          Address of the memory to write to.
 **
 **      ljtUINT32 Data
 **          Data for register.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_WriteMemory(ljkOS Os, ljtPOINTER Address, ljtUINT32 Data)
{
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkHEADER_ARG("Os=%p Address=%p Data=%u", Os, Address, Data);

    /* Verify the arguments. */
    ljmkVERIFY_ARGUMENT(Address != ljvNULL);

    /* Write memory. */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0)
    if (access_ok(Address, 4)) {
#else
    if (access_ok(VERIFY_WRITE, Address, 4)) {
#endif
        /* User address. */
        if (put_user(Data, (ljtUINT32 *)Address))
            ljmkONERROR(ljvSTATUS_INVALID_ADDRESS);
    } else {
        /*
         * don't check the virtual address, maybe it come from
         * io memory or reserved memory.
         */
        /* Kernel address. */
        *(ljtUINT32 *)Address = Data;
    }

OnError:
    ljmkFOOTER();
    return status;
}

ljeSTATUS
ljkOS_ReadMappedPointer(ljkOS Os, ljtPOINTER Address, ljtUINT32_PTR Data)
{
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkHEADER_ARG("Os=%p Address=%p Data=%p", Os, Address, Data);

    /* Verify the arguments. */
    ljmkVERIFY_ARGUMENT(Address != ljvNULL);

    /* Write memory. */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 0, 0)
    if (access_ok(Address, 4)) {
#else
    if (access_ok(VERIFY_READ, Address, 4)) {
#endif
        /* User address. */
        if (get_user(*Data, (ljtUINT32 *)Address))
            ljmkONERROR(ljvSTATUS_INVALID_ADDRESS);
    } else {
        /* Kernel address. */
        *Data = *(ljtUINT32_PTR)Address;
    }

OnError:
    ljmkFOOTER();
    return status;
}

/******************************************************************************
 **
 **  ljkOS_GetBaseAddress
 **
 **  Get the base address for the physical memory.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to the ljkOS object.
 **
 **  OUTPUT:
 **
 **      ljtUINT32_PTR BaseAddress
 **          Pointer to a variable that will receive the base address.
 */
ljeSTATUS
ljkOS_GetBaseAddress(ljkOS Os, ljtUINT32_PTR BaseAddress)
{
    ljmkHEADER_ARG("Os=%p", Os);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(BaseAddress != ljvNULL);

    /* Return base address. */
    *BaseAddress = Os->device->args.baseAddress;

    /* Success. */
    ljmkFOOTER_ARG("*BaseAddress=0x%08x", *BaseAddress);
    return ljvSTATUS_OK;
}

ljeSTATUS
ljkOS_SuspendInterrupt(ljkOS Os)
{
    return ljvSTATUS_OK;
}

ljeSTATUS
ljkOS_SuspendInterruptEx(ljkOS Os, ljeCORE Core)
{
    return ljvSTATUS_OK;
}

ljeSTATUS
ljkOS_ResumeInterrupt(ljkOS Os)
{
    return ljvSTATUS_OK;
}

ljeSTATUS
ljkOS_ResumeInterruptEx(ljkOS Os, ljeCORE Core)
{
    return ljvSTATUS_OK;
}

ljeSTATUS
ljkOS_MemCopy(ljtPOINTER Destination, ljtCONST_POINTER Source, ljtSIZE_T Bytes)
{
#if ljdSKIP_ARM_DC_INSTRUCTION
    ljtINT i = 0;
    ljtUINT8_PTR sptr = ljvNULL;
    ljtUINT8_PTR dptr = ljvNULL;
#endif

    ljmkHEADER_ARG("Destination=%p Source=%p Bytes=0x%zx",
                   Destination, Source, Bytes);

    ljmkVERIFY_ARGUMENT(Destination != ljvNULL);
    ljmkVERIFY_ARGUMENT(Source != ljvNULL);
    ljmkVERIFY_ARGUMENT(Bytes > 0);

    /*
     * Under ARM architecture, memcpy will generate DC instruction, and will
     * generate error when it encounters memory without cache and writebuffer
     */
#if ljdSKIP_ARM_DC_INSTRUCTION
    sptr = (ljtUINT8_PTR)Source;
    dptr = (ljtUINT8_PTR)Destination;
    for (i = 0; i < Bytes; i++)
        dptr[i] = sptr[i];
#else
    memcpy(Destination, Source, Bytes);
#endif

    ljmkFOOTER_NO();
    return ljvSTATUS_OK;
}

ljeSTATUS
ljkOS_ZeroMemory(ljtPOINTER Memory, ljtSIZE_T Bytes)
{
#if ljdSKIP_ARM_DC_INSTRUCTION
    ljtINT i = 0;
    ljtUINT8_PTR ptr = ljvNULL;
#endif

    ljmkHEADER_ARG("Memory=%p Bytes=0x%zx", Memory, Bytes);

    ljmkVERIFY_ARGUMENT(Memory != ljvNULL);
    ljmkVERIFY_ARGUMENT(Bytes > 0);

    /*
     * Under ARM architecture, memset will generate DC instruction, and will
     * generate error when it encounters memory without cache and writebuffer
     */
#if ljdSKIP_ARM_DC_INSTRUCTION
    ptr = (ljtUINT8_PTR)Memory;
    for (i = 0; i < Bytes; i++)
        ptr[i] = 0;
#else
    memset(Memory, 0, Bytes);
#endif

    ljmkFOOTER_NO();
    return ljvSTATUS_OK;
}

/******************************************************************************
 ********************************* Cache Control ******************************
 *****************************************************************************/
static ljeSTATUS
_CacheOperation(ljkOS Os,
                ljtUINT32 ProcessID,
                ljtPHYS_ADDR Handle,
                ljtSIZE_T Offset,
                ljtPOINTER Logical,
                ljtSIZE_T Bytes,
                ljeCACHEOPERATION Operation)
{
    PLINUX_MDL mdl = (PLINUX_MDL)Handle;
    PLINUX_MDL_MAP mdlMap;
    ljkALLOCATOR allocator;
    ljeSTATUS status = ljvSTATUS_OK;
    ljtBOOL acquiredMutex = ljvFALSE;

    if (!mdl || !mdl->allocator) {
        ljmkPRINT("[ljmcore]: %s: Logical=%p no mdl", __func__, Logical);
        return ljvSTATUS_INVALID_ARGUMENT;
    }

    allocator = mdl->allocator;

    if (allocator->ops->Cache) {
        ljmkONERROR(ljkOS_AcquireMutex(Os, &mdl->mapsMutex, ljvINFINITE));
        acquiredMutex = ljvTRUE;

        mdlMap = FindMdlMap(mdl, ProcessID);

#if !ljdENABLE_TTM
        ljmkONERROR(ljkOS_ReleaseMutex(Os, &mdl->mapsMutex));
        acquiredMutex = ljvFALSE;

        if (ProcessID && !mdlMap && !mdl->wrapFromPhysical && !mdl->wrapFromLogical)
            return ljvSTATUS_INVALID_ARGUMENT;
#else
        if (!mdlMap) {
            ljmkONERROR(ljkOS_ReleaseMutex(Os, &mdl->mapsMutex));
            acquiredMutex = ljvFALSE;
            return ljvSTATUS_OK;
        }
        ljmkONERROR(ljkOS_ReleaseMutex(Os, &mdl->mapsMutex));
        acquiredMutex = ljvFALSE;
#endif

#if !ljdENABLE_TTM
        if ((!ProcessID && mdl->cacheable) ||
            (mdlMap && mdlMap->cacheable) ||
            mdl->wrapFromLogical)
#endif
        {
            allocator->ops->Cache(allocator, mdl, Offset, Logical, Bytes, Operation);

            if (Operation == ljvCACHE_CLEAN || Operation == ljvCACHE_FLUSH)
                _ExternalCacheOperation(Os, ljvCACHE_INVALIDATE);

            return ljvSTATUS_OK;
        }
    }

    _MemoryBarrier();

    if (Operation == ljvCACHE_CLEAN || Operation == ljvCACHE_FLUSH)
        _ExternalCacheOperation(Os, ljvCACHE_INVALIDATE);

    return ljvSTATUS_OK;

OnError:
    if (acquiredMutex)
        ljkOS_ReleaseMutex(Os, &mdl->mapsMutex);
    return status;
}

/******************************************************************************
 **  ljkOS_CacheClean
 **
 **  Clean the cache for the specified addresses.  The GPU is going to need
 **  the data.  If the system is allocating memory as non-cachable, this
 **  function can be ignored.
 **
 **  ARGUMENTS:
 **
 **      ljkOS Os
 **          Pointer to ljkOS object.
 **
 **      ljtUINT32 ProcessID
 **          Process ID Logical belongs.
 **
 **      ljtPHYS_ADDR Handle
 **          Physical address handle.  If ljvNULL it is video memory.
 **
 **      ljtSIZE_T Offset
 **          Offset to this memory block.
 **
 **      ljtPOINTER Logical
 **          Logical address to flush.
 **
 **      ljtSIZE_T Bytes
 **          Size of the address range in bytes to flush.
 */

/*
 *  Following patch can be applied to kernel in case cache API is not exported.
 *
 *  diff --git a/arch/arm/mm/proc-syms.c b/arch/arm/mm/proc-syms.c
 *  index 054b491..e9e74ec 100644
 *  --- a/arch/arm/mm/proc-syms.c
 *  +++ b/arch/arm/mm/proc-syms.c
 *  @@ -30,6 +30,9 @@ EXPORT_SYMBOL(__cpuc_flush_user_all);
 *  EXPORT_SYMBOL(__cpuc_flush_user_range);
 *  EXPORT_SYMBOL(__cpuc_coherent_kern_range);
 *  EXPORT_SYMBOL(__cpuc_flush_dcache_area);
 *  +EXPORT_SYMBOL(__glue(_CACHE,_dma_map_area));
 *  +EXPORT_SYMBOL(__glue(_CACHE,_dma_unmap_area));
 *  +EXPORT_SYMBOL(__glue(_CACHE,_dma_flush_range));
 *  #else
 *  EXPORT_SYMBOL(cpu_cache);
 *  #endif
 */
ljeSTATUS
ljkOS_CacheClean(ljkOS Os,
                 ljtUINT32 ProcessID,
                 ljtPHYS_ADDR Handle,
                 ljtSIZE_T Offset,
                 ljtPOINTER Logical,
                 ljtSIZE_T Bytes)
{
    ljeSTATUS status;

    ljmkHEADER_ARG("Os=%p ProcessID=%d Handle=%p Offset=0x%zx Logical=%p Bytes=0x%zx",
                   Os, ProcessID, Handle, Offset, Logical, Bytes);

    ljmkONERROR(_CacheOperation(Os, ProcessID, Handle, Offset,
                                Logical, Bytes, ljvCACHE_CLEAN));

OnError:
    ljmkFOOTER();
    return status;
}

/******************************************************************************
 **  ljkOS_CacheInvalidate
 **
 **  Invalidate the cache for the specified addresses. The GPU is going to need
 **  data.  If the system is allocating memory as non-cachable, this function
 **  can be ignored.
 **
 **  ARGUMENTS:
 **
 **      ljkOS Os
 **          Pointer to ljkOS object.
 **
 **      ljtUINT32 ProcessID
 **          Process ID Logical belongs.
 **
 **      ljtPHYS_ADDR Handle
 **          Physical address handle.  If ljvNULL it is video memory.
 **
 **      ljtPOINTER Logical
 **          Logical address to flush.
 **
 **      ljtSIZE_T Bytes
 **          Size of the address range in bytes to flush.
 */
ljeSTATUS
ljkOS_CacheInvalidate(ljkOS Os,
                      ljtUINT32 ProcessID,
                      ljtPHYS_ADDR Handle,
                      ljtSIZE_T Offset,
                      ljtPOINTER Logical,
                      ljtSIZE_T Bytes)
{
    ljeSTATUS status;

    ljmkHEADER_ARG("Os=%p ProcessID=%d Handle=%p Offset=0x%zx Logical=%p Bytes=0x%zx",
                   Os, ProcessID, Handle, Offset, Logical, Bytes);

    ljmkONERROR(_CacheOperation(Os, ProcessID, Handle, Offset,
                                Logical, Bytes, ljvCACHE_INVALIDATE));

OnError:
    ljmkFOOTER();
    return status;
}

/******************************************************************************
 **  ljkOS_CacheFlush
 **
 **  Clean the cache for the specified addresses and invalidate the lines as
 **  well.  The GPU is going to need and modify the data.  If the system is
 **  allocating memory as non-cachable, this function can be ignored.
 **
 **  ARGUMENTS:
 **
 **      ljkOS Os
 **          Pointer to ljkOS object.
 **
 **      ljtUINT32 ProcessID
 **          Process ID Logical belongs.
 **
 **      ljtPHYS_ADDR Handle
 **          Physical address handle.  If ljvNULL it is video memory.
 **
 **      ljtPOINTER Logical
 **          Logical address to flush.
 **
 **      ljtSIZE_T Bytes
 **          Size of the address range in bytes to flush.
 */
ljeSTATUS
ljkOS_CacheFlush(ljkOS Os,
                 ljtUINT32 ProcessID,
                 ljtPHYS_ADDR Handle,
                 ljtSIZE_T Offset,
                 ljtPOINTER Logical,
                 ljtSIZE_T Bytes)
{
    ljeSTATUS status;

    ljmkHEADER_ARG("Os=%p ProcessID=%d Handle=%p Offset=0x%zx Logical=%p Bytes=0x%zx",
                   Os, ProcessID, Handle, Offset, Logical, Bytes);

    ljmkONERROR(_CacheOperation(Os, ProcessID, Handle, Offset,
                                Logical, Bytes, ljvCACHE_FLUSH));

OnError:
    ljmkFOOTER();
    return status;
}

/******************************************************************************
 ********************************* Broadcasting *******************************
 *****************************************************************************/

/******************************************************************************
 **
 **  ljkOS_Broadcast
 **
 **  System hook for broadcast events from the kernel driver.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to the ljkOS object.
 **
 **      ljkHARDWARE Hardware
 **          Pointer to the ljkHARDWARE object.
 **
 **      ljeBROADCAST Reason
 **          Reason for the broadcast.  Can be one of the following values:
 **
 **              ljvBROADCAST_GPU_IDLE
 **                  Broadcasted when the kernel driver thinks the GPU might
 **                  be idle.  This can be used to handle power management.
 **
 **              ljvBROADCAST_GPU_COMMIT
 **                  Broadcasted when any client process commits a command
 **                  buffer.  This can be used to handle power management.
 **
 **              ljvBROADCAST_GPU_STUCK
 **                  Broadcasted when the kernel driver hits the timeout
 **                  waiting for the GPU.
 **
 **              ljvBROADCAST_FIRST_PROCESS
 **                  First process is trying to connect to the kernel.
 **
 **              ljvBROADCAST_LAST_PROCESS
 **                  Last process has detached from the kernel.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_Broadcast(ljkOS Os, ljkHARDWARE Hardware, ljeBROADCAST Reason)
{
    ljeSTATUS status = ljvSTATUS_OK;
    ljeCHIPPOWERSTATE state;
#if ljdENABLE_PER_DEVICE_PM
    ljtBOOL acquired = ljvFALSE;
#endif

    ljmkHEADER_ARG("Os=%p Hardware=%p Reason=%d", Os, Hardware, Reason);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_OBJECT(Hardware, ljvOBJ_HARDWARE);

    switch (Reason) {
    case ljvBROADCAST_FIRST_PROCESS:
        ljmkTRACE_ZONE(ljvLEVEL_INFO, ljvZONE_OS, "First process has attached");
        break;

    case ljvBROADCAST_LAST_PROCESS:
        ljmkTRACE_ZONE(ljvLEVEL_INFO, ljvZONE_OS, "Last process has detached");

        /* Put GPU OFF. */
        ljmkONERROR(ljkHARDWARE_SetPowerState(Hardware, ljvPOWER_OFF_BROADCAST));
        break;

    case ljvBROADCAST_GPU_IDLE:
        ljmkTRACE_ZONE(ljvLEVEL_INFO, ljvZONE_OS, "GPU idle.");
#if ljdPOWER_SUSPEND_WHEN_IDLE
        state = ljvPOWER_SUSPEND_BROADCAST;
#else
        state = ljvPOWER_IDLE_BROADCAST;
#endif

        /* Put GPU IDLE or SUSPEND. */
        ljmkONERROR(ljkHARDWARE_SetPowerState(Hardware, state));

        /* Add idle process DB. */
        ljmkONERROR(ljkKERNEL_AddProcessDB(Hardware->kernel, 1,
                                           ljvDB_IDLE, ljvNULL, ljvNULL, 0));
        break;

    case ljvBROADCAST_GPU_COMMIT:
        ljmkTRACE_ZONE(ljvLEVEL_INFO, ljvZONE_OS, "COMMIT has arrived.");

        /* Add busy process DB. */
        ljmkONERROR(ljkKERNEL_AddProcessDB(Hardware->kernel, 0,
                                           ljvDB_IDLE, ljvNULL, ljvNULL, 0));

#if ljdENABLE_PER_DEVICE_PM
        if (Hardware->type == ljvHARDWARE_3D ||
            Hardware->type == ljvHARDWARE_3D2D ||
            Hardware->type == ljvHARDWARE_VIP) {
            ljkKERNEL kernel = Hardware->kernel;
            ljkDEVICE device = kernel->device;
            ljtUINT32 broCoreMask;
            ljtUINT i;

            ljmkONERROR(ljkOS_AcquireMutex(Hardware->os,
                                           device->powerMutex, ljvINFINITE));
            acquired = ljvTRUE;

            ljmkVERIFY_OK(ljkOS_AtomGet(Hardware->os,
                                        kernel->atomBroCoreMask,
                                        (ljtINT32_PTR)&broCoreMask));

            /* I am along. */
            if ((ljeCORE)broCoreMask == Hardware->core) {
                /* Put GPU ON. */
                ljmkONERROR(ljkHARDWARE_SetPowerState(Hardware, ljvPOWER_ON_AUTO));
            } else {
                /* Power on all the brother cores. */
                for (i = 0; i < device->coreNum; i++) {
                    kernel = device->coreInfoArray[i].kernel;

                    if ((1 << i) & broCoreMask) {
                        /* Put GPU ON. */
                        ljmkONERROR(ljkHARDWARE_SetPowerState(kernel->hardware,
                                                              ljvPOWER_ON_AUTO));
                    }
                }
            }

            ljmkONERROR(ljkOS_ReleaseMutex(Hardware->os, device->powerMutex));
            acquired = ljvFALSE;
        } else {
#endif /* ljdENABLE_PER_DEVICE_PM */
            /* Put GPU ON. */
            ljmkONERROR(ljkHARDWARE_SetPowerState(Hardware, ljvPOWER_ON_AUTO));
#if ljdENABLE_PER_DEVICE_PM
        }
#endif
        break;

    case ljvBROADCAST_GPU_STUCK:
        ljmkTRACE_N(ljvLEVEL_ERROR, 0, "ljvBROADCAST_GPU_STUCK\n");
        ljmkONERROR(ljkKERNEL_Recovery(Hardware->kernel));
        break;

    case ljvBROADCAST_AXI_BUS_ERROR:
        ljmkTRACE_N(ljvLEVEL_ERROR, 0, "ljvBROADCAST_AXI_BUS_ERROR\n");
        ljmkONERROR(ljkHARDWARE_DumpGPUState(Hardware));
        ljmkONERROR(ljkKERNEL_Recovery(Hardware->kernel));
        break;

    case ljvBROADCAST_OUT_OF_MEMORY:
        ljmkTRACE_N(ljvLEVEL_INFO, 0, "ljvBROADCAST_OUT_OF_MEMORY\n");

        status = _ShrinkMemory(Os);

        if (status == ljvSTATUS_NOT_SUPPORTED)
            goto OnError;

        ljmkONERROR(status);

        break;

    default:
        /* Skip unimplemented broadcast. */
        break;
    }

OnError:
#if ljdENABLE_PER_DEVICE_PM
    if (acquired)
        ljmkVERIFY_OK(ljkOS_ReleaseMutex(Hardware->os, Hardware->kernel->device->powerMutex));
#endif

    /* Return the status. */
    ljmkFOOTER();
    return status;
}

/******************************************************************************
 **
 **  ljkOS_BroadcastHurry
 **
 **  The GPU is running too slow.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to the ljkOS object.
 **
 **      ljkHARDWARE Hardware
 **          Pointer to the ljkHARDWARE object.
 **
 **      ljtUINT Urgency
 **          The higher the number, the higher the urgency to speed up the GPU.
 **          The maximum value is defined by the ljdDYNAMIC_EVENT_THRESHOLD.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_BroadcastHurry(ljkOS Os, ljkHARDWARE Hardware, ljtUINT Urgency)
{
    ljmkHEADER_ARG("Os=%p Hardware=%p Urgency=%u", Os, Hardware, Urgency);

    /* Do whatever you need to do to speed up the GPU now. */
    /*
     * {
     *  ljmkPRINT("Hurry! (%u)", Urgency);
     * }
     */

    /* Success. */
    ljmkFOOTER_NO();
    return ljvSTATUS_OK;
}

/******************************************************************************
 **
 **  ljkOS_BroadcastCalibrateSpeed
 **
 **  Calibrate the speed of the GPU.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to the ljkOS object.
 **
 **      ljkHARDWARE Hardware
 **          Pointer to the ljkHARDWARE object.
 **
 **      ljtUINT Idle, Time
 **          Idle/Time will give the percentage the GPU is idle, so you can use
 **          this to calibrate the working point of the GPU.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_BroadcastCalibrateSpeed(ljkOS Os, ljkHARDWARE Hardware,
                              ljtUINT Idle, ljtUINT Time)
{
    ljmkHEADER_ARG("Os=%p Hardware=%p Idle=%u Time=%u",
                   Os, Hardware, Idle, Time);

    /* Do whatever you need to do to callibrate the GPU speed. */
    /*
     * {
     *  ljtUINT percent = ((Time - Idle) * 100) / Time;
     *
     *  ljmkPRINT("Calibrate to %u%% (%u/%u).", percent, Idle, Time);
     * }
     */

    /* Success. */
    ljmkFOOTER_NO();
    return ljvSTATUS_OK;
}

/******************************************************************************
 ********************************** Semaphores ********************************
 *****************************************************************************/

/******************************************************************************
 **
 **  ljkOS_CreateSemaphore
 **
 **  Create a semaphore.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to the ljkOS object.
 **
 **  OUTPUT:
 **
 **      ljtPOINTER *Semaphore
 **          Pointer to the variable that will receive the created semaphore.
 */
ljeSTATUS
ljkOS_CreateSemaphore(ljkOS Os, ljtPOINTER *Semaphore)
{
    ljeSTATUS status = ljvSTATUS_OK;
    struct semaphore *sem = ljvNULL;

    ljmkHEADER_ARG("Os=%p", Os);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Semaphore != ljvNULL);

    /* Allocate the semaphore structure. */
    sem = kmalloc(ljmSIZEOF(struct semaphore), GFP_KERNEL | ljdNOWARN);

    if (sem == ljvNULL)
        ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);

    /* Initialize the semaphore. */
    sema_init(sem, 1);

    /* Return to caller. */
    *Semaphore = (ljtPOINTER)sem;

OnError:
    /* Return the status. */
    ljmkFOOTER();
    return status;
}

/******************************************************************************
 **
 **  ljkOS_AcquireSemaphore
 **
 **  Acquire a semaphore.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to the ljkOS object.
 **
 **      ljtPOINTER Semaphore
 **          Pointer to the semaphore thet needs to be acquired.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_AcquireSemaphore(ljkOS Os, ljtPOINTER Semaphore)
{
    int down_status;
    ljmkHEADER_ARG("Os=%p Semaphore=%p", Os, Semaphore);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Semaphore != ljvNULL);

    /* Acquire the semaphore. */
    down_status = down_interruptible((struct semaphore *)Semaphore);
    if (down_status == -EINTR) {
        ljmkFOOTER_NO();
        return ljvSTATUS_INTERRUPTED;
    }

    /* Success. */
    ljmkFOOTER_NO();
    return ljvSTATUS_OK;
}

/*******************************************************************************
 **
 **  ljkOS_TryAcquireSemaphore
 **
 **  Try to acquire a semaphore.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to the ljkOS object.
 **
 **      ljtPOINTER Semaphore
 **          Pointer to the semaphore thet needs to be acquired.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_TryAcquireSemaphore(ljkOS Os, ljtPOINTER Semaphore)
{
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkHEADER_ARG("Os=%p", Os);
    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Semaphore != ljvNULL);

    /* Acquire the semaphore. */
    if (down_trylock((struct semaphore *)Semaphore))
        /* Timeout. */
        status = ljvSTATUS_TIMEOUT;

    /* Success. */
    ljmkFOOTER();
    return status;
}

/******************************************************************************
 **
 **  ljkOS_ReleaseSemaphore
 **
 **  Release a previously acquired semaphore.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to the ljkOS object.
 **
 **      ljtPOINTER Semaphore
 **          Pointer to the semaphore thet needs to be released.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_ReleaseSemaphore(ljkOS Os, ljtPOINTER Semaphore)
{
    ljmkHEADER_ARG("Os=%p Semaphore=%p", Os, Semaphore);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Semaphore != ljvNULL);

    /* Release the semaphore. */
    up((struct semaphore *)Semaphore);

    /* Success. */
    ljmkFOOTER_NO();
    return ljvSTATUS_OK;
}

ljeSTATUS
ljkOS_ReleaseSemaphoreEx(ljkOS Os, ljtPOINTER Semaphore)
{
    struct semaphore *sem;
    unsigned long flags;

    ljmkHEADER_ARG("Os=%p Semaphore=%p", Os, Semaphore);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Semaphore != ljvNULL);

    sem = Semaphore;

    raw_spin_lock_irqsave(&sem->lock, flags);

    if (!sem->count) {
        raw_spin_unlock_irqrestore(&sem->lock, flags);
        up((struct semaphore *)Semaphore);
    } else {
        raw_spin_unlock_irqrestore(&sem->lock, flags);
    }

    /* Success. */
    ljmkFOOTER_NO();
    return ljvSTATUS_OK;
}

/******************************************************************************
 **
 **  ljkOS_DestroySemaphore
 **
 **  Destroy a semaphore.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to the ljkOS object.
 **
 **      ljtPOINTER Semaphore
 **          Pointer to the semaphore thet needs to be destroyed.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_DestroySemaphore(ljkOS Os, ljtPOINTER Semaphore)
{
    ljmkHEADER_ARG("Os=%p Semaphore=%p", Os, Semaphore);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Semaphore != ljvNULL);

    /* Free the sempahore structure. */
    kfree(Semaphore);

    /* Success. */
    ljmkFOOTER_NO();
    return ljvSTATUS_OK;
}

/******************************************************************************
 **
 **  ljkOS_GetProcessID
 **
 **  Get current process ID.
 **
 **  INPUT:
 **
 **      Nothing.
 **
 **  OUTPUT:
 **
 **      ljtUINT32_PTR ProcessID
 **          Pointer to the variable that receives the process ID.
 */
ljeSTATUS
ljkOS_GetProcessID(ljtUINT32_PTR ProcessID)
{
    /* Get process ID. */
    *ProcessID = _GetProcessID();

    /* Success. */
    return ljvSTATUS_OK;
}

/*******************************************************************************
 **
 **  ljkOS_GetThreadID
 **
 **  Get current thread ID.
 **
 **  INPUT:
 **
 **      Nothing.
 **
 **  OUTPUT:
 **
 **      ljtUINT32_PTR ThreadID
 **          Pointer to the variable that receives the thread ID.
 */
ljeSTATUS
ljkOS_GetThreadID(ljtUINT32_PTR ThreadID)
{
    /* Get thread ID. */
    if (ThreadID != ljvNULL)
        *ThreadID = _GetThreadID();

    /* Success. */
    return ljvSTATUS_OK;
}

/******************************************************************************
 **
 **  ljkOS_SetClockState
 **
 **  Set the clock state on or off.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to a ljkOS object.
 **
 **      ljKERNEL Kernel
 **          Kernel object whose power is set.
 **
 **      ljeCORE Core
 **          GPU whose power is set.
 **
 **      ljtBOOL Clock
 **          ljvTRUE to turn on the clock, or ljvFALSE to turn off the clock.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_SetClockState(ljkOS Os, ljkKERNEL Kernel, ljtBOOL Clock)
{
    ljtBOOL clockChange = ljvFALSE;
    ljkDEVICE device;

    ljmkHEADER_ARG("Os=%p Kernel=%p Clock=%d", Os, Kernel, Clock);

    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Kernel != ljvNULL);

    device = Kernel->device;

    clockChange = (Clock != Os->clockStates[device->id][Kernel->core]);

    if (clockChange) {
        unsigned long flags;

        if (!Clock) {
            spin_lock_irqsave(&Os->registerAccessLock, flags);

            /* Record clock off, ahead. */
            Os->clockStates[device->id][Kernel->core] = ljvFALSE;

            spin_unlock_irqrestore(&Os->registerAccessLock, flags);
        }

        if (Clock) {
            spin_lock_irqsave(&Os->registerAccessLock, flags);

            /* Record clock on, behind. */
            Os->clockStates[device->id][Kernel->core] = ljvTRUE;

            spin_unlock_irqrestore(&Os->registerAccessLock, flags);
        }
    }

    ljmkFOOTER_NO();
    return ljvSTATUS_OK;
}

/******************************************************************************
 **
 **  ljkOS_GetClockState
 **
 **  Get the clock state on or off.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to a ljkOS object.
 **
 **      ljkKERNEL Kernel
 **          Kernel obj whose power is get.
 **
 **      ljtBOOL Clock
 **          ljvTRUE to turn on the clock, or ljvFALSE to turn off the clock.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_GetClockState(ljkOS Os, ljkKERNEL Kernel, ljtBOOL *Clock)
{
    ljkDEVICE device = Kernel->device;

    *Clock = Os->clockStates[device->id][Kernel->core];

    return ljvSTATUS_OK;
}

/******************************************************************************
 **
 **  ljkOS_SetGPUPower
 **
 **  Set the power of the GPU on or off.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to a ljkOS object.
 **
 **      ljkKERNEL Kernel
 **          Core whose power is set.
 **
 **      ljtBOOL Clock
 **          ljvTRUE to turn on the clock, or ljvFALSE to turn off the clock.
 **
 **      ljtBOOL Power
 **          ljvTRUE to turn on the power, or ljvFALSE to turn off the power.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_SetGPUPower(ljkOS Os, ljkKERNEL Kernel, ljtBOOL Clock, ljtBOOL Power)
{
    ljsPLATFORM *platform;
    ljkDEVICE device;

    ljtBOOL powerChange = ljvFALSE;
    ljtBOOL clockChange = ljvFALSE;

    ljmkHEADER_ARG("Os=%p Kernel=%p Clock=%d Power=%d", Os, Kernel, Clock, Power);
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Kernel != ljvNULL);

    platform = Os->device->platform;
    device = Kernel->device;

    ljmkASSERT(device->id < ljdDEVICE_COUNT);
    ljmkASSERT(Kernel->core < ljdMAX_GPU_COUNT);

    powerChange = (Power != Os->powerStates[device->id][Kernel->core]);

    clockChange = (Clock != Os->clockStates[device->id][Kernel->core]);

    if (powerChange && (Power == ljvTRUE)) {
        if (platform && platform->ops->setPower)
            ljmkVERIFY_OK(platform->ops->setPower(platform, device->id, Kernel->core, Power));

        Os->powerStates[device->id][Kernel->core] = Power;
    }

    if (clockChange) {
        unsigned long flags;

        if (!Clock) {
            spin_lock_irqsave(&Os->registerAccessLock, flags);

            /* Record clock off, ahead. */
            Os->clockStates[device->id][Kernel->core] = ljvFALSE;

            spin_unlock_irqrestore(&Os->registerAccessLock, flags);
        }

        if (platform && platform->ops->setClock)
            ljmkVERIFY_OK(platform->ops->setClock(platform, device->id, Kernel->core, Clock));

        if (Clock) {
            spin_lock_irqsave(&Os->registerAccessLock, flags);

            /* Record clock on, behind. */
            Os->clockStates[device->id][Kernel->core] = ljvTRUE;

            spin_unlock_irqrestore(&Os->registerAccessLock, flags);
        }
    }

    if (powerChange && (Power == ljvFALSE)) {
        if (platform && platform->ops->setPower)
            ljmkVERIFY_OK(platform->ops->setPower(platform, device->id, Kernel->core, Power));

        Os->powerStates[device->id][Kernel->core] = Power;
    }

#if ljdENABLE_AHBXTTD
    if (Kernel->core == ljvCORE_MAJOR && Clock == ljvTRUE && Power == ljvTRUE)
        ljmkVERIFY_OK(ljkHARDWARE_ResetAHBXTTD(ljvNULL, Kernel));
#endif

    ljmkFOOTER_NO();
    return ljvSTATUS_OK;
}

/******************************************************************************
 **
 **  ljkOS_ResetGPU
 **
 **  Reset the GPU.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to a ljkOS object.
 **
 **      ljkKERNEL Kernel
 **          Core whose power is set.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_ResetGPU(ljkOS Os, ljkKERNEL Kernel)
{
    ljeSTATUS status = ljvSTATUS_NOT_SUPPORTED;
    ljsPLATFORM *platform;
    ljkDEVICE device;

    ljmkHEADER_ARG("Os=%p Kernel=%p", Os, Kernel);
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);

    platform = Os->device->platform;
    device = Kernel->device;

    if (platform && platform->ops->reset)
        status = platform->ops->reset(platform, device->id, Kernel->core);

    ljmkFOOTER_NO();
    return status;
}

/******************************************************************************
 **
 **  ljkOS_PrepareGPUFrequency
 **
 **  Prepare to set GPU frequency and voltage.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to a ljkOS object.
 **
 **      ljkCORE Core
 **          GPU whose frequency and voltage will be set.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_PrepareGPUFrequency(ljkOS Os, ljeCORE Core)
{
    return ljvSTATUS_OK;
}

/******************************************************************************
 **
 **  ljkOS_FinishGPUFrequency
 **
 **  Finish GPU frequency setting.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to a ljkOS object.
 **
 **      ljkCORE Core
 **          GPU whose frequency and voltage is set.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_FinishGPUFrequency(ljkOS Os, ljeCORE Core)
{
    return ljvSTATUS_OK;
}

/******************************************************************************
 **
 **  ljkOS_QueryGPUFrequency
 **
 **  Query the current frequency of the GPU.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to a ljkOS object.
 **
 **      ljkCORE Core
 **          GPU whose power is set.
 **
 **      ljtUINT32 *Frequency
 **          Pointer to a ljtUINT32 to obtain current frequency, in MHz.
 **
 **      ljtUINT8 *Scale
 **          Pointer to a ljtUINT8 to obtain current scale(1 - 64).
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_QueryGPUFrequency(ljkOS Os, ljkHARDWARE Hardware, ljtUINT32 *Frequency, ljtUINT8 *Scale)
{
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkHEADER();

    ljmkONERROR(ljkHARDWARE_QueryFrequency(Hardware));

    *Frequency = (ljtUINT32)(Hardware->mcClk / 1024 / 1024);

    ljmkONERROR(ljkHARDWARE_QueryFScale(Hardware, Scale));

    ljmkFOOTER_NO();
    return status;

OnError:
    ljmkFOOTER_NO();
    return status;
}

/******************************************************************************
 **
 **  ljkOS_SetGPUFrequency
 **
 **  Set frequency and voltage of the GPU.
 **
 **      1. DVFS manager gives the target scale of full frequency, BSP must
 **         find a real frequency according to this scale and board's
 **         configure.
 **
 **      2. BSP should find a suitable voltage for this frequency.
 **
 **      3. BSP must make sure setting take effect before this function
 **         returns.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to a ljkOS object.
 **
 **      ljkCORE Core
 **          GPU whose power is set.
 **
 **      ljtUINT8 Scale
 **          Target scale of full frequency, range is [1, 64]. 1 means 1/64 of
 **          full frequency and 64 means 64/64 of full frequency.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_SetGPUFrequency(ljkOS Os, ljkHARDWARE Hardware, ljtUINT8 Scale)
{
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkHEADER();

    ljmkONERROR(ljkHARDWARE_SetClock(Hardware, Scale, Scale));

    ljmkFOOTER_NO();
    return status;

OnError:
    ljmkFOOTER_NO();
    return status;
}

/******************************************************************************
 ******************************** Profile *************************************/

ljeSTATUS
ljkOS_GetProfileTick(ljtUINT64_PTR Tick)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 5, 0)
    struct timespec64 time;

    ktime_get_ts64(&time);
#else
    struct timespec time;

    ktime_get_ts(&time);
#endif
    *Tick = time.tv_nsec + time.tv_sec * 1000000000ULL;

    return ljvSTATUS_OK;
}

ljeSTATUS
ljkOS_QueryProfileTickRate(ljtUINT64_PTR TickRate)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 5, 0)
    struct timespec64 res;
#else
    struct timespec res;
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 2, 0)
    res.tv_sec = 0;
    res.tv_nsec = hrtimer_resolution;
#else
    hrtimer_get_res(CLOCK_MONOTONIC, &res);
#endif

    *TickRate = res.tv_nsec + res.tv_sec * 1000000000ULL;

    return ljvSTATUS_OK;
}

ljtUINT32
ljkOS_ProfileToMS(ljtUINT64 Ticks)
{
#if LINUX_VERSION_CODE > KERNEL_VERSION(2, 6, 23)
    return div_u64(Ticks, 1000000);
#else
    ljtUINT64 rem = Ticks;
    ljtUINT64 b = 1000000;
    ljtUINT64 res, d = 1;
    ljtUINT32 high = rem >> 32;

    /* Reduce the thing a bit first */
    res = 0;
    if (high >= 1000000) {
        high /= 1000000;
        res = (ljtUINT64)high << 32;
        rem -= (ljtUINT64)(high * 1000000) << 32;
    }

    while (((ljtINT64)b > 0) && (b < rem)) {
        b <<= 1;
        d <<= 1;
    }

    do {
        if (rem >= b) {
            rem -= b;
            res += d;
        }

        b >>= 1;
        d >>= 1;
    } while (d);

    return (ljtUINT32)res;
#endif
}

/******************************************************************************
 ******************************* Signal Management ****************************
 *****************************************************************************/

#undef _LJ_OBJ_ZONE
#define _LJ_OBJ_ZONE ljvZONE_SIGNAL

/******************************************************************************
 **
 **  ljkOS_CreateSignal
 **
 **  Create a new signal.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtBOOL ManualReset
 **          If set to ljvTRUE, ljkOS_Signal with ljvFALSE must be called in
 **          order to set the signal to nonsignaled state.
 **          If set to ljvFALSE, the signal will automatically be set to
 **          nonsignaled state by ljkOS_WaitSignal function.
 **
 **  OUTPUT:
 **
 **      ljtSIGNAL *Signal
 **          Pointer to a variable receiving the created ljtSIGNAL.
 */
ljeSTATUS
ljkOS_CreateSignal(ljkOS Os, ljtBOOL ManualReset, ljtSIGNAL *Signal)
{
    ljeSTATUS status = ljvSTATUS_OK;
    ljsSIGNAL_PTR signal = ljvNULL;

    ljmkHEADER_ARG("Os=%p ManualReset=%d", Os, ManualReset);
    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Signal != ljvNULL);

    /* Create an event structure. */
    signal = (ljsSIGNAL_PTR)kmalloc(sizeof(ljsSIGNAL), GFP_KERNEL | ljdNOWARN);

    if (signal == ljvNULL)
        ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);

    /* Save the process ID. */
    signal->process = (ljtHANDLE)(ljtUINTPTR_T)_GetProcessID();

    signal->done = 0;
    init_waitqueue_head(&signal->wait);
    spin_lock_init(&signal->lock);
    signal->manualReset = ManualReset;
    signal->status = ljvSIGNAL_OK;

    atomic_set(&signal->ref, 1);

#if ljdLINUX_SYNC_FILE
#ifndef CONFIG_SYNC_FILE
    signal->timeline = ljvNULL;
# else
    signal->fence = ljvNULL;
# endif
#endif

    ljmkONERROR(_AllocateIntegerId(&Os->signalDB, signal, &signal->id));

    *Signal = (ljtSIGNAL)(ljtUINTPTR_T)signal->id;

OnError:
    if (ljmIS_ERROR(status) && signal)
        kfree(signal);

    ljmkFOOTER_ARG("*Signal=%p", *Signal);
    return status;
}

/******************************************************************************
 **
 **  ljkOS_DestroySignal
 **
 **  Destroy a signal.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtSIGNAL Signal
 **          Pointer to the ljtSIGNAL.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_DestroySignal(ljkOS Os, ljtSIGNAL Signal)
{
    ljeSTATUS status = ljvSTATUS_OK;
    ljsSIGNAL_PTR signal;
    ljtBOOL acquired = ljvFALSE;
    unsigned long flags = 0;

    ljmkHEADER_ARG("Os=%p Signal=%p", Os, Signal);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Signal != ljvNULL);

    if (in_irq())
        spin_lock(&Os->signalLock);
    else
        spin_lock_irqsave(&Os->signalLock, flags);

    acquired = ljvTRUE;

    ljmkONERROR(_QueryIntegerId(&Os->signalDB,
                                (ljtUINT32)(ljtUINTPTR_T)Signal,
                                (ljtPOINTER)&signal));

    ljmkASSERT(signal->id == (ljtUINT32)(ljtUINTPTR_T)Signal);

    if (atomic_dec_and_test(&signal->ref)) {
        ljmkVERIFY_OK(_DestroyIntegerId(&Os->signalDB, signal->id));

        /* Free the sgianl. */
        kfree(signal);
    }

    if (in_irq())
        spin_unlock(&Os->signalLock);
    else
        spin_unlock_irqrestore(&Os->signalLock, flags);

    acquired = ljvFALSE;

OnError:
    if (acquired) {
        /* Release the mutex. */
        if (in_irq())
            spin_unlock(&Os->signalLock);
        else
            spin_unlock_irqrestore(&Os->signalLock, flags);
    }

    ljmkFOOTER();
    return status;
}

/******************************************************************************
 **
 **  ljkOS_Signal
 **
 **  Set a state of the specified signal.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtSIGNAL Signal
 **          Pointer to the ljtSIGNAL.
 **
 **      ljtBOOL State
 **          If ljvTRUE, the signal will be set to signaled state.
 **          If ljvFALSE, the signal will be set to nonsignaled state.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_Signal(ljkOS Os, ljtSIGNAL Signal, ljtBOOL State)
{
    ljeSTATUS status;
    ljsSIGNAL_PTR signal;
#if ljdLINUX_SYNC_FILE
#ifndef CONFIG_SYNC_FILE
    struct sync_timeline *timeline = ljvNULL;
# else
    struct dma_fence *fence = ljvNULL;
# endif
#endif
    unsigned long flags = 0;

    ljmkHEADER_ARG("Os=%p Signal=%p State=%d", Os, Signal, State);
    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Signal != ljvNULL);

    spin_lock_irqsave(&Os->signalLock, flags);

    status = _QueryIntegerId(&Os->signalDB,
                             (ljtUINT32)(ljtUINTPTR_T)Signal,
                             (ljtPOINTER)&signal);

    if (ljmIS_ERROR(status)) {
        spin_unlock_irqrestore(&Os->signalLock, flags);
        ljmkONERROR(status);
    }

    /*
     * Signal saved in event is not referenced. Inc reference here to avoid
     * concurrent issue: signaling the signal while another thread is
     * destroying it.
     */
    atomic_inc(&signal->ref);

    spin_unlock_irqrestore(&Os->signalLock, flags);

    ljmkONERROR(status);

    ljmkASSERT(signal->id == (ljtUINT32)(ljtUINTPTR_T)Signal);

    spin_lock(&signal->lock);

    if (State) {
        signal->done = 1;

        wake_up(&signal->wait);

#if ljdLINUX_SYNC_FILE
#ifndef CONFIG_SYNC_FILE
        timeline = signal->timeline;
# else
        fence = signal->fence;
        signal->fence = NULL;
# endif
#endif
    } else {
        signal->done  = 0;
    }

    spin_unlock(&signal->lock);

#if ljdLINUX_SYNC_FILE
#ifndef CONFIG_SYNC_FILE
    /* Signal timeline. */
    if (timeline)
        sync_timeline_signal(timeline);
# else
    if (fence) {
        dma_fence_signal(fence);
        dma_fence_put(fence);
    }
# endif
#endif

    spin_lock_irqsave(&Os->signalLock, flags);

    if (atomic_dec_and_test(&signal->ref)) {
        ljmkVERIFY_OK(_DestroyIntegerId(&Os->signalDB, signal->id));

        /* Free the sgianl. */
        kfree(signal);
    }

    spin_unlock_irqrestore(&Os->signalLock, flags);

OnError:
    ljmkFOOTER();
    return status;
}

/******************************************************************************
 **
 **  ljkOS_UserSignal
 **
 **  Set the specified signal which is owned by a process to signaled state.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtSIGNAL Signal
 **          Pointer to the ljtSIGNAL.
 **
 **      ljtHANDLE Handle
 **          Handle of process owning the signal.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_UserSignal(ljkOS Os, ljtSIGNAL Signal, ljtHANDLE Handle)
{
    ljeSTATUS status;
    ljsSIGNAL_PTR signal;

    ljmkHEADER_ARG("Os=%p Signal=%p Process=%p", Os, Signal, Handle);

    if (!Handle || Handle == ljvINVALID_HANDLE) {
        status = _QueryIntegerId(&Os->signalDB,
                                 (ljtUINT32)(ljtUINTPTR_T)Signal,
                                 (ljtPOINTER)&signal);

        if (ljmIS_ERROR(status)) {
            ljmkFOOTER();
            return status;
        }

        if (!Handle)
            signal->status = ljvSIGNAL_RECOVERY;
        else
            signal->status = ljvSIGNAL_CANCEL;
    }

    /* Signal. */
    status = ljkOS_Signal(Os, Signal, ljvTRUE);

    ljmkFOOTER();
    return status;
}

/******************************************************************************
 **
 **  ljkOS_WaitSignal
 **
 **  Wait for a signal to become signaled.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtSIGNAL Signal
 **          Pointer to the ljtSIGNAL.
 **
 **      ljtUINT32 Wait
 **          Number of milliseconds to wait.
 **          Pass the value of ljvINFINITE for an infinite wait.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_WaitSignal(ljkOS Os, ljtSIGNAL Signal, ljtBOOL Interruptable, ljtUINT32 Wait)
{
    ljeSTATUS status;
    ljsSIGNAL_PTR signal = ljvNULL;
    int done;

    ljmkHEADER_ARG("Os=%p Signal=%p Wait=0x%08X", Os, Signal, Wait);
    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Signal != ljvNULL);

    ljmkONERROR(_QueryIntegerId(&Os->signalDB,
                                (ljtUINT32)(ljtUINTPTR_T)Signal,
                                (ljtPOINTER)&signal));

    ljmkASSERT(signal->id == (ljtUINT32)(ljtUINTPTR_T)Signal);

    spin_lock(&signal->lock);
    done = signal->done;
    spin_unlock(&signal->lock);

    /*
     * Do not need to lock below:
     * 1. If signal already done, return immediately.
     * 2. If signal not done, wait_event_xxx will handle correctly even
     *    read of signal->done is not atomic.
     *
     * Rest signal->done do not require lock either:
     * No other thread can query/wait auto-reseted signal, because that is
     * logic error.
     */
    if (done) {
        status = ljvSTATUS_OK;

        if (!signal->manualReset)
            signal->done = 0;
    } else if (Wait == 0) {
        status = ljvSTATUS_TIMEOUT;
    } else {
        /* Convert wait to milliseconds. */
        long timeout = (Wait == ljvINFINITE) ?
                       MAX_SCHEDULE_TIMEOUT : msecs_to_jiffies(Wait);

        long ret;

        if (Interruptable)
            ret = wait_event_interruptible_timeout(signal->wait, signal->done, timeout);
        else
            ret = wait_event_timeout(signal->wait, signal->done, timeout);

        if (likely(ret > 0)) {
            status = ljvSTATUS_OK;

            if (!signal->manualReset)
                /* Auto reset. */
                signal->done = 0;
        } else {
            status = (ret == -ERESTARTSYS) ?
                     ljvSTATUS_INTERRUPTED : ljvSTATUS_TIMEOUT;
        }
    }

OnError:
    /* Return status. */
    ljmkFOOTER();
    return status;
}

ljeSTATUS
_QuerySignal(ljkOS Os, ljtSIGNAL Signal)
{
    /*
     * This function is called by 'has_signaled' callback of sync_timeline.
     * By design, 'has_signaled' could be called in interrupt context, but
     * in current driver, it can be called only when 'ljkOS_Signal' and
     * 'ljkOS_CreateNativeFence'. Thus its safe to use normal version of
     * spinlock for 'Os->signalDB.lock' and 'signal->obj.wait.lock'.
     */
    ljeSTATUS status;
    ljsSIGNAL_PTR signal = ljvNULL;

    status = _QueryIntegerId(&Os->signalDB,
                             (ljtUINT32)(ljtUINTPTR_T)Signal,
                             (ljtPOINTER)&signal);

    if (ljmIS_SUCCESS(status)) {
        spin_lock(&signal->lock);
        status = signal->done ? ljvSTATUS_TRUE : ljvSTATUS_FALSE;
        spin_unlock(&signal->lock);
    }

    return status;
}

char *_GetProcessComm(void)
{
    static char name[TASK_COMM_LEN];

    get_task_comm(name, current);
    return name;
}

/******************************************************************************
 **
 **  ljkOS_MapSignal
 **
 **  Map a signal in to the current process space.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtSIGNAL Signal
 **          Pointer to tha ljtSIGNAL to map.
 **
 **      ljtHANDLE Process
 **          Handle of process owning the signal.
 **
 **  OUTPUT:
 **
 **      ljtSIGNAL *MappedSignal
 **          Pointer to a variable receiving the mapped ljtSIGNAL.
 */
ljeSTATUS
ljkOS_MapSignal(ljkOS Os, ljtSIGNAL Signal,
                ljtHANDLE Process, ljtSIGNAL *MappedSignal)
{
    ljeSTATUS status = ljvSTATUS_OK;
    ljsSIGNAL_PTR signal = ljvNULL;
    unsigned long flags = 0;

    ljmkHEADER_ARG("Os=%p Signal=%p Process=%p", Os, Signal, Process);
    ljmkVERIFY_ARGUMENT(Signal != ljvNULL);
    ljmkVERIFY_ARGUMENT(MappedSignal != ljvNULL);

    spin_lock_irqsave(&Os->signalLock, flags);

    ljmkONERROR(_QueryIntegerId(&Os->signalDB,
                                (ljtUINT32)(ljtUINTPTR_T)Signal,
                                (ljtPOINTER)&signal));

    if (atomic_inc_return(&signal->ref) <= 1)
        /* The previous value is 0, it has been deleted. */
        ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);

    *MappedSignal = (ljtSIGNAL)Signal;

OnError:
    spin_unlock_irqrestore(&Os->signalLock, flags);

    ljmkFOOTER_ARG("*MappedSignal=%p", *MappedSignal);
    return status;
}

/******************************************************************************
 **
 **  ljkOS_UnmapSignal
 **
 **  Unmap a signal .
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtSIGNAL Signal
 **          Pointer to that ljtSIGNAL mapped.
 */
ljeSTATUS
ljkOS_UnmapSignal(ljkOS Os, ljtSIGNAL Signal)
{
    return ljkOS_DestroySignal(Os, Signal);
}

/******************************************************************************
 **
 **  ljkOS_CreateUserSignal
 **
 **  Create a new signal to be used in the user space.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtBOOL ManualReset
 **          If set to ljvTRUE, ljkOS_Signal with ljvFALSE must be called in
 **          order to set the signal to nonsignaled state.
 **          If set to ljvFALSE, the signal will automatically be set to
 **          nonsignaled state by ljkOS_WaitSignal function.
 **
 **  OUTPUT:
 **
 **      ljtINT * SignalID
 **          Pointer to a variable receiving the created signal's ID.
 */
ljeSTATUS
ljkOS_CreateUserSignal(ljkOS Os, ljtBOOL ManualReset, ljtINT *SignalID)
{
    ljeSTATUS status;
    ljtSIZE_T signal = 0;

    /* Create a new signal. */
    ljmkONERROR(ljkOS_CreateSignal(Os, ManualReset, (ljtSIGNAL *)&signal));
    *SignalID = (ljtINT)signal;

OnError:
    return status;
}

/******************************************************************************
 **
 **  ljkOS_DestroyUserSignal
 **
 **  Destroy a signal to be used in the user space.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtINT SignalID
 **          The signal's ID.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_DestroyUserSignal(ljkOS Os, ljtINT SignalID)
{
    return ljkOS_DestroySignal(Os, (ljtSIGNAL)(ljtUINTPTR_T)SignalID);
}

/******************************************************************************
 **
 **  ljkOS_WaitUserSignal
 **
 **  Wait for a signal used in the user mode to become signaled.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtINT SignalID
 **          Signal ID.
 **
 **      ljtUINT32 Wait
 **          Number of milliseconds to wait.
 **          Pass the value of ljvINFINITE for an infinite wait.
 **
 **  OUTPUT:
 **
 **      ljeSIGNAL_STATUS *SignalStatus
 **          Return signal status.
 **
 */
ljeSTATUS
ljkOS_WaitUserSignal(ljkOS Os, ljtINT SignalID,
                     ljtUINT32 Wait, ljeSIGNAL_STATUS *SignalStatus)
{
    ljeSTATUS status;
    ljsSIGNAL_PTR signal;

    status = ljkOS_WaitSignal(Os, (ljtSIGNAL)(ljtUINTPTR_T)SignalID, ljvTRUE, Wait);

    if (ljmIS_SUCCESS(status)) {
        if (SignalStatus) {
            status = _QueryIntegerId(&Os->signalDB, SignalID, (ljtPOINTER)&signal);

            if (ljmIS_SUCCESS(status)) {
                *SignalStatus = signal->status;

                /* Clear the signal status. */
                signal->status = ljvSIGNAL_OK;
            }
        }
    }

    return status;
}

/******************************************************************************
 **
 **  ljkOS_SignalUserSignal
 **
 **  Set a state of the specified signal to be used in the user space.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtINT SignalID
 **          SignalID.
 **
 **      ljtBOOL State
 **          If ljvTRUE, the signal will be set to signaled state.
 **          If ljvFALSE, the signal will be set to nonsignaled state.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_SignalUserSignal(ljkOS Os, ljtINT SignalID, ljtBOOL State)
{
    return ljkOS_Signal(Os, (ljtSIGNAL)(ljtUINTPTR_T)SignalID, State);
}

ljeSTATUS
ljkOS_CreateSemaphoreEx(ljkOS Os, ljtSEMAPHORE *Semaphore)
{
    ljeSTATUS status;
    struct semaphore *newSemaphore;

    ljmkHEADER_ARG("Os=%p Semaphore=%p", Os, Semaphore);
    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Semaphore != ljvNULL);

    do {
        /* Allocate the semaphore structure. */
        newSemaphore = kmalloc(ljmSIZEOF(struct semaphore), GFP_KERNEL | ljdNOWARN);

        if (newSemaphore == ljvNULL)
            ljmkERR_BREAK(ljvSTATUS_OUT_OF_MEMORY);

        /* Initialize the semaphore. */
        sema_init(newSemaphore, 0);

        /* Set the handle. */
        *Semaphore = (ljtSEMAPHORE)newSemaphore;

        /* Success. */
        status = ljvSTATUS_OK;
    } while (ljvFALSE);

    ljmkFOOTER();
    /* Return the status. */
    return status;
}

#if ljdENABLE_VG
ljeSTATUS
ljkOS_CreateSemaphoreVG(ljkOS Os, ljtSEMAPHORE *Semaphore)
{
    ljeSTATUS status;
    struct semaphore *newSemaphore;

    ljmkHEADER_ARG("Os=%p Semaphore=%p", Os, Semaphore);
    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Semaphore != ljvNULL);

    do {
        /* Allocate the semaphore structure. */
        newSemaphore = kmalloc(ljmSIZEOF(struct semaphore), GFP_KERNEL | ljdNOWARN);
        if (newSemaphore == ljvNULL)
            ljmkERR_BREAK(ljvSTATUS_OUT_OF_MEMORY);

        /* Initialize the semaphore. */
        sema_init(newSemaphore, 0);

        /* Set the handle. */
        *Semaphore = (ljtSEMAPHORE)newSemaphore;

        /* Success. */
        status = ljvSTATUS_OK;
    } while (ljvFALSE);

    ljmkFOOTER();
    /* Return the status. */
    return status;
}

ljeSTATUS
ljkOS_IncrementSemaphore(ljkOS Os, ljtSEMAPHORE Semaphore)
{
    ljmkHEADER_ARG("Os=%p Semaphore=%p", Os, Semaphore);
    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Semaphore != ljvNULL);

    /* Increment the semaphore's count. */
    up((struct semaphore *)Semaphore);

    ljmkFOOTER_NO();
    /* Success. */
    return ljvSTATUS_OK;
}

ljeSTATUS
ljkOS_DecrementSemaphore(ljkOS Os, ljtSEMAPHORE Semaphore)
{
    ljeSTATUS status;
    ljtINT result;

    ljmkHEADER_ARG("Os=%p Semaphore=%p", Os, Semaphore);
    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Semaphore != ljvNULL);

    do {
        /*
         * Decrement the semaphore's count. If the count is zero, wait
         * until it gets incremented.
         */
        result = down_interruptible((struct semaphore *)Semaphore);

        /* Signal received? */
        if (result != 0) {
            status = ljvSTATUS_TERMINATE;
            break;
        }

        /* Success. */
        status = ljvSTATUS_OK;
    } while (ljvFALSE);

    ljmkFOOTER();
    /* Return the status. */
    return status;
}

/******************************************************************************
 ******************************** Thread Object *******************************
 *****************************************************************************/

ljeSTATUS
ljkOS_StartThread(ljkOS Os,
                  ljtTHREADFUNC ThreadFunction,
                  ljtPOINTER ThreadParameter,
                  ljtTHREAD *Thread)
{
    ljeSTATUS status;
    struct task_struct *thread;

    ljmkHEADER_ARG("Os=%p", Os);
    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(ThreadFunction != ljvNULL);
    ljmkVERIFY_ARGUMENT(Thread != ljvNULL);

    do {
        /* Create the thread. */
        thread = kthread_create(ThreadFunction, ThreadParameter, "Ljmicro Kernel Thread");

        /* Failed? */
        if (IS_ERR(thread)) {
            status = ljvSTATUS_GENERIC_IO;
            break;
        }

        /* Start the thread. */
        wake_up_process(thread);

        /* Set the thread handle. */
        *Thread = (ljtTHREAD)thread;

        /* Success. */
        status = ljvSTATUS_OK;
    } while (ljvFALSE);

    ljmkFOOTER();
    /* Return the status. */
    return status;
}

ljeSTATUS
ljkOS_StopThread(ljkOS Os, ljtTHREAD Thread)
{
    ljmkHEADER_ARG("Os=%p Thread=%p", Os, Thread);
    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Thread != ljvNULL);

    /* Thread should have already been enabled to terminate. */
    kthread_stop((struct task_struct *)Thread);

    ljmkFOOTER_NO();
    /* Success. */
    return ljvSTATUS_OK;
}

ljeSTATUS
ljkOS_VerifyThread(ljkOS Os, ljtTHREAD Thread)
{
    ljmkHEADER_ARG("Os=%p Thread=%p", Os, Thread);
    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Thread != ljvNULL);

    ljmkFOOTER_NO();
    /* Success. */
    return ljvSTATUS_OK;
}
#endif

/******************************************************************************
 ******************************** Software Timer ******************************
 *****************************************************************************/

void
_TimerFunction(struct work_struct *work)
{
    ljsOSTIMER_PTR timer = (ljsOSTIMER_PTR)work;

    ljtTIMERFUNCTION function = timer->function;

    function(timer->data);
}

/******************************************************************************
 **
 **  ljkOS_CreateTimer
 **
 **  Create a software timer.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to the ljkOS object.
 **
 **      ljtTIMERFUNCTION Function.
 **          Pointer to a call back function which will be called when timer is
 **          expired.
 **
 **      ljtPOINTER Data.
 **          Private data which will be passed to call back function.
 **
 **  OUTPUT:
 **
 **      ljtPOINTER *Timer
 **          Pointer to a variable receiving the created timer.
 */
ljeSTATUS
ljkOS_CreateTimer(ljkOS Os,
                  ljtTIMERFUNCTION Function,
                  ljtPOINTER Data,
                  ljtPOINTER *Timer)
{
    ljeSTATUS status = ljvSTATUS_OK;
    ljsOSTIMER_PTR pointer;

    ljmkHEADER_ARG("Os=%p Function=0%p Data=%p", Os, Function, Data);
    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Timer != ljvNULL);

    ljmkONERROR(ljkOS_Allocate(Os, sizeof(ljsOSTIMER), (ljtPOINTER)&pointer));

    pointer->function = Function;
    pointer->data = Data;

    INIT_DELAYED_WORK(&pointer->work, _TimerFunction);

    *Timer = pointer;

OnError:
    ljmkFOOTER();
    return status;
}

/******************************************************************************
 **
 **  ljkOS_DestroyTimer
 **
 **  Destroy a software timer.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to the ljkOS object.
 **
 **      ljtPOINTER Timer
 **          Pointer to the timer to be destroyed.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_DestroyTimer(ljkOS Os, ljtPOINTER Timer)
{
    ljsOSTIMER_PTR timer;

    ljmkHEADER_ARG("Os=%p Timer=%p", Os, Timer);
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Timer != ljvNULL);

    timer = (ljsOSTIMER_PTR)Timer;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 23)
    cancel_delayed_work_sync(&timer->work);
#else
    cancel_delayed_work(&timer->work);
    flush_workqueue(Os->workqueue);
#endif

    ljmkVERIFY_OK(ljmkOS_SAFE_FREE(Os, Timer));

    ljmkFOOTER_NO();
    return ljvSTATUS_OK;
}

/******************************************************************************
 **
 **  ljkOS_StartTimer
 **
 **  Schedule a software timer.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to the ljkOS object.
 **
 **      ljtPOINTER Timer
 **          Pointer to the timer to be scheduled.
 **
 **      ljtUINT32 Delay
 **          Delay in milliseconds.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_StartTimer(ljkOS Os, ljtPOINTER Timer, ljtUINT32 Delay)
{
    ljsOSTIMER_PTR timer;

    ljmkHEADER_ARG("Os=%p Timer=%p Delay=%u", Os, Timer, Delay);

    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Timer != ljvNULL);
    ljmkVERIFY_ARGUMENT(Delay != 0);

    timer = (ljsOSTIMER_PTR)Timer;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 7, 0)
    mod_delayed_work(Os->workqueue, &timer->work, msecs_to_jiffies(Delay));
#else
    if (unlikely(delayed_work_pending(&timer->work))) {
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 23)
        cancel_delayed_work_sync(&timer->work);
# else
        cancel_delayed_work(&timer->work);
        flush_workqueue(Os->workqueue);
# endif
    }

    queue_delayed_work(Os->workqueue, &timer->work, msecs_to_jiffies(Delay));
#endif

    ljmkFOOTER_NO();
    return ljvSTATUS_OK;
}

/******************************************************************************
 **
 **  ljkOS_StopTimer
 **
 **  Cancel a unscheduled timer.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to the ljkOS object.
 **
 **      ljtPOINTER Timer
 **          Pointer to the timer to be cancel.
 **
 **  OUTPUT:
 **
 **      Nothing.
 */
ljeSTATUS
ljkOS_StopTimer(ljkOS Os, ljtPOINTER Timer)
{
    ljsOSTIMER_PTR timer;

    ljmkHEADER_ARG("Os=%p Timer=%p", Os, Timer);
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Timer != ljvNULL);

    timer = (ljsOSTIMER_PTR)Timer;

    cancel_delayed_work(&timer->work);

    ljmkFOOTER_NO();
    return ljvSTATUS_OK;
}

ljeSTATUS
ljkOS_GetProcessNameByPid(ljtINT Pid, ljtSIZE_T Length, ljtUINT8_PTR String)
{
    struct task_struct *task;
    ljeSTATUS status = ljvSTATUS_OK;

    /* Get the task_struct of the task with pid. */
    rcu_read_lock();

    task = FIND_TASK_BY_PID(Pid);
    if (task)
        /* Get name of process. */
        strncpy(String, task->comm, Length);
    else
        status = ljvSTATUS_NOT_FOUND;

    rcu_read_unlock();

    return status;
}

ljeSTATUS
ljkOS_DumpCallStack(ljkOS Os)
{
    ljmkHEADER_ARG("Os=%p", Os);

    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);

    dump_stack();

    ljmkFOOTER_NO();
    return ljvSTATUS_OK;
}

/******************************************************************************
 **
 **  ljkOS_DetectProcessByName
 **
 **      task->comm maybe part of process name, so this function
 **      can only be used for debugging.
 **
 **  INPUT:
 **
 **      ljtCONST_POINTER Name
 **          Pointer to a string to hold name to be check. If the length
 **          of name is longer than TASK_COMM_LEN (16), use part of name
 **          to detect.
 **
 **  OUTPUT:
 **
 **      ljvSTATUS_TRUE if name of current process matches Name.
 **
 */
ljeSTATUS
ljkOS_DetectProcessByName(ljtCONST_POINTER Name)
{
    char comm[sizeof(current->comm)];

    memset(comm, 0, sizeof(comm));

    ljmkVERIFY_OK(ljkOS_GetProcessNameByPid(_GetProcessID(),
                                            sizeof(current->comm), comm));

    return strstr(comm, Name) ? ljvSTATUS_TRUE : ljvSTATUS_FALSE;
}

#if ljdLINUX_SYNC_FILE
#ifndef CONFIG_SYNC_FILE
ljeSTATUS
ljkOS_CreateSyncTimeline(ljkOS Os, ljeCORE Core, ljtHANDLE *Timeline)
{
    ljeSTATUS status = ljvSTATUS_OK;
    struct ljm_sync_timeline *timeline;
    char name[32];

    snprintf(name, 32, "ljcore-%u", (unsigned int)Core);

    /* Create viv sync timeline. */
    timeline = ljm_sync_timeline_create(name, Os);

    if (timeline)
        *Timeline = (ljtHANDLE)timeline;
    else
        /* Out of memory. */
        status = ljvSTATUS_OUT_OF_MEMORY;

    return status;
}

ljeSTATUS
ljkOS_DestroySyncTimeline(ljkOS Os, ljtHANDLE Timeline)
{
    struct ljm_sync_timeline *timeline;

    ljmkASSERT(Timeline != ljvNULL);

    /* Destroy timeline. */
    timeline = (struct ljm_sync_timeline *)Timeline;
    sync_timeline_destroy(&timeline->obj);

    return ljvSTATUS_OK;
}

ljeSTATUS
ljkOS_CreateNativeFence(ljkOS Os, ljtHANDLE Timeline,
                        ljtSIGNAL Signal, ljtINT *FenceFD)
{
    int fd = -1;
    struct ljm_sync_timeline *timeline;
    struct sync_pt *pt = ljvNULL;
    struct sync_fence *fence;
    char name[32];
    ljsSIGNAL_PTR signal;
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkHEADER_ARG("Os=%p Timeline=%p Signal=%p", Os, Timeline, Signal);

    ljmkONERROR(_QueryIntegerId(&Os->signalDB,
                                (ljtUINT32)(ljtUINTPTR_T)Signal,
                                (ljtPOINTER)&signal));

    /* Cast timeline. */
    timeline = (struct ljm_sync_timeline *)Timeline;

    fd = get_unused_fd_flags(O_CLOEXEC);

    if (fd < 0)
        /* Out of resources. */
        ljmkONERROR(ljvSTATUS_OUT_OF_RESOURCES);

    /* Create ljm_sync_pt. */
    pt = ljm_sync_pt_create(timeline, Signal);

    if (pt == ljvNULL)
        ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);

    /* Reference sync_timeline. */
    signal->timeline = &timeline->obj;

    /* Build fence name. */
    snprintf(name, 32, "%.16s-signal_%lu", current->comm, (unsigned long)Signal);

    /* Create sync_fence. */
    fence = sync_fence_create(name, pt);

    if (!fence)
        ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);

    /* Install fence to fd. */
    sync_fence_install(fence, fd);

    *FenceFD = fd;
    return ljvSTATUS_OK;

OnError:
    if (ljmIS_ERROR(status)) {
        /* Error roll back. */
        if (pt)
            sync_pt_free(pt);

        if (fd > 0)
            put_unused_fd(fd);
    }

    ljmkFOOTER_ARG("*FenceFD=%d", fd);
    return status;
}

static void
_NativeFenceSignaled(struct sync_fence *fence, struct sync_fence_waiter *waiter)
{
    kfree(waiter);
    sync_fence_put(fence);
}

ljeSTATUS
ljkOS_WaitNativeFence(ljkOS Os, ljtHANDLE Timeline,
                      ljtINT FenceFD, ljtUINT32 Timeout)
{
    struct sync_timeline *timeline;
    struct sync_fence *fence;
    ljeSTATUS status = ljvSTATUS_OK;

    ljmkHEADER_ARG("Os=%p Timeline=%p FenceFD=%d Timeout=%u",
                   Os, Timeline, FenceFD, Timeout);

    /* Get shortcut. */
    timeline = (struct sync_timeline *)Timeline;

    /* Get sync fence. */
    fence = sync_fence_fdget(FenceFD);

    if (!fence)
        ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);

    if (sync_fence_wait(fence, 0) == 0) {
        /* Already signaled. */
        sync_fence_put(fence);

        goto OnError;
    } else {
        ljtBOOL wait = ljvFALSE;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 17, 0)
        int i;

        for (i = 0; i < fence->num_fences; i++) {
            struct fence *f  = fence->cbs[i].sync_pt;
            struct sync_pt *pt = container_of(f, struct sync_pt, base);

            /* Do not need to wait on same timeline. */
            if ((sync_pt_parent(pt) != timeline) && !fence_is_signaled(f)) {
                wait = ljvTRUE;
                break;
            }
        }
#  else
        struct list_head *pos;

        list_for_each(pos, &fence->pt_list_head) {
            struct sync_pt *pt = container_of(pos, struct sync_pt, pt_list);

            /* Do not need to wait on same timeline. */
            if (pt->parent != timeline) {
                wait = ljvTRUE;
                break;
            }
        }
#  endif /* LINUX_VERSION_CODE >= KERNEL_VERSION(3, 17, 0) */

        if (wait) {
            int err;
            long timeout = (Timeout == ljvINFINITE) ? -1 : (long)Timeout;

            err = sync_fence_wait(fence, timeout);

            /* Put the fence. */
            sync_fence_put(fence);

            switch (err) {
            case 0:
                break;
            case -ETIME:
                status = ljvSTATUS_TIMEOUT;
                break;
            default:
                ljmkONERROR(ljvSTATUS_GENERIC_IO);
                break;
            }
        } else {
            int err;
            struct sync_fence_waiter *waiter;

            waiter = kmalloc(sizeof(*waiter), ljdNOWARN | GFP_KERNEL);

            if (!waiter) {
                sync_fence_put(fence);
                ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);
            }

            /* Schedule a waiter callback. */
            sync_fence_waiter_init(waiter, _NativeFenceSignaled);
            err = sync_fence_wait_async(fence, waiter);

            switch (err) {
            case 0:
                /* Put fence in callback function. */
                break;
            case 1:
                /* already signaled. */
                sync_fence_put(fence);
                break;
            default:
                sync_fence_put(fence);
                ljmkONERROR(ljvSTATUS_GENERIC_IO);
                break;
            }
        }
    }

OnError:
    ljmkFOOTER();
    return status;
}

# else /* else CONFIG_SYNC_FILE */

ljeSTATUS
ljkOS_CreateSyncTimeline(ljkOS Os, ljeCORE Core, ljtHANDLE *Timeline)
{
    struct ljm_sync_timeline *timeline;

    char name[32];

    snprintf(name, 32, "ljcore-%u", (unsigned int)Core);
    timeline = ljm_sync_timeline_create(name, Os);

    if (timeline == ljvNULL)
        /* Out of memory. */
        return ljvSTATUS_OUT_OF_MEMORY;

    *Timeline = (ljtHANDLE)timeline;
    return ljvSTATUS_OK;
}

ljeSTATUS
ljkOS_DestroySyncTimeline(ljkOS Os, ljtHANDLE Timeline)
{
    struct ljm_sync_timeline *timeline;

    /* Destroy timeline. */
    timeline = (struct ljm_sync_timeline *)Timeline;
    ljm_sync_timeline_destroy(timeline);

    return ljvSTATUS_OK;
}

ljeSTATUS
ljkOS_CreateNativeFence(ljkOS Os, ljtHANDLE Timeline,
                        ljtSIGNAL Signal, ljtINT *FenceFD)
{
    struct dma_fence *fence = NULL;
    struct sync_file *sync = NULL;
    int fd = -1;
    struct ljm_sync_timeline *timeline;
    ljsSIGNAL_PTR signal = ljvNULL;
    ljeSTATUS status = ljvSTATUS_OK;

    /* Create fence. */
    timeline = (struct ljm_sync_timeline *)Timeline;

    ljmkONERROR(_QueryIntegerId(&Os->signalDB,
                                (ljtUINT32)(ljtUINTPTR_T)Signal,
                                (ljtPOINTER)&signal));

    fence = ljm_fence_create(timeline, signal);

    if (!fence)
        ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);

    /* Create sync_file. */
    sync = sync_file_create(fence);

    if (!sync)
        ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);

    /* Get a unused fd. */
    fd = get_unused_fd_flags(O_CLOEXEC);

    if (fd < 0)
        ljmkONERROR(ljvSTATUS_OUT_OF_RESOURCES);

    fd_install(fd, sync->file);

    *FenceFD = fd;
    return ljvSTATUS_OK;

OnError:
    if (sync)
        fput(sync->file);

#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 9, 68)
    if (fence)
        dma_fence_put(fence);
#  endif

    if (fd > 0)
        put_unused_fd(fd);

    *FenceFD = -1;
    return status;
}

#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 9, 0)
/*
 * sync_file_fdget() - get a sync_file from an fd
 * @fd:     fd referencing a fence
 *
 * Ensures @fd references a valid sync_file, increments the refcount of the
 * backing file. Returns the sync_file or NULL in case of error.
 */
static struct sync_file *sync_file_fdget(int fd)
{
    struct file *file = fget(fd);

    if (!file)
        return NULL;

    return file->private_data;
}

ljeSTATUS
ljkOS_WaitNativeFence(ljkOS Os, ljtHANDLE Timeline,
                      ljtINT FenceFD, ljtUINT32 Timeout)
{
    struct ljm_sync_timeline *timeline;
    ljeSTATUS status = ljvSTATUS_OK;
    unsigned int i;
    unsigned long timeout;
    unsigned int numFences;
    struct sync_file *sync_file;

    timeline = (struct ljm_sync_timeline *)Timeline;

    sync_file = sync_file_fdget(FenceFD);

    if (!sync_file)
        ljmkONERROR(ljvSTATUS_GENERIC_IO);

    numFences = sync_file->num_fences;

    timeout = msecs_to_jiffies(Timeout);

    for (i = 0; i < numFences; i++) {
        struct fence *f = sync_file->cbs[i].fence;

        fence_get(f);

        if (f->context != timeline->context && !fence_is_signaled(f)) {
            signed long ret;

            ret = fence_wait_timeout(f, 1, timeout);

            if (ret == -ERESTARTSYS) {
                fence_put(f);
                ljmkONERROR(ljvSTATUS_INTERRUPTED);
            } else if (ret <= 0) {
                fence_put(f);
                ljmkONERROR(ljvSTATUS_TIMEOUT);
            } else {
                /* wait success. */
                timeout -= ret;
            }
        }

        fence_put(f);
    }

    return ljvSTATUS_OK;

OnError:
    return status;
}

#  else /* !LINUX_VERSION_CODE < KERNEL_VERSION(4, 9, 0) */

ljeSTATUS
ljkOS_WaitNativeFence(ljkOS Os, ljtHANDLE Timeline,
                      ljtINT FenceFD, ljtUINT32 Timeout)
{
    struct ljm_sync_timeline *timeline;
    ljeSTATUS status = ljvSTATUS_OK;
    unsigned int i;
    unsigned long timeout;
    unsigned int numFences;
    struct dma_fence *fence;
    struct dma_fence **fences;

    timeline = (struct ljm_sync_timeline *)Timeline;

    fence = sync_file_get_fence(FenceFD);

    if (!fence)
        ljmkONERROR(ljvSTATUS_GENERIC_IO);

    if (dma_fence_is_array(fence)) {
        struct dma_fence_array *array = to_dma_fence_array(fence);

        fences = array->fences;
        numFences = array->num_fences;
    } else {
        fences = &fence;
        numFences = 1;
    }

    timeout = msecs_to_jiffies(Timeout);

    for (i = 0; i < numFences; i++) {
        struct dma_fence *f = fences[i];

        if (!dma_fence_is_signaled(fence)) {
            signed long ret;

            ret = dma_fence_wait_timeout(f, 1, timeout);

            if (ret == -ERESTARTSYS) {
                dma_fence_put(fence);
                ljmkONERROR(ljvSTATUS_INTERRUPTED);
            } else if (ret <= 0) {
                dma_fence_put(fence);
                ljmkONERROR(ljvSTATUS_TIMEOUT);
            } else {
                /* wait success. */
                timeout -= ret;
            }
        }
    }

    dma_fence_put(fence);

    return ljvSTATUS_OK;

OnError:
    return status;
}

#  endif /* LINUX_VERSION_CODE < KERNEL_VERSION(4,9,0) */
# endif /* ifndef CONFIG_SYNC_FILE */
#endif /* ljdLINUX_SYNC_FILE */

#if ljdSECURITY
ljeSTATUS
ljkOS_AllocatePageArray(ljkOS Os,
                        ljkKERNEL Kernel,
                        ljtPHYS_ADDR Physical,
                        ljtSIZE_T PageCount,
                        ljtPOINTER *PageArrayLogical,
                        ljtPHYS_ADDR *PageArrayPhysical)
{
    ljeSTATUS status = ljvSTATUS_OK;
    PLINUX_MDL mdl;
    ljtUINT32 *table;
    ljtUINT32 offset;
    ljtSIZE_T bytes;
    ljkALLOCATOR allocator;

    ljmkHEADER_ARG("Os=%p Physical=%p PageCount=%u",
                   Os, Physical, PageCount);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Physical != ljvNULL);
    ljmkVERIFY_ARGUMENT(PageCount > 0);

    bytes = PageCount * ljmSIZEOF(ljtUINT32);
    ljmkONERROR(ljkOS_AllocateNonPagedMemory(Os, Kernel, ljvFALSE,
                                             ljvALLOC_FLAG_CONTIGUOUS, &bytes,
                                             PageArrayPhysical, PageArrayLogical));

    table = *PageArrayLogical;

    /* Convert pointer to MDL. */
    mdl = (PLINUX_MDL)Physical;

    allocator = mdl->allocator;

    /* Get all the physical addresses and store them in the page table. */

    offset = 0;
    PageCount = PageCount / (PAGE_SIZE / 4096);

    /* Try to get the user pages so DMA can happen. */
    while (PageCount-- > 0) {
        unsigned long phys = ~0;

        ljtPHYS_ADDR_T phys_addr;

        allocator->ops->Physical(allocator, mdl, offset * PAGE_SIZE, &phys_addr);

        phys = (unsigned long)phys_addr;

        table[offset] = phys & PAGE_MASK;

        offset += 1;
    }

OnError:

    /* Return the status. */
    ljmkFOOTER();
    return status;
}
#endif

ljeSTATUS
ljkOS_CPUPhysicalToGPUPhysical(ljkOS Os, ljtPHYS_ADDR_T CPUPhysical, ljtPHYS_ADDR_T *GPUPhysical)
{
    ljsPLATFORM *platform;

    ljmkHEADER_ARG("CPUPhysical=%llx", CPUPhysical);

    platform = Os->device->platform;

    if (platform && platform->ops->getGPUPhysical)
        ljmkVERIFY_OK(platform->ops->getGPUPhysical(platform, CPUPhysical, GPUPhysical));
    else
        *GPUPhysical = CPUPhysical;

    ljmkFOOTER_ARG("GPUPhysical=0x%llx", ljmOPT_VALUE(GPUPhysical));
    return ljvSTATUS_OK;
}

ljeSTATUS
ljkOS_GPUPhysicalToCPUPhysical(ljkOS Os, ljtPHYS_ADDR_T GPUPhysical, ljtPHYS_ADDR_T *CPUPhysical)
{
    ljsPLATFORM *platform;

    ljmkHEADER_ARG("Os=%p GPUPhysical=0x%llx", Os, GPUPhysical);

    platform = Os->device->platform;

    if (platform && platform->ops->getCPUPhysical)
        ljmkVERIFY_OK(platform->ops->getCPUPhysical(platform, GPUPhysical, CPUPhysical));
    else
        *CPUPhysical = GPUPhysical;

    ljmkFOOTER_ARG("CPUPhysical=0x%llx", ljmOPT_VALUE(CPUPhysical));
    return ljvSTATUS_OK;
}

static int
fd_release(struct inode *inode, struct file *file)
{
    ljsFDPRIVATE_PTR private = (ljsFDPRIVATE_PTR)file->private_data;

    return (private && private->release) ? private->release(private) : 0;
}

static const struct file_operations fd_fops = {
    .release = fd_release,
};

ljeSTATUS
ljkOS_GetFd(ljtSTRING Name, ljsFDPRIVATE_PTR Private, ljtINT *Fd)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 27)
    *Fd = anon_inode_getfd(Name, &fd_fops, Private, O_RDWR);

    if (*Fd < 0)
        return ljvSTATUS_OUT_OF_RESOURCES;

    return ljvSTATUS_OK;
#else
    return ljvSTATUS_NOT_SUPPORTED;
#endif
}

ljeSTATUS
ljkOS_QueryOption(ljkOS Os, ljtCONST_STRING Option, ljtUINT64 *Value)
{
    ljkGALDEVICE device = Os->device;
    ljeSTATUS status = ljvSTATUS_OK;

    if (!strcmp(Option, "physBase")) {
        *Value = (ljtUINT64)device->args.baseAddress;
    } else if (!strcmp(Option, "physSize")) {
        *Value = (ljtUINT64)device->args.physSize;
    } else if (!strcmp(Option, "mmu")) {
#if ljdSECURITY
        *Value = 0;
#else
        *Value = (ljtUINT64)device->args.enableMmu;
#endif
    } else if (!strcmp(Option, "recovery")) {
        *Value = (ljtUINT64)device->args.recovery;
    } else if (!strcmp(Option, "stuckDump")) {
        *Value = (ljtUINT64)device->args.stuckDump;
    } else if (!strcmp(Option, "powerManagement")) {
        *Value = (ljtUINT64)device->args.powerManagement;
    } else if (!strcmp(Option, "TA")) {
        *Value = 0;
    } else if (!strcmp(Option, "userClusterMasks")) {
        if (ljmSIZEOF(device->args.userClusterMasks) >=
            ljdMAX_MAJOR_CORE_COUNT * ljmSIZEOF(ljtUINT32))
            memcpy(Value, device->args.userClusterMasks, ljdMAX_MAJOR_CORE_COUNT * ljmSIZEOF(ljtUINT32));
        else
            return ljvSTATUS_NOT_SUPPORTED;
    } else if (!strcmp(Option, "smallBatch")) {
        *Value = device->args.smallBatch;
    } else if (!strcmp(Option, "sRAMBases")) {
        if (ljmSIZEOF(device->args.sRAMBases) >=
            ljvSRAM_INTER_COUNT * ljvCORE_COUNT * ljmSIZEOF(ljtUINT64))
            memcpy(Value, device->args.sRAMBases, ljvSRAM_INTER_COUNT * ljvCORE_COUNT * ljmSIZEOF(ljtUINT64));
        else
            return ljvSTATUS_NOT_SUPPORTED;
    } else if (!strcmp(Option, "sRAMSizes")) {
        if (ljmSIZEOF(device->args.sRAMSizes) >=
            ljvSRAM_INTER_COUNT * ljvCORE_COUNT * ljmSIZEOF(ljtUINT32))
            memcpy(Value, device->args.sRAMSizes, ljvSRAM_INTER_COUNT * ljvCORE_COUNT * ljmSIZEOF(ljtUINT32));
        else
            return ljvSTATUS_NOT_SUPPORTED;
    } else if (!strcmp(Option, "sRAMRequested")) {
        *Value = (ljtUINT64)device->args.sRAMRequested;
    } else if (!strcmp(Option, "sRAMLoopMode")) {
        *Value = (ljtUINT64)device->args.sRAMLoopMode;
    } else if (!strcmp(Option, "platformFlagBits")) {
        *Value = (ljtUINT64)device->platform->flagBits;
    } else if (!strcmp(Option, "flatMapping")) {
        *Value = (ljtUINT64)device->args.flatMapping;
    } else if (!strcmp(Option, "mmuPageTablePool")) {
        *Value = (ljtUINT64)device->args.mmuPageTablePool;
    } else if (!strcmp(Option, "mmuCmdPool")) {
        *Value = (ljtUINT64)device->args.mmuCmdPool;
    } else if (!strcmp(Option, "sharedPageTable")) {
        *Value = (ljtUINT64)device->args.sharedPageTable;
    } else if (!strcmp(Option, "mmuDynamicMap")) {
        *Value = (ljtUINT64)device->args.mmuDynamicMap;
    } else if (!strcmp(Option, "allMapInOne")) {
        *Value = (ljtUINT64)device->args.allMapInOne;
    } else if (!strcmp(Option, "isrPoll")) {
        *Value = (ljtUINT64)device->args.isrPoll;
    } else if (!strcmp(Option, "registerAPB")) {
        *Value = (ljtUINT64)device->args.registerAPB;
    } else if (!strcmp(Option, "enableNN")) {
        *Value = (ljtUINT64)device->args.enableNN;
    } else if (!strcmp(Option, "vGPUType")) {
        *Value = (ljtUINT64)device->args.vGPUType;
    } else if (!strcmp(Option, "vGPUId")) {
        *Value = (ljtUINT64)device->args.vGPUId;
    } else if (!strcmp(Option, "softReset")) {
        *Value = (ljtUINT64)device->args.softReset;
    } else if (!strcmp(Option, "devCoreCounts")) {
        if (ljmSIZEOF(device->args.devCoreCounts) >= ljmSIZEOF(ljtUINT32) * ljdDEVICE_COUNT)
            memcpy(Value, device->args.devCoreCounts, ljmSIZEOF(ljtUINT32) * ljdDEVICE_COUNT);
        else
            return ljvSTATUS_NOT_SUPPORTED;
    } else {
        status = ljvSTATUS_NOT_SUPPORTED;
    }

    return status;
}

ljeSTATUS
ljkOS_QueryKernel(ljkKERNEL Kernel, ljtINT index, ljkKERNEL *KernelOut)
{
    if (Kernel && KernelOut) {
        ljkDEVICE device = Kernel->device;

        *KernelOut = device->kernels[index];
    }

    return ljvSTATUS_OK;
}

ljeSTATUS
ljkOS_QueryCPUFrequency(ljkOS Os, ljtUINT32 CPUId, ljtUINT32 *Frequency)
{
    /* The frequency will be KHz */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 16)
    if (Frequency)
        *Frequency = cpufreq_quick_get(0);
#endif
    return ljvSTATUS_OK;
}

ljeSTATUS
ljkOS_MemoryGetSGT(ljkOS Os, ljtPHYS_ADDR Physical,
                   ljtSIZE_T Offset, ljtSIZE_T Bytes, ljtPOINTER *SGT)
{
    PLINUX_MDL mdl;
    ljkALLOCATOR allocator;
    ljeSTATUS status = ljvSTATUS_OK;

    if (!Physical)
        ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);

    mdl = (PLINUX_MDL)Physical;
    allocator = mdl->allocator;

    if (!allocator->ops->GetSGT)
        ljmkONERROR(ljvSTATUS_NOT_SUPPORTED);

    if (Bytes > 0)
        ljmkONERROR(allocator->ops->GetSGT(allocator, mdl, Offset, Bytes, SGT));

OnError:
    return status;
}

ljeSTATUS
ljkOS_MemoryMmap(ljkOS Os, ljtPHYS_ADDR Physical,
                 ljtSIZE_T skipPages, ljtSIZE_T numPages, ljtPOINTER Vma)
{
    PLINUX_MDL mdl;
    PLINUX_MDL_MAP mdlMap;
    ljkALLOCATOR allocator;
    ljeSTATUS status = ljvSTATUS_OK;
    ljtBOOL cacheable = ljvFALSE;
    ljtBOOL acquiredMutex = ljvFALSE;

    if (!Physical)
        ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);

    mdl = (PLINUX_MDL)Physical;
    allocator = mdl->allocator;

    if (!allocator->ops->Mmap)
        ljmkONERROR(ljvSTATUS_NOT_SUPPORTED);

    ljmkONERROR(ljkOS_AcquireMutex(Os, &mdl->mapsMutex, ljvINFINITE));
    acquiredMutex = ljvTRUE;

    mdlMap = FindMdlMap(mdl, _GetProcessID());
    if (mdlMap)
        cacheable = mdlMap->cacheable;

    ljmkONERROR(ljkOS_ReleaseMutex(Os, &mdl->mapsMutex));
    acquiredMutex = ljvFALSE;

    ljmkONERROR(allocator->ops->Mmap(allocator, mdl, cacheable, skipPages, numPages, Vma));

OnError:
    if (acquiredMutex)
        ljkOS_ReleaseMutex(Os, &mdl->mapsMutex);
    return status;
}

/******************************************************************************
 **
 **  ljkOS_WrapMemory
 **
 **  Import a number of pages allocated by other allocator.
 **
 **  INPUT:
 **
 **      ljkOS Os
 **          Pointer to an ljkOS object.
 **
 **      ljtUINT32 Flag
 **          Memory type.
 **
 **  OUTPUT:
 **
 **      ljtSIZE_T *Bytes
 **          Pointer to a variable that hold the number of bytes allocated.
 **
 **      ljtPHYS_ADDR *Physical
 **          Pointer to a variable that will hold the physical address of the
 **          allocation.
 */
ljeSTATUS
ljkOS_WrapMemory(ljkOS Os, ljkKERNEL Kernel, ljsUSER_MEMORY_DESC_PTR Desc,
                 ljtSIZE_T *Bytes, ljtPHYS_ADDR *Physical, ljtBOOL *Contiguous, ljtSIZE_T *PageCountCpu)
{
    PLINUX_MDL mdl = ljvNULL;
    ljeSTATUS status = ljvSTATUS_OUT_OF_MEMORY;
    ljkALLOCATOR allocator;
    ljsATTACH_DESC desc;
    ljtSIZE_T bytes = 0;
    ljtBOOL acquiredMutex = ljvFALSE;

    ljmkHEADER_ARG("Os=%p ", Os);

    /* Verify the arguments. */
    ljmkVERIFY_OBJECT(Os, ljvOBJ_OS);
    ljmkVERIFY_ARGUMENT(Physical != ljvNULL);
    ljmkVERIFY_ARGUMENT(Desc != ljvNULL);

    mdl = _CreateMdl(Os, Kernel);
    if (mdl == ljvNULL)
        ljmkONERROR(ljvSTATUS_OUT_OF_MEMORY);

    mdl->wrapFromPhysical = ljvFALSE;
    mdl->wrapFromLogical = ljvFALSE;
    mdl->fromUser = ljvTRUE;
    mdl->wrapped = ljvTRUE;

    if (Desc->flag & ljvALLOC_FLAG_DMABUF) {
#if defined(CONFIG_DMA_SHARED_BUFFER)
        struct dma_buf *dmabuf;
#endif
        if (IS_ERR(ljmUINT64_TO_PTR(Desc->dmabuf))) {
            /*
             * Won't enter here currently, the caller confirms the
             * dmabuf is valid.
             */
            ljmkPRINT("Wrap memory: invalid dmabuf.\n");
            ljmkONERROR(ljvSTATUS_INVALID_ARGUMENT);
        }

        desc.dmaBuf.dmabuf = ljmUINT64_TO_PTR(Desc->dmabuf);

#if defined(CONFIG_DMA_SHARED_BUFFER)
        dmabuf = (struct dma_buf *)desc.dmaBuf.dmabuf;

        bytes = dmabuf->size;
#endif
    } else if (Desc->flag & ljvALLOC_FLAG_USERMEMORY) {
        desc.userMem.memory = ljmUINT64_TO_PTR(Desc->logical);
        desc.userMem.physical = Desc->physical;
        desc.userMem.size = Desc->size;
        bytes = Desc->size;

        if (Desc->physical == ljvINVALID_PHYSICAL_ADDRESS)
            mdl->wrapFromLogical = ljvTRUE;
        else
            mdl->wrapFromPhysical = ljvTRUE;
    } else {
        ljmkONERROR(ljvSTATUS_NOT_SUPPORTED);
    }

    /* Walk all allocators. */
    list_for_each_entry(allocator, &Os->allocatorList, link) {
        ljmkTRACE_ZONE(ljvLEVEL_INFO, ljvZONE_OS,
                       "%s(%d) Flag = %x allocator->capability = %x",
                       __func__, __LINE__, Desc->flag, allocator->capability);

        if ((Desc->flag & allocator->capability) != Desc->flag) {
            status = ljvSTATUS_NOT_SUPPORTED;
            continue;
        }

        status = allocator->ops->Attach(allocator, &desc, mdl);

        if (ljmIS_SUCCESS(status)) {
            mdl->allocator = allocator;
            break;
        }
    }

    /* Check status. */
    ljmkONERROR(status);

    mdl->dmaHandle = 0;
    mdl->addr = 0;

    mdl->bytes = bytes ? bytes : mdl->numPages * PAGE_SIZE;
    *Bytes = mdl->bytes;

    /* Return physical address. */
    *Physical = (ljtPHYS_ADDR)mdl;

    *Contiguous = mdl->contiguous;

    if (PageCountCpu)
        *PageCountCpu = mdl->numPages;

    /*
     * Add this to a global list.
     * Will be used by get physical address
     * and mapuser pointer functions.
     */
    ljmkONERROR(ljkOS_AcquireMutex(Os, &Os->mdlMutex, ljvINFINITE));
    acquiredMutex = ljvTRUE;

    list_add_tail(&mdl->link, &Os->mdlHead);

    ljmkONERROR(ljkOS_ReleaseMutex(Os, &Os->mdlMutex));
    acquiredMutex = ljvFALSE;

    /* Success. */
    status = ljvSTATUS_OK;

OnError:
    if (acquiredMutex)
        ljkOS_ReleaseMutex(Os, &Os->mdlMutex);

    if (ljmIS_ERROR(status) && mdl)
        /* Free the memory. */
        _DestroyMdl(mdl);

    /* Return the status. */
    ljmkFOOTER_ARG("*Physical=%p", *Physical);
    return status;
}

ljeSTATUS
ljkOS_GetPolicyID(ljkOS Os, ljeVIDMEM_TYPE Type,
                  ljtUINT32_PTR PolicyID, ljtUINT32_PTR AXIConfig)
{
    ljsPLATFORM *platform = Os->device->platform;
    ljeSTATUS status = (platform && platform->ops->getPolicyID) ?
                        platform->ops->getPolicyID(platform, Type, PolicyID, AXIConfig) :
                        ljvSTATUS_NOT_SUPPORTED;

    return status;
}

#if ljdENABLE_MP_SWITCH
ljeSTATUS
ljkOS_SwitchCoreCount(ljkOS Os, ljtUINT32 *Count)
{
    ljeSTATUS status = ljvSTATUS_OK;
    ljsPLATFORM *platform = Os->device->platform;

    ljmkHEADER_ARG("Os=%p", Os);

    status = (platform && platform->ops->switchCoreCount) ?
             platform->ops->switchCoreCount(platform, Count) :
             ljvSTATUS_OK;

    ljmkFOOTER_ARG("*Count=%d", *Count);
    return status;
}
#endif

ljeSTATUS
ljkOS_TraceGpuMemory(ljkOS Os, ljtINT32 ProcessID, ljtINT64 Delta)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 7, 0)
#if defined(CONFIG_TRACE_GPU_MEM)
    traceMem *traceList = ljvNULL;
    ljtBOOL addNodeFlag = ljvTRUE;

    /* if first node is null, create it. */
    if (!memTraceList) {
        ljtPOINTER pointer = ljvNULL;

        ljkOS_Allocate(Os, sizeof(traceMem), &pointer);
        ljkOS_ZeroMemory(pointer, sizeof(traceMem));
        memTraceList = pointer;
        memTraceList->pid = 0;
        atomic64_set(&memTraceList->total, 0);
        memTraceList->next = ljvNULL;
    }

    traceList = memTraceList;

    /* traverse all the node */
    while (ljvTRUE) {
        /* hit the matched node */
        if (ProcessID == traceList->pid) {
            if (Delta > 0)
                atomic64_add(Delta, &traceList->total);
            else
                atomic64_sub(-Delta, &traceList->total);
            addNodeFlag = ljvFALSE;
            break;
        } else if (traceList->pid == 0) {
        /* hit the first node update the global usage */
            if (Delta > 0)
                atomic64_add(Delta, &memTraceList->total);
            else
                atomic64_sub(-Delta, &memTraceList->total);
        }
        if (!traceList->next)
            break;
        else
            traceList = traceList->next;
    }

    /* if there isn't matched node, create one */
    if (!traceList->next && addNodeFlag) {
        ljtPOINTER pointer = ljvNULL;
        traceMem *newNode = ljvNULL;

        ljkOS_Allocate(Os, sizeof(traceMem), &pointer);
        ljkOS_ZeroMemory(newNode, sizeof(traceMem));

        newNode = pointer;
        newNode->pid = ProcessID;
        atomic64_set(&newNode->total, Delta);
        newNode->next = ljvNULL;
        traceList->next = newNode;
        traceList = traceList->next;
    }

    if (trace_gpu_mem_total_enabled()) {
        trace_gpu_mem_total(0, 0, (ljtUINT64)atomic64_read(&memTraceList->total));
        trace_gpu_mem_total(0, ProcessID, (ljtUINT64)atomic64_read(&traceList->total));
    }

    /* clean useless node */
    while (ljvTRUE) {
        traceMem *t_node = traceList->next;

        if (t_node) {
            if (atomic64_read(&t_node->total) <= 0) {
                traceList->next = t_node->next;
                ljkOS_Free(Os, t_node);
                t_node = NULL;
            } else {
                traceList = traceList->next;
            }
        } else {
            break;
        }
    }

    if (memTraceFlag && trace_gpu_mem_total_enabled()) {
       traceMem *t_node = memTraceList;

        while (ljvTRUE) {
           if (t_node)
               trace_gpu_mem_total(0, t_node->pid, (ljtUINT64)atomic64_read(&t_node->total));
           else
               break;
           t_node = t_node->next;
       }
       memTraceFlag = 0;
    }

#endif
#endif
    return ljvSTATUS_OK;
}

void
ljkOS_NodeIdAssign(
    ljkOS Os,
    ljuVIDMEM_NODE_PTR Node)
{
    ljkVIDMEM_BLOCK vidMemBlock = Node->VirtualChunk.parent;
    ljtINT userID = 0x100;

    if (vidMemBlock && vidMemBlock->object.type == ljvOBJ_VIDMEM_BLOCK) {
        Node->VirtualChunk.id = (ljtINT32)atomic_inc_return(&Os->nodeID);
        if (Node->VirtualChunk.id < 1) {
            Node->VirtualChunk.id = userID;
            atomic_set(&Os->nodeID, userID);
        }
    } else {
        Node->VidMem.id = (ljtINT32)atomic_inc_return(&Os->nodeID);
        if (Node->VidMem.id < 1) {
            Node->VidMem.id = userID;
            atomic_set(&Os->nodeID, userID);
        }
    }
}

#if ljdENABLE_CLEAR_FENCE
ljeSTATUS
ljkOS_ClearAllFence(ljkDEVICE Device)
{
    ljsLISTHEAD_PTR list = &Device->fenceList;
    ljsLISTHEAD_PTR nodeHead, nodeTemp;
    ljkVIDMEM_NODE nodeObject = ljvNULL;
    ljtSIZE_T offset;
    ljsUSER_FENCE_INFO_PTR fence_info = ljvNULL;
    ljtUINT32 i, found;
    ljtPOINTER fencePointer;
    ljeSTATUS status;
    ljtBOOL acquired = ljvFALSE;
    ljkKERNEL kernel = ljvNULL;

    ljmkHEADER_ARG("Device=%p ", Device);
    ljmkONERROR(ljkOS_AcquireMutex(Device->os,
                            Device->fenceListMutex,
                            ljvINFINITE));
    acquired = ljvTRUE;

    ljmkLIST_FOR_EACH_SAFE(nodeHead, nodeTemp, list) {
        fence_info = ljmCONTAINEROF(nodeHead, struct _ljsUSER_FENCE_INFO, fenceLink);
        if (fence_info) {
            found = 0;
            for (i = 0; i < ljvCORE_COUNT; i++) {
                status = ljvSTATUS_INVALID_ARGUMENT;
                kernel = Device->kernels[i];
                if (kernel)
                    status = ljkVIDMEM_NODE_Find(kernel, fence_info->addr, &nodeObject, &offset);

                if (ljmIS_SUCCESS(status) && nodeObject->type == ljvVIDMEM_TYPE_FENCE) {
                    ljmkVERIFY_OK(ljkVIDMEM_NODE_LockCPU(kernel, nodeObject, ljvFALSE, ljvFALSE, &fencePointer));
                    fencePointer += offset;
                    *(ljtUINT64 *)fencePointer = fence_info->fenceValue;
                    ljmkVERIFY_OK(ljkVIDMEM_NODE_UnlockCPU(kernel, nodeObject, 0, ljvFALSE, ljvFALSE));
                    found = 1;
                    break;
                }
            }
            if (!found) {
                ljmkPRINT("Failed to write fence value:%llx\n", fence_info->fenceValue);
                ljmkPRINT("Failed to find the Fence vidmem node!\n");
            }
        } else {
            ljmkPRINT("Failed to find the Fence record!\n");
        }
    }

OnError:
    if (acquired)
        ljmkVERIFY_OK(ljkOS_ReleaseMutex(Device->os,
                        Device->fenceListMutex));

    ljmkFOOTER();
    return status;
}

ljtUINT64
ljkOS_AllocFenceRecordId(ljkDEVICE Device, ljsUSER_FENCE_INFO_PTR fence_info)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 9, 0)
    ljtINT result;
    unsigned long flags = 0;

    spin_lock_irqsave(&Device->fenceIdrLock, flags);
    result = idr_alloc((struct idr *)Device->fenceIdr, (ljtPOINTER)fence_info, 0, 0, GFP_ATOMIC);
    spin_unlock_irqrestore(&Device->fenceIdrLock, flags);

    if (result < 0) {
        pr_err("Allocate fence record Id failed, result:%d\n", result);
        ljmkBUG_ON(1);
    }
    return (ljtUINT64)result;
#else
    ljtINT result;
    ljtUINT32 Id;

    spin_lock_irqsave(&Device->fenceIdrLock, flags);
    result = idr_get_new_above((struct idr *)Device->fenceIdr, (ljtPOINTER)fence_info, 0, &Id);
    spin_unlock_irqrestore(&Device->fenceIdrLock, flags);

    if (result < 0) {
        pr_err("Allocate fence record Id failed, result:%d\n", result);
        ljmkBUG_ON(1);
    }
    return (ljtUINT64)Id;
#endif
}

ljsUSER_FENCE_INFO_PTR
ljkOS_ReleaseFenceRecordId(ljkDEVICE Device, ljtUINT64 recordId)
{
    ljtINT Id = (ljtINT)recordId;
    ljtPOINTER fence_info = ljvNULL;
    unsigned long flags = 0;

    spin_lock_irqsave(&Device->fenceIdrLock, flags);
    fence_info = idr_find((struct idr *)Device->fenceIdr, Id);
    spin_unlock_irqrestore(&Device->fenceIdrLock, flags);

    if (!fence_info) {
        pr_err("Find fence record Id failed, Id:%d\n", Id);

        return (ljsUSER_FENCE_INFO_PTR)fence_info;
    }

    spin_lock_irqsave(&Device->fenceIdrLock, flags);
    idr_remove((struct idr *)Device->fenceIdr, Id);
    spin_unlock_irqrestore(&Device->fenceIdrLock, flags);

    return (ljsUSER_FENCE_INFO_PTR)fence_info;
}

void
ljkOS_PreLoadFenceRecId(ljkDEVICE Device)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 9, 0)
    idr_preload(GFP_KERNEL | ljdNOWARN);
#else
    idr_pre_get((struct idr *)Device->fenceIdr, GFP_KERNEL | ljdNOWARN);
#endif
}

void
ljkOS_PreLoadEndFenceRecId(ljkDEVICE Device)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(3, 9, 0)
    idr_preload_end();
#endif
}

#endif
